/* tapeutil for rekey use */
/*****************************************************************************/
/**                                                                         **/
/**                     IBM SCSI Tape Utility Program                       **/
/**                   (C) COPYRIGHT IBM Corporation, 1994-                  **/
/**                                                                         **/
/**                 tapeutil.c - Test Program Source File                   **/
/**                  Licensed Materials - Property of IBM                   **/
/**                           All Rights Reserved                           **/
/**                                                                         **/
/**    US Government Users Restricted Rights  -  Use, duplication, or       **/
/**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.    **/
/**                                                                         **/
/**           ** Notice to users of the source code examples: **            **/
/**                                                                         **/
/**  The source code examples provided by IBM are only intended to assist   **/
/**  in the development of a working software program.  The source code     **/
/**  examples may not function as written; additional code is required.     **/
/**  In addition, the source code examples may not compile and/or bind      **/
/**  successfully as written.                                               **/
/**                                                                         **/
/**  International Business Machines Corporation provides the source code   **/
/**  examples "as is" without warranty of any kind, either expressed or     **/
/**  implied, including, but not limited to the implied warranties of       **/
/**  merchantability and fitness for a particular purpose.  The entire      **/
/**  risk as to the quality and performance of the source code examples     **/
/**  is with you.  Should any part of the source code examples prove        **/
/**  defective, you (and not IBM) assume the entire cost of all necessary   **/
/**  servicing, repair or  correction.                                      **/
/**                                                                         **/
/**  IBM does not warrant that the contents of the source code examples     **/
/**  will meet your requirements or that the source code examples are       **/
/**  error-free.                                                            **/
/**                                                                         **/
/**  IBM may make improvements and/or changes in the source code examples   **/
/**  at any time.  Changes may be made periodically to the information in   **/
/**  the source code examples; these changes may be reported, for the       **/
/**  device drivers included herein, in new editions of the examples.       **/
/**                                                                         **/
/**  References in the source code examples to IBM products, programs, or   **/
/**  services do not imply that IBM intends to make these available in all  **/
/**  countries in which IBM operates.  Any reference to an IBM licensed     **/
/**  program in the source code examples is not intended to state or imply  **/
/**  that only IBM's licensed program may be used.  Any functionally        **/
/**  equivalent program may be used.                                        **/
/**                                                                         **/
/*****************************************************************************/

/*****************************************************************************/
/*                               INCLUDE FILES                               */
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <sys/devinfo.h>
#include <sys/errno.h>
#include <sys/scsi.h>
#include <sys/tape.h>
#include <sys/Atape.h>

/* Added to support files > 2 gigabytes in AIX 4.2 */
#ifndef O_LARGEFILE
#define O_LARGEFILE  0x04000000
#endif

/*****************************************************************************/
/*                            FUNCTION PROTOTYPES                            */
/*****************************************************************************/
void menu(void);
int cmdline(int,char **);

int iocinfo();
int sioc_inquiry(void);
int sioc_reqsense(void);
int sioc_reserve(void);
int sioc_release(void);
int sioc_test_unit_ready(void);
int sioc_log_sense_page(int, int, int);
int sioc_mode_sense_page(int);
int sioc_inquiry_page(int);
int sioc_query_path(void);
int sioc_device_paths(void);
int sioc_reset_path(void);
int sioc_check_path(void);
int sioc_disable_path(int);
int sioc_enable_path(int);
void show_path(struct scsi_path *);
int sioc_query_open(void);
int sioc_kill_open(void);
int sioc_debug(void);
int sioc_reset_device(void);
int sioc_trcstart(void);
int sioc_trcstop(void);

int stioctop(int,int);
int stiocqryp(void);
int stiocsetp(void);
int stiocsync();
int stiocdm(char, char *, char *);
int stiocqrypos(void);
int stiocsetpos(void);
int stiocqrysense(void);
int stiocqryinquiry(void);
int stioc_log_sense(void);
int stioc_recover_buffer(int);
int mtdevice(void);
int stioc_locate(void);
int stioc_read_position(void);
int stioc_dump(char *);
int stioc_force_dump(void);
int stioc_read_dump(char *);
void stioc_load_ucode(char *);
int stioc_reset_drive(void);
int stioc_fmr_tape(void);
int stioc_set_volid(char *);
int stioc_create_partition(int, int);
int stioc_query_partition(void);
int stioc_change_partition(int,int);
int stioc_report_density_support(void);
int stioc_prevent_medium_removal(void);
int stioc_allow_medium_removal(void);
int sioc_passthru_command(void);
int stioc_get_encryption_state(void);
int encryption_test(void);
int rekey_command(void);
int get_sernum(void);

int smcioc_element_info(void);
int smcioc_move_medium(ushort,ushort);
int smcioc_exchange_medium(ushort,ushort,ushort);
int smcioc_load_medium(ushort);
int smcioc_unload_medium(ushort);
int smcioc_pos_to_elem(ushort);
int smcioc_init_elem_stat(void);
int smcioc_init_elem_stat_range(ushort,ushort);
int smcioc_inventory(char *, int);
int smcioc_prevent_medium_removal(void);
int smcioc_allow_medium_removal(void);
int smcioc_read_element_devids(void);

int twrite(char *);
int tread(char *, int);
int rwtest(int,int,int);
int rtest(int,int,int);
int wtest(int,int,int);
int tape_parms(void);
int list_tape(void);
int qvpd(void);
void help(void);
void dump_bytes(uchar *, int,char *);
int htoc(char *, unsigned char);
void genpat(void);

#define min(x,y)     ((x)<(y)?(x):(y))

/*****************************************************************************/
/*                                GLOBAL DATA                                */
/*****************************************************************************/
char       opcode = 0;      /* Menu selected operation              */
short      source;          /* source for move command              */
short      dest;            /* destination for move command         */
short      dest2;           /* destination2 for exchange command    */
short      number;
int        blksize;         /* block size for data transfers        */
int        blocks;          /* number of blocks/bytes to write      */
int        repetition;      /* number of write/read operation       */
short      stop;            /* st_op->stop field (for mt commands)  */
daddr_t    count;           /* st_op->count field (for mt commands) */
char       msg_type;        /* msg type field                       */
char       msg_0[9];        /* message 0 text                       */
char       msg_1[9];        /* message 1 text                       */
char       *sname;          /* source file name                     */
char       *dname;          /* destination file name                */

int        fd;                         /* file descriptor               */
char       opened = 0;                 /* is special file open yet      */
char       devname[40] = "";           /* device special file name      */
int        mode_flags;                 /* flags to use for cmdline open */
int        ext_flags=0;                /* extended flags for  "    open */
uchar      pat[256] = { 0x00 } ;       /* data pattern buffer           */
int err_stop = TRUE;

struct stpos_s stpos;
struct stchgp_s parms;
uint blockid;
struct trace_log trclog;

/*---------------------------------------------------------------------------*/
/* Name:        main                                                         */
/*---------------------------------------------------------------------------*/
int main(int argc, char ** argv)
  {
  int rc;

  if (argc == 1)
    {
    menu();       /* Call the interactive version */
    if (opened)
      close(fd);              /* Close if opened */
    exit(0);
    }

  /* Otherwise assume that it is the cmdline interface version */
  rc = cmdline(argc,argv);             /* Command line version */
  exit(rc);
  }

/*---------------------------------------------------------------------------*/
/* Name:        cmdline                                                      */
/*---------------------------------------------------------------------------*/
int cmdline(int argc,char ** argv)
  {
  int i,j;
  int rc = 0;
  int arg_count = 1;
  char *argp;
  int st_count = 1;
  int source;
  int dest;
  int number,blockid;
  char msg[80];
  int vpd_mode = FALSE;
  int dump_mode = FALSE;
  int read_dump_mode = FALSE;
  int force_dump_mode = FALSE;
  int read_mode = FALSE;
  int wrt_mode = FALSE;
  int append_mode = FALSE;
  int ndelay = FALSE;
  int page;
  int loop_count=0;
  int loop_pass=0;
  int loop_start=0;

  /* help subcommand */
  if ((strcmp(argv[1], "-?") == 0) ||
      (strcmp(argv[1], "?") == 0))
    {
    help();
    exit(0);
    }

  /* check first arg for device */
  if (strncmp("-f",argv[1],2))
    {
    help();
    exit(1);
    }

  if (strlen(argv[1]) == 2)
    {
    arg_count++;
    if (arg_count > argc)
      {
      printf("Missing argument for device\n");
      exit(1);
      }
    strcpy(devname,argv[2]);
    }
  else
    strcpy(devname,argv[1]+2);

  arg_count++;
  if (arg_count > argc)
    {
    printf("Missing subcommand\n");
    exit(1);
    }

  /* see how we should open */
  for (i = arg_count; i < argc; i++)
    {
    if ((strcmp(argv[i],"mount") == 0) ||
        (strcmp(argv[i],"unmount") == 0) ||
        (strcmp(argv[i],"demount") == 0) ||
        (strcmp(argv[i],"position") == 0) ||
        (strcmp(argv[i],"allow") == 0) ||
        (strcmp(argv[i],"prevent") == 0) ||
        (strcmp(argv[i],"exchange") == 0) ||
        (strcmp(argv[i],"move") == 0))
      ndelay = TRUE;

    if ((strcmp(argv[i],"rtest") == 0) ||
        (strcmp(argv[i],"rwtest") == 0) ||
        (strcmp(argv[i],"qrypart") == 0)  ||
        (strcmp(argv[i],"chgpart") == 0)  ||
        (strcmp(argv[i],"list") == 0)  ||
        (strcmp(argv[i],"sili") == 0)  ||
        (strcmp(argv[i],"nosili") == 0)  ||
        (strcmp(argv[i],"read") == 0))
      read_mode = TRUE;

    if ((strcmp(argv[i],"wtest") == 0) ||
        (strcmp(argv[i],"rwtest") == 0) ||
        (strcmp(argv[i],"write") == 0)  ||
        (strcmp(argv[i],"erase") == 0)  ||
        (strcmp(argv[i],"eraseimm") == 0)  ||
        (strcmp(argv[i],"sync") == 0)  ||
        (strcmp(argv[i],"eof") == 0)  ||
        (strcmp(argv[i],"eofimm") == 0)  ||
        (strcmp(argv[i],"erg") == 0)  ||
        (strcmp(argv[i],"sdp") == 0)  ||
        (strcmp(argv[i],"idp") == 0)  ||
        (strcmp(argv[i],"weof") == 0) ||
        (strcmp(argv[i],"rekey") == 0)  ||
        (strcmp(argv[i],"weofimm") == 0))
      wrt_mode = TRUE;

    if (strcmp(argv[i],"append") == 0)
      {
      wrt_mode = TRUE;
      append_mode = TRUE;
      }

    if (strcmp(argv[i],"dump") == 0)
      {
      ext_flags = SC_SERVICE;
      dump_mode = TRUE;
      }
    if (strcmp(argv[i],"readdump") == 0)
      {
      ext_flags = SC_SERVICE;
      read_dump_mode = TRUE;
      }
    if (strcmp(argv[i],"forcedump") == 0)
      {
      ext_flags = SC_SERVICE;
      force_dump_mode = TRUE;
      }
    if (strcmp(argv[i],"vpd") == 0)
      {
      ext_flags = SC_DIAGNOSTIC;
      vpd_mode = TRUE;
      }
    if (strcmp(argv[i],"ucode") == 0 ||
        strcmp(argv[i],"fmrtape") == 0 ||
        strcmp(argv[i],"resetdrive") == 0)
      ext_flags = SC_SERVICE;
    if (strcmp(argv[i],"qrypath") == 0 ||
        strcmp(argv[i],"path") == 0 ||
        strcmp(argv[i],"resetpath") == 0 ||
        strcmp(argv[i],"checkpath") == 0 ||
        strcmp(argv[i],"disablepath") == 0 ||
        strcmp(argv[i],"enablepath") == 0 ||
        strcmp(argv[i],"kill") == 0 ||
        strcmp(argv[i],"trcstart") == 0 ||
        strcmp(argv[i],"trcstop") == 0 ||
        strcmp(argv[i],"reset") == 0 ||
        strcmp(argv[i],"fuser") == 0)
      ext_flags = SC_TMCP;
    if (strcmp(argv[i],"passthru") == 0)
      ext_flags = SC_PASSTHRU;
    if (strcmp(argv[i],"tmcp") == 0)
      ext_flags = SC_TMCP;
    if (strcmp(argv[i],"debug") == 0 && (i + 1) == argc)
      ext_flags = SC_TMCP;
    if (strcmp(argv[i],"encryption") == 0)
      ext_flags = SC_TMCP;
    }

  if (read_mode && wrt_mode)
    mode_flags = O_RDWR;
  else if (read_mode)
    mode_flags = O_RDONLY;
  else if (wrt_mode)
    mode_flags = O_WRONLY;
  else
    mode_flags = O_NDELAY;

  if (append_mode)
    mode_flags |= O_APPEND;

  if (ndelay)
    mode_flags |= O_NDELAY;

  /* Open the device */
  if (ext_flags)
    fd = openx(devname,mode_flags, NULL, ext_flags);
  else
    fd = open(devname,mode_flags);

  /* if we don't have root authority, try non-root Atape passthru */
  if (fd < 0 && errno == EPERM && vpd_mode)
    {
    ext_flags = SC_PASSTHRU;
    fd = openx(devname,mode_flags, NULL, ext_flags);
    }

  /* if device is already open, allow dump using tmcp open */
  if (fd < 0 && errno == EAGAIN && (dump_mode || read_dump_mode || force_dump_mode))
    {
    ext_flags = SC_TMCP;
    fd = openx(devname,mode_flags, NULL, ext_flags);
    }

  if (fd < 0)
    {
    perror("Open failed");
    return errno;
    }

  for (i = arg_count;; i++)
    {
    fflush(stdout);
    if (rc != 0)
      {
      if (errno == EINVAL)
        sprintf(msg,"Operation not supported on device, errno %d",errno);
      else
        sprintf(msg,"Operation failed with errno %d",errno);
      perror (msg);
      if (ext_flags != SC_TMCP)
        stiocqrysense();
      if (err_stop)
        break;
      }

    if (i == argc)
      {
      if (loop_start)
        {
        loop_pass++;
        if (loop_count && (loop_pass > loop_count))
          break;
        else
          i = loop_start;
        }
      else
        break;
      }

    errno = 0;
    rc = 0;
    argp = argv[i];

    /* help subcommand */
    if ((strcmp(argp, "-?") == 0) ||
        (strcmp(argp, "?") == 0))
      {
      help();
      continue;
      }

    /* append subcommand handled on open */
    if (strcmp(argp, "append") == 0)
      {
      printf("Open in append mode...\n");
      continue;
      }

    /* tmcp subcommand handled on open */
    if (strcmp(argp, "tmcp") == 0)
      continue;

    /* passthru subcommand handled on open */
    if (strcmp(argp, "passthru") == 0)
      continue;

    if (strcmp(argp, "nostop") == 0)
      {
      err_stop=FALSE;
      continue;
      }

    if (strcmp(argp, "encryption") == 0)
      {
      rc = stioc_get_encryption_state();
      continue;
      }

    if (strcmp(argp, "ekmtest") == 0)
      {
      rc = encryption_test();
      continue;
      }

    if (strcmp(argp, "reset") == 0)
      {
      rc = sioc_reset_device();
      continue;
      }

    /* query path subcommand */
    if (strcmp(argp, "qrypath") == 0)
      {
      rc = sioc_query_path();
      continue;
      }

    /* path subcommand */
    if (strcmp(argp, "path") == 0)
      {
      rc = sioc_device_paths();
      continue;
      }

    /* fuser subcommand */
    if (strcmp(argp, "fuser") == 0)
      {
      rc = sioc_query_open();
      continue;
      }

    /* kill subcommand */
    if (strcmp(argp, "kill") == 0)
      {
      rc = sioc_kill_open();
      continue;
      }

    /* debug subcommand */
    if (strcmp(argp, "debug") == 0)
      {
      rc = sioc_debug();
      continue;
      }

    /* reset path subcommand */
    if (strcmp(argp, "resetpath") == 0)
      {
      rc = sioc_reset_path();
      continue;
      }

    /* check path subcommand */
    if (strcmp(argp, "checkpath") == 0)
      {
      rc = sioc_check_path();
      continue;
      }

    /* disable path subcommand */
    if (strcmp(argp, "disablepath") == 0)
      {
      i++;
      if (i < argc)
        {
        if (strcmp(argv[i], "primary") == 0)
          rc = sioc_disable_path(PRIMARY_SCSI_PATH);
        else if (strcmp(argv[i], "alternate") == 0)
          rc = sioc_disable_path(ALTERNATE_SCSI_PATH);
        else if (isdigit(argv[i][0]) && atoi(argv[i]) > 0)
          rc = sioc_disable_path(atoi(argv[i]));
        else
          {
          printf("Invalid path argument for disablepath subcommand\n");
          exit(1);
          }
        }
      else
        {
        printf("Missing argument for disablepath subcommand\n");
        exit(1);
        }
      continue;
      }

    /* enable path subcommand */
    if (strcmp(argp, "enablepath") == 0)
      {
      i++;
      if (i < argc)
        {
        if (strcmp(argv[i], "primary") == 0)
          rc = sioc_enable_path(PRIMARY_SCSI_PATH);
        else if (strcmp(argv[i], "alternate") == 0)
          rc = sioc_enable_path(ALTERNATE_SCSI_PATH);
        else if (isdigit(argv[i][0]) && atoi(argv[i]) > 0)
          rc = sioc_enable_path(atoi(argv[i]));
        else
          {
          printf("Invalid path argument for enablepath subcommand\n");
          exit(1);
          }
        }
      else
        {
        printf("Missing argument for enablepath subcommand\n");
        exit(1);
        }
      continue;
      }

    /* loop subcommand */
    if (strcmp(argp, "loop") == 0)
      {
      if ((i + 1) < argc)
        {
        if (isdigit(argv[i+1][0]))
          {
          loop_count = atoi(argv[i+1]);
          i++;
          if ((i + 1) == argc)
            {
            printf("Missing argument for loop subcommand\n");
            exit(1);
            }
          }
        else
          loop_count = 0;
        }
      else
        {
        printf("Missing argument for loop subcommand\n");
        exit(1);
        }
      loop_pass = 1;
      loop_start = i + 1;
      continue;
      }

    /* sleep subcommand */
    if (strcmp(argp, "sleep") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for sleep subcommand\n");
        exit(1);
        }
      if (!isdigit(argv[i][0]))
        {
        printf("Invalid argument for sleep subcommand\n");
        exit(1);
        }
      number = atoi(argv[i]);
      sleep(number);
      continue;
      }

    /* density subcommand */
    if (strcmp(argp, "density") == 0)
      {
      rc = stioc_report_density_support();
      continue;
      }

    /* load ucode subcommand */
    if (strcmp(argp, "ucode") == 0)
      {
      i++;
      if (i < argc)
        stioc_load_ucode(argv[i]);
      else
        {
        printf("Missing argument for ucode subcommand\n");
        exit(1);
        }
      continue;
      }

    /* dump subcommand */
    if (strcmp(argp, "dump") == 0)
      {
      /* if another arg, assume it's a dump filename */
      if ((i + 1) < argc)
        {
        i++;
        rc = stioc_dump(argv[i]);
        }
      else
        rc = stioc_dump(NULL);
      continue;
      }

    /* readdump subcommand */
    if (strcmp(argp, "readdump") == 0)
      {
      /* if another arg, assume it's a dump filename */
      if ((i + 1) < argc)
        {
        i++;
        rc = stioc_read_dump(argv[i]);
        }
      else
        rc = stioc_read_dump(NULL);
      continue;
      }

    /* force dump subcommand */
    if (strcmp(argp, "forcedump") == 0)
      {
      rc = stioc_force_dump();
      continue;
      }

    /* reset drive subcommand */
    if (strcmp(argp, "resetdrive") == 0)
      {
      rc = stioc_reset_drive();
      continue;
      }

    /* create fmr tape subcommand */
    if (strcmp(argp, "fmrtape") == 0)
      {
      rc = stioc_fmr_tape();
      continue;
      }

    /* vpd subcommand */
    if (strcmp(argp, "vpd") == 0)
      {
      rc = qvpd();
      continue;
      }

    /* scsi passthru subcommand */
    if (strcmp(argp, "scsicmd") == 0)
      {
      rc = sioc_passthru_command();
      continue;
      }

    /* list subcommand */
    if (strcmp(argp, "list") == 0)
      {
      rc = list_tape();
      continue;
      }

    /* print subcommand */
    if (strcmp(argp, "print") == 0)
      {
      i++;
      if (i < argc)
        printf("%s\n",argv[i]);
      else
        {
        printf("Missing argument for print subcommand\n");
        exit(1);
        }
      continue;
      }

    /* set volume id subcommand */
    if (strcmp(argp, "volid") == 0)
      {
      i++;
      if (i < argc)
        rc = stioc_set_volid(argv[i]);
      else
        {
        printf("Missing argument for volid subcommand\n");
        exit(1);
        }
      continue;
      }

    /* set autoload subcommand */
    if (strcmp(argp, "autoload") == 0)
      {
      printf("Setting autoload ON\n");
      rc = stiocqryp();
      if (rc == 0)
        {
        parms.autoload = TRUE;
        rc = stiocsetp();
        }
      continue;
      }

    /* set autoload subcommand */
    if (strcmp(argp, "noautoload") == 0)
      {
      printf("Setting autoload OFF\n");
      rc = stiocqryp();
      if (rc == 0)
        {
        parms.autoload = FALSE;
        rc = stiocsetp();
        }
      continue;
      }

    /* set read sili bit subcommand */
    if (strcmp(argp, "sili") == 0)
      {
      printf("Setting Read SILI bit ON\n");
      rc = stiocqryp();
      if (rc == 0)
        {
        parms.read_sili_bit = TRUE;
        rc = stiocsetp();
        }
      continue;
      }

    /* set read sili bit subcommand */
    if (strcmp(argp, "nosili") == 0)
      {
      printf("Setting Read SILI bit OFF\n");
      rc = stiocqryp();
      if (rc == 0)
        {
        parms.read_sili_bit = FALSE;
        rc = stiocsetp();
        }
      continue;
      }

    /* set compression subcommand */
    if (strcmp(argp, "compress") == 0)
      {
      printf("Setting compression ON\n");
      rc = stiocqryp();
      if (rc == 0)
        {
        parms.compression = TRUE;
        rc = stiocsetp();
        }
      continue;
      }

    /* set no compression subcommand */
    if (strcmp(argp, "nocompress") == 0)
      {
      printf("Setting compression OFF\n");
      rc = stiocqryp();
      if (rc == 0)
        {
        parms.compression = FALSE;
        rc = stiocsetp();
        }
      continue;
      }

    /* display message subcommand */
    if (strcmp(argp, "display") == 0)
      {
      i++;
      if (i < argc)
        {
        if (strlen(argv[i]) > 8)
          rc = stiocdm(5, argv[i], argv[i]+8);
        else
          rc = stiocdm(1, argv[i], NULL);
        }
      else
        {
        printf("Missing argument for display subcommand\n");
        exit(1);
        }
      continue;
      }

    /* inquiry subcommand */
    if (strncmp(argp, "inquiry",3) == 0)
      {
      /* check for optional page code */
      if ((i + 1) < argc)
        {
        page = (int)strtol(argv[i+1],NULL,16);
        if (isdigit(argv[i+1][0]) || page)
          {
          i++;
          rc = sioc_inquiry_page(page);
          }
        else
          rc = sioc_inquiry();
        }
      else
        rc = sioc_inquiry();
      continue;
      }

    /* mode sense page subcommand */
    if (strcmp(argp, "modepage") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for modepage subcommand\n");
        exit(1);
        }
      page = (int)strtol(argv[i],NULL,16);
      if (isdigit(argv[i][0]) || page)
        rc = sioc_mode_sense_page(page);
      else
        {
        printf("Invalid argument for modepage subcommand\n");
        exit(1);
        }
      continue;
      }

    /* request sense subcommand */
    if (strncmp(argp, "reqsense", 3) == 0)
      {
      rc = sioc_reqsense();
      continue;
      }

    /* read position subcommand */
    if (strcmp(argp, "readpos") == 0)
      {
      rc = stioc_read_position();
      continue;
      }

    /* locate subcommand */
    if (strcmp(argp, "locate") == 0)
      {
      rc = stioc_locate();
      if (rc == 0)
        rc = stioc_read_position();
      continue;
      }

    /* query position subcommand */
    if (strcmp(argp, "qrypos") == 0)
      {
      rc = stiocqrypos();
      continue;
      }

    /* set position subcommand */
    if (strcmp(argp, "setpos") == 0)
      {
      /* check for optional bockid arg in decimal */
      if (((i + 1) < argc) && isdigit(argv[i+1][0]))
        {
        i++;
        /* set the global position structure used for qrypos/setpos defaults */
        stpos.curpos = atoi(argv[i]);
        stpos.block_type = QP_LOGICAL;
        }
      /* check for optional bockid arg in hex */
      else if (((i + 1) < argc) && argv[i+1][0] == 'x')
        {
        i++;
        /* set the global position structure used for qrypos/setpos defaults */
        stpos.curpos = (int)strtol(argv[i]+1,NULL,16);
        stpos.block_type = QP_LOGICAL;
        }
      rc = stiocsetpos();
      if (rc == 0)
        rc = stiocqrypos();
      continue;
      }

    /* query inquiry subcommand */
    if (strcmp(argp, "qryinquiry") == 0)
      {
      rc = stiocqryinquiry();
      continue;
      }

    /* get devinfo subcommand */
    if (strcmp(argp, "devinfo") == 0)
      {
      rc = iocinfo();
      continue;
      }

    /* test unit ready subcommand */
    if (strcmp(argp, "tur") == 0)
      {
      rc = sioc_test_unit_ready();
      continue;
      }

    /* rekey subcommand */
    if (strcmp(argp, "rekey") == 0)
      {
      rc = rekey_command();
      continue;
      }


    /* reserve subcommand */
    if (strcmp(argp, "reserve") == 0)
      {
      rc = sioc_reserve();
      continue;
      }

    /* release subcommand */
    if (strcmp(argp, "release") == 0)
      {
      rc = sioc_release();
      continue;
      }

    /* get element information subcommand */
    if (strncmp(argp, "elementinfo", 3) == 0)
      {
      rc = smcioc_element_info();
      continue;
      }

    /* initialize element status subcommand */
    if (strcmp(argp, "audit") == 0)
      {
      /* check for audit range */
      if ((i + 1) < argc && isdigit(argv[i+1][0]))
        {
        i++;
        source = atoi(argv[i]);
        if ((i + 1) < argc && isdigit(argv[i+1][0]))
          {
          i++;
          st_count = atoi(argv[i]);
          }
        else
          st_count = 1;
        rc = smcioc_init_elem_stat_range(source,st_count);
        }
      else
        rc = smcioc_init_elem_stat();
      continue;
      }

    /* prevent medium removal subcommand */
    if (strcmp(argp, "prevent") == 0)
      {
      /* check if it's an smc or tape device */
      if (strstr(devname, "smc") != NULL)
        rc = smcioc_prevent_medium_removal();
      else
        rc = stioc_prevent_medium_removal();
      continue;
      }

    /* allow medium removal subcommand */
    if (strcmp(argp, "allow") == 0)
      {
      /* check if it's an smc or tape device */
      if (strstr(devname, "smc") != NULL)
        rc = smcioc_allow_medium_removal();
      else
        rc = stioc_allow_medium_removal();
      continue;
      }

    /* read element status subcommand */
    if (strncmp(argp, "inventory", 3) == 0)
      {
      /* check for optional -v volid option */
      if (((i + 1) < argc) && strncmp("-v",argv[i+1],2) == 0)
        {
        i++;
        if (strlen(argv[i]) == 2)
          {
          i++;
          if (i >= argc)
            {
            printf("Missing argument for volume id option\n");
            exit(1);
            }
          rc = smcioc_inventory(argv[i], 0);
          }
        else
          rc = smcioc_inventory(argv[i]+2, 0);
        }
      /* check for optional -i option */
      else if (((i + 1) < argc) && strncmp("-i",argv[i+1],2) == 0)
        {
        i++;
        rc = smcioc_inventory(NULL, 1);
        }
      else
        rc = smcioc_inventory(NULL, 0);
      continue;
      }

    /* read element device ids subcommand */
    if (strncmp(argp, "devids", 5) == 0)
      {
      rc = smcioc_read_element_devids();
      continue;
      }

    /* move medium subcommand */
    if (strcmp(argp, "move") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for move subcommand\n");
        exit(1);
        }
      /* check for new format */
      if (isdigit(argv[i][0]))
        source = atoi(argv[i]);
      else
        {
        argp = argv[i];
        if (strncmp("-s",argp,2) != 0)  /* source argument */
          {
          printf("Missing argument for move subcommand\n");
          exit(1);
          }
        if (strlen(argp) == 2)
          {
          i++;
          source = atoi(argv[i]);
          }
        else
          source = atoi(argv[i]+2);
        }
      i++;
      if (i >= argc)
        {
        printf("Missing argument for move subcommand\n");
        exit(1);
        }
      /* check for new format */
      if (isdigit(argv[i][0]))
        dest = atoi(argv[i]);
      else
        {
        argp = argv[i];
        if (strncmp("-d",argp,2) != 0)  /* destination argument */
          {
          printf("Missing argument for move subcommand\n");
          exit(1);
          }
        if (strlen(argp) == 2)
          {
          i++;
          dest = atoi(argv[i]);
          }
        else
          dest = atoi(argv[i]+2);
        }
      rc = smcioc_move_medium(source,dest);
      continue;
      }

    /* exchange medium subcommand */
    if (strcmp(argp, "exchange") == 0)
      {
      /* get source arg */
      if (((i + 1) < argc) && isdigit(argv[i+1][0]))
        {
        i++;
        source = atoi(argv[i]);
        }
      else
        {
        printf("Missing argument for exchange subcommand\n");
        exit(1);
        }
      /* get dest1 arg */
      if (((i + 1) < argc) && isdigit(argv[i+1][0]))
        {
        i++;
        dest = atoi(argv[i]);
        }
      else
        {
        printf("Missing argument for exchange subcommand\n");
        exit(1);
        }
      /* get dest2 arg */
      if (((i + 1) < argc) && isdigit(argv[i+1][0]))
        {
        i++;
        dest2 = atoi(argv[i]);
        }
      else
        {
        printf("Missing argument for exchange subcommand\n");
        exit(1);
        }
      rc = smcioc_exchange_medium(source,dest,dest2);
      continue;
      }

    /* mount subcommand */
    if (strcmp(argp, "mount") == 0)
      {
      if (i+1 >= argc || !isdigit(argv[i+1][0]))
        source = 0;
      else
        {
        source = atoi(argv[i+1]);
        if (source == 0)
          source = LOAD_UNLOAD_SLOT_0;
        i++;
        }
      rc = smcioc_load_medium(source);
      continue;
      }

    /* unmount subcommand */
    if (strcmp(argp, "unmount") == 0 || strcmp(argp, "demount") == 0)
      {
      if (i+1 >= argc || !isdigit(argv[i+1][0]))
        dest = 0;
      else
        {
        dest = atoi(argv[i+1]);
        if (dest == 0)
          dest = LOAD_UNLOAD_SLOT_0;
        i++;
        }
      rc = smcioc_unload_medium(dest);
      continue;
      }

    /* position to element subcommand */
    if (strncmp(argp, "position",3) == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for position subcommand\n");
        exit(1);
        }
      /* check for new format */
      if (isdigit(argv[i][0]))
        dest = atoi(argv[i]);
      else
        {
        argp = argv[i];
        if (strncmp("-d",argp,2) != 0)  /* destination argument */
          {
          printf("Missing argument for position subcommand\n");
          exit(1);
          }
        if (strlen(argp) == 2)
          {
          i++;
          dest = atoi(argv[i]);
          }
        else
          dest = atoi(argv[i]+2);
        }
      rc = smcioc_pos_to_elem(dest);
      continue;
      }

    /* synchronize buffers subcommand */
    if (strcmp(argp, "sync") == 0)
      {
      rc = stiocsync();
      continue;
      }

    /* log sense page subcommand */
    if (strcmp(argp, "logpage") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for logpage subcommand\n");
        exit(1);
        }
      page = (int)strtol(argv[i],NULL,16);
      if (isdigit(argv[i][0]) || page)
        rc = sioc_log_sense_page(page, 0, 0);
      else
        {
        printf("Invalid argument for logpage subcommand\n");
        exit(1);
        }
      continue;
      }

    /* log sense subcommand */
    if (strcmp(argp, "logsense") == 0)
      {
      rc = stioc_log_sense();
      continue;
      }

    /* get mtdevice number subcommand */
    if (strcmp(argp, "mtdevice") == 0)
      {
      rc = mtdevice();
      continue;
      }

    /* get serial number subcommand */
    if (strcmp(argp, "sernum") == 0)
      {
      rc = get_sernum();
      continue;
      }

    /* get status subcommand */
    if (strcmp(argp, "status") == 0 ||
        strcmp(argp, "parms") == 0)
      {
      rc = status();
      continue;
      }

    /* load tape subcommand */
    if (strcmp(argp, "load") == 0)
      {
      rc = stioctop(STINSRT,0);
      continue;
      }

    /* unload tape subcommand */
    if (strcmp(argp, "offline") == 0  ||
        strcmp(argp, "rewoffl") == 0  ||
        strcmp(argp, "unload") == 0)
      {
      rc = stioctop(STOFFL,0);
      continue;
      }

    /* rewind tape subcommand */
    if (strcmp(argp, "rewind") == 0  ||
        strcmp(argp, "retension") == 0)
      {
      rc = stioctop(STREW,0);
      continue;
      }

    /* erase tape subcommand */
    if (strcmp(argp, "erase") == 0)
      {
      rc = stioctop(STERASE,0);
      continue;
      }

    /* eraseimm tape subcommand */
    if (strcmp(argp, "eraseimm") == 0)
      {
      rc = stioctop(STERASE_IMM,0);
      continue;
      }

    /* erase record gap subcommand */
    if (strcmp(argp, "erg") == 0)
      {
      rc = stioctop(STERASEGAP,0);
      continue;
      }

    /* space to end-of-data subcommand */
    if (strcmp(argp, "seod") == 0)
      {
      rc = stioctop(STSEOD,0);
      continue;
      }

    /* write filemarks subcommand */
    if (strcmp(argp, "eof") == 0  ||
        strcmp(argp, "weof") == 0)
      {
      if ((i + 1) < argc)
        {
        st_count = atoi(argv[i+1]);
        if (st_count > 0)
          i++;
        else
          st_count = 1;
        }
      else
        st_count = 1;
      rc = stioctop(STWEOF,st_count);
      continue;
      }

    /* write filemarks immediate subcommand */
    if (strcmp(argp, "eofimm") == 0  ||
        strcmp(argp, "weofimm") == 0)
      {
      if ((i + 1) < argc)
        {
        st_count = atoi(argv[i+1]);
        if (st_count > 0)
          i++;
        else
          st_count = 1;
        }
      else
        st_count = 1;
      rc = stioctop(STWEOF_IMM,st_count);
      continue;
      }

    /* forward space filemarks subcommand */
    if (strcmp(argp, "fsf") == 0)
      {
      if ((i + 1) < argc)
        {
        st_count = atoi(argv[i+1]);
        if (st_count > 0)
          i++;
        else
          st_count = 1;
        }
      else
        st_count = 1;
      rc = stioctop(STFSF,st_count);
      continue;
      }

    /* backward space filemarks subcommand */
    if (strcmp(argp, "bsf") == 0)
      {
      if ((i + 1) < argc)
        {
        st_count = atoi(argv[i+1]);
        if (st_count > 0)
          i++;
        else
          st_count = 1;
        }
      else
        st_count = 1;
      rc = stioctop(STRSF,st_count);
      continue;
      }

    /* forward space records subcommand */
    if (strcmp(argp, "fsr") == 0)
      {
      if ((i + 1) < argc)
        {
        st_count = atoi(argv[i+1]);
        if (st_count > 0)
          i++;
        else
          st_count = 1;
        }
      else
        st_count = 1;
      rc = stioctop(STFSR,st_count);
      continue;
      }

    /* backward space records subcommand */
    if (strcmp(argp, "bsr") == 0)
      {
      if ((i + 1) < argc)
        {
        st_count = atoi(argv[i+1]);
        if (st_count > 0)
          i++;
        else
          st_count = 1;
        }
      else
        st_count = 1;
      rc = stioctop(STRSR,st_count);
      continue;
      }

    /* read and write tests */
    if (strcmp(argp, "wtest") == 0 ||
        strcmp(argp, "rtest") == 0 ||
        strcmp(argp, "rwtest") == 0)
      {
      blksize = 10240;
      blocks = 20;
      repetition = 1;
      /* look for r/w options */
      rc = 0;
      for (j = 0; (j < 3) && ((i + 1) < argc); j++)
        {
        if (strncmp("-b",argv[i+1],2) == 0)  /* blocksize argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for blocksize option\n");
              exit(1);
              }
            blksize = atoi(argv[i]);
            }
          else
            blksize = atoi(argv[i]+2);
          continue;
          }
        if (strncmp("-c",argv[i+1],2) == 0)  /* count argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for count option\n");
              exit(1);
              }
            blocks = atoi(argv[i]);
            }
          else
            blocks = atoi(argv[i]+2);
          continue;
          }
        if (strncmp("-r",argv[i+1],2) == 0)  /* repetition argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for repetition option\n");
              exit(1);
              }
            repetition = atoi(argv[i]);
            }
          else
            repetition = atoi(argv[i]+2);
          continue;
          }
        break;
        }
      if (rc)
        continue;
      if (strcmp(argp, "wtest") == 0)
        rc = wtest(blksize,blocks,repetition);
      else if (strcmp(argp, "rtest") == 0)
        rc = rtest(blksize,blocks,repetition);
      else
        rc = rwtest(blksize,blocks,repetition);
      continue;
      }

    /* write  */
    if (strcmp(argp, "write") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for write subcommand\n");
        exit(1);
        }
      argp = argv[i];
      if (strncmp("-s",argp,2) != 0)  /* source argument */
        {
        printf("Missing argument for write subcommand\n");
        exit(1);
        }
      if (strlen(argp) == 2)
        {
        i++;
        sname = argv[i];
        }
      else
        sname = argv[i]+2;
      rc = twrite(sname);
      continue;
      }

    /* read  */
    if (strcmp(argp, "read") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for read subcommand\n");
        exit(1);
        }
      argp = argv[i];
      if (strncmp("-d",argp,2) != 0)  /* destination argument */
        {
        printf("Missing argument for read subcommand\n");
        exit(1);
        }
      if (strlen(argp) == 2)
        {
        i++;
        dname = argv[i];
        }
      else
        dname = argv[i]+2;
      /* check for optional -c count  */
      if (((i + 1) < argc) && strncmp("-c",argv[i+1],2) == 0)
        {
        i++;
        if (strlen(argv[i]) == 2)
          {
          i++;
          if (i >= argc)
            {
            printf("Missing argument for count option\n");
            exit(1);
            }
          number = atoi(argv[i]);
          }
        else
          number = atoi(argv[i]+2);
        if (number == 0)
          {
          printf("Invalid argument for count option\n");
          exit(1);
          }
        }
      else
        number = 0;
      rc = tread(dname, number);
      continue;
      }

    /* create sdp partition subcommand */
    if (strcmp(argp, "sdp") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for sdp subcommand\n");
        exit(1);
        }
      number = atoi(argv[i]);
      rc = stioc_create_partition(SDP_PARTITION,number);
      continue;
      }

    /* create idp partition subcommand */
    if (strcmp(argp, "idp") == 0)
      {
      rc = stioc_create_partition(IDP_PARTITION,1);
      continue;
      }

    /* change partition subcommand */
    if (strcmp(argp, "chgpart") == 0)
      {
      i++;
      if (i >= argc)
        {
        printf("Missing argument for chgpart subcommand\n");
        exit(1);
        }
      number = atoi(argv[i]);
      /* check for optional bockid arg */
      if (((i + 1) < argc) && isdigit(argv[i+1][0]))
        {
        i++;
        blockid = atoi(argv[i]);
        }
      else
        blockid = 0;
      rc = stioc_change_partition(number,blockid);
      continue;
      }

    /* query partition subcommand */
    if (strcmp(argp, "qrypart") == 0)
      {
      rc = stioc_query_partition();
      continue;
      }

    /* start trace logging */
    if (strcmp(argp, "trcstart") == 0)
      {
      bzero(trclog,sizeof(struct trace_log));
      /* look for trace options */
      rc = 0;
      for (j = 0; (j < 6) && ((i + 1) < argc); j++)
        {
        if (strncmp("-s",argv[i+1],2) == 0)  /* size argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for size option\n");
              exit(1);
              }
            trclog.size = atoi(argv[i]);
            }
          else
            trclog.size = atoi(argv[i]+2);
          continue;
          }
        if (strncmp("-n",argv[i+1],2) == 0)  /* files argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for number files option\n");
              exit(1);
              }
            trclog.files = atoi(argv[i]);
            }
          else
            trclog.files = atoi(argv[i]+2);
          continue;
          }
        if (strncmp("-c",argv[i+1],2) == 0)  /* current file argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for current file option\n");
              exit(1);
              }
            trclog.cur_file = atoi(argv[i]);
            }
          else
            trclog.cur_file = atoi(argv[i]+2);
          continue;
          }
        if (strncmp("-d",argv[i+1],2) == 0)  /* directory argument */
          {
          i++;
          if (strlen(argv[i]) == 2)
            {
            i++;
            if (i >= argc)
              {
              printf("Missing argument for directory option\n");
              exit(1);
              }
            strcpy(trclog.dir,argv[i]);
            }
          else
            strcpy(trclog.dir,argv[i]+2);
          continue;
          }
        if (strncmp("-D",argv[i+1],2) == 0)  /* debug argument */
          {
          i++;
          trclog.debug_trace = 1;
          continue;
          }
        if (strncmp("-q",argv[i+1],2) == 0)  /* query trace log argument */
          {
          i++;
          trclog.query_trace = 1;
          continue;
          }
        break;
        }
      rc = sioc_trcstart();
      continue;
      }

    /* stop trace logging */
    if (strcmp(argp, "trcstop") == 0)
      {
      rc = sioc_trcstop();
      continue;
      }

    printf("Invalid argument: %s\n",argp);
    exit(1);
    }

  close(fd);                /* Close the device */
  return rc;
  }

/*---------------------------------------------------------------------------*/
/* Name:        menu                                                         */
/*---------------------------------------------------------------------------*/
void menu(void)
  {
  char   buf[80] = "";
  char   fname[80] = "";
  int    option,blockid;
  static int pass = 0;
  int rc = 0;
  struct element_info element_info;
  int page, len, parm, count;

  while(1)
    {
    if (rc != 0)
      {
      if (errno == EINVAL)
        sprintf(buf,"Operation not supported on device, errno %d",errno);
      else
        sprintf(buf,"Operation failed with errno %d",errno);
      perror (buf);
      stiocqrysense();
      }

    errno = 0;
    rc = 0;
    if (pass)
      {
      printf ("\nHit <enter> to continue...");
      getchar();
      }
    else
      pass++;

    printf("\n General Commands:\n");
    printf("    1. Open a Device                  5. Inquiry             9. Log Sense Page\n");
    printf("    2. Close a Device                 6. Test Unit Ready    10. Mode Sense Page\n");
    printf("    3. Device Info                    7. Reserve Device     11. Release Device\n");
    printf("    4. Tape Drive Service Aids        8. Request Sense       Q. Quit Program\n");
    printf("\n Medium Changer Commands:\n");
    printf("   12. Element Information           16. Move Medium\n");
    printf("   13. Position To Element           17. Load/Unload Medium\n");
    printf("   14. Element Inventory             18. Initialize Element Status\n");
    printf("   15. Exchange Medium               19. Prevent/Allow Medium Removal\n");
    printf("\n Tape Commands:\n");
    printf("   20. Query/Set Parameters          30. Read and Write Tests\n");
    printf("   21. Prevent/Allow Medium Removal  31. Unload/Load Tape\n");
    printf("   22. Rewind                        32. Erase\n");
    printf("   23. Erase Gap                     33. Write Filemarks\n");
    printf("   24. Forward Space Filemarks       34. Backward Space Filemarks\n");
    printf("   25. Forward Space Records         35. Backward Space Records\n");
    printf("   26. Space to End of Data          36. Query/Set Tape Position\n");
    printf("   27. Log Sense                     37. Read or Write Files    \n");
    printf("   28. Display Message               38. Get Drive Encryption Settings\n");
    printf("   29. Synchronize Buffers           39. Report Density Support  \n");
    printf("                                     40. Test Encryption Key Path/Setup\n");
    printf("\n");
    if (opened)
      printf("Enter Selection for %s: ",devname);
    else
      printf("Enter Selection: ");
    fflush(stdin);
    gets(buf);
    printf("\n");
    fflush(stdout);

    if (*buf == '\0')
      continue;                 /* Get a valid input */

    if (*buf == 'Q' || *buf == 'q')
      return;

    opcode = atoi(buf);
    /* Require open to be first command, except for service aids */
    if (!opened && opcode != 1 && opcode != 4)
      {
      printf("Nothing is opened yet.\n");
      continue;
      }

    switch (opcode)
      {
      case 1:
        if (opened)
          {
          printf("There is already an open file: %s\n",devname);
          break;
          }

        printf("Enter the device special file name (Example: /dev/rmt0): ");
        gets(devname);
        if (strstr(devname, "smc") != NULL)
          mode_flags = O_NDELAY;
        else
          {
          printf ("\nSelect mode (1=Read/Write, 2=Read Only, 3=Write Only, 4=Append): ");
          gets (buf);
          option = atoi(buf);
          if (option == 1)
            mode_flags = O_RDWR | O_NDELAY;
          else if (option == 2)
            mode_flags = O_RDONLY | O_NDELAY;
          else if (option == 3)
            mode_flags = O_WRONLY | O_NDELAY;
          else if (option == 4)
            mode_flags = O_WRONLY | O_APPEND | O_NDELAY;
          else
            break;
          printf("\n");
          }

        printf("Opening device...\n");
        if ((fd = open(devname,mode_flags)) < 0)
          rc = errno;
        else
          opened = 1;
        break;

      case 2:
        if (close(fd) < 0)
          {
          sprintf(buf,"Device closed with errno %d",errno);
          perror (buf);
          }
        else
          printf("Device closed...\n");
        opened = 0;
        break;

      case 3:
        rc = iocinfo();
        break;

      case 4:
        /* close device first */
        if (opened)
          close(fd);
        rc = system("/usr/lpp/diagnostics/bin/uAtape");
        if (rc)
          {
          printf("Error initializing Service Aid screen, check window size and terminal type.\n");
          rc = 0;
          }
        if (opened)
          {
          fd = open(devname,mode_flags);
          if (fd == -1)
            opened = FALSE;
          }
        break;

      case 5:
        printf ("\nEnter page code in hex or <enter> for standard inquiry: ");
        gets (buf);
        if (strlen(buf) == 0)
          rc = sioc_inquiry();
        else
          {
          page = (int)strtol(buf,NULL,16);
          rc = sioc_inquiry_page(page);
          }
        break;

        rc = sioc_inquiry();
        break;

      case 6:
        rc = sioc_test_unit_ready();
        break;

      case 7:
        rc = sioc_reserve();
        break;

      case 8:
        rc = sioc_reqsense();
        break;

      case 9:
        printf ("\nEnter page code in hex: ");
        gets (buf);
        page = (int)strtol(buf,NULL,16);
        printf ("\nEnter parameter pointer in hex or <enter> for all parameters: ");
        gets (buf);
        parm = (int)strtol(buf,NULL,16);
        if (parm)
          {
          printf ("\nEnter length (including header) or <enter> for entire parameter: ");
          gets (buf);
          len = atoi(buf);
          }
        else
          len=0;

        rc = sioc_log_sense_page(page, len, parm);
        break;

      case 10:
        printf ("\nEnter page code in hex: ");
        gets (buf);
        page = (int)strtol(buf,NULL,16);
        rc = sioc_mode_sense_page(page);
        break;

      case 11:
        rc = sioc_release();
        break;

      case 12:
        rc = smcioc_element_info();
        break;

      case 13:
        printf ("Enter Destination Address: ");
        gets (buf);
        dest = (ushort)atoi(buf);
        printf ("\n");
        rc = smcioc_pos_to_elem(dest);
        break;

      case 14:
        rc = smcioc_inventory(NULL, 0);
        break;

      case 15:
        printf ("Enter Source Address: ");
        gets (buf);
        source = (ushort)atoi(buf);

        printf ("Enter First Destination Address: ");
        gets (buf);
        dest = (ushort)atoi(buf);

        printf ("Enter Second Destination Address: ");
        gets (buf);
        dest2 = (ushort)atoi(buf);
        printf ("\n");
        rc = smcioc_exchange_medium(source,dest,dest2);
        break;

      case 16:
        printf ("Enter Source Address: ");
        gets (buf);
        source = (ushort)atoi(buf);

        printf ("Enter Destination Address: ");
        gets (buf);
        dest = (ushort)atoi(buf);
        printf ("\n");
        rc = smcioc_move_medium(source,dest);
        break;

      case 17:
        printf("\nSelect operation (1=Load, 2=Unload, <enter>=Quit): ");
        gets(buf);
        if (strlen(buf) == 0)
          break;
        option = atoi(buf);
        rc = ioctl (fd, SMCIOC_ELEMENT_INFO, &element_info);
        if (rc)
          break;
        if (option == 1)
          printf ("Enter Slot %d-%d or <enter> for first full slot: ",
                   element_info.slot_addr,element_info.slot_addr+element_info.slots-1);
        else
          printf ("Enter Slot %d-%d or <enter> for first empty slot: ",
                   element_info.slot_addr,element_info.slot_addr+element_info.slots-1);
        gets (buf);
        if (strlen(buf) == 0)
          source = 0;
        else
          {
          source = (ushort)atoi(buf);
          if (source == 0)
            source = LOAD_UNLOAD_SLOT_0;
          }
        if (option == 1)
          rc = smcioc_load_medium(source);
        else if (option == 2)
          rc = smcioc_unload_medium(source);
        break;

      case 18:
        rc = smcioc_init_elem_stat();
        break;

      case 19:
        printf("\nSelect operation (1=Prevent, 2=Allow, <enter>=Quit): ");
        gets(buf);
        if (strlen(buf) == 0)
          break;
        option = atoi(buf);
        if (option == 1)
          rc = smcioc_prevent_medium_removal();
        else
          rc = smcioc_allow_medium_removal();
        break;

      case 20:
        rc = tape_parms();
        break;

      case 21:
        printf("\nSelect operation (1=Prevent, 2=Allow, <enter>=Quit): ");
        gets(buf);
        if (strlen(buf) == 0)
          break;
        option = atoi(buf);
        if (option == 1)
          rc = stioc_prevent_medium_removal();
        else
          rc = stioc_allow_medium_removal();
        break;

      case 22:
        rc = stioctop(STREW,0);
        break;

      case 23:
        rc = stioctop(STERASEGAP,0);
        break;

      case 24:
        printf("Enter number of filemarks to space forward: ");
        gets(buf);
        printf("\n");
        count = atoi(buf);
        rc = stioctop(STFSF,count);
        break;

      case 25:
        printf("Enter number of records to space forward: ");
        gets(buf);
        printf("\n");
        count = atoi(buf);
        rc = stioctop(STFSR,count);
        break;

      case 26:
        rc = stioctop(STSEOD,0);
        break;

      case 27:
        rc = stioc_log_sense();
        break;

      case 28:  /* display message */
        printf ("Select message type (1=msg0, 2=msg1, 3=flash0, 4=flash1, 5=alt): ");
        gets (buf);
        msg_type = (char)atoi (buf);
        if ((msg_type <= 0) || (msg_type > 5))
          {
          printf ("Invalid message type selection.\n");
          break;
          }
        printf ("Enter message 0 text (8 characters or less): ");
        gets (buf);
        if ((strlen (buf) <= 0) || (strlen (buf) > 8))
          {
          printf ("Invalid message length.\n");
          break;
          }
        strcpy (msg_0, buf);
        if (msg_type == 2 || msg_type == 4 ||
            msg_type == 5)
          {
          printf ("Enter message 1 text (8 characters or less): ");
          gets (buf);
          if ((strlen (buf) <= 0) || (strlen (buf) > 8))
            {
            printf ("Invalid message length.\n");
            break;
            }
          strcpy (msg_1, buf);
          rc = stiocdm(msg_type, msg_0, msg_1);
          }
        else
          rc = stiocdm(msg_type, msg_0, NULL);
        break;

      case 29:
        rc = stiocsync();
        break;

      case 30:
        printf ("\nSelect test (1=Read/Write, 2=Read Only, 3=Write Only): ");
        gets (buf);
        option = atoi(buf);
        printf("\nHit <Enter> for default 10240 block size\n");
        printf("or enter new block size: ");
        gets(buf);
        if (strlen(buf) == 0)
          blksize = 10240;
        else
          blksize = (int)atoi(buf);

        if (blksize == 0)
          {
          printf("\nHit <Enter> for default 65536 bytes per read/write\n");
          printf("or enter new bytes: ");
          gets(buf);
          blocks = (int)atoi(buf);
          if (blocks == 0)
            blocks = 65536;
          }
        else
          {
          printf("\nHit <Enter> for default 20 blocks per read/write\n");
          printf("or enter new blocks: ");
          gets(buf);
          blocks = (int)atoi(buf);
          if (blocks == 0)
            blocks = 20;
          }

        printf("\nHit <Enter> for default 1 repetition\n");
        printf("or enter new repetition: ");
        gets(buf);
        repetition = (int)atoi(buf);
        if (repetition == 0)
          repetition = 1;
        printf("\n");
        if (option == 1)
          rc = rwtest(blksize,blocks,repetition);
        else if (option == 2)
          rc = rtest(blksize,blocks,repetition);
        else if (option == 3)
          rc = wtest(blksize,blocks,repetition);
        break;

      case 31:
        printf("\nSelect operation (1=Unload, 2=Load, <enter>=Quit): ");
        gets(buf);
        if (strlen(buf) == 0)
          break;
        option = atoi(buf);
        if (option == 1)
          rc = stioctop(STOFFL,0);
        else
          rc = stioctop(STINSRT,0);
        break;

      case 32:
        rc = stioctop(STERASE,0);
        break;

      case 33:
        printf("Enter number of filemarks to write: ");
        gets(buf);
        printf("\n");
        count = atoi(buf);
        rc = stioctop(STWEOF,count);
        break;

      case 34:
        printf("Enter number of filemarks to space backward: ");
        gets(buf);
        printf("\n");
        count = atoi(buf);
        rc = stioctop(STRSF,count);
        break;

      case 35:
        printf("Enter number of records to space backward: ");
        gets(buf);
        printf("\n");
        count = atoi(buf);
        rc = stioctop(STRSR,count);
        break;

      case 36:
        rc = stiocqrypos();
        if (rc != 0)
          break;
        printf("\nEnter position to set (1=Logical, 2=Physical, <enter>=Quit): ");
        gets(buf);
        if (strlen(buf) == 0)
          break;
        option = atoi(buf);
        printf("\nEnter Block ID in decimal or hex (Example: 16 or x10): ");
        gets(buf);
        if (strlen(buf) == 0)
          break;
        if (buf[0] == 'x')
          stpos.curpos = (int)strtol(buf+1,NULL,16);
        else
          stpos.curpos = atoi(buf);
        if (option == 1)
          stpos.block_type = QP_LOGICAL;
        else
          stpos.block_type = QP_PHYSICAL;
        rc = stiocsetpos();
        if (rc == 0)
          rc = stiocqrypos();
        break;

      case 37:
        printf ("\nSelect (1=Read file from tape, 2=Write file to tape): ");
        gets (buf);
        option = atoi(buf);
        if (option == 1)
          {
          printf ("Enter Destination File: ");
          gets (fname);
          if (strlen(fname) == 0)
            break;
          printf ("Enter number of records to read or <enter> for entire file: ");
          gets (buf);
          if (strlen(buf) != 0)
            {
            count = atoi(buf);
            if (count == 0)
              {
              printf ("Invalid number entered");
              break;
              }
            }
          else
            count = 0;
          rc = tread(fname,count);
          }
        else if (option == 2)
          {
          printf ("Enter Source File: ");
          gets (buf);
          if (strlen(buf) == 0)
            break;
          rc = twrite(buf);
          }
        break;

      case 38:
        rc = stioc_get_encryption_state();
        break;

      case 39:
        rc = stioc_report_density_support();
        break;

      case 40:
        encryption_test();
        break;
      }
    }
  }

/*---------------------------------------------------------------------------*/
/* Name: help                                                                */
/*---------------------------------------------------------------------------*/
void help()
  {
  printf("\n");
  printf("Usage: tapeutil [-f Device Subcommand [Subcommand ...]]\n");
  printf("\n");
  printf(" General Subcommands:\n");
  printf("   devinfo       inquiry [Page]   print \"Text\"   \n");
  printf("   reserve       release          reqsense         \n");
  printf("   reset         logpage \"Page\"   modepage \"Page\"\n");
  printf("   qrypath       resetpath        disablepath \"Primary | Alternate | Number\"\n");
  printf("   path          checkpath        enablepath \"Primary | Alternate | Number\"\n");
  printf("   tur           vpd              fuser\n");
  printf("   passthru      loop [Count]     sleep \"Seconds\"  \n");
  printf("   kill                                              \n");
  printf("\n");
  printf(" Medium Changer Subcommands:\n");
  printf("   allow                     prevent          audit [Address [Count]]     \n");
  printf("   devids                    mount [Slot]     position \"Destination\"\n");
  printf("   elementinfo               unmount [Slot]   move \"Source\" \"Destination\"\n");
  printf("   inventory [-i | -v Volid]                  exchange \"Source\" \"Dest1\" \"Dest2\"\n");
  printf("\n");
  printf(" Tape Subcommands:\n");
  printf("   append      bsf [Count]     bsr [Count] \n");
  printf("   autoload    eof [Count]     weof [Count]\n");
  printf("   noautoload  eofimm [Count]  weofimm [Count]\n");
  printf("   compress    fsf [Count]     fsr [Count]   \n");
  printf("   nocompress  erg             logsense      \n");
  printf("   load        erase           display \"Message\"\n");
  printf("   mtdevice    eraseimm        read -d Destination [-c Count]\n");
  printf("   qrypos      retension       write -s Source \n");
  printf("   seod        status          rtest  [-b Blocksize] [-c Count] [-r Repetition]\n");
  printf("   offline     parms           wtest  [-b Blocksize] [-c Count] [-r Repetition]\n");
  printf("   rewoffl     sync            rwtest [-b Blocksize] [-c Count] [-r Repetition]\n");
  printf("   unload      volid \"Name\"    setpos [Blockid]          \n");
  printf("   list        sdp \"Number\"    chgpart \"Number\" [Blockid] \n");
  printf("   density     idp             qrypart    \n");
  printf("   sili        prevent         allow      \n");
  printf("   nosili      rewind          encryption \n");
  printf("\n");
  printf(" Service Aid Subcommands:\n");
  printf("   dump [Filename]   fmrtape   resetdrive   ucode \"Name\"\n");
  printf("   trcstop           trcstart [-q | [-d Directory] [-n Files] [-s Size] [-D]]  \n");
  printf("   ekmtest           forcedump    readdump [Filename] \n");
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_inquiry                                                 */
/*---------------------------------------------------------------------------*/
int sioc_inquiry()
  {
  struct inquiry_data inquiry_data;

  printf("Issuing inquiry...\n");

  if (!ioctl (fd, SIOC_INQUIRY, &inquiry_data))
    dump_bytes ((uchar *)&inquiry_data, inquiry_data.len+5, "Inquiry Data");

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_log_sense                                              */
/*---------------------------------------------------------------------------*/
int stioc_log_sense()
  {
  struct log_sense logp;
  struct tm *time;
  char work[25];
  int page_cnt, i;
  struct log_page_header *page_header;
  char *data;
  time_t when;

  printf("Issuing log sense...\n");

  if (ioctl (fd, STIOC_LOG_SENSE, &logp) < 0)
    return errno;

  when = logp.header.when;
  time = localtime(&when);
  strftime(work, 20,"%m/%d/%y %H:%M:%S", time);
  printf("\n%s  Device: %0.8s  S/N: %0.12s  Volid: %0.16s\n", work,
         logp.header.device_type,logp.header.serial,logp.header.volid);

  /* format the log sense pages */
  page_header = (struct log_page_header *)logp.data;

  /* page length from first page defines how many pages are following */
  page_cnt = page_header->len;
  for (i = 0; i < page_cnt; i++)
    {
    /* skip a page with no data */
    if (page_header->len == 0)
      continue;
    sprintf(work,"Log Sense Page 0x%02X",page_header->code);
    data = (char *)page_header;
    dump_bytes(data,page_header->len + sizeof(struct log_page_header),work);
    if (opcode && (i+1 < page_cnt))
      {
      printf ("\nHit <enter> to continue...");
      getchar();
      }

    /* increment to next page header */
    page_header = (struct log_page_header *)(data +
                  sizeof(struct log_page_header) + page_header->len);
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_reserve                                                 */
/*---------------------------------------------------------------------------*/
int sioc_reserve()
  {
  printf("Reserving device...\n");

  ioctl (fd, SIOC_RESERVE, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_release                                                 */
/*---------------------------------------------------------------------------*/
int sioc_release()
  {
  printf("Releasing device...\n");

  ioctl (fd, SIOC_RELEASE, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_test_unit_ready                                         */
/*---------------------------------------------------------------------------*/
int sioc_test_unit_ready()
  {
  printf("Issuing test unit ready...\n");

  ioctl (fd, SIOC_TEST_UNIT_READY, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_reqsense                                                */
/*---------------------------------------------------------------------------*/
int sioc_reqsense()
  {
  struct request_sense sense_data;

  printf("Issuing request sense...\n");

  if (!ioctl (fd, SIOC_REQSENSE, &sense_data))
    dump_bytes ((uchar *)&sense_data, sense_data.addlen+8,
                "Sense Data");

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_log_sense_page                                          */
/*---------------------------------------------------------------------------*/
int sioc_log_sense_page(int page, int len, int parm)
  {
  struct log_sense_page logdata;
  struct log_page_header *page_header;
  char text[80];
  int count;

  logdata.page_code = page;
  logdata.len = len;
  logdata.parm_pointer = parm;
  page_header = (struct log_page_header *)logdata.data;

  printf("Issuing log sense for page 0x%02X...\n",page);

  if (!ioctl (fd, SIOC_LOG_SENSE_PAGE, &logdata))
    {
    count = page_header->len+sizeof(struct log_page_header);
    sprintf(text,"Log Sense Page 0x%02X,  Page Length %d,  Data",page,count);
    if (count > LOGSENSEPAGE)
      count = LOGSENSEPAGE;
    dump_bytes(logdata.data, count, text);
    }
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_mode_sense_page                                         */
/*---------------------------------------------------------------------------*/
int sioc_mode_sense_page(int page)
  {
  struct mode_sense_page modedata;
  char text[80];

  modedata.page_code = page;

  printf("Issuing mode sense for page 0x%02X...\n",page);

  if (!ioctl (fd, SIOC_MODE_SENSE_PAGE, &modedata))
    {
    sprintf(text,"Mode Sense Page 0x%02X",page);
    dump_bytes(modedata.data, modedata.data[1]+2, text);
    }
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_inquiry_page                                            */
/*---------------------------------------------------------------------------*/
int sioc_inquiry_page(int page)
  {
  struct inquiry_page inqdata;
  char text[80];

  inqdata.page_code = page;

  printf("Issuing inquiry for page 0x%02X...\n",page);

  if (!ioctl (fd, SIOC_INQUIRY_PAGE, &inqdata))
    {
    sprintf(text,"Inquiry Page 0x%02X",page);
    dump_bytes(inqdata.data, inqdata.data[3]+4, text);
    }
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_element_info                                          */
/*---------------------------------------------------------------------------*/
int smcioc_element_info()
  {
  struct element_info element_info;

  printf("Getting element information...\n");

  if (!ioctl (fd, SMCIOC_ELEMENT_INFO, &element_info))
    {
    printf("\n");
    printf("  Number of Robots .............. %d\n",element_info.robots);
    if (element_info.robots)
      printf("  First Robot Address ........... %d\n",element_info.robot_addr);

    printf("  Number of Slots ............... %d\n",element_info.slots);
    if (element_info.slots)
      printf("  First Slot Address ............ %d\n",element_info.slot_addr);

    printf("  Number of Import/Exports ...... %d\n",element_info.ie_stations);
    if (element_info.ie_stations)
      printf("  First Import/Export Address ... %d\n",element_info.ie_addr);

    printf("  Number of Drives .............. %d\n",element_info.drives);
    if (element_info.drives)
      printf("  First Drive Address ........... %d\n",element_info.drive_addr);
    printf("\n");
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_move_medium                                           */
/*---------------------------------------------------------------------------*/
int smcioc_move_medium (ushort source, ushort dest)
  {
  struct move_medium move_medium;

  move_medium.robot = 0;
  move_medium.invert = 0;
  move_medium.source = source;
  move_medium.destination = dest;

  printf("Moving medium: source %u, destination %u...\n",source,dest);

  ioctl (fd, SMCIOC_MOVE_MEDIUM, &move_medium);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_exchange_medium                                       */
/*---------------------------------------------------------------------------*/
int smcioc_exchange_medium (ushort source, ushort dest, ushort dest2)
  {
  struct exchange_medium exchange_medium;

  exchange_medium.robot = 0;
  exchange_medium.invert1 = 0;
  exchange_medium.invert2 = 0;
  exchange_medium.source = source;
  exchange_medium.destination1 = dest;
  exchange_medium.destination2 = dest2;

  printf("Exchanging medium: source %u, first destination %u, second destination %u...\n",
          source,dest,dest2);

  ioctl (fd, SMCIOC_EXCHANGE_MEDIUM, &exchange_medium);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_load_medium                                           */
/*---------------------------------------------------------------------------*/
int smcioc_load_medium (ushort source)
  {

  if (source)
    {
    if (source == LOAD_UNLOAD_SLOT_0)
      printf("Loading medium from slot 0...\n");
    else
      printf("Loading medium from slot %u...\n",source);
    }
  else
    printf("Loading medium from first full slot...\n");

  ioctl (fd, SMCIOC_LOAD_MEDIUM, source);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_unload_medium                                         */
/*---------------------------------------------------------------------------*/
int smcioc_unload_medium (ushort source)
  {

  if (source)
    {
    if (source == LOAD_UNLOAD_SLOT_0)
      printf("Unloading medium to slot 0...\n");
    else
      printf("Unloading medium to slot %u...\n",source);
    }
  else
    printf("Unloading medium to first empty slot...\n");

  ioctl (fd, SMCIOC_UNLOAD_MEDIUM, source);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_pos_to_elem                                           */
/*---------------------------------------------------------------------------*/
int smcioc_pos_to_elem (ushort dest)
  {
  struct pos_to_elem pos_to_elem;

  pos_to_elem.robot = 0;
  pos_to_elem.invert = 0;
  pos_to_elem.destination = dest;

  printf("Positioning to element, destination %u...\n",dest);

  ioctl (fd, SMCIOC_POS_TO_ELEM, &pos_to_elem);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_init_elem_stat                                        */
/*---------------------------------------------------------------------------*/
int smcioc_init_elem_stat()
  {

  printf("Initializing element status...\n");

  ioctl (fd, SMCIOC_INIT_ELEM_STAT, 0);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_init_elem_stat_range                                  */
/*---------------------------------------------------------------------------*/
int smcioc_init_elem_stat_range(ushort address, ushort number)
  {
  struct element_range elem_range;

  elem_range.element_address = address;
  elem_range.number_elements = number;

  printf("Initializing element status for address %d, %d elements...\n",address,number);

  ioctl (fd, SMCIOC_INIT_ELEM_STAT_RANGE, &elem_range);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_prevent_medium_removal                                */
/*---------------------------------------------------------------------------*/
int smcioc_prevent_medium_removal()
  {

  printf("Issuing prevent medium removal...\n");

  ioctl (fd, SMCIOC_PREVENT_MEDIUM_REMOVAL, 0);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_allow_medium_removal                                  */
/*---------------------------------------------------------------------------*/
int smcioc_allow_medium_removal()
  {

  printf("Issuing allow medium removal...\n");

  ioctl (fd, SMCIOC_ALLOW_MEDIUM_REMOVAL, 0);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_inventory                                             */
/*---------------------------------------------------------------------------*/
int smcioc_inventory(char *volid, int ie_only)
  {
  int i,j;
  struct element_status *robot_status = NULL;
  struct element_status *slot_status = NULL;
  struct element_status *ie_status = NULL;
  struct element_status *drive_status = NULL;
  struct element_status *elemp;
  struct inventory inventory;
  struct element_info element_info;
  int found_volid = FALSE;

  if (ioctl(fd, SMCIOC_ELEMENT_INFO, &element_info))
    return errno;

  if (ie_only)
    {
    element_info.robots = 0;
    element_info.slots = 0;
    element_info.drives = 0;
    }

  if (element_info.robots)
    {
    robot_status = malloc(element_info.robots * sizeof(struct element_status));
    if (robot_status == NULL)
      {
      errno = ENOMEM;
      return errno;
      }
    bzero((caddr_t)robot_status,element_info.robots * sizeof(struct element_status));
    }

  if (element_info.slots)
    {
    slot_status = malloc(element_info.slots * sizeof(struct element_status));
    if (slot_status == NULL)
      {
      if (robot_status)
        free(robot_status);
      errno = ENOMEM;
      return errno;
      }
    bzero((caddr_t)slot_status,element_info.slots * sizeof(struct element_status));
    }

  if (element_info.drives)
    {
    drive_status = malloc(element_info.drives * sizeof(struct element_status));
    if (drive_status == NULL)
      {
      if (robot_status)
        free(robot_status);
      if (slot_status)
        free(slot_status);
      errno = ENOMEM;
      return errno;
      }
    bzero((caddr_t)drive_status,element_info.drives * sizeof(struct element_status));
    }

  if (element_info.ie_stations)
    {
    ie_status = malloc(element_info.ie_stations * sizeof(struct element_status));
    if (ie_status == NULL)
      {
      if (robot_status)
        free(robot_status);
      if (slot_status)
        free(slot_status);
      if (drive_status)
        free(drive_status);
      errno = ENOMEM;
      return errno;
      }
    bzero((caddr_t)ie_status,element_info.ie_stations * sizeof(struct element_status));
    }

  inventory.robot_status = robot_status;
  inventory.slot_status = slot_status;
  inventory.ie_status = ie_status;
  inventory.drive_status = drive_status;

  if (volid != NULL)
    printf("Reading element status for Volume %s...\n",volid);
  else
    printf("Reading element status...\n");

  if (!ioctl (fd, SMCIOC_INVENTORY, &inventory))
    {
    j=0;
    elemp = robot_status;
    for (i = 0; (i < element_info.robots && !found_volid); i++, elemp++)
      {
      if (opcode && j == 2)
        {
        j=0;
        printf ("\nHit <enter> to continue...");
        getchar();
        }

      if (volid != NULL)
        {
        if (strncmp(volid,elemp->volume,strlen(volid)) == 0)
          found_volid = TRUE;
        else
          continue;
        }


      j++;
      printf("\nRobot Address %d\n",elemp->address);
      if (elemp->except)
        printf("  Robot State .................... Abnormal\n");
      else
        printf("  Robot State .................... Normal\n");
      printf("  ASC/ASCQ ....................... %02X%02X\n",
             elemp->asc,elemp->ascq);
      if (elemp->full)
        printf("  Media Present .................. Yes\n");
      else
        printf("  Media Present .................. No\n");
      if (elemp->svalid)
        printf("  Source Element Address ......... %d\n",elemp->source);
      else
        printf("  Source Element Address Valid ... No\n");
      if (elemp->invert)
        printf("  Media Inverted ................. Yes\n");
      else
        printf("  Media Inverted ................. No\n");
      printf("  Volume Tag ..................... %0.36s\n", elemp->volume);
      }

    elemp = ie_status;
    for (i = 0; (i < element_info.ie_stations && !found_volid); i++, elemp++)
      {
      if (opcode && j == 2)
        {
        j=0;
        printf ("\nHit <enter> to continue...");
        getchar();
        }

      if (volid != NULL)
        {
        if (strncmp(volid,elemp->volume,strlen(volid)) == 0)
          found_volid = TRUE;
        else
          continue;
        }

      j++;
      printf("\nImport/Export Station Address %d\n",elemp->address);
      if (elemp->except)
        printf("  Import/Export State ............ Abnormal\n");
      else
        printf("  Import/Export State ............ Normal\n");
      printf("  ASC/ASCQ ....................... %02X%02X\n",
             elemp->asc,elemp->ascq);
      if (elemp->full)
        {
        printf("  Media Present .................. Yes\n");
        if (elemp->impexp)
          printf("  Media Placed by Operator ....... Yes\n");
        else
          printf("  Media Placed by Operator ....... No\n");
        }
      else
        printf("  Media Present .................. No\n");
      if (elemp->inenab)
        printf("  Import Enabled ................. Yes\n");
      else
        printf("  Import Enabled ................. No\n");
      if (elemp->exenab)
        printf("  Export Enabled ................. Yes\n");
      else
        printf("  Export Enabled ................. No\n");
      if (elemp->access)
        printf("  Robot Access Allowed ........... Yes\n");
      else
        printf("  Robot Access Allowed ........... No\n");
      if (elemp->svalid)
        printf("  Source Element Address ......... %d\n",elemp->source);
      else
        printf("  Source Element Address Valid ... No\n");
      if (elemp->invert)
        printf("  Media Inverted ................. Yes\n");
      else
        printf("  Media Inverted ................. No\n");
      printf("  Volume Tag ..................... %0.36s\n", elemp->volume);
      }

    elemp = drive_status;
    for (i = 0; (i < element_info.drives && !found_volid); i++, elemp++)
      {
      /* don't overflow screen if menu mode */
      if (opcode && j == 2)
        {
        j=0;
        printf ("\nHit <enter> to continue...");
        getchar();
        }

      if (volid != NULL)
        {
        if (strncmp(volid,elemp->volume,strlen(volid)) == 0)
          found_volid = TRUE;
        else
          continue;
        }

      j++;
      printf("\nDrive Address %d\n",elemp->address);
      if (elemp->except)
        printf("  Drive State .................... Abnormal\n");
      else
        printf("  Drive State .................... Normal\n");
      printf("  ASC/ASCQ ....................... %02X%02X\n",
             elemp->asc,elemp->ascq);
      if (elemp->full)
        printf("  Media Present .................. Yes\n");
      else
        printf("  Media Present .................. No\n");
      if (elemp->access)
        printf("  Robot Access Allowed ........... Yes\n");
      else
        printf("  Robot Access Allowed ........... No\n");
      if (elemp->svalid)
        printf("  Source Element Address ......... %d\n",elemp->source);
      else
        printf("  Source Element Address Valid ... No\n");
      if (elemp->invert)
        printf("  Media Inverted ................. Yes\n");
      else
        printf("  Media Inverted ................. No\n");
      if (elemp->notbus)
        printf("  Same Bus as Medium Changer ..... No\n");
      else
        printf("  Same Bus as Medium Changer ..... Yes\n");
      if (elemp->idvalid)
        printf("  SCSI Bus Address ............... %d\n",elemp->scsi);
      else
        printf("  SCSI Bus Address Vaild ......... No\n");
      if (elemp->luvalid)
        printf("  Logical Unit Number ............ %d\n",elemp->lun);
      else
        printf("  Logical Unit Number Valid ...... No\n");
      printf("  Volume Tag ..................... %0.36s\n", elemp->volume);
      }

    elemp = slot_status;
    for (i = 0; (i < element_info.slots && !found_volid); i++, elemp++)
      {
      if (opcode && j == 2)
        {
        j=0;
        printf ("\nHit <enter> to continue...");
        getchar();
        }

      if (volid != NULL)
        {
        if (strncmp(volid,elemp->volume,strlen(volid)) == 0)
          found_volid = TRUE;
        else
          continue;
        }

      j++;
      printf("\nSlot Address %d\n",elemp->address);
      if (elemp->except)
        printf("  Slot State ..................... Abnormal\n");
      else
        printf("  Slot State ..................... Normal\n");
      printf("  ASC/ASCQ ....................... %02X%02X\n",
             elemp->asc,elemp->ascq);
      if (elemp->full)
        printf("  Media Present .................. Yes\n");
      else
        printf("  Media Present .................. No\n");
      if (elemp->access)
        printf("  Robot Access Allowed ........... Yes\n");
      else
        printf("  Robot Access Allowed ........... No\n");
      if (elemp->svalid)
        printf("  Source Element Address ......... %d\n",elemp->source);
      else
        printf("  Source Element Address Valid ... No\n");
      if (elemp->invert)
        printf("  Media Inverted ................. Yes\n");
      else
        printf("  Media Inverted ................. No\n");
      printf("  Volume Tag ..................... %0.36s\n", elemp->volume);
      }
    }

  if (robot_status)
    free(robot_status);
  if (slot_status)
    free(slot_status);
  if (drive_status)
    free(drive_status);
  if (ie_status)
    free(ie_status);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        smcioc_read_element_devids                                   */
/*---------------------------------------------------------------------------*/
int smcioc_read_element_devids()
  {
  int i,j;
  struct element_devid *elem_devid, *elemp;
  struct read_element_devids devids;
  struct element_info element_info;

  if (ioctl(fd, SMCIOC_ELEMENT_INFO, &element_info))
    return errno;

  if (element_info.drives)
    {
    elem_devid = malloc(element_info.drives * sizeof(struct element_devid));
    if (elem_devid == NULL)
      {
      errno = ENOMEM;
      return errno;
      }
    bzero((caddr_t)elem_devid,element_info.drives * sizeof(struct element_devid));

    devids.drive_devid = elem_devid;
    devids.element_address = element_info.drive_addr;
    devids.number_elements = element_info.drives;

    printf("Reading element device ids...\n");

    if (ioctl (fd, SMCIOC_READ_ELEMENT_DEVIDS, &devids))
      {
      free(elem_devid);
      return errno;
      }

    j=0;
    elemp = elem_devid;
    for (i = 0; i < element_info.drives; i++, elemp++)
      {
      /* don't overflow screen if menu mode */
      if (opcode && j == 2)
        {
        j=0;
        printf ("\nHit <enter> to continue...");
        getchar();
        }

      j++;
      printf("\nDrive Address %d\n",elemp->address);
      if (elemp->except)
        printf("  Drive State .................... Abnormal\n");
      else
        printf("  Drive State .................... Normal\n");
      if (elemp->asc == 0x81 && elemp->ascq ==0x00)
        printf("  ASC/ASCQ ....................... %02X%02X (Drive Present)\n",
             elemp->asc,elemp->ascq);
      else if (elemp->asc == 0x82 && elemp->ascq ==0x00)
        printf("  ASC/ASCQ ....................... %02X%02X (Drive Not Present)\n",
             elemp->asc,elemp->ascq);
      else
        printf("  ASC/ASCQ ....................... %02X%02X\n",
             elemp->asc,elemp->ascq);
      if (elemp->full)
        printf("  Media Present .................. Yes\n");
      else
        printf("  Media Present .................. No\n");
      if (elemp->access)
        printf("  Robot Access Allowed ........... Yes\n");
      else
        printf("  Robot Access Allowed ........... No\n");
      if (elemp->svalid)
        printf("  Source Element Address ......... %d\n",elemp->source);
      else
        printf("  Source Element Address Valid ... No\n");
      if (elemp->invert)
        printf("  Media Inverted ................. Yes\n");
      else
        printf("  Media Inverted ................. No\n");
      if (elemp->notbus)
        printf("  Same Bus as Medium Changer ..... No\n");
      else
        printf("  Same Bus as Medium Changer ..... Yes\n");
      if (elemp->idvalid)
        printf("  SCSI Bus Address ............... %d\n",elemp->scsi);
      else
        printf("  SCSI Bus Address Vaild ......... No\n");
      if (elemp->luvalid)
        printf("  Logical Unit Number ............ %d\n",elemp->lun);
      else
        printf("  Logical Unit Number Valid ...... No\n");
      printf("  Device ID ...................... %0.36s\n", elemp->identifier);
      }
    }
  else
    {
    printf("\nNo drives found in element information\n");
    if (opcode)
      {
      printf ("\nHit <enter> to continue...");
      getchar();
      }
    }

  free(elem_devid);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_prevent_medium_removal                                 */
/*---------------------------------------------------------------------------*/
int stioc_prevent_medium_removal()
  {

  printf("Issuing prevent medium removal...\n");

  ioctl (fd, STIOC_PREVENT_MEDIUM_REMOVAL, 0);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_allow_medium_removal                                   */
/*---------------------------------------------------------------------------*/
int stioc_allow_medium_removal()
  {

  printf("Issuing allow medium removal...\n");

  ioctl (fd, STIOC_ALLOW_MEDIUM_REMOVAL, 0);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_load_ucode                                             */
/*---------------------------------------------------------------------------*/
void stioc_load_ucode(char *ucode)
  {
  char cmd[132];

  printf("Downloading microcode...\n");

  if (ioctl (fd, STIOC_LOAD_UCODE, ucode) == 0)
    {
    close(fd);
    /* check if it's a SAN Gateway  */
    if (strstr(devname, ".sdg") == NULL)
      {
      printf("Updating VPD data...\n");
      sprintf(cmd,"/bin/cfgAtape -V -l %s",devname+5);
      errno=system(cmd);
      if (errno)
        errno = errno >> 8;
      }
    }

  if (errno)
    {
    sprintf(cmd,"Operation failed with errno %d",errno);
    perror (cmd);
    }

  exit(errno);
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_dump                                                   */
/*---------------------------------------------------------------------------*/
int stioc_dump(char *fname)
  {
  if (fname == NULL)
    printf("Dumping device to /var/adm/ras directory...\n");
  else
    printf("Dumping device to %s...\n",fname);

  ioctl (fd, STIOC_DUMP, fname);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_read_dump                                              */
/*---------------------------------------------------------------------------*/
int stioc_read_dump(char *fname)
  {
  if (fname == NULL)
    printf("Reading dump from device to /var/adm/ras directory...\n");
  else
    printf("Reading dump from device to %s...\n",fname);

  ioctl (fd, STIOC_READ_DUMP, fname);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_force_dump                                             */
/*---------------------------------------------------------------------------*/
int stioc_force_dump()
  {
  printf("Forcing drive dump...\n");

  ioctl (fd, STIOC_FORCE_DUMP, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_reset_drive                                            */
/*---------------------------------------------------------------------------*/
int stioc_reset_drive()
  {
  printf("Resetting drive...\n");

  ioctl (fd, STIOC_RESET_DRIVE, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_fmr_tape                                               */
/*---------------------------------------------------------------------------*/
int stioc_fmr_tape()
  {
  printf("Creating FMR tape...\n");

  ioctl (fd, STIOC_FMR_TAPE, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        stioc_set_volid                                              */
/*---------------------------------------------------------------------------*/
int stioc_set_volid(char *volid)
  {
  printf("Setting volume id to %0.16s\n",volid);

  ioctl (fd, STIOC_SET_VOLID, volid);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioctop                                                            */
/*---------------------------------------------------------------------------*/
int stioctop(int stop, int count)
  {
  struct stop st;

  st.st_op = stop;
  st.st_count = count;

  switch (stop)
    {
    case STREW:
      printf("Rewinding tape...\n");
      break;

    case STINSRT:
      printf("Loading tape...\n");
      break;

    case STOFFL:
      printf("Unloading tape...\n");
      break;

    case STERASE:
      printf("Erasing tape...\n");
      break;

    case STERASE_IMM:
      printf("Erasing tape with immediate bit...\n");
      break;

    case STERASEGAP:
      printf("Erasing record gap...\n");
      break;

    case STWEOF:
      if (count == 1)
        printf("Writing filemark...\n");
      else
        printf("Writing %d filemarks...\n",count);
      break;

    case STWEOF_IMM:
      if (count == 1)
        printf("Writing filemark with immediate bit...\n");
      else
        printf("Writing %d filemarks with immediate bit...\n",count);
      break;

    case STFSF:
      if (count == 1)
        printf("Forward spacing filemark...\n");
      else
        printf("Forward spacing %d filemarks...\n",count);
      break;

    case STRSF:
      if (count == 1)
        printf("Backward spacing filemark...\n");
      else
        printf("Backward spacing %d filemarks...\n",count);
      break;

    case STFSR:
      if (count == 1)
        printf("Forward spacing record...\n");
      else
        printf("Forward spacing %d records...\n",count);
      break;

    case STRSR:
      if (count == 1)
        printf("Backward spacing record...\n");
      else
        printf("Backward spacing %d records...\n",count);
      break;

    case STSEOD:
      printf("Spacing to end of data...\n");
      break;
    }

  ioctl (fd, STIOCTOP, &st);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: iocinfo                                                             */
/*---------------------------------------------------------------------------*/
int iocinfo()
  {
  struct devinfo devinfo;

  bzero((caddr_t)&devinfo,sizeof(struct devinfo));

  printf("Issuing iocinfo...\n");

  if (!ioctl (fd, IOCINFO, &devinfo))
    {
    if (devinfo.devtype == DD_SCTAPE)
      {
      printf("devtype = SCSI Tape\n");
      printf("devsubtype = %02X\n",devinfo.devsubtype);
      if (devinfo.un.scmt.type == DT_STREAM)
        printf("tapetype = Streaming Tape Drive\n");
      else
        printf("tapetype = %c\n",devinfo.un.scmt.type);
      printf("block size = %d\n",devinfo.un.scmt.blksize);
      }
    else if (devinfo.devtype == DD_MEDIUM_CHANGER)
      {
      printf("devtype = SCSI Medium Changer\n");
      printf("devsubtype = %02X\n",devinfo.devsubtype);
      }
    else
      {
      printf("devtype = %c\n",devinfo.devtype);
      if (devinfo.devsubtype == 'S')
        {
        printf("devsubtype = SCSI Adapter\n");
        printf("max transfer size = %d\n",devinfo.un.scsi.max_transfer);
        }
      else
        printf("devsubtype = %c\n",devinfo.devsubtype);
      }
    printf("\n");
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: mtdevice                                                            */
/*---------------------------------------------------------------------------*/
int mtdevice()
  {
  int seq_no;

  printf("Issuing mtdevice...\n");

  if (!ioctl(fd, MTDEVICE, &seq_no))
    printf("\n  Library Sequence Number...... %08X\n",seq_no);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocqrysense                                                       */
/*---------------------------------------------------------------------------*/
int stiocqrysense()
  {
  char msg[80];
  struct stsense_s stsense;

  stsense.sense_type = LASTERROR;
  if (ioctl(fd, STIOCQRYSENSE, &stsense) < 0)
    return errno;

  if (stsense.residual_count > 0)
    printf("Residual count: %d\n", stsense.residual_count);

  if (stsense.len > 0)
    dump_bytes(stsense.sense,stsense.len,"Error Sense Data");

  return 0;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocqryinquiry                                                     */
/*---------------------------------------------------------------------------*/
int stiocqryinquiry()
  {
  struct st_inquiry inqdata;

  printf("Issuing inquiry...\n");

  if (!ioctl(fd, STIOCQRYINQUIRY, &inqdata))
    dump_bytes((char *)&inqdata,sizeof(struct st_inquiry),"Inquiry Data");

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocsync                                                           */
/*---------------------------------------------------------------------------*/
int stiocsync(void)
  {
  printf("Synchronizing buffers...\n");

  ioctl(fd, STIOCSYNC, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocqryp                                                           */
/*---------------------------------------------------------------------------*/
int stiocqryp(void)
  {
  printf("Querying tape parameters...\n");

  ioctl(fd, STIOCQRYP, &parms);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocsetp                                                           */
/*---------------------------------------------------------------------------*/
int stiocsetp(void)
  {
  printf("Setting tape parameters...\n");

  ioctl(fd, STIOCSETP, &parms);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocdm                                                             */
/*---------------------------------------------------------------------------*/
int stiocdm(char type, char *msg_0, char *msg_1)
  {
  struct stdm_s msg_data;
  char msg_type[] = { DMMSG0, DMMSG1, DMFLASHMSG0, DMFLASHMSG1, DMALTERNATE };
  int rc;

  strncpy (msg_data.dm_msg0, msg_0,8);
  if (msg_1 != NULL)
    {
    strncpy (msg_data.dm_msg1, msg_1,8);
    printf("Displaying message, msg0: %0.8s msg1: %s\n",msg_0, msg_1);
    }
  else
    printf("Displaying message, msg0: %0.8s\n",msg_0);

  /* Display demount message if tape is loaded */
  msg_data.dm_func = msg_type[type-1] | DMDVMSG;
  rc = ioctl (fd, STIOCDM, &msg_data);

  if (rc == 0)
    {
    /* Display mount message if tape is unloaded */
    msg_data.dm_func = msg_type[type-1] | DMMIMMED;
    ioctl (fd, STIOCDM, &msg_data);
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        tape_parms                                                   */
/*---------------------------------------------------------------------------*/
int tape_parms(void)
  {
  int rc = 0;
  char   buf[80] = "";
  int option;
  int work;
  int abort;

  while(1)
    {
    if ((rc = stiocqryp()) != 0)
      break;

    printf("\n Nonchangeable parameters:\n\n");
    if (parms.trace)
      printf("   Trace.................... On     ");
    else
      printf("   Trace.................... Off    ");

    printf("   Medium Type............ %02X\n",parms.medium_type);
    printf("   Trace Hookword........... %x    ",parms.hkwrd);
    printf("   Density Code........... %02X\n",parms.density_code);
    printf("   Minimum Block Size....... %d\n",parms.min_blksize);
    printf("   Maximum Block Size....... %d\n",parms.max_blksize);
    printf("   Maximum SCSI Xfer........ %d\n",parms.max_scsi_xfer);
    if (parms.write_protect)
      printf("   Physical Write Protect... Yes\n");
    else
      printf("   Physical Write Protect... No\n");

    if (parms.acf_mode == ACF_NONE)
      printf("   ACF Mode................. N/A\n");
    else if (parms.acf_mode == ACF_SYSTEM)
      printf("   ACF Mode................. SYSTEM\n");
    else if (parms.acf_mode == ACF_MANUAL)
      printf("   ACF Mode................. MANUAL\n");
    else if (parms.acf_mode == ACF_AUTOMATIC)
      printf("   ACF Mode................. AUTOMATIC\n");
    else if (parms.acf_mode == ACF_RANDOM)
      printf("   ACF Mode................. RANDOM\n");
    else if (parms.acf_mode == ACF_ACCUMULATE)
      printf("   ACF Mode................. ACCUMULATE\n");
    if (parms.retain_reservation)
      printf("   Retain Reservation....... Yes\n");
    else
      printf("   Retain Reservation....... No\n");
    if (parms.alt_pathing)
      printf("   Alternate Pathing........ Yes\n");
    else
      printf("   Alternate Pathing........ No\n");

    printf("\n Changeable parameters (* effective during open, only):\n\n");

    if (parms.autoload)
      printf("  *1. Autoload.............. Yes");
    else
      printf("  *1. Autoload.............. No ");

    if (parms.read_sili_bit)
      printf("  *10. Read SILI Bit.......... Yes\n");
    else
      printf("  *10. Read SILI Bit.......... No \n");

    if (parms.compression)
      printf("  *2. Compression........... Yes");
    else
      printf("  *2. Compression........... No ");

    printf("   11. Volume ID.............. %0.16s\n",parms.volid);

    if (parms.buffered_mode)
      printf("  *3. Buffered Mode......... Yes");
    else
      printf("  *3. Buffered Mode......... No ");

    if (parms.capacity_scaling == SCALE_100)
      printf("   12. Capacity Scaling....... 100%c\n",'%');
    else if (parms.capacity_scaling == SCALE_75)
      printf("   12. Capacity Scaling....... 75%c\n",'%');
    else if (parms.capacity_scaling == SCALE_50)
      printf("   12. Capacity Scaling....... 50%c\n",'%');
    else if (parms.capacity_scaling == SCALE_25)
      printf("   12. Capacity Scaling....... 25%c\n",'%');
    else if (parms.capacity_scaling == SCALE_VALUE)
      printf("   12. Capacity Scaling....... x%02X\n",parms.capacity_scaling_value);

    if (parms.trailer_labels)
      printf("  *4. Trailer Labels........ Yes");
    else
      printf("  *4. Trailer Labels........ No ");

    if (parms.logical_write_protect == WORM_PROTECT)
      printf("   13. Logical Write Protect.. WORM\n");
    else if (parms.logical_write_protect == ASSOCIATED_PROTECT)
      printf("   13. Logical Write Protect.. ASSOCIATED\n");
    else if (parms.logical_write_protect == PERSISTENT_PROTECT)
      printf("   13. Logical Write Protect.. PERSISTENT\n");
    else
      printf("   13. Logical Write Protect.. No\n");

    if (parms.rewind_immediate)
      printf("  *5. Rewind Immediate...... Yes\n");
    else
      printf("  *5. Rewind Immediate...... No\n");

    printf("  *6. Block Size............ %d\n",parms.blksize);

    if (parms.record_space_mode == SCSI_SPACE_MODE)
      printf("  *7. Record Space Mode..... SCSI\n");
    else if (parms.record_space_mode == AIX_SPACE_MODE)
      printf("  *7. Record Space Mode..... AIX\n");
    else
      printf("  *7. Record Space Mode..... Unknown\n");

    if (parms.logging)
      printf("  *8. Volume Logging........ Yes\n");
    else
      printf("  *8. Volume Logging........ No\n");

    if (parms.emulate_autoloader)
      printf("  *9. Emulate Autoloader.... Yes\n");
    else
      printf("  *9. Emulate Autoloader.... No\n");

    printf("\nEnter parameter to change or <enter> to quit: ");
    gets(buf);
    if (strlen(buf) == 0)
      break;
    option = atoi(buf);
    abort = FALSE;
    switch (option)
      {
      case 1:
        if (parms.autoload)
          parms.autoload = FALSE;
        else
          parms.autoload = TRUE;
        break;

      case 2:
        if (parms.compression)
          parms.compression = FALSE;
        else
          parms.compression = TRUE;
        break;

      case 3:
        if (parms.buffered_mode)
          parms.buffered_mode = FALSE;
        else
          parms.buffered_mode = TRUE;
        break;

      case 4:
        if (parms.trailer_labels)
          parms.trailer_labels = FALSE;
        else
          parms.trailer_labels = TRUE;
        break;

      case 5:
        if (parms.rewind_immediate)
          parms.rewind_immediate = FALSE;
        else
          parms.rewind_immediate = TRUE;
        break;

      case 6:
        printf("\nEnter block size: ");
        gets(buf);
        parms.blksize = atoi(buf);
        break;

      case 7:
        printf("\nEnter record space mode (1=SCSI, 2=AIX): ");
        gets(buf);
        if (strlen(buf) == 0)
          {
          abort = TRUE;
          break;
          }
        parms.record_space_mode = atoi(buf);
        break;

      case 8:
        if (parms.logging)
          parms.logging = FALSE;
        else
          parms.logging = TRUE;
        break;

      case 9:
        if (parms.emulate_autoloader)
          parms.emulate_autoloader = FALSE;
        else
          parms.emulate_autoloader = TRUE;
        break;

      case 10:
        if (parms.read_sili_bit)
          parms.read_sili_bit = FALSE;
        else
          parms.read_sili_bit = TRUE;
        break;

      case 11:
        printf("\nEnter volume id: ");
        gets(buf);
        strncpy(parms.volid,buf,sizeof(parms.volid));
        break;

      case 12:
        printf("\nEnter capacity scaling (1=100%c, 2=75%c, 3=50%c, 4=25%c, 5=Hex Value): ",'%','%','%','%');
        gets(buf);
        if (atoi(buf) == 0)
          {
          abort = TRUE;
          break;
          }
        work = atoi(buf) - 1;
        parms.capacity_scaling = work;
        if (work == SCALE_VALUE)
         {
         printf("\nEnter hex value: ");
         gets(buf);
         if (strlen(buf) == 0)
           {
           abort = TRUE;
           break;
           }
          parms.capacity_scaling_value = (int)strtol(buf,NULL,16);
         }
        printf("\nWarning!  All data on tape will be lost with this change!\n");
        printf("Continue (y or n): ");
        gets(buf);
        if (strlen(buf) == 0 || buf[0] != 'y')
          {
          abort = TRUE;
          break;
          }
        break;

      case 13:
        if (parms.logical_write_protect == WORM_PROTECT)
          {
          printf("\nWORM protection can not be changed, hit any key to continue...\n");
          getchar();
          break;
          }
        printf("\nEnter write protect (1=NONE, 2=ASSOCIATED, 3=PERSISTENT, 4=WORM): ");
        gets(buf);
        if (atoi(buf) == 0)
          {
          abort = TRUE;
          break;
          }
        work = atoi(buf) - 1;
        if (work == WORM_PROTECT)
          {
          printf("\nWarning!  WORM protection is a permanent change to the tape!\n");
          printf("Continue (y or n): ");
          gets(buf);
          if (strlen(buf) == 0 || buf[0] != 'y')
            {
            abort = TRUE;
            break;
            }
          }
        parms.logical_write_protect = work;
        break;
      }
    if (abort)
      continue;
    if ((rc=stiocsetp()) != 0)
      break;
    }
  return rc;
  }

/*---------------------------------------------------------------------------*/
/* Name: status                                                              */
/*---------------------------------------------------------------------------*/
int status(void)
  {

  if (stiocqryp() != 0)
    return errno;

  printf ("\n");
  if (parms.trace)
    printf("   Trace.................... On\n");
  else
    printf("   Trace.................... Off\n");
  printf("   Trace Hookword........... %x\n",parms.hkwrd);
  printf("   Minimum Block Size....... %d\n",parms.min_blksize);
  printf("   Maximum Block Size....... %d\n",parms.max_blksize);
  printf("   Maximum SCSI Xfer........ %d\n",parms.max_scsi_xfer);
  if (parms.write_protect)
    printf("   Physical Write Protect... Yes\n");
  else
    printf("   Physical Write Protect... No\n");

  if (parms.logical_write_protect == ASSOCIATED_PROTECT)
    printf("   Logical Write Protect.... ASSOCIATED\n");
  else if (parms.logical_write_protect == PERSISTENT_PROTECT)
    printf("   Logical Write Protect.... PERSISTENT\n");
  else if (parms.logical_write_protect == WORM_PROTECT)
    printf("   Logical Write Protect.... WORM\n");
  else
    printf("   Logical Write Protect.... No\n");

  if (parms.autoload)
    printf("   Autoload................. Yes\n");
  else
    printf("   Autoload................. No\n");

  if (parms.compression)
    printf("   Compression.............. Yes\n");
  else
    printf("   Compression.............. No\n");

  if (parms.buffered_mode)
    printf("   Buffered Mode............ Yes\n");
  else
    printf("   Buffered Mode............ No\n");

  if (parms.trailer_labels)
    printf("   Trailer Labels........... Yes\n");
  else
    printf("   Trailer Labels........... No\n");

  if (parms.rewind_immediate)
    printf("   Rewind Immediate......... Yes\n");
  else
    printf("   Rewind Immediate......... No\n");

  printf("   Block Size............... %d\n",parms.blksize);

  if (parms.record_space_mode == SCSI_SPACE_MODE)
    printf("   Record Space Mode........ SCSI\n");
  else if (parms.record_space_mode == AIX_SPACE_MODE)
    printf("   Record Space Mode........ AIX\n");
  else
    printf("   Record Space Mode........ Unknown\n");

  if (parms.logging)
    printf("   Volume Logging........... Yes\n");
  else
    printf("   Volume Logging........... No\n");
  printf("   Volume ID................ %0.16s\n",parms.volid);

  if (parms.capacity_scaling == SCALE_100)
    printf("   Capacity Scaling......... 100%c\n",'%');
  else if (parms.capacity_scaling == SCALE_75)
    printf("   Capacity Scaling......... 75%c\n",'%');
  else if (parms.capacity_scaling == SCALE_50)
    printf("   Capacity Scaling......... 50%c\n",'%');
  else if (parms.capacity_scaling == SCALE_25)
    printf("   Capacity Scaling......... 25%c\n",'%');
  else if (parms.capacity_scaling == SCALE_VALUE)
    printf("   Capacity Scaling......... x%02X\n",parms.capacity_scaling_value);

  printf("   Medium Type.............. %02X\n",parms.medium_type);
  printf("   Density Code............. %02X\n",parms.density_code);

  if (parms.acf_mode == ACF_NONE)
    printf("   ACF Mode................. N/A\n");
  else if (parms.acf_mode == ACF_SYSTEM)
    printf("   ACF Mode................. SYSTEM\n");
  else if (parms.acf_mode == ACF_MANUAL)
    printf("   ACF Mode................. MANUAL\n");
  else if (parms.acf_mode == ACF_AUTOMATIC)
    printf("   ACF Mode................. AUTOMATIC\n");
  else if (parms.acf_mode == ACF_RANDOM)
    printf("   ACF Mode................. RANDOM\n");
  else if (parms.acf_mode == ACF_ACCUMULATE)
    printf("   ACF Mode................. ACCUMULATE\n");

  if (parms.retain_reservation)
    printf("   Retain Reservation....... Yes\n");
  else
    printf("   Retain Reservation....... No\n");

  if (parms.alt_pathing)
    printf("   Alternate Pathing........ Yes\n");
  else
    printf("   Alternate Pathing........ No\n");

  if (parms.emulate_autoloader)
    printf("   Emulate Autoloader....... Yes\n");
  else
    printf("   Emulate Autoloader....... No\n");

  if (parms.read_sili_bit)
    printf("   Read SILI bit............ Yes\n");
  else
    printf("   Read SILI bit............ No\n");

  return 0;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocqrypos                                                         */
/*---------------------------------------------------------------------------*/
int stiocqrypos(void)
  {
  char   buf[80] = "";
  int option;

  printf("Querying tape position....\n");

  stpos.block_type = QP_LOGICAL;
  if (ioctl (fd, STIOCQRYPOS, &stpos) <0)
   return errno;

  printf("\nLogical Tape position:\n\n");
  printf("    Partition Number ........... %d\n",stpos.partition_number);
  if (stpos.bot)
    printf("    Beginning of Partition ..... Yes\n");
  else
    printf("    Beginning of Partition ..... No\n");

  if (stpos.eot)
    printf("    Leading End of Partition ... Yes\n");
  else
    printf("    Leading End of Partition ... No\n");

    printf("    Current Block ID ........... %u (0x%08X)\n",
       stpos.curpos,stpos.curpos);
    printf("    Next Block on Tape ......... %u (0x%08X)\n",
         stpos.tapepos,stpos.tapepos);
  if (stpos.lbot == LBOT_NONE)
    printf("    Last Block Written ......... NONE\n");
  else if (stpos.lbot == LBOT_UNKNOWN)
    printf("    Last Block Written ......... UNKNOWN\n");
  else
    printf("    Last Block Written ......... %u (0x%08X)\n",
           stpos.lbot, stpos.lbot);

  if (stpos.bcu)
    printf("    Blocks in Buffer ........... UNKNOWN\n");
  else
    printf("    Blocks in Buffer ........... %u\n", stpos.num_blocks);
  if (stpos.bycu)
    printf("    Bytes in Buffer ............ UNKNOWN\n");
  else
    printf("    Bytes in Buffer ............ %u\n", stpos.num_bytes);

  stpos.block_type = QP_PHYSICAL;
  if (ioctl (fd, STIOCQRYPOS, &stpos) <0)
   return errno;

  printf("\nPhysical Tape position:\n\n");
  printf("    Partition Number ........... %d\n",stpos.partition_number);
  if (stpos.bot)
    printf("    Beginning of Partition ..... Yes\n");
  else
    printf("    Beginning of Partition ..... No\n");

  if (stpos.eot)
    printf("    Leading End of Partition ... Yes\n");
  else
    printf("    Leading End of Partition ... No\n");
    printf("    Current Block ID ........... %u (0x%08X)\n",
           stpos.curpos,stpos.curpos);
    printf("    Next Block on Tape ......... %u (0x%08X)\n",
         stpos.tapepos,stpos.tapepos);
  if (stpos.lbot == LBOT_NONE)
    printf("    Last Block Written ......... NONE\n");
  else if (stpos.lbot == LBOT_UNKNOWN)
    printf("    Last Block Written ......... UNKNOWN\n");
  else
    printf("    Last Block Written ......... %u (0x%08X)\n",
            stpos.lbot, stpos.lbot);

  if (stpos.bcu)
    printf("    Blocks in Buffer ........... UNKNOWN\n");
  else
    printf("    Blocks in Buffer ........... %u\n", stpos.num_blocks);
  if (stpos.bycu)
    printf("    Bytes in Buffer ............ UNKNOWN\n");
  else
    printf("    Bytes in Buffer ............ %u\n", stpos.num_bytes);
  printf("\n");

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stiocsetpos                                                         */
/*---------------------------------------------------------------------------*/
int stiocsetpos(void)
  {

  printf("Setting tape position to block id %u (0x%X)....\n",stpos.curpos,stpos.curpos);
  ioctl (fd, STIOCSETPOS, &stpos);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_read_position                                                 */
/*---------------------------------------------------------------------------*/
int stioc_read_position(void)
  {

  printf("Reading tape position....\n");

  if (ioctl (fd, STIOC_READ_POSITION, &blockid) <0)
   return errno;

  printf("\n    Current Block ID ............... %u (0x%08X)\n",blockid,blockid);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_locate                                                        */
/*---------------------------------------------------------------------------*/
int stioc_locate(void)
  {

  printf("\nLocating to Block ID %u (0x%08X)...\n",blockid,blockid);
  ioctl (fd, STIOC_LOCATE, blockid);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_recover_buffer                                                */
/*---------------------------------------------------------------------------*/
int stioc_recover_buffer(int size)
  {
  struct buffer_data bufdata;

  bufdata.bufsize = size;
  bufdata.buffer = malloc(size);
  if (bufdata.buffer == NULL)
    {
    errno = ENOMEM;
    return errno;
    }

  printf("Recovering buffered data....\n");
  if (ioctl (fd, STIOC_RECOVER_BUFFER, &bufdata) <0)
    {
    free(bufdata.buffer);
    return errno;
    }

  printf("\n");
  printf("  Buffer address ..... %08X\n",bufdata.buffer);
  printf("  Buffer size ........ %d\n",bufdata.bufsize);
  printf("  Returned Length .... %d\n",bufdata.ret_len);
  printf("\n");

  free(bufdata.buffer);
  return 0;
  }

/*---------------------------------------------------------------------------*/
/* Name:        rwtest                                                       */
/*---------------------------------------------------------------------------*/
int rwtest(int blksize, int blocks, int repetition)
  {
  int i, j, k;
  char *wbuf, *temp;
  char *rbuf;
  long long total = 0;
  struct stop stop;
  int bytes;
  struct stchgp stchgp;
  int count, blockid, tempid;

  printf("\nRead/Write test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

  if (blksize == 0)
    {
    count = blocks;
    blocks = 1;
    }
  else
    count = blocks * blksize;

  wbuf = malloc(count);
  if (wbuf == NULL)
    {
    errno = ENOMEM;
    return errno;
    }

  rbuf = malloc(count);
  if (rbuf == NULL)
    {
    errno = ENOMEM;
    free(wbuf);
    return errno;
    }

  printf("Reading current tape position...\n");
  if (ioctl (fd, STIOC_READ_POSITION, &blockid) <0)
    {
    free(wbuf);
    free(rbuf);
    return errno;
    }

  stchgp.st_blksize = blksize;
  printf("Setting block size...\n");
  if ((ioctl(fd, STIOCHGP, &stchgp)) < 0)
    {
    free(wbuf);
    free(rbuf);
    return errno;
    }

  /* generate data pattern */
  genpat();
  j=01;
  for (i=0;i<count;i++)
    {
    wbuf[i] = j;
    j=pat[j];
    }

  for (j=0,total=0;j<repetition;j++)
    {
    errno = 0;
    /* put current block id in bytes 0-3 of each block */
    temp = wbuf;
    tempid = blockid;
    for (i=0; i < blocks; i++)
      {
      temp[0] = ((char *)&tempid)[0];
      temp[1] = ((char *)&tempid)[1];
      temp[2] = ((char *)&tempid)[2];
      temp[3] = ((char *)&tempid)[3];
      temp += blksize;
      tempid++;
      }

    printf("\nRead/Write test pass %d\n",j+1);
    printf("Current block id %u (0x%08X)\n",blockid,blockid);
    printf("Writing %d bytes, %d blocks...\n",count,blocks);
    bytes = write(fd,wbuf,count);
    if (bytes == -1)
      break;

    total += bytes;
    if (blksize == 0)
      {
      printf("Write complete, bytes written %d, blocks 1\n", bytes);
      blockid++;
      }
    else
      {
      printf("Write complete, bytes written %d, blocks %d\n", bytes, bytes/blksize);
      blockid += bytes/blksize;
      }

    if (bytes != count)
      {
      printf("Error on write, bytes written not equal to bytes requested\n");
      errno = 999;
      break;
      }

    stop.st_op = STWEOF;
    stop.st_count = 2;
    printf("Writing 2 file marks...\n");
    if ((ioctl(fd, STIOCTOP, &stop)) < 0)
      break;

    stop.st_op = STRSF;
    stop.st_count = 2;
    printf("Backward spacing 2 file marks...\n");
    if ((ioctl(fd, STIOCTOP, &stop)) < 0)
      break;

    stop.st_op = STRSR;
    stop.st_count = blocks;
    printf("Backward spacing %d records...\n",blocks);
    if ((ioctl(fd, STIOCTOP, &stop)) < 0)
      break;

    /* read the data back from tape and compare to original */
    bzero(rbuf,count);
    printf("Reading %d bytes, %d blocks...\n",count,blocks);
    bytes = read(fd,rbuf,count);
    if (bytes == -1)
      break;

    if (blksize == 0)
      printf("Read complete, bytes read %d, blocks 1\n", bytes);
    else
      printf("Read complete, bytes read %d, blocks %d\n", bytes, bytes/blksize);

    if (bytes != count)
      {
      printf("Error on read, bytes read not equal to bytes written\n");
      errno = 999;
      break;
      }

    for (k=0; k < count; k++)
      {
      if (wbuf[k] != rbuf[k])
        {
        printf("Read data miscompare, ");
        printf("byte read at position %d is %02X, should be %02X\n", k,rbuf[k],wbuf[k]);
        errno = 999;
        break;
        }
      }
    if (k != count && err_stop)
      break;

    if (k == count)
      printf("Data verification complete, no discrepancies found\n");
    printf("Forward spacing file mark...\n");
    stop.st_op = STFSF;
    stop.st_count = 1;
    if ((ioctl(fd, STIOCTOP, &stop)) < 0)
      break;
    blockid++;
    }

  if (errno)
    {
    printf("\nCurrent block id %u (0x%08X)\n",blockid,blockid);
    printf("Read/Write test terminated, total bytes written %lld\n",total);
    }
  else
    {
    printf("\nCurrent block id %u (0x%08X)\n",blockid,blockid);
    printf("Read/Write test completed, total bytes written %lld\n",total);
    }
  free(wbuf);
  free(rbuf);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        wtest                                                        */
/*---------------------------------------------------------------------------*/
int wtest(int blksize, int blocks, int repetition)
  {
  int i, j, k;
  char *wbuf, *temp;
  long long total = 0;
  int bytes;
  struct stchgp stchgp;
  int count, blockid, tempid;

  printf("\nWrite test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

  if (blksize == 0)
    {
    count = blocks;
    blocks = 1;
    }
  else
    count = blocks * blksize;

  wbuf = malloc(count);
  if (wbuf == NULL)
    {
    errno = ENOMEM;
    return errno;
    }

  printf("Reading current tape position...\n");
  if (ioctl (fd, STIOC_READ_POSITION, &blockid) <0)
    {
    free(wbuf);
    return errno;
    }

  stchgp.st_blksize = blksize;
  printf("Setting block size...\n");
  if ((ioctl(fd, STIOCHGP, &stchgp)) < 0)
    {
    free(wbuf);
    return errno;
    }

  /* generate data pattern */
  genpat();
  j=01;
  for (i=0;i < count;i++)
    {
    wbuf[i] = j;
    j=pat[j];
    }

  for (j=0,total=0;j<repetition;j++)
    {
    /* put current block id in bytes 0-3 of each block */
    temp = wbuf;
    tempid = blockid;
    for (i=0; i < blocks; i++)
      {
      temp[0] = ((char *)&tempid)[0];
      temp[1] = ((char *)&tempid)[1];
      temp[2] = ((char *)&tempid)[2];
      temp[3] = ((char *)&tempid)[3];
      temp += blksize;
      tempid++;
      }

    printf("\nWrite test pass %d\n",j+1);
    printf("Current block id %u (0x%08X)\n",blockid,blockid);

    printf("Writing %d bytes, %d blocks...\n",count,blocks);
    bytes = write(fd,wbuf,count);
    if (bytes == -1)
      break;

    total += bytes;
    if (blksize == 0)
      {
      printf("Write complete, bytes written %d, blocks 1\n", bytes);
      blockid++;
      }
    else
      {
      printf("Write complete, bytes written %d, blocks %d\n", bytes, bytes/blksize);
      blockid += bytes/blksize;
      }

    if (bytes != count)
      {
      printf("Error on write, bytes written not equal to bytes requested\n");
      errno = 999;
      break;
      }
    }

  if (errno)
    {
    printf("\nCurrent block id %u (0x%08X)\n",blockid,blockid);
    printf("Write test terminated, total bytes written %lld\n",total);
    }
  else
    {
    printf("\nCurrent block id %u (0x%08X)\n",blockid,blockid);
    printf("Write test completed, total bytes written %lld\n",total);
    }
  free(wbuf);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        rtest                                                        */
/*---------------------------------------------------------------------------*/
int rtest(int blksize, int blocks, int repetition)
  {
  int i, j, k;
  char *wbuf, *temp;
  char *rbuf;
  long long total = 0;
  int bytes;
  struct stchgp stchgp;
  int count, blockid, tempid;

  printf("\nRead test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

  if (blksize == 0)
    {
    count = blocks;
    blocks = 1;
    }
  else
    count = blocks * blksize;

  wbuf = malloc(count);
  if (wbuf == NULL)
    {
    errno = ENOMEM;
    return errno;
    }

  rbuf = malloc(count);
  if (rbuf == NULL)
    {
    errno = ENOMEM;
    free(wbuf);
    return errno;
    }

  printf("Reading current tape position...\n");
  if (ioctl (fd, STIOC_READ_POSITION, &blockid) <0)
    {
    free(wbuf);
    free(rbuf);
    return errno;
    }

  stchgp.st_blksize = blksize;
  printf("Setting block size...\n");
  if ((ioctl(fd, STIOCHGP, &stchgp)) < 0)
    {
    free(wbuf);
    free(rbuf);
    return errno;
    }

  /* generate data pattern */
  genpat();
  j=01;
  for (i=0;i < count;i++)
    {
    wbuf[i] = j;
    j=pat[j];
    }

  for (j=0,total=0;j<repetition;j++)
    {
    /* put current block id in bytes 0-3 of each block */
    temp = wbuf;
    tempid = blockid;
    for (i=0; i < blocks; i++)
      {
      temp[0] = ((char *)&tempid)[0];
      temp[1] = ((char *)&tempid)[1];
      temp[2] = ((char *)&tempid)[2];
      temp[3] = ((char *)&tempid)[3];
      temp += blksize;
      tempid++;
      }

    printf("\nRead test pass %d\n",j+1);
    printf("Current block id %u (0x%08X)\n",blockid,blockid);

    /* read the data back from tape and compare to expected write data */
    bzero(rbuf,count);
    printf("Reading %d bytes, %d blocks...\n",count,blocks);
    bytes = read(fd,rbuf,count);
    if (bytes == -1)
      break;

    total += bytes;
    if (blksize == 0)
      {
      printf("Read complete, bytes read %d, blocks 1\n", bytes);
      blockid++;
      }
    else
      {
      printf("Read complete, bytes read %d, blocks %d\n", bytes, bytes/blksize);
      blockid += bytes/blksize;
      }

    for (k=0; k < count; k++)
      {
      if (wbuf[k] != rbuf[k])
        {
        printf("Read data miscompare, ");
        printf("byte read at position %d is %02X, should be %02X\n", k,rbuf[k],wbuf[k]);
        errno = 999;
        break;
        }
      }
    if (k != count)
      {
      if (err_stop)
        break;
      else
        {
        errno=0;
        continue;
        }
      }
    printf("Data verification complete, no discrepancies found\n");
    }

  if (errno)
    {
    printf("\nCurrent block id %u (0x%08X)\n",blockid,blockid);
    printf("Read test terminated, total bytes read %lld\n",total);
    }
  else
    {
    printf("\nCurrent block id %u (0x%08X)\n",blockid,blockid);
    printf("Read test completed, total bytes read %lld\n",total);
    }
  free(wbuf);
  free(rbuf);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: twrite                                                              */
/*---------------------------------------------------------------------------*/
int twrite(char *fname)
  {
  int wrtfd, rc=0;
  int bytes,count,blksize;
  char *wbuf;
  long long total = 0;
  struct stchgp stchgp;

  printf("Opening source file %s...\n",fname);
  wrtfd = open(fname, O_RDONLY|O_LARGEFILE);
  if (wrtfd == -1)
    return errno;

  if (stiocqryp() != 0)
    {
    close(wrtfd);
    return errno;
    }

  blksize = min(parms.max_blksize,parms.max_scsi_xfer);

  /* set actual block size to variable */
  stchgp.st_blksize = 0;
  printf("Setting block size to variable...\n");
  if ((ioctl(fd, STIOCHGP, &stchgp)) < 0)
    {
    close(wrtfd);
    return errno;
    }

  wbuf = malloc(blksize);
  if (wbuf == NULL)
    {
    close(wrtfd);
    errno = ENOMEM;
    return errno;
    }

  printf("Writing file to tape...\n");

  while(1)
    {
    bytes = read(wrtfd,wbuf,blksize);
    if (bytes == -1)
      {
      rc = errno;
      break;
      }
    if (bytes == 0)
      break;
    count = write(fd,wbuf,bytes);
    if (count == -1)
      {
      rc = errno;
      break;
      }
    total += count;
    if (count != bytes)
      {
      rc = EIO;
      break;
      }
    }

  if (rc)
    printf("Write terminated, %lld total bytes written...\n",total);
  else
    printf("Write complete, %lld total bytes written...\n",total);
  close(wrtfd);
  free(wbuf);
  return rc;
  }

/*---------------------------------------------------------------------------*/
/* Name: tread                                                               */
/*---------------------------------------------------------------------------*/
int tread(char *fname, int number)
  {
  int rdfd, rc=0;
  int bytes,count,blksize;
  char *rbuf;
  long long total = 0;
  struct stchgp stchgp;
  int rec_count = 0;

  printf("Opening destination file %s...\n",fname);

  rdfd = open(fname, O_RDWR|O_CREAT|O_TRUNC|O_LARGEFILE,0666);
  if (rdfd == -1)
    return errno;

  if (stiocqryp() != 0)
    {
    close(rdfd);
    return errno;
    }

  blksize = min(parms.max_blksize,parms.max_scsi_xfer);

  /* set actual block size to variable */
  stchgp.st_blksize = 0;
  printf("Setting block size to variable...\n");
  if ((ioctl(fd, STIOCHGP, &stchgp)) < 0)
    {
    close(rdfd);
    return errno;
    }

  rbuf = malloc(blksize);
  if (rbuf == NULL)
    {
    close(rdfd);
    errno = ENOMEM;
    return errno;
    }

  if (number)
    printf("Reading %d records from tape...\n",number);
  else
    printf("Reading file from tape...\n");

  while(1)
    {
    if (number && rec_count == number)
      break;
    bytes = read(fd,rbuf,blksize);
    if (bytes == -1)
      {
      rc = errno;
      break;
      }
    total += bytes;
    if (bytes == 0)
      break;
    rec_count++;
    count = write(rdfd,rbuf,bytes);
    if (count == -1)
      {
      rc = errno;
      break;
      }
    if (count != bytes)
      {
      rc = EIO;
      break;
      }

    /* don't loop if reading from null tape device */
    if (strstr(devname, ".null"))
      break;
    }

  if (rc || (number && rec_count != number))
    printf("Read terminated, %d records, %lld total bytes read...\n",
            rec_count,total);
  else
    printf("Read complete, %d records, %lld total bytes read...\n",
            rec_count,total);
  close(rdfd);
  free(rbuf);
  return rc;
  }

/*---------------------------------------------------------------------------*/
/* Name: list_tape                                                           */
/*---------------------------------------------------------------------------*/
int list_tape(void)
  {
  int rc=0;
  int count=0;
  int fm=0;
  int bytes,blksize;
  char *rbuf;
  struct stchgp_s parms;
  int save_errno = 0;
  struct stsense_s stsense;

  if (stioctop(STREW,0) != 0)
    return errno;

  if ((ioctl(fd, STIOCQRYP, &parms)) < 0)
    return errno;

  blksize = min(parms.max_blksize,parms.max_scsi_xfer);

  /* set actual block size to variable */
  printf("Setting block size to variable...\n");
  parms.autoload = FALSE;
  parms.blksize = 0;
  if ((ioctl(fd, STIOCSETP, &parms)) < 0)
    return errno;

  rbuf = malloc(blksize);
  if (rbuf == NULL)
    {
    errno = ENOMEM;
    return errno;
    }

  printf("Scanning tape...\n");

  while(1)
    {
    bytes = read(fd,rbuf,blksize);
    if (bytes == -1)
      break;
    if (bytes == 0)
      {
      fm++;
      count = 0;
      printf("*** Filemark %d ***\n",fm);
      }
    else
      {
      count++;
      printf("Record %d  Length %d\n",count,bytes);
      }

    /* don't loop if reading from null tape device */
    if (strstr(devname, ".null"))
      break;
    }

  if (bytes == -1)
    {
    save_errno = errno;
    stsense.sense_type = LASTERROR;

    /* check for normal end of data, sense key 8 and asc/ascq=0005 */
    if (ioctl(fd, STIOCQRYSENSE, &stsense) == 0 && stsense.len > 12 &&
        stsense.sense[2] == 0x08 && stsense.sense[12] == 0x00 &&
        stsense.sense[13] == 0x05)
      {
      printf("\nTape listing completed.\n");
      errno = 0;
      }
    else
      {
      printf("*** Unexpected error listing tape ***\n\n");
      errno = save_errno;
      }
    }

  free(rbuf);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: qvpd                                                               */
/*---------------------------------------------------------------------------*/
int qvpd(void)
  {
  struct sc_iocmd sciocmd;
  struct inquiry_data inquiry_data;

  bzero(&sciocmd, sizeof(struct sc_iocmd));
  bzero(&inquiry_data, sizeof(struct inquiry_data));

  /* build inquiry */
  sciocmd.scsi_cdb[0]=SCSI_INQUIRY;
  sciocmd.timeout_value=200;    /* SECONDS */
  sciocmd.command_length=6;
  sciocmd.buffer=(char *)&inquiry_data;
  sciocmd.data_length=sizeof(struct inquiry_data);
  sciocmd.scsi_cdb[4]=sizeof(struct inquiry_data);
  sciocmd.flags=B_READ;

  printf ("Getting vpd data from device...\n");

  if (!ioctl (fd, STIOCMD, &sciocmd))
    {
    printf("\n  Manufacturer .............. %0.8s\n",inquiry_data.vid);
    printf("  Product Identification .... %0.16s\n",inquiry_data.pid);
    printf("  Revision Level ............ %0.4s\n",inquiry_data.revision);
    printf("\n");
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: genpat                                                              */
/*---------------------------------------------------------------------------*/
void genpat(void)
  {
  pat[0x00] = 0x01;
  pat[0x01] = 0x13;
  pat[0x13] = 0x37;
  pat[0x37] = 0x7F;
  pat[0x7F] = 0xFE;
  pat[0xFE] = 0xFC;
  pat[0xFC] = 0xE8;
  pat[0xE8] = 0xC0;
  pat[0xC0] = 0x81;
  pat[0x81] = 0x03;
  pat[0x03] = 0x07;
  pat[0x07] = 0x1F;
  pat[0x1F] = 0x3E;
  pat[0x3E] = 0x7C;
  pat[0x7C] = 0xF8;
  pat[0xF8] = 0xF0;
  pat[0xF0] = 0xE1;
  pat[0xE1] = 0xC3;
  pat[0xC3] = 0x87;
  pat[0x87] = 0x0F;
  pat[0x0F] = 0x0E;
  pat[0x0E] = 0x1C;
  pat[0x1C] = 0x38;
  pat[0x38] = 0x70;
  pat[0x70] = 0xF1;
  pat[0xF1] = 0xF3;
  pat[0xF3] = 0xE7;
  pat[0xE7] = 0xCF;
  pat[0xCF] = 0x8E;
  pat[0x8E] = 0x0C;
  pat[0x0C] = 0x08;
  pat[0x08] = 0x10;
  pat[0x10] = 0x31;
  pat[0x31] = 0x73;
  pat[0x73] = 0xF7;
  pat[0xF7] = 0xFF;
  pat[0xFF] = 0xEE;
  pat[0xEE] = 0xCC;
  pat[0xCC] = 0x88;
  pat[0x88] = 0x00;
  }

/*---------------------------------------------------------------------------*/
/* Name: htoc                                                                */
/*---------------------------------------------------------------------------*/
int  htoc(char *s,unsigned char h)
  {
  char *ttable="0123456789ABCDEF"; /* hex character translate table    */
  short  x;

  x = (h & 0xf0) >> 4;             /* isolate hi nibble for conversion */
  s[0] = ttable[x];                /* index into table for ascii char  */
  x = h & 0x0f;                    /* isolate lo nibble for conversion */
  s[1] = ttable[x];                /* index into table for ascii char  */
  }

/*---------------------------------------------------------------------------*/
/* Name:        dump_bytes                                                   */
/*---------------------------------------------------------------------------*/
void dump_bytes(char *b, int len, char *title)
  {
  short str_index;
  size_t buf_index;
  size_t off;
  size_t lines;
  size_t frames;
  size_t remainder;
  size_t index = 0;
  char smsg[48];
  char tmsg[48];
  unsigned char *ptr;
  char  ttext[16];
  int start = 0;

  ptr = b + start;        /* point to buffer location to start  */
  buf_index = 0;
  frames    = (len/256);       /* calculate number of frames         */
  remainder = (len%256);
  if (remainder)               /* if there are bytes left over       */
    {
    frames++;                  /* then we need another frame and..   */
    remainder = (remainder-1) / 16; /* calculate how many lines are  */
    remainder++;               /* in the last frame                  */
    }

  if (title)                 /* if a title has been supplied       */
    printf("\n%s,  Length %u\n\n", title,len);
  else
    printf("\nData,  Length %u\n\n", len);

  printf( "            0 1  2 3  4 5  6 7  8 9  A B  C D  E F   0123456789ABCDEF\n");

  while (frames > 0)
    {
    /* if last frame and more lines are required get number of lines */
    if (frames == 1 && remainder > 0)
      lines = remainder;
    else
      lines = 16;

    while (lines>0)
      {
      memset(smsg,' ',48);
      memset(tmsg,' ',48);

      for (str_index=0,off=0;str_index<16;
           str_index++,buf_index++,off+=3)
        {
        if (buf_index >= len)
          break;               /* if we are at the end of the buffer */
        htoc(&tmsg[off],ptr[buf_index]); /* convert hex to string    */
        smsg[str_index]=ptr[buf_index];  /* save in ascii string     */
        if (smsg[str_index] < ' ')
          smsg[str_index]='.'; /* convert non printable chars to '.' */
        buf_index++;           /* index to next byte in buffer       */
        if (buf_index >= len)
          break;               /* if we are at the end of buffer     */
        str_index++;           /* index to next position in string   */
        off+=2;                /* index to next position in hex str  */
        htoc(&tmsg[off],ptr[buf_index]); /* convert hex to string    */
        smsg[str_index]=ptr[buf_index];  /* save in ascii string     */
        if (smsg[str_index] < ' ')
          smsg[str_index]='.'; /* convert non printable chars to '.' */
        }
      tmsg[39]=0;
      smsg[16]=0;
      printf("    %04X - %s  [%s]\n",index,tmsg,smsg);
      index += 16;
      lines--;                 /* finished with this line            */
      }
    frames--;                  /* finished with this frame           */
    }
  printf("\n");
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_create_partition                                              */
/*---------------------------------------------------------------------------*/
int stioc_create_partition(int type, int number)
  {
  struct tape_partition partition;

  partition.type = type;
  partition.number = number;

  if (type == SDP_PARTITION)
    {
    if (number == 1)
      printf("Creating 1 SDP partition....\n");
    else
      printf("Creating %d SDP partitions....\n",number);
    }
  else
    printf("Creating IDP partition....\n",number);

  ioctl(fd, STIOC_CREATE_PARTITION, &partition);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_change_partition                                              */
/*---------------------------------------------------------------------------*/
int stioc_change_partition(int number,int blockid)
  {
  struct change_partition partition;

  partition.partition_number = number;
  partition.block_id = blockid;

  if (blockid)
    printf("Changing to partition %d, block id %d....\n",number,blockid);
  else
    printf("Changing to partition %d....\n",number);

  ioctl(fd, STIOC_CHANGE_PARTITION, &partition);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_query_partition                                               */
/*---------------------------------------------------------------------------*/
int stioc_query_partition(void)
  {
  struct query_partition partition;
  int i,j;

  printf("Querying tape partitioning....\n");

  if (ioctl(fd, STIOC_QUERY_PARTITION, &partition) < 0)
    return errno;

  printf("\n");
  if (partition.type == IDP_PARTITION)
    printf("  Type of Partitioning ....... IDP\n");
  else if (partition.type == SDP_PARTITION)
    printf("  Type of Partitioning ....... SDP\n");
  else
    printf("  Type of Partitioning ....... UNKNOWN\n");

  printf("  Number of Partitions ....... %d\n",partition.number);
  j=0;
  for (i=0; i < partition.number; i++)
    {
    if (opcode && j == 16)
      {
      j=0;
      printf ("\nHit <enter> to continue...");
      getchar();
      }

    if (i < 10)
      printf("  Partition %d Size (Meg) ..... %d\n",i,partition.size[i]);
    else
      printf("  Partition %d Size (Meg) .... %d\n",i,partition.size[i]);
    j++;
    }
  printf("\n");

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_query_path                                                     */
/*---------------------------------------------------------------------------*/
int sioc_query_path(void)
  {
  struct scsi_path path;

  printf("Querying SCSI paths...\n");

  if (ioctl(fd, SIOC_QUERY_PATH, &path) == 0)
    show_path(&path);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_device_paths                                                   */
/*---------------------------------------------------------------------------*/
int sioc_device_paths(void)
  {
  struct device_paths device_path;
  int i;

  printf("Querying device paths...\n");

  if (ioctl(fd, SIOC_DEVICE_PATHS, &device_path) == 0)
    {
    printf("\n");
    for (i=0; i < device_path.number_paths; i++)
      {
      if (i == 0)
        printf("Primary Path Number 1\n");
      else
        printf("Alternate Path Number %d\n", i+1);
      printf("  Logical Device................. %s\n",device_path.path[i].name);
      printf("  Product Identification......... %0.16s\n",device_path.path[i].type);
      printf("  Serial Number.................. %0.16s\n",device_path.path[i].serial);
      printf("  SCSI Parent.................... %s\n",device_path.path[i].parent);
      if (device_path.path[i].fcp_id_valid)
        {
        if (device_path.path[i].dynamic_tracking)
          printf("  Dynamic Tracking Enabled....... Yes\n");
        else
          printf("  Dynamic Tracking Enabled....... No \n");
        printf("  FCP SCSI ID.................... 0x%llx\n",device_path.path[i].fcp_scsi_id);
        printf("  FCP Logical Unit............... 0x%llx\n",device_path.path[i].fcp_lun_id);
        printf("  FCP World Wide Port Name....... 0x%llx\n",device_path.path[i].fcp_ww_name);
        printf("  FCP World Wide Node Name....... 0x%llx\n",device_path.path[i].fcp_node_name);
        }
      else if (device_path.path[i].sas_id_valid)
        {
        printf("  SAS SCSI ID.................... 0x%llx\n",device_path.path[i].fcp_scsi_id);
        printf("  SAS Logical Unit............... 0x%llx\n",device_path.path[i].fcp_lun_id);
        }
      else
        {
        printf("  Target ID...................... %d\n",device_path.path[i].id);
        printf("  Logical Unit................... %d\n",device_path.path[i].lun);
        }
      if (device_path.path[i].drive_port_valid)
        printf("  Drive Port Number.............. %d\n",device_path.path[i].drive_port);
      if (device_path.path[i].enabled)
        printf("  Path Enabled................... Yes\n");
      else
        printf("  Path Enabled................... No \n");
      if (device_path.path[i].fenced)
        printf("  Path Manually Disabled......... Yes\n");
      else
        printf("  Path Manually Disabled......... No \n");

      printf("\n");
      }
    printf("Current path..................... %d\n",device_path.path[0].current_path);
    printf("Total paths configured........... %d\n",device_path.number_paths);
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_reset_path                                                     */
/*---------------------------------------------------------------------------*/
int sioc_reset_path(void)
  {
  struct scsi_path path;

  printf("Resetting SCSI paths...\n");

  if (ioctl(fd, SIOC_RESET_PATH, &path) == 0)
    show_path(&path);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_check_path                                                     */
/*---------------------------------------------------------------------------*/
int sioc_check_path(void)
  {
  struct scsi_path path;

  printf("Checking SCSI paths...\n");

  if (ioctl(fd, SIOC_CHECK_PATH, &path) == 0)
    show_path(&path);

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_disable_path                                                   */
/*---------------------------------------------------------------------------*/
int sioc_disable_path(int path)
  {

  if (path == PRIMARY_SCSI_PATH)
    printf("Disabling primary SCSI path 1...\n");
  else
    printf("Disabling alternate SCSI path %d...\n",path);

  ioctl(fd, SIOC_DISABLE_PATH, path);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_enable_path                                                    */
/*---------------------------------------------------------------------------*/
int sioc_enable_path(int path)
  {

  if (path == PRIMARY_SCSI_PATH)
    printf("Enabling primary SCSI path 1...\n");
  else
    printf("Enabling alternate SCSI path %d...\n",path);

  ioctl(fd, SIOC_ENABLE_PATH, path);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: show_path                                                           */
/*---------------------------------------------------------------------------*/
void show_path(struct scsi_path *path)
  {

  printf("\n");
  if (path->alternate_configured)
    printf("Primary Path Information:\n");
  printf("  Logical Device................. %s\n",path->primary_name);
  printf("  SCSI Parent.................... %s\n",path->primary_parent);
  if (path->primary_fcp_id_valid)
    {
    printf("  FCP SCSI ID.................... 0x%llx\n",path->primary_fcp_scsi_id);
    printf("  FCP Logical Unit............... 0x%llx\n",path->primary_fcp_lun_id);
    printf("  FCP World Wide Port Name....... 0x%llx\n",path->primary_fcp_ww_name);
    }
  else if (path->primary_sas_id_valid)
    {
    printf("  SAS SCSI ID.................... 0x%llx\n",path->primary_fcp_scsi_id);
    printf("  SAS Logical Unit............... 0x%llx\n",path->primary_fcp_lun_id);
    }
  else
    {
    printf("  Target ID...................... %d\n",path->primary_id);
    printf("  Logical Unit................... %d\n",path->primary_lun);
    }
  if (path->primary_drive_port_valid)
    printf("  Drive Port Number.............. %d\n",path->primary_drive_port);
  if (path->primary_enabled)
    printf("  Path Enabled................... Yes\n");
  else
    printf("  Path Enabled................... No \n");
  if (path->primary_fenced)
    printf("  Path Manually Disabled......... Yes\n");
  else
    printf("  Path Manually Disabled......... No \n");
  if (path->current_path == PRIMARY_SCSI_PATH)
    printf("  Current Path................... Yes\n");
  else
    printf("  Current Path................... No\n");

  if (!path->alternate_configured)
    printf("  Alternate Path Configured...... No\n");
  else
    {
    printf("  Alternate Path Configured...... Yes\n");
    printf("\nAlternate Path Information:\n");
    printf("  Logical Device................. %s\n",path->alternate_name);
    printf("  SCSI Parent.................... %s\n",path->alternate_parent);
    if (path->alternate_fcp_id_valid)
      {
      printf("  FCP SCSI ID.................... 0x%llx\n",path->alternate_fcp_scsi_id);
      printf("  FCP Logical Unit............... 0x%llx\n",path->alternate_fcp_lun_id);
      printf("  FCP World Wide Port Name....... 0x%llx\n",path->alternate_fcp_ww_name);
      }
    else if (path->alternate_sas_id_valid)
      {
      printf("  SAS SCSI ID.................... 0x%llx\n",path->alternate_fcp_scsi_id);
      printf("  SAS Logical Unit............... 0x%llx\n",path->alternate_fcp_lun_id);
      }
    else
      {
      printf("  Target ID...................... %d\n",path->alternate_id);
      printf("  Logical Unit................... %d\n",path->alternate_lun);
      }
    if (path->alternate_drive_port_valid)
      printf("  Drive Port Number.............. %d\n",path->alternate_drive_port);
    if (path->alternate_enabled)
      printf("  Path Enabled................... Yes\n");
    else
      printf("  Path Enabled................... No \n");
    if (path->alternate_fenced)
      printf("  Path Manually Disabled......... Yes\n");
    else
      printf("  Path Manually Disabled......... No \n");
    if (path->current_path == ALTERNATE_SCSI_PATH)
      printf("  Current Path................... Yes\n");
    else
      printf("  Current Path................... No\n");

    }
  }

/*---------------------------------------------------------------------------*/
/* Name: stoc_report_density_support                                         */
/*---------------------------------------------------------------------------*/
int stioc_report_density_support(void)
  {
  int i;
  struct report_density_support density;

  printf("Issuing Report Density Support for ALL supported media...\n");

  density.media = ALL_MEDIA_DENSITY;

  if (ioctl(fd, STIOC_REPORT_DENSITY_SUPPORT, &density) != 0)
    return errno;

  printf("Total number of densities reported: %d\n",density.number_reports);
  for (i = 0; i < density.number_reports; i++)
    {
    printf("\n");
    printf("  Density Name................. %0.8s\n",density.reports[i].density_name);
    printf("  Assigning Organization....... %0.8s\n",density.reports[i].assigning_org);
    printf("  Description.................. %0.20s\n",density.reports[i].description);
    printf("  Primary Density Code......... %02X\n",density.reports[i].primary_density_code);
    printf("  Secondary Density Code....... %02X\n",density.reports[i].secondary_density_code);

    if (density.reports[i].wrtok)
      printf("  Write OK..................... Yes\n");
    else
      printf("  Write OK..................... No\n");

    if (density.reports[i].dup)
      printf("  Duplicate.................... Yes\n");
    else
      printf("  Duplicate.................... No\n");

    if (density.reports[i].deflt)
      printf("  Default...................... Yes\n");
    else
      printf("  Default...................... No\n");

    printf("  Bits per MM.................. %d\n",density.reports[i].bits_per_mm);
    printf("  Media Width.................. %d\n",density.reports[i].media_width);
    printf("  Tracks....................... %d\n",density.reports[i].tracks);
    printf("  Capacity (megabytes)......... %d\n",density.reports[i].capacity);
    if (opcode)
      {
      printf ("\nHit <enter> to continue...");
      getchar();
      }
    }

  printf("\nIssuing Report Density Support for CURRENT media...\n");

  density.media = CURRENT_MEDIA_DENSITY;

  if (ioctl(fd, STIOC_REPORT_DENSITY_SUPPORT, &density) != 0)
    return errno;

  for (i = 0; i < density.number_reports; i++)
    {
    printf("\n");
    printf("  Density Name................. %0.8s\n",density.reports[i].density_name);
    printf("  Assigning Organization....... %0.8s\n",density.reports[i].assigning_org);
    printf("  Description.................. %0.20s\n",density.reports[i].description);
    printf("  Primary Density Code......... %02X\n",density.reports[i].primary_density_code);
    printf("  Secondary Density Code....... %02X\n",density.reports[i].secondary_density_code);

    if (density.reports[i].wrtok)
      printf("  Write OK..................... Yes\n");
    else
      printf("  Write OK..................... No\n");

    if (density.reports[i].dup)
      printf("  Duplicate.................... Yes\n");
    else
      printf("  Duplicate.................... No\n");

    if (density.reports[i].deflt)
      printf("  Default...................... Yes\n");
    else
      printf("  Default...................... No\n");

    printf("  Bits per MM.................. %d\n",density.reports[i].bits_per_mm);
    printf("  Media Width.................. %d\n",density.reports[i].media_width);
    printf("  Tracks....................... %d\n",density.reports[i].tracks);
    printf("  Capacity (megabytes)......... %d\n",density.reports[i].capacity);
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_query_open                                                     */
/*---------------------------------------------------------------------------*/
int sioc_query_open(void)
  {
  int pid = 0;


  if (ioctl(fd, SIOC_QUERY_OPEN, &pid) == 0)
    {
    if (pid)
      printf("Device is currently open by process id %d\n",pid);
    else
      printf("Device is not open\n");
    }
  else
    printf("Error querying device open...\n");

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_kill_open                                                      */
/*---------------------------------------------------------------------------*/
int sioc_kill_open(void)
  {

  printf("Killing open processes...\n");
  ioctl(fd, SIOC_KILL_OPEN, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        sioc_reset_device                                            */
/*---------------------------------------------------------------------------*/
int sioc_reset_device()
  {
  printf("Resetting device...\n");

  ioctl (fd, SIOC_RESET_DEVICE, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_debug                                                          */
/*---------------------------------------------------------------------------*/
int sioc_debug(void)
  {

  printf("Issuing debug command...\n");
  ioctl(fd, SIOC_DEBUG, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_trcstop                                                        */
/*---------------------------------------------------------------------------*/
int sioc_trcstop(void)
  {

  printf("Trace logging stopped...\n");
  ioctl(fd, SIOC_TRCSTOP, NULL);
  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_trcstart                                                       */
/*---------------------------------------------------------------------------*/
int sioc_trcstart(void)
  {

  if (trclog.query_trace)
    printf("Querying trace logging...\n");
  else
    printf("Starting trace logging...\n");

  if (ioctl(fd,SIOC_TRCSTART, &trclog) == 0)
    {
    if (trclog.query_trace && !trclog.trace_active)
      printf("Trace logging not active...\n");
    else
      {
      printf("\n");
      printf("  Trace log directory...... %s \n",trclog.dir);
      printf("  Number of trace files.... %d \n",trclog.files);
      printf("  Trace file size.......... %d \n",trclog.size);
      if (trclog.cur_file != 0)
        printf("  Current trace file....... %d \n",trclog.cur_file);
      else
        printf("  Current trace file....... 1 \n");
      if (trclog.debug_trace)
        printf("  Debug trace.............. On \n");
      else
        printf("  Debug trace.............. Off \n");
      }
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name: sioc_passthru_command                                               */
/*---------------------------------------------------------------------------*/
int sioc_passthru_command(void)
  {
  struct scsi_passthru_cmd passthru;
  char buffer[255];

  /* build inquiry */
  bzero(&passthru, sizeof(struct scsi_passthru_cmd));
  passthru.scsi_cdb[0]=SCSI_INQUIRY;
  passthru.scsi_cdb[4]=255;
  passthru.command_length=6;
  passthru.timeout_value=200;
  passthru.buffer=buffer;
  passthru.buffer_length=255;
  passthru.trace_length=128;

  printf ("Issuing SCSI passthru Inquiry command...\n");
  errno = 0;
  ioctl(fd, SIOC_PASSTHRU_COMMAND, &passthru);

  printf("  Errno....................... %d\n",errno);
  printf("  Sense length ............... %d\n",passthru.sense_length);
  printf("  Timeout value .............. %d\n",passthru.timeout_value);
  printf("  Buffer length .............. %d\n",passthru.buffer_length);
  printf("  Number bytes transferred.... %d\n",passthru.number_bytes);
  if (passthru.sense_length)
    dump_bytes ((uchar *)passthru.sense,passthru.sense_length,"Sense Data");
  if (passthru.number_bytes)
    dump_bytes ((uchar *)passthru.buffer,passthru.number_bytes,"Buffer Data");
  printf("\n");

  /* build test uit ready */
  bzero(&passthru, sizeof(struct scsi_passthru_cmd));
  passthru.scsi_cdb[0]=0;
  passthru.command_length=6;
  passthru.timeout_value=30;
  passthru.buffer=NULL;
  passthru.buffer_length=0;

  printf ("Issuing SCSI passthru Test Unit Ready command...\n");
  errno = 0;
  ioctl(fd, SIOC_PASSTHRU_COMMAND, &passthru);

  printf("  Errno....................... %d\n",errno);
  printf("  Sense length ............... %d\n",passthru.sense_length);
  printf("  Timeout value .............. %d\n",passthru.timeout_value);
  printf("  Buffer length .............. %d\n",passthru.buffer_length);
  printf("  Number bytes transferred.... %d\n",passthru.number_bytes);
  if (passthru.sense_length)
    dump_bytes ((uchar *)passthru.sense,passthru.sense_length,"Sense Data");
  if (passthru.number_bytes)
    dump_bytes ((uchar *)passthru.buffer,passthru.number_bytes,"Buffer Data");
  printf("\n");

  return 0;
  }

/*---------------------------------------------------------------------------*/
/* Name: stioc_get_encryption_state                                          */
/*---------------------------------------------------------------------------*/
int stioc_get_encryption_state(void)
  {
  struct encryption_status  ep_state;

  printf ("Getting drive encryption settings...\n");
  if (!ioctl (fd, GET_ENCRYPTION_STATE, &ep_state))
    {
    printf ("Encryption settings:\n\n");
    printf ("  Drive Encryption Capable.... %s\n", (ep_state.encryption_capable == 1 ? "Yes" : "No"));
    printf ("  Encryption Method........... ");
    switch (ep_state.encryption_method)
      {
      case 0: printf ("None\n"); break;
      case 1: printf ("Library\n"); break;
      case 2: printf ("System\n"); break;
      case 3: printf ("Application\n"); break;
      case 4: printf ("Custom\n"); break;
      case 5: printf ("Unknown\n"); break;
      }
    printf ("  Encryption State............ ");
    switch (ep_state.encryption_state)
      {
      case 0: printf ("Off\n"); break;
      case 1: printf ("On\n"); break;
      case 2: printf ("NA\n"); break;
      }
    }

  return errno;
  }

/*---------------------------------------------------------------------------*/
/* Name:        encryption_test                                              */
/*---------------------------------------------------------------------------*/
int encryption_test()
  {
  struct encryption_diagnostics diag;

  /* test server config and connections */
  printf("Testing server configuration and connections...\n");
  diag.diag_type = SERVER_PING_DIAG;

  if (ioctl (fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag))
    return errno;

  if (diag.diag_errno)
    {
    errno = diag.diag_errno;
    perror("Server test failed");
    errno = 0;
    return errno;
    }
  else
    printf("Testing complete, servers available %d\n",diag.diag_cc);

  /* run basic drive to server encryption test */
  printf("Running basic drive to server encryption test...\n");
  diag.diag_type = BASIC_ENCRYPTION_DIAG;

  if (ioctl (fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag))
    {
    perror("Basic Encryption ioctl failed");
    errno =0;
    }
  else
    {
    if (diag.diag_errno)
      {
      errno = diag.diag_errno;
      perror("Basic Encryption test failed");
      errno = 0;
      }
    else
      printf("Testing complete, completion code %d\n",diag.diag_cc);
    }

  /* run full drive to server encryption test */
  printf("Running full drive to server encryption test...\n");
  diag.diag_type = FULL_ENCRYPTION_DIAG;

  if (ioctl (fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag))
    {
    perror("Full Encryption ioctl failed");
    errno =0;
    }
  else
    {
    if (diag.diag_errno)
      {
      errno = diag.diag_errno;
      perror("Full Encryption test failed");
      errno = 0;
      }
    else
      printf("Testing complete, completion code %d\n",diag.diag_cc);
   }

  return 0;
  }

/*---------------------------------------------------------------------------*/
/* Name: rekey_command                                                       */
/*---------------------------------------------------------------------------*/
int rekey_command(void)
  {
  struct scsi_passthru_cmd passthru;
  char buffer[255];

  /* build mode select x25/xEE in short form */
  bzero(&passthru, sizeof(struct scsi_passthru_cmd));
  bzero(&buffer,sizeof(buffer));
  passthru.scsi_cdb[0]=SCSI_MODE_SELECT_10;
  passthru.scsi_cdb[1]=0x10;
  passthru.scsi_cdb[8]=56;
  passthru.command_length=10;
  passthru.timeout_value=600;
  passthru.buffer=buffer;
  passthru.buffer_length=56;
  passthru.number_bytes=56;
  passthru.trace_length=128;

  passthru.buffer[3]=0x10;
  passthru.buffer[8]=0x65;
  passthru.buffer[9]=0xEE;
  passthru.buffer[11]=0x2C;
  passthru.buffer[12]=0x21;
  passthru.buffer[13]=0xEE;
  passthru.buffer[15]=0x28;
  passthru.buffer[20]=0x18;
  passthru.buffer[28]=0x18;
  passthru.buffer[31]=0x04;
  passthru.buffer[39]=0x02;
  passthru.buffer[40]=0x10;
  passthru.buffer[44]=0x19;
  passthru.buffer[48]=0x10;
  passthru.buffer[52]=0x19;


  printf ("Issuing rekey command...\n");
  errno = 0;
  ioctl(fd, SIOC_PASSTHRU_COMMAND, &passthru);

  if (errno)
     printf("rekey command failed (errno=%d)!\n", errno);
  else
     printf("rekey command succeeded!\n");

  if (passthru.sense_length)
    dump_bytes ((uchar *)passthru.sense,passthru.sense_length,"Sense Data");
  printf("\n");

  return 0;
  }

/*---------------------------------------------------------------------------*/
/* Name: get_sernum command                                                  */
/*---------------------------------------------------------------------------*/
int get_sernum()
{
   struct inquiry_page inqdata;
   char serial_num[20];

   inqdata.page_code = (uchar) 0x80;
   if (!ioctl(fd, SIOC_INQUIRY_PAGE, &inqdata)) {
      bcopy(&inqdata.data[4], serial_num, inqdata.data[3]);
      serial_num[inqdata.data[3]] = '\0';
      printf("%s\n", serial_num);
   }
   else {
      printf ("Get serial number failed (errno=%d).\n", errno);
      sioc_reqsense();
   }

   return errno;
}


