/* tapeutil for rekeying use                                                 */
/*****************************************************************************/
/*****************************************************************************/
/**                                                                         **/
/**      IBM SCSI Tape & Medium Changer Utility Program / HP-UX             **/
/**              (C) COPYRIGHT IBM Corporation, 1998-                       **/
/**                                                                         **/
/**                tapeutil.c - Utility Program Source File                 **/
/**                  Licensed Materials - Property of IBM                   **/
/**                           All Rights Reserved                           **/
/**                                                                         **/
/**    US Government Users Restricted Rights  -  Use, duplication, or       **/
/**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.    **/
/**                                                                         **/
/**           ** Notice to users of the source code examples: **            **/
/**                                                                         **/
/**  The source code examples provided by IBM are only intended to assist   **/
/**  in the development of a working software program.  The source code     **/
/**  examples may not function as written; additional code is required.     **/
/**  In addition, the source code examples may not compile and/or bind      **/
/**  successfully as written.                                               **/
/**                                                                         **/
/**  International Business Machines Corporation provides the source code   **/
/**  examples "as is" without warranty of any kind, either expressed or     **/
/**  implied, including, but not limited to the implied warranties of       **/
/**  merchantability and fitness for a particular purpose.  The entire      **/
/**  risk as to the quality and performance of the source code examples     **/
/**  is with you.  Should any part of the source code examples prove        **/
/**  defective, you (and not IBM) assume the entire cost of all necessary   **/
/**  servicing, repair or correction.                                       **/
/**                                                                         **/
/**  IBM does not warrant that the contents of the source code examples     **/
/**  will meet your requirements or that the source code examples are       **/
/**  error-free.                                                            **/
/**                                                                         **/
/**  IBM may make improvements and/or changes in the source code examples   **/
/**  at any time.  Changes may be made periodically to the information in   **/
/**  the source code examples; these changes may be reported, for the       **/
/**  device drivers included herein, in new editions of the examples.       **/
/**                                                                         **/
/**  References in the source code examples to IBM products, programs, or   **/
/**  services do not imply that IBM intends to make these available in all  **/
/**  countries in which IBM operates.  Any reference to an IBM licensed     **/
/**  program in the source code examples is not intended to state or imply  **/
/**  that only IBM's licensed program may be used.  Any functionally        **/
/**  equivalent program may be used.                                        **/
/**                                                                         **/
/*****************************************************************************/


/*****************************************************************************/
/*                               INCLUDE FILES                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <fcntl.h>
#include <time.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/scsi.h>

#include "st.h"
#include "svc.h"
#include "smc.h"

/*****************************************************************************/
/*                                GLOBAL DATA                                */
/*****************************************************************************/

#define VERSION         4.3.6.0

#define TRUE            1
#define FALSE           0
#define EXTENDED        1

#define IOBUFSZ         0x10000
#define RWBUFSZ         0x500000

#define END_BUFLIST     -1

#define PRINTF          if (verbose)  printf
#define PERROR          if (verbose)  perror
#define FPRINTF         if (verbose)  fprintf

#ifndef MIN
#define MIN(a,b) ((a) < (b)) ? (a) : (b)
#endif

#ifndef _SC_KERNEL_BITS
#define _SC_KERNEL_BITS 10013
#endif

#ifndef _CS_MACHINE_MODEL
#define _CS_MACHINE_MODEL 10000
#endif

#define INQ_DRV_TYPE 0x01
#define INQ_SMC_TYPE 0x08

#define MT_MODE                     1
#define ST_MODE                     2

typedef struct {
  int  opcode;                          /* user selected operation             */
  int  mode;                            /* open mode flag                      */
  int  src;                             /* source for move/exchange command    */
  int  dst;                             /* first dest for move/exchange cmd    */
  int  dst2;                            /* second destination for exchange cmd */
  int  blk_size;                        /* block size for data transfers       */
  int  blk_fact;                        /* blocking factor                     */
  int  multiplier;                      /* multiple of buffer to read/write    */
  int  seed;                            /* random number generator seed        */
  int  mt_op;                           /* MTIOCTOP op field                   */
  int  mt_count;                        /* MTIOCTOP count field                */
  int  st_op;                           /* STIOC_TAPE_OP op field              */
  int  st_count;                        /* STIOC_TAPE_OP count field           */
  int  type;                            /* msg/pos/prm type field              */
  int  value;                           /* pos/prm value field                 */
  int  parm_code;                       /* log sense parameter code            */
  char msg_0[MSGLEN+1];                 /* message 0 text                      */
  char msg_1[MSGLEN+1];                 /* message 1 text                      */
  char filename[64];                    /* ucode/dump/pattern file name        */
  uchar default_density;                /* default write density               */
  uchar pending_density;                /* pending write density               */
} parm_t;

typedef struct {          /* device type/model indicator       */
   uint
        t3490     : 1,
        t3570     : 1,
        t3590     : 1,
        t3592     : 1,
        t3592j1a  : 1,
        ultrium   : 1,
        lto1      : 1,
        lto2      : 1,
        lto3      : 1,
        lto4      : 1,
        ult3581   : 1,
        ult3582   : 1,
        ult3583   : 1,
        ult3584   : 1,
        ny3581    : 1,
        ult3576   : 1,
        ult3573   : 1,
        lib3577   : 1,
        reserved  : 14;
} dev_typemod_t;
static dev_typemod_t device;

typedef struct {                       /* element descriptors                */
   uint
       medium_elem   : 1,
       storage_elem  : 1,
       impexp_elem   : 1,
       data_elem     : 1,
       reserved      : 28;
} elem_type_t;
static elem_type_t elem_type;

#define PIDLEN          16
static char pid[PIDLEN + 1];            /* Product ID from Inquiry Data     */

static uchar   tape_drv    = 0,
               medium_chgr = 0;

extern int  errno;                      /* where error values are kept       */
static int  interactive;                /* execute mode (menu or cmd)        */
static int  verbose;                    /* output mode (noisy or quiet)      */
static int  opened;                     /* special file status               */
static int  tape_op_mode;               /* MT/ST command mode flag           */
static int  out_to_file;                /* Output the inventory data in a file */
static int  dev_fd;                     /* device special file descriptor    */
static char dev_name[128] = "";         /* device special file name          */
static char ucode_level[32] = "";       /* microcode level                   */
static char driver_level[25] = "";      /* driver level                      */
static char message[128] = "";          /* display message text buffer       */
static char addr[128]= "";              /* move/exchnage medium addr buffer  */
static element_info_t element_info;     /* medium changer element status     */
static inquiry_data_t inquiry_data;
static int devt;                        /*major/minor number of path to en/disable*/
static char path_name[128] = "";        /* device special file name to en/disable */

/*****************************************************************************/
/*                            FUNCTION PROTOTYPES                            */
/*****************************************************************************/

/*------------------------*/
/* Main Control Functions */
/*------------------------*/
extern int  main (int, char **);
static int  cmd_line (int, char **);
static int  new_cmdline(int argc,char ** argv);
static int  interact (void);
static int  get_cmd (parm_t *);
static int  exec_cmd (parm_t *);
static void menu (void);
static void usage (int extended);
static void dump_bytes (char *, int);
static void elem_descriptor (char *, char *, int);
static int  get_pid(char *pid );

/*-------------------*/
/* Service Functions */
/*-------------------*/
static int svc_device_sn (void);
static int svc_ucode_level (void);
static int svc_force_dump (void);
static int svc_store_dump (char *);
static int svc_load_ucode (char *filename);
static int svc_format_tape (void);
static int svc_check_device (void);
static int get_driver_level (void);
static long long read_buflist ( int fp, buffer_io_t *biop, int desc_mode, int data_mode, int
 bufid, ...);

/*----------------------*/
/* Basic SCSI Functions */
/*----------------------*/
static int scsi_test_unit_ready (void);
static int scsi_inquiry (void);
static int scsi_inquiry_page (int);
static int scsi_request_sense (void);
static int scsi_reserve (void);
static int scsi_release (void);
static int scsi_log_sense_page (int, int, int);
static int scsi_mode_sense_page (int);
static int scsi_prevent_medium_rm (void);
static int scsi_allow_medium_rm (void);

/*--------------------------*/
/* Medium Changer Functions */
/*--------------------------*/
static int smc_move_medium (int, int);
static int smc_pos_to_elem (int);
static int smc_element_info (void);
static int smc_inventory (void);
static int smc_audit (void);
static int smc_lock_door (int);
static int smc_exchange_medium (int, int, int);
static int smc_read_element_devids (void);
static int smc_init_elem_stat_range (int, int);
static int smc_read_cartridge_location (int, int);
static int device_paths(void);

/*----------------------*/
/* Tape Drive Functions */
/*----------------------*/
static int st_read (int, int, int);
static int st_write (int, int, int, int, char *);
static int mt_tape_op (int, int);
static int st_tape_op (int, int);
static int st_get_device_status (void);
static int st_get_device_info (void);
static int st_get_media_info (void);
static int st_get_position (int);
static int st_set_position (int, int);
static int st_get_parm (int);
static int st_set_parm (int, int);
static int st_sync_buffer (void);
static int st_display_msg (int, char *, char *);
static int st_report_density (void);
static int rwtest(int blk_size, int blk_fact, int multiplier, int seed);
static int st_get_density(void);
static int st_get_encryption_state (void);
static int rekey_command (void);
static int get_sernum (void);


/*****************************************************************************/
/*                          MAIN CONTROL FUNCTIONS                           */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* Name:        main                                                         */
/* Synopsis:    Determine run mode (interactive or command line)             */
/*              and invoke appropriate controlling function.                 */
/* Returns:     RC returned by interact() or cmd_line().                     */
/*---------------------------------------------------------------------------*/

extern int main ( int argc, char **argv )
{
  int rc;
  int i;


  (void) memset (pid, 0, PIDLEN + 1);
   *((int *)&device) = 0;

  setbuf (stdout, '\0');  /* turn off buffering */
  opened  = FALSE;  /* special file not opened yet */

  if (argc == 1) {  /* interactive mode execution */
    interactive = TRUE;
    verbose = TRUE;
    rc = interact ();
  }
  else {  /* command line mode execution */
    interactive = FALSE;
    verbose = FALSE;
    rc = cmd_line (argc, argv);
  }

  return (rc);
}
/* main */


/*---------------------------------------------------------------------------*/
/* Name:        cmd_line                                                     */
/* Synopsis:    Central controller for command line mode.                    */
/* Returns:     RC returned by exec_cmd().                                   */
/*---------------------------------------------------------------------------*/

static int cmd_line ( int argc, char **argv )
{
  int  rc = 0;
  int  i, j;
  int  opt;
  long longarg;
  char *first_char;
  char **last_char;
  extern int opterr;
  extern char *optarg;
  char dst_addr[128]= "";
  char dst2_addr[128]= "";
  parm_t parm;

  char flags[] = "f:o:w:s:d:c:b:n:m:r:z:t:x:y:p:vh";

  char *cmds[] = {
    "" ,           /*  0 undefined                           */
    "" ,           /*  1 open                                */
    "" ,           /*  2 close                               */
    "qsn",         /*  3 Query Subsystem Number For use with */
                   /*    lmcpd and IBM 3494 Tape Subsystem   */
    "qmc",         /*  4 Query microcode level               */
    "fdp",         /*  5 Force microcode dump                */
    "sdp",         /*  6 Save microcode dump to file         */
    "dmc",         /*  7 Download microcode to drive         */
    "drv",         /*  8 Query driver level                  */
    "tur",         /*  9 SCSI Test unit ready                */
    "inq",         /* 10 SCSI Inquiry/Inquiry Page           */
    "req",         /* 11 SCSI Request sense                  */
    "log",         /* 12 SCSI Log sense page                 */
    "mod",         /* 13 SCSI Mode page                      */
    "res",         /* 14 SCSI Reserve                        */
    "rel",         /* 15 SCSI Release                        */
    "rem",         /* 16 SCSI Prevent/Allow medium removal   */
    "mov",         /* 17 SCSI Move medium                    */
    "exh",         /* 18 SCSI Exchange Medium                */
    "pos",         /* 19 SCSI Position to element            */
    "ele",         /* 20 SCSI Element info from mode data    */
    "inv",         /* 21 SCSI Read element status            */
    "dvc",         /* 22 SCSI Read Elem Status(DVCID=1)      */
    "aud",         /* 23 SCSI Init element status            */
    "ier",         /* 24 SCSI Initial Element Status Range   */
    "lck",         /* 25 SCSI Prevent/Allow medium access    */
    "rea",         /* 26 SCSI Read                           */
    "wri",         /* 27 SCSI Write                          */
    "eof",         /* 28 SCSI Write file mark                */
    "era",         /* 29 SCSI Erase                          */
    "rew",         /* 30 SCSI Rewind                         */
    "ret",         /* 31 SCSI Retension                      */
    "off",         /* 32 Offline: Rewind and unload tape     */
    "lod",         /* 33 SCSI Load/Unload                    */
    "fsf",         /* 34 SCSI Forward space file             */
    "bsf",         /* 35 SCSI Backward space file            */
    "fsr",         /* 36 SCSI Forward space record           */
    "bsr",         /* 37 SCSI Backward space record          */
    "eod",         /* 38 SCSI Space to end of data           */
    "gpa",         /* 39 Get parameter                       */
    "spa",         /* 40 Set parameter                       */
    "syn",         /* 41 Sync; Flush buffered data to tape   */
    "msg",         /* 42 Display message                     */
    "den",         /* 43 Report tape density                 */
    "chk",         /* 44 Check device                        */
    "grs",         /* 45 Get Record size                     */
    "srs",         /* 46 Set Record Size                     */
    "gds",         /* 47 Get Device Status                   */
    "gdi",         /* 48 Get Device informaton               */
    "gmi",         /* 49 Get Media informaton                */
    "gpo",         /* 50 Get position                        */
    "spo",         /* 51 Set position                        */
    "",            /* 52 Set MT/ST command mode              */
    "phs",         /* 53 Display paths to device             */
    "ena",         /* 54 Enable CPF path                     */
    "dis",         /* 55 Disable CPF path                    */
    "rwt",         /* 56 Read/Write test                     */
    "gdn",         /* 57 Get Density                         */
    "",            /* 58 Get Encryption State                */
    NULL           /* End of List                            */
  };

  opterr = 0;
  memset ((char *)&parm, 0, sizeof (parm_t));
  parm.type = -1;

  while ((opt = getopt (argc, argv, flags)) != EOF) {

    first_char = optarg;
    last_char = &first_char;

    switch (opt) {

      case 'f':  /* device special file name */
        strcpy (dev_name, optarg);
      break;

      case 'o':  /* operation */
        for (i = 1; cmds[i]; i++) {
          if ((*optarg != '\0') && (strncmp (cmds[i], optarg, 3) == 0)) {
            parm.opcode = i;
            break;
          }
        }
      break;

      case 'w':  /* open mode */
        i = atoi (optarg);
        if (i == 1)  parm.mode = O_RDWR;
        if (i == 2)  parm.mode = O_RDONLY;
        if (i == 3)  parm.mode = O_WRONLY;
        if (i == 4)  parm.mode = O_APPEND | O_WRONLY;
      break;

      case 's':  /* move/exchange medium/IER source */
        parm.src = atoi (optarg);
      break;

      case 'd':  /* move/exchange medium destination(s) */
        strcpy (addr, optarg);
        for (i=0, j=0; i<16; i++, j++) {
          if ((addr[j] == ' ') || (addr[j] == '\0'))  break;
             dst_addr[i] = addr[j];
        }
        parm.dst = atoi (dst_addr);
        for (i=0, j++; i<16; i++, j++) {
          if ((addr[j] == ' ') || (addr[j] == '\0'))  break;
             dst2_addr[i] = addr[j];
        }
        parm.dst2 = atoi (dst2_addr);
      break;

      case 'c':  /* tape op count */
         if ( tape_op_mode == MT_MODE )
            parm.mt_count = atoi (optarg);
         else
            parm.st_count = atoi (optarg);
      break;

      case 'b':  /* block size for data transfers */
        parm.blk_size = atoi (optarg);
      break;

      case 'n':  /* blocking factor for data transfers */
        parm.blk_fact = atoi (optarg);
      break;

      case 'm':  /* multiple of buffer to read/write */
        parm.multiplier = atoi (optarg);
      break;

      case 'r':  /* random seed */
        parm.seed = atoi (optarg);
      break;

      case 'z':  /* ucode/dump/pattern file name */
        strcpy (parm.filename, optarg);
      break;

      case 't':  /* msg/gpo/spo/gpa/spa/inq/log/inv/dvc type */
        longarg = strtol(optarg, last_char, 10);
        if (**last_char == 0) {
           parm.type = (int)longarg;
        }
       else {
           printf("-t flag requires an integer argument; \"%s\" is invalid\n", optarg);
       }
      break;

      case 'x':  /* lck/lod/rem/spo/spa/page code/IER value */
        longarg = strtol(optarg, last_char, 10);
        if (**last_char == 0) {
           parm.value = (int)longarg;
        }
        else {
           last_char = &optarg;
           longarg = strtol(optarg, last_char, 16);
           if (**last_char == 0) {
               parm.value = (int)longarg;
           }
           else {
               printf("-x flag requires an integer argument in hex; \"%s\" is invalid\n", optarg);
           }
        }
      break;

      case 'y':  /* display message text */
        strcpy (message, optarg);
        for (i=0, j=0; i<8; i++, j++) {
          if ((message[j] == ',') || (message[j] == '\0'))  break;
          parm.msg_0[i] = message[j];
        }
        for (i=0, j++; i<8; i++, j++) {
          if ((message[j] == ',') || (message[j] == '\0'))  break;
          parm.msg_1[i] = message[j];
        }
      break;

    case 'p':
      strcpy (path_name, optarg);
      break;

      case 'v':  /* verbose */
        verbose = TRUE;
      break;

      case 'h':
      case '?':  /* help flag */
        verbose = TRUE;
        usage (EXTENDED);
        return (0);
        /* NOTREACHED */
      break;

      default:  /*  invalid flag */
        verbose = TRUE;
        usage (EXTENDED);
        return (0);
        /* NOTREACHED */
      break;
    }
  }

  /* validate input parameters */

  if (dev_name[0] == '\0') {
    PRINTF ("\nSpecial File Name [-f] always required.\n");
    usage (EXTENDED);
    return (EINVAL);
  }

  if (parm.opcode < 1) {
     rc = new_cmdline(argc,argv);
     return(rc);
  }

  if (parm.opcode > 60) {
    PRINTF ("\nOperation [-o] missing or invalid.\n");
    usage (EXTENDED);
    return (EINVAL);
  }

  if (parm.blk_fact * parm.blk_size > RWBUFSZ) {
    PRINTF ("\nBlock Size [-b] or Blocking Factor [-n] out of range.\n");
    usage (EXTENDED);
    return (EINVAL);
  }

  if (parm.multiplier > RWBUFSZ) {
    PRINTF ("\nMultiplier [-m] out of range.\n");
    usage (EXTENDED);
    return (EINVAL);
  }

  /* check for missing parameters required by specific  */
  /* operations and initialize op field for mt or st commands */

  switch (parm.opcode) {

    case 1:  /* open device */
    break;

    case 2:  /* close device */
    break;

    case 3:  /* query serial number */
    break;

    case 4:  /* query microcode level */
    break;

    case 5:  /* force dump */
    break;

    case 6:  /* store dump */
      if (parm.filename[0] == '\0') {
        PRINTF ("\nSpecified operation requires [-z] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 7:  /* download microcode */
      if (parm.filename[0] == '\0') {
        PRINTF ("\nSpecified operation requires [-z] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 8:  /* query driver level */
    break;

    case 9:  /* test unit ready */
    break;

    case 10:  /* inquiry */
      if ((parm.type < -1) || ( parm.type > 1 )) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 11:  /* request sense */
    break;

    case 12:  /* log sense page */
    break;

    case 13:  /* mode page */
    break;

    case 14:  /* reserve */
    break;

    case 15:  /* release */
    break;

    case 16:  /* prevent/allow medium removal */
      if ((parm.value < 1) || (parm.value > 2)) {
        PRINTF ("\nSpecified operation requires [-x] parm.\n");
      usage (EXTENDED);
      return (EINVAL);
      }
    break;

    case 17:  /* move medium */
      if ((parm.src == '\0') || (parm.dst == '\0')) {
       PRINTF ("\nSpecified operation requires [-s] and [-d] parms for address(s).\n");
       usage (EXTENDED);
       return (EINVAL);
     }
    break;

    case 18:  /* exchange medium */
     if ((parm.src == '\0') || (parm.dst == '\0') || (parm.dst2 == '\0')) {
      PRINTF ("\nSpecified operation requires [-s] and [-d] parms for address(s).\n");
      usage (EXTENDED);
      return (EINVAL);
    }
    break;

    case 19:  /* position to element */
     if ((parm.dst == '\0')) {
      PRINTF ("\nSpecified operation requires [-d] parm for address.\n");
      usage (EXTENDED);
      return (EINVAL);
    }
    break;

    case 20:  /* element information */
    break;

    case 21:  /* inventory */
      if (parm.type < 0)
         out_to_file = 0;
      else
         out_to_file = 1;
    break;

    case 22:  /* read element status (DVCID=1) */
      if (parm.type < 0)
         out_to_file = 0;
      else
         out_to_file = 1;
    break;

    case 23:  /* audit */
    break;

    case 24:  /* initial element status range */
     if ((parm.src == '\0')) {
      PRINTF ("\nSpecified operation requires [-s] parm for address.\n");
      usage (EXTENDED);
      return (EINVAL);
    }
    break;

    case 25:  /* lock/unlock door */
      if ((parm.value < 1) || (parm.value > 2)) {
        PRINTF ("\nSpecified operation requires [-x] parm.\n");
      usage (EXTENDED);
      return (EINVAL);
      }
      if (parm.value == 1)  parm.value = SMC_LOCK;
      else  parm.value = SMC_UNLOCK;
    break;

    case 26:  /* read data */
      if ((parm.blk_size == 0) ||
          (parm.blk_fact == 0) ||
          (parm.multiplier == 0)) {
        PRINTF ("\nSpecified operation requires [-b], [-n] and [-m] parms.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 27:  /* write data */
      if ((parm.blk_size == 0) ||
          (parm.blk_fact == 0) ||
          (parm.multiplier == 0)) {
        PRINTF ("\nSpecified operation requires [-b], [-n] and [-m] parms.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 28:  /* write file mark */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1 .\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTWEOF;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_WEOF;
         tape_op_mode = ST_MODE;
      }
    break;

    case 29:  /* erase tape */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTERASE;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_ERASE;
         tape_op_mode = ST_MODE;
      }
    break;

    case 30:  /* rewind */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTREW;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_REW;
         tape_op_mode = ST_MODE;
      }
    break;

    case 31:  /* retension */
      parm.st_op = ST_OP_RETEN;
      tape_op_mode = ST_MODE;
    break;

    case 32:  /* offline */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTOFFL;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_OFFL;
         tape_op_mode = ST_MODE;
      }
    break;

    case 33:  /* load/unload tape */
      if ((parm.value < 1) || (parm.value > 2)) {
        PRINTF ("\nSpecified operation requires [-x] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
      if (parm.value == 1) {
         parm.st_op = ST_OP_LOAD;
      }
      else {
         parm.st_op = ST_OP_UNLOAD;
      }
      tape_op_mode = ST_MODE;
    break;

    case 34:  /* forward space file */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTFSF;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_FSF;
         tape_op_mode = ST_MODE;
      }
    break;

    case 35:  /* backward space file */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTBSF;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_BSF;
         tape_op_mode = ST_MODE;
      }
    break;

    case 36:  /* forward space record */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTFSR;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_FSR;
         tape_op_mode = ST_MODE;
      }
    break;

    case 37:  /* backward space record */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTBSR;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_BSR;
         tape_op_mode = ST_MODE;
      }
    break;

    case 38:  /* locate end of data */
      if ((parm.type < -1) || ( parm.type > 1)) {
        PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
        usage (EXTENDED);
        return (EINVAL);
      }
      if ( parm.type == 0 ) {
         parm.mt_op = MTEOD;
         tape_op_mode = MT_MODE;
      }
      else {
         parm.st_op = ST_OP_EOD;
         tape_op_mode = ST_MODE;
      }
    break;

    case 39:  /* get parameter */
      if (parm.type == -1) {
        PRINTF ("\nSpecified operation requires [-t] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 40:  /* set parameter */
      if (parm.type == -1) {
        PRINTF ("\nSpecified operation requires [-t] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 41:  /* sync buffer */
    break;

    case 42:  /* display message */
      if ((parm.msg_0[0] == '\0') || (parm.msg_1[0] == '\0')) {
        PRINTF ("\nSpecified operation requires [-y] parm for message(s).\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 43:  /* report density support */
    break;

    case 44:  /* check device */
    break;

    case 45:  /* get record size */
      parm.st_op = ST_OP_GRSZ;
      tape_op_mode = ST_MODE;
    break;

    case 46:  /* set record size */
      parm.st_op = ST_OP_SRSZ;
      tape_op_mode = ST_MODE;
    break;

    case 47:  /* get device status */
    break;

    case 48:  /* get device information */
    break;

    case 49:  /* get media information */
    break;

    case 50:  /* get position */
      if (parm.type == -1) {
        PRINTF ("\nSpecified operation requires [-t] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 51:  /* set position */
      if (parm.type == -1) {
        PRINTF ("\nSpecified operation requires [-t] parm.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 52:  /* set MT/ST command mode */
    break;

    case 53:  /* display paths */
    break;

    case 54:  /* enable path */
    break;

    case 55:  /* disable path */
    break;

    case 56:  /* read/write test */
      if ((parm.blk_size == 0) ||
          (parm.blk_fact == 0) ||
          (parm.multiplier == 0)) {
        PRINTF ("\nSpecified operation requires [-b], [-n] and [-m] parms.\n");
        usage (EXTENDED);
        return (EINVAL);
      }
    break;

    case 57: /*get density */
    break;

    case 58: /*get encryption state */
    break;


  }

  /* open device, issue command, and close device */
  if ((dev_fd = open (dev_name, parm.mode | O_NDELAY)) < 0) {
    PERROR ("\nFailure opening device special file");
    PRINTF ("\n");
    return (errno);
  }

  rc = get_pid(pid);
  if (rc != 0) {
    (void) fprintf (stderr, "Failure obtaining device PID\n");
    return (EINVAL);
  }

  PRINTF ("\n");
  rc = exec_cmd (&parm);
  PRINTF ("\n");
  close (dev_fd);
  return (rc);
}
/* cmd_line */

/*---------------------------------------------------------------------------*/
/* Name:        new_cmdline                                                  */
/*---------------------------------------------------------------------------*/
static int new_cmdline(int argc,char ** argv)
   {
   int i;
   int rc = 0;
   int arg_count = 1;
   char *argp;
   parm_t parm;
   int err_stop = TRUE;


   memset ((char *)&parm, 0, sizeof (parm_t));
   parm.type = -1;
   verbose = TRUE;


   /* check first arg for device */
   if (strncmp("-f",argv[1],2))
      {
      exit(1);
      }

   if (strlen(argv[1]) == 2)
      {
      arg_count++;
      if (arg_count > argc)
      {
      printf("Missing argument for device\n");
      exit(1);
      }
      strcpy(dev_name,argv[2]);
      }
   else
      strcpy(dev_name,argv[1]+2);

   arg_count++;
   if (arg_count > argc)
      {
      printf("Missing subcommand\n");
      exit(1);
      }

   /* open device, issue command, and close device */
   dev_fd = -1;
   if (strlen(dev_name) > 0) {
      dev_fd = open (dev_name, O_RDONLY | O_NDELAY);
      if (dev_fd > 0) {
         opened    = TRUE;
      }
   }

     if (opened) {
      rc = get_pid(pid);
      if (rc != 0) {
         (void) fprintf (stderr, "Failure obtaining device PID\n");
         return (EINVAL);
      }
   }
   else {
      if (errno && (errno != ENOENT)) {
         (void) fprintf (stderr, "Error %d (%s) opening %s.\n",
            errno, strerror(errno), dev_name);
         return (errno);
      }
      else if (errno == ENOENT) {
         (void) fprintf (stderr, "Error %d (%s) opening %s.\n",
            errno, strerror(errno), dev_name);
         return (errno);
      }
      else {
         (void) fprintf (stderr, "tapeutil error opening %s.\n", dev_name);
         return (-1);
      }
   }

   for (i = arg_count;; i++)
   {
           fflush(stdout);
           if (rc != 0)
           {
              /* sprintf(msg,"Operation failed with errno %d",errno);
              perror (msg);
              scsi_request_sense(); */
              if (err_stop)
                  break;
           }

           errno = 0;
           rc = 0;
           argp = argv[i];

           /* encryption subcommand */
           if (strcmp(argp, "encryption") == 0)
           {
              rc = st_get_encryption_state();
              continue;
           }

           /* get serial number subcommand */
           if (strcmp(argp, "sernum") == 0)
           {
              rc = get_sernum ();
              continue;
           }

           /* rekey subcommand */
           if (strcmp(argp, "rekey") == 0)
           {
                   rc = rekey_command();
                   continue;
           }

           /* read element slot ids subcommand */
           if (strncmp(argp, "cartridgelocation", 17) == 0)
           {
              i++;
              if (i >= argc)
              {
                 printf("Missing argument for cartridgelocation subcommand\n");
                 exit(1);
              }
              /* check for slot argument */
              if (isdigit(argv[i][0]))
                 parm.src = atoi(argv[i]);

              if ((i + 1) < argc && isdigit(argv[i+1][0]))
              {
                 i++;
                 parm.value = atoi(argv[i]);
              }
              else
                 parm.value = 1;

              rc = smc_read_cartridge_location (parm.src,parm.value);
              continue;
           }


          /* printf("Invalid argument: %s\n",argp); */
           exit(1);
   }

   close(dev_fd);                /* Close the device */
   return rc;
}


/*---------------------------------------------------------------------------*/
/* Name:        interact                                                     */
/* Synopsis:    Central controller for interactive mode.                     */
/* Returns:     RC returned by exec_cmd().                                   */
/*---------------------------------------------------------------------------*/

static int interact ()
{
  int rc = 0;
  parm_t parm;

  printf ("\n");
  menu ();

  do {
    memset ((char *)&parm, 0, sizeof (parm_t));

    while (!get_cmd (&parm));  /* get valid command and all parms */
    if (parm.opcode) {
      rc = exec_cmd (&parm);  /* execute cmd if not quit */
    }
  } while (parm.opcode);  /* if 0, then command was quit */

  if (opened) {
    close (dev_fd);  /* close special file if opened */
  }

  return (rc);
}
/* interact */


/*---------------------------------------------------------------------------*/
/* Name:        get_cmd                                                      */
/* Synopsis:    Obtain command selection and appropriate parameters.         */
/* Returns:     TRUE for valid option, FALSE for invalid/missing option.     */
/*---------------------------------------------------------------------------*/

static int get_cmd ( parm_t *parm )
{
  int temp;
  char buf[50] = "";

  printf ("\nENTER COMMAND ('M' for Menu) ==> ");

  fflush (stdin);
  gets (buf);
  printf ("\n");
  fflush (stdout);

  if (*buf == '\0') {
    return (FALSE);
  }

  if ((*buf == 'D') || (*buf == 'd')) {  /* device type */
    sprintf (buf, "%d", 44);
   }

  if ((*buf == 'M') || (*buf == 'm')) {  /* menu refresh */
    menu ();
    return (FALSE);
  }

  if ((*buf == 'Q') || (*buf == 'q')) {  /* quit program */
    parm->opcode = 0;
    return (TRUE);
  }

  if (*buf == '!') {  /* Pass a command to the system. */
    (void) system ((char *)(&buf[1]));
    return (FALSE);
  }

  parm->opcode = atoi (buf);

  if ((parm->opcode < 1) || (parm->opcode > 62)) {
    printf ("Invalid command selection.\n");
    return (FALSE);
  }

  if ((!opened) && (parm->opcode != 1)) { /* must be opened first */
    printf ("Device not opened.\n");
    return (FALSE);
  }

  switch (parm->opcode) {

    case 1:  /* open device */
      if (opened) {
        printf ("Device %s already open.\n", dev_name);
        return (FALSE);
      }
      dev_name[0] = '\0';
      printf ("Enter device special file name (e.g., /dev/rmt/1mnb): ");
      gets (dev_name);
      printf ("Select open mode");
      printf (" (1=read-write, 2=read, 3=write, 4=append): ");
      gets (buf);
      temp = atoi (buf);
      if ((temp < 1) || (temp > 4)) {
        printf ("\nInvalid open mode selection.\n");
        return (FALSE);
      }
      if (temp == 1)  parm->mode = O_RDWR;
      if (temp == 2)  parm->mode = O_RDONLY;
      if (temp == 3)  parm->mode = O_WRONLY;
      if (temp == 4)  parm->mode = O_APPEND | O_WRONLY;
      printf ("\n");
    break;

    case 2:  /* close device */
    break;

    case 3:  /* query serial number */
    break;

    case 4:  /* query microcode level */
    break;

    case 5:  /* force dump */
    break;

    case 6:  /* store dump */
      printf ("Enter dump destination file name\n");
      printf (" ('tapedrive' = store on cartridge in drive): ");
      gets (parm->filename);
      if (parm->filename[0] == '\0') {
        printf ("\nInvalid file name.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 7:  /* download microcode */
      printf ("Enter microcode source file name: ");
      gets (parm->filename);
      if (parm->filename[0] == '\0') {
        printf ("\nInvalid file name.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 8:  /* query driver level */
    break;

    case 9:  /* test unit ready */
    break;

    case 10:  /* inquiry */
      printf ("Enter Inquiry Page code in hex or <enter> for standard inquiry: ");
      gets (buf);
      if ( buf[0] == '\0' )
        parm->type = 0;
      else {
        parm->type = 1;
        parm->value = (int) strtol(buf, NULL, 16);
      }
      printf ("\n");
    break;

    case 11:  /* request sense */
    break;

    case 12:  /* log sense page */
      printf ("Enter Log Page code in hex: ");
      gets (buf);
      parm->value = (int) strtol (buf, NULL, 16);
      printf ("\nEnter parameter code in hex or <enter> for all parameters: ");
      gets (buf);
       if ( buf[0] == '\0' )
        parm->type = 0;
      else {
        parm->type = 1;
        parm->parm_code = strtol(buf,NULL,16);
      }
      printf ("\n");
    break;

    case 13:  /* mode page */
      printf ("Enter Mode Page code in hex: ");
      gets (buf);
      parm->value = (int) strtol (buf, NULL, 16);
      printf ("\n");
    break;

    case 14:  /* reserve */
    break;

    case 15:  /* release */
    break;

    case 16:  /* prevent/allow medium removal */
      printf ("Enter selection");
      printf (" (1=prevent, 2=allow): ");
      gets (buf);
      parm->value = atoi (buf);
      if ((parm->value < 1) || (parm->value > 2)) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 17:  /* move medium */
      printf ("Enter source address: ");
      gets (buf);
      parm->src = atoi (buf);
      if ((parm->src < 0) || (parm->src == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("Enter destination address: ");
      gets (buf);
      parm->dst = atoi (buf);
      if ((parm->dst < 0) || (parm->dst == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 18:  /* exchange medium */
      printf ("Enter source address: ");
      gets (buf);
      parm->src = atoi (buf);
      if ((parm->src < 0) || (parm->src == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("Enter the first destination address: ");
      gets (buf);
      parm->dst = atoi (buf);
      if ((parm->dst < 0) || (parm->dst == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("Enter the second destination address: ");
      gets (buf);
      parm->dst2 = atoi (buf);
      if ((parm->dst2 < 0) || (parm->dst2 == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 19:  /* position to element */
      printf ("Enter destination address: ");
      gets (buf);
      parm->dst = atoi (buf);
      printf ("\n");
    break;

    case 20:  /* element information */
    break;

    case 21:  /* inventory */
    break;

    case 22:  /* read elem status with DVCID=1 */
    break;

    case 23:  /* audit */
    break;

    case 24:  /* initial elem status range */
      printf ("Enter the first slot address: ");
      gets (buf);
      parm->src = atoi (buf);
      if ((parm->src < 0) || (parm->src == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("Enter the slot number checked: ");
      gets (buf);
      parm->value = atoi (buf);
      if ((parm->value <= 0) || (parm->value == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    break;

    case 25:  /* lock/unlock door */
      printf ("Enter selection");
      printf (" (1=lock, 2=unlock): ");
      gets (buf);
      parm->value = atoi (buf);
      if ((parm->value < 1) || (parm->value > 2)) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      if (parm->value == 1)  parm->value = SMC_LOCK;
      else  parm->value = SMC_UNLOCK;
      printf ("\n");
    break;

    case 26:  /* read data */
      printf ("Total Bytes Read = ");
      printf ("(block size) * (blocking factor) * (multiplier).\n\n");
      printf ("Enter block size: ");
      gets (buf);
      parm->blk_size = atoi (buf);
      if ((parm->blk_size < 1) || (parm->blk_size > RWBUFSZ)) {
        printf ("\nBlock size out of range.\n");
        return (FALSE);
      }
      printf("Enter blocking factor: ");
      gets (buf);
      parm->blk_fact = atoi (buf);
      if ((parm->blk_fact < 1) ||
          ((parm->blk_fact * parm->blk_size) > RWBUFSZ)) {
        printf ("\nBlocking factor out of range.\n");
        return (FALSE);
      }
      printf ("Enter multiplier: ");
      gets (buf);
      parm->multiplier = atoi (buf);
      if ((parm->multiplier < 1) || (parm->multiplier > RWBUFSZ)) {
        printf ("\nMultiplier out of range.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 27:  /* write data */
      printf ("Total Bytes Written = ");
      printf ("(block size) * (blocking factor) * (multiplier).\n\n");
      printf ("Enter block size: ");
      gets (buf);
      parm->blk_size = atoi (buf);
      if ((parm->blk_size < 1) || (parm->blk_size > RWBUFSZ)) {
        printf ("\nBlock size out of range.\n");
        return (FALSE);
      }
      printf ("Enter blocking factor: ");
      gets (buf);
      parm->blk_fact = atoi (buf);
      if ((parm->blk_fact < 1) ||
          ((parm->blk_fact * parm->blk_size) > RWBUFSZ)) {
        printf ("\nBlocking factor out of range.\n");
        return (FALSE);
      }
      printf ("Enter multiplier: ");
      gets (buf);
      parm->multiplier = atoi (buf);
      if ((parm->multiplier < 1) || (parm->multiplier > RWBUFSZ)) {
        printf ("\nMultiplier out of range.\n");
        return (FALSE);
      }
      printf ("Enter pattern file name");
      printf (" (<ret> = random pattern): ");
      gets (parm->filename);
      if (parm->filename[0] == '\0') {  /* create our own random pattern */
        printf ("Enter random seed");
        printf (" (0 = auto generated seed): ");
        gets (buf);
        parm->seed = atoi (buf);
        if (parm->seed < 0) {
          printf ("\nRandom seed out of range.\n");
          return (FALSE);
        }
      }
      printf ("\n");
    break;

    case 28:  /* write file mark */
      printf ("Enter number of file marks to write: ");
      gets (buf);
      if ( tape_op_mode == MT_MODE ) {
          parm->mt_op = MTWEOF;
          parm->mt_count = atoi (buf);
      }
      else {
          parm->st_op = ST_OP_WEOF;
          parm->st_count = atoi (buf);
      }
      printf ("\n");
    break;

    case 29:  /* erase tape */
      if ( tape_op_mode == MT_MODE ) parm->mt_op = MTERASE;
      else parm->st_op = ST_OP_ERASE;
    break;

    case 30:  /* rewind */
      if ( tape_op_mode == MT_MODE ) parm->mt_op = MTREW;
      else parm->st_op = ST_OP_REW;
    break;

    case 31:  /* retension */
      parm->st_op = ST_OP_RETEN;
    break;

    case 32:  /* offline */
      if ( tape_op_mode == MT_MODE ) parm->mt_op = MTOFFL;
      else parm->st_op = ST_OP_OFFL;
    break;

    case 33:  /* load/unload tape */
      printf ("Enter selection");
      printf (" (1=load, 2=unload): ");
      gets (buf);
      parm->value = atoi (buf);
      if ((parm->value < 1) || (parm->value > 2)) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      if (parm->value == 1)  parm->st_op = ST_OP_LOAD;
      else  parm->st_op = ST_OP_UNLOAD;
      printf ("\n");
    break;

    case 34:  /* forward space file */
      printf ("Enter number of files to space forward: ");
      gets (buf);
      if ( tape_op_mode == MT_MODE ) {
          parm->mt_op = MTFSF;
          parm->mt_count = atoi (buf);
      }
      else {
          parm->st_op = ST_OP_FSF;
          parm->st_count = atoi (buf);
      }
      printf ("\n");
    break;

    case 35:  /* backward space file */
      printf ("Enter number of files to space backward: ");
      gets (buf);
      if ( tape_op_mode == MT_MODE ) {
           parm->mt_op = MTBSF;
           parm->mt_count = atoi (buf);
      }
      else {
           parm->st_op = ST_OP_BSF;
           parm->st_count = atoi (buf);
      }
      printf ("\n");
    break;

    case 36:  /* forward space record */
      printf ("Enter number of records to space forward: ");
      gets (buf);
      if ( tape_op_mode == MT_MODE ) {
           parm->mt_op = MTFSR;
           parm->mt_count = atoi (buf);
      }
      else {
           parm->st_op = ST_OP_FSR;
           parm->st_count = atoi (buf);
      }
      printf ("\n");
    break;

    case 37:  /* backward space record */
      printf ("Enter number of records to space backward: ");
      gets (buf);
      if ( tape_op_mode == MT_MODE ) {
           parm->mt_op = MTBSR;
           parm->mt_count = atoi (buf);
      }
      else {
           parm->st_op = ST_OP_BSR;
           parm->st_count = atoi (buf);
      }
      printf ("\n");
    break;

    case 38:  /* locate end of data */
      if ( tape_op_mode == MT_MODE ) parm->mt_op = MTEOD;
      else parm->st_op = ST_OP_EOD;
    break;

    case 39:  /* get parameter */
      printf ("Get Parameter:\n");
      printf ("Select parameter type\n");
      printf (" (0=all, 1=blocksize, 2=compression, 3=buffering, 4=immediate 5=trailer)\n");
      printf (" (6=writeprotect, 7=acfmode, 8=capacity, 9=sili, 10=release): ");
      gets (buf);
      parm->type = atoi (buf);
      if ((parm->type < 0) || (parm->type > 10)) {
        printf ("\nInvalid parameter type selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 40:  /* set parameter */
      st_get_parm (0);
      printf ("\nSet Parameter:\n");
      printf ("Select parameter type\n");
      printf (" (1=blocksize, 2=compression, 3=buffering, 4=immediate 5=trailer)\n");
      printf (" (6=writeprotect, 7=acfmode, 8=capacity, 9=sili, 10=release): ");
      gets (buf);
      parm->type = atoi (buf);
      if ((parm->type < 1) || (parm->type > 10) || (parm->type == 7)) {
        printf ("\nInvalid parameter type selection.\n");
        return (FALSE);
      }
      if (parm->type == 6)
         printf ("Protection Modes: 0=NO_PROTECT, 1=WORM_PROTECT, 16=PERS_PROTECT, 64=ASSC_PROTECT\n");
      if (parm->type == 10)
         printf ("Release options: 0=No special release, 6=RELEASE(6), 10=RELEASE(10)\n");
      printf ("Enter parameter value: ");
      if (parm->type == 8) {
            if (device.t3592)
               printf("Enter capacity scaling (1=100%%, 2=75%%, 3=50%%, 4=25%% or other hexadecimal value): ");
            else
               printf("Enter capacity scaling (1=100%%, 2=75%%, 3=50%%, 4=25%%): ");
            gets (buf);
            switch(atoi(buf)) {
            case 1:
               parm->value = SCALE_100;
               break;
            case 2:
               parm->value = SCALE_75;
               break;
            case 3:
               parm->value = SCALE_50;
               break;
            case 4:
               parm->value = SCALE_25;
               break;
            default:
               parm->value = (int)strtol(buf, NULL, 16);
               break;
            }
      }
      else {
         gets (buf);
         parm->value = atoi (buf);
      }
      printf ("\n");
    break;

    case 41:  /* sync buffer */
    break;

    case 42:  /* display message */
      printf ("Select message type");
      printf (" (1=msg0, 2=msg1, 3=flash0, 4=flash1, 5=alternate): ");
      gets (buf);
      parm->type = atoi (buf);
      if ((parm->type < 1) || (parm->type > 5)) {
        printf ("\nInvalid message type selection.\n");
        return (FALSE);
      }
      switch(parm->type){
        case 1:
        case 3:
        case 5:
          printf ("Enter message-0 text (8 characters or less): ");
          gets (buf);
          if ((strlen (buf) < 1) || (strlen (buf) > 8)) {
            printf ("\nInvalid message-0 length.\n");
            return (FALSE);
          }
          strcpy (parm->msg_0, buf);
        break;
      } /*switch*/

      switch(parm->type){
        case 2:
        case 4:
        case 5:
          printf ("Enter message-1 text (8 characters or less): ");
          gets (buf);
          if ((strlen (buf) < 1) || (strlen (buf) > 8)) {
            printf ("\nInvalid message-1 length.\n");
            return (FALSE);
          }
          strcpy (parm->msg_1, buf);
        break;
      } /*switch*/
      printf ("\n");
    break;

    case 43:  /* report density support */
    break;

    case 44:  /* check device */
    break;

    case 45:  /* get record size */
      parm->st_op = ST_OP_GRSZ;
    break;

    case 46:  /* set record size */
      printf ("Enter new record size: ");
      gets (buf);
      parm->st_count = atoi (buf);
      parm->st_op = ST_OP_SRSZ;
      printf ("\n");
    break;

    case 47:  /* get device status */
    break;

    case 48:  /* get device information */
    break;

    case 49:  /* get media information */
    break;

    case 50:  /* get position */
      printf ("Get Position:\n");
      printf ("Select block type");
      printf (" (1=logical, 2=physical): ");
      gets (buf);
      parm->type = atoi (buf);
      if ((parm->type < 1) || (parm->type > 2)) {
        printf ("\nInvalid block type selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;

    case 51:  /* set position */
      printf ("Set Position:\n");
      printf ("Select block type");
      printf (" (1=logical, 2=physical): ");
      gets (buf);
      parm->type = atoi (buf);
      if ((parm->type < 1) || (parm->type > 2)) {
        printf ("\nInvalid block type selection.\n");
        return (FALSE);
      }
      printf ("Enter position value: ");
      gets (buf);
      parm->value = atoi (buf);
      printf ("\n");
    break;

    case 52:  /* set MT/ST command mode */
      printf ("Choose MT/ST command mode (0 for MT and 1 or <enter> for ST mode): ");
      gets (buf);
      parm->value = atoi (buf);
      if ( parm->value == 0 )
        tape_op_mode = MT_MODE;
      else
        tape_op_mode = ST_MODE;
    break;

    case 53: /* display path*/
    break;

    case 54: /* enable path */
      path_name[0]='\0';
      printf ("Enter device special file name (e.g., /dev/rmt/1mnb): ");
      gets (path_name);
    break;

    case 55: /* disable path */
      path_name[0]='\0';
      printf ("Enter device special file name (e.g., /dev/rmt/1mnb): ");
      gets (path_name);
    break;

    case 56: /* read/write test */
      printf ("Total Bytes Written/Read = ");
      printf ("(block size) * (blocking factor) * (multiplier).\n\n");
      printf ("Enter block size: ");
      gets (buf);
      parm->blk_size = atoi (buf);
      if ((parm->blk_size < 1) || (parm->blk_size > RWBUFSZ)) {
        printf ("\nBlock size out of range.\n");
        return (FALSE);
      }
      printf ("Enter blocking factor: ");
      gets (buf);
      parm->blk_fact = atoi (buf);
      if ((parm->blk_fact < 1) ||
          ((parm->blk_fact * parm->blk_size) > RWBUFSZ)) {
        printf ("\nBlocking factor out of range.\n");
        return (FALSE);
      }
      printf ("Enter multiplier: ");
      gets (buf);
      parm->multiplier = atoi (buf);
      if ((parm->multiplier < 1) || (parm->multiplier > RWBUFSZ)) {
        printf ("\nMultiplier out of range.\n");
        return (FALSE);
      }
      printf ("Enter random seed");
      printf (" (0 = auto generated seed): ");
      gets (buf);
      parm->seed = atoi (buf);
      if (parm->seed < 0)
        {
          printf ("\nRandom seed out of range.\n");
          return (FALSE);
        }
      printf ("\n");
    break;

    case 57: /* get density */
    break;

    case 58: /* get encryption state */
    break;

    case 59:  /* read element slot id */
      printf ("Enter address of first element: ");
      gets (buf);
      parm->src = atoi (buf);
      printf ("Enter number of elements: ");
      gets (buf);
      parm->value = atoi (buf);
      if ((parm->value <= 0) || (parm->value == '\0')) {
        printf ("\nInvalid selection.\n");
        return (FALSE);
      }
      printf ("\n");
    break;


  }

  return (TRUE);
}
/* get_cmd */


/*---------------------------------------------------------------------------*/
/* Name:        exec_cmd                                                     */
/* Synopsis:    Execute command indicated by parm->opcode.                   */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int exec_cmd ( parm_t *parm )
{
  int page;
  int rc;
  int path_fd=0;
  struct stat statbuf;

  switch (parm->opcode) {
    case 1:  /* open device */
      if ((dev_fd = open (dev_name, parm->mode | O_NDELAY)) < 0) {
        PERROR ("Open failed");
        opened = FALSE;
        rc = errno;
      }
      else {
        PRINTF ("Device %s opened.\n", dev_name);
        opened = TRUE;
        rc = get_pid(pid);
        if (rc != 0) {
           (void) fprintf (stderr, "Failure obtaining device PID\n");
           return (EINVAL);
        }
        menu();
        rc = 0;
      }
    break;

    case 2:  /* close device */
      rc = close (dev_fd);
      if (rc < 0) {
         (void) fprintf (stderr, "Error %d (%s) closing %s\n",
             errno, strerror(errno), dev_name);
         opened = FALSE;
      }
      else {
         PRINTF ("Device %s closed.\n", dev_name);
         opened = FALSE;
      }
    break;

    case 3:  /* query serial number */
      rc = svc_device_sn ();
    break;

    case 4:  /* query microcode level */
      rc = svc_ucode_level ();
    break;

    case 5:  /* force dump */
      rc = svc_force_dump ();
    break;

    case 6:  /* store dump */
      rc = svc_store_dump (parm->filename);
    break;

    case 7:  /* download microcode */
      rc = svc_load_ucode (parm->filename);
    break;

    case 8:  /* query driver level */
      rc = get_driver_level ();
    break;

    case 9:  /* test unit ready */
      rc = scsi_test_unit_ready ();
    break;

    case 10:  /* inquiry */
      if (( parm->type == 0 ) || ( parm->type == -1 ))
          rc = scsi_inquiry ();
      else {
          page = parm->value & 0x00FF;    /* Mask for page code */
          rc = scsi_inquiry_page (page);
      }
    break;

    case 11:  /* request sense */
      rc = scsi_request_sense ();
    break;

    case 12: /* log sense page */
      page = parm->value & 0x00FF;    /* Mask for page code */
      rc = scsi_log_sense_page (page, parm->type, parm->parm_code);
    break;

    case 13: /* mode page */
      page = parm->value & 0x00FF;    /* Mask for page code */
      rc = scsi_mode_sense_page (page);
    break;

    case 14:  /* reserve */
      rc = scsi_reserve ();
    break;

    case 15:  /* release */
      rc = scsi_release ();
    break;

    case 16:  /* prevent/allow medium removal */
      if ( parm->value == 1 )
        rc = scsi_prevent_medium_rm ();
      else
        rc = scsi_allow_medium_rm ();
    break;

    case 17:  /* move medium */
      rc = smc_move_medium (parm->src, parm->dst);
    break;

    case 18:  /* exchange medium */
      rc = smc_exchange_medium (parm->src, parm->dst, parm->dst2);
    break;

    case 19:  /* position to element */
      rc = smc_pos_to_elem (parm->dst);
    break;

    case 20:  /* element information */
      rc = smc_element_info ();
    break;

    case 21:  /* inventory */
      rc = smc_inventory ();
    break;

    case 22:  /* read elem status with DVCID=1 */
      rc = smc_read_element_devids ();
    break;

    case 23:  /* audit */
      rc = smc_audit ();
    break;

    case 24:  /* initial element status range */
      rc = smc_init_elem_stat_range (parm->src, parm->value);
    break;

    case 25:  /* lock/unlock door */
      rc = smc_lock_door (parm->value);
    break;

    case 26:  /* read data */
      rc = st_read (parm->blk_size, parm->blk_fact, parm->multiplier);
    break;

    case 27:  /* write data */
      rc = st_write (parm->blk_size, parm->blk_fact,
         parm->multiplier, parm->seed, parm->filename);
    break;

    case 28:  /* write file mark */
    case 29:  /* erase tape */
    case 30:  /* rewind */
    case 32:  /* offline */
    case 34:  /* forward space file */
    case 35:  /* backward space file */
    case 36:  /* forward space record */
    case 37:  /* backward space record */
    case 38:  /* locate end of data */
      if ( tape_op_mode == MT_MODE )
         rc = mt_tape_op (parm->mt_op, parm->mt_count);
      else
         rc = st_tape_op (parm->st_op, parm->st_count);
    break;

    case 31:  /* retension */
    case 33:  /* load/unload tape */
      rc = st_tape_op (parm->st_op, parm->st_count);
    break;

    case 39:  /* get parameter */
      rc = st_get_parm (parm->type);
    break;

    case 40:  /* set parameter */
      rc = st_set_parm (parm->type, parm->value);
    break;

    case 41:  /* sync buffer */
      rc = st_sync_buffer ();
    break;

    case 42:  /* display_message */
      rc = st_display_msg (parm->type, parm->msg_0, parm->msg_1);
    break;

    case 43:  /* report density support */
      rc = st_report_density ();
    break;

    case 44:  /* check device */
      rc = svc_check_device ();
    break;

    case 45:  /* get record size */
    case 46:  /* set record size */
      rc = st_tape_op (parm->st_op, parm->st_count);
    break;

    case 47:  /* get device status */
      rc = st_get_device_status ();
    break;

    case 48:  /* get device information */
      rc = st_get_device_info ();
    break;

    case 49:  /* get media information */
      rc = st_get_media_info ();
    break;

    case 50:  /* get position */
      rc = st_get_position (parm->type);
    break;

    case 51:  /* set position */
      rc = st_set_position (parm->type, parm->value);
    break;

    case 52:  /* set MT/ST command mode */
    break;

    case 53: /* dispaly paths*/
      rc = device_paths();
    break;

    case 54:  /* enable path*/
      if (stat(path_name,&statbuf)!=0)
        {
          printf("Unable to stat path.\n");
          return -1;
        }

      if ((statbuf.st_rdev)&0xF00)
        {
          dev_t tempdev=(statbuf.st_rdev)&0xE00;
          tempdev>>=1; // this is the same as shift left 1 and 0xF00
          (statbuf.st_rdev)&=0xFFFFF0FF;
          (statbuf.st_rdev)|=tempdev;
        }

      devt=statbuf.st_rdev;

      if(!(rc = ioctl(dev_fd, STIOC_ENABLE_PATH, &devt)))
        {
          PRINTF("SCSI path enabled. \n");
        }
      else
        {
          PRINTF("Unabled to enable SCSI path, make sure this path is to the same library as the opened path. \n Run Display Paths to see what paths are connected to the opened path.\n");
        }

    break;

    case 55: /* disable path*/
      if (!strcmp(dev_name, path_name))
      {
        printf("Can not disable primary path.\n");
          return -1;
        }

      if (stat(path_name,&statbuf)!=0)
        {
          printf("Unable to stat path.\n");
          return -1;
        }
      if ((statbuf.st_rdev)&0xF00)
        {
          dev_t tempdev=(statbuf.st_rdev)&0xE00;
          tempdev>>=1; // this is the same as shift left 1 and 0xF00
          (statbuf.st_rdev)&=0xFFFFF0FF;
          (statbuf.st_rdev)|=tempdev;
        }
      devt=statbuf.st_rdev;

      if(!(rc = ioctl(dev_fd, STIOC_DISABLE_PATH, &devt)))
        {
          PRINTF("SCSI path disabled. \n");
        }
      else
        {
          PRINTF("Unabled to disable SCSI path, make sure this path is to the same library as the opened path.\n Run Display Paths to see what paths are connected to the opened path.\n");
        }

    break;

    case 56:  /* read/write test */
       rc=rwtest(parm->blk_size, parm->blk_fact, parm->multiplier,parm->seed);
    break;

    case 57: /* get density */
       rc=st_get_density();
    break;

    case 58: /* get encryption state */
       rc=st_get_encryption_state ();
    break;

    case 59:  /* read element slot id */
       rc = smc_read_cartridge_location (parm->src, parm->mt_count);
    break;
  }

  return (rc);
}
/* exec_cmd */


/*---------------------------------------------------------------------------*/
/* Name:        menu                                                         */
/* Synopsis:    Display menu.                                                */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/

static void menu ()
{
          printf (" +---------------------------------------------------------------+ \n");
          printf (" |      IBM SCSI TAPE & MEDIUM CHANGER UTILITY PROGRAM           | \n");
          printf (" +------------------------------+--------------------------------+ \n");
          printf (" |  << GENERAL COMMANDS >>      |    << SERVICE COMMANDS >>      | \n");
          printf (" |   1: Open Device             |   3: Query Serial Number       | \n");
          printf (" |   2: Close Device            |   4: Query Microcode Level     | \n");
          printf (" |   D: Device Type             |   5: Force Dump                | \n");
          printf (" |   M: Menu Refresh            |   6: Store Dump                | \n");
          printf (" |   Q: Quit Program            |   7: Download Microcode        | \n");
          printf (" |                              |   8: Query Driver Level        | \n");
          printf (" +------------------------------+--------------------------------+ \n");
          printf (" |                   << BASIC SCSI COMMANDS >>                   | \n");
          printf (" |   9: Test Unit Ready   10: Inquiry      11: Request Sense     | \n");
          printf (" |  12: Log Sense Page    13: Mode Page    14: Reserve on close  | \n");
          printf (" |  15: Release           16: Prevent/Allow Media Removal        | \n");
          printf (" +------------------------------+--------------------------------+ \n");

       if (opened && medium_chgr) {
          printf (" |                 << MEDIUM CHANGER COMMANDS >>                 | \n");
          printf (" |  17: Move Medium    18: Exchange Medium  19: Position To Elem | \n");
          printf (" |  20: Element Info   21: Inventory        22: Read Device IDs  | \n");
          printf (" |  23: Audit          24: Audit Range      25: Lock/Unlock Door | \n");
          printf (" |  59: Read Cartridge Location                                  | \n");
          printf (" +------------------------------+--------------------------------+ \n");
          printf (" |                      << CPF COMMANDS >>                       | \n");
          printf (" |  53: Display Paths  54: Enable Path      55: Disable Path     | \n");
          printf (" +------------------------------+--------------------------------+ \n");
       }

      if (opened && tape_drv) {
          printf (" |                 << TAPE DRIVE COMMANDS >>                     | \n");
          printf (" |  26: Read Data                  27: Write Data                | \n");
          printf (" |  28: Write File Mark            29: Erase Tape                | \n");
          printf (" |  30: Rewind                     31: Retension                 | \n");
          printf (" |  32: Offline                    33: Load/Unload Tape          | \n");
          printf (" |  34: Forward Space File         35: Backward Space File       | \n");
          printf (" |  36: Forward Space Record       37: Backward Space Record     | \n");
          printf (" |  38: Locate End Of Data         39: Get Parameter             | \n");
          printf (" |  40: Set Parameter              41: Sync Buffer               | \n");
          printf (" |  42: Display Message            43: Report Tape Density       | \n");
          printf (" |  44: Check Device Type          45: Get Record Size           | \n");
          printf (" |  46: Set Record Size            47: Get Device Status         | \n");
          printf (" |  48: Get Device Info            49: Get Media Info            | \n");
          printf (" |  50: Get Position               51: Set Position              | \n");
          printf (" |  52: Set MT/ST Mode             56: Read/Write Test           | \n");
          printf (" |  57: Get Density                58: Gen Encryption State      | \n");
          printf (" +---------------------------------------------------------------+ \n");
          printf (" |                      << DPF COMMANDS >>                       | \n");
          printf (" |  53: Display Paths  54: Enable Path      55: Disable Path     | \n");
          printf (" +------------------------------+--------------------------------+ \n");
      }
}
/* menu */


/*---------------------------------------------------------------------------*/
/* Name:        usage                                                        */
/* Synopsis:    Display program usage information.                           */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/

static void usage (int extended)
{
  if (!verbose)  return;
  printf ("                                                               \n");
  printf (" USAGE:                                                        \n");
  printf ("   tapeutil -f device-file -o operation [-w open-mode]         \n");
  printf ("     [-s move-source] [-d move-destination] [-c op-count]      \n");
  printf ("     [-b block-size] [-n blocking-factor] [-m multiplier]      \n");
  printf ("     [-r random-seed] [-z ucode/dump/pattern-filename]         \n");
  printf ("     [-t msg/gpo/spo/gpa/spa/inq/inv/mt-st -type]              \n");
  printf ("     [-x lck/lod/rem/spo/spa/inq/log/mod-value]                \n");
  printf ("     [-y msg0,msg1] [-k res_key, ser_key] [-p path]            \n");
  printf ("     [-v (verbose-mode)] [-h,-? (usage-help)]                  \n");
  printf (" OPERATIONS:                                                   \n");
  printf ("   Service Commands:        qsn, qmc, fdp, sdp, dmc, drv       \n");
  printf ("   Basic SCSI Commands:     tur, inq, req, res, rel, log       \n");
  printf ("                            mod, rem                           \n");
  printf ("   Medium Changer Commands: mov, exc, pos, ele, inv, dvc       \n");
  printf ("                            aud, ier, lck                      \n");
  printf ("   CPF Commands:            qry, phs, ena, dis                 \n");
  printf ("   Tape Drive Commands:     rea, wri, eof, era, rew, ret       \n");
  printf ("                            off, lod, fsf, bsf, fsr, bsr       \n");
  printf ("                            eod, grs, srs, gds, gdi, gmi       \n");
  printf ("                            gpo, spo, gpa, spa, syn, msg       \n");
  printf ("                            den, chk                           \n");
  printf (" OPEN-MODES:                                                   \n");
  printf ("   1=r/w, 2=rdo, 3=wro, 4=app                                  \n");
  printf (" TYPES:                                                        \n");
  printf ("   msg:     1=ms0, 2=ms1, 3=fl0, 4=fl1, 5=alt                  \n");
  printf ("   gpo/spo: 1=log, 2=phy                                       \n");
  printf ("   gpa/spa: 0=all(gpa only), 1=blk, 2=cmp, 3=buf, 4=imm        \n");
  printf ("            5=trl, 6=wpr, 7=acf, 8=cap, 9=sil, 10=rel          \n");
  printf ("   inq:     0=std, 1=pgc                                       \n");
  printf ("   inv/dvc: 1=out_to_file                                      \n");
  printf ("   eof/era/rew/off/fsf/bsf/fsr/bsr/eod:                        \n");
  printf ("            0=mt,  1=st                                        \n");
  printf ("                                                               \n");
}
/* usage */


/*---------------------------------------------------------------------------*/
/* Name:        dump_bytes                                                   */
/* Synopsis:    Display byte data.                                           */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/

static void dump_bytes ( char *data, int count)
{
  int i;
  int hex_index;
  int ascii_index;
  char hex_buf[50];
  char ascii_buf[25];

  if (!verbose)  return;

  printf ("\n");
  printf ("   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F");
  printf ("  0123456789ABCDEF\n");
  printf ("  -----------------------------------------------");
  printf ("  ----------------\n");

  hex_index   = 0;
  ascii_index = 0;
  memset (hex_buf  , 0, 50);
  memset (ascii_buf, 0, 25);

  for (i = 0; i < count; i++) {

    sprintf (&hex_buf[hex_index], " %02X", (uchar)data[i]);
    if (data[i] < 32)  ascii_buf[ascii_index] = '.';
    else  ascii_buf[ascii_index] = data[i];
    hex_index += 3;
    ascii_index += 1;

    if (((i + 1) % 16 == 0) || (i + 1 == count)) {
      printf (" %-48s", hex_buf);
      printf ("  %-16s\n", ascii_buf);
      hex_index   = 0;
      ascii_index = 0;
      memset (hex_buf  , 0, 50);
      memset (ascii_buf, 0, 25);
    }
  }
}
/* dump_bytes */


/*---------------------------------------------------------------------------*/
/* Name:        elem_descriptor                                               */
/* Synopsis:    Display Element Descriptors data.                            */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/

static void elem_descriptor( char *str_elem, char *elem_p, int elem_size )
{
      element_status_t data;
      bcopy(elem_p, (char *)&data, elem_size);

      printf ("%s Address %d\n", str_elem, data.address);

      if ( elem_type.impexp_elem ) {
         if (data.inenab)
            printf ("Import Enable ............... Yes\n");
         else
            printf ("Import Enable ............... No\n");
         if (data.exenab)
            printf ("Export Enable ............... Yes\n");
         else
            printf ("Export Enable ............... No\n");
       }

      if (data.except) {
         printf ("%s State ......... Abnormal\n", str_elem);
         printf ("ASC/ASCQ ................... %2.2X/%2.2X\n", data.asc, data.ascq);
      }
      else {
         printf ("%s State ......... Normal\n", str_elem);
         printf ("ASC/ASCQ ................... %2.2X/%2.2X\n", data.asc, data.ascq);
      }

      if (data.full)
         printf ("Media Present ............... Yes\n");
      else
         printf ("Media Present ............... No\n");

      if (!elem_type.medium_elem) {
         if (data.access)
            printf ("Robot Access Allowed ........ Yes\n");
         else
            printf ("Robot Access Allowed ........ No\n");
      }

      if ( elem_type.impexp_elem ) {
         if (data.impexp)
            printf ("The media was placed by..... the operator\n");
         else
            printf ("The media was placed by..... Medium Changer\n");
      }

      if ( elem_type.data_elem ) {
         if (data.notbus)
            printf ("Same Bus As Medium Changer .. Yes\n");
         else
            printf ("Same Bus As Medium Changer .. No\n");

         if (data.idvalid)
            printf ("SCSI Bus Addressi ............ %d\n", data.scsi);
         else
            printf ("SCSI Bus Address ............ No\n");

         if (data.luvalid)
            printf ("Logical Unit Number ......... %d\n",data.lun);
         else
            printf ("Logical Unit Number ......... No\n");
      }

      if (data.svalid)
         printf ("Source Element Address ...... %d\n", data.source);
      else
         printf ("Source Element Address ...... No\n");

      if (data.invert)
         printf ("Media Inverted .............. Yes\n");
      else
         printf ("Media Inverted .............. No\n");

      if (data.full) {
         if ((data.volume[0] == ' ') || (data.volume[0] == 0))
            printf ("Volume Tag .................. Unknown\n");
         else
            printf ("Volume Tag .................. %0.36s\n", data.volume);
      }
      else
         printf ("Volume Tag .................. No\n");

  return;
}
/* elem_descriptor */


/*****************************************************************************/
/*                             SERVICE FUNCTIONS                             */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* Name:        srv_device_sn                                                */
/* Synopsis:    Display tape device serial number.                           */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_device_sn ()
{
  int rc;
  device_sn_t device_sn;
  inquiry_page_t inquiry_page;
  char serial_num[20];

  memset ((char *)&inquiry_page, (char)0, sizeof (inquiry_page_t));
  inquiry_page.page_code = (uchar) 0x80;
  if (!(rc = ioctl (dev_fd, IOC_INQUIRY_PAGE, &inquiry_page))) {
     bcopy(&inquiry_page.data[4], serial_num, inquiry_page.data[3]);
     serial_num[inquiry_page.data[3]] = '\0';
     PRINTF ("Serial Number: %s\n", serial_num);
  }
  else {
     PRINTF ("IOC_INQUIRY_PAGE for page 0x%X failed.\n", 0x80);
     PRINTF ("\n");
     scsi_request_sense ();
   }

  if (device.ultrium) {
      return (1);
   }
   else if (medium_chgr) {
      return (1);
   }

   if (!(rc = ioctl (dev_fd, STIOC_DEVICE_SN, &device_sn))) {
      PRINTF ("%s IBM 3494 Device Subsystem Number: %08X\n", dev_name, device_sn);
   }
   else {
    PERROR ("Failure obtaining tape device serial number");
    PRINTF ("\n");
    scsi_request_sense ();
   }

  return (rc);
}
/* svc_device_sn */


/*---------------------------------------------------------------------------*/
/* Name:        svc_ucode_level                                              */
/* Synopsis:    Display tape device microcode level.                         */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_ucode_level ()
{
   int i, rc;
   sense_data_t sense_data;
   int save_verbose;

   save_verbose = verbose;
   verbose = FALSE;
   (void) scsi_inquiry ();
   verbose = save_verbose;
   if ((device.ultrium)
      || (device.ult3581)
      || (device.ult3582)
      || (device.ult3583)
      || (device.ult3573)
      || (device.ult3576)
      || (device.ult3584)
      || (device.ny3581)
      || (device.t3490)
      || (device.lib3577)){
      strncpy (ucode_level, (const char *)inquiry_data.rev, 4);
      ucode_level[4] = '\0';
      PRINTF ("%s Firmware Revision Level: %s\n", dev_name, ucode_level);
      PRINTF ("   (from Inquiry Standard Data)\n");
   }
   else if ((device.t3570) || (device.t3590) || (device.t3592)) {
      if (!(rc = ioctl (dev_fd, IOC_REQUEST_SENSE, &sense_data))) {
         strncpy (ucode_level, (const char *)&sense_data.vendor[54], 7);
         ucode_level[7] = '\0';
         PRINTF ("%s microcode level: %s\n", dev_name, ucode_level);
         PRINTF ("   (from Sense Data)\n");
         strncpy (ucode_level, (const char *)inquiry_data.rev, 4);
         ucode_level[4] = '\0';
         PRINTF ("Firmware Revision Level: %s.\n", ucode_level);
         PRINTF ("   (from Inquiry Standard Data)\n");
      }
      else {
         PERROR ("Failure obtaining device microcode level");
         PRINTF ("\n");
         scsi_request_sense ();
      }
   }
   else {
      (void) fprintf (stderr, "Cannot identify device type.\n");
      return (-1);
   }

   return (rc);
}
/* svc_ucode_level */


/*---------------------------------------------------------------------------*/
/* Name:        svc_force_dump                                               */
/* Synopsis:    Force tape device dump.                                      */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_force_dump ()
{
  int rc;

  if ((device.ult3581)
     || (device.ult3582)
     || (device.ult3583)
     || (device.ult3576)
     || (device.ult3573)
     || (device.ult3584)
     || (device.ny3581)
     || (device.lib3577)
     || (device.t3490)
     || ((device.t3590) && (medium_chgr))
     || ((device.t3570) && (medium_chgr))
      ) {
      fprintf (stderr, "%s does not support force_dump.\n", dev_name);
      return (1);
   }

  if (!(rc = ioctl (dev_fd, STIOC_FORCE_DUMP, 0))) {
    PRINTF ("Dump completed successfully on tape device %s.\n", dev_name);
  }
  else {
    PERROR ("Failure performing tape device dump");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* svc_force_dump */


/*------------------------------------------------------------------------------------*/
/* Name:       read_buflist                                                           */
/* Synopsis:   Read from list of device buffers to specified file descriptor.         */
/* Returns:    Total bytes read from all specified device buffers.                    */
/* Called by:  svd_store_dump()                                                       */
/* Call with:  read_buflist(int fp, buffer_io_t *biop, int desc_mode,int data_mode,   */
/*                              int  bufid, ..., END_BUFLIST );                       */
/*------------------------------------------------------------------------------------*/

static long long read_buflist ( int fp, buffer_io_t *biop, int desc_mode, int data_mode, int  bufid, ...)
{
   #define     MAXBUFS 20
   va_list     ap;
   uchar       uc_desc_mode;
   uchar       uc_data_mode;
   uint        next_read;
   int         tmpbufid;
   uint        offset;
   uint        tot_buflen;
   int         bufcount    = 0;
   int         rc          = 0;
   long long   tot_dumplen = 0;
   uchar       *bufp;
   uchar       bufid_list[MAXBUFS];
   int         bufidx      = 1;

   uc_desc_mode = (uchar) desc_mode;
   uc_data_mode = (uchar) data_mode;
   va_start(ap, bufid);
   bufid_list[0] = (uchar) bufid;

   while((bufidx < MAXBUFS) && ((tmpbufid = va_arg(ap, int)) >= 0)) {
     bufid_list[bufidx++] = (uchar) tmpbufid;
   }
   va_end(ap);
   bufcount = bufidx;
   bufp =  biop->buffer;

   for (bufidx = 0; bufidx < bufcount; bufidx++) {
      biop->mode   = uc_desc_mode;
      biop->offset = 0;
      biop->size   = 4;
      biop->id     = bufid_list[bufidx];
      if (rc = ioctl (dev_fd, STIOC_READ_BUFFER, biop)) {
         PERROR ("Failure retrieving dump size");
         PRINTF ("\n");
         errno = rc;
         return (-1);
      }
      tot_buflen = ((bufp[1] << 16) + (bufp[2] << 8) + (bufp[3]));
      offset     = 0;
      next_read  = IOBUFSZ;
      biop->mode = uc_data_mode;

      while (offset < tot_buflen) {
         if ((tot_buflen - offset) < next_read)  next_read = tot_buflen - offset;

         biop->offset = offset;
         biop->size   = next_read;

         PRINTF ("Reading Buffer 0x%2.2X,  Offset %6d,  Bytes %6d \n",
            bufid_list[bufidx], offset, next_read );
         if (rc = ioctl (dev_fd, STIOC_READ_BUFFER, biop)) {
            PERROR ("Failure retrieving dump buffer");
            PRINTF ("\n");
            errno = rc;
            return (-1);
         }
         if (write (fp, bufp, next_read) != next_read) {
            PERROR ("Failure writing dump to destination file");
            errno = rc;
            return (-1);
         }
         offset += next_read;
      }
      tot_dumplen += tot_buflen;
   }
   return (tot_dumplen);
}
/* read_buflist */


/*---------------------------------------------------------------------------*/
/* Name:        svc_store_dump                                               */
/* Synopsis:    Store tape device dump to tape cartridge or host file.       */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_store_dump ( char *filename )
{
  int rc;
  int file;
  long long dump_size;
  buffer_io_t buffer_io;
  uchar *buf;
  int log_data_len, data_len;
  log_sns_pg_t   log_sns_pg;
  log_page_hdr_t *log_page_hdr_p;


  if ( ((device.t3590) && (medium_chgr))
      || ((device.t3570) && (medium_chgr))
      ) {
      FPRINTF (stderr, "%s does not support store_dump.\n", dev_name);
      FPRINTF (stderr, "Use the associated tape drive for this device, instead.\n");
      return (1);
  }

  if ((device.t3490) || (device.ult3576)) {
      FPRINTF (stderr, "%s does not support store_dump.\n", dev_name);
      return (1);
  }

  if (!strcmp (filename, "tapedrive")) {
    if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) {
      if (!(rc = ioctl (dev_fd, STIOC_STORE_DUMP, 0))) {
        PRINTF ("Dump stored successfully on tape device %s.\n", dev_name);
      }
      else {
        PERROR ("Failure storing dump on tape device");
        PRINTF ("\n");
        scsi_request_sense ();
      }
    }
    else {
      PERROR ("Tape device not ready or tape not mounted");
      PRINTF ("\n");
      scsi_request_sense ();
    }
  }
  else {
    if (!(buf = malloc (IOBUFSZ))) {
      PERROR ("Failure allocating IO buffer");
      return (errno);
    }
   memset (buf, 0, IOBUFSZ);

    if ((file = open (filename, O_WRONLY | O_CREAT, 0644)) < 0) {
        (void) fprintf (stderr, "Error %d (%s) opening dump destination file.\n",
            errno, strerror(errno));
         if (errno == EEXIST) {
            (void) fprintf (stderr, "Erase this file or specify another name.\n");
         }
         free (buf);
         return (errno);
    }

    memset ((char *)&buffer_io, 0, sizeof(buffer_io_t));
    buffer_io.buffer = buf;
        if (device.ultrium) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
            ULT_DESC_ONLY_MODE, ULT_DATA_ONLY_MODE, ULT_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ult3581) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
            ULT3581_DESC_ONLY_MODE, ULT3581_DATA_ONLY_MODE,
            ULT3581_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ult3582) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
            ULT3582_DESC_ONLY_MODE, ULT3582_DATA_ONLY_MODE,
            ULT3582_DUMP_BUF, ULT3582_CMD_BUF, END_BUFLIST );
      }
      else if (device.ult3583) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
            ULT3583_DESC_ONLY_MODE, ULT3583_DATA_ONLY_MODE,
            ULT3583_DUMP_BUF, ULT3583_CMD_BUF, END_BUFLIST );
      }
      else if (device.ult3584) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
           ULT3584_DESC_ONLY_MODE, ULT3584_DATA_ONLY_MODE,
           ULT3584_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ny3581) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
           NY3581_DESC_ONLY_MODE, NY3581_DATA_ONLY_MODE,
           NY3581_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ult3576) {
        (void) fprintf (stderr, "The 3576 Doesn't support this command.\n");
        return (-1);
      }
      else if ((device.t3570) || (device.t3590) || (device.t3592)) {
         /*LINTED argument mismatch*/
         dump_size = read_buflist( file, &buffer_io,
           MAG_DESC_ONLY_MODE, MAG_DATA_ONLY_MODE,
           MAG_DUMP_BUF, END_BUFLIST );
      }
      else if (device.lib3577 || device.ult3573) {
         memset ((char *)&log_sns_pg, 0, sizeof (log_sns_pg_t));
         if (device.lib3577) log_sns_pg.page_code = 0x6;
         if (device.ult3573) log_sns_pg.page_code = 0x33;

         rc = ioctl (dev_fd, IOC_LOG_SENSE_PAGE, &log_sns_pg);
         if (!rc) {
            log_data_len = (int)((log_sns_pg.data[2] << 8) + log_sns_pg.data[3]) + 4;
            dump_size = write (file, (uchar *)(&log_sns_pg.data), log_data_len);
            if (dump_size == -1) {
               PERROR ("Failure writing dump to destination file");
               rc = errno;
            }
         }
      }
      else {
         (void) fprintf (stderr, "Cannot identify device type.\n");
         return (-1);
      }

    free (buf);
      close (file);
      if (dump_size > 0) {
         PRINTF ("Dump stored to %s successfully (%lld bytes).\n", filename, dump_size);
      }
      else {
         PRINTF ("Store dump failed. \n");
         return (errno);
      }
  }
  return (rc);
}
/* svc_store_dump */


/*---------------------------------------------------------------------------*/
/* Name:        svc_load_ucode                                               */
/* Synopsis:    Download microcode to tape device.                           */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_load_ucode ( char *filename )
{
  int rc = 0;
  int i;
  int file;
  uint ucode_size = 0;
  uint buf_size;
  uint remaining;
  uint offset = 0;
  long uc_chunksz;  /* Ucode chunksize */
  buffer_io_t buffer_io;
  tape_op_t tape_op;
  uchar *buf;
  char cmdbuf[64];
  int wait_reset = 30;
  struct stat f_info;
  uint retry_time = 3;

  if (device.ult3573)
    {
      buf_size=IOBUFSZ-1;
    }
  else
    {
      buf_size=IOBUFSZ;
    }


  if (!(buf = malloc (buf_size))) {
    PERROR ("Failure allocating IO buffer");
    return (errno);
  }

  if ((file = open (filename, O_RDONLY)) < 0) {
    PERROR ("Failure opening microcode source file");
    rc = errno;
    free (buf);
    return (rc);
  }

  if (fstat(file, &f_info)) {
      PERROR ("Failure statting microcode source file");
      rc = errno;
      free (buf);
      return (rc);
   }

  if ((uc_chunksz = read (file, buf, buf_size)) < 0) {
    PERROR ("Failure reading microcode image");
    rc = errno;
    free (buf);
    close (file);
    return (rc);
  }

  if ((! medium_chgr) && (! device.ultrium) && (! device.t3592)) {
   if (!(ioctl (dev_fd, IOC_TEST_UNIT_READY))) {
    tape_op.st_op = ST_OP_OFFL;
    tape_op.st_count = 0;
    PRINTF ("Unloading tape device.\n");
    if (rc = ioctl (dev_fd, STIOC_TAPE_OP, &tape_op)) {
      PERROR ("Failure unloading tape device");
      PRINTF ("\n");
      scsi_request_sense ();
      free (buf);
      close (file);
      return (rc);
     }
   }
  }

  PRINTF ("\nCurrent microcode level...\n");
  (void) svc_ucode_level();

  if (device.ultrium) {
      wait_reset       = 60;
      buffer_io.mode   = ULT_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3581) {
      wait_reset       = 60;
      buffer_io.mode   = ULT3581_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3581_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3582) {
      wait_reset       = 90;
      buffer_io.mode   = ULT3582_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3582_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3583) {
      wait_reset       = 90;
      buffer_io.mode   = ULT3583_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3583_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3584) {
      buffer_io.mode   = ULT3584_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3584_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ny3581) {
      wait_reset       = 60;
      buffer_io.mode   = NY3581_DWNLD_UCODE_MODE;
      buffer_io.id     = NY3581_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3576) {
      wait_reset       = 120;
      buffer_io.mode   = ULT3583_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3583_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3573) {
      wait_reset       = 60;
      buffer_io.mode   = ULT3573_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3573_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.lib3577) {
      wait_reset       = 120;
      buffer_io.mode   = LIB3577_DATA_ONLY_MODE;
      buffer_io.id     = LIB3577_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if ((device.t3570) || (device.t3590) || (device.t3592)) {
      buffer_io.mode   = MAG_DWNLD_UCODE_MODE;
      buffer_io.id     = MAG_UCODE_BUF;
   ucode_size = ((buf[0]) + (buf[1] << 8) + (buf[2] << 16) + (buf[3] << 24));
   }
   else {
      (void) fprintf (stderr, "Cannot identify device type.\n");
      return (-1);
   }

  buffer_io.buffer = buf;
  remaining = ucode_size;

  if (device.ult3583) {
      if (strcmp(ucode_level, "2.10.0007") < 0) {  /* automated download from ucode level ibm-v2_10_7.lif */
         (void) fprintf (stdout, "Place library in \"Update Firmware\" mode.\n");
         (void) fprintf (stdout, "Press ENTER when ready, or type QUIT to abort.> ");
         gets(cmdbuf);
         if (cmdbuf[0] != '\0') return (-1);
         for (i = 0; i < 36; i++) {
            if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) break;
         }
         if (rc) {
            (void) fprintf (stderr, "Device not ready.\n");
         }
      }
      else {
         (void) fprintf (stdout, "Waiting the library into \"booting\" status ...\n");
      }
  }

  while (remaining) {
    buffer_io.offset = offset;
    buffer_io.size   = (uint) MIN(uc_chunksz, remaining);

    PRINTF ("Downloading %7d bytes at offset %7d.\n", uc_chunksz, offset);

    if (remaining <= buf_size) {
      PRINTF ("Waiting for tape device to write flash-prom.\n");
    }

    if (strcmp(ucode_level, "2.10.0007") < 0) {
       rc = ioctl (dev_fd, STIOC_WRITE_BUFFER, &buffer_io);
    }
    else {
       for (i = 0; i < 72; i++) {
          if (!(rc = ioctl (dev_fd, STIOC_WRITE_BUFFER, &buffer_io))) break;
       }
    }
    if (rc) {
      PERROR ("Downloading microcode to device");
      PRINTF ("\n");
      scsi_request_sense ();
      free (buf);
      close (file);
      return (rc);
    }

    offset += uc_chunksz;

    if ((uc_chunksz = read (file, buf, buf_size)) < 0) {
      PERROR ("Failure reading microcode file");
      rc = errno;
      free (buf);
      close (file);
      return (rc);
    }

    if ((ucode_size - offset) < uc_chunksz)  uc_chunksz = ucode_size - offset;
    remaining = ucode_size - offset;

    if (!uc_chunksz) break;
  }

  if (device.lib3577) {
     buffer_io.mode   = LIB3577_DWNLD_UCODE_MODE;
     buffer_io.id     = LIB3577_UCODE_BUF;
     buffer_io.size   = 0;
     buffer_io.offset = 0;

     if (rc = ioctl (dev_fd, STIOC_WRITE_BUFFER, &buffer_io)) {
        PERROR ("Flush microcode from buffer to device");
        PRINTF ("\n");
        scsi_request_sense ();
        free (buf);
        close (file);
        return (rc);
     }
  }

  free (buf);
  close (file);
  PRINTF ("Microcode downloaded to %s successfully (%d bytes).\n",
      dev_name, offset);
  rc = 1;

  if (device.ult3584) {
     PRINTF ("Waiting 5 minutes for device to reset ...\n");
     (void) sleep(270);
  }
  else {
     PRINTF ("Waiting %d seconds for device to reset (will retry up to %d times)...\n", wait_reset, retry_time);
  }

  for (i = 0; i < retry_time && rc != 0; i++) {
      if (i > 0) PRINTF ("Waiting additional %d seconds for device to reset...\n", wait_reset);
      (void) sleep(wait_reset);
      PRINTF ("Trying INQUIRY ...\n");
      rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data);
  }

  if (rc != 0) {
      rc = errno;
      (void) fprintf (stdout, "INQUIRY rc %d (%s)\n", rc, strerror(rc));
  }
  else {
      PRINTF ("\nNew microcode level...\n");
      (void) svc_ucode_level();
  }
  PRINTF ("The tape is unloaded from the drive while updating the microcode for the tape drive.\n");
  return (rc);
}
/* svc_load_ucode */


/*---------------------------------------------------------------------------*/
/* Name:        svc_format_tape                                              */
/* Synopsis:    Format tape cartridge mounted in tape device.                */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_format_tape ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) {
    if (!(rc = ioctl (dev_fd, STIOC_FORMAT_TAPE, 0))) {
      PRINTF ("Tape cartridge formatted successfully.\n");
    }
    else {
      PERROR ("Failure formatting tape cartridge");
      PRINTF ("\n");
      scsi_request_sense ();
    }
  }
  else {
    PERROR ("Tape device not ready or tape not mounted");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* svc_format_tape */


/*---------------------------------------------------------------------------*/
/* Name:        svc_check_device                                             */
/* Synopsis:    Check device attachment.                                     */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int svc_check_device ()
{
  int rc;
  char vid[9];

  if (!(rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data))) {
    strncpy (vid, (const char*)inquiry_data.vid, 8);
    vid[8] = '\0';
    if (((inquiry_data.type == 1) || (inquiry_data.type == 8)) &&
        (!(strncmp (vid, "IBM", 3)))) {
      vid[3] = '\0';
      PRINTF ("%s %s configured at %s.\n", vid, pid, dev_name);
    }
    else {
      PRINTF ("Unsupported %s %s detected at %s.\n", vid, pid, dev_name);
      rc = ENODEV;
    }
  }
  else {
    PERROR ("Unable to verify device attachment");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* svc_check_device */

/*---------------------------------------------------------------------------*/
/* Name:        get_driver_level                                             */
/* Synopsis:    Display tape driver name and version.                        */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int get_driver_level ()
{
  int rc;
  get_driver_info_t get_driver_info;

  if (!(rc = ioctl (dev_fd, IOC_GET_DRIVER_INFO, &get_driver_info))) {
    strncpy (driver_level, get_driver_info.version, 9);
    PRINTF ("The version of %s(Advanced Tape Device Driver): %s\n", get_driver_info.driver_id, driver_level);
  }
  else {
    PERROR ("Failure obtaining the version of ATDD");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* get_driver_level */

/*****************************************************************************/
/*                         BASIC SCSI FUNCTIONS                              */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* Name:        scsi_test_unit_ready                                         */
/* Synopsis:    Verify device ready.                                         */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_test_unit_ready ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) {
    PRINTF ("The IOC_TEST_UNIT_READY ioctl succeeded.\n");
  }
  else {
    PERROR ("The IOC_TEST_UNIT_READY ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* scsi_test_unit_ready */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_inquiry                                                 */
/* Synopsis:    Display device inquiry data.                                 */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_inquiry ()
{
  int rc;
  inquiry_data_t inquiry_data;

  memset ((char *)&inquiry_data, (char)0, sizeof (inquiry_data_t));
  if (!(rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data))) {
    PRINTF ("The IOC_INQUIRY ioctl succeeded.\n");
    PRINTF ("\nThe inquiry data is:\n");
    dump_bytes ((char *)&inquiry_data, sizeof (inquiry_data_t));
  }
  else {
    PERROR ("The IOC_INQUIRY ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* scsi_inquiry */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_inquiry_page                                            */
/* Synopsis:    Display device inquiry page data.                            */
/*              See also scsi_inquiry().                                     */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_inquiry_page (int page)
{
   int rc;
   inquiry_page_t inquiry_page;

   memset ((char *)&inquiry_page, (char)0, sizeof (inquiry_page_t));
   inquiry_page.page_code = (uchar) page;
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY_PAGE, &inquiry_page))) {
     PRINTF ("The IOC_INQUIRY_PAGE ioctl succeeded.\n");
     PRINTF ("Inquiry Data (Page 0x%02x):\n", page);
     dump_bytes ((char *)&inquiry_page.data, inquiry_page.data[3]+4);
   }
   else {
     PRINTF ("IOC_INQUIRY_PAGE for page 0x%02x failed.\n", page);
     PRINTF ("\n");
     scsi_request_sense ();
   }

   return (rc);
}
/* scsi_inquiry_page */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_request_sense                                           */
/* Synopsis:    Display device sense data.                                   */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_request_sense ()
{
  int rc;
  sense_data_t sense_data;

  if (!(rc = ioctl (dev_fd, IOC_REQUEST_SENSE, &sense_data)))
  {
    PRINTF ("The IOC_REQUEST_SENSE ioctl succeeded.\n");
    PRINTF ("\nThe request sense data is:\n");
    dump_bytes ((char *)&sense_data, sizeof (sense_data_t));
  }
  else {
    PERROR ("The IOC_REQUEST_SENSE ioctl failed");
  }

  return (rc);
}
/* scsi_request_sense */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_reserve                                                 */
/* Synopsis:    Reserve device on CLOSE().                                   */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_reserve ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, IOC_RESERVE, 0))) {
    PRINTF ("The IOC_RESERVE ioctl succeeded.\n");
  }
  else {
    PERROR ("The IOC_RESERVE ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* scsi_reserve */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_release                                                 */
/* Synopsis:    Release device.                                              */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_release ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, IOC_RELEASE, 0))) {
    PRINTF ("The IOC_RELEASE ioctl succeeded.\n");
  }
  else {
    PERROR ("The IOC_RELEASE ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* scsi_release */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_log_sense_page                                          */
/* Synopsis:    Display Device Log Sense Page.                               */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_log_sense_page (int page, int type, int parmcode)
{
  int i, j=0;
  int rc;
  int true;
  int len, parm_len;
  int parm_code;
  log_sns_pg_t   log_sns_page;
  log_page_hdr_t page_header;

  memset ((char *)&log_sns_page, (char)0, sizeof(log_sns_pg_t));
  log_sns_page.page_code = (uchar) page;

  if (!(rc = ioctl (dev_fd, IOC_LOG_SENSE_PAGE, &log_sns_page))) {
     len =(int) ((log_sns_page.data[2] << 8) + log_sns_page.data[3]) + 4;
     if ( type != 1) {
        PRINTF ("Log Sense Data (Page 0x%02x):\n", page);
        dump_bytes ((char *)&log_sns_page.data, len);
     }
     else {
        for(i=4; i<=len; i=(parm_len+4)){
           j += i;
           parm_code = (int) ((log_sns_page.data[j] << 8) + log_sns_page.data[j+1]);
           parm_len = (int) (log_sns_page.data[j+3]);
           if (true = (parm_code == parmcode)) {
              PRINTF ("Log Sense Data (Page 0x%02x, Parameter Code 0x%04x):\n", page, parmcode);
              dump_bytes ((char *)&log_sns_page.data[j], (parm_len+4));
              break;
           }
        }
        if (!true)
          PRINTF ("IOC_LOG_SENSE_PAGE for Page 0x%02x, Parameter Code 0x%04x failed.\n", page, parmcode);
     }
   }
   else {
     PRINTF ("IOC_LOG_SENSE_PAGE for page 0x%X failed.\n", page);
     PRINTF ("\n");
     scsi_request_sense ();
   }

   return (rc);
   }
/* scsi_log_sense_page */


/*---------------------------------------------------------------------------*/
/* Name:        scsi_mode_sense_page                                         */
/* Synopsis:    Display Device Mode Sense Page.                              */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int scsi_mode_sense_page (page)
{
   int               rc;
   int               header_len;
   int               blk_dsc_len = 0;
   int               mode_data_len = 0;
   int               mode_data_returned_len = 0;
   int               max_mdsnspg_data_len = 0;
   uchar             cmd_code;
   uchar             medium_type;
   uchar             density_code;
   uchar             wrt_prot;
   char              *header_p;
   char              *blkdsc_p;
   void              *mode_data_p;
   char              * dens_str_p;
   char              * med_str_p;
   char              * wrt_prot_p;
   char              err_msg[1024];
   mode_sns_t        mode_page;

   memset (err_msg,            (char)0, sizeof(err_msg));
   memset ((char *)&mode_page, 0, sizeof(mode_sns_t));

   mode_page.page_code = page;
   if (!(rc = ioctl (dev_fd, IOC_MODE_SENSE, &mode_page))) {
      PRINTF ("IOC_MODE_SENSE succeeded.\n");
      header_p = (char *)&mode_page.data;
      cmd_code = mode_page.cmd_code;
      if ( cmd_code == MODESNS_6_CMD ) {
         header_len             = sizeof(mode_hdr6_t);
         mode_data_len          = (uint) ((mode_hdr6_t *)header_p)->data_len;
         blk_dsc_len            = (uint) ((mode_hdr6_t *)header_p)->blk_dsc_len;
         max_mdsnspg_data_len   = MAX_MSDATA - header_len - blk_dsc_len;
         mode_data_returned_len = MIN( mode_data_len + 1, max_mdsnspg_data_len);
         medium_type            = (uchar)((mode_hdr6_t *)(header_p))->medium_type;
         wrt_prot                = (uchar)((mode_hdr6_t *)(header_p))->wrt_prot;
      }
      else if ( cmd_code == MODESNS_10_CMD ) {
         header_len    = sizeof(mode_hdr10_t);
         mode_data_len = (uint) ((((mode_hdr10_t *)header_p)->data_len[0] << 8)
               | ((mode_hdr10_t *)header_p)->data_len[1] );
         blk_dsc_len   = (uint) ((((mode_hdr10_t *)header_p)->blk_dsc_len[0] << 8)
               | ((mode_hdr10_t *)header_p)->blk_dsc_len[1] );
         max_mdsnspg_data_len   = MAX_MSDATA - header_len - blk_dsc_len;
         mode_data_returned_len = MIN( mode_data_len + 2, max_mdsnspg_data_len);
         medium_type            = (uchar)((mode_hdr10_t *)(header_p))->medium_type;
         wrt_prot                = (uchar)((mode_hdr10_t *)(header_p))->wrt_prot;
      }
      else {
         fprintf (stderr, "mode sense: Unknown mode sense command code '0x%X'.\n", cmd_code);
         return (1);
      }

      blkdsc_p     = header_p + header_len;
      mode_data_p  = blkdsc_p + blk_dsc_len;
      density_code = (blk_dsc_len
                     ? ( unsigned char )((blkdsc_t *)(blkdsc_p))->density_code
                     : 0 );

      switch (density_code) {
         case 0x28:
         case 0x80:
            dens_str_p = (char *)&("IBM 3480-2 XF Density (36 Track, 1491 cpmm)");
            break;

         case 0x29:
            dens_str_p = (char *)&("IBM TotalStorage 3590 Model B Density");
            break;

         case 0x2A:
            dens_str_p = (char *)&("IBM TotalStorage 3590 Model E Density");
            break;

         case 0x2c:
            dens_str_p = (char *)&("IBM TotalStorage 3590 Model H Density");
            break;

         case 0x51:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model J1A Density");
            break;

         case 0x52:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model E05 Density");
            break;

         case 0x71:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model J1A Encryption Density");
            break;

         case 0x72:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model E05 Encryption Density");
            break;

         case 0x82:
            dens_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, B-Format Density");
            break;

         case 0x83:
            dens_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, C-Format Density");
            break;

         case 0x40:
            dens_str_p = (char *)&("IBM Ultrium LTO-1 1/8T Density");
            break;

         case 0x42:
            dens_str_p = (char *)&("IBM Ultrium LTO-2 2/8T Density");
            break;

         case 0x44:
            dens_str_p = (char *)&("IBM Ultrium LTO-3 3/16T Density");
            break;

         case 0x46:
            dens_str_p = (char *)&("IBM Ultrium LTO-4 4/16T Density");
            break;

         case 0x00:
            dens_str_p = (char *)&("Medium Not Present");
            break;

         default:
            dens_str_p = (char *)&("? Unrecognized Medium Density Code");
            break;
      }

      switch (medium_type) {
         case 0x81:
            med_str_p = (char *)&("IBM TotalStorage 3590 High Performance Cartridge");
         break;

         case 0x82:
            med_str_p = (char *)&("IBM TotalStorage 3590 High Performance Extended Length Cartridge");
         break;

         case 0x91:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A Tape Cartridge(JA)");
         break;

         case 0x92:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B Tape Cartridge(JB)");
         break;

         case 0xB1:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A Short Cartridge(JJ)");
         break;

         case 0xB2:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B Short Cartridge(JK)");
         break;

         case 0xA1:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A WORM Cartridge(JW)");
         break;

         case 0xA2:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B WORM Cartridge(JX)");
         break;

         case 0xC1:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A Short WORM Cartridge(JR)");
         break;

         case 0xC2:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B Short WORM Cartridge(JS)");
         break;

         case 0x42:
            med_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, B-Format");
         break;

         case 0x43:
            med_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, C-Format");
         break;

         case 0x44:
            med_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, C-Format XL");
         break;

         case 0x18:
            med_str_p = (char *)&("IBM Ultrium LTO-1 1/8T Cartridge");
         break;

         case 0x28:
            med_str_p = (char *)&("IBM Ultrium LTO-2 2/8T Cartridge");
         break;

         case 0x38:
            med_str_p = (char *)&("IBM Ultrium LTO-3 3/16T Cartridge");
         break;

         case 0x3C:
            med_str_p = (char *)&("IBM Ultrium LTO-3 WORM Cartridge");
         break;

         case 0x48:
            med_str_p = (char *)&("IBM Ultrium LTO-4 4/16T Cartridge");
         break;

         case 0x4C:
            med_str_p = (char *)&("IBM Ultrium LTO-4 WORM Cartridge");
         break;

         case 0x00:
            if (device.ultrium) med_str_p = (char *)&("Unknown");
            else med_str_p = (char *)&("Medium Not Present");
         break;

         default:
            med_str_p = (char *)&("? Unrecognized Medium Type Code");
         break;
      }

      switch (wrt_prot) {
         case 0x01:
            wrt_prot_p = (char *)&("Medium is Read Only");
         break;

         default:
            wrt_prot_p = (char *)&("Medium is Write-Enabled");
         break;
      }

      PRINTF ("\n");
      PRINTF ("Page Code      x'%2.2X' \n", page );
      PRINTF ("Command Code   x'%2.2X' \n", mode_page.cmd_code );
      PRINTF ("Mode Data Len  %4d \n",      mode_data_len);
      PRINTF ("Mode Page Len  %4d\n",     ((*((modep_prefix_t *)mode_data_p)).page_len));
      PRINTF ("Blk Desc Len   %4d \n",      blk_dsc_len );
      PRINTF ("Returned Len   %4d \n",      mode_data_returned_len );
      /*
      */
      PRINTF ("Write Protect  x'%2.2X'\t(%s) \n", wrt_prot, wrt_prot_p );
      PRINTF ("Medium Type    x'%2.2X'\t(%s) \n", medium_type, med_str_p );
      PRINTF ("Density Code   x'%2.2X'\t(%s) \n", density_code, dens_str_p );
      PRINTF ("\nHeader:\n");
      dump_bytes ((char *)(header_p), header_len);
      PRINTF ("\nBlock Descriptor:\n");
      dump_bytes ((char *)(blkdsc_p), blk_dsc_len);
      PRINTF ("\nMode Page:\n");
      dump_bytes ((char *)(mode_data_p), mode_data_returned_len - header_len - blk_dsc_len);
   }
   else {
      PERROR ("mode sense");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}
/* scsi_mode_sense_page */

/*---------------------------------------------------------------------------*/
/* Name:        scsi_prevent_medium_rm                                         */
/* Synopsis:    Prevent medium removal                                       */
/* Returns:     Error code from /usr/include/sys/errno.h                     */
/*---------------------------------------------------------------------------*/

static int scsi_prevent_medium_rm ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, IOC_PREVENT_MEDIUM_REMOVAL, 0))) {
     PRINTF ("The IOC_PREVENT_MEDIUM_REMOVAL ioctl succeeded.\n");
  }
  else {
     PERROR ("The IOC_PREVENT_MEDIUM_REMOVAL ioctl failed");
     PRINTF ("\n");
     scsi_request_sense ();
  }

   return (rc);
}
/* scsi_prevent_medium_rm */

/*---------------------------------------------------------------------------*/
/* Name:        scsi_allow_medium_rm                                           */
/* Synopsis:    Allow medium removal                                         */
/* Returns:     Error code from /usr/include/sys/errno.h                     */
/*---------------------------------------------------------------------------*/

static int scsi_allow_medium_rm ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, IOC_ALLOW_MEDIUM_REMOVAL, 0))) {
     PRINTF ("The IOC_ALLOW_MEDIUM_REMOVAL ioctl succeeded.\n");
  }
  else {
     PERROR ("The IOC_ALLOW_MEDIUM_REMOVAL ioctl failed");
     PRINTF ("\n");
     scsi_request_sense ();
  }

  return (rc);
}
/* scsi_allow_medium_rm */

/*****************************************************************************/
/*                         MEDIUM CHANGER FUNCTIONS                          */
/*****************************************************************************/


/*---------------------------------------------------------------------------*/
/* Name:        smc_move_medium                                              */
/* Synopsis:    Move cartridge from source to destination.                   */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_move_medium ( int src, int dst )
{
  int rc;
  move_medium_t move_medium;

  move_medium.robot = 0;
  move_medium.invert = NO_FLIP;
  move_medium.source = (short)src;
  move_medium.destination = (short)dst;


  if (!(rc = ioctl (dev_fd, SMCIOC_MOVE_MEDIUM, &move_medium))) {
    PRINTF ("The SMCIOC_MOVE_MEDIUM ioctl succeeded.\n");
  }
  else {
    PERROR ("The SMCIOC_MOVE_MEDIUM ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_move_medium */

/*---------------------------------------------------------------------------*/
/* Name:        smc_exchange_medium                                          */
/* Synopsis:    Exchange cartridges from the first to second destination and */
/*              then from the source to the first destination.               */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_exchange_medium ( int src, int dst, int dst2 )
{
  int rc;
  exchange_medium_t exchange_medium;

  exchange_medium.robot = 0;
  exchange_medium.invert1 = NO_FLIP;
  exchange_medium.invert2 = NO_FLIP;
  exchange_medium.source = (short)src;
  exchange_medium.destination1 = (short)dst;
  exchange_medium.destination2 = (short)dst2;

  if (!(rc = ioctl (dev_fd, SMCIOC_EXCHANGE_MEDIUM, &exchange_medium))) {
    PRINTF ("The SMCIOC_EXCHANGE_MEDIUM ioctl succeeded.\n");
  }
  else {
    PERROR ("The SMCIOC_EXCHANGE_MEDIUM ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_exchange_medium */

/*---------------------------------------------------------------------------*/
/* Name:        smc_pos_to_elem                                              */
/* Synopsis:    Move robot to element.                                       */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_pos_to_elem ( int dst )
{
  int rc;
  pos_to_elem_t pos_to_elem;

  pos_to_elem.robot = 0;
  pos_to_elem.invert = NO_FLIP;
  pos_to_elem.destination = (short)dst;

  if (!(rc = ioctl (dev_fd, SMCIOC_POS_TO_ELEM, &pos_to_elem))) {
    PRINTF ("The SMCIOC_POS_TO_ELEM ioctl succeeded.\n");
  }
  else {
    PERROR ("The SMCIOC_POS_TO_ELEM ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_pos_to_elem */


/*---------------------------------------------------------------------------*/
/* Name:        smc_element_info                                             */
/* Synopsis:    Display device element information.                          */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_element_info ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, SMCIOC_ELEMENT_INFO, &element_info))) {
    PRINTF ("The SMCIOC_ELEMENT_INFO ioctl succeeded.\n");
    PRINTF ("\nThe element information data is:\n");
    PRINTF ("Number of Robots........... %d\n", element_info.robot_count);
    PRINTF ("First Robot Address........ %d\n", element_info.robot_address);
    PRINTF ("Number of Slots............ %d\n", element_info.cell_count);
    PRINTF ("First Slot Address......... %d\n", element_info.cell_address);
    PRINTF ("Number of Imp/Exports...... %d\n", element_info.port_count);
    PRINTF ("First Imp/Export Address... %d\n", element_info.port_address);
    PRINTF ("Number of Drives........... %d\n", element_info.drive_count);
    PRINTF ("First Drive Address........ %d\n", element_info.drive_address);
  }
  else {
    PERROR ("The SMCIOC_ELEMENT_INFO ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_element_info */


/*---------------------------------------------------------------------------*/
/* Name:        smc_inventory                                                */
/* Synopsis:    Display device element status.                               */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_inventory ()
{
  int rc;
  int elem_size;
  ushort i;
  inventory_t inventory;

  /* get element information first */
  if (rc = ioctl (dev_fd, SMCIOC_ELEMENT_INFO, &element_info)) {
    PERROR ("The SMCIOC_INVENTORY ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
    return (rc);
  }

  inventory.robot_status = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.robot_count);
  inventory.cell_status  = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.cell_count );
  inventory.port_status  = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.port_count );
  inventory.drive_status = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.drive_count);

  if (!inventory.robot_status || !inventory.cell_status ||
      !inventory.port_status  || !inventory.drive_status) {
    PRINTF ("The SMCIOC_INVENTORY ioctl failed: Memory allocation failure.\n");
    return (ENOMEM);
  }

  if ((device.ult3581) || (device.ult3582) || (device.ult3583) || (device.ult3584) || (device.ny3581) || (device.t3570) || (device.ult3576) || (device.ult3573) || (device.lib3577))
     elem_size = 56;
  else if (device.t3590)
     elem_size = 12;

  if (!(rc = ioctl (dev_fd, SMCIOC_INVENTORY, &inventory))) {

    PRINTF ("\nThe SMCIOC_INVENTORY ioctl succeeded.\n");
    PRINTF ("\nThe robot status pages are:\n");

    for (i = 0; i < element_info.robot_count; i++) {
      elem_type.medium_elem = TRUE;
      elem_descriptor ("Robot", (char *)(&inventory.robot_status[i]), elem_size);
      if (!out_to_file) {
         PRINTF ("\n--- more ---");
         getchar ();
      }
      else
         printf ("\n");
    }
    elem_type.medium_elem = FALSE;

    PRINTF ("\nThe cell status pages are:\n");

    for (i = 0; i < element_info.cell_count; i++) {
      elem_type.storage_elem = TRUE;
      elem_descriptor ("Storage Cell", (char *)(&inventory.cell_status[i]), elem_size);
      if (!out_to_file) {
         PRINTF ("\n--- more ---");
         getchar ();
      }
      else
         printf ("\n");
    }
    elem_type.storage_elem = FALSE;

    PRINTF ("\nThe port status pages are:\n");

    for (i = 0; i < element_info.port_count; i++) {
      elem_type.impexp_elem = TRUE;
      elem_descriptor ("Import/Export", (char *)(&inventory.port_status[i]), elem_size);
      if (!out_to_file) {
         PRINTF ("\n--- more ---");
         getchar ();
      }
      else
         printf ("\n");
    }
    elem_type.impexp_elem = FALSE;

    PRINTF ("\nThe drive status pages are:\n");

    for (i = 0; i < element_info.drive_count; i++) {
      elem_type.data_elem = TRUE;
      elem_descriptor ("Tape Drive", (char *)(&inventory.drive_status[i]), elem_size);
      if (!out_to_file) {
         PRINTF ("\n--- more ---");
         getchar ();
      }
      else
         printf ("\n");
    }
    elem_type.data_elem = FALSE;
  }
  else {
    PERROR ("The SMCIOC_INVENTORY ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_inventory */


/*---------------------------------------------------------------------------*/
/* Name:        smc_read_element_devids                                      */
/* Synopsis:    Display drive element status with DVCID=1.                   */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_read_element_devids ()
{
  int rc;
  int i;

  element_devid_t *elem_devid, *elemp;
  read_element_devids_t devids;
  element_info_t element_info;

  if (rc = ioctl (dev_fd, SMCIOC_ELEMENT_INFO, &element_info)) {
    PERROR ("The SMCIOC_READ_ELEMENT_DEVIDS ioctl failed: Get the element info failure.\n");
    PRINTF ("\n");
    scsi_request_sense ();
    return (rc);
  }

  if (element_info.drive_count) {
    elem_devid = malloc(element_info.drive_count * sizeof(element_devid_t));

    if (elem_devid == NULL) {
      PRINTF ("The SMCIOC_READ_ELEMENT_DEVIDS ioctl failed: Memory allocation failure.\n");
      return (ENOMEM);
    }
    bzero(elem_devid, element_info.drive_count * sizeof(element_devid_t));

    devids.drive_devid = elem_devid;
    devids.element_address = element_info.drive_address;
    devids.number_elements = element_info.drive_count;

    printf("Reading element device ids...\n");

    if (!(rc = ioctl (dev_fd, SMCIOC_READ_ELEMENT_DEVIDS, &devids))) {
      elemp = elem_devid;

      PRINTF ("\nThe SMCIOC_READ_ELEMENT_DEVIDS ioctl succeeded.\n");
      PRINTF ("\nThe drives status datas are:\n");

      for (i = 0; i < element_info.drive_count; i++, elemp++) {
         printf("\n  Drive Address .................. %d\n",elemp->address);
         if (elemp->except)
            printf("  Drive State .................... Abnormal\n");
         else
            printf("  Drive State .................... Normal\n");
         if (elemp->asc == 0x81 && elemp->ascq ==0x00)
            printf("  ASC/ASCQ ....................... %02X%02X (Drive Present)\n",
              elemp->asc,elemp->ascq);
        else if (elemp->asc == 0x82 && elemp->ascq ==0x00)
            printf("  ASC/ASCQ ....................... %02X%02X (Drive Not Present)\n",
               elemp->asc,elemp->ascq);
        else
            printf("  ASC/ASCQ ....................... %02X%02X\n",
               elemp->asc,elemp->ascq);
        if (elemp->full)
            printf("  Media Present .................. Yes\n");
        else
            printf("  Media Present .................. No\n");
        if (elemp->access)
           printf("  Robot Access Allowed ........... Yes\n");
        else
           printf("  Robot Access Allowed ........... No\n");
        if (elemp->svalid)
           printf("  Source Element Address ......... %d\n",elemp->source);
        else
           printf("  Source Element Address Valid ... No\n");
        if (elemp->invert)
           printf("  Media Inverted ................. Yes\n");
        else
           printf("  Media Inverted ................. No\n");
        if (elemp->notbus)
           printf("  Same Bus as Medium Changer ..... No\n");
        else
           printf("  Same Bus as Medium Changer ..... Yes\n");
        if (elemp->idvalid)
           printf("  SCSI Bus Address ............... %d\n",elemp->scsi);
        else
           printf("  SCSI Bus Address Vaild ......... No\n");
        if (elemp->luvalid)
           printf("  Logical Unit Number ............ %d\n",elemp->lun);
        else
           printf("  Logical Unit Number Valid ...... No\n");
        if (elemp->dev_id[0] == '\0')
           printf("  Device ID ...................... No\n");
        else
           printf("  Device ID ...................... %0.36s\n", elemp->dev_id);

        if (!out_to_file) {
           PRINTF ("\n--- more ---");
           getchar ();
        }
        else
           printf ("\n");
        }
      }
      else {
         PERROR ("The SMCIOC_READ_ELEMENT_DEVIDS ioctl failed");
         PRINTF ("\n");
         scsi_request_sense ();
      }
     }
     else {
        printf("\nNo drives found in element information\n");
     }

  free (elem_devid);
  return (rc);
}
/* smc_read_element_devids */


/*---------------------------------------------------------------------------*/
/* Name:        smc_audit                                                    */
/* Synopsis:    Initialize device element status.                            */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_audit ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, SMCIOC_AUDIT, 0))) {
    PRINTF ("The SMCIOC_AUDIT ioctl succeeded.\n");
  }
  else {
    PERROR ("The SMCIOC_AUDIT ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_audit */

/*---------------------------------------------------------------------------*/
/* Name:        smc_init_elem_stat_range                                     */
/* Synopsis:    Initialize device element status with range.                 */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_init_elem_stat_range ( int src, int number)
{
  int rc;
  element_range_t elem_range;
  elem_range.element_address = (short)src;
  elem_range.number_elements = (short)number;

  if (!(rc = ioctl (dev_fd, SMCIOC_INIT_ELEM_STAT_RANGE, &elem_range))) {
    PRINTF ("The SMCIOC_INIT_ELEM_STAT_RANGE ioctl succeeded.\n");
  }
  else {
    PERROR ("The SMCIOC_INIT_ELEM_STAT_RANGE ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_init_elem_stat_range */

/*---------------------------------------------------------------------------*/
/* Name:        smc_lock_door                                                */
/* Synopsis:    Lock or unlock device door.                                  */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int smc_lock_door ( int lock )
{
  int rc;
  lock_door_t lock_door;

  lock_door = lock;

  if (!(rc = ioctl (dev_fd, SMCIOC_LOCK_DOOR, &lock_door))) {
    PRINTF ("The SMCIOC_LOCK_DOOR ioctl succeeded.\n");
  }
  else {
    PERROR ("The SMCIOC_LOCK_DOOR ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* smc_lock_door */


/*****************************************************************************/
/*                           TAPE DRIVE FUNCTIONS                            */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/* Name:        st_read                                                      */
/* Synopsis:    Read data from device.                                       */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_read ( int blk_size, int blk_fact, int multiplier)
{
  int rc;
  int i;
  int count;
  long long total_bytes;
  struct timeval start;
  struct timeval stop;
  struct timezone zone;
  char *data;

  if (!(data = malloc (RWBUFSZ))) {
    PERROR ("Failure allocating read data buffer");
    return (errno);
  }

  total_bytes = (long long)0;
  gettimeofday (&start, &zone);

  for (i = 1; i <= multiplier; i++) {
    count = read (dev_fd, data, (uint)(blk_size * blk_fact));
    rc = errno;

    if (count < 0) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("The READ command failed (request %d, count = %d, errno = %d).",
          i, count, errno);
      break;
    }

    if (count != (blk_size * blk_fact)) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("Short READ (request %d, count = %d).", i, count);
      count = -1;
      break;
    }

    total_bytes = total_bytes + (long long)count;
    PRINTF ("*");

    if (!(i % 32)) {
      PRINTF ("  %lld\n", total_bytes);
    }
  }

  gettimeofday (&stop, &zone);

  if ((i % 32) || (count < 0)) {
    PRINTF ("\n");
  }

  PRINTF ("\n");

  if (count < 0) {
    scsi_request_sense ();
  }
  else {
    PRINTF ("Read %lld bytes \n", total_bytes );
  }

  free (data);
  return (rc);
}
/* st_read  */


/*---------------------------------------------------------------------------*/
/* Name:        st_write                                                     */
/* Synopsis:    Write data to device.                                        */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_write
  (
  int blk_size,
  int blk_fact,
  int multiplier,
  int seed,
  char *filename
  )
{
  int rc;
  int i;
  int fd;
  int count;
  int length;
  long long total_bytes;
  struct timeval start;
  struct timeval stop;
  struct timezone zone;
  char *data;
  char *ptrn;

  if (!(data = malloc (RWBUFSZ))) {
    PERROR ("Failure allocating write data buffer");
    return (errno);
  }

  if (filename[0] != '\0') {  /* generate data using pattern file */
    if ((fd = open (filename, O_RDWR)) < 0) {  /* open pattern file */
      PERROR ("Failure opening pattern file");
      rc = errno;
      free (data);
      return (rc);
    }
    if (!(ptrn = malloc (RWBUFSZ))) {
      PERROR ("Failure allocating pattern data buffer");
      rc = errno;
      close (fd);
      free (data);
      return (rc);
    }
    if ((length = read (fd, &ptrn, RWBUFSZ)) <= 0) {  /* read pattern */
      PERROR ("Failure reading pattern file");
      rc = errno;
      close (fd);
      free (ptrn);
      free (data);
      return (rc);
    }

    close (fd);

    /* build data buffer as multiple of pattern */
    for (i = 0; i < (blk_size * blk_fact); i++) {
      data[i] = ptrn[i % length];
    }

    free (ptrn);
  }
  else {  /* generate data using random pattern */
    while (!seed) {  /* 0 < seed < 32768 */
      gettimeofday (&start, &zone);
      seed = (32768 + start.tv_usec) % 32768;
    }
    srand (seed);
    for (i = 0; i < (int)(blk_size * blk_fact); i++) {
      data[i] = (char)rand ();
    }
    PRINTF ("Generated %d bytes of random data using seed %d.\n\n",
        (blk_size * blk_fact), seed);
  }

  total_bytes = (long long)0;
  gettimeofday (&start, &zone);

  for (i = 1; i <= multiplier; i++) {
    count = write (dev_fd, data, (blk_size * blk_fact));
    rc = errno;

    if (count < 0) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("The WRITE command failed (request %d, count = %d, errno = %d).",
          i, count, errno);
      break;
    }

    if (count != (blk_size * blk_fact)) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("Short WRITE (request %d, count = %d).", i, count);
      count = -1;
      break;
    }

    total_bytes = total_bytes + (long long)count;
    PRINTF ("*");

    if (!(i % 32)) {
      PRINTF ("  %lld\n", total_bytes);
    }
  }

  gettimeofday (&stop, &zone);

  if ((i % 32) || (count < 0)) {
    PRINTF ("\n");
  }

  PRINTF ("\n");

  if (count < 0) {
    scsi_request_sense ();
  }
  else {
    PRINTF ("Wrote %lld bytes \n", total_bytes );
  }

  free (data);
  return (rc);
}
/* st_write */


/*---------------------------------------------------------------------------*/
/* Name:        st_tape_op                                                   */
/* Synopsis:    Perform tape movement operations issuing STIOC_TAPE_OP.      */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_tape_op ( int st_op, int st_count )
{
  int rc;
  tape_op_t st_tape_op;

  st_tape_op.st_op = (short)st_op;
  st_tape_op.st_count = st_count;

  if (!(rc = ioctl (dev_fd, STIOC_TAPE_OP, &st_tape_op))) {
    PRINTF ("The STIOC_TAPE_OP ioctl succeeded.\n");
    if (st_tape_op.st_op == ST_OP_GRSZ) {
      PRINTF ("\nThe record size data is:\n");
      dump_bytes ((char *)&st_tape_op.st_count, sizeof (long));
    }
  }
  else {
    PERROR ("The STIOC_TAPE_OP ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_tape_op */

/*---------------------------------------------------------------------------*/
/* Name:        mt_tape_op                                                   */
/* Synopsis:    Perform tape movement operations issuing MTIOCTOP ioctl.     */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int mt_tape_op ( int mt_op, int mt_count )
{
  int rc;
  struct mtop mt_tape_op;

  mt_tape_op.mt_op = mt_op;
  mt_tape_op.mt_count = mt_count;

  if (!(rc = ioctl (dev_fd, MTIOCTOP, &mt_tape_op))) {
    PRINTF ("The MTIOCTOP ioctl succeeded.\n");
  }
  else {
    PERROR ("The MTIOCTOP ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* mt_tape_op */

/*---------------------------------------------------------------------------*/
/* Name:        st_get_device_status                                         */
/* Synopsis:    Display device status.                                       */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_get_device_status ()
{
  int rc;
  device_status_t device_status;

  if (!(rc = ioctl (dev_fd, STIOC_GET_DEVICE_STATUS, &device_status))) {
    PRINTF ("The STIOC_GET_DEVICE_STATUS ioctl succeeded.\n");
    PRINTF ("\nThe device status data is:\n");
    dump_bytes ((char *)&device_status, sizeof (device_status_t));
  }
  else {
    PERROR ("The STIOC_GET_DEVICE_STATUS ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_get_device_status */


/*---------------------------------------------------------------------------*/
/* Name:        st_get_device_info                                           */
/* Synopsis:    Display device information.                                  */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_get_device_info ()
{
  int rc;
  device_info_t device_info;

  if (!(rc = ioctl (dev_fd, STIOC_GET_DEVICE_INFO, &device_info))) {
    PRINTF ("The STIOC_GET_DEVICE_INFO ioctl succeeded.\n");
    PRINTF ("\nThe device information is:\n");
    dump_bytes ((char *)&device_info, sizeof (device_info_t));
  }
  else {
    PERROR ("The STIOC_GET_DEVICE_INFO ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_get_device_info */


/*---------------------------------------------------------------------------*/
/* Name:        st_get_media_info                                            */
/* Synopsis:    Display media information.                                   */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/

static int st_get_media_info ()
{
  int rc;
  media_info_t media_info;

  if (!(rc = ioctl (dev_fd, STIOC_GET_MEDIA_INFO, &media_info))) {
    PRINTF ("The STIOC_GET_MEDIA_INFO ioctl succeeded.\n");
    PRINTF ("\nThe loaded media information is:\n");
    if (device.ultrium) {
      if (media_info.write_protect)
        PRINTF ("The tape is in Physical Write Protect.\n");
      else
        PRINTF ("The tape isn't in Write Protect\n");
      dump_bytes ((char *)&media_info, sizeof (media_info_t));
      return (1);
    }

    PRINTF ("Media Type:    ");
    switch (media_info.media_type) {
     case  0x0:
       if (device.ultrium)
          PRINTF ("Unknown Medium Present\n");
       else
          PRINTF ("No Medium Present\n");
     break;

     case  0x0018:
        PRINTF ("Ultrium LTO-1 1/8T Tape Cartridge\n");
     break;

     case  0x0028:
        PRINTF ("Ultrium LTO-2 2/8T Tape Cartridge\n");
     break;

     case  0x0038:
        PRINTF ("Ultrium LTO-3 3/16T Tape Cartridge\n");
     break;

     case  0x003C:
        PRINTF ("Ultrium LTO-3 3/16T WORM Tape Cartridge\n");
     break;

     case  0x0048:
        PRINTF ("Ultrium LTO-4 3/16T Tape Cartridge\n");
     break;

     case  0x004C:
        PRINTF ("Ultrium LTO-4 3/16T WORM Tape Cartridge\n");
     break;

     case  0x0190:
        PRINTF ("IBM 3480/3490E/3490C Tape Cartridge\n");
     break;

     case  0x0081:
        PRINTF ("TotalStorage 3590 High Performance Cartridge\n");
     break;

     case 0x0082:
         PRINTF ("TotalStorage 3590 Extended High Performance Cartridge\n");
     break;

     case 0x0091:
         PRINTF ("TotalStorage 3592 J1A Enterprise Tape Cartridge\n");
     break;

     case 0x00B1:
         PRINTF ("TotalStorage 3592 J1A Enterprise Tape Short Cartridge\n");
     break;

     case 0x00A1:
         PRINTF ("TotalStorage 3592 J1A Enterprise Tape WORM Cartridge\n");
     break;

     case 0x00C1:
         PRINTF ("TotalStorage 3592 J1A Enterprise Tape Short WORM Cartridge");
         break;

     case 0x0092:
         PRINTF ("TotalStorage 3592 E05 Enterprise Tape Cartridge\n");
     break;

     case 0x00B2:
         PRINTF ("TotalStorage 3592 E05 Enterprise Tape Short Cartridge\n");
     break;

     case 0x00A2:
         PRINTF ("TotalStorage 3592 E05 Enterprise Tape WORM Cartridge\n");
     break;

     case 0x00C2:
         PRINTF ("TotalStorage 3592 E05 Enterprise Tape Short WORM Cartridge");
         break;

     case 0x0042:
         PRINTF ("Magstar MP Fast Access Linear Tape B-Format\n");
     break;

     case 0x0043:
         PRINTF ("Magstar MP Fast Access Linear Tape C-Format\n");
     break;

     case 0x0044:
         PRINTF ("Magstar MP Fast Access Linear Tape C-Format XL\n");
     break;

     default:
         PRINTF ("Unknown\n");
     break;
     }

    PRINTF ("Media Format:  ");
    switch (media_info.media_format) {
     case 0x0:
          PRINTF ("No Medium Present\n");
     break;

     case 0x82:
         PRINTF ("Magstar 3570 MP B-Series Model - B-Format\n");
     break;

     case 0x83:
         PRINTF ("Magstar 3570 MP C-Series Model - C-Format\n");
     break;

     case 0x29:
         PRINTF ("TotalStorage 3590 Model-B Format\n");
     break;

     case 0x2A:
         PRINTF ("TotalStorage 3590 Model-E Format\n");
     break;

     case 0x2C:
         PRINTF ("TotalStorage 3590 Model-H Format\n");
     break;

     case 0x51:
         PRINTF ("TotalStorage 3592 Model-J1A Format\n");
     break;

     case 0x52:
         PRINTF ("TotalStorage 3592 Model-E05 Format\n");
     break;

     case 0x71:
         PRINTF ("IBM TotalStorage 3592 Model J1A Encryption Format\n");
     break;

     case 0x72:
         PRINTF ("IBM TotalStorage 3592 Model E05 Encryption Format\n");
     break;

     case 0x40:
         PRINTF ("Ultrium_1/8T Format\n");
     break;

     case 0x42:
         PRINTF ("Ultrium_2/8T Format\n");
     break;

     case 0x44:
         PRINTF ("Ultrium_3/16T Format\n");
     break;

     case 0x46:
         PRINTF ("Ultrium_4/16T Format\n");
     break;

     default:
         PRINTF ("Unknown Medium Format\n");
     break;
    }

    PRINTF ("Write Protect: ");
    switch (media_info.write_protect) {
      case 0x0:
        PRINTF ("No write protect\n");
      break;

      case 0x01:
        PRINTF ("Worm write protect\n");
      break;

      case 0x10:
        PRINTF ("Persistant write protect\n");
      break;

      case 0x40:
        PRINTF ("Associated write protect\n");
      break;

      case 0x80:
        PRINTF ("Physical write protect\n");
      break;
    }
  }
  else {
    PERROR ("The STIOC_GET_MEDIA_INFO ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }
  return (rc);
}
/* st_get_media_info */

/*---------------------------------------------------------------------------*/
/* Name:        st_get_position                                              */
/* Synopsis:    Display tape position information.                           */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/

static int st_get_position ( int type )
{
  int rc;
  position_data_t position_data;

  position_data.block_type = type;

  if (!(rc = ioctl (dev_fd, STIOC_GET_POSITION, &position_data))) {
    PRINTF ("The STIOC_GET_POSITION ioctl succeeded.\n");
    PRINTF ("\nThe tape position data is:\n");
    dump_bytes ((char *)&position_data, sizeof (position_data_t));
  }
  else {
    PERROR ("The STIOC_GET_POSITION ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_get_position */


/*---------------------------------------------------------------------------*/
/* Name:        st_set_position                                              */
/* Synopsis:    Set tape position.                                           */
/* Returns:     Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static int st_set_position ( int type, int value )
{
  int rc;
  position_data_t position_data;

  position_data.block_type = type;
  position_data.position = value;

  if (!(rc = ioctl (dev_fd, STIOC_SET_POSITION, &position_data))) {
    PRINTF ("The STIOC_SET_POSITION ioctl succeeded.\n");
  }
  else {
    PERROR ("The STIOC_SET_POSITION ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_set_position */


/*---------------------------------------------------------------------------*/
/* Name:        st_get_parm                                                  */
/* Synopsis:    Display device specific parameter.                           */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_get_parm ( int type )
{
  int rc;
  parm_data_t parm_data;

  parm_data.type = type;

  if (medium_chgr) {
     printf("Invalid command for medium changer.\n");
     return (1);
  }

  if (device.ultrium) {
     if (parm_data.type == 7) {
        FPRINTF (stderr, "%s does not support get acfmode parameter.\n", dev_name);
        return (1);
     }
     if (parm_data.type == 8) {
        FPRINTF (stderr, "%s does not support get capacity parameter.\n", dev_name);
        return (1);
     }
  }

   printf("\n-- Tape Parameters ---------------------\n\n");

   parm_data.type = BLOCKSIZE;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    1. Block Size......... %-10d\n",
         parm_data.value);
   }

   parm_data.type = COMPRESSION;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    2. Compression........ %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }

   parm_data.type = BUFFERING;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    3. Buffering.......... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }

   parm_data.type = IMMEDIATE;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    4. Rewind Immediate... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }

   parm_data.type = TRAILER;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    5. Trailer Labels..... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }

   parm_data.type = WRITEPROTECT;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    6. Write Protection... ");
      switch(parm_data.value) {
      case NO_PROTECT:
         printf("None\n");
         break;
      case PHYS_PROTECT:
         printf("Physical\n");
         break;
      case WORM_PROTECT:
         printf("Logical (WORM)\n");
         break;
      case PERS_PROTECT:
         printf("Logical (PERS)\n");
         break;
      case ASSC_PROTECT:
         printf("Logical (ASSC)\n");
         break;
      default:
         printf("Unsupported\n");
      }
   }

   parm_data.type = ACFMODE;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    7. ACF Mode........... ");
      switch(parm_data.value) {
      case NO_ACF:
         printf("No ACF\n");
         break;
      case SYSTEM_MODE:
         printf("System\n");
         break;
      case RANDOM_MODE:
         printf("Random\n");
         break;
      case MANUAL_MODE:
         printf("Manual\n");
         break;
      case ACCUM_MODE:
         printf("Accumulate\n");
         break;
      case AUTO_MODE:
         printf("Automatic\n");
         break;
      case LIB_MODE:
         printf("Library\n");
         break;
      default:
         printf("Unsupported\n");
      }
   }

   parm_data.type = SCALING;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    8. Capacity Scaling... ");
      switch(parm_data.value) {
      case SCALE_100:
         printf("100%%\n");
         break;
      case SCALE_75:
         printf("75%%\n");
         break;
      case SCALE_50:
         printf("50%%\n");
         break;
      case SCALE_25:
         printf("25%%\n");
         break;
      default:
         if (device.t3592) {
            if (parm_data.value == 0)
               printf("100%%\n");
            else {
               printf("%2.2f%%\n", (float)parm_data.value/256*100);
               break;
            }
         }
         else
            printf("100%%\n");
      }
   }

   parm_data.type = SILI;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    9. SILI Mode.......... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }

   printf("\n----------------------------------------\n\n");

   if (rc) {
      PERROR ("STIOC_GET_PARM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }

   return (rc);

}
/* st_get_parm */


/*---------------------------------------------------------------------------*/
/* Name:        st_set_parm                                                  */
/* Synopsis:    Set device specific parameter.                               */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_set_parm ( int type, int value )
{
  int rc;
  char buf[10];
  parm_data_t parm_data;

  parm_data.type = type;
  parm_data.value = value;

  if ( device.ultrium ) {
     if ( parm_data.type == 7 ) {
        FPRINTF (stderr, "%s does not support set acfmode parameter.\n", dev_name);
        return (1);
     }
     if ( parm_data.type == 8 ) {
        FPRINTF (stderr, "%s does not support set capacity parameter.\n", dev_name);
        return (1);
     }
  }

  if ( parm_data.type == 8 ) {
     printf("Warning!  All data on tape will be lost with this change!\n");
     printf("Continue (y or n): ");
     gets(buf);
     if(buf[0] == 'y') {
       printf("Rewinding the tape...\n");
       if (st_tape_op(ST_OP_REW, 0))  return errno;
       verbose = TRUE;
     }
  }

  if (!(rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data))) {
    PRINTF ("The STIOC_SET_PARM ioctl succeeded.\n\n");
    st_get_parm (0);
  }
  else {
    PERROR ("The STIOC_SET_PARM ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_set_parm */


/*---------------------------------------------------------------------------*/
/* Name:        st_sync_buffer                                               */
/* Synopsis:    Flush device buffer.                                         */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_sync_buffer ()
{
  int rc;

  if (!(rc = ioctl (dev_fd, STIOC_SYNC_BUFFER, 0))) {
    PRINTF ("The STIOC_SYNC_BUFFER ioctl succeeded.\n");
  }
  else {
    PERROR ("The STIOC_SYNC_BUFFER ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_sync_buffer */


/*---------------------------------------------------------------------------*/
/* Name:        st_display_msg                                               */
/* Synopsis:    Display message on device operator panel.                    */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int st_display_msg ( int type, char *msg_0, char *msg_1 )
{
  int rc;

  msg_data_t msg_data;
  char msg_type[] = { DISPMSG0, DISPMSG1, FLASHMSG0, FLASHMSG1, ALTERNATE };

  msg_data.function = msg_type[type-1];
  memcpy (msg_data.msg_0, msg_0, 8);
  memcpy (msg_data.msg_1, msg_1, 8);

  if ( device.ultrium ) {
     FPRINTF (stderr, "%s does not support display message.\n", dev_name);
     return (1);
  }

  if (!(rc = ioctl (dev_fd, STIOC_DISPLAY_MSG, &msg_data))) {
    PRINTF ("The STIOC_DISPLAY_MSG ioctl succeeded.\n");
  }
  else {
    PERROR ("The STIOC_DISPLAY_MSG ioctl failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* st_display_msg */


/*---------------------------------------------------------------------------*/
/* Name:        st_report_density                                            */
/* Synopsis:    Report Media density                                         */
/* Returns:     Error code from /usr/include/sys/errno.h                     */
/*---------------------------------------------------------------------------*/

static int st_report_density (void)
{
  int  i;
  int  rc;
  int  width, media_tracks;
  long bits, media_capacity;
  rpt_dens_sup_t     density;
  density_report_t   report;

  density.media = ALL_MEDIA_DENSITY;
  PRINTF ("Issuing Report Density Support for ALL media... \n");

  if (!(rc = ioctl(dev_fd, STIOC_REPORT_DENSITY_SUPPORT, &density))) {
     PRINTF ("The STIOC_REPORT_DENSITY_SUPPORT ioctl succeeded.\n");
     PRINTF ( "Total number of densities reported: %d \n", density.count);
     for ( i=0; i<density.count; i++ ) {
        PRINTF ("\n");
        PRINTF ("Density Name................. %0.8s \n", density.reports[i].density_name);
        PRINTF ("Assigning Organization....... %0.8s \n", density.reports[i].assigning_org) ;
        PRINTF ("Description.................. %0.20s \n", density.reports[i].description);
        PRINTF ("Primary Density Code......... %02X \n", density.reports[i].pri_density_code);
        PRINTF ("Secondary Density Code....... %02X \n", density.reports[i].sec_density_code);
        if (density.reports[i].wrtok)
           PRINTF ("Write OK..................... Yes \n");
        else
           PRINTF ("Write OK..................... No \n");
        if (density.reports[i].dup)
           PRINTF ("Duplicate.................... Yes \n");
        else
           PRINTF ("Duplicate.................... No \n");
        if (density.reports[i].deflt)
           PRINTF ("Default...................... Yes \n");
        else
           PRINTF ("Default...................... No \n");
        bits = (density.reports[i].bits_per_mm[0] << 16) + (density.reports[i].bits_per_mm[1] << 8) + density.reports[i].bits_per_mm[2];
        PRINTF("Bits per MM.................. %ld \n", bits);
        width = ((density.reports[i].media_width[0] << 8) + density.reports[i].media_width[1]);
        PRINTF ("Media Width (millimeters).... %d \n", width);
        media_tracks = ((density.reports[i].tracks[0] << 8) + density.reports[i].tracks[1]);
        PRINTF ("Tracks....................... %d \n", media_tracks);
        media_capacity = (density.reports[i].capacity[0] << 24) + (density.reports[i].capacity[1] << 16) + (density.reports[i].capacity[2] << 8) + (density.reports[i].capacity[3]);
        PRINTF ("Capacity (megabytes)......... %ld \n", media_capacity);
        PRINTF ("\n--- more ---");
        getchar ();
     }
  }
  else {
     PERROR ("The STIOC_REPORT_DENSITY_SUPPORT ioctl for all media failed");
     scsi_request_sense ();
  }

  density.media = CURRENT_MEDIA_DENSITY;
  PRINTF ("\nIssuing Report Density Support for CURRENT media... \n");

  if (!(rc = ioctl(dev_fd, STIOC_REPORT_DENSITY_SUPPORT, &density))) {
     PRINTF ("The STIOC_REPORT_DENSITY_SUPPORT ioctl succeeded.\n");
     PRINTF ( "Total number of densities reported: %d \n", density.count);
     for (i=0; i<density.count; i++) {
        PRINTF ("\n");
        PRINTF ("Density Name................. %0.8s \n", density.reports[i].density_name);
        PRINTF ("Assigning Organization....... %0.8s \n", density.reports[i].assigning_org);
        PRINTF ("Description.................. %0.20s \n", density.reports[i].description);
        PRINTF ("Primary Density Code......... %02X \n", density.reports[i].pri_density_code);
        PRINTF ("Secondary Density Code....... %02X \n", density.reports[i].sec_density_code);
        if (density.reports[i].wrtok)
          PRINTF ("Write OK..................... Yes \n");
        else
          PRINTF ("Write OK..................... No \n");
        if (density.reports[i].dup)
          PRINTF ("Duplicate.................... Yes \n");
        else
          PRINTF ("Duplicate.................... No \n");
        if (density.reports[i].deflt)
          PRINTF ("Default...................... Yes \n");
        else
          PRINTF ("Default...................... No \n");
        bits = (density.reports[i].bits_per_mm[0] << 16) + (density.reports[i].bits_per_mm[1] << 8) + density.reports[i].bits_per_mm[2];
        PRINTF("Bits per MM.................. %ld \n", bits);
        width = ((density.reports[i].media_width[0] << 8) + density.reports[i].media_width[1]);
        PRINTF ("Media Width (millimeters).... %d \n", width);
        media_tracks = ((density.reports[i].tracks[0] << 8) + density.reports[i]
.tracks[1]);
        PRINTF ("Tracks....................... %d \n", media_tracks);
        media_capacity = (density.reports[i].capacity[0] << 24) + (density.reports[i].capacity[1] << 16) + (density.reports[i].capacity[2] << 8) + (density.reports[i].capacity[3]);
        PRINTF ("Capacity (megabytes)......... %ld \n", media_capacity);
        PRINTF ("\n--- more ---");
        getchar ();
     }
  }
  else {
     PERROR ("The STIOC_REPORT_DENSITY_SUPPORT ioctl for current media failed");
     scsi_request_sense ();
  }

   return (rc);
}
/* st_report_density */


/*---------------------------------------------------------------------------*/
/* Name:        get_pid                                                      */
/* Synopsis:    Get product ID into null-terminated                          */
/*              string, then convert all trailing blanks to nulls.           */
/*---------------------------------------------------------------------------*/
static int get_pid(char *pid )
{
   int rc = 0;
   int i;

   rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data);
   if (rc != 0) {
      (void) fprintf (stderr, "INQUIRY rc %d (%s)\n", errno, strerror(errno));
      scsi_request_sense ();
      return errno;
   }
   tape_drv = inquiry_data.type & INQ_DRV_TYPE;
   medium_chgr = inquiry_data.type & INQ_SMC_TYPE;
   (void) memset (pid, 0, PIDLEN + 1);
   *((int *)&device) = 0;
   bcopy ((const char *)inquiry_data.pid, pid, PIDLEN);
   /* Start at last char before final \0, look for blanks, change them to \0s. */
   for ( i = PIDLEN - 1; i >= 0; i-- ) {
     if (pid[i] == ' ') pid[i] = '\0';
     else break;
   }
   if (!strncmp (pid,"ULT3580-TD",10))          device.ultrium = TRUE;
   if (!strncmp (pid,"ULTRIUM-TD",10))          device.ultrium = TRUE;
   if (!strncmp (pid,"ULT3580-HH",10))          device.ultrium = TRUE;
   if (!strncmp (pid,"ULTRIUM-HH",10))          device.ultrium = TRUE;
   if (!strncmp (pid,"ULT3580-TD1",11))         device.lto1    = TRUE;
   if (!strncmp (pid,"ULTRIUM-TD1",11))         device.lto1    = TRUE;

   if (!strncmp (pid,"ULT3580-TD2",11))         device.lto2    = TRUE;
   if (!strncmp (pid,"ULTRIUM-TD2",11))         device.lto2    = TRUE;

   if (!strncmp (pid,"ULT3580-TD3",11))         device.lto3    = TRUE;
   if (!strncmp (pid,"ULTRIUM-TD3",11))         device.lto3    = TRUE;
   if (!strncmp (pid,"ULT3580-HH3",11))         device.lto3    = TRUE;
   if (!strncmp (pid,"ULTRIUM-HH3",11))         device.lto3    = TRUE;

   if (!strncmp (pid,"ULT3580-TD4",11))         device.lto4    = TRUE;
   if (!strncmp (pid,"ULTRIUM-TD4",11))         device.lto4    = TRUE;
   if (!strncmp (pid,"ULT3580-HH4",11))         device.lto4    = TRUE;
   if (!strncmp (pid,"ULTRIUM-HH4",11))         device.lto4    = TRUE;

   if (!strcmp (pid, "ULT3581-TA"))             device.ult3581 = TRUE;
   if (!strcmp (pid, "ULT3582-TL"))             device.ult3582 = TRUE;
   if (!strcmp (pid, "ULT3583-TL"))             device.ult3583 = TRUE;
   if (!strncmp (pid, "03584", 5))              device.ult3584 = TRUE;
   if (!(strncmp (pid, "03490", 5)))          { device.t3490   = TRUE;
                                                pid[8] = '\0'; }
   if (!(strncmp (pid, "03570", 5)))            device.t3570   = TRUE;
   if (!(strncmp (pid, "03590", 5)))            device.t3590   = TRUE;
   if (!(strncmp (pid, "03592", 5)))            device.t3592   = TRUE;
   if (!(strncmp (pid, "ULT3581-TA2",11)))      device.ny3581  = TRUE;
   if (!strncmp (pid, "3576-MTL",8))            device.ult3576 = TRUE;
   if (!strncmp (pid, "3573-TL",7))             device.ult3573 = TRUE;
   if (!strncmp (pid, "3577-TL",7))             device.lib3577 = TRUE;

   return (rc);
}
/* get_pid */


/*---------------------------------------------------------------------------*/
/*        Name: device_path                                                  */
/*    Synopsis: Display all of paths                                         */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int device_paths()
{

  int rc = 0;
  struct device_paths paths;
  int i;

  PRINTF("Querying device paths...\n");

memset(&paths, 0, sizeof(struct device_paths));

  if(!(rc = ioctl(dev_fd, STIOC_DEVICE_PATH, &paths)))
    {
      PRINTF("\n");
      for (i=0; i < paths.number_paths; i++)
        {
          if (i == 0)
            {
              PRINTF("Primary Path Number %d\n", i);
            }
          else
            {
              PRINTF("Alternate Path Number %d\n", i);
            }
          PRINTF("  Logical Device................. %s\n",paths.device_path[i].name);
          PRINTF("  Host Bus Adapter............... %s\n",paths.device_path[i].parent);
          if (paths.device_path[i].id_valid)
            {
              PRINTF("  SCSI Channel................... %d\n",paths.device_path[i].bus);
              PRINTF("  Target ID...................... %d\n",paths.device_path[i].id);
              PRINTF("  Logical Unit................... %d\n",paths.device_path[i].lun);
            }
          if (paths.device_path[i].fcp_id_valid)
            {
              PRINTF("  WWNN........................... %lld\n",paths.device_path[i].fcp_ww_name);
            }
          if (paths.device_path[i].drive_port_valid)
            {
              PRINTF("  Drive Port..................... %d\n",paths.device_path[i].drive_port);
            }
          if (paths.device_path[i].enabled)
            {
              PRINTF("  Path Enabled................... Yes\n");
            }
          else
            {
              PRINTF("  Path Enabled................... No \n");
            }
          if (paths.device_path[i].fenced)
            {
              PRINTF("  Path Manually Disabled......... Yes\n");
            }
          else
            {
              PRINTF("  Path Manually Disabled......... No \n");
            }

          PRINTF("\n");
        }

      PRINTF("Total paths configured........... %d\n",paths.number_paths);
      PRINTF("The current active path.......... %d\n",paths.cur_path);
    }
  return rc;

}


static int rwtest(int blk_size, int blk_fact, int multiplier, int seed)
{
  int rc;
  int i,j;
  int count;
  int length;
  long long total_bytes;
  struct timeval start;
  struct timeval stop;
  struct timezone zone;
  char *r_data;
  char *w_data;
  char *ptrn;
  int error;

  /*rewinding */
  PRINTF ("Rewinding \n", total_bytes );
  if ( tape_op_mode == MT_MODE )
    rc = mt_tape_op (MTREW,1);
  else
    rc = st_tape_op (ST_OP_REW,1);

  if (!(w_data = malloc (RWBUFSZ)))
    {
      PERROR ("Failure allocating write data buffer");
      return (errno);
    }

  if (!(r_data = malloc (RWBUFSZ)))
    {
      PERROR ("Failure allocating read data buffer");
      return (errno);
    }

  /* generate data using random pattern */
  while (!seed)
    {
      /* 0 < seed < 32768 */
      gettimeofday (&start, &zone);
      seed = (32768 + start.tv_usec) % 32768;
    }
  srand (seed);
  for (i = 0; i < (int)(blk_size * blk_fact); i++)
    {
      w_data[i] = (char)rand ();
    }
  PRINTF ("Generated %d bytes of random data using seed %d.\n\n",
          (blk_size * blk_fact), seed);


  total_bytes = (long long)0;
  gettimeofday (&start, &zone);

  /*write section */

  for (i = 1; i <= multiplier; i++)
    {
      count = write (dev_fd, w_data, (blk_size * blk_fact));
      rc = errno;

      if (count < 0)
        {
          if (i > 1)  PRINTF ("\n\n");
          PRINTF ("The WRITE command failed (request %d, count = %d, errno = %d).",
                  i, count, errno);
          break;
        }

      if (count != (blk_size * blk_fact))
        {
          if (i > 1)  PRINTF ("\n\n");
          PRINTF ("Short WRITE (request %d, count = %d).", i, count);
          count = -1;
          break;
        }

      total_bytes = total_bytes + (long long)count;
      PRINTF ("*");

      if (!(i % 32)) {
        PRINTF ("  %lld\n", total_bytes);
      }
    }/*end of  for (i = 1; i <= multiplier; i++) for write section*/

  gettimeofday (&stop, &zone);

  if ((i % 32) || (count < 0))
    {
      PRINTF ("\n");
    }

  PRINTF ("\n");

  if (count < 0)
    {
      scsi_request_sense ();
    }
  else
    {
      PRINTF ("Wrote %lld bytes \n", total_bytes );
    }

  /*rewinding */
  PRINTF ("Rewinding \n", total_bytes );
  if ( tape_op_mode == MT_MODE )
    rc = mt_tape_op (MTREW,1);
  else
    rc = st_tape_op (ST_OP_REW,1);


  total_bytes= (long long)0;
  error=0;
  i=1;

  /*read (check) section */
  while (i<=multiplier && !error)
    {
      count = read (dev_fd, r_data, (uint)(blk_size * blk_fact));
      rc = errno;

      if (count < 0)
        {
          if (i > 1)  PRINTF ("\n\n");
          PRINTF ("The READ command failed (request %d, count = %d, errno = %d).",
                  i, count, errno);
          error=1;
          break;

        }

      if (count != (blk_size * blk_fact))
        {
          if (i > 1)  PRINTF ("\n\n");
          PRINTF ("Short READ (request %d, count = %d).", i, count);
          count = -1;
          error=1;
          break;
        }

      for (j=0;j<(blk_size * blk_fact);j++)
        {
          if(r_data[j]!= w_data[j])
            {
              PRINTF ("\n  ***Data mismatch!***\n");
              error=1;
              break;
            }
        }

      total_bytes = total_bytes + (long long)count;
      PRINTF ("*");

    if (!(i % 32))
      {
        PRINTF ("  %lld\n", total_bytes);
      }

    i++;
    }/*end of for while loop for read section*/

  gettimeofday (&stop, &zone);

  if ((i % 32) || (count < 0)) {
    PRINTF ("\n");
  }

  PRINTF ("\n");

  if (count < 0) {
    scsi_request_sense ();
  }
  else {
    PRINTF ("Read %lld bytes \n", total_bytes );
  }

  free (r_data);
  free (w_data);
  return (rc);
}
/* wr_test */

/*---------------------------------------------------------------------------*/
/*        Name: st_get_density                                               */
/*    Synopsis: Get tape density.                                            */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_density(void)
{
  density_data_t* data;
  int rc=0;

  data = malloc (sizeof(density_data_t));
  if (data == NULL)
  {
     return -ENOMEM;
  }

  memset(data, 0, sizeof(density_data_t));

  rc = ioctl(dev_fd, STIOC_GET_DENSITY, data);
  if(rc != 0)
    {
      printf("The STIOC_GET_DENSITY ioctl for current media failed, rc=%d\n",rc);
      scsi_request_sense ();
      return rc;
    }

  printf("density code = %x\n", data->density_code);
  printf("default_density = %x\n", data->default_density);
  printf("pending_density = %x\n", data->pending_density);

  return rc;
}

/*---------------------------------------------------------------------------*/
/*        Name: st_get_encryption_state                                      */
/*    Synopsis: Query drive encrytion status; print out result.              */
/*     Returns: Returns: Error code from /usr/include/sys/errno.h.           */
/*---------------------------------------------------------------------------*/
static int st_get_encryption_state ( )
{
   int rc;
   struct encryption_status  ep_state;

   if (!(rc = ioctl (dev_fd, GET_ENCRYPTION_STATE, &ep_state))) {
      PRINTF ("\nGET_ENCRYPTION_STATE command succeeded.\n");
      PRINTF ("Encryption settings:\n");
      PRINTF ("%20s %s\n", "  Encryption Capable...", (ep_state.encryption_capable == 1 ? "Yes" : "No"));
      PRINTF ("%20s ", "  Encryption Method....");
      switch (ep_state.encryption_method) {
         case 0: PRINTF ("None\n"); break;
         case 1: PRINTF ("Library\n"); break;
         case 2: PRINTF ("System\n"); break;
         case 3: PRINTF ("Application\n"); break;
         case 4: PRINTF ("Custom\n"); break;
         case 5: PRINTF ("Unknown\n"); break;
      }
      PRINTF ("%20s ", "  Encryption State.....");
      switch (ep_state.encryption_state) {
         case 0: PRINTF ("Off\n"); break;
         case 1: PRINTF ("On\n"); break;
         case 2: PRINTF ("NA\n"); break;
      }
   }
   else {
      PERROR ("GET_ENCRYPTION_STATE failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}

/*---------------------------------------------------------------------------*/
/*        Name: smc_read_cartridge_location                                  */
/*    Synopsis: Read cartridge location with slot devid on                   */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_read_cartridge_location (int element_address, int number_elements)
{
   int rc;
   int i,j;
   cartridge_location_data_t *slot_devid, *slotp;
   read_cartridge_location_t slot_devids;

   slot_devids.element_address = (ushort)element_address;
   slot_devids.number_elements = (ushort)number_elements;


   slot_devid = malloc(slot_devids.number_elements * sizeof(cartridge_location_data_t));
   if (slot_devid == NULL)
      {
      errno = ENOMEM;
      return errno;
      }
   bzero((caddr_t)slot_devid,slot_devids.number_elements * sizeof(cartridge_location_data_t));

   slot_devids.data = slot_devid;

   printf("Reading cartridge location ...\n");

   if (rc = ioctl (dev_fd, SMCIOC_READ_CARTRIDGE_LOCATION, &slot_devids))
      {
      free(slot_devid);
      PERROR ("SMCIOC_READ_CARTRIDGE_LOCATION failed");
      PRINTF ("\n");
      scsi_request_sense ();

      return rc;
      }

   j=0;
   slotp = slot_devid;

   for (i = 0; i < slot_devids.number_elements; i++, slotp++)
      {
      /* don't overflow screen if menu mode */
      if (interactive && j == 2)
         {
         j=0;
         printf ("\nHit <enter> to continue...");
         getchar();
         }

      j++;
      printf("\nStorage Cell Address %d\n",slotp->address);
      if (slotp->except)
         printf("  Slot State ..................... See ASC/ASCQ\n");
      else
         printf("  Slot State ..................... Normal\n");
      if (slotp->asc == 0x11 && slotp->ascq ==0x00)
         printf("  ASC/ASCQ ....................... %02X%02X (Unable to Read Bar Code Label)\n",
               slotp->asc,slotp->ascq);
      else if (slotp->asc == 0x30 && slotp->ascq ==0x03)
         printf("  ASC/ASCQ ....................... %02X%02X (Cleaning Cartridge Installed)\n",
               slotp->asc,slotp->ascq);
      else if (slotp->asc == 0x81 && slotp->ascq ==0x00)
         printf("  ASC/ASCQ ....................... %02X%02X (Status is Questionable)\n",
               slotp->asc,slotp->ascq);
      else
         printf("  ASC/ASCQ ....................... %02X%02X\n",
               slotp->asc,slotp->ascq);
      if (slotp->full)
         printf("  Media Present .................. Yes\n");
      else
         printf("  Media Present .................. No\n");
      if (slotp->access)
         printf("  Robot Access Allowed ........... Yes\n");
      else
         printf("  Robot Access Allowed ........... No\n");
      if (slotp->svalid)
         printf("  Source Element Address ......... %d\n",slotp->source);
      else
         printf("  Source Element Address Valid ... No\n");
      if (slotp->invert)
         printf("  Media Inverted ................. Yes\n");
      else
         printf("  Media Inverted ................. No\n");
      printf("  Volume Tag ..................... %0.36s\n", slotp->volume);
      printf("  Code Set ....................... %x\n", slotp->code_set);
      printf("  Identifier Type ................ %d\n", slotp->ident_type);
      printf("  Identifier Length .............. %d\n", slotp->ident_len);
      printf("  Slot ID Info ................... %0.24s\n", slotp->identifier);
   }

   free(slot_devid);
   return errno;
}

/*---------------------------------------------------------------------------*/
/* Name: get_sernum command                                                  */
/*---------------------------------------------------------------------------*/
static int get_sernum()
{
   int rc;
   inquiry_page_t inquiry_page;
   char serial_num[20];

   inquiry_page.page_code = (uchar) 0x80;
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY_PAGE, &inquiry_page))) {
      bcopy(&inquiry_page.data[4], serial_num, inquiry_page.data[3]);
      serial_num[inquiry_page.data[3]] = '\0';
      printf("%s\n", serial_num);
   }
   else {
      printf ("Get serial number failed\n");
      scsi_request_sense ();
   }

   return (rc);
}

/*---------------------------------------------------------------------------*/
/* Name: rekey_command                                                       */
/*---------------------------------------------------------------------------*/
static int rekey_command()
  {
  #define SCMD_MODE_SEL_10                0x55

  int rc;
  struct sctl_io passthru;
  char *buffer;

  buffer = malloc(255);
  if (buffer == NULL) {
     errno = ENOMEM;
     printf ("Error %d (%s) Fail to allocate memory.\n",
            errno, strerror(errno));
     return errno;
   }

  /* build mode select x25/xEE in short form */
  bzero(&passthru, sizeof(struct sctl_io));
  bzero(buffer,255);
  passthru.cdb[0]=SCMD_MODE_SEL_10;
  passthru.cdb[1]=0x10;
  passthru.cdb[8]=56;
  passthru.cdb_length=10;
  passthru.data=buffer;
  passthru.data_length=56;
  passthru.max_msecs=600000;  /* allow 600 seconds for cmd */

  buffer[3]=0x10;
  buffer[8]=0x65;
  buffer[9]=(char)0xEE;
  buffer[11]=0x2C;
  buffer[12]=0x21;
  buffer[13]=(char)0xEE;
  buffer[15]=0x28;
  buffer[20]=0x18;
  buffer[28]=0x18;
  buffer[31]=0x04;
  buffer[39]=0x02;
  buffer[40]=0x10;
  buffer[44]=0x19;
  buffer[48]=0x10;
  buffer[52]=0x19;

  printf ("Issuing rekey command...\n");
  if ((rc = ioctl(dev_fd, SIOC_IO, &passthru)) <0 )
     printf("rekey command failed!\n");
  else if (passthru.cdb_status != S_GOOD) {
     printf("rekey command failed!\n");
     printf("   cdb_status = %x\n", passthru.cdb_status);
     printf("   sense_xfer=%d\n", passthru.sense_xfer);
     dump_bytes ((char *)passthru.sense,passthru.sense_xfer);
  }
  else
     printf("rekey command succeeded!\n");

  printf("\n");

  free(buffer);

  return (rc);
  }

