/***************************************************************************/
/**                                                                       **/
/**                   IBM SCSI Tape Utility Program                       **/
/**                      Copyright (C) 2001-                              **/
/**               International Business Machines Corporation             **/
/**                                                                       **/
/**               IBMtapeutil.c - Test Program Source File                **/
/**                Licensed Materials - Property of IBM                   **/
/**                         All Rights Reserved                           **/
/**                                                                       **/
/**    US Government Users Restricted Rights  -  Use, duplication, or     **/
/**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.  **/
/**                                                                       **/
/***************************************************************************/

#ifndef IBMTAPEUTIL_H
#define IBMTAPEUTIL_H

/* need manually toggle the mode */
#define DEBUG    0
#define ERR_STOP 1         /* do not continue if there is an error */

/* type definitions */
#ifndef unchar
 #define unchar unsigned char
#endif

/* include files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/mtio.h>
#include <scsi/scsi_ioctl.h>
#include <ctype.h>
#include <sys/stat.h>
#include <unistd.h>
#include "IBM_tape.h"

/*---------function prototypes---------- */
void device_menu(void);
void menu(void);
int  cmdline(int, char **);
void help(void);

/* General Commands */
int sioc_inquiry(void);
int sioc_reqsense(void);
int sioc_reserve(void);
int sioc_release(void);
int sioc_test_unit_ready(void);
int sioc_log_sense_page(int, int);
int sioc_mode_sense_page(int);
int sioc_inquiry_page(int);
int sioc_query_path(void);
int sioc_device_paths(void);
int sioc_enable_path(int);
int sioc_disable_path(int);

/* Tape Commands */
int mtioctop(int, int);
int mtiocqrypos(void);
int mtiocsetpos(void);
int mtdevice(void);
int stiocdm(char, char *, char *);
int display_msg(void);
int stioc_prevent_medium_removal(void);
int stioc_allow_medium_removal(void);
int status(void);
int twrite(char *);
int tread(char *, int);
int rwtest(int, int, int);
int rtest(int, int, int);
int wtest(int, int, int);

int stioctop(int, int);
int stiocqryp(void);
int stiocsetp(void);
int stiocsync(void);
int stiocqrypos(void);
int stiocsetpos(void);
int stiocqrysense(void);
int stiocqryinquiry(void);
int stioc_locate(void);
int stioc_read_position(void);
int stioc_report_density_support(void);
int tape_parms(void);
int list_tape(void);
int stioc_qry_driver_version(void);
int qry_encryption_status(void);
int ekm_test(void);

/* Medium Changer Commands */
int smcioc_element_info(void);
int smcioc_move_medium(ushort, ushort);
int smcioc_exchange_medium(ushort, ushort, ushort);
int smcioc_load_medium(ushort);
int smcioc_unload_medium(ushort);
int smcioc_pos_to_elem(ushort);
int smcioc_init_elem_stat(void);
int smcioc_init_elem_stat_range(ushort, ushort);
int smcioc_inventory(void);
int smcioc_prevent_medium_removal(void);
int smcioc_allow_medium_removal(void);
int smcioc_read_element_devids(void);
int smcioc_read_cartridge_location(void);

/* Service Aid Commands */
int stioc_read_dump(char *fname);
int stioc_force_dump(void);
int stioc_reset_drive_Atape(void);
int stioc_load_ucode(char *);
int stioc_read_keys(void);
int stioc_read_reservations(void);
int stioc_register_key(void);
int stioc_remove_registration(void);
int stioc_clear_all_registrations(void);

/* Support Functions */
int  search_drive_id(void);
void genpat(void);
void htoc(char *, unsigned char);
void dump_bytes(char *, int, char *);
int  isnumber(char *);
void show_path(struct scsi_path*);

/* Rekey Functions */
int rekey_command(void);
int get_sernum(void);

/*-------------------------- global variables -------------------------- */

int        exec_device;     /* menu execution device                */
int        opcode;          /* menu selected operation              */
int        mt_st_mode;      /* MT or ST ioctl                       */

short      source;          /* source for move command              */
short      dest;            /* destination for move command         */
short      dest2;           /* destination2 for exchange command    */
int        blksize;         /* block size for data transfers        */
int        blocks;          /* number of blocks/bytes to write      */
int        repetition;      /* number of write/read operation       */
short      stop;            /* st_op->stop field (for mt commands)  */
daddr_t    count;           /* st_op->count field (for mt commands) */
char       msg_type;        /* msg type field                       */
char       msg_0[9];        /* message 0 text                       */
char       msg_1[9];        /* message 1 text                       */
char       *sname;          /* source file name                     */
char       *dname;          /* destination file name                */

int        fd;                    /* file descriptor                 */
char       opened;                /* is special file open yet        */
char       devname[80];           /* device special file name        */
int        mode_flags;            /* flags to use for cmdline open   */
unchar     pat[256];              /* data pattern buffer             */

struct     mtpos mtpos;            /* structure for block position   */
struct     stchgp_s parms;         /* structure for tape parameters  */
struct     stpos_s stpos;          /* structure for tape position    */
uint       blockid;                /* for locate/read_position       */

struct     encryption_status encryption_status_t; /* for query/set encryption */
struct     data_key encryption_data_key; /* for set encryption data key */

int        saved_errno;            /* errno for the last command to  */
                                   /* the device                     */

/*-------------------- definitions and micros ---------------*/

/* device type */
#define    TAPE      1
#define    CHANGER   2

/* ST or MT mode */
#define    MTMODE    1
#define    STMODE    2

/* maximum size per transfer */
#define    MAX_TRANSFER_SIZE            65536    /* 64K */

/* data for passthrough */
#define    CDB_10_GENERIC               10       /* CDB length*/
#define    CDB_6_GENERIC                6

#define    SCSIOP_READ_BUFFER           0x3C     /* operation code */
#define    SCSIOP_WRITE_BUFFER          0x3B
#define    SCSIOP_SEND_DIAGNOSTIC       0x1D

#define    SCSI_PASSTHRU_TIMEOUT_VALUE  5000


/* microcode size */
#define MORGAN_MICROCODESIZE      63488

/* LTO device IDs and magstar devices IDs */
#define IBM_INVALID_DEVICE   0
#define IBM_3590             1
#define IBM_3592             2
#define IBM_3580             3
#define IBM_3581             4
#define IBM_3582             5
#define IBM_3583             6
#define IBM_3584             7
#define BDT_3581             8
#define IBM_3576             9
#define IBM_3573            10
#define IBM_3577            11

static char *SUPPORTED_VENDORS[] = { "IBM", "ADIC", NULL };

static char *SUPPORTED_TAPE_DRIVES[] = {  "03590",           /* 1  Magstar 3590  */
                                          "03592",           /* 2  Jaguar 3592   */
                                          "LTO TAPE DRIVE",  /* 3  Old Ultrium Inquiry String */
                                          "FastStor LTO",    /* 4  Old Morgan Inquiry String */
                                          "Scalar 100",      /* 5  Old Belgian Inquiry String */
                                          "ULT3580",         /* 6  LTO 3580 Ultrium  */
                                          "ULTRIUM",         /* 7  LTO 3580 Ultrium  */
                                          "ULT3581-TA2",     /* 8  LTO 3581 BDT     */
                                          "ULT3581-TA",      /* 9 LTO 3581 Morgan   */
                                          "ULT3582-TL",      /* 10 LTO 3582 Belgian  */
                                          "ULT3583-TL",      /* 11 LTO 3583 Belgian  */
                                          "03584L",          /* 12 LTO 3584 Anaconda */
                                          "3576-MTL",        /* 13 3576 Orion        */
                                          "3573-TL",         /* 14 3573 library      */
                                          "3577-TL",         /* 15 3577 library      */
                                          NULL };

static int SUPPORTED_TAPE_DRIVE_ID[] = {  IBM_3590,          /* 1  Magstar 3590  */
                                          IBM_3592,          /* 2  Jaguar  3592  */
                                          IBM_3580,          /* 3  Old Ultrium Inquiry String */
                                          IBM_3581,          /* 4  Old Morgan Inquiry String */
                                          IBM_3583,          /* 5  Old Belgian Inquiry String */
                                          IBM_3580,          /* 6  LTO 3580 Ultrium  */
                                          IBM_3580,          /* 7  LTO 3580 Ultrium  */
                                          BDT_3581,          /* 8 LTO 3581 BDT      */
                                          IBM_3581,          /* 9 LTO 3581 Morgan   */
                                          IBM_3582,          /* 10 LTO 3582 Belgian  */
                                          IBM_3583,          /* 11 LTO 3583 Belgian  */
                                          IBM_3584,          /* 12 LTO 3584 Anaconda */
                                          IBM_3576,          /* 13 LTO 3576 Orion    */
                                          IBM_3573,          /* 14 LTO 3573 library  */
                                          IBM_3577           /* 15 3577 library      */
                                          };

#endif







