/*******************************************************************************
********************************************************************************
**                                                                            **
**              IBM SCSI Tape & Medium Changer Utility Program                **
**                   (C) COPYRIGHT IBM Corporation, 2008-                     **
**                                                                            **
**                ntutil.h - Utility Program Include File                     **
**                  Licensed Materials - Property of IBM                      **
**                           All Rights Reserved                              **
**                                                                            **
**    US Government Users Restricted Rights  -  Use, duplication, or          **
**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.       **
**                                                                            **
**           ** Notice to users of the source code examples: **               **
**                                                                            **
**  The source code examples provided by IBM are only intended to assist      **
**  in the development of a working software program.  The source code        **
**  examples may not function as written; additional code is required.        **
**  In addition, the source code examples may not compile and/or bind         **
**  successfully as written.                                                  **
**                                                                            **
**  International Business Machines Corporation provides the source code      **
**  examples "as is" without warranty of any kind, either expressed or        **
**  implied, including, but not limited to the implied warranties of          **
**  merchantability and fitness for a particular purpose.  The entire         **
**  risk as to the quality and performance of the source code examples        **
**  is with you.  Should any part of the source code examples prove           **
**  defective, you (and not IBM) assume the entire cost of all necessary      **
**  servicing, repair or correction.                                          **
**                                                                            **
**  IBM does not warrant that the contents of the source code examples        **
**  will meet your requirements or that the source code examples are          **
**  error-free.                                                               **
**                                                                            **
**  IBM may make improvements and/or changes in the source code examples      **
**  at any time.  Changes may be made periodically to the information in      **
**  the source code examples; these changes may be reported, for the          **
**  device drivers included herein, in new editions of the examples.          **
**                                                                            **
**  References in the source code examples to IBM products, programs, or      **
**  services do not imply that IBM intends to make these available in all     **
**  countries in which IBM operates.  Any reference to an IBM licensed        **
**  program in the source code examples is not intended to state or imply     **
**  that only IBM's licensed program may be used.  Any functionally           **
**  equivalent program may be used.                                           **
**                                                                            **
**   COMPILER:        Microsoft Visual C++                                    **
**                                                                            **
**   PRIVATE FUNCTIONS: N/A                                                   **
**                                                                            **
**   PUBLIC  FUNCTIONS:                                                       **
**                                                                            **
**                                                                            **
********************************************************************************

/*******************************************************************************
*                               INCLUDE FILES                                  *
********************************************************************************/

// Needs manually toggle these compiler flags
#define debug 0
#define changer_debug 0
#define NT4 0
#define W2K 0
#define WNET 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <direct.h>
#include <errno.h>
#include <fcntl.h>
#include <io.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <windows.h>

#include <winioctl.h>
#include <winnt.h>
#include <wtypes.h>

/* NT DD includes from DDK\src\storage\inc */
#include <ntddscsi.h>

/* grab the right copy of the magvers file, depending on WNET, W2K, or WNT compile */
#if NT4

#include <magvers.h>

#else

  #if W2K

  #define MAGCHANGER    1 /* dummy variable for compatibility with magvers.h */
  #include "..\..\..\..\ntddk\src\storage\inc\magvers.h"

  #else

    #if WNET
    #define IBM_CHANGER 1 /* dummy variable for compatibility with ibmrmss.h */
    /* #include "ibmrmss.h" */ /* be sure to place this file in the right include directory */
                          /* a dummy build.h file is also necessary */
    #else

    Error -- cannot include correct version file

    #endif

  #endif

#endif


/*
** sizeof various types
** char   1
** UCHAR  1
** USHORT 2
** int    4
** long   4
** ULONG  4
** DWORD  4
** double 8
*/

/*
** macros and definitions
*/
#define iffeature(x)  if(gcp->Features0 &x)   {printf("    ");printf(#x);printf("\n");}
#define LINEFEED            '\x0a'
#define MAX_TOKEN               50
#define MAX_TRANSFER        655360  // read/write ops
#define MAX_TRANSFER_INDEX    1024  //
#define MAX_TRANSFER_BLOCKS    640  // for init of write data buffer index max_transfer/max_transfer_index

#define MAX_TRANSFER_SPT     65536  // can not exceede Adapters MaxScatter limit
                                    // else invalid parm on scsi pass thru....
#define NOERROR                  0
#define TESTFAIL                -1
#define ERRSYNTAX               -2
#define ERREXIT                 -3
#define ERRMOVE                 -4
#define ERRDATA                 -5
#define BAD_OPCODE            8182
#define OMODE_INVALID           -1
#define VENDLEN                 20
#define MAX_BARCODE_LEN         36              /* BLD, 06/07/04 */
#define BARCODE_LEN             8               /* BLD, 06/07/04 */
#define UNSET               0xFFFF
#define MAX_LOG_SENSE         1024      /* maximum data xfer for log sense */
#define LOG_HEADER_SIZE         4

#define MAGSTAR                  1
#define DLT                      2
#define BUFFER_SIZE            512
#define LTO_MICROCODESIZE  1048576
#define MORGAN_MICROCODESIZE 63488
#define MAX_SERIAL_NUMBER_LENGTH 20
#define MALLOC(x)   Malloc(x)
#define FREE(x)     \
{                   \
    if(x) free(x);  \
    x = NULL;       \
}

#ifndef INT_MAX
#define INT_MAX (0x7FFFFFFF)
#endif

#define MIN(x, y) (x < y ? x : y)
#define MAX(x, y) (x > y ? x : y)
#define STRLNCMP(x,y) strncmp(x,y,strlen(y)) /* BLD 5/12/04 */

#ifdef _WIN64

#define SETBUF(stream, buf) setbuf(stream, buf)
#define STRCPY(dst, src) strcpy(dst, src)
#define GETS(buf) gets(buf)
#define STRCAT(dst, src) strcat(dst, src)
#define STRNCPY(dst, src, count) strncpy(dst, src, count)
#define SCANF(x) scanf x
#define sprintf_s sprintf
#define MEMCPY(dest, src, length) memcpy(dest, src, length)

#else

#define SETBUF(stream, buf) setvbuf(stream, buf, _IOFBF, (buf ? sizeof(buf) : INT_MAX))
#define STRCPY(dst, src) strcpy_s(dst, INT_MAX, src)
#define GETS(buf) gets_s(buf, INT_MAX)
#define STRCAT(dst, src) strcat_s(dst, INT_MAX, src)
#define STRNCPY(dst, src, count) strncpy_s(dst, INT_MAX, src, count)
#define SCANF(x) scanf_s x
#define MEMCPY(dest, src, length) memcpy_s(dest, length, src, length)

#endif

#define ONE_K 1024

/* BLD, 09/22/04.  Need to use strcat to ensure that a null
   terminator is placed at the end of the string after each iteration */
#define PAD_OR_TRUNCATE(sz, c, iSize)   \
{                                       \
    char szTmp[2] = {0};                \
    szTmp[0] = c;                       \
    while(strlen(sz) < iSize)           \
        STRCAT(sz, szTmp);              \
                                        \
    sz[iSize] = '\0';                   \
}

#define BZERO(sz, iSize) memset(sz, '\0', iSize) /* BLD, 06/08/04.  Set up to mimick UNIX bzero function */
#define ARGV_STRING_LENGTH 2  /* BLD, 06/09/04 */

/*
** this is the largest element address in the box
** it allows an array of that size even though some members are not used
**          Transport       Mags            I/E             Drive
** 3575     0x1F41-0x1F42   0x0020-0x0163   0x001E-0x001F   0x0010-0x0015
** 3570     0x0000          0x0020-0x0032   0x001F          0x0010-0x0011
** 3590     0x0000          0x0020-0x0029   0x001F          0x0010
**
** for 3570 its  50  0x32
** for 3575 its 357  0x163 + 2
** for 3590 its  41  0x29
*/
#define MAX_ELEM  7200

/* tracing macros conditional on -t flag */
#define TTRACE(fp,x) if(gp->trace_flag==1) {fprintf(fp,x);fflush(fp);}
#define TTRACE1(fp,x,s) if(gp->trace_flag==1) {fprintf(fp,x,s);fflush(fp);}

#define TF(x)    (x==0 ? "False" : "True" )
#define OF(x)    (x==0 ? "OFF" : "ON" )
#define NOTTF(x) ((x<0) || (x>1))
#define FULL(x)  (x==0 ? "EMPTY" : "FULL " )

/* used to set the buffer address in system calls */
#define BUFADR(x) buffer_address_set(x)

/* used to print device info */

#define iflow(fp,x)  if(dip->FeaturesLow &x)   {fprintf(fp,"    ");fprintf(fp,#x);fprintf(fp,"\n");}
#define ifhigh(fp,x) if(dip->FeaturesHigh &x)  {fprintf(fp,"    ");fprintf(fp,#x);fprintf(fp,"\n");}


/********************************************************************
**
** command definitions
**
********************************************************************/

/*
** tape operations
*/

#define STOFFL      5      /* rewind and unload tape */
#define STREW       6      /* rewind */
#define STERASE     7      /* erase tape, leave at load point */
#define STRETEN     8      /* retension tape, leave at load point */
#define STWEOF     10      /* write an end-of-file record */
#define STFSF      11      /* forward space file */
#define STRSF      12      /* reverse space file */
#define STFSR      13      /* forward space record */
#define STRSR      14      /* reverse space record */
#define ST_LOAD    21      /* load volume */
#define ST_FLUSH   22      /* flush buffer to media */
#define ST_EOD     23      /* position to end of data */
#define ST_WSS     24      /* write setmarks */
#define ST_BSS     25      /* reverse space setmarks */
#define ST_FSS     26      /* forward space setmarks */
#define STTUR      27      /* test unit ready */

/********************************************************************
**
** IOCTL command definitions
**
********************************************************************/

/*
** Define the method codes for how buffers are passed for I/O and FS controls
*/

#define METHOD_BUFFERED                 0
#define METHOD_IN_DIRECT                1
#define METHOD_OUT_DIRECT               2
#define METHOD_NEITHER                  3

//
// Macro definition for defining IOCTL and FSCTL function control codes.  Note
// that function codes 0-2047 are reserved for Microsoft Corporation, and
// 2048-4095 are reserved for customers.
//

/*
macro defined in ntddk.h and devioctl.h
#define CTL_CODE(DeviceType, Function, Method, Access) \
                (((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method))
*/


#define IOCTL_TAPE_BASE        FILE_DEVICE_TAPE

#define IOCTL_TAPE_OBTAIN_SENSE           CTL_CODE(IOCTL_TAPE_BASE, 0x0819, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_TAPE_OBTAIN_VERSION         CTL_CODE(IOCTL_TAPE_BASE, 0x081a, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_TAPE_LOG_SELECT             CTL_CODE(IOCTL_TAPE_BASE, 0x081c, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)
#define IOCTL_TAPE_LOG_SENSE              CTL_CODE(IOCTL_TAPE_BASE, 0x081d, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_TAPE_REPORT_MEDIA_DENSITY   CTL_CODE(IOCTL_TAPE_BASE, 0x081e, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_TAPE_STANDARD_INQUIRY       CTL_CODE(IOCTL_TAPE_BASE, 0x081f, METHOD_BUFFERED, FILE_READ_ACCESS )

#define IOCTL_TAPE_SCALE_CAPACITY         CTL_CODE(IOCTL_TAPE_BASE, 0x000b, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)
#define IOCTL_TAPE_GET_ENCRYPTION_STATE   CTL_CODE(IOCTL_TAPE_BASE, 0x0820, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_TAPE_SET_ENCRYPTION_STATE   CTL_CODE(IOCTL_TAPE_BASE, 0x0821, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)
#define IOCTL_TAPE_SET_DATA_KEY           CTL_CODE(IOCTL_TAPE_BASE, 0x0822, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)
/*
** dd_trace_flag values
*/

#define DD_TRACE_CAPACITY  100000
#define DD_TRACE_DUMP_SIZE 10000
#define DD_TRACE_PRINTF    0x00000001 /* standard printf              */
#define DD_TRACE_PS        0x00000002 /* TRCGEN AIX,cmn_err SUN,N/A HP*/
#define DD_TRACE_LOCK      0x00000004 /* enable use of locks in macros*/
#define DD_TRACE_CC        0x00000008 /* enable common code trace     */
#define DD_TRACE_ENTRY     0x00000010 /* enable all ENTRY macros      */
#define DD_TRACE_EXIT      0x00000020 /* enable all EXIT  macros      */
#define DD_TRACE_DEBUG     0x00000040 /* enable all DEBUG macros      */
#define DD_TRACE_DV        0x00000080 /* enable developer trace       */
#define DD_TRACE_PAUSE     0x00000100 /* suspend (and resume) tracing */
#define DD_TRACE_COMPRESS  0x00000200 /* enable compress trace data   */
#define DD_TRACE_TIMESTAMP 0x00000400 /* enable timestamp in data     */
#define DD_TRACE_DUMP      0x00001000 /* indicate internal dmp buf use*/

//
// Command Descriptor Block constants for SCSI PASSTHRU
//

#define CDB6GENERIC_LENGTH                   6
#define CDB10GENERIC_LENGTH                  10

#define SCSIOP_TEST_UNIT_READY     0x00
#define SCSIOP_REZERO_UNIT         0x01
#define SCSIOP_REWIND              0x01
#define SCSIOP_REQUEST_BLOCK_ADDR  0x02
#define SCSIOP_REQUEST_SENSE       0x03
#define SCSIOP_FORMAT_UNIT         0x04
#define SCSIOP_READ_BLOCK_LIMITS   0x05
#define SCSIOP_REASSIGN_BLOCKS     0x07
#define SCSIOP_READ6               0x08
#define SCSIOP_RECEIVE             0x08
#define SCSIOP_WRITE6              0x0A
#define SCSIOP_PRINT               0x0A
#define SCSIOP_SEND                0x0A
#define SCSIOP_SEEK6               0x0B
#define SCSIOP_TRACK_SELECT        0x0B
#define SCSIOP_SLEW_PRINT          0x0B
#define SCSIOP_SEEK_BLOCK          0x0C
#define SCSIOP_PARTITION           0x0D
#define SCSIOP_READ_REVERSE        0x0F
#define SCSIOP_WRITE_FILEMARKS     0x10
#define SCSIOP_FLUSH_BUFFER        0x10
#define SCSIOP_SPACE               0x11
#define SCSIOP_INQUIRY             0x12
#define SCSIOP_VERIFY6             0x13
#define SCSIOP_RECOVER_BUF_DATA    0x14
#define SCSIOP_MODE_SELECT         0x15
#define SCSIOP_RESERVE_UNIT        0x16
#define SCSIOP_RELEASE_UNIT        0x17
#define SCSIOP_COPY                0x18
#define SCSIOP_ERASE               0x19
#define SCSIOP_MODE_SENSE          0x1A
#define SCSIOP_START_STOP_UNIT     0x1B
#define SCSIOP_STOP_PRINT          0x1B
#define SCSIOP_LOAD_UNLOAD         0x1B
#define SCSIOP_RECEIVE_DIAGNOSTIC  0x1C
#define SCSIOP_SEND_DIAGNOSTIC     0x1D
#define SCSIOP_MEDIUM_REMOVAL      0x1E
#define SCSIOP_READ_CAPACITY       0x25
#define SCSIOP_READ                0x28
#define SCSIOP_WRITE               0x2A
#define SCSIOP_SEEK                0x2B
#define SCSIOP_LOCATE              0x2B
#define SCSIOP_WRITE_VERIFY        0x2E
#define SCSIOP_VERIFY              0x2F
#define SCSIOP_SEARCH_DATA_HIGH    0x30
#define SCSIOP_SEARCH_DATA_EQUAL   0x31
#define SCSIOP_SEARCH_DATA_LOW     0x32
#define SCSIOP_SET_LIMITS          0x33
#define SCSIOP_READ_POSITION       0x34
#define SCSIOP_SYNCHRONIZE_CACHE   0x35
#define SCSIOP_COMPARE             0x39
#define SCSIOP_COPY_COMPARE        0x3A
#define SCSIOP_WRITE_DATA_BUFF     0x3B
#define SCSIOP_READ_DATA_BUFF      0x3C
#define SCSIOP_CHANGE_DEFINITION   0x40
#define SCSIOP_READ_SUB_CHANNEL    0x42
#define SCSIOP_READ_TOC            0x43
#define SCSIOP_READ_HEADER         0x44
#define SCSIOP_PLAY_AUDIO          0x45
#define SCSIOP_PLAY_AUDIO_MSF      0x47
#define SCSIOP_PLAY_TRACK_INDEX    0x48
#define SCSIOP_PLAY_TRACK_RELATIVE 0x49
#define SCSIOP_PAUSE_RESUME        0x4B
#define SCSIOP_LOG_SELECT          0x4C
#define SCSIOP_LOG_SENSE           0x4D

/*
 * Common completion code values (for all device classes)
 * returned to caller
 */

typedef enum erp_compcode_e {
 DD_UNKNOWN_ERROR = -1,         /*  -1 */
 DD_NO_ERROR = 0,               /*   0 */
 DD_NO_SENSE = 200,             /* 200 */
 DD_DEVICE_DRIVER_FAILURE,      /* 201 */
 DD_EEPROM_FAILURE,             /* 202 */
 DD_MANUAL_INTERVENTION,        /* 203 */
 DD_RECOVERED_ERROR,            /* 204 */
 DD_SCSI_ADAPTER_ERROR,         /* 205 */
 DD_SCSI_ERROR,                 /* 206 */
 DD_ILLEGAL_REQUEST,            /* 207 */
 DD_COMMAND_ABORTED,            /* 208 */
 DD_HARDWARE_MICROCODE,         /* 209 */
 DD_UNIT_ATTENTION,             /* 210 */
 DD_SCSI_BUSY,                  /* 211 */

/*
 * Completion code values for media changer
 */

 DD_CARTRIDGE_ENTRY_EXIT = 300, /* 300 */
 DD_CARTRIDGE_LOAD_FAILURE,     /* 301 */
 DD_CARTRIDGE_IN_FAILED_DRIVE,  /* 302 */
 DD_CAROUSEL_NOT_LOADED,        /* 303 */
 DD_CHANGER_FAILURE,            /* 304 */
 DD_DRIVE_FAILURE,              /* 305 */
 DD_DRIVE_OR_MEDIA_FAILURE,     /* 306 */
 DD_ENTRY_EXIT_FAILURE,         /* 307 */
 DD_ENTRY_EXIT_NOT_PRESENT,     /* 308 */
 DD_LIBRARY_AUDIT,              /* 309 */
 DD_LIBRARY_FULL,               /* 310 */
 DD_MEDIA_EXPORT,               /* 311 */
 DD_SLOT_FAILURE,               /* 312 */
 DD_SLOT_OR_MEDIA_FAILURE,      /* 313 */
 DD_SOURCE_EMPTY,               /* 314 */
 DD_DESTINATION_FULL,           /* 315 */
 DD_CLEANER_INST,               /* 316 */
 DD_MEDIA_NOT_EJECTED,          /* 317 */
 DD_IOPORT_NOT_CONFIG,          /* 318 */
 DD_FIRST_DEST_EMPTY,           /* 319 */


/*
 * Completion code values for tape drives
 */

 DD_END_PHYSICAL_MEDIA = 400,   /* 400 */
 DD_MEDIA_BLANK,                /* 401 */
 DD_MEDIA_CORRUPTED,            /* 402 */
 DD_MEDIA_FAILURE,              /* 403 */
 DD_MEDIA_INCOMPATIBILITY,      /* 404 */
 DD_SECTOR_RELOCATION,          /* 405 */
 DD_SECTOR_OUT_OF_RANGE,        /* 406 */
 DD_WRITE_PROTECT,              /* 407 */
 DD_CLEAN_MEDIA,                /* 408 */
 DD_MEDIA_FAULT,                /* 409 */
 DD_CLEANING_COMPLETE,          /* 410 */
 DD_LOGICAL_END_OF_MEDIA,       /* 411 */
 DD_MEDIA_NOT_PRESENT,          /* 412 */
 DD_BEGINNING_OF_MEDIA,         /* 413 */
 DD_ERASE_FAILURE,              /* 414 */
 DD_WRITE_TO_WRITTEN_WORM,      /* 415 */
 DD_WRONG_LENGTH_BLOCK,         /* 416 */
 DD_OPEN_READ_ONLY,             /* 417 */
 DD_OPEN_WRITE_ONLY             /* 418 */

} erp_compcode_t;

/*
** open mode settings on this platform
*/

#define PS_OMODE_INVALID       -1
#define PS_OMODE_RD            GENERIC_READ
#define PS_OMODE_RW            GENERIC_READ | GENERIC_WRITE

/*
** STRUCTUREs ETC
*/

/* generic syntax error code */
#define SYNTAX_ERROR gp->return_error_when_fail=1; \
                     gp->exit_on_unexpected_result=1; \
                     printf("batch file syntax error, line %d \n",gp->file_line); \
                     gp->final_rc=ERRSYNTAX

/*
** LOG SENSE
*/

typedef struct _LOG_PAGE_HEADER
  {
    char page_code;
    char res;
    char page_len[2];
  }LOG_PAGE_HEADER, *PLOG_PAGE_HEADER;

typedef struct _LOG_PARAMETER
  {
    char parm_code[2];
    char control;
    char parm_len;
    char parm_value[1];
  }LOG_PARAMETER, *PLOG_PARAMETER;

typedef struct _LOG_SENSE
  {
    LOG_PAGE_HEADER header;
    LOG_PARAMETER   parameter[MAX_LOG_SENSE];
  }LOG_SENSE, *PLOG_SENSE;

//
// IOCTL_TAPE_OBTAIN_MTDEVICE definitions
//

typedef ULONG TAPE_OBTAIN_MTDEVICE, *PTAPE_OBTAIN_MTDEVICE;

//
// IOCTL_TAPE_OBTAIN_SENSE definitions - specific to our driver
//

#define SENSE_BUFFER_SIZE 96    // from scsi.h

typedef struct _TAPE_OBTAIN_SENSE
{
  LONG SenseDataLength;
  CHAR SenseData[SENSE_BUFFER_SIZE];
} TAPE_OBTAIN_SENSE, *PTAPE_OBTAIN_SENSE;

//
// IOCTL_TAPE_OBTAIN_VERSION definitions - specific to our driver
//
#define MAX_DRIVER_VERSIONID_LENGTH 12 // from magcomm.h

typedef struct _TAPE_OBTAIN_VERSION
{
  CHAR VersionId[MAX_DRIVER_VERSIONID_LENGTH];
} TAPE_OBTAIN_VERSION, *PTAPE_OBTAIN_VERSION;

/* see ntddscsi.h and spti.c for passthrough structures etc */

#define SCSI_PASSTHRU_TIMEOUT_VALUE    (850)

typedef struct _SCSI_PASS_THROUGH_WITH_BUFFERS
    {
      SCSI_PASS_THROUGH spt;
      ULONG             Filler;                  // realign buffers to double word boundary
      UCHAR             ucSenseBuf[SENSE_BUFFER_SIZE];          // sense length 3570/90
      UCHAR             ucDataBuf[MAX_TRANSFER_SPT]; // max data xfer length on NT with Adaptec 2944 is 65536 bytes
    } SCSI_PASS_THROUGH_WITH_BUFFERS, *PSCSI_PASS_THROUGH_WITH_BUFFERS;

typedef struct _SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER
    {
      SCSI_PASS_THROUGH_DIRECT sptd;
      ULONG                    Filler;             // realign buffers to double word boundary
      UCHAR                    ucSenseBuf[SENSE_BUFFER_SIZE];     // sense length 3570/90
      // data buffer accessed thru pointer in structure
    } SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER, *PSCSI_PASS_THROUGH_DIRECT_WITH_BUFFER;

typedef struct _TAPE_LOG_SENSE_PARAMETERS{
    UCHAR       PageCode;
    UCHAR PC;
    UCHAR       PageLength[2];
    UCHAR       LogData[MAX_LOG_SENSE];
} TAPE_LOG_SENSE_PARAMETERS, *PTAPE_LOG_SENSE_PARAMETERS;

typedef struct { /* mapping for standard 36 bytes of INQUIRY data */
        UCHAR device_type;                /* bits 7-5, peripheral qualifier, bits 4-0, periferal device type. */
        UCHAR removable_media_device_modifier;   /* rmb - bit 7 = 1 means media is removable */
        UCHAR iso_ecma_ansi;                      /* bits 7 and 6 - ISO Version */
                                          /* bits 5 through 3 - ECMA Version */
                                          /* bits 2 through 0 - ANSI version */
        UCHAR rdf;
        UCHAR additional_length;          /* Additional bytes of INQUIRY data. 0x33 for 3570/90 */
        UCHAR reserved;
        UCHAR scsi3;                      /* SCSI 3 changes */
        UCHAR flags;
        UCHAR vendor_id[8];               /* ASCII representation of vendor identification  */
        UCHAR product_id[16];             /* ASCII representation of the product name */
        UCHAR product_revision_level[4];  /* ASCII representation of the revision level.  */
        UCHAR mfg_plant[2];
        UCHAR seq_number[12];
        UCHAR filler[2];
        UCHAR equipment_flags;
        UCHAR scsi_custom_byte;
        UCHAR reserved2;
        UCHAR msg_display_type;
} Inquiry_Data;

#define SN_PAGE_CODE 0x80
#define SUPPORTED_PAGES_PAGE_CODE   0x00
#define DEV_CONFIG_PAGE_CODE        0x83

/* p. 0x83 id types */
#define DEV_CONFIG_ID_STANDARD      0x01
#define DEV_CONFIG_ID_WWNN          0x03
#define DEV_CONFIG_ID_WWPN          0x13
#define DEV_CONFIG_ID_PID           0x14
#define DEV_CONFIG_ID_ALT_WWPN      0x23
#define DEV_CONFIG_ID_ALT_PID       0x24
#define EVPD_ON                     1
#define EVPD_OFF                    0
typedef struct
{
    UCHAR device_type;
    UCHAR page_code;
    UCHAR reserved;
    UCHAR page_length;
    UCHAR unit_serial_number[MAX_SERIAL_NUMBER_LENGTH+1];
} UNIT_SERIAL_NUMBER_PAGE;

typedef struct
{
    UCHAR device_type;
    UCHAR page_code;
    UCHAR reserved;
    UCHAR page_length;
    UCHAR szData[1];
} GENERIC_INQUIRY_STRUCT;

typedef struct
{
    UCHAR CodeSet;
    UCHAR IdType;
    UCHAR Reserved;
    UCHAR IdLength;
    UCHAR szData[1];
} DEV_CONFIG_ID;

/* struct created 4/9/04 by BLD.  Not sure how extensible the Inquiry_Data
   struct is, but this one should be extensible if future development is needed */
typedef struct
{
    Inquiry_Data scInqData;
    UCHAR szSerialNumber[MAX_SERIAL_NUMBER_LENGTH];
} Inquiry_Data_Plus;
/*
****************  IOCTL_TAPE_REPORT_MEDIA_DENSITY definitions *********
*/

typedef struct _TAPE_REPORT_DENSITY{
  ULONG   PrimaryDensityCode;
  ULONG   SecondaryDensityCode;
  BOOLEAN WriteOk;
  ULONG   BitsPerMM;
  ULONG   MediaWidth;
  ULONG   Tracks;
  ULONG   Capacity;
} TAPE_REPORT_DENSITY, *PTAPE_REPORT_DENSITY;

/*
** general data/variable structure for this utility
*/

struct global_data
{
  char    library_mode_on;                        /* base or library mode */
  char    library_mode_only;                      /* library only mode */
  char    s[256];                                 /* work string area */
  char    global_string[256];                     /* used by vartos */
  char    *cp;
  double  total_bytes;                            /* total bytes transfered */
  int     rv;
  int     return_error_when_fail;                 /* if 1(true), set the rc on exit */
  int     exit_on_unexpected_result;              /* stop test on unexpected result? */
  int     skip_count;                             /* count of lines to skip in batch file */

  char    batch_input[200];                       /* default batch input file - default ntutil.in */
  char    batch_output[200];                      /* default batch output file - defalut ntutil.out */
  char    batch_special[200];                     /* special file to be opened if -d on command line */
  char    batch_special_c[200];                   /* special file to be opened if -c on command line */
  char    pass_special[200];                      /* special file to pass to other test tool */
  int     batch_input_flag;                       /* 0 means no -f option specified */
  int     batch_output_flag;                      /* 0 means no -o option specified */
  int     batch_special_flag;                     /* 0 means no -d option specified */
  int     batch_special_flag_c;                   /* 0 means no -c option specified */
  int     trace_flag;                             /* 0 means no tracing, 1 means -t specified */
  int     file_line;                              /* line number in the input file */
  int     final_rc;                               /* return code when lbtest completes */
  int     fail_line;                              /* last line of test that failed */
  time_t  start_time;                             /* time batch test started */
  UCHAR   block[MAX_TRANSFER];                    /* data block */
  UCHAR   block1[MAX_TRANSFER];                   /* data block */
  int     countflag;                              /* 0 means no element count exists yet */
  int     invflag;                                /* 0 means no inventory exists, 1 means we have it */

  int     dosioctl;                               /* used in call to DosDevIOCtl */

  HANDLE  ddHandle0;                              /* LUN 0 handle */
  HANDLE  ddHandle1;                              /* LUN 1 handle */
  HANDLE  fd0;                                    /* special file descriptor LUN0, filled in at open */
  HANDLE  fd1;                                    /* special file descriptor LUN1, filled in at open */
  char    device;                                 /* MAGSTAR or DLT - manual mode only */
  int     id;                                     /* IBM_3590, ...70, ...LTO */
  int     iTapeID;                                /* Fix tape capacity bug w/o rearchitecting code, BLD, 06/21/06*/
  int     generation0;                            /* To distinguish between generations of Ultrium drives BLD, 10/03/02*/
  int     generation1;                            /* To distinguise between generations of Ultrium changers, BLD 07/13/04 */
  char*   supported_inq_pages_tape;               /* holds the data from inq. p. 0x0 (EVPD = 1), BLD 04/19/04 */
  char*   supported_inq_pages_cgr;                /* holds the data from inq. p. 0x0 (EVPD = 1), BLD 04/19/04 */
  Inquiry_Data_Plus inquiry_data;                 /* An extensible standard inquiry data structure, BLD 04/09/04 */

};

typedef long daddr_t;

/* allowable statement types */
enum instruction_type {comment, tcommand, set, symbols,
                       typeit, tpause, texit, ifstmt, skip,
                       passthru, system_call,delay};

/*
** global data
*/

struct global_data               mt_global;              /* define global data used by subroutines */
struct global_data               *gp = &mt_global;       /* set global data pointer */


enum element_types {TPUNKNOWN ,TPDRIVE, TPSLOT, TPTRANS, TPEE };
int                              elem_map[MAX_ELEM];        /* element map, 0=unused 1=exists in box  */
int                              elem_type[MAX_ELEM];       /* element type */
int                              elem_state[MAX_ELEM];      /* element state, 0=empty, 1=full */
int                              elem_accessible[MAX_ELEM]; /* 1 if accessible, 0 if NOT accessible */
int                              elem_min;                  /* smallest element in use */
int                              elem_max;                  /* largest element in use */
int                              elem_min_slot;             /* smallest slot element */
int                              elem_max_slot;             /* largest  slot element */
int                              elem_num;                  /* number of elements in use */
int                              elem_num_slot;             /* number of slots in use */
int                              lastslot = -1;             /* slot we just moved from */
int                              lastdrive = -1;            /* drive we just moved to */

/* NOTE
** due to INTEL byte swapping - the relation between lblock and lblock_id are swapped
**
**  lblock[0] is block offset low      lblock[1] is block offset high -
**  lblock[0] == 0 for MAGSTAR         lblock[1] contains block id for MAGSTAR
**  lblock[0] bytes     0  1  2  3     lblock[1] bytes 0  1  2  3
**  lblock_id[x]bytes   3  2  1  0                     7  6  5  4
*/

DWORD          lblock[2];                         /* lblock[0] is offset low, lblock[1] is offset high */
UCHAR          *lblock_id = (UCHAR *)&lblock[0];  /* set by read_block_id, used by locate_block_id */
DWORD          tape_partition=1;
SCSI_ADDRESS   scsi_addr;                  /* see ntddscsi.h tape drive */
SCSI_ADDRESS   scsi_addr1;                 /* see ntddscsi.h changer */
int            global_write_count = 0;     /* used as default data - 1st 4 bytes on a write */
long           bytes_xfer;                 /* bytes read or writen */
long           block_size  = 1024;         /* either 0 (variable mode), or physical tape blocksize */
long           block_count = 1;            /* blocks transferred per I/O, (or bytes in variable mode) */

#if NT4 || W2K

IO_SCSI_CAPABILITIES                 capabilities; //note change for WNET below

#else

    #if WNET

//the following defines are from ntddstor.h
#define IOCTL_STORAGE_QUERY_PROPERTY   CTL_CODE(IOCTL_STORAGE_BASE, 0x0500, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef enum _STORAGE_PROPERTY_ID {
    StorageDeviceProperty = 0,
    StorageAdapterProperty
} STORAGE_PROPERTY_ID, *PSTORAGE_PROPERTY_ID;

typedef enum _STORAGE_QUERY_TYPE {
    PropertyStandardQuery = 0,          // Retrieves the descriptor
    PropertyExistsQuery,                // Used to test whether the descriptor is supported
    PropertyMaskQuery,                  // Used to retrieve a mask of writeable fields in the descriptor
    PropertyQueryMaxDefined     // use to validate the value
} STORAGE_QUERY_TYPE, *PSTORAGE_QUERY_TYPE;


typedef struct _STORAGE_PROPERTY_QUERY {

    //
    // ID of the property being retrieved
    //

    STORAGE_PROPERTY_ID PropertyId;

    //
    // Flags indicating the type of query being performed
    //

    STORAGE_QUERY_TYPE QueryType;

    //
    // Space for additional parameters if necessary
    //

    UCHAR AdditionalParameters[1];

} STORAGE_PROPERTY_QUERY, *PSTORAGE_PROPERTY_QUERY;

typedef struct _STORAGE_ADAPTER_DESCRIPTOR {

    ULONG Version;

    ULONG Size;

    ULONG MaximumTransferLength;

    ULONG MaximumPhysicalPages;

    ULONG AlignmentMask;

    BOOLEAN AdapterUsesPio;

    BOOLEAN AdapterScansDown;

    BOOLEAN CommandQueueing;

    BOOLEAN AcceleratedTransfer;

    BOOLEAN BusType;

    USHORT BusMajorVersion;

    USHORT BusMinorVersion;

} STORAGE_ADAPTER_DESCRIPTOR, *PSTORAGE_ADAPTER_DESCRIPTOR;

STORAGE_ADAPTER_DESCRIPTOR           capabilities = {0};
STORAGE_PROPERTY_QUERY               stAdptProp = {0};

    #else

        Behavior Undefined

    #endif

#endif

SCSI_PASS_THROUGH_WITH_BUFFERS       sptwb;
SCSI_PASS_THROUGH_DIRECT_WITH_BUFFER sptdwb;
UCHAR                                *pucDataBuf, *pucFreePtr;
TAPE_OBTAIN_VERSION                  code_version;
TAPE_OBTAIN_SENSE                    sense_data;
TAPE_OBTAIN_MTDEVICE                 ext_inquiry_data;
TAPE_LOG_SENSE_PARAMETERS            logsense;

/* LTO device IDs must have higher value than magstar devices IDs */
#define IBM_INVALID_DEVICE   0
#define IBM_3570             1
#define IBM_3570C            2
#define IBM_3575             3
#define IBM_3490             4
#define IBM_3590             5
#define IBM_3592             6
#define IBM_3580             7
#define IBM_3581             8
#define IBM_3582             9
#define IBM_3583             10
#define IBM_3584             11
#define IBM_3576             12
#define IBM_3573             13
#define IBM_3577             14

static char *SUPPORTED_VENDORS[] = { "IBM", "ADIC", NULL };

static char *SUPPORTED_TAPE_DRIVES[] = {  "03570",           /* 1  Magstar 3570  */
                                          "03570C",          /* 2  Magstar 3570C */
                                          "03575",           /* 3  Magstar 3575  */
                                          "03590",           /* 4  Magstar 3590  */
                                          "03490",           /* 5  Magstar VTS   */
                                          "LTO TAPE DRIVE",  /* 6  Old Ultrium Inquiry String */
                                          "FastStor LTO",    /* 7  Old Morgan  Inquiry String */
                                          "Scalar 100",      /* 8  Old Belgian Inquiry String */
                                          "ULT3580-TD1",     /* 9  LTO 3580 Ultrium  */
                                          "ULTRIUM-TD1",     /* 10 LTO 3580 Ultrium  */
                                          "ULT3580-TD2",     /* 11 LTO 3580 Ultrium 2 */
                                          "ULTRIUM-TD2",     /* 12 LTO 3580 Ultrium 2 */
                                          "ULT3580-TD3",     /* 13 LTO 3580 Ultrium 3 */
                                          "ULTRIUM-TD3",     /* 14 LTO 3580 Ultrium 3 */
                                          "ULT3580-HH3",     /* 15 LTO 3580 Ultrium 3 */
                                          "ULTRIUM-HH3",     /* 16 LTO 3580 Ultrium 3 */
                                          "HH LTO Gen 3",    /* 17 LTO 3580 Ultrium 3 */
                                          "ULT3580-TD4",     /* 18 LTO 3580 Ultrium 4 */
                                          "ULTRIUM-TD4",     /* 19 LTO 3580 Ultrium 4 */
                                          "ULT3580-HH4",     /* 20 LTO 3580 Ultrium 4 */
                                          "ULTRIUM-HH4",     /* 21 LTO 3580 Ultrium 4 */
                                          "HH LTO Gen 4",    /* 22 LTO 3580 Ultrium 4 */
                                          "ULT3581-TA2",     /* 23 LTO 3581 New York   */
                                          "36232LX",         /* 24 LTO 3581 New York (OEM) */
                                          "36232SX",         /* 25 LTO 3581 New York (OEM) */
                                          "33622LX",         /* 26 LTO 3573 Library */
                                          "33614LX",         /* 27 LTO 3573 Library */
                                          "ULT3581-TA",      /* 28 LTO 3581 Morgan   */
                                          "ULT3582-TL",      /* 29 LTO 3582  Pegasus */
                                          "ULT3583-TL",      /* 30 LTO 3583 Belgian  */
                                          "03584L22",        /* 31 LTO 3584 L22 (w/ 3592 support) */
                                          "03584L32",        /* 32 LTO 3584 Anaconda */
                                          "03584L52",        /* 33 LTO 3584 L52 (w/ 3592 support) */
                                          "03592J1A",        /* 34 TotalStorage 3592 */
                                          "03592E05",        /* 35 TotalStorage 3592 */
                                          "03592E06",        /* 36 TotalStorage 3592 */
                                          "3576-MTL",        /* 37 3576 Tape Library (Orion) */
                                          "3573-TL",         /* 38 3573 Tape Library */
                                          "3577-TL",         /* 39 3577 Tape Library */
                                          NULL };

static int SUPPORTED_TAPE_DRIVE_ID[] = {  IBM_3570,          /* 1  Magstar 3570  */
                                          IBM_3570C,         /* 2  Magstar 3570C */
                                          IBM_3575,          /* 3  Magstar 3575  */
                                          IBM_3590,          /* 4  Magstar 3590  */
                                          IBM_3490,          /* 5  Magstar VTS   */
                                          IBM_3580,          /* 6  Old Ultrium Inquiry String */
                                          IBM_3581,          /* 7  Old Morgan  Inquiry String */
                                          IBM_3583,          /* 8  Old Belgian Inquiry String */
                                          IBM_3580,          /* 9  LTO 3580 Ultrium  */
                                          IBM_3580,          /* 10 LTO 3580 Ultrium  */
                                          IBM_3580,          /* 11 LTO 3580 Ultrium 2 */
                                          IBM_3580,          /* 12 LTO 3580 Ultrium 2 */
                                          IBM_3580,          /* 13 LTO 3580 Ultrium 3 */
                                          IBM_3580,          /* 14 LTO 3580 Ultrium 3 */
                                          IBM_3580,          /* 15 LTO 3580 Ultrium 3 */
                                          IBM_3580,          /* 16 LTO 3580 Ultrium 3 */
                                          IBM_3580,          /* 17 LTO 3580 Ultrium 4 */
                                          IBM_3580,          /* 18 LTO 3580 Ultrium 4 */
                                          IBM_3580,          /* 19 LTO 3580 Ultrium 4 */
                                          IBM_3580,          /* 20 LTO 3580 Ultrium 4 */
                                          IBM_3580,          /* 21 LTO 3580 Ultrium 4 */
                                          IBM_3580,          /* 22 LTO 3580 Ultrium 4 */
                                          IBM_3581,          /* 23 LTO 3581 New York */
                                          IBM_3581,          /* 24 LTO 3581 New York (OEM) */
                                          IBM_3581,          /* 25 LTO 3581 New York (OEM) */
                                          IBM_3573,          /* 26 LTO 3573 */
                                          IBM_3573,          /* 27 LTO 3573 */
                                          IBM_3581,          /* 28 LTO 3581 Morgan   */
                                          IBM_3582,          /* 29 LTO 3582 Pegasus */
                                          IBM_3583,          /* 30 LTO 3583 Belgian  */
                                          IBM_3584,          /* 31 LTO 3584 Anaconda */
                                          IBM_3584,          /* 32 LTO 3584 Anaconda */
                                          IBM_3584,          /* 33 LTO 3584 Anaconda */
                                          IBM_3592,          /* 34 TotalStorage 3592 */
                                          IBM_3592,          /* 35 TotalStorage 3592 */
                                          IBM_3592,          /* 36 TotalStorage 3592 */
                                          IBM_3576,          /* 37 3576 Tape Library */
                                          IBM_3573,          /* 38 3573 Tape Library */
                                          IBM_3577,          /* 39 3573 Tape Library */
                                        };


char dlt_id[] = {'Q','U','A','N', 'T', 'U', 'M', ' ', 'D', 'L', 'T'};
char dlt_device_id[] = {'D','L','T'};
char message_display[] = {0x00,0x10,0x00,0x00,0x00,0x00,0x00,0x10,
                          'T', 'E', 'S', 'T', 'M', 'S', 'G', '*',
                          '*', 'T', '*', 'E', '*', 'S', '*', 'T'};

struct vartab
{
  /* A string and the correspondinginteger value. */
  char *varstr; /* variable name  */
  int  varvalue;/* variable value */
} var_table[] =
  {
    "SUCCESS",ERROR_SUCCESS,                                       //    0
    "ERROR_INVALID_FUNCTION",ERROR_INVALID_FUNCTION,               //    1
    "ERROR_FILE_NOT_FOUND",ERROR_FILE_NOT_FOUND,                   //    2
    "ERROR_PATH_NOT_FOUND",ERROR_PATH_NOT_FOUND,                   //    3
    "ERROR_TOO_MANY_OPEN_FILES",ERROR_TOO_MANY_OPEN_FILES,         //    4
    "ERROR_ACCESS_DENIED",ERROR_ACCESS_DENIED,                     //    5
    "ERROR_INVALID_HANDLE",ERROR_INVALID_HANDLE,                   //    6
    "ERROR_NOT_ENOUGH_MEMORY",ERROR_NOT_ENOUGH_MEMORY,             //    8
    "ERROR_BAD_FORMAT",ERROR_BAD_FORMAT,                           //    9
    "ERROR_INVALID_BLOCK",ERROR_INVALID_BLOCK,                     //   10
    "ERROR_BAD_ENVIRONMENT",ERROR_BAD_ENVIRONMENT,                 //   11
    "ERROR_INVALID_ACCESS",ERROR_INVALID_ACCESS,                   //   12
    "ERROR_INVALID_DATA",ERROR_INVALID_DATA,                       //   13
    "ERROR_OUTOFMEMORY",ERROR_OUTOFMEMORY,                         //   14
    "ERROR_INVALID_DRIVE",ERROR_INVALID_DRIVE,                     //   15
    "ERROR_WRITE_PROTECT",ERROR_WRITE_PROTECT,                     //   19
    "ERROR_BAD_UNIT",ERROR_BAD_UNIT,                               //   20
    "ERROR_NOT_READY",ERROR_NOT_READY,                             //   21
    "ERROR_BAD_COMMAND",ERROR_BAD_COMMAND,                         //   22
    "ERROR_CRC",ERROR_CRC,                                         //   23
    "ERROR_HANDLE_EOF",ERROR_HANDLE_EOF,                           //   38
    "ERROR_NOT_SUPPORTED",ERROR_NOT_SUPPORTED,                     //   50
    "ERROR_DEV_NOT_EXIST",ERROR_DEV_NOT_EXIST,                     //   55
    "ERROR_ALREADY_ASSIGNED",ERROR_ALREADY_ASSIGNED,               //   85
    "ERROR_INVALID_PARAMETER",ERROR_INVALID_PARAMETER,             //   87
    "ERROR_OPEN_FAILED",ERROR_OPEN_FAILED,                         //  110
    "ERROR_INSUFFICIENT_BUFFER",ERROR_INSUFFICIENT_BUFFER,         //  122
    "ERROR_INVALID_NAME",ERROR_INVALID_NAME,                       //  123
    "ERROR_BUSY_DRIVE",ERROR_BUSY_DRIVE,                           //  142
    "DD_NO_SENSE",DD_NO_SENSE,                                     //  200
    "DD_DEVICE_DRIVER_FAILURE",DD_DEVICE_DRIVER_FAILURE,           //  201
    "DD_EEPROM_FAILURE",DD_EEPROM_FAILURE,                         //  202
    "DD_MANUAL_INTERVENTION",DD_MANUAL_INTERVENTION,               //  203
    "DD_RECOVERED_ERROR",DD_RECOVERED_ERROR,                       //  204
    "DD_SCSI_ADAPTER_ERROR",DD_SCSI_ADAPTER_ERROR,                 //  205
    "DD_SCSI_ERROR",DD_SCSI_ERROR,                                 //  206
    "DD_SCSI_BUSY",DD_SCSI_BUSY,                                   //  211
    "DD_ILLEGAL_REQUEST",DD_ILLEGAL_REQUEST,                       //  207
    "DD_COMMAND_ABORTED",DD_COMMAND_ABORTED,                       //  208
    "DD_HARDWARE_MICROCODE",DD_HARDWARE_MICROCODE,                 //  209
    "DD_UNIT_ATTENTION",DD_UNIT_ATTENTION,                         //  210
    "ERROR_MORE_DATA",ERROR_MORE_DATA,                             //  234
    "DD_CARTRIDGE_ENTRY_FAILURE",DD_CARTRIDGE_ENTRY_EXIT,          //  300
    "DD_CARTRIDGE_LOAD_FAILURE",DD_CARTRIDGE_LOAD_FAILURE,         //  301
    "DD_CARTRIDGE_IN_FAILED_DRIVE",DD_CARTRIDGE_IN_FAILED_DRIVE,   //  302
    "DD_CAROUSEL_NOT_LOADED",DD_CAROUSEL_NOT_LOADED,               //  303
    "DD_CHANGER_FAILURE",DD_CHANGER_FAILURE,                       //  304
    "DD_DRIVE_FAILURE",DD_DRIVE_FAILURE,                           //  305
    "DD_DRIVE_OR_MEDIA_FAILURE",DD_DRIVE_OR_MEDIA_FAILURE,         //  306
    "DD_ENTRY_EXIT_FAILURE",DD_ENTRY_EXIT_FAILURE,                 //  307
    "DD_ENTRY_EXIT_NOT_PRESENT",DD_ENTRY_EXIT_NOT_PRESENT,         //  308
    "DD_LIBRARY_AUDIT",DD_LIBRARY_AUDIT,                           //  309
    "DD_LIBRARY_FULL",DD_LIBRARY_FULL,                             //  310
    "DD_MEDIA_EXPORT",DD_MEDIA_EXPORT,                             //  311
    "DD_SLOT_FAILURE",DD_SLOT_FAILURE,                             //  312
    "DD_SLOT_OR_MEDIA_FAILURE",DD_SLOT_OR_MEDIA_FAILURE,           //  313
    "DD_SOURCE_EMPTY",DD_SOURCE_EMPTY,                             //  314
    "DD_DESTINATION_FULL",DD_DESTINATION_FULL,                     //  315
    "DD_CLEANER_INST",DD_CLEANER_INST,                             //  316
    "DD_MEDIA_NOT_EJECTED",DD_MEDIA_NOT_EJECTED,                   //  317
    "DD_IOPORT_NOT_CONFIG",DD_IOPORT_NOT_CONFIG,                   //  318
    "DD_FIRST_DEST_EMPTY",DD_FIRST_DEST_EMPTY,                     //  319
    "DD_END_PHYSICAL_MEDIA",DD_END_PHYSICAL_MEDIA,                 //  400
    "DD_MEDIA_BLANK",DD_MEDIA_BLANK,                               //  401
    "DD_MEDIA_CORRUPTED",DD_MEDIA_CORRUPTED,                       //  402
    "DD_MEDIA_FAILURE",DD_MEDIA_FAILURE,                           //  403
    "DD_MEDIA_INCOMPATIBILITY",DD_MEDIA_INCOMPATIBILITY,           //  404
    "DD_SECTOR_RELOCATION",DD_SECTOR_RELOCATION,                   //  405
    "DD_SECTOR_OUT_OF_RANGE",DD_SECTOR_OUT_OF_RANGE,               //  406
    "DD_WRITE_PROTECT",DD_WRITE_PROTECT,                           //  407
    "DD_CLEAN_MEDIA",DD_CLEAN_MEDIA,                               //  408
    "DD_MEDIA_FAULT",DD_MEDIA_FAULT,                               //  409
    "DD_CLEANING_COMPLETE",DD_CLEANING_COMPLETE,                   //  410
    "DD_LOGICAL_END_OF_MEDIA",DD_LOGICAL_END_OF_MEDIA,             //  411
    "DD_MEDIA_NOT_PRESENT",DD_MEDIA_NOT_PRESENT,                   //  412
    "DD_BEGINNING_OF_MEDIA",DD_BEGINNING_OF_MEDIA,                 //  413
    "DD_ERASE_FAILURE",DD_ERASE_FAILURE,                           //  414
    "DD_WRITE_TO_WRITTEN_WORM",DD_WRITE_TO_WRITTEN_WORM,           //  415
    "DD_WRONG_LENGTH_BLOCK",DD_WRONG_LENGTH_BLOCK,                 //  416
    "ERROR_IO_INCOMPLETE",ERROR_IO_INCOMPLETE,                     //  996
    "ERROR_IO_PENDING",ERROR_IO_PENDING,                           //  997
    "ERROR_NOACCESS",ERROR_NOACCESS,                               //  998
    "ERROR_CANTOPEN",ERROR_CANTOPEN,                               // 1011
    "ERROR_CANTREAD",ERROR_CANTREAD,                               // 1012
    "ERROR_CANTWRITE",ERROR_CANTWRITE,                             // 1013
    "ERROR_END_OF_MEDIA",ERROR_END_OF_MEDIA,                       // 1100
    "ERROR_FILEMARK_DETECTED",ERROR_FILEMARK_DETECTED,             // 1101
    "ERROR_BEGINNING_OF_MEDIA",ERROR_BEGINNING_OF_MEDIA,           // 1102
    "ERROR_SETMARK_DETECTED",ERROR_SETMARK_DETECTED,               // 1103
    "ERROR_NO_DATA_DETECTED",ERROR_NO_DATA_DETECTED,               // 1104
    "ERROR_PARTITION_FAILURE",ERROR_PARTITION_FAILURE,             // 1105
    "ERROR_INVALID_BLOCK_LENGTH",ERROR_INVALID_BLOCK_LENGTH,       // 1106
    "ERROR_DEVICE_NOT_PARTITIONED",ERROR_DEVICE_NOT_PARTITIONED,   // 1107
    "ERROR_UNABLE_TO_LOCK_MEDIA",ERROR_UNABLE_TO_LOCK_MEDIA,       // 1108
    "ERROR_UNABLE_TO_UNLOAD_MEDIA",ERROR_UNABLE_TO_UNLOAD_MEDIA,   // 1109
    "ERROR_MEDIA_CHANGED",ERROR_MEDIA_CHANGED,                     // 1110
    "ERROR_BUS_RESET",ERROR_BUS_RESET,                             // 1111
    "ERROR_NO_MEDIA_IN_DRIVE",ERROR_NO_MEDIA_IN_DRIVE,             // 1112
    "ERROR_IO_DEVICE",ERROR_IO_DEVICE,                             // 1117
    "ERROR_TOO_MANY_LINKS",ERROR_TOO_MANY_LINKS,                   // 1142
    "UNKNOWN",-1 /* end of table marker */         //   -1
  };

struct cmdtab
{
  /* A string and the correspondinginteger value. */
  char *varstr; /* variable name  */
  int  varvalue;/* variable value */
} cmd_table[] =
  {
    "IOCTL_TAPE_OBTAIN_VERSION",IOCTL_TAPE_OBTAIN_VERSION,
    "IOCTL_TAPE_OBTAIN_SENSE",IOCTL_TAPE_OBTAIN_SENSE,
    "IOCTL_SCSI_GET_CAPABILITIES",IOCTL_SCSI_GET_CAPABILITIES,
#if WNET
    "IOCTL_STORAGE_QUERY_PROPERTY",IOCTL_STORAGE_QUERY_PROPERTY,
#endif
    "IOCTL_SCSI_GET_ADDRESS",IOCTL_SCSI_GET_ADDRESS,
    "IOCTL_SCSI_PASS_THROUGH",IOCTL_SCSI_PASS_THROUGH,
    "IOCTL_SCSI_PASS_THROUGH_DIRECT",IOCTL_SCSI_PASS_THROUGH_DIRECT,
    "CMD_TAB VALUE NOT FOUND",-1 /* end of table marker */         //   -1
  };

struct key_entry
{
  unsigned char kv;
  char keystr[50];
} key_table[] =
{
  0x00,"no sense data",
  0x01,"recovered error",
  0x02,"not ready",
  0x03,"medium error",
  0x04,"hardware error",
  0x05,"illegal request",
  0x06,"unit attention!",
  0x07,"data protect",
  0x08,"blank check",
  0x09,"vendor specific",
  0x0a,"copy aborted",
  0x0b,"aborted command",
  0x0c,"equal compare",
  0x0d,"volume overflow",
  0x0e,"miscompare",
  0x0f,"reserved"
};

struct asc_entry
{
  unsigned char asc;
  unsigned char ascq;
  char ascstr[50];
} asc_table[] =
  {
    0x00,    0x00,"no additional sense           ",
    0x00,    0x01,"filemark detected             ",
    0x00,    0x02,"end-of-medium detected early warning ",
    0x00,    0x03,"setmark detected              ",
    0x00,    0x04,"beginning of medium           ",
    0x00,    0x05,"end of data                   ",
    0x00,    0x06,"I/O process terminated        ",
    0x00,    0x17,"Drive Needs Cleaning          ",
    0x02,    0x00,"no seek complete              ",
    0x03,    0x00,"device write fault            ",
    0x03,    0x01,"no write current              ",
    0x03,    0x02,"excessive write errors        ",
    0x04,    0x00,"logical unit not ready        ",
    0x04,    0x01,"becoming ready                ",
    0x04,    0x02,"not ready, init command rqd   ",
    0x04,    0x03,"not ready, manual int rqd     ",
    0x04,    0x04,"not ready, formating          ",
    0x04,    0x85,"Open Door                     ",
    0x04,    0x86,"Cartridge Magazine Missing    ",
    0x04,    0x8D,"In Manual (sequential) Mode   ",
    0x04,    0x8E,"In Auto (sequential) Mode     ",
    0x05,    0x00,"no response to select         ",
    0x06,    0x00,"no reference pos found        ",
    0x07,    0x00,"multiple devices selected     ",
    0x08,    0x00,"communication failure         ",
    0x08,    0x01,"communication timeout         ",
    0x08,    0x02,"communication parity error    ",
    0x09,    0x00,"track following error         ",
    0x0A,    0x00,"error log overflow            ",
    0x0C,    0x00,"write error                   ",
    0x11,    0x00,"unrecovered read error        ",
    0x11,    0x01,"read retries exhausted        ",
    0x11,    0x02,"error too long to correct     ",
    0x11,    0x03,"multiple read errors          ",
    0x11,    0x08,"incomplete block read         ",
    0x11,    0x09,"no gap found                  ",
    0x11,    0x0A,"miscorrected error            ",
    0x14,    0x00,"recorded entity not found     ",
    0x14,    0x01,"record not found              ",
    0x14,    0x02,"filemark/setmark not found    ",
    0x14,    0x03,"end-of-data not found         ",
    0x14,    0x04,"block sequence error          ",
    0x15,    0x00,"random positioning error      ",
    0x15,    0x01,"mechanical positioning error  ",
    0x15,    0x02,"read positioning error        ",
    0x17,    0x00,"no error correction applied   ",
    0x17,    0x01,"recovered with retries        ",
    0x17,    0x02,"recovered with pos head offset",
    0x17,    0x03,"recovered with neg head offset",
    0x18,    0x00,"ECC applied                   ",
    0x1A,    0x00,"parm list length error        ",
    0x1B,    0x00,"sync data xfer error          ",
    0x20,    0x00,"invalid op code               ",
    0x21,    0x00,"block out of range            ",
    0x21,    0x01,"invalid element address       ",
    0x24,    0x00,"invalid field in CDB          ",
    0x25,    0x00,"LUN not supported             ",
    0x26,    0x00,"invalid field in parm list    ",
    0x26,    0x01,"parameter not supported       ",
    0x26,    0x02,"parameter value invalid       ",
    0x26,    0x03,"threshold parms not supported ",
    0x27,    0x00,"write protected               ",
    0x28,    0x00,"not-ready to ready            ",
    0x28,    0x01,"imp/exp element accessed      ",
    0x28,    0x80,"Not Ready to Ready Medium Changer ",
    0x29,    0x00,"power-on, reset, bus reset!   ",
    0x2A,    0x00,"parameters changed            ",
    0x2A,    0x01,"mode parameters changed       ",
    0x2A,    0x02,"log parameters changed        ",
    0x2B,    0x00,"copy cannot execute           ",
    0x2C,    0x00,"command sequence error        ",
    0x2D,    0x00,"overwrite error on update     ",
    0x2F,    0x00,"command cleared by another init ",
    0x30,    0x00,"incompatible media            ",
    0x30,    0x01,"media unknown format          ",
    0x30,    0x02,"media incompatible format     ",
    0x30,    0x03,"cleaning cartridge installed  ",
    0x31,    0x00,"media format corrupted        ",
    0x31,    0x01,"format command failed         ",
    0x33,    0x00,"tape length error             ",
    0x37,    0x00,"rounded parameter             ",
    0x39,    0x00,"saving parms not supported    ",
    0x3A,    0x00,"medium not present            ",
    0x3B,    0x00,"sequential positioning error  ",
    0x3B,    0x01,"positioning error at BOT      ",
    0x3B,    0x02,"positioning error at EOT      ",
    0x3B,    0x08,"reposition error              ",
    0x3B,    0x0D,"Medium Destination Element Full ",
    0x3B,    0x0E,"Medium Source Element Empty   ",
    0x3B,    0x11,"Medium Magazine Not Accessible ",
    0x3B,    0x12,"Medium Magazine Removed        ",
    0x3B,    0x13,"Medium Magazine Inserted       ",
    0x3B,    0x14,"Medium Magazine Locked         ",
    0x3B,    0x15,"Medium Magazine Unlocked       ",
    0x3B,    0x80,"Medium Transport Element Full ",
    0x3D,    0x00,"invalid bits in message       ",
    0x3E,    0x00,"LUN not self-configured       ",
    0x3F,    0x00,"operating conditions changed  ",
    0x3F,    0x01,"microcode has been changed    ",
    0x3F,    0x02,"changed operating definition  ",
    0x3F,    0x03,"Inquiry data has changed      ",
    0x40,    0x00,"Diagnostic Failure            ",
    0x43,    0x00,"message error                 ",
    0x44,    0x00,"internal target failure       ",
    0x45,    0x00,"select/reselect failure       ",
    0x46,    0x00,"unsuccessful soft reset       ",
    0x47,    0x00,"SCSI parity error             ",
    0x48,    0x00,"init detected msg received    ",
    0x49,    0x00,"invalid message error         ",
    0x4A,    0x00,"command phase error           ",
    0x4B,    0x00,"data phase error              ",
    0x4C,    0x00,"LUN failed self-configuration ",
    0x4E,    0x00,"overlapped commands attempt   ",
    0x50,    0x00,"write append error            ",
    0x50,    0x01,"write append position error   ",
    0x50,    0x02,"position error (timing)       ",
    0x51,    0x00,"erase failure                 ",
    0x52,    0x00,"cartridge fault               ",
    0x53,    0x00,"media load or eject failed    ",
    0x53,    0x01,"unload tape failure           ",
    0x53,    0x02,"media removal prevented       ",
    0x5A,    0x00,"operator state changed        ",
    0x5A,    0x01,"operator media removal request ",
    0x5A,    0x02,"operator write protect        ",
    0x5A,    0x03,"operator write permit         ",
    0x5B,    0x00,"log exception                 ",
    0x5B,    0x01,"threshold condition met       ",
    0x5B,    0x02,"log counter at maximum        ",
    0x5B,    0x03,"log list codes exhausted      ",
    0x80,    0x00,"CU Mode, Vendor Unique        ",
    0x82,    0x83,"Cleaning Completed            ",
    0x85,    0x00,"Write Protect tape/drive fail ",
    0x85,    0x01,"Write Protect tape fail       ",
    0x85,    0x02,"Write Protect drive fail      ",
    0x85,    0x03,"Write Protect due to Current Tape Position ",
    0xFF,    0xFF,"End of Table "
  };

// Windows 2000 only

typedef  struct _GET_CHANGER_PARAMETERS_W2K {

    //
    // Size of the structure. Can be used for versioning.
    //

    ULONG Size;

    //
    // Number of N element(s) as defined by the Element Address Page (or equivalent...).
    //

    USHORT NumberTransportElements;
    USHORT NumberStorageElements;                // for data cartridges only
    USHORT NumberCleanerSlots;                   // for cleaner cartridges
    USHORT NumberIEElements;
    USHORT NumberDataTransferElements;

    //
    // Number of doors/front panels (allows user entry into the cabinet).
    //

    USHORT NumberOfDoors;

    //
    // The device-specific address (from user manual of the device) of the first N element. Used
    // by the UI to relate the various elements to the user.
    //

    USHORT FirstSlotNumber;
    USHORT FirstDriveNumber;
    USHORT FirstTransportNumber;
    USHORT FirstIEPortNumber;
    USHORT FirstCleanerSlotAddress;

    //
    // Indicates the capacity of each magazine, if they exist.
    //

    USHORT MagazineSize;

    //
    // Specifies the approximate number of seconds for when a cleaning should be completed.
    // Only applicable if drive cleaning is supported. See Features0.
    //

    ULONG DriveCleanTimeout;

    //
    // See features bits, above.
    //

    ULONG Features0;
    ULONG Features1;

    //
    // Bitmask defining Move from N element to element. Defined by Device Capabilities Page (or equivalent).
    // AND-masking with the TO_XXX values will indicate legal destinations.
    //

    UCHAR MoveFromTransport;
    UCHAR MoveFromSlot;
    UCHAR MoveFromIePort;
    UCHAR MoveFromDrive;

    //
    // Bitmask defining Exchange from N element to element. Defined by Device Capabilities Page (or equivalent).
    // AND-masking with the TO_XXX values will indicate legal destinations.
    //

    UCHAR ExchangeFromTransport;
    UCHAR ExchangeFromSlot;
    UCHAR ExchangeFromIePort;
    UCHAR ExchangeFromDrive;

    //
    // Bitmask defining which elements are capable of lock/unlock. Valid only if
    // CHANGER_LOCK_UNLOCK is set in Features0.
    //

    UCHAR LockUnlockCapabilities;

    //
    // Bitmask defining which elements valid for positioning operations. Valid only if
    // CHANGER_POSITION_TO_ELEMENT is set in Features0.
    //

    UCHAR PositionCapabilities;

    //
    // For future expansion.
    //

    UCHAR Reserved1[2];
    ULONG Reserved2[2];

} GET_CHANGER_PARAMETERS_W2K, * PGET_CHANGER_PARAMETERS_W2K;

//
// Definition for Features0 of GET_CHANGER_PARAMETERS
//
#define CHANGER_CLEANER_ACCESS_NOT_VALID    0x00040000 // The access bit in GES isn't valid for cleaner cartridges.


//
// Definitions for Features1 of GET_CHANGER_PARAMETERS
//

#define CHANGER_PREDISMOUNT_ALIGN_TO_SLOT   0x80000001 // The transport must be prepositioned to the slot prior to ejecting the media.
#define CHANGER_PREDISMOUNT_ALIGN_TO_DRIVE  0x80000002 // The transport must be prepositioned to the drive prior to ejecting the media.
//#define CHANGER_CLEANER_AUTODISMOUNT        0x80000004 // The device will move the cleaner cartridge back into the slot when cleaning has completed.
#define CHANGER_TRUE_EXCHANGE_CAPABLE       0x80000008 // Device can do src -> dest2 exchanges.
#define CHANGER_SLOTS_USE_TRAYS             0x80000010 // Slots have removable trays, requiring multiple moves for inject/eject.
#define CHANGER_RTN_MEDIA_TO_ORIGINAL_ADDR  0x80000020 // Media must be returned to the slot from which it originated after a move to another element.
#define CHANGER_CLEANER_OPS_NOT_SUPPORTED   0x80000040 // Automated cleaning operations are not supported on this device.
#define CHANGER_IEPORT_USER_CONTROL_OPEN    0x80000080 // Indicates that user action is necessary to open a closed ieport.
#define CHANGER_IEPORT_USER_CONTROL_CLOSE   0x80000100 // Indicates that user action is necessary to close an opened ieport.
#define CHANGER_MOVE_EXTENDS_IEPORT         0x80000200 // Indicates that a move media to the ieport extends the tray.
#define CHANGER_MOVE_RETRACTS_IEPORT        0x80000400 // Indicates that a move media from the ieport retracts the tray.

/*
** function prototype declarations
*/

int  parse(char *string,
           FILE *fp,
           int  *token_position,
           int  *token_length,
           int  *command_position,
           int  *analyze_position,
           int  *set_position,
           enum instruction_type *type_ptr);

int  execute_command(char   *input_string,
                     FILE   *out_fp,
                     int    *position,
                     int    *length,
                     int    first_cmd_token,
                     int    cmd_token_count,
                     int    *rc_ptr,
                     DWORD  *err_ptr,
                     int    *data);

int  analyze(char  *string,
             FILE  *fp,
             int   *position,
             int   *length,
             int    first_res,
             int    res_count,
             int    rc,
             DWORD  err_ptr,
             int    data_in);

int as_expected(int   left,
                char *relation ,
                int   right);

void batch_test_menu();

int  ctoi(FILE *fp,
          char *cp,
          int   count);

long ctol(char *cp,
          int   count);

char *decode_asc(UCHAR,
                 UCHAR);

char *decode_key(UCHAR);

void execute_delay(FILE *fp,
                   int   d_sec);

int  execute_set(char *string,
                 FILE *fp,
                 int  *position,
                 int  *length,
                 int  first_token,
                 int  token_count);

int  execute_tape_control(FILE    *fp,
                          int     st_op,
                          daddr_t st_count,
                          DWORD   *err_ptr);

void fprintf_device_info (FILE                      *fp,
                          TAPE_GET_DRIVE_PARAMETERS *dip);

void fprintf_devinfo     (FILE           *fp,
                          struct devinfo *dip);

void fprintf_medium_info (FILE                      *fp,
                          TAPE_GET_MEDIA_PARAMETERS *mip);

VOID fprintf_inquiry_data(FILE               *fp,
                          Inquiry_Data_Plus  *DataBuffer,
                          int                dcount);

char * itovar (int wanted);

void manual_test_menu(void);

char * ntdecodeerror(LONG);

LONG NTlseek(DWORDLONG);

void process(char *string,
             FILE *fp);

BOOL VerifyId(FILE *fp,
              char *temp_path,
              SCSI_ADDRESS *scsi_address);

BOOLEAN ReadNTRegistry(FILE *fp,
                       int  PortNumber,
                       int  PathId,
                       int  TargetId,
                       int  Lun,
                       char *DeviceName,
                       char *Identifier);

UCHAR stoh(char * );

void strfix(char  *string,
            int   length);

char * vartos (char *,
               int);


void dump_map(FILE *fp);

int add_map(FILE *fp,
            int  element);

int clear_map(FILE *fp,
              int  element);

VOID dump_bytes(FILE *fp,
                UCHAR *ucDataBuf,
                DWORD DataTransferLength,
                BOOL  truncate_print);

VOID PrintSenseInfo(FILE *fp,
                    UCHAR ScsiStatus,
                    UCHAR *ucSenseBuf,
                    UCHAR SenseInfoLength);

char *decode_asc(unsigned char asc,
                 unsigned char ascq);

void MallocAlignedDataBuffer(unsigned int size,
                             unsigned int align,
                             UCHAR        **pucBP,
                             UCHAR        **pucFP);
// Windows 2000 or .net only
#if W2K || WNET

void fprintf_w2k_inv(FILE *fp,
                     PCHANGER_ELEMENT_STATUS_EX ces,
                     ULONG ulNumberofElements);

void fprint_w2k_changer_parameters(FILE *fp,
                                   GET_CHANGER_PARAMETERS_W2K *gcp);

#endif

static int NTversion;

#define Vendor_Length 8
#define Product_Length 16
#define ProductRev_Length 4
#define SerialNumber_Length 16

typedef struct _TAPE_STANDARD_INQUIRY{
  UCHAR VendorID[Vendor_Length+1];                  // Vendor ID plus string terminator
  UCHAR ProductID[Product_Length+1];                // Product ID plus string terminator
  UCHAR ProductRevisionLevel[ProductRev_Length+1];  // Product Revision plus string terminator
  UCHAR SerialNumber[SerialNumber_Length+1];        // Serial Number plus string terminator
} TAPE_STANDARD_INQUIRY, *PTAPE_STANDARD_INQUIRY;

//
//      Extended Inquiry definitions
//

#define DD_DRIVER_LEVEL_LENGTH 10
#define DD_MAX_DEVICE_WWNAME 8
#define DD_MAX_DEVICE_SERIAL 36

typedef struct _INQUIRYDATA {
    UCHAR DeviceType : 5;
    UCHAR DeviceTypeQualifier : 3;
    UCHAR DeviceTypeModifier : 7;
    UCHAR RemovableMedia : 1;
    UCHAR Versions;
    UCHAR ResponseDataFormat;
    UCHAR AdditionalLength;
    UCHAR Reserved[2];
    UCHAR SoftReset : 1;
    UCHAR CommandQueue : 1;
    UCHAR Reserved2 : 1;
    UCHAR LinkedCommands : 1;
    UCHAR Synchronous : 1;
    UCHAR Wide16Bit : 1;
    UCHAR Wide32Bit : 1;
    UCHAR RelativeAddressing : 1;
    UCHAR VendorId[8];
    UCHAR ProductId[16];
    UCHAR ProductRevisionLevel[4];
    UCHAR VendorSpecific[20];
    UCHAR Reserved3[40];
} INQUIRYDATA, *PINQUIRYDATA;

typedef struct _dd_valid_flag_t{
    UCHAR ww_name_valid:1,
        dd_device_serial_valid:1,
        reserved:6;
} dd_valid_flag_t, *pdd_valid_flag_t;


// ****************** IOCTL_TAPE_SCALE_CAPACITY definitions *****************

typedef enum _CAP_SCAL_VALUES
{
    CAP_SCAL_25_PCT,
    CAP_SCAL_50_PCT,
    CAP_SCAL_75_PCT,
    CAP_SCAL_100_PCT,
    CAP_SCAL_60_GB,
    CAP_SCAL_USR_DEFINED
} CAP_SCAL_VALUES;

#define SCALING_60_GB   0x35
#define SCALING_300_GB  0x00
#define SCALING_25_PCT  0x40
#define SCALING_50_PCT  0x80
#define SCALING_75_PCT  0xC0
#define SCALING_100_PCT SCALING_300_GB

// ***************** IOCTL_TAPE_GET_ENCRYPTION_STATE definitions **************

#define INTERNAL_ENCRYPTION_IOCTLS 0

typedef struct _ENCRYPTION_STATUS
{
    UCHAR ucEncryptionCapable;
    UCHAR ucEncryptionMethod;
    UCHAR ucEncryptionState;
    UCHAR aucReserved[13];
} ENCRYPTION_STATUS, *PENCRYPTION_STATUS;

/* Keep in sync with char* array below */
typedef enum _ENCRYPTION_METHOD
{
    ENCRYPTION_METHOD_NONE, /* must be 0 */
    ENCRYPTION_METHOD_LIBRARY,
    ENCRYPTION_METHOD_SYSTEM,
    ENCRYPTION_METHOD_APPLICATION,
    ENCRYPTION_METHOD_CUSTOM,
    ENCRYPTION_METHOD_UNKNOWN,

    ENCRYPTION_METHOD_LARGEST_VALUE
} ENCRYPTION_METHOD, *PENCRYPTION_METHOD;

/* Keep in sync with enum above */
static char* aszEncryptionMethodStrings[] =
{
    "None",
    "Library Managed",
    "System Managed",
    "Application Managed",
    "Customized",
    "Unknown",

    NULL
};

typedef enum _ENCRYPTION_STATE
{
    ENCRYPTION_STATE_OFF, /* must be 0 */
    ENCRYPTION_STATE_ON,
    ENCRYPTION_STATE_NA
} ENCRYPTION_STATE, *PENCRYPTION_STATE;

#define DATA_KEY_INDEX_LENGTH       12
#define DATA_KEY_RESERVED1_LENGTH   15
#define DATA_KEY_LENGTH             32
#define DATA_KEY_RESERVED2_LENGTH   48

typedef struct _DATA_KEY
{
    UCHAR aucDataKeyIndex[DATA_KEY_INDEX_LENGTH];
    UCHAR ucDataKeyIndexLength;
    UCHAR aucReserved1[DATA_KEY_RESERVED1_LENGTH];
    UCHAR aucDataKey[DATA_KEY_LENGTH];
    UCHAR aucReserved2[DATA_KEY_RESERVED2_LENGTH];
} DATA_KEY, *PDATA_KEY;

#define SERVER_PING_DIAG        (1)
#define BASIC_ENCRYPTION_DIAG   (2)
#define FULL_ENCRYPTION_DIAG    (3)

typedef struct _ENCRYPTION_DIAGS
{
    int     iDiagType;
    int     iDiagErrno;
    int     iDiagCC;
    UCHAR   aucReserved[8];
} ENCRYPTION_DIAGS, *PENCRYPTION_DIAGS;

// ******************** IOCTL_TAPE_SCALE_CAPACITY definitions *****************

typedef struct _CAPACITY_SCALE
{
    UCHAR Value;
    UCHAR fSet;
} CAPACITY_SCALE, *PCAPACITY_SCALE;

/* Failover descriptor: contains the device name, sernum, and whether CPF is supported */
typedef struct _FAILOVER_DEVICE_DESCRIPTOR
{
    CHAR                szDeviceName[ONE_K/16];     /* "changerx" */

    CHAR                szSerNum[ONE_K/16];         /* serial number of device */

    BOOLEAN             fFailoverSupported;         /* CPF supported */

    SCSI_ADDRESS        scScsiAddr;                 /* SCSI address of path */

    BOOLEAN             fActive;                    /* Can we use this path? */

    BOOLEAN             fPrimary;                   /* Is this the primary path? */

}FAILOVER_DEVICE_DESCRIPTOR, *PFAILOVER_DEVICE_DESCRIPTOR;

// ****************** FAILOVER_IOCTL definitions *****************
#define IOCTL_CPF_BASE  0x84c9
#define IOCTL_DPF_BASE  0x84ca

#define DEFINE_CPF_IOCTL(function) CTL_CODE(IOCTL_CPF_BASE, function, \
    METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)

#define DEFINE_DPF_IOCTL(function) CTL_CODE(IOCTL_DPF_BASE, function, \
    METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)

#define CPF_IOCTL_BUS_QUERY_DEVICES     DEFINE_CPF_IOCTL(0x7)

#define DPF_IOCTL_BUS_QUERY_DEVICES      DEFINE_DPF_IOCTL(0x7)
#define DPF_IOCTL_ENCRYPTION_DIAGNOSTICS DEFINE_DPF_IOCTL(0xC)

/* Container for the failover device descriptors */
typedef struct _FAILOVER_DEVICE_DESCRIPTORS
{
    ULONG                       ulSize;             /* I/O: The size of this structure */

    FAILOVER_DEVICE_DESCRIPTOR  ascDescriptors[1];  /* the descriptors */

} FAILOVER_DEVICE_DESCRIPTORS, *PFAILOVER_DEVICE_DESCRIPTORS;

#define ChangerTransport 1
#define ChangerSlot      2
#define ChangerIEPort    3
#define ChangerDrive     4

void AdditionalChangerIoctls();
void IoctlsMenu();
void DisplayModeSense(PGET_CHANGER_PARAMETERS_W2K changerParameters);

int ReturnCount( HANDLE hDeviceMC);
int LibraryInventory( HANDLE hDeviceMC, long drives, long slots, long eeport, long transp_elems);
int MoveMedium( HANDLE hDeviceMC,USHORT transport,USHORT source,USHORT destination);
int LibraryAudit(HANDLE hDeviceMC, USHORT barcode);
int LibraryInfo(HANDLE hDeviceMC);
int Position(HANDLE hDeviceMC, USHORT transport, USHORT destination);
int LibraryInventoryRange( HANDLE hDeviceMC, int typecode , long startingAddress, long elements);
int ExtendedDriveInventory(HANDLE hDeviceMC, USHORT addr, long drives);
int ReturnError (HANDLE hDeviceMC, BOOL *error);
int GetInquiry (HANDLE hDeviceMC);
int ExtendedInquiry (HANDLE hDeviceMC);
int GetElementStatus(HANDLE hDeviceMC);
int GetInqByPage(char** szBuf, HANDLE handle, FILE* fp, UCHAR page_code);
int PreparePassThruForInq(SCSI_PASS_THROUGH_WITH_BUFFERS* sptwb,
    HANDLE handle, int iSize, int page_code, FILE* fp);

int PrintDevConfigPage(FILE* fp, char* pDevConfigPage);
void PrintConfigStandard(FILE* fp, DEV_CONFIG_ID* pID, int iIDLen);
void PrintConfigWWNN(FILE* fp, DEV_CONFIG_ID* pID, int iIDLen);
void PrintConfigWWPN(FILE* fp, DEV_CONFIG_ID* pID, int iIDLen);
void PrintConfigPID(FILE* fp, DEV_CONFIG_ID* pID, int iIDLen);
void PrintConfigAltWWPN(FILE* fp, DEV_CONFIG_ID* pID, int iIDLen);
void PrintConfigAltPID(FILE* fp, DEV_CONFIG_ID* pID, int iIDLen);

void* Malloc(size_t size);
int WhichDevice();
