/*******************************************************************************
********************************************************************************
**                                                                            **
**                    IBM Rekey Utility Program                               **
**              (C) COPYRIGHT IBM Corporation, 2008                           **
**                                                                            **
**                rekey_win.c - Rekey Utility Program Source File             **
**                  Licensed Materials - Property of IBM                      **
**                           All Rights Reserved                              **
**                                                                            **
**    US Government Users Restricted Rights  -  Use, duplication, or          **
**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.       **
**                                                                            **
**           ** Notice to users of the source code examples: **               **
**                                                                            **
**  The source code examples provided by IBM are only intended to assist      **
**  in the development of a working software program.  The source code        **
**  examples may not function as written; additional code is required.        **
**  In addition, the source code examples may not compile and/or bind         **
**  successfully as written.                                                  **
**                                                                            **
**  International Business Machines Corporation provides the source code      **
**  examples "as is" without warranty of any kind, either expressed or        **
**  implied, including, but not limited to the implied warranties of          **
**  merchantability and fitness for a particular purpose.  The entire         **
**  risk as to the quality and performance of the source code examples        **
**  is with you.  Should any part of the source code examples prove           **
**  defective, you (and not IBM) assume the entire cost of all necessary      **
**  servicing, repair or correction.                                          **
**                                                                            **
**  IBM does not warrant that the contents of the source code examples        **
**  will meet your requirements or that the source code examples are          **
**  error-free.                                                               **
**                                                                            **
**  IBM may make improvements and/or changes in the source code examples      **
**  at any time.  Changes may be made periodically to the information in      **
**  the source code examples; these changes may be reported, for the          **
**  device drivers included herein, in new editions of the examples.          **
**                                                                            **
**  References in the source code examples to IBM products, programs, or      **
**  services do not imply that IBM intends to make these available in all     **
**  countries in which IBM operates.  Any reference to an IBM licensed        **
**  program in the source code examples is not intended to state or imply     **
**  that only IBM's licensed program may be used.  Any functionally           **
**  equivalent program may be used.                                           **
**                                                                            **
**   ENVIRONMENT:       Windows Servers                                       **
**                                                                            **
**   COMPILER:        Microsoft Visual C++                                    **
**                                                                            **
**   PRIVATE FUNCTIONS: N/A                                                   **
**                                                                            **
**   PUBLIC  FUNCTIONS:                                                       **
**                                                                            **
**                                                                            **
********************************************************************************
* Change History                                                               *
* 1/09/2008      Init version                                                  *
********************************************************************************/


/*****************************************************************************
*                               INCLUDE FILES                                *
******************************************************************************/
#include <time.h>
#include "ntutil.h"

#define MIN(a,b) (a < b ? a : b)
#define SCSIOP_MOVE_MEDIUM 0xA5
#define MODE_PAGE_ELEMENT_ADDRESS 0x1D

#define MAX_FILENAME 255  /* for files (248 for directories), entire path. */

/*******************************************************************
** NAME:        main
**
** DESCRIPTION: Test tool main entry point.
**
** SYNOPSIS:    Parse any parameters and control further processing.
**
** RETURNS:
**
**              NOERROR - All tests were successful.
**              TESTFAIL - At least one test failed.
**
*******************************************************************/

int main(int argc,char *argv[])
{
  FILE * fp;
  char logfile[30];
  char vval[512] = {0};
  DWORD vtype, vlen = 10;
  HKEY  RegistryKey;
  ULONG i, j;
  int   rc;
  int   errno;
  char  temp_drive_name[MAX_FILENAME-1], drive_name[MAX_FILENAME];
  char  temp_lib_name[MAX_FILENAME-1], lib_name[MAX_FILENAME];
  int   slot, slot1, slot2;
  HANDLE drive_handle, lib_handle;
  DWORD  cb;
  struct tm *newtime;
  time_t long_time;

  ENCRYPTION_STATUS scEncryptStat = {0};
  SCSI_PASS_THROUGH_WITH_BUFFERS sptwb;
  unsigned char drive_sn[20];
  int elem, dlen, find_drive=0;
  char *devid_data;
  unsigned short drive_address;         /* drive element address */
  unsigned char  ident_len;             /* device identifier length */
  unsigned char  drive_devid[20];       /* device sn in devid */


  /* structure for get_lib_elements function */
  struct lib_element_data
  {
    unsigned short robot_addr;          /* first robot address */
    unsigned short robots;              /* number robot elements */
    unsigned short slot_addr;           /* first medium storage element address */
    unsigned short slots;               /* number medium storage elements */
    unsigned short ie_addr;             /* first import/export element address */
    unsigned short ie_stations;         /* number import-export elements */
    unsigned short drive_addr;          /* first data transfer element address */
    unsigned short drives;              /* number data transfer elements */
  };

  struct lib_element_data *lib_elements;

  CHANGER_READ_ELEMENT_STATUS rel;
  PCHANGER_ELEMENT_STATUS     ces;
  ULONG                       ulNumberofElements;

  /* the number of element types windows defines is 7*/
  unsigned short              count[7];
  unsigned char               type;
  unsigned short              offset[7];

  /*
  ** initialization
  */

  /* Check the OS version */
  /* Opening Hardware\devicemap\scsi */
  rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
                        "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion",
                        0,
                        KEY_READ,
                        &RegistryKey);

  if(rc != ERROR_SUCCESS)
  {
    printf("Fatal Error: Unable to open HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\n");
    return (-1);
  }

  rc = RegQueryValueEx(RegistryKey,
                           "CurrentVersion",
                           0,
                           &vtype,
                           vval,
                           &vlen);

  if(rc != ERROR_SUCCESS)
  {
    printf("Current Version is %s, return code is %d\n", vval, rc);
    printf("Fatal Error: Unable to query Current NT Version\n");
    return (-1);
  }

  RegCloseKey(RegistryKey);

  if(strcmp(vval, "4.0") == 0)
  {
    NTversion = 40;
  }
  else if(strcmp(vval, "5.0") == 0)
  {
    NTversion = 50;
  }
  else if(strcmp(vval, "5.2") == 0) /* corresponds to Windows Server 2003 */
  {
    NTversion = 52;
  }
  else
  {
    NTversion = 0;
  }

  setbuf(stdout, NULL); /* no buffering of standard output */

  if (argc != 5)
  {
    printf ("Missing argument.\n");
    printf ("\n");
    printf ("rekey:\n");
    printf ("- a program is used to replace the encryption key for tape cartridge\n");
    printf ("located at the storage slot only on Windows servers.\n\n");
    printf ("Usage: rekey drive_dsf changer_dsf 1st_slot_addr last_slot_addr\n");
    printf ("       for example, command-line: rekey tape0 Changer0 1027 2050 \n\n");
    printf ("Note:\n");
    printf ("1. The utility supports for 3592 drive in Library Managed Encryption (LME) mode only.\n");
    printf ("2. The utility runs with IBM tape drivers only.\n");
    printf ("3. The changer must be configured with IBM tape driver.\n");
    printf ("4. The drive must be in the same logical library.\n");
    printf ("5. The rekey tape cartridges must be located in storage slots.\n");
    printf ("6. Make sure the tape isn't loaded in the drive before to run the program.\n\n");
    return (0);
  }
  else
  {
    strcpy(temp_drive_name, argv[1]);
    for (i=0; i<strlen(temp_drive_name); i++)
    {
      temp_drive_name[i] = tolower(temp_drive_name[i]);
    }

    if(STRLNCMP(temp_drive_name, "tape") != 0)
    {
      printf("Argument error, invalid tape special file name on command line\n");
      SYNTAX_ERROR;
      return (0);
    }

    /* handle device name translation */
    drive_name[0] = '\0';
    strcpy(drive_name, "\\\\.\\");
    strcat(drive_name, temp_drive_name);

    strcpy(temp_lib_name, argv[2]);
    for (i=0; i<strlen(temp_lib_name); i++)
    {
      temp_lib_name[i] = tolower(temp_lib_name[i]);
    }

    if(STRLNCMP(temp_lib_name, "changer") != 0)
    {
      printf("Argument error, invalid changer special file name on command line\n");
      SYNTAX_ERROR;
      return (0);
    }

    /* handle device name translation */
    lib_name[0] = '\0';
    strcpy(lib_name, "\\\\.\\");
    strcat(lib_name, temp_lib_name);

    slot1 = atoi(argv[3]);
    slot2 = atoi(argv[4]);

    if (slot1 > slot2)
    {
      printf ("Invalid last slot address!\n");
      return (0);
    }
  }

  time( &long_time );                /* Get time as long integer. */
  newtime = localtime( &long_time ); /* Convert to local time. */

  sprintf(logfile, "Windows\\rekey_%s.%04d-%02d-%02d-%02dh-%02dm-%02ds.log",
          temp_drive_name, newtime->tm_year + 1900, newtime->tm_mon + 1,
          newtime->tm_mday, newtime->tm_hour, newtime->tm_min, newtime->tm_sec);

  if(!(fp = fopen(logfile, "w+")))
  {
    fprintf(stderr, "Error opening log file '%s'\n", logfile);
    return (0);
  }

  /* Get UNIX-style time and display as number and string. */
  fprintf(fp, "Rekey time and date:\t%s\n", ctime(&long_time));


  errno = 0;
  fprintf(fp, "tape special file (%s) will be opened\n", drive_name);

  /* open the drive file and set the file handle drive_hand */
  drive_handle = CreateFile(drive_name,                   /* file/device name */
                            GENERIC_READ|GENERIC_WRITE,   /* access */
                            0,                            /* share mode */
                            NULL,                         /* security attributes */
                            OPEN_EXISTING,                /* create options */
                            0,                            /* attrs and flags */
                            (HANDLE) NULL);               /* template file */

  if (drive_handle == INVALID_HANDLE_VALUE)
  {
    errno = GetLastError();
        printf("Error! Open failed for tape drive(errno=%d)\n", errno);
    fprintf(fp, "Error! Open failed for tape drive(errno=%d)\n", errno);
        fflush(fp);
    fclose(fp);
        printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
    return (-1);
  }
  else
  {
    fprintf(fp, "Open tape drive successful - get the handle.\n");
  }

  rc = DeviceIoControl(drive_handle,
                       IOCTL_SCSI_GET_ADDRESS,
                       NULL,
                       0,
                       &scsi_addr,
                       (long)sizeof(scsi_addr),
                       &cb,
                       (LPOVERLAPPED) NULL);
  if (!(rc))
  {  // rc will be == 0 in this case
        errno = GetLastError();
    printf("Error! open failed for tape drive - get_scsi_address(errno=%d).\n", errno);
        fprintf(fp, "Error! open failed for tape drive - get_scsi_address(errno=%d).\n", errno);
        fflush(fp);
    fclose(fp);
        printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
    CloseHandle(drive_handle);

    /* Returns if the requested resource is in use */
    if (errno == 170){ return 0; }
    return (-1);
  }
  else
  {
    fprintf(fp, "Open tape drive successful - get scsi address.\n");
  }

  /* get the tape drive open, now check encryption and any tape loading, get sn */
  errno = 0;
  fprintf (fp, "get_encrypt_state\n");

  rc = DeviceIoControl(drive_handle,
                            IOCTL_TAPE_GET_ENCRYPTION_STATE,
                            &scEncryptStat,
                            sizeof(ENCRYPTION_STATUS),
                            &scEncryptStat,
                            sizeof(ENCRYPTION_STATUS),
                            &cb,
                            (LPOVERLAPPED) NULL);
  if (!(rc))
  {
     errno = GetLastError();
     fprintf(fp, "execute_command() get_encrypt_state: ");
     fprintf(fp, "DeviceIoControl() failed with error %d.\n", errno);
  }
  else
  {
     fprintf(fp, "execute_command() get_encrypt_state: DeviceIoControl() succeeded.\n");

     if (scEncryptStat.ucEncryptionMethod != ENCRYPTION_METHOD_LIBRARY)
     {
        printf("Unsupport the encryption managed method.\n");
                fprintf(fp, "Unsupport the encryption managed method.\n");
                fflush(fp);
        fclose(fp);
        CloseHandle(drive_handle);
                printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
        return (0);
     }
  }

  /* Retrieve inquiry data p. 0x80, serial number */
  errno = 0;
  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

  sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.CdbLength          = CDB6GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn             = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = sizeof(UNIT_SERIAL_NUMBER_PAGE);
  sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
  sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
  sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
  sptwb.spt.Cdb[0]             = SCSIOP_INQUIRY;
  sptwb.spt.Cdb[1]            |= 1; /* EVPD */
  sptwb.spt.Cdb[2]             = SN_PAGE_CODE;
  sptwb.spt.Cdb[4]             = sizeof(UNIT_SERIAL_NUMBER_PAGE);
  sptwb.spt.PathId             = scsi_addr.PathId;
  sptwb.spt.TargetId           = scsi_addr.TargetId;
  sptwb.spt.Lun                = scsi_addr.Lun;

  rc = DeviceIoControl(drive_handle,
                       IOCTL_SCSI_PASS_THROUGH,
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &cb,
                       (LPOVERLAPPED) NULL);

  if (rc)
  {
    /* Check if SCSI status is good */
    if (sptwb.spt.ScsiStatus)
    {
      fprintf(fp, "Inquiry command (p. 0x80) resulted in check condition!\n");
      PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
      printf("Error! Get drive SN failed.\n");
      printf("Exit to run the program!\n");
      fprintf(fp, "Error! Get drive SN failed.\n");
      fprintf(fp, "Exit to run the program!\n");
      fflush(fp);
      fclose(fp);
      CloseHandle(drive_handle);
          printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
      return 0;
    } /* if */
    else
    {
      memcpy(drive_sn, ((UNIT_SERIAL_NUMBER_PAGE*)(&sptwb.ucDataBuf))->unit_serial_number, MAX_SERIAL_NUMBER_LENGTH);
      fprintf(fp, "The SN of the tape drive %s is %s.\n", drive_name, drive_sn);
    }
  }
  else
  {
     rc = -1;
     errno = GetLastError();
     printf("Error! Get drive SN failed (errno = %d).\n", errno);
     printf("Exit to run the program!\n");
     fprintf(fp, "Error! Get drive SN failed (errno = %d).\n", errno);
     fprintf(fp, "Exit to run the program!\n");
     fflush(fp);
     fclose(fp);
     CloseHandle(drive_handle);
         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
     return (0);
  }

  /* Check if any tape is loaded in the drive */
  errno = 0;
  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

  sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId             = scsi_addr.PathId;
  sptwb.spt.TargetId           = scsi_addr.TargetId;
  sptwb.spt.Lun                = scsi_addr.Lun;
  sptwb.spt.CdbLength          = CDB6GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn             = SCSI_IOCTL_DATA_UNSPECIFIED;
  sptwb.spt.DataTransferLength = 0;
  sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
  sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
  sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
  sptwb.spt.Cdb[0]             = SCSIOP_TEST_UNIT_READY;

  rc = DeviceIoControl(drive_handle,
                       IOCTL_SCSI_PASS_THROUGH,
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &cb,
                       (LPOVERLAPPED) NULL);

  if (rc)
  {
    /* Check if SCSI status is good */
    if (sptwb.spt.ScsiStatus)
    {
      fprintf(fp, "TUR resulted in check condition!\n");
      PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
      fprintf(fp, "Continue to run the program!\n");
    } /* if */
    else
    {
      /* tape is loaded */
      printf("A tape is loaded in the drive. Please unload it before to run the progem!\n");
      printf("Exit to run the program!\n");
      fprintf(fp, "A tape is loaded in the drive. Please unload it before to run the progem!\n");
      fprintf(fp, "Exit to run the program!\n");
      fflush(fp);
      fclose(fp);
          CloseHandle(drive_handle);
          printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
      return (0);
    }
  }
  else
  {
     rc = -1;
     errno = GetLastError();
     fprintf(fp, "Error! Fail to issue TUR command(errno = %d).\n", errno);
     fprintf(fp, "Continue to run the program anyway!\n");
  }

  /* we are done to prepare the tape drive for the rekeying */
  /* it is time to open the changer */
  errno = 0;
  fprintf(fp, "changer special file (%s) will be opened\n", lib_name);

  /* open the drive file and set the file handle drive_hand */
  lib_handle = CreateFile(lib_name,                   /* file/device name */
                          GENERIC_READ|GENERIC_WRITE,   /* access */
                          0,                            /* share mode */
                          NULL,                         /* security attributes */
                          OPEN_EXISTING,                /* create options */
                          0,                            /* attrs and flags */
                          (HANDLE) NULL);               /* template file */

  if (lib_handle == INVALID_HANDLE_VALUE)
  {
    errno = GetLastError();
    printf("Error! Open failed for changer (errno=%d)\n", errno);
    fprintf(fp, "Error! Open failed for changer (errno=%d)\n", errno);
    fflush(fp);
    fclose(fp);
        CloseHandle(drive_handle);
        printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
    return (-1);
  }
  else
  {
    fprintf(fp, "Open changer successful - get the handle.\n");
  }

  rc = DeviceIoControl(lib_handle,
                       IOCTL_SCSI_GET_ADDRESS,
                       NULL,
                       0,
                       &scsi_addr1,
                       (long)sizeof(scsi_addr1),
                       &cb,
                       (LPOVERLAPPED) NULL);
  if (!(rc))
  {  // rc will be == 0 in this case
    errno = GetLastError();
        printf("Error! open failed for changer - get_scsi_address(errno=%d).\n",errno);
    fprintf(fp, "Error! open failed for changer - get_scsi_address(errno=%d).\n",errno);
    fflush(fp);
    fclose(fp);
    CloseHandle(drive_handle);
        printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
    /* Returns if the requested resource is in use */
    if (errno == 170){ return 0; }
    return (-1);
  }
  else
  {
    fprintf(fp, "Open changer successful - get scsi address.\n");
  }

  /* Check if the tape drive is in the library */
  /* Get library element info first */
  errno = 0;
  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
  lib_elements = malloc(sizeof(struct lib_element_data));

  sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId             = scsi_addr1.PathId;
  sptwb.spt.TargetId           = scsi_addr1.TargetId;
  sptwb.spt.Lun                = scsi_addr1.Lun;
  sptwb.spt.CdbLength          = CDB6GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn             = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = 0x38;
  sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
  sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
  sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
  sptwb.spt.Cdb[0]             = SCSIOP_MODE_SENSE;
  sptwb.spt.Cdb[1]             = 0x08;         /* DBD */
  sptwb.spt.Cdb[2]             = MODE_PAGE_ELEMENT_ADDRESS;
  sptwb.spt.Cdb[4]             = 0x38;

  rc = DeviceIoControl(lib_handle,
                       IOCTL_SCSI_PASS_THROUGH,
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &cb,
                       (LPOVERLAPPED) NULL);

  if (rc)
  {
    /* Check if SCSI status is good */
    if (sptwb.spt.ScsiStatus)
    {
          printf("Fail to get lib element information!\n");
      fprintf(fp, "Get Mode Page resulted in check condition!\n");
      PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
      printf("Exit to run the program!\n");
      fprintf(fp, "Exit to run the program!\n");
          fflush(fp);
      fclose(fp);
      CloseHandle(drive_handle);
      CloseHandle(lib_handle);
          printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
      return (0);
    }
    else
    {
      fprintf(fp, "Get Library Element Info.\n");
          // dump_bytes(fp, sptwb.ucDataBuf, sptwb.spt.DataTransferLength, TRUE);

      lib_elements->drives     = sptwb.ucDataBuf[20] << 8 | sptwb.ucDataBuf[21];
      lib_elements->drive_addr = sptwb.ucDataBuf[18] << 8 | sptwb.ucDataBuf[19];
      lib_elements->slots      = sptwb.ucDataBuf[12] << 8 | sptwb.ucDataBuf[13];
      lib_elements->slot_addr  = sptwb.ucDataBuf[10] << 8 | sptwb.ucDataBuf[11];
      lib_elements->robots     = sptwb.ucDataBuf[ 8] << 8 | sptwb.ucDataBuf[ 9];
      lib_elements->robot_addr = sptwb.ucDataBuf[ 6] << 8 | sptwb.ucDataBuf[ 7];
      lib_elements->ie_stations= sptwb.ucDataBuf[16] << 8 | sptwb.ucDataBuf[17];
      lib_elements->ie_addr    = sptwb.ucDataBuf[14] << 8 | sptwb.ucDataBuf[15];
    }
  }
  else
  {
     rc = -1;
     errno = GetLastError();
         printf("Fail to issue Mode Page command to get lib elem info(errno=%d).\n", errno);
     printf("Exit to run the program!\n");
     fprintf(fp, "Error! Fail to issue Mode Page command(errno=%d).\n", errno);
     fprintf(fp, "Exit to run the program!\n");
         fflush(fp);
     fclose(fp);
     CloseHandle(drive_handle);
     CloseHandle(lib_handle);
         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
     return (0);
  }

  /* Get drive devids */
  errno = 0;
  ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

  sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
  sptwb.spt.PathId             = scsi_addr1.PathId;
  sptwb.spt.TargetId           = scsi_addr1.TargetId;
  sptwb.spt.Lun                = scsi_addr1.Lun;
  sptwb.spt.CdbLength          = CDB10GENERIC_LENGTH;
  sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
  sptwb.spt.DataIn             = SCSI_IOCTL_DATA_IN;
  sptwb.spt.DataTransferLength = lib_elements->drives*125+16;
  sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
  sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
  sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
  sptwb.spt.Cdb[0]             = 0xB8;
  sptwb.spt.Cdb[1]             = (unsigned char)((scsi_addr1.Lun << 5) + 4);
  sptwb.spt.Cdb[2]             = (unsigned char)(lib_elements->drive_addr >> 8);
  sptwb.spt.Cdb[3]             = (unsigned char)(lib_elements->drive_addr & 0xff);
  sptwb.spt.Cdb[4]             = (unsigned char)(lib_elements->drives >> 8);
  sptwb.spt.Cdb[5]             = (unsigned char)(lib_elements->drives & 0xff);
  sptwb.spt.Cdb[6]             = 1;
  sptwb.spt.Cdb[7]             = sizeof(sptwb.ucDataBuf) >> 16;
  sptwb.spt.Cdb[8]             = sizeof(sptwb.ucDataBuf) >> 8 & 0xff;
  sptwb.spt.Cdb[9]             = sizeof(sptwb.ucDataBuf) & 0xff;

  rc = DeviceIoControl(lib_handle,
                       IOCTL_SCSI_PASS_THROUGH,
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &sptwb,
                       sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                       &cb,
                       (LPOVERLAPPED) NULL);

  find_drive = 0;

  if (rc)
  {
    /* Check if SCSI status is good */
    if (sptwb.spt.ScsiStatus)
    {
          fprintf(fp, "Get devids resulted in check condition!\n");
      PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
      printf("Can not verify if the tape drive in the library!\n");
      printf("Exit to run the program!\n");
      fprintf(fp, "Can not verify if the tape drive in the library!\n");
      fprintf(fp, "Exit to run the program!\n");
          fflush(fp);
      fclose(fp);
      CloseHandle(drive_handle);
      CloseHandle(lib_handle);
          printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
      return (0);
    }
    else
    {
      elem = (sptwb.ucDataBuf[2]<<8) + sptwb.ucDataBuf[3];
      dlen = (sptwb.ucDataBuf[10]<<8) + sptwb.ucDataBuf[11];
      devid_data = &(sptwb.ucDataBuf[16]);

      for (i=0; i<(unsigned int)elem && i<(unsigned int)lib_elements->drives; i++, devid_data += dlen)
      {
          drive_address = (devid_data[0] << 8) + devid_data[1];
          ident_len =(unsigned char)devid_data[15];

          memcpy(&(drive_devid[0]), &(devid_data[40]), (ident_len == 36)?12:10);
          if (!strcmp (drive_devid, drive_sn))
          {
             fprintf(fp, "Find the drive with sn (%s) and devid (%s) at the addr %d!\n",
                    drive_sn, drive_devid, drive_address);
             find_drive = 1;
             break;
          }
      }

      if (!find_drive)
      {
        printf("Can not find the tape drive in the library!\n");
        printf("Exit to run the program!\n");
            fprintf(fp, "Can not find the tape drive in the library!\n");
        fprintf(fp, "Exit to run the program!\n");
            fflush(fp);
        fclose(fp);
        CloseHandle(drive_handle);
        CloseHandle(lib_handle);
                printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
        return (0);
      }
    }
  }
  else
  {
     rc = -1;
     errno = GetLastError();
     fprintf(fp, "Error! Fail to issue devids command(errno = %d).\n", errno);
     printf("Can not verify if the tape drive in the library!\n");
     printf("Exit to run the program!\n");
         fprintf(fp, "Can not verify if the tape drive in the library!\n");
     fprintf(fp, "Exit to run the program!\n");
         fflush(fp);
     fclose(fp);
     CloseHandle(drive_handle);
     CloseHandle(lib_handle);
         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
     return (0);
  }

  /* Get library inventory */
  ZeroMemory(&rel, sizeof(CHANGER_READ_ELEMENT_STATUS));

  ulNumberofElements =
      lib_elements->drives + lib_elements->slots +
      lib_elements->robots + lib_elements->ie_stations;

  ces = (CHANGER_ELEMENT_STATUS *)malloc(sizeof(CHANGER_ELEMENT_STATUS) * ulNumberofElements);
  ZeroMemory(ces, sizeof(CHANGER_ELEMENT_STATUS) * ulNumberofElements);

  rel.ElementList.Element.ElementType         = (ELEMENT_TYPE) 2;    /* Storage Slot only */
  rel.ElementList.Element.ElementAddress      = 0;
  rel.ElementList.NumberOfElements            = ulNumberofElements;
  rel.VolumeTagInfo                           = TRUE;

  errno = 0;
  /* now we need to build a IOCTL_CHANGER_GET_ELEMENT_STATUS
     calc the needed data length  */
  rc = DeviceIoControl(lib_handle,
                      IOCTL_CHANGER_GET_ELEMENT_STATUS,
                      &rel,
                      (long)sizeof(CHANGER_READ_ELEMENT_STATUS),
                      ces,
                      (long)sizeof(CHANGER_ELEMENT_STATUS) * ulNumberofElements,
                      &cb,
                      (LPOVERLAPPED) NULL);

   if (!(rc))
   {
     rc = -1;
     errno = GetLastError();
     printf("Fail to storage slot information(errno = %d).\n", errno);
     printf("Exit to run the program!\n");
         fprintf(fp, "Error! Fail to issue inventory command(errno = %d).\n", errno);
     fprintf(fp, "Exit to run the program!\n");
         fflush(fp);
     fclose(fp);
     free(ces);
     CloseHandle(drive_handle);
     CloseHandle(lib_handle);
         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
     return (errno?errno:EIO);
   }

   if (cb != ((long)sizeof(CHANGER_ELEMENT_STATUS) * ulNumberofElements))
   {
     rc = -1;
     errno = ENOMEM;
     printf("Error! Fail to get all of inventory info(errno=%d).\n", errno);
     printf("Exit to run the program!\n");
     fprintf(fp, "Error! Fail to get all of inventory info(errno=%d).\n", errno);
     fprintf(fp, "Exit to run the program!\n");
         fflush(fp);
     fclose(fp);
     free(ces);
     CloseHandle(drive_handle);
     CloseHandle(lib_handle);
         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
     return (errno);
   }

   ZeroMemory(count,sizeof(count));

   count[ChangerDrive]         = lib_elements->drives;
   count[ChangerSlot]          = lib_elements->slots;
   count[ChangerIEPort]        = lib_elements->ie_stations;
   count[ChangerTransport]     = lib_elements->robots;

   offset[ChangerDrive]        = lib_elements->drive_addr;
   offset[ChangerSlot]         = lib_elements->slot_addr;
   offset[ChangerIEPort]       = lib_elements->ie_addr;
   offset[ChangerTransport]    = lib_elements->robot_addr;

   for(i=0; i<ulNumberofElements; i++)
   {
      type = ces[i].Element.ElementType;
      if( count[type] == 0 ) continue;
      count[type]--;

      slot = (unsigned short)(ces[i].Element.ElementAddress + offset[type]);

          /* check the slot address, move the tape, rekey, unload and move it back here */
      if ((slot >= slot1) && (slot <= slot2))
      {
                 fprintf(fp, "\nCheck slot %d in the range of slot1 %d and slot2 %d\n", slot, slot1, slot2);
         if ((unsigned char)(ces[i].Flags & ELEMENT_STATUS_FULL))
         {
           /* move the tape into the drive */
           printf("\nMoving the tape with %0.8s from slot %d to drive %d\n",
                     (char *)ces[i].PrimaryVolumeID, slot, drive_address);
           fprintf(fp, "\nMoving the tape with %0.8s from slot %d to drive %d\n",
                     (char *)ces[i].PrimaryVolumeID, slot, drive_address);
           errno = 0;
           ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

           sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
           sptwb.spt.PathId             = scsi_addr1.PathId;
           sptwb.spt.TargetId           = scsi_addr1.TargetId;
           sptwb.spt.Lun                = scsi_addr1.Lun;
           sptwb.spt.CdbLength          = CDB10GENERIC_LENGTH;
           sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
           sptwb.spt.DataIn             = SCSI_IOCTL_DATA_UNSPECIFIED;
           sptwb.spt.DataTransferLength = 0;
           sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
           sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
           sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
           sptwb.spt.Cdb[0]             = SCSIOP_MOVE_MEDIUM;
           sptwb.spt.Cdb[1]             = scsi_addr1.Lun << 5;
           sptwb.spt.Cdb[2]             = 0; // robot = 0
           sptwb.spt.Cdb[3]             = 0;
           sptwb.spt.Cdb[4]             = (unsigned char)(slot >> 8);
           sptwb.spt.Cdb[5]             = (unsigned char)(slot & 0xff);
           sptwb.spt.Cdb[6]             = (unsigned char)(drive_address >> 8);
           sptwb.spt.Cdb[7]             = (unsigned char)(drive_address & 0xff);

           rc = DeviceIoControl(lib_handle,
                            IOCTL_SCSI_PASS_THROUGH,
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &cb,
                            (LPOVERLAPPED) NULL);

           if (rc)
           {
             /* Check if SCSI status is good */
             if (sptwb.spt.ScsiStatus)
             {
               fprintf(fp, "Move tape into drive resulted in check condition!\n");
               PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
               continue;
             } /* if */
             else
             {
               printf("Move tape to drive done!\n");
                           fprintf(fp, "Move tape to drive done!\n");
             }
           }
           else
           {
             rc = -1;
             errno = GetLastError();
             fprintf(fp, "Error! Fail to move the tape to the drive(errno = %d).\n", errno);
             continue;
           }

           /* run TUR to clean out UA */
           Sleep (15000);

           errno = 0;
           ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

           sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
           sptwb.spt.PathId             = scsi_addr.PathId;
           sptwb.spt.TargetId           = scsi_addr.TargetId;
           sptwb.spt.Lun                = scsi_addr.Lun;
           sptwb.spt.CdbLength          = CDB6GENERIC_LENGTH;
           sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
           sptwb.spt.DataIn             = SCSI_IOCTL_DATA_UNSPECIFIED;
           sptwb.spt.DataTransferLength = 0;
           sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
           sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
           sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
           sptwb.spt.Cdb[0]             = SCSIOP_TEST_UNIT_READY;

           for (j=0; j<20; j++)
                   {
             rc = DeviceIoControl(drive_handle,
                            IOCTL_SCSI_PASS_THROUGH,
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &cb,
                            (LPOVERLAPPED) NULL);
             if (!rc) break;
                   }

           if (rc)
                   {
             /* Check if SCSI status is good */
             if (sptwb.spt.ScsiStatus)
                         {
                fprintf(fp, "TUR resulted in check condition!\n");
                PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
                fprintf(fp, "Continue to run the program!\n");
                         } /* if */
             else
                         {
                /* tape is ready */
                fprintf(fp, "tape is ready for rekey!\n");
                         }
                   }
           else
                   {
              rc = -1;
              errno = GetLastError();
              fprintf(fp, "Error! Fail to issue TUR command(errno=%d).\n", errno);
              fprintf(fp, "Continue to run the program anyway!\n");
                   }

           /* rekey the tape */
           printf("Rekey the tape...\n");
           fprintf(fp, "Rekey the tape...\n");

                   errno = 0;
           ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));
           ZeroMemory(&sptwb.ucDataBuf, sizeof(sptwb.ucDataBuf));

           sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
           sptwb.spt.PathId             = scsi_addr.PathId;
           sptwb.spt.TargetId           = scsi_addr.TargetId;
           sptwb.spt.Lun                = scsi_addr.Lun;
           sptwb.spt.CdbLength          = CDB10GENERIC_LENGTH;
           sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
           sptwb.spt.DataIn             = SCSI_IOCTL_DATA_IN;
           sptwb.spt.DataTransferLength = sizeof(sptwb.ucDataBuf);
           sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
           sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
           sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
           sptwb.spt.Cdb[0]             = 0x55;         /* SCSIOP_MODE_SELECT */
           sptwb.spt.Cdb[1]             = 0x10;
           sptwb.spt.Cdb[2]             = 0;
           sptwb.spt.Cdb[3]             = 0;
           sptwb.spt.Cdb[4]             = 0;
           sptwb.spt.Cdb[5]             = 0;
           sptwb.spt.Cdb[6]             = 0;
           sptwb.spt.Cdb[7]             = 0;
           sptwb.spt.Cdb[8]             = 56;
           sptwb.spt.Cdb[9]             = 0;

           sptwb.ucDataBuf[3]           = 0x10;
           sptwb.ucDataBuf[8]           = 0x65;
           sptwb.ucDataBuf[9]           = 0xEE;
           sptwb.ucDataBuf[11]          = 0x2C;
           sptwb.ucDataBuf[12]          = 0x21;
           sptwb.ucDataBuf[13]          = 0xEE;
           sptwb.ucDataBuf[15]          = 0x28;
           sptwb.ucDataBuf[20]          = 0x18;
           sptwb.ucDataBuf[28]          = 0x18;
           sptwb.ucDataBuf[31]          = 0x04;
           sptwb.ucDataBuf[39]          = 0x02;
           sptwb.ucDataBuf[40]          = 0x10;
           sptwb.ucDataBuf[44]          = 0x19;
           sptwb.ucDataBuf[48]          = 0x10;
           sptwb.ucDataBuf[52]          = 0x19;

                   // dump_bytes(fp, sptwb.ucDataBuf, 56, TRUE);

                   for (j=0; j<5; j++)
                   {
              rc = DeviceIoControl(drive_handle,
                            IOCTL_SCSI_PASS_THROUGH,
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &cb,
                            (LPOVERLAPPED) NULL);
              if (!rc) break;
                   }

           if (rc)
           {
             /* Check if SCSI status is good */
             if (sptwb.spt.ScsiStatus)
             {
                           printf("Rekey failed in check condition!\n");
               fprintf(fp, "Rekey resulted in check condition!\n");
               PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
             } /* if */
             else
             {
               printf("Rekey tape succeed!\n");
               fprintf(fp, "Rekey tape succeed!\n");
             }
           }
           else
           {
             rc = -1;
             errno = GetLastError();
             printf("Error! Fail to rekey the tape (errno = %d).\n", errno);
                         fprintf(fp, "Error! Fail to rekey the tape (errno=%d).\n", errno);
           }

           /* Unload the tape */
           printf("Unload the tape\n");
           fprintf(fp, "Unload the tape\n");

           errno = 0;
           ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

           sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
           sptwb.spt.PathId             = scsi_addr.PathId;
           sptwb.spt.TargetId           = scsi_addr.TargetId;
           sptwb.spt.Lun                = scsi_addr.Lun;
           sptwb.spt.CdbLength          = CDB10GENERIC_LENGTH;
           sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
           sptwb.spt.DataIn             = SCSI_IOCTL_DATA_UNSPECIFIED;
           sptwb.spt.DataTransferLength = 0;
           sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
           sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
           sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
           sptwb.spt.Cdb[0]             = 0x1B;
           sptwb.spt.Cdb[4]             = 0;

                   for (j=0; j<5; j++)
                   {
              rc = DeviceIoControl(drive_handle,
                            IOCTL_SCSI_PASS_THROUGH,
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &cb,
                            (LPOVERLAPPED) NULL);

                          if (!rc) break;
                   }

           if(rc)
           {
             /* Check if SCSI status is good */
             if (sptwb.spt.ScsiStatus)
             {
                           if ((sptwb.ucSenseBuf[2]==0x02) && (sptwb.ucSenseBuf[12]==0x04) && (sptwb.ucSenseBuf[13]==0x03))
                           {
                  printf("Unload tape done!\n");
                  fprintf(fp, "Unload tape done!\n");
                           }
                           else
                           {
                                  fprintf(fp, "Unload tape from drive resulted in check condition!\n");
                  PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
                                  printf("Fail to unload the tape\n");
                  printf("Exit to run the program!\n");
                  fprintf(fp, "Fail to unload the tape\n");
                  fprintf(fp, "Exit to run the program!\n");
                                  fflush(fp);
                  fclose(fp);
                  free(ces);
                  CloseHandle(drive_handle);
                  CloseHandle(lib_handle);
                                  printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
                  return (0);

                           } /* if */
             }
                         else
             {
               printf("Unload tape done!\n");
               fprintf(fp, "Unload tape done!\n");
             }
           }
           else
           {
             rc = -1;
             errno = GetLastError();
             printf("Error! Fail to unload the tape(errno = %d).\n", errno);
             printf("Exit to run the program!\n");
             fprintf(fp, "Error! Fail to unload the tape(errno = %d).\n", errno);
             fprintf(fp, "Exit to run the program!\n");
                         fflush(fp);
             fclose(fp);
             free(ces);
             CloseHandle(drive_handle);
             CloseHandle(lib_handle);
                         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
             return (errno);
           }


           /* Move the tape back to slot */
           printf("Moving the tape with %0.8s from drive %d to slot %d\n",
                  (char *)ces[i].PrimaryVolumeID, drive_address, slot);
                   fprintf(fp, "Moving the tape with %0.8s from drive %d to slot %d\n",
                  (char *)ces[i].PrimaryVolumeID, drive_address, slot);

           errno = 0;
           ZeroMemory(&sptwb, sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS));

           sptwb.spt.Length             = sizeof(SCSI_PASS_THROUGH);
           sptwb.spt.PathId             = scsi_addr1.PathId;
           sptwb.spt.TargetId           = scsi_addr1.TargetId;
           sptwb.spt.Lun                = scsi_addr1.Lun;
           sptwb.spt.CdbLength          = CDB10GENERIC_LENGTH;
           sptwb.spt.SenseInfoLength    = SENSE_BUFFER_SIZE;
           sptwb.spt.DataIn             = SCSI_IOCTL_DATA_UNSPECIFIED;
           sptwb.spt.DataTransferLength = 0;
           sptwb.spt.TimeOutValue       = SCSI_PASSTHRU_TIMEOUT_VALUE;
           sptwb.spt.DataBufferOffset   = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucDataBuf);
           sptwb.spt.SenseInfoOffset    = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS,ucSenseBuf);
           sptwb.spt.Cdb[0]             = SCSIOP_MOVE_MEDIUM;
           sptwb.spt.Cdb[1]             = scsi_addr1.Lun << 5;
           sptwb.spt.Cdb[2]             = 0; // robot = 0
           sptwb.spt.Cdb[3]             = 0;
           sptwb.spt.Cdb[4]             = (unsigned char)(drive_address >> 8);
           sptwb.spt.Cdb[5]             = (unsigned char)(drive_address & 0xff);
           sptwb.spt.Cdb[6]             = (unsigned char)(slot >> 8);
           sptwb.spt.Cdb[7]             = (unsigned char)(slot & 0xff);

           rc = DeviceIoControl(lib_handle,
                            IOCTL_SCSI_PASS_THROUGH,
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &sptwb,
                            sizeof(SCSI_PASS_THROUGH_WITH_BUFFERS),
                            &cb,
                            (LPOVERLAPPED) NULL);

           if (rc)
           {
             /* Check if SCSI status is good */
             if (sptwb.spt.ScsiStatus)
             {
               fprintf(fp, "Move tape back to slot resulted in check condition!\n");
               PrintSenseInfo(fp, sptwb.spt.ScsiStatus, sptwb.ucSenseBuf, sptwb.spt.SenseInfoLength);
                           printf("Fail to move tape back to slot %d\n", slot);
               printf("Exit to run the program!\n");
               fprintf(fp, "Fail to move tape back to slot %d\n", slot);
               fprintf(fp, "Exit to run the program!\n");
                           fflush(fp);
               fclose(fp);
               free(ces);
               CloseHandle(drive_handle);
               CloseHandle(lib_handle);
                           printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
               return (0);
             } /* if */
             else
             {
               printf("Done to move tape back to slot %d!\n", slot);
               fprintf(fp, "Done to move tape back to slot %d!\n", slot);
             }
           }
           else
           {
             rc = -1;
             errno = GetLastError();
             printf("Error! Fail to move the tape back to slot %d(errno = %d).\n", slot, errno);
             printf("Exit to run the program!\n");
             fprintf(fp, "Error! Fail to move the tape back to slot %d(errno = %d).\n", slot, errno);
             fprintf(fp, "Exit to run the program!\n");
             fflush(fp);
             fclose(fp);
             free(ces);
             CloseHandle(drive_handle);
             CloseHandle(lib_handle);
                         printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
             return (errno);
           }
         } /* slot is full */
         else
                 {
                    fprintf(fp, "slot %d is empty.\n", slot);
                 }
       }  /* slot in range */
           else if (slot > slot2)
           {
                  break;
           }
    } /* for */

        printf("\nDone to rekey!\n");
        fprintf(fp, "\nDone to rekey!\n");
        fflush(fp);
    fclose(fp);
    free(ces);
    CloseHandle(drive_handle);
    CloseHandle(lib_handle);
        printf("Please take a look %-15s to find any error report in ths file!!\n", logfile);
    return 0;

}

/*******************************************************************
** NAME:        PrintSenseInfo
**
** DESCRIPTION:
**
** SYNOPSIS:
**
** RETURNS:     void
**
*******************************************************************/

VOID PrintSenseInfo(FILE  *fp,
                                        UCHAR ScsiStatus,
                    UCHAR *ucSenseBuf,
                    UCHAR SenseInfoLength)
{
  int i;

  fprintf(fp, "Scsi status: %02Xh, Sense Info Length %xh\n\n", ScsiStatus, SenseInfoLength);

  if (SenseInfoLength)
  {
    fprintf(fp, "Sense KEY/ASC/ASCQ = %02x/%02x/%02x \n", ucSenseBuf[2], ucSenseBuf[12], ucSenseBuf[13]);

    fprintf(fp, "Sense Info -- consult SCSI spec for details\n");
    fprintf(fp, "      00  01  02  03  04  05  06  07   08  09  0A  0B  0C  0D  0E  0F\n");
    fprintf(fp, "      ---------------------------------------------------------------\n");
    for(i = 0; i < SenseInfoLength; i++)
    {
      if((i) % 16 == 0)
      {
        fprintf(fp, " %03X  ", i);
      }
      fprintf(fp, "%02X  ", ucSenseBuf[i]);
      if((i + 1) % 8 == 0)
      {
        fprintf(fp, " ");
      }
      if((i + 1) % 16 == 0)
      {
        fprintf(fp, "\n");
      }
    }
    fprintf(fp, "\n\n");
  }
}


/*******************************************************************
** NAME:        dump_bytes
**
** DESCRIPTION: Print data buffer
**
** SYNOPSIS:
**
** RETURNS:     void
**
*******************************************************************/
#define HEX_BUFFER_SIZE 50
#define ASCII_BUFFER_SIZE 50

VOID dump_bytes(FILE  *fp,
                                UCHAR *ucDataBuf,
                DWORD DataTransferLength,
                BOOL  truncate_print)
{
    ULONG i;
    int   hex_index   = 0;
    int   ascii_index = 0;
    char  hex_buffer[HEX_BUFFER_SIZE];
    char  ascii_buf[ASCII_BUFFER_SIZE];
    unsigned int index = 0;

    fprintf(fp, "Dump Bytes - number to dump: %Xh\n", DataTransferLength);

    if(DataTransferLength)
    {

      if((DataTransferLength > 256) && truncate_print)
      {
        DataTransferLength = 256;
      }

      fprintf(fp, "Data Buffer Info\n");
      fprintf(fp, "        0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F");
      fprintf(fp, "     0123456789ABCDEF\n");
      fprintf(fp, "        -----------------------------------------------");
      fprintf(fp, "    ----------------\n");
      BZERO(hex_buffer, HEX_BUFFER_SIZE);
      BZERO(ascii_buf, ASCII_BUFFER_SIZE);
      for(i = 0; i < DataTransferLength; i++)
      {
        sprintf(&hex_buffer[hex_index],
            " %02x", ucDataBuf[i]);
        if(ucDataBuf[i] < 32)
        {
          ascii_buf[ascii_index] = '.';
        }
        else
        {
          ascii_buf[ascii_index] = ucDataBuf[i];
        }
        hex_index += 3;
        ascii_index += 1;

        if(((i + 1) % 16 == 0) || ((i + 1) == DataTransferLength))
        {
          fprintf(fp, "%05x ", index);
          fprintf(fp, " %-48s", hex_buffer);
          fprintf(fp, "    %-16s\n", ascii_buf);
          hex_index = 0;
          ascii_index = 0;
          index += 16;
          BZERO(hex_buffer, HEX_BUFFER_SIZE);
          BZERO(ascii_buf, ASCII_BUFFER_SIZE);
        }
      }
      fprintf(fp, "\n\n");
    }
} /* dump_bytes() */
