/* IBMtapeutil for rekeying use                                              */
/*****************************************************************************/
/**                   IBM SCSI Tape Utility Program                         **/
/**                        Copyright (C) 2001-                              **/
/**                International Business Machines Corporation              **/
/**                                                                         **/
/**               IBMtapeutil.c - Test Program Source File                  **/
/**                Licensed Materials - Property of IBM                     **/
/**                         All Rights Reserved                             **/
/**                                                                         **/
/**    US Government Users Restricted Rights  -  Use, duplication, or       **/
/**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.    **/
/**                                                                         **/
/**           ** Notice to users of the source code examples: **            **/
/**                                                                         **/
/**  The source code examples provided by IBM are only intended to assist   **/
/**  in the development of a working software program.  The source code     **/
/**  examples may not function as written; additional code is required.     **/
/**  In addition, the source code examples may not compile and/or bind      **/
/**  successfully as written.                                               **/
/**                                                                         **/
/**  International Business Machines Corporation provides the source code   **/
/**  examples "as is" without warranty of any kind, either expressed or     **/
/**  implied, including, but not limited to the implied warranties of       **/
/**  merchantability and fitness for a particular purpose.  The entire      **/
/**  risk as to the quality and performance of the source code examples     **/
/**  is with you.  Should any part of the source code examples prove        **/
/**  defective, you (and not IBM) assume the entire cost of all necessary   **/
/**  servicing, repair or  correction.                                      **/
/**                                                                         **/
/**  IBM does not warrant that the contents of the source code examples     **/
/**  will meet your requirements or that the source code examples are       **/
/**  error-free.                                                            **/
/**                                                                         **/
/**  IBM may make improvements and/or changes in the source code examples   **/
/**  at any time.  Changes may be made periodically to the information in   **/
/**  the source code examples; these changes may be reported, for the       **/
/**  device drivers included herein, in new editions of the examples.       **/
/**                                                                         **/
/**  References in the source code examples to IBM products, programs, or   **/
/**  services do not imply that IBM intends to make these available in all  **/
/**  countries in which IBM operates.  Any reference to an IBM licensed     **/
/**  program in the source code examples is not intended to state or imply  **/
/**  that only IBM's licensed program may be used.  Any functionally        **/
/**  equivalent program may be used.                                        **/
/**                                                                         **/
/*****************************************************************************/

static const char *version = "IBMtapeutil for Linux, Version 1.4.6, Dec. 17, 2007";

#include "IBMtapeutil.h"


#define min(x,y)       ((x)<(y)?(x):(y))

/* use fgets() instead of gets() to get user input */
#define get_input(buf) fgets(buf,80,stdin);\
                       if(strlen(buf) < 1){ printf("Input error\n"); break; }\
                       buf[strlen(buf)-1] = '\0'


/*- name:      main
 *  function:  invoke command line mode or interative mode
 *  arguments: command line arguments
 *  exit code: if no error, exit 0; otherwise, exit non-zero
-*/
int main(int argc, char ** argv)
{
  int  rc;
  char buf[80];

  /* initialize global data */
  mt_st_mode = MTMODE;    /* by default, use MTIOCTOP ioctl for tape operations*/
  opcode     = 0;
  opened     = 0;
  mode_flags = 0;
  fd         = -1;
  bzero(devname, 80);
  bzero(pat, 256);
  memset(&mtpos, 0, sizeof(struct mtpos));
  memset(&stpos, 0, sizeof(struct stpos_s));
  memset(&parms, 0, sizeof(struct stchgp_s));

  if(argc == 1)
  {
    system("clear");
    device_menu();          /* Call the interactive version */
    if(exec_device != 0 )
      {
        system("clear");
        menu();
      }

    saved_errno = errno;

    /* close if device is open */
    if(opened)
      {
        rc = close(fd);
        if(rc != 0)
          {
            sprintf(buf,"Device closed with errno %d", errno);
            perror (buf);
          }
        else
          printf("Device %s closed...\n", devname);
      }

    if(saved_errno)
      exit(saved_errno);
    else
      exit(errno);
  }

  /* Otherwise assume that it is the cmdline interface version */
  rc = cmdline(argc, argv);     /* Command line version */

  if(saved_errno)
    exit(saved_errno);
  else
    exit(errno);
}

/*- name:     help
 *  function: show cmdline format
 *  argument: void
 *  return:   none
 -*/
void help()
{
  printf("\n");
  printf("Usage: IBMtapeutil [-f Device Subcommand [Subcommand ...]]\n");
  printf("\n");
  printf("%s\n", version);

  /* General subcommands */
  printf(" General Subcommands:\n");
  printf("   tur              inquiry [Page]   print \"Text\"   \n");
  printf("   reserve          release          reqsense         \n");
  printf("   qryversion       logpage \"Page\"   modepage \"Page\"\n");
  printf("   qrypath          enablepath \"primary\" | number\n");
  printf("   path             disablepath \"primary\" | number\n");

  /* Tape subcommands */
  printf(" Tape Subcommands:\n");
  printf("   bsf [Count]      bsr [Count]    eof [Count]\n");
  printf("   fsf [Count]      fsr [Count]    weof [Count]\n");
  printf("   fsfm [Count]     bsfm [Count]   asf [Count]\n");
  printf("   compress         tell           seek [Count]\n");
  printf("   nocompress       rewind         sync\n");
  printf("   load             erase          display \"Message\"\n");
  printf("   unload           retension      read -d Destination [-c Count]\n");
  printf("   qrypos           seod           write -s Source \n");
  printf("   setpos [Blockid] status         rtest  [-b Blocksize] [-c Count] [-r Repetition]\n");
  printf("   offline          parms          wtest  [-b Blocksize] [-c Count] [-r Repetition]\n");
  printf("   rewoffl          list           rwtest [-b Blocksize] [-c Count] [-r Repetition]\n");
  printf("   prevent          lock           setblk [Count]\n");
  printf("   allow            unlock         density\n");
  printf("   qryinquiry       qrysense       append         mtdevice   encryption\n");

  /* Medium changer subcommands */
  printf(" Medium Changer Subcommands:\n");
  printf("   allow         prevent          audit [Address [Count]]     \n");
  printf("   inventory     mount [Slot]     position \"Destination\"\n");
  printf("   elementinfo   unmount [Slot]   move \"Source\" \"Destination\"\n");
  printf("   devids        exchange \"Source\" \"Dest1\" \"Dest2\"\n");

  /* Service aid subcommands */
  printf(" Service Aid Subcommands:\n");
  printf("   dump [Filename]  forcedump     ucode \"Name\"   resetdrive\n\n");
}


/*- name:      cmdline
 *  function:  parse command line arguments
 *  arguments: command line arguments
 *  return:    0 if succeeded, -1 or -2 if failed
-*/
int cmdline(int argc,char ** argv)
{
  char *argp;
  char msg[80];
  int  rc           = 0;
  int  arg_count    = 1;
  int  st_count     = 1;
  int  read_mode    = FALSE;
  int  wrt_mode     = FALSE;
  int  append_mode  = FALSE;
  int  ndelay       = FALSE;
  int  number, page;
  int  i, j;

  /* help subcommand */
  if((strcmp(argv[1], "-?") == 0) ||
      (strcmp(argv[1], "?") == 0) )
    {
      help();
      exit(0);
    }

  /* check first arg for device */
  if(strncmp("-f",argv[1],2))
    {
      help();
      exit(1);
    }

  if(strlen(argv[1]) == 2)
    {
      arg_count++;
      if(arg_count >= argc)
        {
          printf("Missing argument for device\n");
          exit(1);
        }
      strcpy(devname, argv[2]);
    }
  else
    strcpy(devname, argv[1]+2);

  arg_count++;
  if(arg_count >= argc)
    {
      printf("Missing subcommand\n");
      exit(1);
    }

  /* find out open flags */
  for (i = arg_count; i < argc; i++)
    {
      if((strcmp(argv[i], "mount") == 0)     ||
          (strcmp(argv[i], "unmount") == 0)   ||
          (strcmp(argv[i], "demount") == 0)   ||
          (strcmp(argv[i], "position") == 0)  ||
          (strcmp(argv[i], "allow") == 0)     ||
          (strcmp(argv[i], "prevent") == 0)   ||
          (strcmp(argv[i], "exchange") == 0)  ||
          (strcmp(argv[i], "move") == 0)      ||
          (strcmp(argv[i], "lock") == 0)      ||
          (strcmp(argv[i], "unlock") == 0)    ||
          (strcmp(argv[i], "ucode") == 0)     ||
          (strcmp(argv[i], "dump") == 0)      ||
          (strcmp(argv[i], "forcedump") == 0) ||
          (strcmp(argv[i], "resetdrive") == 0))
        ndelay = TRUE;

      if((strcmp(argv[i], "rtest") == 0)   ||
          (strcmp(argv[i], "rwtest") == 0) ||
          (strcmp(argv[i], "list") == 0)   ||
          (strcmp(argv[i], "read") == 0))
        read_mode = TRUE;

      if((strcmp(argv[i], "wtest") == 0)   ||
          (strcmp(argv[i], "rwtest") == 0) ||
          (strcmp(argv[i], "write") == 0)  ||
          (strcmp(argv[i], "erase") == 0)  ||
          (strcmp(argv[i], "sync") == 0)   ||
          (strcmp(argv[i], "eof") == 0)    ||
          (strcmp(argv[i], "weof") == 0))
        wrt_mode = TRUE;

      if(strcmp(argv[i],"append") == 0)
        {
          wrt_mode = TRUE;
          append_mode = TRUE;
        }
    }


  if(read_mode && wrt_mode)
    mode_flags = O_RDWR;
  else if(read_mode)
    mode_flags = O_RDONLY;
  else if(wrt_mode)
    mode_flags = O_WRONLY;
  else
    mode_flags = O_NDELAY;

  if(append_mode)
    mode_flags |= O_APPEND;

  if(ndelay)
    mode_flags |= O_NDELAY;

  /* open device */
  fd = open(devname, mode_flags);
  if(fd < 0)
  {
    perror("Open failed");
    exit(errno);
  }

  /* get device type */
  if(strstr(devname, "changer") != NULL)
    exec_device = CHANGER;
  else
    exec_device = TAPE;

  /* execute subcommands */
  for (i = arg_count;; i++)
    {
      /* save errno before fflush call */
      saved_errno = errno;
      fflush(stdout);
      if(rc != 0)
        {
          sprintf(msg,"Operation failed with errno %d", saved_errno);
          perror (msg);
          if(fd >= 0 && rc == -1)
            {
              rc = sioc_reqsense();
              if(rc != 0)
                {
                  sprintf(msg,"Operation failed with errno %d", saved_errno);
                  perror (msg);
                }
            }
          if(ERR_STOP)
            break;
        }

      if(i == argc) break;

      errno = 0;
      rc = 0;
      argp = argv[i];

      /* help subcommand */
      if((strcmp(argp, "-?") == 0) ||
          (strcmp(argp, "-h") == 0) ||
          (strcmp(argp, "?") == 0))
        {
          help();
          continue;
        }

      /*------------------General Subcommands--------------------*/
      /* inquiry subcommand */
      if(strncmp(argp, "inquiry", 3) == 0)
        {
          /* check for optional page code */
          if((i + 1) < argc)
            {
              page = (int)strtol(argv[i+1],NULL,16);
              if(isdigit(argv[i+1][0]) || page)
                {
                  i++;
                  rc = sioc_inquiry_page(page);
                }
              else
                rc = sioc_inquiry();
            }
          else
            rc = sioc_inquiry();
          continue;
        }

      /* print subcommand */
      if(strcmp(argp, "print") == 0)
        {
          i++;
          if(i < argc)
            printf("%s\n",argv[i]);
          else
            {
              printf("Missing argument for print subcommand\n");
              exit(1);
            }
          continue;
        }

      /* reserve subcommand */
      if(strcmp(argp, "reserve") == 0)
        {
          rc = sioc_reserve();
          continue;
        }

      /* release subcommand */
      if(strcmp(argp, "release") == 0)
        {
          rc = sioc_release();
          continue;
        }

      /* request sense subcommand */
      if(strncmp(argp, "reqsense", 3) == 0)
        {
          rc = sioc_reqsense();
          continue;
        }

      /* log sense page subcommand */
      if(strcmp(argp, "logpage") == 0)
        {
          i++;
          if(i >= argc)
            {
              printf("Missing argument for logpage subcommand\n");
              exit(1);
            }
          page = (int)strtol(argv[i], NULL, 16);
          if(isdigit(argv[i][0]) || page)
            rc = sioc_log_sense_page(page, 0);
          else
            {
              printf("Invalid argument for logpage subcommand\n");
              exit(1);
            }
          continue;
        }

      /* mode sense page subcommand */
      if(strcmp(argp, "modepage") == 0)
        {
          i++;
          if(i >= argc)
            {
              printf("Missing argument for modepage subcommand\n");
              exit(1);
            }
          page = (int)strtol(argv[i],NULL,16);
          if(isdigit(argv[i][0]) || page)
            rc = sioc_mode_sense_page(page);
          else
            {
              printf("Invalid argument for modepage subcommand\n");
              exit(1);
            }
          continue;
        }

       /* test unit ready subcommand */
      if(strcmp(argp, "tur") == 0)
        {
          rc = sioc_test_unit_ready();
          continue;
        }


      /*--------------Medium Changer Subcommands----------------*/
      /* prevent medium removal subcommand */
      if(strcmp(argp, "prevent") == 0)
        {
          /* check if it's a medium changer or tape device */
          if(strstr(devname, "changer") != NULL)
            rc = smcioc_prevent_medium_removal();
          else
            rc = stioc_prevent_medium_removal();
          continue;
        }

      /* allow medium removal subcommand */
      if(strcmp(argp, "allow") == 0)
        {
          /* check if it's a medium changer or tape device */
          if(strstr(devname, "changer") != NULL)
            rc = smcioc_allow_medium_removal();
          else
            rc = stioc_allow_medium_removal();
          continue;
        }

      /* initialize element status subcommand */
      if(strcmp(argp, "audit") == 0)
        {
          /* check for audit range */
          if((i + 1) < argc && isdigit(argv[i+1][0]))
            {
              i++;
              source = atoi(argv[i]);
              if((i + 1) < argc && isdigit(argv[i+1][0]))
                {
                  i++;
                  st_count = atoi(argv[i]);
                }
              else
                st_count = 1;
              rc = smcioc_init_elem_stat_range(source, st_count);
            }
          else
            rc = smcioc_init_elem_stat();
          continue;
        }

      /* read element status subcommand */
      if(strncmp(argp, "inventory", 3) == 0)
        {
          rc = smcioc_inventory();
          continue;
        }

      /* mount subcommand */
      if(strcmp(argp, "mount") == 0)
        {
          if(i+1 >= argc || !isdigit(argv[i+1][0]))
            source = 0;
          else
            {
              source = atoi(argv[i+1]);
              i++;
            }
          rc = smcioc_load_medium(source);
          continue;
        }

      /* unmount subcommand */
      if(strcmp(argp, "unmount") == 0 || strcmp(argp, "demount") == 0)
        {
          if(i+1 >= argc || !isdigit(argv[i+1][0]))
            dest = 0;
          else
            {
              dest = atoi(argv[i+1]);
              i++;
            }
          rc = smcioc_unload_medium(dest);
          continue;
        }

      /* position to element subcommand */
      if(strncmp(argp, "position", 3) == 0)
        {
          i++;
          if(i >= argc)
            {
              printf("Missing argument for position subcommand\n");
              exit(1);
            }
          /* check for new format */
          if(isdigit(argv[i][0]))
            dest = atoi(argv[i]);
          else
            {
              argp = argv[i];
              if(strncmp("-d", argp, 2) != 0)  /* destination argument */
                {
                  printf("Missing argument for position subcommand\n");
                  exit(1);
                }
              if(strlen(argp) == 2)
                {
                  i++;
                  dest = atoi(argv[i]);
                }
              else
                dest = atoi(argv[i]+2);
            }
          rc = smcioc_pos_to_elem(dest);
          continue;
      }

      /* get element information subcommand */
      if(strncmp(argp, "elementinfo", 3) == 0)
        {
          rc = smcioc_element_info();
          continue;
        }

      /* move medium subcommand */
      if(strcmp(argp, "move") == 0)
        {
          i++;
          if(i >= argc)
            {
              printf("Missing argument for move subcommand\n");
              exit(1);
            }
          /* check for new format */
          if(isdigit(argv[i][0]))
            source = atoi(argv[i]);
          else
            {
              argp = argv[i];
              if(strncmp("-s", argp, 2) != 0)  /* source argument */
                {
                  printf("Missing argument for move subcommand\n");
                  exit(1);
                }
              if(strlen(argp) == 2)
                {
                  i++;
                  source = atoi(argv[i]);
                }
              else
                source = atoi(argv[i]+2);
            }
          i++;
          if(i >= argc)
            {
              printf("Missing argument for move subcommand\n");
              exit(1);
            }
          /* check for new format */
          if(isdigit(argv[i][0]))
            dest = atoi(argv[i]);
          else
            {
              argp = argv[i];
              if(strncmp("-d", argp, 2) != 0)  /* destination argument */
                {
                  printf("Missing argument for move subcommand\n");
                  exit(1);
                }
              if(strlen(argp) == 2)
                {
                  i++;
                  dest = atoi(argv[i]);
                }
              else
                dest = atoi(argv[i]+2);
            }
          rc = smcioc_move_medium(source,dest);
          continue;
        }

      /* read element device ids subcommand */
      if(strncmp(argp, "devids", 5) == 0)
        {
          rc = smcioc_read_element_devids();
          continue;
        }

      /* exchange medium subcommand */
      if(strcmp(argp, "exchange") == 0)
        {
          /* get source arg */
          if(((i + 1) < argc) && isdigit(argv[i+1][0]))
            {
              i++;
              source = atoi(argv[i]);
            }
          else
            {
              printf("Missing argument for exchange subcommand\n");
              exit(1);
            }
          /* get dest1 arg */
          if(((i + 1) < argc) && isdigit(argv[i+1][0]))
            {
              i++;
              dest = atoi(argv[i]);
            }
          else
          {
            printf("Missing argument for exchange subcommand\n");
            exit(1);
          }
          /* get dest2 arg */
          if(((i + 1) < argc) && isdigit(argv[i+1][0]))
            {
              i++;
              dest2 = atoi(argv[i]);
            }
          else
            {
              printf("Missing argument for exchange subcommand\n");
              exit(1);
            }
          rc = smcioc_exchange_medium(source,dest,dest2);
          continue;
        }

      /* query primary and first alternate paths */
      if(strncmp(argp, "qrypath", 7) == 0)
        {
          rc = sioc_query_path();
          continue;
        }

      /* query all paths */
      if(strncmp(argp, "path", 4) == 0)
        {
          rc = sioc_device_paths();
          continue;
        }

      /* enable path subcommand */
      if (strcmp(argp, "enablepath") == 0)
        {
          i++;
          if (i < argc)
            {
              if (strcmp(argv[i], "primary") == 0)
                rc = sioc_enable_path(PRIMARY_SCSI_PATH);
              else if (isdigit(argv[i][0]) && atoi(argv[i]) > 0)
                rc = sioc_enable_path(atoi(argv[i]));
              else
                {
                  printf("Invalid path argument for enablepath subcommand\n");
                  exit(1);
                }
            }
          else
            {
              printf("Missing argument for enablepath subcommand\n");
              exit(1);
            }
          continue;
        }

      /* disable path subcommand */
      if (strcmp(argp, "disablepath") == 0)
        {
          i++;
          if (i < argc)
            {
              if (strcmp(argv[i], "primary") == 0)
                rc = sioc_disable_path(PRIMARY_SCSI_PATH);
              else if (isdigit(argv[i][0]) && atoi(argv[i]) > 0)
                rc = sioc_disable_path(atoi(argv[i]));
              else
                {
                  printf("Invalid path argument for disablepath subcommand\n");
                  exit(1);
                }
            }
          else
            {
              printf("Missing argument for disablepath subcommand\n");
              exit(1);
            }
          continue;
        }

      /*--------------------Tape Subcommands---------------------*/

      /* append subcommand handled on open */
      if(strcmp(argp, "append") == 0)
        {
          printf("Open in append mode...\n");
          continue;
        }

      /* set compression subcommand */
      if(strcmp(argp, "compress") == 0)
        {
          printf("Setting compression ON...\n");
          rc = mtioctop(MTCOMPRESSION, 1);
          continue;
        }

      /* set no compression subcommand */
      if(strcmp(argp, "nocompress") == 0)
        {
          printf("Setting compression OFF...\n");
          rc = mtioctop(MTCOMPRESSION, 0);
          continue;
        }

      /* query position subcommand */
      if(strcmp(argp, "qrypos") == 0)
        {
          rc = stiocqrypos();
          continue;
        }

      /* tell block subcommand */
      if(strcmp(argp, "tell") == 0)
        {
          rc = mtioctop(MTTELL, 0);
          continue;
        }

      /* seek subcommand */
      if((strcmp(argp, "seek") == 0))
        {

          /* check for optional bockid arg in decimal */
          if(((i + 1) < argc) && isdigit(argv[i+1][0]))
            {
              i++;
              /* set the global position structure used for qrypos/setpos defaults */
              mtpos.mt_blkno = atoi(argv[i]);
            }
          /* check for optional bockid arg in hex */
          else if(((i + 1) < argc) && argv[i+1][0] == 'x')
            {
              i++;
              /* set the global position structure used for qrypos/setpos defaults */
              mtpos.mt_blkno = (int)strtol(argv[i]+1, NULL, 16);
            }

          rc = mtiocsetpos();
          if(rc == 0)
            rc = mtiocqrypos();

          continue;
        }

       /* setpos subcommand */
      if((strcmp(argp, "setpos") == 0))
        {

          /* check for optional bockid arg in decimal */
          if(((i + 1) < argc) && isdigit(argv[i+1][0]))
            {
              i++;
              /* set the global position structure used for qrypos/setpos defaults */
              stpos.curpos = atoi(argv[i]);
              stpos.block_type = QP_LOGICAL;
            }
          /* check for optional bockid arg in hex */
          else if(((i + 1) < argc) && argv[i+1][0] == 'x')
            {
              i++;
              /* set the global position structure used for qrypos/setpos defaults */
              stpos.curpos = (int)strtol(argv[i]+1, NULL, 16);
              stpos.block_type = QP_PHYSICAL;
            }

          rc = stiocsetpos();
          if(rc == 0)
            rc = stiocqrypos();

          continue;
        }


      /* get mtdevice number subcommand */
      if(strcmp(argp, "mtdevice") == 0)
        {
          rc = mtdevice();
          continue;
        }

      /* get status subcommand */
      if(strcmp(argp, "status") == 0)
        {
          rc = status();
          continue;
        }

      /* get/set tape parameters subcommand */
      if(strcmp(argp, "parms") == 0)
        {
          rc = tape_parms();
          continue;
        }

      /* synchronize buffers subcommand */
      if(strcmp(argp, "sync") == 0)
        {
          rc = stiocsync();
          continue;
        }

      /* load tape subcommand */
      if(strcmp(argp, "load") == 0)
        {
          rc = mtioctop(MTLOAD, 0);
          continue;
        }

      /* offline tape subcommand */
      if(strcmp(argp, "offline") == 0  ||
          strcmp(argp, "rewoffl") == 0)
        {
          rc = mtioctop(MTOFFL,0);
          continue;
        }

      /* unload tape subcommand */
      if(strcmp(argp, "unload") == 0)
        {
          mtioctop(MTUNLOCK, 0); /* unlock the drive door first */
          rc = mtioctop(MTUNLOAD, 0);
          continue;
        }

      /* rewind tape subcommand */
      if(strcmp(argp, "rewind") == 0 )
        {
          rc = mtioctop(MTREW,0);
          continue;
        }

      /* retension tape subcommand */
      if(strcmp(argp, "retension") == 0 )
        {
          rc = mtioctop(MTRETEN, 0);
          continue;
        }

      /* erase tape subcommand */
      if(strcmp(argp, "erase") == 0)
        {
          rc = mtioctop(MTERASE,0);
          continue;
        }

      /* space to end-of-data subcommand */
      if(strcmp(argp, "eom") == 0
          || strcmp(argp, "seod") == 0 )
        {
          rc = mtioctop(MTEOM,0);
          continue;
        }

      /* setblock size subcommand */
      if(strcmp(argp, "setblk") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if((st_count > 0) || (strcmp(argv[i+1], "0") == 0))
                i++;
              else
                st_count = 0;
            }
          else
            st_count = 0;

          rc = mtioctop(MTSETBLK, st_count);
          continue;
        }

      /* write filemarks subcommand */
      if(strcmp(argp, "eof") == 0  ||
          strcmp(argp, "weof") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;
          rc = mtioctop(MTWEOF,st_count);
          continue;
        }


      /* forward space filemarks subcommand */
      if(strcmp(argp, "fsf") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;
          rc = mtioctop(MTFSF,st_count);
          continue;
        }

      /* backward space filemarks subcommand */
      if(strcmp(argp, "bsf") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;
          rc = mtioctop(MTBSF,st_count);
          continue;
        }

      /* forward space records subcommand */
      if(strcmp(argp, "fsr") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;

          rc = mtioctop(MTFSR, st_count);
          continue;
        }

      /* backward space records subcommand */
      if(strcmp(argp, "bsr") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;
          rc = mtioctop(MTBSR,st_count);
          continue;
        }

      /* forward space filemarks and position tape
       * to the last block of previous file subcommand */
      if(strcmp(argp, "fsfm") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;
          rc = mtioctop(MTFSFM, st_count);
          continue;
        }

      /* backward space filemarks and position tape
       * to the first block of next file subcommand */
      if(strcmp(argp, "bsfm") == 0)
        {
          if((i + 1) < argc)
            {
              st_count = atoi(argv[i+1]);
              if(st_count > 0)
                i++;
              else
                st_count = 1;
            }
          else
            st_count = 1;
          rc = mtioctop(MTBSFM, st_count);
          continue;
        }

      /* set tape position subcommand */
      if(strcmp(argp, "asf") == 0)
        {
          rc = mtioctop(MTREW, 0);  /* rewind tape first */
          if(rc == 0)
            {
              if((i + 1) < argc)
                {
                  st_count = atoi(argv[i+1]);
                  if(st_count > 0)
                    i++;
                  else
                    st_count = 1;
                }
              else
                st_count = 1;
              rc = mtioctop(MTFSF, st_count);
            }
          continue;
        }

      /* read and write tests */
      if(strcmp(argp, "wtest") == 0 ||
          strcmp(argp, "rtest") == 0 ||
          strcmp(argp, "rwtest") == 0)
        {
          blksize = 10240;
          blocks = 20;
          repetition = 1;
          /* look for r/w options */
          rc = 0;
          for (j = 0; (j < 3) && ((i + 1) < argc); j++)
            {
              if(strncmp("-b",argv[i+1],2) == 0)  /* blocksize argument */
                {
                  i++;
                  if(strlen(argv[i]) == 2)
                    {
                      i++;
                      if(i >= argc)
                        {
                          printf("Missing argument for blocksize option\n");
                          exit(1);
                        }
                      blksize = atoi(argv[i]);
                    }
                  else
                    blksize = atoi(argv[i]+2);
                  continue;
                }
              if(strncmp("-c", argv[i+1], 2) == 0)  /* count argument */
                {
                  i++;
                  if(strlen(argv[i]) == 2)
                    {
                      i++;
                      if(i >= argc)
                        {
                          printf("Missing argument for count option\n");
                          exit(1);
                        }
                      blocks = atoi(argv[i]);
                    }
                  else
                    blocks = atoi(argv[i]+2);
                  continue;
                }
              if(strncmp("-r", argv[i+1], 2) == 0)  /* repetition argument */
                {
                  i++;
                  if(strlen(argv[i]) == 2)
                    {
                      i++;
                      if(i >= argc)
                        {
                          printf("Missing argument for repetition option\n");
                          exit(1);
                        }
                      repetition = atoi(argv[i]);
                    }
                  else
                    repetition = atoi(argv[i]+2);
                  continue;
                }
              break;
            }
          if(rc)
            continue;
          if(strcmp(argp, "wtest") == 0)
            rc = wtest(blksize, blocks, repetition);
          else if(strcmp(argp, "rtest") == 0)
            rc = rtest(blksize, blocks, repetition);
          else
            rc = rwtest(blksize, blocks, repetition);
          continue;
        }

      /* write  */
      if(strcmp(argp, "write") == 0)
        {
          i++;
          if(i >= argc)
            {
              printf("Missing argument for write subcommand\n");
          exit(1);
            }
          argp = argv[i];
          if(strncmp("-s", argp, 2) != 0)  /* source argument */
            {
              printf("Missing argument for write subcommand\n");
              exit(1);
            }
          if(strlen(argp) == 2)
            {
              i++;
              sname = argv[i];
            }
          else
            sname = argv[i]+2;
          rc = twrite(sname);
          continue;
        }

      /* read  */
      if(strcmp(argp, "read") == 0)
        {
          i++;
          if(i >= argc)
            {
              printf("Missing argument for read subcommand\n");
              exit(1);
            }
          argp = argv[i];
          if(strncmp("-d", argp, 2) != 0)  /* destination argument */
            {
              printf("Missing argument for read subcommand\n");
              exit(1);
            }
          if(strlen(argp) == 2)
            {
              i++;
              dname = argv[i];
            }
          else
            dname = argv[i]+2;
          /* check for optional -c count  */
          if(((i + 1) < argc) && strncmp("-c",argv[i+1],2) == 0)
            {
              i++;
              if(strlen(argv[i]) == 2)
                {
                  i++;
                  if(i >= argc)
                    {
                      printf("Missing argument for count option\n");
                      exit(1);
                    }
                  number = atoi(argv[i]);
                }
              else
                number = atoi(argv[i]+2);
              if(number == 0)
                {
                  printf("Invalid argument for count option\n");
                  exit(1);
                }
            }
          else
            number = 0;
          rc = tread(dname, number);
          continue;
        }

      /* list subcommand */
      if(strcmp(argp, "list") == 0)
        {
          rc = list_tape();
          continue;
        }

      /* density subcommand */
      if(strcmp(argp, "density") == 0)
        {
          rc = stioc_report_density_support();
          continue;
        }


      /* display message subcommand */
      if(strcmp(argp, "display") == 0)
        {
          i++;
          if(i < argc)
            {
              if(strlen(argv[i]) > 8)
                rc = stiocdm(5, argv[i], argv[i]+8);
              else
                rc = stiocdm(1, argv[i], NULL);
            }
          else
            {
              printf("Missing argument for display subcommand\n");
              exit(1);
            }
          continue;
        }

      /* lock subcommand */
      if(strcmp(argp, "lock") == 0)
        {
          rc = mtioctop(MTLOCK, 0);
          continue;
        }

      /* unlock subcommand */
      if(strcmp(argp, "unlock") == 0)
        {
          rc = mtioctop(MTUNLOCK, 0);
          continue;
        }

      /* qryinquiry subcommand */
      if(strcmp(argp, "qryinquiry") == 0)
        {
          rc = stiocqryinquiry();
          continue;
        }

      /* qrysense subcommand */
      if(strcmp(argp, "qrysense") == 0)
        {
          rc = stiocqrysense();
          continue;
        }

      /* query driver version subcommand */
      if(strcmp(argp, "qryversion") == 0)
        {
          rc = stioc_qry_driver_version();
          continue;
        }

      /* get encryption state subcommand */
      if(strcmp(argp, "encryption") == 0)
        {
          rc = qry_encryption_status();
          continue;
        }

      /* get serial number subcommand */
      if(strcmp(argp, "sernum") == 0)
        {
          rc = get_sernum();
          continue;
        }

      /* rekey subcommand */
      if(strcmp(argp, "rekey") == 0)
        {
          rc = rekey_command();
          continue;
        }

      /*----------------Service Aid Subcommands-----------------*/

      /* read dump subcommand */
      if(strcmp(argp, "dump") == 0)
        {
          int id = search_drive_id();

          /* if another arg, assume it's a dump filename */
          if((i + 1) < argc)
            {
              i++;
              if(id == IBM_3590 && exec_device == CHANGER)
                printf("dump is not executed for IBM3590 LUN2.\n");
              else
                rc = stioc_read_dump(argv[i]);
            }/* use user specified file name */
          else
            {
              if(id == IBM_3590 && exec_device == CHANGER)
                printf("dump is not executed for IBM3590 LUN2.\n");
              else
                {
                  printf("dump file is dump0001.dmp.\n");
                  rc = stioc_read_dump("dump0001.dmp");
                }
            }/* use default file name */
          continue;
        }

      /* force dump subcommand */
      if(strcmp(argp, "forcedump") == 0)
        {
          if(search_drive_id() == IBM_3590 && exec_device == CHANGER)
              printf("forcedump is not executed for IBM3590 LUN2.\n");
          else
            rc = stioc_force_dump();
          continue;
        }

      /* reset drive subcommand */
      if(strcmp(argp, "resetdrive") == 0)
        {
          rc = stioc_reset_drive_Atape();
          continue;
        }

      /* load ucode subcommand */
      if(strcmp(argp, "ucode") == 0)
        {
          int id = search_drive_id();
          i++;
          if(i < argc)
            {
              if(id == IBM_3590 && exec_device == CHANGER)
                printf("ucode is not executed for IBM3590 LUN2.\n");
              else
                rc = stioc_load_ucode(argv[i]);
            }
          else
            {
              if(id == IBM_3590 && exec_device == CHANGER)
                printf("ucode is not executed for IBM3590 LUN2.\n");
              else
                printf("Missing argument for ucode subcommand\n");
              exit(1);
            }
          continue;
        }

      printf("Invalid argument: %s\n", argp);
      exit(1);
    } /* end of for loop */

  close(fd);
  return rc;
}


/*- name:      device_menu
 *  function:  let user choose a Tape/Changer device; set global exec_device
 *  arguments: void
 *  return:    none
 -*/
void device_menu()
{
  char buf[10] = "";

  printf("\n%s\n", version);
  printf("=========================================\n");
  printf("         1. Tape\n");
  printf("         2. Changer\n");
  printf("         3. Quit\n");
  printf("=========================================\n\n");

  printf("Enter your choice: ");
  fgets(buf, 10, stdin);
  buf[strlen(buf)-1] = '\0';
  if(strlen(buf) == 0)
    return;

  exec_device = atoi(buf);
  if(exec_device != 1 && exec_device != 2)
    exec_device = 0;
}

/*- name:     menu
 *  function: interative with user
 *  argument: void
 *  return:   none
 -*/
void menu()
{
  char   buf[80]   = "";
  char   fname[80] = "";
  static int pass  = 0;
  int    rc        = 0;
  int    count, option;
  int    page, parm;
  struct element_info element_info;

  while(1)
  {
    if(rc != 0)
    {
      sprintf(buf,"Operation failed with errno %d", errno);
      perror(buf);
      saved_errno = errno;
      if(fd >= 0 && rc == -1)
        {
          rc = sioc_reqsense();
          if(rc != 0)
           {
             sprintf(buf, "Operation failed with errno %d",errno);
             perror (buf);
           }
        }
    }

    if(exec_device == 0)
      return;

    errno = 0;
    rc = 0;
    if(pass)
    {
      if(opcode != 99)
        {
          printf("\nHit <enter> to continue...");
          get_input(buf);
        }
    }
    else
      pass++;

    printf("\n-------------------------- General Commands: -----------------------------\n");
    printf("    1. Open a Device                  9. Mode Sense Pag\n");
    printf("    2. Close a Device                10. Switch Tape/Changer Devic\n");
    printf("    3. Inquiry                       11. Create Special Files\n");
    printf("    4. Test Unit Ready               12. Query Driver Version\n");
    printf("    5. Reserve Device                13. Query Device Path\n");
    printf("    6. Release Device                14. Display All Paths\n");
    printf("    7. Request Sense                 15. Enable A Path\n");
    printf("    8. Log Sense Page                16. Disable A Path\n");
    printf("    Q. Quit IBMtapeutil\n");
    if(exec_device == TAPE)
      {
        printf("--------------------------- Tape Commands: --------------------------------\n");
        printf("   20. Rewind                        33. Set Block Size\n");
        printf("   21. Forward Space Filemarks       34. Retension Tape\n");
        printf("   22. Backward Space Filemarks      35. Query/Set Tape Position\n");
        printf("   23. Forward Space Records         36. Query Tape Status\n");
        printf("   24. Backward Space Records        37. Load Tape\n");
        printf("   25. FSFM                          38. Unload Tape\n");
        printf("   26. BSFM                          39. Lock Tape Drive Door\n");
        printf("   27. Space to End of Data          40. Unlock Tape Drive Door\n");
        printf("   28. Read and Write Tests          41. Take Tape Offline\n");
        printf("   29. Write Filemarks               42. Enable/Disable Compression\n");
        printf("   30. Read or Write Files           43. Flush Drive's Buffer\n");
        printf("   31. Erase                         44. Self Test\n");
        printf("   32. Reset Drive                   45. Display Message\n");
        printf("--------------------------- IBMtape Commands: -------------------------------\n");
        printf("   46. Query Sense                   53. Read Tape Position\n");
        printf("   47. Query Inquiry                 54. Query Mtdevice Number\n");
        printf("   48. Query/Set Tape Parameters     55. Synchronize Buffers\n");
        printf("   49. Query/Set Tape Position       56. List Tape Filemarks\n");
        printf("   50. Query/Set MT/ST Mode          57. Query Encryption Status\n");
        printf("   51. Report Density Support        58. EKM Test\n");
        printf("   52. Locate Tape Position\n");
        printf("--------------------------- Service Aid Commands: -------------------------\n");
        printf("   70. Dump Device                   74. Read Persistent Reservation Keys\n");
        printf("   71. Force Dump                    75. Read Persistent Reservations\n");
        printf("   72. Load Ucode                    76. Register Persistent Reservation Key\n");
        printf("   73. Reset Drive                   77. Remove Persistent Reservation\n");
        printf("                                     78. Clear All Persistent Registrations\n");
        printf("---------------------------------------------------------------------------\n");
        printf("   99. Back To Main Menu\n");
      }
    else /* it is a changer device */
      {
        printf("------------------------- Medium Changer Commands: -----------------------\n");
        printf("   60. Element Information           65. Load/Unload Medium\n");
        printf("   61. Position To Element           66. Initialize Element Status\n");
        printf("   62. Element Inventory             67. Prevent/Allow Medium Removal\n");
        printf("   63. Exchange Medium               68. Initialize Element Status Range\n");
        printf("   64. Move Medium                   69. Read Device Identifiers\n");
        printf("-------------------------- Service Aid Commands: -------------------------\n");
        printf("   70. Dump Device                   72. Load Ucode\n");
        printf("   71. Force Dump                    73. Reset Drive\n");
        printf("------------------- Additional Medium Changer Commands: ------------------\n");
        printf("   80. Read Cartridge Location\n");
        printf("--------------------------------------------------------------------------\n");
        printf("   99. Back To Main Menu\n");
      }

    printf("\n");
    if(opened)
      printf("Enter Selection for %s: ", devname);
    else
      printf("Enter Selection: ");
    fflush(stdin);
    get_input(buf);
    printf("\n");
    fflush(stdout);

    /* get a valid input */
    if(*buf == '\0')
      {
        printf("Please enter a selection before hit <enter>.\n");
        continue;
      }

    if(*buf == 'Q' || *buf == 'q')
      return;

    opcode = atoi(buf);

    /* require open to be first command */
    if(!opened && opcode != 1 && opcode != 10
        && opcode != 11 && opcode != 99)
    {
      printf("Nothing is opened yet.\n");
      continue;
    }

    switch (opcode)
      {
      /* General Commands */
      case 1: /* open a device */
        if(opened)
          {
            printf("There is already an open file: %s\n", devname);
            break;
          }

        if(exec_device == TAPE)
          printf("Enter device name (<enter> for /dev/IBMtape0): ");
        else
          printf("Enter device name (<enter> for /dev/IBMchanger0): ");

        get_input(devname);

        if(strlen(devname) == 0)
          {
            if(exec_device == TAPE)
              strcpy(devname, "/dev/IBMtape0");
            else
              strcpy(devname, "/dev/IBMchanger0");
          }

        /* make sure device type and devname match */
        if((strstr(devname, "changer") == NULL) && exec_device == CHANGER)
          {
            printf("Please open a medium changer device.\n");
            break;
          }
        if((strstr(devname, "changer") != NULL) && exec_device == TAPE)
          {
            printf("Please open a tape device.\n");
            break;
          }

        /* get open flags */
        mode_flags = O_NDELAY;
        if(exec_device == TAPE)
          {
            printf("Select mode (<enter> or 1=Read/Write, 2=Read Only, 3=Write Only, 4=Append): ");
            get_input(buf);
            if(strlen(buf) == 0)
              mode_flags |= O_RDWR;
            else
              {
                option = atoi(buf);
                if(option == 1)
                  mode_flags |= O_RDWR;
                else if(option == 2)
                  mode_flags |= O_RDONLY;
                else if(option == 3)
                  mode_flags |= O_WRONLY;
                else if(option == 4)
                  mode_flags |= O_WRONLY | O_APPEND;
                else
                  {
                    printf("Please make a selection on open mode.\n");
                    break;
                  }
              }
            printf("\n");
          }

        printf("Opening device...\n");
        fd = open(devname, mode_flags);
        if(fd < 0)
          rc = fd;
        else
          opened = 1;
        break;

      case 2: /* close a device */
        rc = close(fd);
        if(rc != 0)
        {
          sprintf(buf,"Device closed with errno %d", errno);
          perror (buf);
        }
        else
          printf("Device closed...\n");

        /* reinitialize globals */
        opened     = 0;
        mode_flags = 0;
        fd         = -1;
        bzero(devname, 80);
        bzero(pat, 256);
        memset(&mtpos, 0, sizeof(struct mtpos));
        memset(&stpos, 0, sizeof(struct stpos_s));
        memset(&parms, 0, sizeof(struct stchgp_s));
        break;

      case 3: /* inquiry */
        printf("\nEnter page code in hex or <enter> for standard inquiry: ");
        get_input(buf);
        if(strlen(buf) == 0)
          rc = sioc_inquiry();
        else
          {
            page = (int)strtol(buf, NULL, 16);
            rc = sioc_inquiry_page(page);
          }
        break;

      case 4: /* test unit ready */
        if(mt_st_mode == STMODE)
          {
            rc = stioctop(STTUR, 0);
            if(rc == 0)
              printf("Unit ready.\n");
          }
        else
          rc = sioc_test_unit_ready();
        break;

      case 5: /* reserve */
        rc = sioc_reserve();
        break;

      case 6: /* release device */
        rc = sioc_release();
        break;

      case 7: /* request sense */
        rc = sioc_reqsense();
        break;

      case 8:
         /* log sense page */
        printf("\nEnter page code in hex (e.g. 0x02 or <enter> for default 0): ");
        get_input(buf);
        if(strlen(buf) == 0)
          page = 0;
        else
          page = (int)strtol(buf, NULL, 16);
        printf("\nEnter parameter pointer in hex or <enter> for all parameters: ");
        get_input(buf);
        parm = (int)strtol(buf, NULL, 16);

        rc = sioc_log_sense_page(page, parm);
        break;

      case 9: /* mode sense page */
        printf("\nEnter page code in hex (e.g. 0x01 or <enter> for default 0): ");
        get_input(buf);
        if(strlen(buf) == 0)
          page = 0;
        else
          page = (int)strtol(buf, NULL, 16);

        rc = sioc_mode_sense_page(page);
        break;

      case 10: /* switch tape/changer device */
        if(opened)
          {
            printf("closing device %s\n", devname);
            if(close(fd) < 0)
              {
                sprintf(buf,"Device %s closed with errno %d", devname, errno);
                perror (buf);
              }
            else
              printf("Device %s closed...\n", devname);

            /* reinitialize globals */
            opened     = 0;
            mode_flags = 0;
            fd         = -1;
            bzero(devname, 80);
            bzero(pat, 256);
            memset(&mtpos, 0, sizeof(struct mtpos));
            memset(&stpos, 0, sizeof(struct stpos_s));
            memset(&parms, 0, sizeof(struct stchgp_s));
          }

        if(exec_device == TAPE)
          {
            printf("Switching to changer device...\n");
            exec_device = CHANGER;
          }
        else
          {
            printf("Switching to tape device...\n");
            exec_device = TAPE;
          }

        system("clear");
        break;

      case 11: /* create tape and changer special files */
        printf("This operation requires root authority.\n");
        printf("Continue? (y for YES, n for NO): ");
        get_input(buf);
        if(buf[0] != 'y' && buf[0] != 'Y')
          break;

        if(opened)
          {
            rc = close(fd);
            if(rc < 0)
              {
                sprintf(buf,"Device closed with errno %d",errno);
                perror (buf);
              }
            else
              printf("Device %s closed...\n", devname);

            /* reinitialize globals */
            opened     = 0;
            mode_flags = 0;
            fd         = -1;
            bzero(devname, 80);
            bzero(pat, 256);
            memset(&mtpos, 0, sizeof(struct mtpos));
            memset(&stpos, 0, sizeof(struct stpos_s));
            memset(&parms, 0, sizeof(struct stchgp_s));
          }

        rc = system("IBMtapeconfig");
        if(rc != 0)
          {
            printf("Create special files failed.\n");
            errno = 999;
          }
        break;

      case 12: /* query driver version */
        rc = stioc_qry_driver_version();
        break;

      case 13: /* query primary and first alternate paths */
        rc = sioc_query_path();
        break;

      case 14: /* query all device paths */
        rc = sioc_device_paths();
        break;

      case 15: /* enable a path */
        printf("Enter path number(hit <enter> for primary path): ");
        get_input(buf);
        printf("\n");
        if(strlen(buf) == 0)
          rc = sioc_enable_path(PRIMARY_SCSI_PATH);
        else if(isnumber(buf) && atoi(buf) > 0)
          rc = sioc_enable_path(atoi(buf));
        else
          {
            printf("Invalid path argument for enable path subcommand\n");
          }
        break;

      case 16: /* disable a path */
        printf("Enter path number(hit <enter> for primary path): ");
        get_input(buf);
        printf("\n");
        if(strlen(buf) == 0)
          rc = sioc_disable_path(PRIMARY_SCSI_PATH);
        else if(isnumber(buf) && atoi(buf) > 0)
          rc = sioc_disable_path(atoi(buf));
        else
          {
            printf("Invalid path argument for disable path subcommand\n");
          }
        break;

      /* Tape Commands */
      case 20: /* rewind */
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTREW,0);
        else
          rc = stioctop(STREW, 0);
        break;

      case 21: /* forward space filemarks */
        printf("Enter number of filemarks to space forward: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);

        if(count > 0)
          {
            if(mt_st_mode == MTMODE)
              rc = mtioctop(MTFSF, count);
            else
              rc = stioctop(STFSF, count);
          }
        else
          printf("Operation is not executed.\n");
        break;

      case 22: /* backward space filemarks */
        printf("Enter number of filemarks to space backward: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);

        if(count > 0)
          {
            if(mt_st_mode == MTMODE)
              rc = mtioctop(MTBSF, count);
            else
              rc = stioctop(STRSF, count);
          }
        else
          printf("Operation is not executed.\n");
        break;

      case 23: /* forward space records */
        printf("Enter number of records to space forward: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);

        if(count > 0)
          {
            if(mt_st_mode == MTMODE)
              rc = mtioctop(MTFSR, count);
            else
              rc = stioctop(STFSR, count);
          }
        else
          printf("Operation is not executed.\n");
        break;

      case 24: /* backward space records */
        printf("Enter number of records to space backward: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);

        if(count > 0)
          {
            if(mt_st_mode == MTMODE)
              rc = mtioctop(MTBSR, count);
            else
              rc = stioctop(STRSR, count);
          }
        else
          printf("Operation is not executed.\n");
        break;

      case 25: /* FSFM */
        printf("Enter number of filemarks to space forward: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);
        if(count > 0)
          rc = mtioctop(MTFSFM, count);
        else
          printf("Operation is not executed.\n");
        break;

      case 26: /* BSFM */
        printf("Enter number of filemarks to space backward: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);
        if(count > 0)
          rc = mtioctop(MTBSFM, count);
        else
          printf("Operation is not executed.\n");
        break;

      case 27: /* space to end of data */
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTEOM,0);
        else
          rc = stioctop(STSEOD, 0);
        break;

      case 28: /* read/write tests */
        printf("\nSelect test (1=Read/Write, 2=Read Only, 3=Write Only): ");
        get_input(buf);
        option = atoi(buf);

        if(option != 1 && option != 2 && option != 3)
          {
            printf("Not a valid selection.\n");
            break;
          }

        /* Read Only assumes that user has done the write test and rewined tape */
        if(option == 2)
          {
            printf("\n Read Only option assumes that you have done option 1 or 3\n");
            printf(" and rewinded tape\n");
            printf(" Do you want to continue?(hit <enter> or 'y' for YES, 'n' for NO): ");
            get_input(buf);
            if(buf[0] == 'n' || buf[0] == 'N' || buf[0] == 'q' || buf[0] == 'Q')
              break;
          }

        printf("\nHit <Enter> for default 10240 block size\n");
        printf("or enter new block size: ");
        get_input(buf);
        if(strlen(buf) == 0)
          blksize = 10240;
        else if (isnumber(buf))
          blksize = atoi(buf);
        else
          {
            printf("Not a valid number.\n");
            break;
          }

        if(blksize == 0)
        {
          /* get maximum transfer size */
          rc = stiocqryp();
          if(rc != 0)
            break;
          blocks = min(parms.max_scsi_xfer, MAX_TRANSFER_SIZE);
          printf("\nHit <Enter> for default %d bytes per read/write\n", blocks);
          printf("or enter new bytes: ");
          get_input(buf);
          if(strlen(buf) != 0)
            blocks = atoi(buf);
          if(blocks <= 0)
            {
              printf("Not a valid number.\n");
              break;
            }
        }
        else
        {
          printf("\nHit <Enter> for default 20 blocks per read/write\n");
          printf("or enter new blocks: ");
          get_input(buf);
          if(strlen(buf) == 0)
            blocks = 20;
          else if(isnumber(buf))
            blocks = atoi(buf);
          else
            {
              printf("Not a valid number.\n");
              break;
            }
          if(blocks == 0)
            break;
        }

        printf("\nHit <Enter> for default 1 repetition\n");
        printf("or enter new repetition: ");
        get_input(buf);
        if(strlen(buf) == 0)
          repetition = 1;
        else if(isnumber(buf))
          repetition = atoi(buf);
        else
          {
            printf("Not a valid number.\n");
            break;
          }
        if(repetition == 0)
          break;
        printf("\n");

        if(option == 1)
          rc = rwtest(blksize, blocks, repetition);
        else if(option == 2)
          rc = rtest(blksize, blocks, repetition);
        else if(option == 3)
          rc = wtest(blksize, blocks, repetition);
        break;

      case 29: /* write filemarks */
        printf("Enter number of filemarks to write: ");
        get_input(buf);
        printf("\n");
        count = atoi(buf);

        if(count > 0)
          {
            if(mt_st_mode == MTMODE)
              rc = mtioctop(MTWEOF, count);
            else
              rc = stioctop(STWEOF, count);
          }
        else
          printf("Operation is not executed.\n");
        break;

      case 30: /* read/write files */
        printf("\nSelect (1=Read file from tape, 2=Write file to tape): ");
        get_input(buf);
        option = atoi(buf);
        if(option == 1)
          {
            printf("Enter destination file: ");
            get_input(fname);
            if(strlen(fname) == 0)
              {
                printf("No file name entered.\n");
                break;
              }
            printf("Enter number of records to read or <enter> for entire file: ");
            get_input(buf);
            if(strlen(buf) != 0)
              {
                count = atoi(buf);
                if(count <= 0)
                  {
                    printf("Not a valid number.\n");
                    break;
                  }
              }
            else
              count = 0;
           rc = tread(fname, count);
          }
        else if(option == 2)
          {
            printf("Enter Source File: ");
            get_input(buf);
            if(strlen(buf) == 0)
              {
                printf("No file name entered.\n");
                break;
              }
            rc = twrite(buf);
          }
        else
          printf("Not a valid selection.\n");

        break;

      case 31: /* erase */
        printf("Do you really want to erase the tape?\n");
        printf("Hit 'y' for YES, 'n' for NO: ");
        get_input(buf);
        if(strlen(buf) == 0)
          break;
        if(buf[0] != 'y' && buf[0] != 'Y')
          break;

        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTERASE, 0);
        else
          rc = stioctop(STERASE, 0);
        break;

      case 32: /* reset drive */
        rc = mtioctop(MTRESET,0);
        break;

      case 33: /* set block size */
        printf("Enter block size to set: ");
        get_input(buf);
        printf("\n");
        if(isnumber(buf))
          count = atoi(buf);
        else
          {
            printf("Not a valid number.\n");
            break;
          }
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTSETBLK, count);
        else
          {
            struct stchgp stchgp_blk;
            memset(&stchgp_blk, 0, sizeof(struct stchgp));
            stchgp_blk.st_blksize = count;
            printf("Setting block size to %d...\n", count);
            rc = ioctl(fd, STIOCHGP, &stchgp_blk);
          }
        break;
      case 34: /* retension tape */
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTRETEN, 0);
        else
          rc = stioctop(STRETEN, 0);
        break;

      case 35: /* query/set block position */

        rc = mtiocqrypos();
        if(rc != 0)
          break;

        printf("\nEnter position to set (1=Logical, 2=Physical, <enter>=Quit): ");
        get_input(buf);
        if(strlen(buf) == 0)
          break;
        else if(isnumber(buf))
          option = atoi(buf);
        else
          {
             printf("Not a valid selection.\n");
             break;
          }
        if(option != 1 && option != 2)
          {
             printf("Not a valid selection.\n");
             break;
          }

        printf("\nEnter Block ID in decimal or hex (Example: 16 or 0x10): ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No block id entered.\n");
            break;
          }
        if(buf[0] == 'x')
          mtpos.mt_blkno = (int)strtol(buf+1, NULL, 16);
        else if(buf[0] == '0' && buf[1] == 'x')
          mtpos.mt_blkno = (int)strtol(buf, NULL, 16);
        else
          mtpos.mt_blkno = atoi(buf);

        rc = mtiocsetpos();
        if(rc == 0)
          rc = mtiocqrypos();
        break;

      case 36: /* query tape status */
        rc = status( );
        break;

      case 37: /* load tape */
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTLOAD, 0);
        else
          rc = stioctop(STLOAD, 0);
        break;

      case 38: /* unload tape */
        mtioctop(MTUNLOCK, 0); /* unlock the drive door first */
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTUNLOAD, 0);
        else
          rc = stioctop(STEJECT, 0);
        break;

      case 39: /* lock tape drive door */
        rc = mtioctop(MTLOCK, 0);
        break;

      case 40: /* unlock tape drive door */
        rc = mtioctop(MTUNLOCK, 0);
        break;

      case 41: /* offline tape */
        if(mt_st_mode == MTMODE)
          rc = mtioctop(MTOFFL, 0);
        else
          rc = stioctop(STOFFL, 0);
        break;

      case 42: /* enable/disable compression */
        printf("Select operation (1=enable, 2=disable, <enter>=Quit): ");
        get_input(buf);
        printf("\n");
        if(strlen(buf) == 0)
          break;

        count = atoi(buf);
        if(count == 2)
          count = 0;
        else if(count != 1)
          {
            printf("Not a valid selection.\n");
            break;
          }
        rc = mtioctop(MTCOMPRESSION, count);
        break;

      case 43: /* flush buffer */
        rc = mtioctop(MTNOP, 0);
        break;

      case 44: /* self test */
        rc = mtioctop(MTRAS1, 0);
        break;

      case 45: /* display messages */
        rc = display_msg();
        break;

      /* Atape Commands */
      case 46: /* query sense (initiate a new sense) */
        rc = stiocqrysense();
        break;

      case 47: /* query inquiry */
        rc = stiocqryinquiry();
        break;

      case 48: /* query/set tape parameters */
        rc = tape_parms();
        break;

      case 49:   /* query/set tape position */
        rc = stiocqrypos();
        if(rc != 0)
          break;
        printf("\nEnter position to set (1=Logical, 2=Physical, <enter>=Quit): ");
        get_input(buf);
        if(strlen(buf) == 0)
          break;
        option = atoi(buf);
        if(option != 1 && option != 2)
          {
            printf("Not a valid selection.\n");
            break;
          }

        printf("\nEnter Block ID in decimal or hex (Example: 16 or 0x10): ");
        get_input(buf);
        if(strlen(buf) == 0)
          break;
        if(buf[0] == 'x')
          stpos.curpos = (int)strtol(buf+1, NULL, 16);
        else if(buf[0] == '0' && buf[1] == 'x')
          stpos.curpos = (int)strtol(buf, NULL, 16);
        else
          stpos.curpos = atoi(buf);

        if(option == 1)
          stpos.block_type = QP_LOGICAL;
        else
          stpos.block_type = QP_PHYSICAL;

        rc = stiocsetpos();
        if(rc == 0)
          rc = stiocqrypos();
        break;

      case 50: /* query/set MT/ST */
        if(mt_st_mode == MTMODE)
          {
            printf("Current mode is MT. IOCTL MTIOCTOP is used for some tape operations.\n");
            printf("Toggle to ST mode? (<enter> for YES, 'n' for NO): ");
          }
        else
          {
            printf("Current mode is ST. IOCTL STIOCTOP is used for some tape operations.\n");
            printf("Toggle to MT mode? (<enter> for YES, 'n' for NO): ");
          }

        get_input(buf);
        if(strlen(buf) == 0 || buf[0] == 'y' || buf[0] == 'Y')
          {
            if(mt_st_mode == MTMODE)
              {
                mt_st_mode = STMODE;
                printf("\nSet to ST. IOCTL STIOCTOP will be used for some tape operations.\n");
              }
            else
              {
                mt_st_mode = MTMODE;
                printf("\nSet to MT. IOCTL MTIOCTOP will be used for some tape operations.\n");
              }
          }
        break;

      case 51: /* report density support */
        rc = stioc_report_density_support();
        break;

      case 52: /* locate block position */
        printf("\nEnter Block ID in decimal or hex (Example: 16 or 0x10): ");
        get_input(buf);
        if(strlen(buf) == 0)
          break;
        if(buf[0] == 'x')
          blockid = (int)strtol(buf+1, NULL, 16);
        else if(buf[0] == '0' && buf[1] == 'x')
          blockid = (int)strtol(buf, NULL, 16);
        else
          blockid = atoi(buf);
        rc = stioc_locate();
        break;

      case 53: /* read current block position */
        rc = stioc_read_position();
        break;

      case 54: /* query mtdevice number */
        rc = mtdevice();
        break;

      case 55: /* synchronize buffer */
        rc = stiocsync();
        break;

      case 56: /* list tape */
        rc = list_tape();
        break;

      case 57: /* query encryption status */
        rc = qry_encryption_status();
        break;

      case 58: /* EKM test */
        rc = ekm_test();
        break;

        /* Medium Changer Commands */
      case 60: /* get element info */
        rc = smcioc_element_info();
        break;

      case 61: /* position to element */
        printf("Enter destination address: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No address entered.\n");
            break;
          }
        dest = (ushort)atoi(buf);
        printf("\n");
        rc = smcioc_pos_to_elem(dest);
        break;

      case 62: /* get inventory */
        rc = smcioc_inventory();
        break;

      case 63: /* exchange medium */
        printf("Enter source address: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No address entered.\n");
            break;
          }
        source = (ushort)atoi(buf);

        printf("Enter first destination address: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No address entered.\n");
            break;
          }
        dest = (ushort)atoi(buf);

        printf("Enter second destination address: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No address entered.\n");
            break;
          }
        dest2 = (ushort)atoi(buf);
        printf("\n");

        rc = smcioc_exchange_medium(source, dest, dest2);
        break;

      case 64: /* move medium */
        printf("Enter source address: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No address entered.\n");
            break;
          }
        source = (ushort)atoi(buf);

        printf("Enter destination address: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No address entered.\n");
            break;
          }
        dest = (ushort)atoi(buf);
        printf("\n");

        rc = smcioc_move_medium(source,dest);
        break;

      case 65: /* load/unload medium */
        printf("\nSelect operation (1=Load, 2=Unload, <enter>=Quit): ");
        get_input(buf);
        if(strlen(buf) == 0)
          break;

        option = atoi(buf);
        rc = ioctl (fd, SMCIOC_ELEMENT_INFO, &element_info);
        if(rc)
          break;
        if(option == 1)
          printf("Enter slot %d-%d or 0 for first full slot: ",
                  element_info.slot_addr,
                  element_info.slot_addr+element_info.slots-1);
        else if(option == 2)
          printf("Enter slot %d-%d or 0 for first empty slot: ",
                  element_info.slot_addr,
                  element_info.slot_addr+element_info.slots-1);
        else
          {
            printf("Not a valid selection.\n");
            break;
          }

        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No slot number entered.\n");
            break;
          }
        source = (ushort)atoi(buf);
        if(option == 1)
          rc = smcioc_load_medium(source);
        else if(option == 2)
          rc = smcioc_unload_medium(source);
        break;

      case 66: /* initialize element */
        rc = smcioc_init_elem_stat();
        break;

      case 67: /* prevent/allow medium removal */
        printf("\nSelect operation (1=Prevent, 2=Allow, <enter>=Quit): ");
        get_input(buf);
        if(strlen(buf) == 0)
            break;

        option = atoi(buf);
        if(option == 1)
          rc = smcioc_prevent_medium_removal();
        else if(option == 2)
          rc = smcioc_allow_medium_removal();
        else
          printf("Not a valid selection.\n");

        break;

      case 68: /* initialize element status range */
        printf("Enter element address: ");
        get_input(buf);
        if((strlen(buf) == 0))
          {
            printf("No address entered.\n");
            break;
          }
        source = (ushort)atoi(buf);
        printf("\nEnter number of elements: ");
        get_input(buf);
        if((strlen(buf) == 0))
          {
            printf("No number of elements entered.\n");
            break;
          }
        count = (ushort)atoi(buf);
        rc = smcioc_init_elem_stat_range(source, count);
        break;

      case 69: /* read device identifiers */
        rc = smcioc_read_element_devids();
        break;

        /* Service aid commands */
      case 70: /* read dump */
        if(search_drive_id() == IBM_3590 && exec_device == CHANGER)
          {
            printf("Operation is not executed for IBM3590 LUN2.\n");
            break;
          }
        printf("Enter destination filename (<enter> for dump0001.dmp): ");
        get_input(buf);
        printf("\n");
        if(strlen(buf) == 0)
          strcpy(buf, "dump0001.dmp");
        rc = stioc_read_dump(buf);
        break;

      case 71: /* force dump */
        if(search_drive_id() == IBM_3590 && exec_device == CHANGER)
          {
            printf("Operation is not executed for IBM3590 LUN2.\n");
            break;
          }
        rc = stioc_force_dump();
        break;

      case 72: /* load ucode */
        if(search_drive_id() == IBM_3590 && exec_device == CHANGER)
          {
            printf("Operation is not executed for IBM3590 LUN2.\n");
            break;
          }

        printf("Enter ucode file name: ");
        get_input(buf);
        if(strlen(buf) == 0)
          {
            printf("No file name entered.\n");
            break;
          }
        printf("\n");
        rc = stioc_load_ucode(buf);
        break;

      case 73: /* reset drive */
        rc = stioc_reset_drive_Atape();
        break;

      case 74: /* read persistent reservation keys */
        rc = stioc_read_keys();
        break;

      case 75: /* read persistent reservations */
        rc = stioc_read_reservations();
        break;

      case 76: /* register a reservation key */
        rc = stioc_register_key();
        break;

      case 77: /* remove current host's key */
        rc = stioc_remove_registration();
        break;

      case 78: /* clear all persistent registration keys*/
        rc = stioc_clear_all_registrations();
        break;

      case 80: /* read cartridge location */
        rc = smcioc_read_cartridge_location();
        break;

      case 99: /* back to main menu */
        if(opened)
          {
            if(close(fd) < 0)
              {
                sprintf(buf,"Device %s closed with errno %d", devname, errno);
                perror (buf);
              }
            else
              printf("Device %s closed...\n", devname);


            /* reinitialize globals */
            opened     = 0;
            mode_flags = 0;
            fd         = -1;
            bzero(devname, 80);
            bzero(pat, 256);
            memset(&mtpos, 0, sizeof(struct mtpos));
            memset(&stpos, 0, sizeof(struct stpos_s));
            memset(&parms, 0, sizeof(struct stchgp_s));
          }
        system("clear");
        device_menu();
        system("clear");
        break;

      default:
        printf("You enter an invalid number. Please try again.\n");
        break;
      }  /* end of switch */
  } /* end of while(1) */
}




/****************************************************************************/
/*                  General Command Functions                               */
/****************************************************************************/

/*- name:     sioc_inquiry
 *  function: invoke IOCTL SIOC_INQUIRY and print result
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_inquiry()
{
  int  rc;
  char vid[9];
  char pid[17];
  char revision[5];
  struct inquiry_data inqdata;

  printf("Issuing inquiry...\n");

  memset(&inqdata, 0, sizeof(struct inquiry_data));
  rc = ioctl(fd, SIOC_INQUIRY, &inqdata);
  if(rc == 0)
    {
      /* print out structure data field */
      printf("\nInquiry Data:\n");
      printf("Peripheral Qualifier-------------------0x%02x\n", inqdata.qual);
      printf("Peripheral Device Type----------------0x%02x\n", inqdata.type);
      printf("Removal Medium Bit--------------------%d\n", inqdata.rm);
      printf("Device Type Modifier------------------0x%02x\n", inqdata.mod);
      printf("ISO version---------------------------0x%02x\n", inqdata.iso);
      printf("ECMA version--------------------------0x%02x\n", inqdata.ecma);
      printf("ANSI version--------------------------0x%02x\n", inqdata.ansi);
      printf("Asynchronous Event Notification Bit---%d\n", inqdata.aenc);
      printf("Terminate I/O Process Message Bit-----%d\n", inqdata.trmiop);
      printf("Response Data Format------------------0x%02x\n", inqdata.rdf);
      printf("Additional Length---------------------0x%02x\n", inqdata.len);
      printf("Medium Changer Mode-------------------0x%02x\n", inqdata.mchngr);
      printf("Relative Addressing Bit---------------%d\n", inqdata.reladr);
      printf("32 Bit Wide Data Transfers Bit--------%d\n", inqdata.wbus32);
      printf("16 Bit Wide Data Transfers Bit--------%d\n", inqdata.wbus16);
      printf("Synchronous Data Transfers Bit--------%d\n", inqdata.sync);
      printf("Linked Commands Bit-------------------%d\n", inqdata.linked);
      printf("Command Queueing Bit------------------%d\n", inqdata.cmdque);
      printf("Soft Reset Bit------------------------%d\n", inqdata.sftre);

      strncpy(vid, (char*)inqdata.vid, 8);
      vid[8] = '\0';
      strncpy(pid, (char*)inqdata.pid, 16);
      pid[16] = '\0';
      strncpy(revision, (char*)inqdata.revision, 4);
      revision[4] = '\0';

      printf("Vendor ID-----------------------------%s\n", vid);
      printf("Product ID----------------------------%s\n", pid);
      printf("Product Revision Level----------------%s\n", revision);

      dump_bytes((char*)inqdata.vendor1, 20, "vendor1");
      dump_bytes((char*)inqdata.vendor2, 31, "vendor2");
    }

  return rc;
}

/*- name:     sioc_test_unit_ready
 *  function: invoke IOCTL SIOC_TEST_UNIT_READY
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_test_unit_ready()
{
  int rc;

  printf("Issuing test unit ready...\n");

  rc = ioctl(fd, SIOC_TEST_UNIT_READY, NULL);
  if(rc == 0)
    printf("Unit ready.\n");

  return rc;
}

/*- name:     sioc_reserve
 *  function: invoke IOCTL SIOC_RESERVE
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_reserve()
{
  int rc;

  printf("Reserving device...\n");

  rc = ioctl (fd, SIOC_RESERVE, NULL);
  return rc;
}

/*- name:     sioc_log_sense_page
 *  function: invoke IOCTL SIOC_LOG_SENSE_PAGE and print log sense page data
 *  argument: page--page number
 *            parm--paramer pointer
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_log_sense_page(int page, int parm)
{
  struct log_sense_page  logdata;
  char   text[80];
  int    count;
  int    rc;

  memset(&logdata, 0, sizeof(struct log_sense_page));

  logdata.page_code = page & 0xFF;
  logdata.parm_pointer = parm;

  printf("Issuing log sense for page 0x%02X...\n", page & 0xFF);

  rc = ioctl(fd, SIOC_LOG_SENSE_PAGE, &logdata);
  if(rc == 0)
    {
      count = logdata.len;
      sprintf(text,"Log Sense Page 0x%02X,  Page Length %d,  Data",page,count);
      if(count > LOGSENSEPAGE)
        count = LOGSENSEPAGE;
      dump_bytes((char*)logdata.data, count, text);
    }

  return rc;
}

/*- name:     sioc_mode_sense_page
 *  function: invoke IOCTL SIOC_MODE_SENSE_PAGE and print date
 *  argument: page number
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_mode_sense_page(int page)
{
  struct mode_sense_page modedata;
  char   text[80];
  int    rc;

  memset(&modedata, 0, sizeof(struct mode_sense_page));
  modedata.page_code = page & 0xFF;

  printf("Issuing mode sense for page 0x%02X...\n",page & 0xFF);

  rc = ioctl(fd, SIOC_MODE_SENSE_PAGE, &modedata);
  if(rc == 0)
    {
      sprintf(text,"Mode Sense Page 0x%02X",page);
      dump_bytes((char*)modedata.data, MODESENSEPAGE, text);
    }

  return rc;
}

/*- name:     sioc_inquiry_page
 *  function: invoke IOCTL SIOC_INQUIRY_PAGE and print inquiry info for this page
 *  argument: page number
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_inquiry_page(int page)
{
  struct inquiry_page inqdata;
  char   text[80];
  int    rc;

  memset(&inqdata, 0, sizeof(struct inquiry_page));
  inqdata.page_code = page & 0xFF;

  printf("Issuing inquiry for page 0x%02X...\n",page);

  rc = ioctl(fd, SIOC_INQUIRY_PAGE, &inqdata);
  if(rc == 0)
    {
      sprintf(text,"Inquiry Page 0x%02X", page & 0xFF);
      dump_bytes((char*)inqdata.data, (inqdata.data[3] & 0xFF)+4, text);
    }

  return rc;
}


/*- name:     sioc_release
 *  function: invoke IOCTL SIOC_RELEASE
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_release()
{
  int rc;

  printf("Releasing device...\n");

  rc = ioctl (fd, SIOC_RELEASE, NULL);
  return rc;
}

/*- name:     sioc_reqsense
 *  function: invoke IOCTL SIOC_REQSENSE and print the sense data
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int sioc_reqsense()
{
  struct request_sense sense_data;
  int rc;

  printf("Issuing request sense...\n");

  memset(&sense_data, 0, sizeof(struct request_sense));
  rc = ioctl(fd, SIOC_REQSENSE, &sense_data);
  if(rc == 0)
  {
    if(!sense_data.err_code)
      printf("No valid sense data returned.\n");
    else
      {
        /* print out data fields */
        printf("Information Field Valid Bit-------%d\n", sense_data.valid);
        printf("Error Code------------------------0x%02x\n", sense_data.err_code);
        printf("Segment Number--------------------0x%02x\n", sense_data.segnum);
        printf("File Mark Detected Bit------------%d\n", sense_data.fm);
        printf("End Of Medium Bit-----------------%d\n", sense_data.eom);
        printf("Illegal Length Indicator Bit------%d\n", sense_data.ili);
        printf("Sense Key-------------------------0x%02x\n", sense_data.key);
        if(sense_data.valid)
          printf("Information Bytes--------------- %d\n", sense_data.info);
        printf("Additional Sense Length-----------0x%02x\n", sense_data.addlen);
        printf("Command Specific Information------0x%02x 0x%02x 0x%02x 0x%02x\n",
               sense_data.cmdinfo >> 24, sense_data.cmdinfo >> 16,
               sense_data.cmdinfo >> 8, sense_data.cmdinfo & 0xFF);
        printf("Additional Sense Code-------------0x%02x\n", sense_data.asc);
        printf("Additional Sense Code Qualifier---0x%02x\n", sense_data.ascq);
        printf("Field Replaceable Unit Code-------0x%02x\n", sense_data.fru);
        printf("Sense Key Specific Valid Bit------%d\n", sense_data.sksv);
        if(sense_data.sksv)
          {
            printf("    Command Data Block Bit--------%d\n", sense_data.cd);
            printf("    Bit Pointer Valid Bit---------%d\n", sense_data.bpv);
            if(sense_data.bpv)
              printf("    System Information Message----0x%02x\n", sense_data.sim);
            printf("    Field Pointer-----------------0x%02x%02x\n",
                        sense_data.field[0], sense_data.field[1]);
          }

        dump_bytes((char*)sense_data.vendor, 109, "Vendor");

      }
  }
  return rc;
}


/*- name:     stioc_qry_driver_version
 *  function: print out the driver IBMtape's major, minor, and increment
 *            of version numbers
 *  argument: none
 *  return:   0 if success, -1 if failed.
 -*/
int stioc_qry_driver_version()
{
  int rc;
  QueryDriverVersion version;

  printf("Issuing query driver version...\n");
  rc = ioctl(fd, QUERY_DRIVER_VERSION, &version);
  if(rc < 0)
    return rc;

  printf("IBMtape driver version: %d.%d.%d\n",
         version.MajorRelease,
         version.MinorRelease,
         version.IncrementRelease);

  return rc;
}


/****************************************************************************/
/*                    Tape Command Functions                                */
/****************************************************************************/

/*- name:     mtiocqrypos
 *  function: invoke IOCTL MTIOCPOS and print current tape position
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int mtiocqrypos(void)
{
  int rc;

  printf("Querying tape position....\n");

  rc = ioctl(fd, MTIOCPOS, &mtpos);
  if(rc == 0)
    printf("Current block ........ %ld (0x%lx)\n", mtpos.mt_blkno, mtpos.mt_blkno);
  return rc;
}

/*- name:     mtiocsetpos
 *  function: invoke IOCTL MTSEEK
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int mtiocsetpos(void)
{
  struct mtop mmtop;
  int         rc;

  printf("Setting tape position to block id %ld (0x%lx)....\n",mtpos.mt_blkno, mtpos.mt_blkno);
  mmtop.mt_op    = MTSEEK;
  mmtop.mt_count = mtpos.mt_blkno;

  rc = ioctl (fd, MTIOCTOP, &mmtop);
  return rc;
}

/*- name:     mtioctop
 *  function: invoke 23 st IOCTLs
 *  argument: IOCTL name and count
 *  return:   0 if succeeded, -1 if failed
 -*/
int mtioctop(int mmtop, int count)
{
  struct mtop mt;
  int         rc;

  mt.mt_op    = mmtop;
  mt.mt_count = count;

  switch (mmtop)
    {
    case MTREW:
      printf("Rewinding tape...\n");
      break;

    case MTOFFL:
      printf("Taking tape offline...\n");
      break;

    case MTERASE:
      printf("Erasing tape...\n");
      break;

    case MTWEOF:
      if(count == 1)
        printf("Writing filemark...\n");
      else
        printf("Writing %d filemarks...\n", count);
      break;

    case MTFSF:
      if(count == 1)
        printf("Forward spacing filemark...\n");
      else
        printf("Forward spacing %d filemarks...\n", count);
      break;

    case MTBSF:
      if(count == 1)
        printf("Backward spacing filemark...\n");
      else
        printf("Backward spacing %d filemarks...\n", count);
      break;

    case MTFSR:
      if(count == 1)
        printf("Forward spacing record...\n");
      else
        printf("Forward spacing %d records...\n", count);
      break;

    case MTBSR:
      if(count == 1)
        printf("Backward spacing record...\n");
      else
        printf("Backward spacing %d records...\n", count);
      break;

    case MTEOM:
      printf("Spacing to end of data...\n");
      break;

    case MTBSFM:
      printf("Backward spacing filemarks and positioning the tape on\n");
      printf("the first block of next file...\n");
      break;

    case MTCOMPRESSION:
      if(count == 0)
        printf("Disabling compression...\n");
      else
        printf("Enabling compression...\n");
      break;

    case MTFSFM:
      printf("Forward spacing filemarks and positioning the tape on\n");
      printf("the last block of the previous file...\n");
      break;

    case MTLOAD:
      printf("Loading tape...\n");
      break;

    case MTLOCK:
      printf("Locking tape drive door...\n");
      break;

    case MTNOP:
      printf("Flushing buffer...\n");
      break;

    case MTRESET:
      printf("Resetting drive...\n");
      break;

    case MTRETEN:
      printf("Retensioning tape...\n");
      break;

    case MTSEEK:
      printf("Seeking block to %d...\n", count);
      break;

    case MTSETBLK:
      printf("Setting block size to %d...\n", count);
      break;

    case MTSETDENSITY:
      printf("Setting density to 0x%x...\n", count);
      break;

    case MTUNLOAD:
      printf("Unloading tape...\n");
      break;

    case MTUNLOCK:
      printf("Unlocking drive door...\n");
      break;

    case MTRAS1:
      printf("Self testing...\n");
      break;

    case MTTELL:
      break;

    default:
      printf("Operation is not supported by ioctl MTIOCTOP.\n");
    } /* end of switch */

  rc = ioctl (fd, MTIOCTOP, &mt);
  if(rc == 0 && mmtop == MTTELL)
    {
      mtpos.mt_blkno = mt.mt_count;
      printf("At block %d.\n", mt.mt_count);
    }

  return rc;
}

/*- name:      mtdevice
 *  function:  get mtdevice number
 *  arguments: void
 *  return:    0 if succeeded, -1 if failed
 */
int mtdevice(void)
{
  int seq_no;
  int rc;

  printf("Issuing mtdevice...\n");
  rc = ioctl(fd, MTDEVICE, &seq_no);
  if(rc == 0)
    {
      printf("\n  Library Sequence Number....%X\n", seq_no);
      printf("This number is valid only when IBM 3590 is attached in a library.\n");
    }

  return rc;
}

/*- name:      stiocdm
 *  function:  invoke IOCTL STIOCDM to display messages
 *  arguments: type  -- message type
 *             msg_0 -- message 0
 *             msg_1 -- message 1
 *  return:    0 if succeeded, -1 if failed
 -*/
int stiocdm(char type, char *msg_0, char *msg_1)
{
  struct stdm_s msg_data;
  char   msg_type[] = { DMMSG0, DMMSG1, DMFLASHMSG0, DMFLASHMSG1, DMALTERNATE };
  int    rc;

  memset(&msg_data, 0, sizeof(struct stdm_s));

  strncpy (msg_data.dm_msg0, msg_0, 8);
  if(msg_1 != NULL)
    {
      strncpy (msg_data.dm_msg1, msg_1, 8);
      printf("Displaying message, msg0: %.8s msg1: %.8s\n", msg_0, msg_1);
    }
  else
    printf("Displaying message, msg0: %.8s\n", msg_0);

  /* Display demount message if tape is loaded */
  msg_data.dm_func = msg_type[type-1] | DMDVMSG;
  rc = ioctl(fd, STIOCDM, &msg_data);

  if(rc == 0)
    {
      /* Display mount message if tape is unloaded */
      msg_data.dm_func = msg_type[type-1] | DMMIMMED;
      rc = ioctl(fd, STIOCDM, &msg_data);
    }

  return rc;
}

/*- name:     display_msg
 *  function: display messages using stiocdm
 *  argument: void
 *  return:   0 if ioctl succeeded or message invalid, -1 if failed
 */
int display_msg(void)
{
  char msg_0[9];
  char msg_1[9];
  char msg_type;
  int  rc;
  char buf[80] = "";

  printf("Select message type (1=msg0, 2=msg1, 3=flash0, 4=flash1, 5=alt): ");
  fgets(buf, 80, stdin);
  buf[strlen(buf)-1] = '\0';
  msg_type = (char)atoi(buf);
  if((msg_type <= 0) || (msg_type > 5))
    {
      printf("Invalid message type selection.\n");
      return 0;
    }
  printf("Enter message 0 text (8 characters or less): ");
  fgets(buf, 80, stdin);
  buf[strlen(buf)-1] = '\0';
  if((strlen (buf) <= 0) || (strlen (buf) > 8))
    {
      printf("Invalid message length.\n");
      return 0;
    }
  strcpy (msg_0, buf);
  if(msg_type == 2 || msg_type == 4 ||
     msg_type == 5)
    {
      printf("Enter message 1 text (8 characters or less): ");
      fgets(buf, 80, stdin);
      buf[strlen(buf)-1] = '\0';
      if((strlen (buf) <= 0) || (strlen (buf) > 8))
        {
          printf("Invalid message length.\n");
          return 0;
        }
      strcpy (msg_1, buf);
      rc = stiocdm(msg_type, msg_0, msg_1);
    }
  else
    rc = stiocdm(msg_type, msg_0, NULL);

  return rc;
}

/*- name:     stioc_prevent_medium_removal
 *  function: invoke IOCTL STIOC_PREVENT_MEDIUM_REMOVAL
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_prevent_medium_removal()
{
  int rc;

  printf("Issuing prevent medium removal...\n");
  rc = ioctl(fd, STIOC_PREVENT_MEDIUM_REMOVAL, 0);
  return rc;
}

/*- name:     stioc_allow_medium_removal
 *  function: invoke IOCTL STIOC_ALLOW_MEDIUM_REMOVAL
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_allow_medium_removal()
{
  int rc;

  printf("Issuing allow medium removal...\n");
  rc = ioctl(fd, STIOC_ALLOW_MEDIUM_REMOVAL, 0);
  return rc;
}


/*- name:     status
 *  function: invoke IOCTL MTIOCGET and print tape status
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int status(void)
{
  struct mtget mtstatus;
  long         gstat;
  int          rc;

  memset(&mtstatus, 0, sizeof(struct mtget));

  rc = ioctl(fd, MTIOCGET, &mtstatus);
  if(rc == 0)
    {
      /* tape type info */
      if(mtstatus.mt_type == MT_ISSCSI1)
        printf("Generic ANSI SCSI-1 tape drive\n");
      else if(mtstatus.mt_type == MT_ISSCSI2)
        printf("Generic ANSI SCSI-2 tape drive\n");
      else if(mtstatus.mt_type == MT_ISDDS2)
        printf("DDS device with partitions\n");
      else if(mtstatus.mt_type == MT_ISDDS1)
        printf("DDS device without partitions\n");
      else if(mtstatus.mt_type == MT_ISEVEREX_FT40A)
        printf("Everex FT40A (QIC-40)\n");
      else if(mtstatus.mt_type == MT_ISTEAC_MT2ST)
        printf("Teac MT-2ST 155mb drive, Teac DC-1 card (Wangtek type)\n");
      else if(mtstatus.mt_type == MT_ISWT5099EEN24)
        printf("Wangtek 5099-een24, 60MB, QIC-24\n");
      else if(mtstatus.mt_type == MT_ISQIC02_ALL_FEATURES)
        printf("Generic QIC-02 with all features\n");
      else if(mtstatus.mt_type == MT_ISARCHIVESC499)
        printf("Archive SC-499 QIC-36 controller\n");
      else if(mtstatus.mt_type == MT_ISARCHIVE_2060L)
        printf("Archive Viper 2060L\n");
      else if(mtstatus.mt_type == MT_ISARCHIVE_2150L)
        printf("Archive Viper 2150L\n");
      else if(mtstatus.mt_type == MT_ISARCHIVE_VP60I)
        printf("Archive VP60i, QIC-02\n");
      else if(mtstatus.mt_type ==  MT_ISTDC3610 )
        printf("Tandberg 6310, QIC-24\n");
      else if(mtstatus.mt_type == MT_ISCMSJ500)
        printf("CMS Jumbo 500 (QIC-02?)\n");
      else if(mtstatus.mt_type == MT_ISARCHIVE_5945L2)
        printf("Archive 5945L-2, QIC-24, QIC-02?\n");
      else if(mtstatus.mt_type & MT_ISWT5150)
        printf("Wangtek 5150EQ, QIC-150, QIC-02\n");
      else if(mtstatus.mt_type == MT_ISQIC02)
        printf("Generic QIC-02 tape streamer\n");
      else
        printf("Unknown tape drive.\n");

      /* tape partition, block size, and density info */
      printf("File number=%d, block number=%d, partition=%ld.\n",
             mtstatus.mt_fileno, mtstatus.mt_blkno, mtstatus.mt_resid);
      printf("Tape block size %ld bytes. ",
             (mtstatus.mt_dsreg & MT_ST_BLKSIZE_MASK) >> MT_ST_BLKSIZE_SHIFT);
      printf("Density code 0x%lx\n",
             (mtstatus.mt_dsreg & MT_ST_DENSITY_MASK) >> MT_ST_DENSITY_SHIFT);

      /* Recovered softerror info */
      printf("Soft error count since last status=%ld\n", mtstatus.mt_erreg & MT_ST_SOFTERR_MASK >> MT_ST_SOFTERR_SHIFT);

      /* tape position info */
      gstat = mtstatus.mt_gstat;
      printf("General status bits on (%lx):\n", gstat);
      if(GMT_EOF(gstat))
        printf(" EOF");
      if(GMT_BOT(gstat))
        printf(" BOT");
      if(GMT_EOT(gstat))
        printf(" EOT");
      if(GMT_SM(gstat))
        printf(" SM");
      if(GMT_EOD(gstat))
        printf(" EOD");
      if(GMT_WR_PROT(gstat))
        printf(" WR_PROT");
      if(GMT_ONLINE(gstat))
        printf(" ONLINE");
      if(GMT_DR_OPEN(gstat))
        printf(" DR_OPEN");
      if(GMT_IM_REP_EN(gstat))
        printf(" IM_REP_EN");

      /* reports the current density seeting for 9-track 1/2" tape drive only */
      if(GMT_D_6250(gstat))
        printf("Current density for 9-track 1/2 tape drive: 6250\n");
      else if(GMT_D_1600(gstat))
        printf("Current density for 9-track 1/2 tape drive: 1600\n");
      else if(GMT_D_800(gstat))
        printf("Current density for 9-track 1/2 tape drive: 800\n");
      else
        ;

      printf("\n");
    }

  return rc;
}


/*- name:     stiocqryp
 *  function: invoke IOCTL STIOCQRYP and set global structure variable parms
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocqryp(void)
{
  int rc;

  printf("Querying tape parameters...\n");
  rc = ioctl(fd, STIOCQRYP, &parms);
  return rc;
}

/*- name:     stiocsetp
 *  function: invoke IOCTL STIOCSETP passing global structure variable parms
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocsetp(void)
{
  int rc;

  printf("Setting tape parameters...\n");
  rc = ioctl(fd, STIOCSETP, &parms);
  return rc;
}

/*- name:     stiocsync
 *  function: invoke IOCTL STIOCSYNC
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocsync(void)
{
  int rc;

  printf("Synchronizing buffers...\n");
  rc = ioctl(fd, STIOCSYNC, NULL);
  return rc;
}

/*- name:     rwtest
 *  function: test write and read premitives.
 *  argument: blksize    -- size of a block
 *            blocks     -- number of blocks; if blksize is 0, blocks = 1
 *            repetition -- number of repeated tests
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int rwtest(int blksize, int blocks, int repetition)
{
  long long   total = 0;
  struct mtop mt;
  char        *wbuf, *rbuf;
  int         bytes, count;
  int         i, j, k;
  int         rc;

  memset(&mt, 0, sizeof(struct mtop));

  printf("\nRead/Write test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

  if(blksize == 0)
  {
    count = blocks;
    blocks = 1;
  }
  else
    count = blocks * blksize;

  wbuf = malloc(count);
  if(wbuf == NULL)
  {
    errno = ENOMEM;
    return -2;
  }

  rbuf = malloc(count);
  if(rbuf == NULL)
  {
    errno = ENOMEM;
    free(wbuf);
    return -2;
  }

  /* set block size */
  mt.mt_op = MTSETBLK;
  mt.mt_count = blksize;
  printf("Setting block size to %d...\n", blksize);
  rc = ioctl(fd, MTIOCTOP, &mt);
  if(rc != 0)
  {
    free(wbuf);
    free(rbuf);
    return rc;
  }

  /* generate data pattern */
  genpat();
  j = 01;
  for (i=0; i<count; i++)
  {
    wbuf[i] = j;
    j = pat[j];
  }

  for (j=0,total=0; j<repetition; j++)
    {
      errno = 0;
      printf("\nRead/Write test pass %d\n", j+1);
      printf("Writing %d bytes, %d blocks...\n", count, blocks);
      bytes = write(fd, wbuf, count);
      if(bytes == -1)
        {
          rc = -1;
          break;
        }

      total += bytes;
      if(blksize == 0)
        printf("Write complete, bytes written %d, blocks 1\n", bytes);
      else
        printf("Write complete, bytes written %d, blocks %d\n", bytes, bytes/blksize);

      if(bytes != count)
        {
          printf("Error on write, bytes written not equal to bytes requested\n");
          errno = 999;
          rc = -2;
          break;
        }

      mt.mt_op = MTWEOF;
      mt.mt_count = 2;
      printf("Writing 2 file marks...\n");
      rc = ioctl(fd, MTIOCTOP, &mt);
      if(rc != 0)
        break;

      mt.mt_op = MTBSF;
      mt.mt_count = 2;
      printf("Backward spacing 2 file marks...\n");
      rc = ioctl(fd, MTIOCTOP, &mt);
      if(rc != 0)
        break;

      mt.mt_op = MTBSR;
      mt.mt_count = blocks;
      printf("Backward spacing %d records...\n", blocks);
      rc = ioctl(fd, MTIOCTOP, &mt);
      if(rc != 0)
        break;

      /* read the data back from tape and compare to original */
      bzero(rbuf, count);
      printf("Reading %d bytes, %d blocks...\n", count, blocks);
      bytes = read(fd, rbuf, count);
      if(bytes == -1)
        {
          rc = -1;
          break;
        }

      if(blksize == 0)
        printf("Read complete, bytes read %d, blocks 1\n", bytes);
      else
        printf("Read complete, bytes read %d, blocks %d\n", bytes, bytes/blksize);

      if(bytes != count)
        {
          printf("Error on read, bytes read not equal to bytes written\n");
          errno = 999;
          rc = -2;
          break;
        }

      printf("Forward spacing file mark...\n");
      mt.mt_op = MTFSF;
      mt.mt_count = 1;
      rc = ioctl(fd, MTIOCTOP, &mt);
      if(rc != 0)
        break;

      /* compare data */
      for (k=0; k < count; k++)
        {
          if(wbuf[k] != rbuf[k])
            {
              printf("Read data miscompare, ");
              printf("byte read at position %d is %02X, should be %02X\n", k,rbuf[k],wbuf[k]);
              errno = 999;
              rc = -2;
              break;
            }
        }
      if(k != count)
        {
          if(ERR_STOP)
            break;
          else
            {
              errno = 0;
              continue;
            }
        }
      printf("Data verification complete, no discrepancies found\n");
    }

  if(errno)
    printf("\nRead/Write test terminated, total bytes written %lld\n", total);
  else
    printf("\nRead/Write test completed, total bytes written %lld\n", total);
  free(wbuf);
  free(rbuf);

  return rc;
}

/*- name:     wtest
 *  function: test write premitives.
 *  argument: blksize    -- size of a block
 *            blocks     -- number of blocks; if blksize is 0, blocks = 1
 *            repetition -- number of repeated tests
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int wtest(int blksize, int blocks, int repetition)
{
  struct mtop mt;
  long long   total = 0;
  char        *wbuf;
  int         count, bytes;
  int         i, j;
  int         rc;

  memset(&mt, 0, sizeof(struct mtop));

  printf("\nWrite test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

  if(blksize == 0)
  {
    count = blocks;
    blocks = 1;
  }
  else
    count = blocks * blksize;

  wbuf = malloc(count);
  if(wbuf == NULL)
  {
    errno = ENOMEM;
    return -2;
  }

  /* set block size */
  mt.mt_op = MTSETBLK;
  mt.mt_count = blksize;
  printf("Setting block size...\n");
  rc = ioctl(fd, MTIOCTOP, &mt);
  if(rc != 0)
  {
    free(wbuf);
    return rc;
  }

  /* generate data pattern */
  genpat();
  j = 01;
  for (i = 0; i < count; i++)
  {
    wbuf[i] = j;
    j=pat[j];
  }

  for (j=0,total=0; j<repetition; j++)
  {
    printf("\nWrite test pass %d\n",j+1);

    printf("Writing %d bytes, %d blocks...\n", count, blocks);
    bytes = write(fd,wbuf,count);
    if(bytes == -1)
      {
        rc = -1;
        break;
      }

    total += bytes;
    if(blksize == 0)
      printf("Write complete, bytes written %d, blocks 1\n", bytes);
    else
      printf("Write complete, bytes written %d, blocks %d\n", bytes, bytes/blksize);

    if(bytes != count)
    {
      printf("Error on write, bytes written not equal to bytes requested\n");
      errno = 999;
      rc = -2;
      break;
    }
  }


  if(errno)
    printf("\nWrite test terminated, total bytes written %lld\n",total);
  else
    printf("\nWrite test completed, total bytes written %lld\n",total);

  if(wbuf)
    free(wbuf);

  return rc;
}


/*- name:     rtest
 *  function: test read premitives. rtest() assumes that user has done wtest or rwtest and rewinded tape
 *            to the start positision of last write
 *  argument: blksize    -- size of a block
 *            blocks     -- number of blocks; if blksize is 0, blocks = 1
 *            repetition -- number of repeated tests
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int rtest(int blksize, int blocks, int repetition)
{
  long long   total = 0;
  struct mtop mt;
  char        *wbuf, *rbuf;
  int         bytes, count;
  int         i, j, k;
  int         rc;

  memset(&mt, 0, sizeof(struct mtop));

  printf("\nRead test, blocksize %u count %u repetition %u\n",
         blksize, blocks, repetition);

  if(blksize == 0)
  {
    count = blocks;
    blocks = 1;
  }
  else
    count = blocks * blksize;

  wbuf = malloc(count);
  if(wbuf == NULL)
  {
    errno = ENOMEM;
    return -2;
  }

  rbuf = malloc(count);
  if(rbuf == NULL)
  {
    errno = ENOMEM;
    free(wbuf);
    return -2;
  }

  /* set block size */
  mt.mt_op = MTSETBLK;
  mt.mt_count = blksize;
  printf("Setting block size to %d...\n", blksize);
  rc = ioctl(fd, MTIOCTOP, &mt);
  if(rc != 0)
  {
    free(wbuf);
    free(rbuf);
    return rc;
  }

  /* generate data pattern */
  genpat();
  j = 01;
  for (i=0; i < count; i++)
  {
    wbuf[i] = j;
    j = pat[j];
  }

  /* start read */
  for (j=0,total=0; j<repetition; j++)
  {
    printf("\nRead test pass %d\n", j+1);

    /* read the data back from tape and compare to expected write data */
    bzero(rbuf,count);
    printf("Reading %d bytes, %d blocks...\n", count, blocks);
    bytes = read(fd, rbuf, count);
    if(bytes == -1)
      {
        if(errno == ENOMEM)
          {
            /* this is overlength read */
            dump_bytes((char*)rbuf, count, "Overlength read");
          }

        rc = -1;
        break;
      }

    total += bytes;
    if(blksize == 0)
      printf("Read complete, bytes read %d, blocks 1\n", bytes);
    else
      printf("Read complete, bytes read %d, blocks %d\n", bytes, bytes/blksize);

    for (k=0; k < bytes; k++)
    {
      if(wbuf[k] != rbuf[k])
      {
        printf("Read data miscompare, ");
        printf("byte read at position %d is %02X, should be %02X\n", k, rbuf[k], wbuf[k]);
        errno = 999;
        rc = -2;
        break;
       }
      }
    if(k != bytes)
    {
      if(ERR_STOP)
        break;
      else
        {
          errno = 0;
          continue;
        }
    }

    if(k != 0)
      printf("Data verification complete, no discrepancies found\n");
  }

  if(errno)
    printf("\nRead test terminated, total bytes read %lld\n", total);
  else
    printf("\nRead test completed, total bytes read %lld\n", total);

  if(rbuf)
    free(rbuf);
  if(wbuf)
    free(wbuf);

  return rc;
}

/*- name:     twrite
 *  function: test write a file on tape
 *  argument: file name
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int twrite(char *fname)
{
  struct mtop mt;
  long long   total = 0;
  char        *wbuf;
  int         wrtfd;
  int         bytes, count, blksize;
  int         rc;

  memset(&mt, 0, sizeof(struct mtop));

  printf("Opening source file %s...\n",fname);
  wrtfd = open(fname, O_RDONLY);
  if(wrtfd == -1)
    return -2;

  /* get max_blksize and max_scsi_xfer in parms */
  rc = stiocqryp();
  if(rc != 0)
    {
      close(wrtfd);
      return rc;
    }

  blksize = min(parms.max_blksize, parms.max_scsi_xfer);
  blksize = min(blksize, MAX_TRANSFER_SIZE);

  /* set actual block size to variable */
  mt.mt_op = MTSETBLK;
  mt.mt_count = 0;
  printf("Setting block size to variable...\n");
  rc = ioctl(fd, MTIOCTOP, &mt);
  if(rc != 0)
  {
    close(wrtfd);
    return rc;
  }

  wbuf = malloc(blksize);
  if(wbuf == NULL)
  {
    close(wrtfd);
    errno = ENOMEM;
    return -2;
  }

  printf("Writing file to tape...\n");

  while(1)
  {
    bytes = read(wrtfd, wbuf, blksize);
    if(bytes == -1)
    {
      rc = -1;
      break;
    }
    if(bytes == 0)
      break;
    count = write(fd,wbuf,bytes);
    if(count == -1)
    {
      rc = -1;
      break;
    }
    total += count;
    if(count != bytes)
    {
      errno = EIO;
      rc = -2;
      break;
    }
  }

  if(rc)
    printf("Write terminated, %lld total bytes written...\n", total);
  else
    printf("Write complete, %lld total bytes written...\n", total);

  close(wrtfd);
  if(wbuf)
    free(wbuf);

  /* write 2 file marks */
  mt.mt_op = MTWEOF;
  mt.mt_count = 2;
  printf("Writing 2 file marks...\n");
  rc = ioctl(fd, MTIOCTOP, &mt);
  if(rc != 0)
    return rc;

  /* back space 1 filemark */
  mt.mt_op = MTBSF;
  mt.mt_count = 1;
  printf("Backward spacing 1 file mark...\n");
  rc = ioctl(fd, MTIOCTOP, &mt);

  return rc;
}

/*- name:     tread
 *  function: test read a file or specified number of blocks
 *  argument: fname  -- file name to read
 *            number -- number of blocks to read
 *  return:   0 if succeeded, -1 if failed
 -*/
int tread(char *fname, int number)
{
  struct mtop mt;
  long long   total = 0;
  char        *rbuf;
  int         rec_count = 0;
  int         rdfd;
  int         bytes, count, blksize;
  int         rc = 0;

  printf("Opening destination file %s...\n", fname);

  rdfd = open(fname, O_RDWR|O_CREAT|O_TRUNC, 0666);
  if(rdfd == -1)
    return -2;

  /* get max_blksize and max_scsi_xfer in parms */
  rc = stiocqryp();
  if(rc != 0)
    {
      close(rdfd);
      return rc;
    }

  blksize = min(parms.max_blksize, parms.max_scsi_xfer);
  blksize = min(blksize, MAX_TRANSFER_SIZE);

  /* set actual block size to variable */
  mt.mt_op = MTSETBLK;
  mt.mt_count = 0;
  printf("Setting block size to variable...\n");
  rc = ioctl(fd, MTIOCTOP, &mt);
  if(rc != 0)
  {
    close(rdfd);
    return rc;
  }

  rbuf = malloc(blksize);
  if(rbuf == NULL)
  {
    close(rdfd);
    errno = ENOMEM;
    return -2;
  }

  if(number)
    printf("Reading %d records from tape...\n", number);
  else
    printf("Reading file from tape...\n");

  while(1)
  {
    if(number && rec_count == number)
      break;
    bytes = read(fd, rbuf, blksize);
    if(bytes == -1)
    {
      rc = -1;
      break;
    }
    total += bytes;
    if(bytes == 0)
      break;
    rec_count++;
    count = write(rdfd, rbuf, bytes);
    if(count == -1)
      {
        rc = -1;
        break;
      }
    if(count != bytes)
      {
        printf("count = %d, bytes = %d\n", count, bytes);
        rc = -2;
        break;
      }
  }

  if(rc || (number && rec_count != number))
    printf("Read terminated, %d records, %lld total bytes read...\n",
            rec_count, total);
  else
    printf("Read complete, %d records, %lld total bytes read...\n",
            rec_count, total);

  close(rdfd);
  if(rbuf)
    free(rbuf);

  return rc;
}

/*- name:      stioctop
 *  function:  invoke IOCTL STIOCTOP
 *  arguments: IOCTL name and count
 *  return:    0 if succeeded, -1 if failed
 */
int stioctop(int sstop, int count)
{
  struct stop st;
  int         rc;

  st.st_op    = sstop;
  st.st_count = count;

  switch(sstop)
    {
    case STREW:
      printf("Rewinding tape...\n");
      break;

    case STOFFL:
      printf("Taking tape offline...\n");
      break;

    case STERASE:
      printf("Erasing tape...\n");
      break;

    case STRETEN:
      printf("Retensioning tape...\n");
      break;

    case STWEOF:
      if(count == 1)
        printf("Writing filemark...\n");
      else
        printf("Writing %d filemarks...\n",count);
      break;

    case STFSF:
      if(count == 1)
        printf("Forward spacing filemark...\n");
      else
        printf("Forward spacing %d filemarks...\n",count);
      break;

    case STRSF:
      if(count == 1)
        printf("Backward spacing filemark...\n");
      else
        printf("Backward spacing %d filemarks...\n",count);
      break;

    case STFSR:
      if(count == 1)
        printf("Forward spacing record...\n");
      else
        printf("Forward spacing %d records...\n",count);
      break;

    case STRSR:
      if(count == 1)
        printf("Backward spacing record...\n");
      else
        printf("Backward spacing %d records...\n",count);
      break;

    case STTUR:
      printf("Test unit ready...\n");
      break;

    case STLOAD:
      printf("Loading tape...\n");
      break;

    case STSEOD:
      printf("Forward spacing to the end of data...\n");
      break;

    case STEJECT:
      printf("Unload tape...\n");
      break;

    } /* end of switch */

  rc = ioctl(fd, STIOCTOP, &st);
  return rc;
}


/*- name:     tape_parms
 *  function: print/set the parameters in global structure variable parm
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int tape_parms(void)
{
  int  rc = 0;
  char buf[80] = "";
  int  option, select;
  int  work;

  while(1)
    {
      rc = stiocqryp();
      if(rc != 0)
        break;

      printf("\n Nonchangeable parameters:\n\n");

      printf("   Medium Type.............. %02X\n", parms.medium_type);
      printf("   Trace Hookword........... %x\n", parms.hkwrd);
      printf("   Density Code............. %02X\n", parms.density_code);
      printf("   Minimum Block Size....... %d\n", parms.min_blksize);
      printf("   Maximum Block Size....... %d\n", parms.max_blksize);

      if(parms.write_protect)
        printf("   Physical Write Protect... Yes\n");
      else
        printf("   Physical Write Protect... No\n");

      if(parms.acf_mode == ACF_NONE)
        printf("   ACF Mode................. N/A\n");
      else if(parms.acf_mode == ACF_SYSTEM)
        printf("   ACF Mode................. SYSTEM\n");
      else if(parms.acf_mode == ACF_MANUAL)
        printf("   ACF Mode................. MANUAL\n");
      else if(parms.acf_mode == ACF_AUTOMATIC)
        printf("   ACF Mode................. AUTOMATIC\n");
      else if(parms.acf_mode == ACF_RANDOM)
        printf("   ACF Mode................. RANDOM\n");
      else if(parms.acf_mode == ACF_ACCUMULATE)
        printf("   ACF Mode................. ACCUMULATE\n");

      if(parms.retain_reservation)
        printf("   Retain Reservation....... Yes\n");
      else
        printf("   Retain Reservation....... No\n");

      if(parms.alt_pathing)
        printf("   Alternate Pathing........ Yes\n");
      else
        printf("   Alternate Pathing........ No\n");

      if(parms.read_sili_bit)
        printf("   Read SILI Bit............ Yes\n");
      else
        printf("   Read SILI Bit............ No \n");

      if(parms.autoload)
        printf("   Autoload................. Yes\n");
      else
        printf("   Autoload................. No\n");

      if(parms.record_space_mode == SCSI_SPACE_MODE)
        printf("   Record Space Mode........ SCSI\n");
      else if(parms.record_space_mode == AIX_SPACE_MODE)
        printf("   Record Space Mode........ AIX\n");
      else
        printf("   Record Space Mode........ Unknown\n");

      if(parms.emulate_autoloader)
        printf("   Emulate Autoloader....... Yes\n");
      else
        printf("   Emulate Autoloader....... No\n");


      printf("\n Changeable parameters:\n\n");

      if(parms.compression)
        printf("   1. Compression........... Yes\n");
      else
        printf("   1. Compression........... No\n");

      if(parms.buffered_mode)
        printf("   2. Buffered Mode......... Yes\n");
      else
        printf("   2. Buffered Mode......... No\n");

      if(parms.trailer_labels)
        printf("   3. Trailer Labels........ Yes\n");
      else
        printf("   3. Trailer Labels........ No\n");

      if(parms.rewind_immediate)
        printf("   4. Rewind Immediate...... Yes\n");
      else
        printf("   4. Rewind Immediate...... No\n");

      printf("   5. Block Size............ %d\n", parms.blksize);

      printf("   6. Maximum SCSI Xfer..... %d\n", parms.max_scsi_xfer);

      if(parms.read_past_filemark)
        printf("   7. Read Past Filemark.... Yes\n");
      else
        printf("   7. Read Past Filemark.... No\n");

      if(parms.capacity_scaling == SCALE_100)
        printf("   8. Capacity Scaling...... 100%c\n",'%');
      else if(parms.capacity_scaling == SCALE_75)
        printf("   8. Capacity Scaling...... 75%c\n",'%');
      else if(parms.capacity_scaling == SCALE_50)
        printf("   8. Capacity Scaling...... 50%c\n",'%');
      else if(parms.capacity_scaling == SCALE_25)
        printf("   8. Capacity Scaling...... 25%c\n",'%');
      else if(parms.capacity_scaling == SCALE_VALUE)
        printf("   8. Capacity Scaling Hex Value ... 0x%02X\n", parms.capacity_scaling_value);

      if(parms.logical_write_protect == WORM_PROTECT)
        printf("   9. Logical Write Protect. WORM\n");
      else if(parms.logical_write_protect == ASSOCIATED_PROTECT)
        printf("   9. Logical Write Protect. ASSOCIATED\n");
      else if(parms.logical_write_protect == PERSISTENT_PROTECT)
        printf("   9. Logical Write Protect. PERSISTENT\n");
      else
        printf("   9. Logical Write Protect. No\n");

      /* changerable parameters for tracing */
      if(parms.trace)
        printf("  10. Trace................. On\n");
      else
        printf("  10. Trace................. Off\n");

      if(parms.disable_sim_logging)
        printf("  11. SimMim Logging ....... Off\n");
      else
        printf("  11. SimMim Logging ....... On\n");

      if(parms.disable_auto_drive_dump)
        printf("  12. Auto Drive Dump ...... Off\n");
      else
        printf("  12. Auto Drive Dump ...... On\n");

      if(parms.logging)
        printf("  13. Volume Logging........ On\n");
      else
        printf("  13. Volume Logging........ Off\n");

      /* setting tape parameters or quit */
      printf("\nEnter parameter to change or <enter> to quit: ");
      get_input(buf);
      if(strlen(buf) == 0)
        return rc;
      if(buf[0] == 'q' || buf[0] == 'Q')
        return rc;

      option = atoi(buf);
      switch (option)
        {
        case 1:
          if(parms.compression)
            parms.compression = FALSE;
          else
            parms.compression = TRUE;
          break;

        case 2:
          if(parms.buffered_mode)
            parms.buffered_mode = FALSE;
          else
            parms.buffered_mode = TRUE;
          break;

        case 3:
          if(parms.trailer_labels)
            parms.trailer_labels = FALSE;
          else
            parms.trailer_labels = TRUE;
          break;

        case 4:
          if(parms.rewind_immediate)
            parms.rewind_immediate = FALSE;
          else
            parms.rewind_immediate = TRUE;
          break;

        case 5:
          printf("\nEnter block size: ");
          get_input(buf);
          parms.blksize = atoi(buf);
          break;

        case 6:
          printf("\nEnter maximum transfer bytes (<enter> for default 262144 bytes ): ");
          get_input(buf);
          parms.max_scsi_xfer = atoi(buf);
          if(parms.max_scsi_xfer == 0)
            parms.max_scsi_xfer = 256 * 1024;
          break;

        case 7:
          if(parms.read_past_filemark)
            parms.read_past_filemark = FALSE;
          else
            parms.read_past_filemark = TRUE;
          break;

        case 8:
          printf("\nEnter capacity scaling (1=100%c, 2=75%c, 3=50%c, 4=25%c, 5=other): ",'%','%','%','%');
          get_input(buf);
          select = atoi(buf);
          if(select != 1 && select != 2 && select != 3 && select != 4 && select != 5)
            {
              printf("No valid selection entered.\n");
              break;
            }

          memset(buf, 0, 80);
          if(select == 5)
            {
              printf("\nEnter hex value of capacity scaling (e.g. 0xD0 or hit <enter> for default 0):");
              get_input(buf);
              if(strlen(buf) == 0)
                parms.capacity_scaling_value = 0;
              else
                parms.capacity_scaling_value = (unchar)strtol(buf, NULL, 16);
            }

          work = select - 1;
          printf("\nWarning!  All data on tape will be lost with this change!\n");
          printf("Continue (y or n): ");
          get_input(buf);
          if(strlen(buf) == 0 || buf[0] != 'y')
            break;
          parms.capacity_scaling = work;

          break;

        case 9:
          if(parms.logical_write_protect == WORM_PROTECT)
            {
              printf("\nWORM protection can not be changed, hit any key to continue...\n");
              get_input(buf);
              break;
            }

          printf("\nEnter write protect (1=NONE, 2=ASSOCIATED, 3=PERSISTENT, 4=WORM): ");
          get_input(buf);
          select = atoi(buf);
          if(select != 1 && select != 2 && select != 3 && select !=4)
            {
              printf("No valid selection entered.\n");
              break;
            }
          work = select - 1;
          if(work == WORM_PROTECT)
            {
              printf("\nWarning!  WORM protection is a permanent change to the tape!\n");
              printf("Continue (y or n): ");
              get_input(buf);
              if(strlen(buf) == 0 || buf[0] != 'y')
                break;
            }
          parms.logical_write_protect = work;
          break;

          /* trace parameters */
        case 10:
          if(parms.trace)
            parms.trace = FALSE;
          else
            parms.trace = TRUE;
          break;
        case 11:
          if(parms.disable_sim_logging)
            parms.disable_sim_logging = FALSE;
          else
            parms.disable_sim_logging = TRUE;
          break;
        case 12:
          if(parms.disable_auto_drive_dump)
            parms.disable_auto_drive_dump = FALSE;
          else
            parms.disable_auto_drive_dump = TRUE;
          break;
        case 13:
          if(parms.logging)
            parms.logging = FALSE;
          else
            parms.logging = TRUE;
          break;

        default:
          printf("You entered an invalid number.\n");
          return rc;
        }  /* end of switch */

      rc = stiocsetp();
      if(rc != 0)
        break;
    } /* end of while(1) */

  return rc;
}

/*- name:     stiocqrypos
 *  function: invoke IOCTL STIOCQRYPOS and print result
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocqrypos(void)
{
  int rc;

  /* Logical position */
  stpos.block_type = QP_LOGICAL;
  rc = ioctl(fd, STIOCQRYPOS, &stpos);
  if(rc != 0)
    return rc;

  printf("\nLogical Tape position:\n\n");
  printf("    Partition Number ........... %d\n",stpos.partition_number);
  if(stpos.bot)
    printf("    Beginning of Partition ..... Yes\n");
  else
    printf("    Beginning of Partition ..... No\n");

  if(stpos.eot)
    printf("    Leading End of Partition ... Yes\n");
  else
    printf("    Leading End of Partition ... No\n");

    printf("    Current Block ID ........... %u (0x%08X)\n",
       stpos.curpos,stpos.curpos);
    printf("    Next Block on Tape ......... %u (0x%08X)\n",
         stpos.tapepos,stpos.tapepos);
  if(stpos.lbot == LBOT_NONE)
    printf("    Last Block Written ......... NONE\n");
  else if(stpos.lbot == LBOT_UNKNOWN)
    printf("    Last Block Written ......... UNKNOWN\n");
  else
    printf("    Last Block Written ......... %u (0x%08X)\n",
           stpos.lbot, stpos.lbot);

  printf("    Blocks in Buffer ........... %u\n", stpos.num_blocks);
  printf("    Bytes in Buffer ............ %u\n", stpos.num_bytes);

  /* Physical position */
  stpos.block_type = QP_PHYSICAL;
  rc = ioctl(fd, STIOCQRYPOS, &stpos);
  if(rc != 0)
   return rc;

  printf("\nPhysical Tape position:\n\n");
  printf("    Partition Number ........... %d\n",stpos.partition_number);
  if(stpos.bot)
    printf("    Beginning of Partition ..... Yes\n");
  else
    printf("    Beginning of Partition ..... No\n");

  if(stpos.eot)
    printf("    Leading End of Partition ... Yes\n");
  else
    printf("    Leading End of Partition ... No\n");
    printf("    Current Block ID ........... %u (0x%08X)\n",
           stpos.curpos,stpos.curpos);
    printf("    Next Block on Tape ......... %u (0x%08X)\n",
         stpos.tapepos,stpos.tapepos);
  if(stpos.lbot == LBOT_NONE)
    printf("    Last Block Written ......... NONE\n");
  else if(stpos.lbot == LBOT_UNKNOWN)
    printf("    Last Block Written ......... UNKNOWN\n");
  else
    printf("    Last Block Written ......... %u (0x%08X)\n",
            stpos.lbot, stpos.lbot);

  printf("    Blocks in Buffer ........... %u\n", stpos.num_blocks);
  printf("    Bytes in Buffer ............ %u\n", stpos.num_bytes);
  printf("\n");

  return rc;
}

/*- name:     stiocsetpos
 *  function: invoke IOCTL STIOCSETPOS
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocsetpos(void)
{
  int rc;

  printf("Setting tape position to block id %u (0x%x)...\n", stpos.curpos, stpos.curpos);
  rc = ioctl(fd, STIOCSETPOS, &stpos);

  return rc;
}

/*- name:     stiocqrysense
 *  function: invoke IOCTL STIOCQRYSENSE and print new sense data
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocqrysense()
{
  struct stsense_s stsense;
  int    rc;

  printf("Issuing query sense...\n");

  memset(&stsense, 0, sizeof(struct stsense_s));
  stsense.sense_type = FRESH;
  rc = ioctl(fd, STIOCQRYSENSE, &stsense);
  if(rc != 0)
    return rc;

  if(stsense.residual_count > 0)
    printf("Residual count: %d\n", stsense.residual_count);

  if(stsense.len > 0)
    dump_bytes((char*)stsense.sense, stsense.len, "Sense Data");

  return rc;
}

/*- name:     stiocqryinquiry
 *  function: invoke IOCTL STIOCQRYINQUIRY and print inquiry data
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stiocqryinquiry()
{
  struct st_inquiry  stinquiry;
  struct inq_data_s  *inqdata;
  int                rc;
  char               vid[9];
  char               pid[17];
  char               revision[5];
  BYTE               b0, b1, b2, b3, b8;

  printf("Issuing qryinquiry...\n");

  memset(&stinquiry, 0, sizeof(struct st_inquiry));
  rc = ioctl(fd, STIOCQRYINQUIRY, &stinquiry);
  if(rc == 0)
    {
      /* print out data field */
      printf("\nInquiry Data:\n");
      inqdata = &(stinquiry.standard);

      /* evaluate the first byte */
      printf("Peripheral Qualifier--------------");
      b0 = PERIPHERAL_QUALIFIER(inqdata);
      if(b0 == PERIPHERAL_CONNECTED)
        printf("Peripheral connected\n");
      else if(b0 == PERIPHERAL_NOT_CONNECTED)
        printf("Peripheral not connected\n");
      else if(b0 == LUN_NOT_SUPPORTED)
        printf("LUN not supported\n");
      else
        printf("\n");

      printf("Peripheral Device Type-----------");
      b0 = PERIPHERAL_DEVICE_TYPE(inqdata);
      if(b0 == DIRECT_ACCESS)
        printf("Direct access\n");
      else if(b0 == SEQUENTIAL_DEVICE)
        printf("Sequential\n");
      else if(b0 == PRINTER_DEVICE)
        printf("Printer\n");
      else if(b0 == PROCESSOR_DEVICE)
        printf("Processor\n");
      else if(b0 == CD_ROM_DEVICE)
        printf("CD ROM\n");
      else if(b0 == OPTICAL_MEMORY_DEVICE)
        printf("Optical memory\n");
      else if(b0 == MEDIUM_CHANGER_DEVICE)
        printf("Medium changer\n");
      else if(b0 == UNKNOWN)
        printf("Unknown\n");
      else
        printf("\n");

      /* evaluate the second byte */
      printf("Removable Medium Bit-------------");
      b1 = RMB(inqdata);
      if(b1 == FIXED)
        printf("Fixed\n");
      else if(b1 == REMOVABLE)
        printf("Removable\n");
      else
        printf("\n");

      printf("Device Type Modifier-------------0x%02x\n",
             device_type_qualifier(inqdata));

      /* evalutate the third byte */
      printf("ISO Version----------------------0x%02x\n", ISO_Version(inqdata));
      printf("ECMA Version---------------------0x%02x\n", ECMA_Version(inqdata));
      printf("ANSI Version---------------------");
      b2 = ANSI_Version(inqdata);
      if(b2 == 0)
        printf("Nonstandard\n");
      else if(b2 == SCSI1)
        printf("SCSI1\n");
      else if(b2 == SCSI2)
        printf("SCSI2\n");
      else if(b2 == SCSI3)
        printf("SCSI3\n");
      else
        printf("\n");

      /* evaluate the 4th byte */
      printf("Asyschronous Event Notification--");
      b3 = AENC(inqdata);
      if(b3 == TRUE)
        printf("True\n");
      else if(b3 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("Terminate I/O Process Message----");
      b3 = TrmIOP(inqdata);
      if(b3 == TRUE)
        printf("True\n");
      else if(b3 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("Response Data Format-------------");
      b3 = Response_Data_Format(inqdata);
      if(b3 == SCSI1INQ)
        printf("SCSI1 standard\n");
      else if(b3 == CCSINQ)
        printf("CCS standard\n");
      else if(b3 == SCSI2INQ)
        printf("SCSI2 standard\n");
      else
        printf("\n");

      /* evaluate the 5th byte */
      printf("Additional Length----------------0x%02x\n", inqdata->additional_length);

      /* evaluate the 7th byte */
      printf("Medium Changer Mode--------------0x%02x\n", MChngr(inqdata));

      /* evaluate the 8th byte */
      printf("Relative Addressing--------------");
      b8 = RelAdr(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("32 Bit Wide Data Transfer--------");
      b8 = WBus32(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("16 Bit Wide Data Transfer--------");
      b8 = WBus16(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("Synchronous Data Transfer--------");
      b8 = Sync(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("Linked Commands------------------");
      b8 = Linked(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");
      else
        printf("\n");

      printf("Command Queueing-----------------");
      b8 = CmdQue(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");

      printf("Soft Reset-----------------------");
      b8 = SftRe(inqdata);
      if(b8 == TRUE)
        printf("True\n");
      else if(b8 == FALSE)
        printf("False\n");
      else
        printf("\n");

      /* evaluate other bytes */
      strncpy(vid, inqdata->vendor_identification, 8);
      vid[8] = '\0';
      strncpy(pid, inqdata->product_identification, 16);
      pid[16] = '\0';
      strncpy(revision, inqdata->product_revision_level, 4);
      revision[4] = '\0';

      printf("Vendor ID------------------------%s\n", vid);
      printf("Product ID-----------------------%s\n", pid);
      printf("Product revision level-----------%s\n", revision);

      dump_bytes((char*)stinquiry.vendor_specific, 255-sizeof(struct inq_data_s), "Vendor Specific");
    }

  return rc;
}

/*- name:     stioc_report_density_support
 *  function: invoke IOCTL STIOC_REPORT_DENSITY_SUPPORT and print density support data
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_report_density_support(void)
{
  struct report_density_support density;
  int i;
  int rc;

  printf("Issuing Report Density Support for ALL supported media...\n");

  memset(&density, 0, sizeof(struct report_density_support));
  density.media = ALL_MEDIA_DENSITY;

  rc = ioctl(fd, STIOC_REPORT_DENSITY_SUPPORT, &density);
  if(rc != 0)
    return rc;

  printf("Total number of densities reported: %d\n", density.number_reports);
  for (i = 0; i < density.number_reports; i++)
    {
      printf("\n");
      dump_bytes((char*)density.reports[i].density_name, 8, "Density Name");
      dump_bytes((char*)density.reports[i].assigning_org, 8, "Assigning Organization");
      dump_bytes((char*)density.reports[i].description, 20, "Description");

      printf("  Primary Density Code......... %02X\n", density.reports[i].primary_density_code);
      printf("  Secondary Density Code....... %02X\n", density.reports[i].secondary_density_code);

      if(density.reports[i].wrtok)
        printf("  Write OK..................... Yes\n");
      else
        printf("  Write OK..................... No\n");

      if(density.reports[i].dup)
        printf("  Duplicate.................... Yes\n");
      else
        printf("  Duplicate.................... No\n");

      if(density.reports[i].deflt)
        printf("  Default...................... Yes\n");
      else
        printf("  Default...................... No\n");

      printf("  Bits per MM.................. %d\n",density.reports[i].bits_per_mm);
      printf("  Media Width (millimeters).... %d\n",density.reports[i].media_width);
      printf("  Tracks....................... %d\n",density.reports[i].tracks);
      printf("  Capacity (megabytes)......... %d\n",density.reports[i].capacity);
      if(opcode)
        {
          printf("\nHit <enter> to continue...");
          getchar();
        }
    }

  /* check if the drive is loaded */
  printf("\n");
  rc = sioc_test_unit_ready();
  if (rc != 0)
    return rc;

  /* drive is loaded; issue report density support for current media */
  printf("\nIssuing Report Density Support for CURRENT media...\n");

  density.media = CURRENT_MEDIA_DENSITY;

  rc = ioctl(fd, STIOC_REPORT_DENSITY_SUPPORT, &density);
  if(rc != 0)
    return rc;

  for (i = 0; i < density.number_reports; i++)
    {
      printf("\n");
      dump_bytes((char*)density.reports[i].density_name, 8, "Density Name");
      dump_bytes((char*)density.reports[i].assigning_org, 8, "Assigning Organization");
      dump_bytes((char*)density.reports[i].description, 20, "Description");

      printf("  Primary Density Code......... %02X\n",density.reports[i].primary_density_code);
      printf("  Secondary Density Code....... %02X\n",density.reports[i].secondary_density_code);

      if(density.reports[i].wrtok)
        printf("  Write OK..................... Yes\n");
      else
        printf("  Write OK..................... No\n");

      if(density.reports[i].dup)
        printf("  Duplicate.................... Yes\n");
      else
        printf("  Duplicate.................... No\n");

      if(density.reports[i].deflt)
        printf("  Default...................... Yes\n");
      else
        printf("  Default...................... No\n");

      printf("  Bits per MM.................. %d\n", density.reports[i].bits_per_mm);
      printf("  Media Width (millimeters).... %d\n", density.reports[i].media_width);
      printf("  Tracks....................... %d\n", density.reports[i].tracks);
      printf("  Capacity (megabytes)......... %d\n", density.reports[i].capacity);
    }

  return rc;
}

/*- name:     stioc_locate
 *  function: invoke IOCTL STIOC_LOCATE using global variable blockid
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_locate(void)
{
  int rc;

  printf("\nLocating to Block ID %u (0x%08X)...\n", blockid, blockid);
  rc = ioctl (fd, STIOC_LOCATE, blockid);
  return rc;
}

/*- name:     stioc_read_position
 *  function: invoke IOCTL STIOC_READ_POSITION
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_read_position(void)
{
  int rc;

  printf("Reading tape position....\n");

  rc = ioctl(fd, STIOC_READ_POSITION, &blockid);
  if(rc == 0)
    printf("\n    Current Block ID ............... %u (0x%08X)\n", blockid, blockid);
  return rc;
}


/*- name:     list_tape
 *  function: print contents of tape
 *  argument: void
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int list_tape()
{
  struct stsense_s stsense;
  struct mtop      mmtop;
  char             *rbuf;
  int              save_errno = 0;
  int              fm         = 0;
  int              count      = 0;
  int              bytes, transfer_size;
  int              rc;

  /* rewind tape */
  rc = mtioctop(MTREW, 0);
  if(rc != 0)
    return rc;

  /* get the maximum scsi transfer size */
  rc = stiocqryp();
  if(rc != 0)
    return rc;

  transfer_size = parms.max_scsi_xfer;

  /* set actual block size to variable */
  printf("Setting block size to variable...\n");
  mmtop.mt_op = MTSETBLK;
  mmtop.mt_count = 0;
  rc = ioctl(fd, MTIOCTOP, &mmtop);

  if(rc != 0)
    return rc;

  rbuf = malloc(transfer_size);
  if(rbuf == NULL)
    {
      errno = ENOMEM;
      return -2;
    }

  printf("Scanning tape...\n");

  while(1)
    {
      bytes = read(fd, rbuf, transfer_size);
      if(bytes == -1)
        {
          rc = -1;
          break;
        }
      if(bytes == 0)
        {
          fm++;
          count = 0;
          printf("*** Filemark %d ***\n", fm);
        }
      else
        {
          count++;
          printf("Record %d  Length %d\n", count, bytes);
        }
    } /* end of while(1) */

  if(bytes == -1)
    {
      save_errno = errno;
      stsense.sense_type = LASTERROR;

      /* check for normal end of data, sense key 8 and asc/ascq=0005 */
      rc = ioctl(fd, STIOCQRYSENSE, &stsense);
      if(rc == 0 && stsense.len > 12 &&
          stsense.sense[2] == 0x08 && stsense.sense[12] == 0x00 &&
          stsense.sense[13] == 0x05)
        {
          printf("\nTape listing completed.\n");
          errno = 0;
        }
      else
        {
          printf("*** Unexpected error listing tape ***\n\n");
          errno = save_errno;
          rc = -1;
        }
    }

  if(rbuf)
    free(rbuf);
  return rc;
}

int qry_encryption_status(void)
{
   int rc = 0;

   printf("issuing query encryption status...\n");
   memset(&encryption_status_t, 0, sizeof(struct encryption_status));
   rc = ioctl(fd, GET_ENCRYPTION_STATE, &encryption_status_t);
   if(rc == 0)
   {
      if(encryption_status_t.encryption_capable)
          printf("Encryption Capable...... Yes\n");
      else
          printf("Encryption Capable...... No\n");
      switch(encryption_status_t.encryption_method)
      {
      case METHOD_NONE:
              printf("Encryption Method....... None\n");
              break;
      case METHOD_LIBRARY:
              printf("Encryption Method....... Library\n");
              break;
      case METHOD_SYSTEM:
              printf("Encryption Method....... System\n");
              break;
      case METHOD_APPLICATION:
              printf("Encryption Method....... Application\n");
              break;
      case METHOD_CUSTOM:
              printf("Encyrpiton Method....... Custom\n");
              break;
      case METHOD_UNKNOWN:
              printf("Encryption Method....... Unknown\n");
              break;

      default:
              printf("Encrption Method....... Error\n");
      }

      switch(encryption_status_t.encryption_state)
      {
      case STATE_OFF:
              printf("Encryption State........ Off\n");
              break;
      case STATE_ON:
              printf("Encryption State........ On\n");
              break;
      case STATE_NA:
              printf("Encryption State........ NA\n");
              break;

      default:
              printf("Encryption State...... Error\n");
      }
   }

   return rc;
}


/**************************************************************************************/
/*               Service Aid Command Functions                                        */
/**************************************************************************************/

/*- name:     stioc_read_dump
 *  function: use IOCTL SCSI_PASS_THROUGH to read dump into a file
 *  argument: fname -- file name to dump into
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int stioc_read_dump(char *fname)
{
  struct _SCSIPassThrough spt;
  struct log_sense_page logdata;
  long long               data_length, buf_offset;
  int                     dumpfd = -1;
  int                     transfer_size, num_transfers, excess_transfer;
  int                     rc, dev_id;
  int                     bytes;
  int                     i;

  printf("Issuing read dump...\n");

  /* get the drive ID */
  dev_id = search_drive_id();

  if(dev_id == -1)
      return -1;

  if(dev_id == IBM_INVALID_DEVICE)
    {
      errno = 999;
      return -2;
    }

  if(dev_id == IBM_3576)
  {
          printf("read dump is not supported for 3576 library.\n");
          return -2;
  }

  /* some libraries use log sense pages for library logs */
  memset(&logdata, 0, sizeof(struct log_sense_page));
  if(dev_id == IBM_3573)
    {
      logdata.page_code = 0x33;
      rc = ioctl(fd, SIOC_LOG_SENSE_PAGE, &logdata);
      if(rc == 0)
        {
           /* open the dump file */
           dumpfd = open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
           if(dumpfd < 0)
             {
               perror("Open dump file failed.");
               return -2;
             }
        }
      /* write buffer data into dump file */
      bytes = write(dumpfd, &logdata.data[4], logdata.len-4);
      if(bytes == -1)
        {
           close(dumpfd);
           return -1;
        }
       printf("Wrote %d bytes for log sense page 0x33.\n", bytes);
       return 0;
    }
   if(dev_id == IBM_3577)
    {
      logdata.page_code = 0x06;
      rc = ioctl(fd, SIOC_LOG_SENSE_PAGE, &logdata);
      if(rc == 0)
        {
           /* open the dump file */
           dumpfd = open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
           if(dumpfd < 0)
             {
               perror("Open dump file failed.");
               return -2;
             }
        }
      /* write buffer data into dump file */
      bytes = write(dumpfd, &logdata.data[4], logdata.len-4);
      if(bytes == -1)
        {
           close(dumpfd);
           return -1;
        }
       printf("Wrote %d bytes for log sense page 0x06.\n", bytes);
       return 0;
    }


  /* get maximum transfer size */
  transfer_size = MAX_TRANSFER_SIZE;

  /* initialize spt */
  memset(&spt, 0, sizeof(struct _SCSIPassThrough));

  if(dev_id != IBM_3583)
    {
      spt.CommandLength = CDB_10_GENERIC;
      spt.DataDirection = SCSI_DATA_IN;
      spt.TimeOut       = SCSI_PASSTHRU_TIMEOUT_VALUE;

      spt.CDB[0]  = SCSIOP_READ_BUFFER;
      if(dev_id >= IBM_3580 && exec_device == TAPE)
        {
          spt.CDB[1] = 3;        /* Just ask the descriptor to be returned */
          spt.CDB[2] = 1;        /* Buffer ID is Dump Data (Read Only) */
        }
      else if(dev_id == IBM_3584 || dev_id == IBM_3581)
        {
          spt.CDB[1] = 3;        /* Just ask the descriptor to be returned */
          spt.CDB[2] = 1;        /* Buffer ID is Dump Data (Read Only) */
        }
      else if(dev_id == IBM_3582)
        {
          spt.CDB[1] = 3;        /* Just ask the descriptor to be returned */
          spt.CDB[2] = 0x0A;     /* Buffer ID is Error Event Log(Read Only) */
        }
      else
        {
          spt.CDB[1] = 3;        /* Just ask the descriptor to be returned */
          spt.CDB[2] = 0;        /* Buffer ID is Dump Data (Read Only) */
        }
      spt.CDB[8] = 4;            /* Allocation length is 4 bytes */

      spt.BufferLength = 4;
      spt.Buffer = malloc(sizeof(char) * spt.BufferLength);
      if(spt.Buffer == NULL)
        {
          errno = ENOMEM;
          return -2;
        }
      memset(spt.Buffer, 0, spt.BufferLength);

      /* invoke ioctl to get buffer capacity */
      rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
      if(rc != 0){
        if(DEBUG)
          printf("Get buffer capacity failed.\n");

        free(spt.Buffer);
        return rc;
      }

      /* check returned status */
      if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
          || spt.DriverStatus)
        {
          printf("Error in returned status.\n");
          free(spt.Buffer);
          errno = 999;

          if(spt.SenseDataValid)
            printf("SenseKey %02x, ASC %02x, ASCQ %02x\n",
                   spt.SenseKey, spt.ASC, spt.ASCQ);

          return -2;
        }


      /* compute the buffer capacity */
      data_length = (spt.Buffer[1] << 16) + (spt.Buffer[2] << 8) + (int)spt.Buffer[3];
      free(spt.Buffer);
      spt.Buffer = NULL;

      /* open dump file for write only*/
      dumpfd    = open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
      if(dumpfd < 0)
        {
          perror("Open dump file failed.");
          return -2;
        }

      /* get the total number of transfers */
      num_transfers   = data_length / transfer_size;
      excess_transfer = data_length % transfer_size;
      if(excess_transfer)
        num_transfers += 1;

      printf("Total dump data length is %lld.\n", data_length);
      printf("Total number of transfers is %d.\n", num_transfers);

      /* start to transfer data */
      buf_offset = 0;
      i = 0;
      printf("Transfering dump data...\n");
      while(num_transfers)
        {
          i++;
          printf("Current transfer number is %d.\n", i);

          memset(&spt, 0, sizeof(struct _SCSIPassThrough));

          spt.CommandLength = CDB_10_GENERIC;
          spt.DataDirection = SCSI_DATA_IN;
          spt.TimeOut = SCSI_PASSTHRU_TIMEOUT_VALUE;

          spt.CDB[0] = SCSIOP_READ_BUFFER;
          if(dev_id >= IBM_3580 && exec_device == TAPE)
            {
              spt.CDB[1] = 2;        /* Just ask data to be returned */
              spt.CDB[2] = 1;        /* Buffer ID is Dump Data (Read Only) */
            }
          else if(dev_id == IBM_3581)
            {
              spt.CDB[1] = 2;        /* Just ask data to be returned */
              spt.CDB[2] = 1;        /* Buffer ID is Dump Data (Read Only) */
            }
          else if(dev_id == IBM_3584)
            {
              spt.CDB[1] = 1;        /* Just ask data to be returned */
              spt.CDB[2] = 1;        /* Buffer ID is Dump Data (Read Only) */
            }
          else if(dev_id == IBM_3582)
            {
              spt.CDB[1] = 2;        /* Just ask data to be returned */
              spt.CDB[2] = 0x0A;     /* Buffer ID is Error Event Log (Read Only) */
            }
          else
            {
              spt.CDB[1] = 1;        /* Just ask data to be returned */
              spt.CDB[2] = 0;        /* Buffer ID is Dump Data (Read Only) */
            }

          /* Buffer Offset is set by buf_offset */
          spt.CDB[3] = (unchar)(buf_offset >> 16);
          spt.CDB[4] = (unchar)(buf_offset >> 8);
          spt.CDB[5] = (unchar)(buf_offset & 0xFF);

          /* Allocation Length is transfer_size or excess_transfer*/
          if(excess_transfer && num_transfers == 1)
            {
              spt.CDB[6] = (unchar)(excess_transfer >> 16);
              spt.CDB[7] = (unchar)(excess_transfer >> 8);
              spt.CDB[8] = (unchar)(excess_transfer & 0xFF);
            }
          else
            {
              spt.CDB[6] = (unchar)(transfer_size >> 16);
              spt.CDB[7] = (unchar)(transfer_size >> 8);
              spt.CDB[8] = (unchar)(transfer_size & 0xFF);
            }

          if(excess_transfer && num_transfers == 1)
            spt.BufferLength = excess_transfer;
          else
            spt.BufferLength = transfer_size;

          spt.Buffer = malloc(sizeof(char) * spt.BufferLength);
          if(spt.Buffer == NULL)
            {
              errno = ENOMEM;
              close(dumpfd);
              return -2;
            }
          memset(spt.Buffer, 0, spt.BufferLength);

          /* invoke ioctl to get data */
          rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
          if(rc != 0)
            {
              free(spt.Buffer);
              close(dumpfd);
              return rc;
            }

          /* check returned status */
          if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
              || spt.DriverStatus)
            {
              printf("Error in returned status.\n");
              free(spt.Buffer);
              close(dumpfd);
              errno = 999;

              if(spt.SenseDataValid)
                printf("SenseKey %02x, ASC %02x, ASCQ %02x\n",
                       spt.SenseKey, spt.ASC, spt.ASCQ);

              return -2;
            }

          /* write buffer data into dump file */
          bytes = write(dumpfd, spt.Buffer, spt.BufferLength);
          if(bytes == -1)
            {
              free(spt.Buffer);
              close(dumpfd);
              return -1;
            }

          printf("Wrote %d bytes.\n", bytes);

          if(bytes != spt.BufferLength)
            {
              printf("Error in writing dump file. Bytes written do not match buffer length.\n");
              errno = 999;
              free(spt.Buffer);
              close(dumpfd);
              return -2;
            }

          /* update offset and num_transfers, free buffer */
          buf_offset += transfer_size;
          num_transfers -= 1;
          free(spt.Buffer);
        } /* end of while(num_transfers) */

      close(dumpfd);
    } /* dev_id is not IBM_3583 */
  else
    {
      int readingbuffer = 2;
      int bufferID;
      while(readingbuffer)
        {
          if(readingbuffer == 2)
            bufferID = 0x09;
          else
            bufferID = 0x0A;

           spt.CommandLength = CDB_10_GENERIC;
           spt.DataDirection = SCSI_DATA_IN;
           spt.TimeOut       = SCSI_PASSTHRU_TIMEOUT_VALUE;

           spt.CDB[0] = SCSIOP_READ_BUFFER;
           spt.CDB[1] = 3;
           spt.CDB[2] = bufferID;
           spt.CDB[8] = 4;

           spt.BufferLength = 4;
           spt.Buffer = malloc(sizeof(char) * spt.BufferLength);
           if(spt.Buffer == NULL)
             {
               errno = ENOMEM;
               return -2;
             }
           memset(spt.Buffer, 0, spt.BufferLength);

           /* invoke ioctl to get buffer capacity */
           rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
           if(rc != 0){
             if(DEBUG)
               printf("Get buffer capacity failed.\n");

             free(spt.Buffer);
             return rc;
           }

           /* check returned status */
           if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
               || spt.DriverStatus)
            {
              printf("Error in returned status.\n");
              free(spt.Buffer);
              errno = 999;

              if(spt.SenseDataValid)
                printf("SenseKey %02x, ASC %02x, ASCQ %02x\n",
                       spt.SenseKey, spt.ASC, spt.ASCQ);

                return -2;
            }

           /* compute the buffer capacity */
           data_length = (spt.Buffer[1] << 16) + (spt.Buffer[2] << 8) + (int)spt.Buffer[3];
           free(spt.Buffer);
           spt.Buffer = NULL;

           /* open dump file for write only*/
           if(readingbuffer == 2)
             {
               dumpfd = open(fname, O_WRONLY|O_CREAT|O_TRUNC, 0666);
               if(dumpfd < 0)
                 {
                   perror("Open dump file failed.");
                   return -2;
                 }
             }

           /* get the total number of transfers */
           num_transfers   = data_length / transfer_size;
           excess_transfer = data_length % transfer_size;
           if(excess_transfer)
             num_transfers += 1;

           printf("Total dump data length is %lld.\n", data_length);
           printf("Total number of transfers is %d.\n", num_transfers);

           /* start to transfer data */
           buf_offset = 0;
           i = 0;
           printf("Transfering dump data...\n");
           while(num_transfers)
             {
               i++;
               printf("Current transfer number is %d.\n", i);

               memset(&spt, 0, sizeof(struct _SCSIPassThrough));

               spt.CommandLength = CDB_10_GENERIC;
               spt.DataDirection = SCSI_DATA_IN;
               spt.TimeOut = SCSI_PASSTHRU_TIMEOUT_VALUE;

               spt.CDB[0] = SCSIOP_READ_BUFFER;
               spt.CDB[1] = 2;    /* request data to be returned */
               spt.CDB[2] = bufferID;

               /* Buffer Offset is set by buf_offset */
               spt.CDB[3] = (unchar)(buf_offset >> 16);
               spt.CDB[4] = (unchar)(buf_offset >> 8);
               spt.CDB[5] = (unchar)(buf_offset & 0xFF);

               /* Allocation Length is transfer_size or excess_transfer*/
               if(excess_transfer && num_transfers == 1)
                 {
                   spt.CDB[6] = (unchar)(excess_transfer >> 16);
                   spt.CDB[7] = (unchar)(excess_transfer >> 8);
                   spt.CDB[8] = (unchar)(excess_transfer & 0xFF);
                 }
               else
                 {
                   spt.CDB[6] = (unchar)(transfer_size >> 16);
                   spt.CDB[7] = (unchar)(transfer_size >> 8);
                   spt.CDB[8] = (unchar)(transfer_size & 0xFF);
                 }

               if(excess_transfer && num_transfers == 1)
                 spt.BufferLength = excess_transfer;
               else
                 spt.BufferLength = transfer_size;

               spt.Buffer = malloc(sizeof(char) * spt.BufferLength);
               if(spt.Buffer == NULL)
                 {
                   errno = ENOMEM;
                   close(dumpfd);
                   return -2;
                 }
               memset(spt.Buffer, 0, spt.BufferLength);

               /* invoke ioctl to get data */
               rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
               if(rc != 0)
                 {
                   free(spt.Buffer);
                   close(dumpfd);
                   return rc;
                 }

               /* check returned status */
               if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
                   || spt.DriverStatus)
                 {
                   printf("Error in returned status.\n");
                   free(spt.Buffer);
                   close(dumpfd);
                   errno = 999;

                   if(spt.SenseDataValid)
                     printf("SenseKey %02x, ASC %02x, ASCQ %02x\n",
                            spt.SenseKey, spt.ASC, spt.ASCQ);

                   return -2;
                 }

               /* write buffer data into dump file */
               bytes = write(dumpfd, spt.Buffer, spt.BufferLength);
               if(bytes == -1)
                 {
                   free(spt.Buffer);
                   close(dumpfd);
                   return -1;
                 }

               printf("Wrote %d bytes.\n", bytes);

               if(bytes != spt.BufferLength)
                 {
                   printf("Error in writing dump file. Bytes written mismatch buffer length.\n");
                   errno = 999;
                   free(spt.Buffer);
                   close(dumpfd);
                   return -2;
                 }

               /* update offset and num_transfers, free buffer */
               buf_offset += transfer_size;
               num_transfers -= 1;
               free(spt.Buffer);
             } /* end of while(num_transfers) */

           readingbuffer -= 1;
        }/* end of while(readingbuffer) */

      close(dumpfd);
    } /* dev_id is IBM_3583 */

  return rc;
}

/*- name:     stioc_force_dump
 *  function: invoke IOCTL SCSI_PASS_THROUGH to force dump
 *  argument: none
 *  return:   0 if succeeded, -1 or -2 if failed
 */
int stioc_force_dump()
{
  struct _SCSIPassThrough spt;
  int                     dev_id;
  int                     rc;

  printf("Issuing force dump...\n");

  /* get the drive ID */
  dev_id = search_drive_id();

  if(dev_id == -1)
    return -1;

  if(dev_id == IBM_INVALID_DEVICE)
    {
      errno = 999;
      return -2;
    }

  /* initialize spt */
  memset(&spt, 0, sizeof(struct _SCSIPassThrough));

  spt.CDB[0]        = SCSIOP_SEND_DIAGNOSTIC;
  spt.CommandLength = CDB_6_GENERIC;
  spt.DataDirection = SCSI_DATA_OUT;
  spt.TimeOut       = SCSI_PASSTHRU_TIMEOUT_VALUE;

  spt.BufferLength = sizeof(char) * 8;        /* force dump list size */
  spt.Buffer = malloc(spt.BufferLength);
  if(spt.Buffer == NULL)
    {
      errno = ENOMEM;
      return -2;
    }
  memset(spt.Buffer, 0, spt.BufferLength);

  spt.CDB[1]       = 0x10;              /* PF bit is set to 1 */
  spt.CDB[3]       = 0x00;
  spt.CDB[4]       = 0x08;              /* parameter length is 0x0008 */
  spt.Buffer[0]    = 0x80;              /* page code */
  spt.Buffer[2]    = 0x00;
  spt.Buffer[3]    = 0x04;              /* page length */
  spt.Buffer[4]    = 0x01;
  spt.Buffer[5]    = 0x60;              /* diagnostic id */

  /* invoke ioctl to send a page */
  rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
  if(rc != 0)
    {
      free(spt.Buffer);
      return rc;
    }

  /* check returned status */
  if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
      || spt.DriverStatus)
    {
      printf("Error in returned status.\n");
      free(spt.Buffer);
      errno = 999;

      if(spt.SenseDataValid)
        printf("SenseKey %02x, ASC %02x, ASCQ %02x\n",
               spt.SenseKey, spt.ASC, spt.ASCQ);

      return -2;
    }

  free(spt.Buffer);
  return rc;
}


/*- name:     stioc_load_ucode
 *  function: invoke IOCTL SCSI_PASS_THROUGH to write an image file into drive
 *  argument: ucode -- name of image file
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int stioc_load_ucode(char *ucode)
{
  struct _SCSIPassThrough spt;
  struct mtop             mt;
  struct stat             stbuf;         /* for inode of image file */
  int                     image;
  long long               data_length, buf_offset;
  int                     transfer_size, num_transfers, excess_transfer, count;
  int                     dev_id, bytes;
  int                     retry = 0;     /* for IBM_3583 */
  int                     rc;
  int                     i;

  /* get the drive ID */
  dev_id = search_drive_id();

  if(dev_id == -1)
     return -1;

  if(dev_id == IBM_INVALID_DEVICE)
    {
      errno = 999;
      return -2;
    }

  /* get maximum transfer size */
  if(exec_device == TAPE)
    {
      rc = stiocqryp();
      if(rc != 0)
        return rc;

      transfer_size = min(parms.max_blksize, parms.max_scsi_xfer);
      transfer_size = min(transfer_size, MAX_TRANSFER_SIZE);
    }
  else
    transfer_size = MAX_TRANSFER_SIZE;

  if(dev_id == IBM_3573)
          transfer_size = 65535;

  /* conduct test unit ready */
  for(i = 0; i < 35; i++)
    ioctl(fd, SIOC_TEST_UNIT_READY, NULL);

  /* open image file */
  image = open(ucode, O_RDONLY);
  if(image < 0)
    return -2;

  /* get image size*/
  stat(ucode, &stbuf);
  if(S_ISDIR(stbuf.st_mode))
    {
      printf("%s is a directory, not a file.\n", ucode);
      close(image);
      errno = 999;
      return -2;
    }

  data_length = (long long)(stbuf.st_size);

  /* make sure image size is ok */
  if(dev_id == IBM_3581 && exec_device == CHANGER
     && data_length > MORGAN_MICROCODESIZE)
    {
      printf("Error: The size of image file does not match IBM 3581 spec!\n");
      errno = 999;
      close(image);
      return -2;
    }

  /* get total number of transfers */
  num_transfers = data_length / transfer_size;
  excess_transfer = data_length % transfer_size;
  if(excess_transfer)
    num_transfers += 1;

  printf("Image size is %lld.\n", data_length);
  printf("Total number of transfers is %d.\n", num_transfers);

  /* start to transfer data */
  printf("Loading image data, please wait...\n");
  buf_offset = 0;
  i = 0;
  count = num_transfers;
  while(count)
    {
      i++;
      printf("Current transfer number is %d.\n", i);

      memset(&spt, 0, sizeof(struct _SCSIPassThrough));

      spt.CommandLength = CDB_10_GENERIC;
      spt.DataDirection = SCSI_DATA_OUT;
      spt.TimeOut       = SCSI_PASSTHRU_TIMEOUT_VALUE;

      spt.CDB[0] = SCSIOP_WRITE_BUFFER;
      if(dev_id >= IBM_3580 && exec_device == TAPE)
        {
          spt.CDB[1] = 2;        /* Vendor specific mode */
          spt.CDB[2] = 4;        /* Buffer ID */
          spt.CDB[3] = (unchar)(buf_offset >> 16);
          spt.CDB[4] = (unchar)(buf_offset >> 8);
          spt.CDB[5] = (unchar)(buf_offset & 0xFF);
        } /* IBM LTO 3580 microcode updata */
      else if(dev_id == IBM_3581)
        {
          spt.CDB[1] = 5;        /* Vendor specific mode */
          spt.CDB[2] = 0;        /* Buffer ID */
          spt.CDB[3] = 0;
          spt.CDB[4] = 0;
          spt.CDB[5] = 0;
        } /* IBM 3581 Autoloader microcode update */
      else if(dev_id == IBM_3584
              || dev_id == IBM_3583
              || dev_id == BDT_3581
              || dev_id == IBM_3576)
        {
          spt.CDB[1] = 5;        /* Vendor specific mode */
          spt.CDB[2] = 0;        /* Buffer ID */
          spt.CDB[3] = (unchar)(buf_offset >> 16);
          spt.CDB[4] = (unchar)(buf_offset >> 8);
          spt.CDB[5] = (unchar)(buf_offset & 0xFF);
        } /* IBM 3584, IBM 3583, BDT 3581 microcode update */
      else if(dev_id == IBM_3573)
        {
          spt.CDB[1] = 4;        /* Vendor specific mode */
          spt.CDB[2] = 0;        /* Buffer ID */
          spt.CDB[3] = (unchar)(buf_offset >> 16);
          spt.CDB[4] = (unchar)(buf_offset >> 8);
          spt.CDB[5] = (unchar)(buf_offset & 0xFF);
        }
      else if(dev_id == IBM_3577)
        {
          spt.CDB[1] = 2;       /* Data mode */
          spt.CDB[2] = 0;       /* Buffer ID */
          spt.CDB[3] = (unchar)(buf_offset >> 16);
          spt.CDB[4] = (unchar)(buf_offset >> 8);
          spt.CDB[5] = (unchar)(buf_offset & 0xFF);
        }

      else
        {
          spt.CDB[1] = 5;        /* Vendor specific mode */
          spt.CDB[2] = 1;        /* Buffer ID */
          spt.CDB[3] = (unchar)(buf_offset >> 16);
          spt.CDB[4] = (unchar)(buf_offset >> 8);
          spt.CDB[5] = (unchar)(buf_offset & 0xFF);
        } /* Non-LTO devices: 3590 and 3592  microcode update
           * Also IBM LTO 3582 library microcode update */


      /* Allocation length is transfer_size or excess_transfer*/
      if(excess_transfer && count == 1)
        {
          spt.CDB[6] = (unchar)(excess_transfer >> 16);
          spt.CDB[7] = (unchar)(excess_transfer >> 8);
          spt.CDB[8] = (unchar)(excess_transfer & 0xFF);
        }
      else
        {
          spt.CDB[6] = (unchar)(transfer_size >> 16);
          spt.CDB[7] = (unchar)(transfer_size >> 8);
          spt.CDB[8] = (unchar)(transfer_size & 0xFF);
        }

      if(excess_transfer && count == 1)
        spt.BufferLength = excess_transfer;
      else
        spt.BufferLength = transfer_size;

      spt.Buffer = malloc(sizeof(char) * spt.BufferLength);
      if(spt.Buffer == NULL)
        {
          errno = ENOMEM;
          close(image);
          return -2;
        }
      memset(spt.Buffer, 0, spt.BufferLength);

      /* read image file and write into buffer */
      bytes = read(image, spt.Buffer, spt.BufferLength);
      if(bytes == -1)
        {
          perror("read image file failed.");
          free(spt.Buffer);
          close(image);
          return -1;
        }
      if(bytes != spt.BufferLength)
        {
          printf("Error in reading image file.\n");
          errno = 999;
          free(spt.Buffer);
          close(image);
          return -2;
        }

      /* invoke ioctl to write buffer*/
      rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);

      /* the old 3583 with SDG will reset library after the first write buffer */
      if(dev_id == IBM_3583)
       {
         if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
            || spt.DriverStatus)
           retry = 0; //need retry the first write buffer transfer
         else
           retry = 1; //first retry succeed, no need to retry
       }

      /* retry after first write buffer if it is IBM_3583 */
      if(dev_id == IBM_3583 && retry == 0)
        {
          retry = 1;
          printf("Retry load microcode for 3583....\n");
          sleep(20);

          /* issue one Test_Unit_Ready command to retrieve the check condition
           * for power-on reset
           */
          ioctl(fd, SIOC_TEST_UNIT_READY, NULL);

          /* now reissure the Pass_Through command */
          rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
          if(rc != 0)
            {
              free(spt.Buffer);
              close(image);
              return rc;
            }
        }

      /* check returned status */
      if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus
          || spt.DriverStatus)
        {
          printf("Error in returned status.\n");
          free(spt.Buffer);
          close(image);
          errno = 999;

          if(spt.SenseDataValid)
            printf("SenseKey %02x, ASC %02x, ASCQ %02x\n",
                   spt.SenseKey, spt.ASC, spt.ASCQ);

          return -2;
        }

      /* update offset and num_transfers, free buffer */
      buf_offset += transfer_size;
      count -= 1;
      free(spt.Buffer);
    }
  close(image);

  /* send mode id 5 for 3577 */
  if(dev_id == IBM_3577)
  {
    spt.CommandLength = CDB_10_GENERIC;
    spt.DataDirection = SCSI_DATA_NONE;
    spt.TimeOut = SCSI_PASSTHRU_TIMEOUT_VALUE;
    spt.CDB[0] = SCSIOP_WRITE_BUFFER;
    spt.CDB[1] = 5;
    spt.CDB[2] = 0;
    spt.CDB[3] = 0;
    spt.CDB[4] = 0;
    spt.CDB[5] = 0;
    spt.CDB[6] = 0;
    spt.CDB[7] = 0;
    spt.CDB[8] = 0;
    spt.Buffer = NULL;
    spt.BufferLength = 0;

    printf("issuing write buffer with mode id 5 for 3577 library...\n");
    rc = ioctl(fd, SCSI_PASS_THROUGH, &spt);
    if(rc != 0)
      return rc;
    if(spt.TargetStatus || spt.MessageStatus || spt.HostStatus || spt.DriverStatus)
    {
       printf("error in returned status.\n");
       errno = 999;
       if(spt.SenseDataValid)
       {
          printf("SenseKey: %02x, ASC: %02x, ASCQ: %02x\n",
                  spt.SenseKey, spt.ASC, spt.ASCQ);
       }
       return -2;
    }
  }

  /* wait for device to reset */
  printf("Wait 3 minutes to reset device...\n");
  sleep(3*60);
  rc = ioctl(fd, SIOC_TEST_UNIT_READY, NULL);

  /* wait for bus to reset */
  printf("Wait 2 minutes to reset bus...\n");
  sleep(2*60);
  for(i = 0; i<30; i++)
    ioctl(fd, SIOC_TEST_UNIT_READY, NULL);

  /* unlock the drive door */
  mt.mt_op = MTUNLOCK;
  mt.mt_count = 0;
  ioctl(fd, MTIOCTOP, &mt);

  /* issue inquiry */
  rc = sioc_inquiry();

  if((rc == 0) && (errno != 0))
    errno = 0;

  return rc;
}

/*- name:     stioc_reset_drive_Atape
 *  function: invoke IOCTL STIOC_RESET_DRIVE
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_reset_drive_Atape()
{
  int rc;

  printf("Resetting drive...\n");
  rc = ioctl(fd, STIOC_RESET_DRIVE, NULL);
  return rc;
}

/*- name:     stioc_read_keys
 *  function: invoke STIOC_READ_RESERVEKEYS ioctl and print out all the persistent reservation keys
 *            registered to the drive
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 */
int stioc_read_keys(void)
{
  int rc, i;
  struct read_keys key_data;
  int num_keys;

  printf("issuing read keys......\n");

  /* issue the first ioctl to get the number of all existing reservation keys */
  memset(&key_data, 0, sizeof(struct read_keys));
  key_data.reserve_key_list = NULL;

  rc = ioctl(fd, STIOC_READ_RESERVEKEYS, &key_data);
  if(rc < 0)
    return rc;
  num_keys = key_data.length / RESERVE_KEY_LENGTH;

  /* issue the second ioctl to get each key value */
  if(num_keys)
    {
      memset(&key_data, 0, sizeof(struct read_keys));
      key_data.reserve_key_list = (char*)malloc(key_data.length);
      if (key_data.reserve_key_list == NULL)
        {
          errno = ENOMEM;
          return -2;
        }
      memset(key_data.reserve_key_list, 0, key_data.length);
      rc = ioctl(fd, STIOC_READ_RESERVEKEYS, &key_data);
      if(rc == 0)
        {
          /* print out each key value */
          for(i=0; i<num_keys; i++)
            {
              printf("Reservation Key %d", i+1);

              dump_bytes((char*)(key_data.reserve_key_list + i*RESERVE_KEY_LENGTH), RESERVE_KEY_LENGTH, NULL);
            }
        }
      if(key_data.reserve_key_list)
        free(key_data.reserve_key_list);
    }
  else
    printf("No reservation key is registered.\n");

  return rc;
}

/*- name:     stioc_read_reserves
 *  function: invoke IOCTL STIOC_READ_RESERVATIONS and print out all existing reservations in detail
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_read_reservations()
{
  int rc, i;
  struct read_reserves reserve_data;
  struct reserve_descriptor *reserve;
  int num_reserves;
  char type;

  printf("issuing read reservations......\n");

  /* issue the first ioctl to get the number of reservation keys */
  memset(&reserve_data, 0, sizeof(struct read_reserves));
  rc = ioctl(fd, STIOC_READ_RESERVATIONS, &reserve_data);
  if(rc < 0)
    return rc;
  num_reserves = reserve_data.length / sizeof(struct reserve_descriptor);

  /* issue the second ioctl to get the description of each existing reservation */
  if(num_reserves)
    {
      memset(&reserve_data, 0, sizeof(struct read_reserves));
      reserve_data.reserve_list = (struct reserve_descriptor*)malloc(num_reserves * sizeof(struct reserve_descriptor));
      if(reserve_data.reserve_list == NULL)
        {
          errno = ENOMEM;
          return -2;
        }

      rc = ioctl(fd, STIOC_READ_RESERVATIONS, &reserve_data);
      if(rc < 0)
        {
          free(reserve_data.reserve_list);
          return rc;
        }

      /* print out the details of each reservation */
      if(reserve_data.reserve_list)
        reserve = (struct reserve_descriptor*)reserve_data.reserve_list;
      for(i=0; i<num_reserves; i++)
        {
          printf("Reservation Key %d", i+1);

          dump_bytes((char*) (reserve_data.reserve_list + i), RESERVE_KEY_LENGTH, NULL);

          printf("Scope Specific Address..................... %u\n", reserve->scope_spec_addr);

          printf("Scope...................................... %d\n", reserve->scope);
          type = (reserve->type) & 0x0F;
          if (type == 0x03)
            printf("Reservation Type........................... Exclusive Access\n");
          else if (type == 0x06)
            printf("Reservation Type........................... Exclusive Access for Registrants Only\n");
          else
            printf("Reservation Type........................... %d\n", type);

          reserve++;
        }

      if(reserve_data.reserve_list)
        free(reserve_data.reserve_list);
    }

  else
    printf("No reservation is available.\n");
  return rc;
}


/*- name:     stioc_register_key
 *  function: invoke IOCTL STIOC_REGISTER_KEY to register a reservation key
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_register_key()
{
  int rc;

  printf("issuing register a persistent reservation key......\n");
  rc = ioctl(fd, STIOC_REGISTER_KEY, NULL);
  if(rc == 0)
    rc = stioc_read_keys();

  return rc;
}

/*- name:     stioc_remove_registration
 *  function: invoke IOCTL STIOC_REMOVE_REGISTRATION to remove the current host's key
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_remove_registration()
{
  int rc;
   struct read_keys key_data;
  int num_keys;

  /* issue the first ioctl to get the number of all existing reservation keys */
  printf("issuing read keys......\n");

  memset(&key_data, 0, sizeof(struct read_keys));
  key_data.reserve_key_list = NULL;

  rc = ioctl(fd, STIOC_READ_RESERVEKEYS, &key_data);
  if(rc < 0)
    return rc;
  num_keys = key_data.length / RESERVE_KEY_LENGTH;

  if(num_keys)
    {

      printf("issuing remove a persistent reservation key......\n");
      rc = ioctl(fd, STIOC_REMOVE_REGISTRATION, NULL);
      if(rc == 0)
        rc = stioc_read_reservations();
    }
  else
    printf("No reservation key is registered.\n");

  return rc;
}

/*- name:     stioc_clear_all_registrations
 *  function: invoke IOCTL STIOC_CLEAR_ALL_REGISTRATIONS to clear all the keys
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int stioc_clear_all_registrations()
{
  int rc;
  struct read_keys key_data;
  int num_keys;

  /* issue the first ioctl to get the number of all existing reservation keys */
  printf("issuing read keys......\n");

  memset(&key_data, 0, sizeof(struct read_keys));
  key_data.reserve_key_list = NULL;

  rc = ioctl(fd, STIOC_READ_RESERVEKEYS, &key_data);
  if(rc < 0)
    return rc;
  num_keys = key_data.length / RESERVE_KEY_LENGTH;

  if(num_keys)
    {
      printf("issuing clear registration......\n");
      rc = ioctl(fd, STIOC_CLEAR_ALL_REGISTRATIONS, NULL);
      if(rc < 0)
        return rc;

      rc = stioc_read_keys();
    }
  else
    printf("No reservation key is registered.\n");

  return rc;
}



/*****************************************************************************/
/*                Medium Changer Functions                                   */
/*****************************************************************************/

/*- name:     smcioc_element_info
 *  function: invoke medium changer IOCTL SMCIOC_ELEMENT_INFO and print element info
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_element_info()
{
  struct element_info element_info;
  int    rc;

  printf("Getting element information...\n");

  memset(&element_info, 0, sizeof(struct element_info));
  rc = ioctl(fd, SMCIOC_ELEMENT_INFO, &element_info);
  if(rc == 0)
    {
      printf("\n");
      printf("  Number of Robots .............. %d\n",element_info.robots);
      if(element_info.robots)
        printf("  First Robot Address ........... %d\n",element_info.robot_addr);

      printf("  Number of Slots ............... %d\n",element_info.slots);
      if(element_info.slots)
        printf("  First Slot Address ............ %d\n",element_info.slot_addr);

      printf("  Number of Import/Exports ...... %d\n",element_info.ie_stations);
      if(element_info.ie_stations)
        printf("  First Import/Export Address ... %d\n",element_info.ie_addr);

      printf("  Number of Drives .............. %d\n",element_info.drives);
      if(element_info.drives)
        printf("  First Drive Address ........... %d\n",element_info.drive_addr);
      printf("\n");
    }

  return rc;
}

/*- name:     smcioc_move_medium
 *  function: invoke medium changer IOCTOL SMCIOC_MOVE_MEDIUM
 *  argument: source slot and destination slot
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_move_medium (ushort source, ushort dest)
{
  struct move_medium move_medium;
  int    rc;

  memset(&move_medium, 0, sizeof(struct move_medium));
  move_medium.source = source;
  move_medium.destination = dest;

  printf("Moving medium: source %u, destination %u...\n",source,dest);

  rc = ioctl (fd, SMCIOC_MOVE_MEDIUM, &move_medium);
  return rc;
}

/*- name:     smcioc_exchange_medium
 *  function: invoke medium changer IOCTOL SMCIOC_EXCHANGE_MEDIUM
 *  argument: source slot, first and second destionation slotss
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_exchange_medium (ushort source, ushort dest, ushort dest2)
{
  struct exchange_medium exchange_medium;
  int    rc;

  memset(&exchange_medium, 0, sizeof(struct exchange_medium));
  exchange_medium.source = source;
  exchange_medium.destination1 = dest;
  exchange_medium.destination2 = dest2;

  printf("Exchanging medium: source %u, first destination %u, second destination %u...\n",
          source,dest,dest2);

  rc = ioctl (fd, SMCIOC_EXCHANGE_MEDIUM, &exchange_medium);
  return rc;
}


/*- name:     smcioc_load_medium
 *  function: invoke medium changer IOCTOL SMCIOC_LOAD_MEDIUM
 *  argument: source slot
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_load_medium (ushort source)
{
  int rc;

  if(source)
    printf("Loading medium from slot %u...\n", source);
  else
    printf("Loading medium from first full slot...\n");

  rc = ioctl (fd, SMCIOC_LOAD_MEDIUM, source);
  return rc;
}

/*- name:     smcioc_unload_medium
 *  function: invoke medium changer IOCTOL SMCIOC_UNLOAD_MEDIUM
 *  argument: source slot
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_unload_medium (ushort source)
{
  int rc;

  if(source)
    printf("Unloading medium to slot %u...\n", source);
  else
    printf("Unloading medium to first empty slot...\n");

  rc = ioctl (fd, SMCIOC_UNLOAD_MEDIUM, source);
  return rc;
}

/*- name:     smcioc_pos_to_elem
 *  function: invoke medium changer IOCTOL SMCIOC_POS_TO_ELEM
 *  argument: destination slot
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_pos_to_elem (ushort dest)
{
  struct pos_to_elem pos_to_elem;
  int    rc;

  memset(&pos_to_elem, 0, sizeof(struct pos_to_elem));
  pos_to_elem.destination = dest;

  printf("Positioning to element, destination %u...\n",dest);

  rc = ioctl (fd, SMCIOC_POS_TO_ELEM, &pos_to_elem);
  return rc;
}


/*- name:     smcioc_init_elem_stat
 *  function: invoke medium changer IOCTOL SMCIOC_INIT_ELEM_STAT
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_init_elem_stat()
{
  int rc;

  printf("Initializing element status...\n");

  rc = ioctl (fd, SMCIOC_INIT_ELEM_STAT, 0);
  return rc;
}

/*- name:     smcioc_init_elem_stat_range
 *  function: invoke medium changer IOCTOL SMCIOC_INI_ELEM_STAT_RANGE
 *  argument: element address and number of elements
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_init_elem_stat_range(ushort address, ushort number)
{
  int    rc;
  struct element_range elem_range;

  memset(&elem_range, 0, sizeof(struct element_range));
  elem_range.element_address = address;
  elem_range.number_elements = number;

  printf("Initializing element status for address %d, %d elements...\n",address,number);

  rc = ioctl (fd, SMCIOC_INIT_ELEM_STAT_RANGE, &elem_range);
  return rc;
}

/*- name:     smcioc_prevent_medium_removal
 *  function: invoke medium changer IOCTOL SMCIOC_PREVENT_MEDIUM_REMOVAL
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_prevent_medium_removal()
{
  int rc;

  printf("Issuing prevent medium removal...\n");

  rc = ioctl (fd, SMCIOC_PREVENT_MEDIUM_REMOVAL, 0);
  return rc;
}

/*- name:     smcioc_allow_medium_removal
 *  function: invoke medium changer IOCTOL SMCIOC_ALLOW_MEDIUM_REMOVAL
 *  argument: void
 *  return:   0 if succeeded, -1 if failed
 -*/
int smcioc_allow_medium_removal()
{
  int rc;

  printf("Issuing allow medium removal...\n");

  rc = ioctl (fd, SMCIOC_ALLOW_MEDIUM_REMOVAL, 0);
  return rc;
}

/*- name:     smcioc_inventory
 *  function: invoke medium changer IOCTOL SMCIOC_INVENTORY and print inventory info
 *  argument: void
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int smcioc_inventory()
{
  int    i,j;
  int    rc;
  struct element_status *robot_status = NULL;
  struct element_status *slot_status = NULL;
  struct element_status *ie_status = NULL;
  struct element_status *drive_status = NULL;
  struct element_status *elemp;
  struct inventory inventory;
  struct element_info element_info;

  memset(&inventory, 0, sizeof(struct inventory));
  memset(&element_info, 0, sizeof(struct element_info));

  rc = ioctl(fd, SMCIOC_ELEMENT_INFO, &element_info);
  if(rc != 0)
    return rc;

  if(element_info.robots)
    {
    robot_status = malloc(element_info.robots * sizeof(struct element_status));
    if(robot_status == NULL)
      {
      errno = ENOMEM;
      return -2;
      }
    bzero((caddr_t)robot_status,element_info.robots * sizeof(struct element_status));
    }

  if(element_info.slots)
    {
    slot_status = malloc(element_info.slots * sizeof(struct element_status));
    if(slot_status == NULL)
      {
      if(robot_status)
        free(robot_status);
      errno = ENOMEM;
      return -2;
      }
    bzero((caddr_t)slot_status,element_info.slots
          * sizeof(struct element_status));
    }

  if(element_info.drives)
    {
    drive_status = malloc(element_info.drives * sizeof(struct element_status));
    if(drive_status == NULL)
      {
      if(robot_status)
        free(robot_status);
      if(slot_status)
        free(slot_status);
      errno = ENOMEM;
      return -2;
      }
    bzero((caddr_t)drive_status,element_info.drives
          * sizeof(struct element_status));
    }

  if(element_info.ie_stations)
    {
    ie_status = malloc(element_info.ie_stations
                       * sizeof(struct element_status));
    if(ie_status == NULL)
      {
      if(robot_status)
        free(robot_status);
      if(slot_status)
        free(slot_status);
      if(drive_status)
        free(drive_status);
      errno = ENOMEM;
      return -2;
      }
    bzero((caddr_t)ie_status,element_info.ie_stations
          * sizeof(struct element_status));
    }

  inventory.robot_status = robot_status;
  inventory.slot_status = slot_status;
  inventory.ie_status = ie_status;
  inventory.drive_status = drive_status;

  printf("Reading element status...\n");

  rc = ioctl(fd, SMCIOC_INVENTORY, &inventory);
  if(rc == 0)
    {
      j=0;
      elemp = robot_status;
      for (i = 0; i < element_info.robots; i++, elemp++)
        {
          if(opcode && j == 2)
            {
              j=0;
              printf("\nHit <enter> to continue...");
              getchar();
            }

          j++;
          printf("\nRobot Address %d\n",elemp->address);
          if(elemp->except)
            printf("  Robot State .................... Abnormal\n");
          else
            printf("  Robot State .................... Normal\n");
          printf("  ASC/ASCQ ....................... %02X%02X\n",
                 elemp->asc,elemp->ascq);
          if(elemp->full)
            printf("  Media Present .................. Yes\n");
          else
            printf("  Media Present .................. No\n");
          if(elemp->svalid)
            printf("  Source Element Address ......... %d\n",elemp->source);
          else
            printf("  Source Element Address Valid ... No\n");
          if(elemp->invert)
            printf("  Media Inverted ................. Yes\n");
          else
            printf("  Media Inverted ................. No\n");
           printf("  Volume Tag ............. %s\n", (char*)elemp->volume);
         //  dump_bytes((char*)elemp->volume, 36, "Volume Tag");
        }

      elemp = ie_status;
      for (i = 0; i < element_info.ie_stations; i++, elemp++)
        {
          if(opcode && j == 2)
            {
              j=0;
              printf("\nHit <enter> to continue...");
              getchar();
            }

          j++;
          printf("\nImport/Export Station Address %d\n",elemp->address);
          if(elemp->except)
            printf("  Import/Export State ............ Abnormal\n");
          else
            printf("  Import/Export State ............ Normal\n");
          printf("  ASC/ASCQ ....................... %02X%02X\n",
                 elemp->asc,elemp->ascq);
          if(elemp->full)
            printf("  Media Present .................. Yes\n");
          else
            printf("  Media Present .................. No\n");
          if(elemp->inenab)
            printf("  Import Enabled ................. Yes\n");
          else
            printf("  Import Enabled ................. No\n");
          if(elemp->exenab)
            printf("  Export Enabled ................. Yes\n");
          else
            printf("  Export Enabled ................. No\n");
          if(elemp->access)
            printf("  Robot Access Allowed ........... Yes\n");
          else
            printf("  Robot Access Allowed ........... No\n");
          if(elemp->svalid)
            printf("  Source Element Address ......... %d\n",elemp->source);
          else
            printf("  Source Element Address Valid ... No\n");
          if(elemp->invert)
            printf("  Media Inverted ................. Yes\n");
          else
            printf("  Media Inverted ................. No\n");
          printf("  Volume Tag ............. %s\n", (char*)elemp->volume);
         // dump_bytes((char*)elemp->volume, 36, "Volume Tag");
        }

      elemp = drive_status;
      for (i = 0; i < element_info.drives; i++, elemp++)
        {
          /* don't overflow screen if menu mode */
          if(opcode && j == 2)
            {
              j=0;
              printf("\nHit <enter> to continue...");
              getchar();
            }

          j++;
          printf("\nDrive Address %d\n",elemp->address);
          if(elemp->except)
            printf("  Drive State .................... Abnormal\n");
          else
            printf("  Drive State .................... Normal\n");
          printf("  ASC/ASCQ ....................... %02X%02X\n",
                 elemp->asc,elemp->ascq);
          if(elemp->full)
            printf("  Media Present .................. Yes\n");
          else
            printf("  Media Present .................. No\n");
          if(elemp->access)
        printf("  Robot Access Allowed ........... Yes\n");
          else
            printf("  Robot Access Allowed ........... No\n");
          if(elemp->svalid)
        printf("  Source Element Address ......... %d\n",elemp->source);
          else
            printf("  Source Element Address Valid ... No\n");
          if(elemp->invert)
            printf("  Media Inverted ................. Yes\n");
          else
            printf("  Media Inverted ................. No\n");
          if(elemp->notbus)
            printf("  Same Bus as Medium Changer ..... No\n");
          else
            printf("  Same Bus as Medium Changer ..... Yes\n");
          if(elemp->idvalid)
            printf("  SCSI Bus Address ............... %d\n",elemp->scsi);
          else
            printf("  SCSI Bus Address Vaild ......... No\n");
          if(elemp->luvalid)
            printf("  Logical Unit Number ............ %d\n",elemp->lun);
          else
            printf("  Logical Unit Number Valid ...... No\n");
          printf("  Volume Tag ............. %s\n", (char*)elemp->volume);
        //  dump_bytes((char*)elemp->volume, 36, "Volume Tag");
      }

      elemp = slot_status;
      for (i = 0; i < element_info.slots; i++, elemp++)
        {
          if(opcode && j == 2)
            {
              j=0;
            printf("\nHit <enter> to continue...");
            getchar();
            }

          j++;
          printf("\nSlot Address %d\n",elemp->address);
          if(elemp->except)
            printf("  Slot State ..................... Abnormal\n");
          else
            printf("  Slot State ..................... Normal\n");
          printf("  ASC/ASCQ ....................... %02X%02X\n",
                 elemp->asc,elemp->ascq);
          if(elemp->full)
            printf("  Media Present .................. Yes\n");
          else
            printf("  Media Present .................. No\n");
          if(elemp->access)
            printf("  Robot Access Allowed ........... Yes\n");
          else
            printf("  Robot Access Allowed ........... No\n");
          if(elemp->svalid)
            printf("  Source Element Address ......... %d\n",elemp->source);
          else
            printf("  Source Element Address Valid ... No\n");
          if(elemp->invert)
            printf("  Media Inverted ................. Yes\n");
          else
            printf("  Media Inverted ................. No\n");
          printf("  Volume Tag ............. %s\n", (char*)elemp->volume);
         // dump_bytes((char*)elemp->volume, 36, "Volume Tag");
        }
    }

  if(robot_status)
    free(robot_status);
  if(slot_status)
    free(slot_status);
  if(drive_status)
    free(drive_status);
  if(ie_status)
    free(ie_status);

  return rc;
}


/*- name:     smcioc_read_cartridge_location
 *  function: invoke medium changer IOCTL SMCIOC_READ_CARTRIDGE_LOCATION and print
 *  argument: void
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int smcioc_read_cartridge_location()
{
  int    i = 0, j = 0;
  int    rc = 0;
  cartridge_location_data* p_slot_status = NULL, * p_begin = NULL;
  read_cartridge_location  sc_read_cart_loc = {0};
  struct element_info sc_element_info = {0};

  memset(&sc_element_info, 0, sizeof(struct element_info));

  rc = ioctl(fd, SMCIOC_ELEMENT_INFO, &sc_element_info);
  if(rc != 0)
    return rc;

  if(sc_element_info.slots)
  {
    p_slot_status = p_begin = malloc(sc_element_info.slots
      * sizeof(cartridge_location_data));

    if(p_slot_status == NULL)
    {
      errno = ENOMEM;
      return -2;
    } /* if */

    bzero((caddr_t)p_slot_status, sc_element_info.slots
      * sizeof(cartridge_location_data));

  } /* if */

  sc_read_cart_loc.element_address = sc_element_info.slot_addr;
  sc_read_cart_loc.number_elements = sc_element_info.slots;
  sc_read_cart_loc.data = p_slot_status;

  printf("Reading element status...\n");

  rc = ioctl(fd, SMCIOC_READ_CARTRIDGE_LOCATION, &sc_read_cart_loc);

  if(rc == 0)
  {
    for(i = 0; i < sc_element_info.slots; i++, p_slot_status++)
    {
      if(opcode && j == 2)
      {
        j=0;
        printf("\nHit <enter> to continue...");
        getchar();
      } /* if */

      j++;
      printf("\nSlot Address %d\n",p_slot_status->address);

      printf("  Slot State ..................... %sormal\n",
        p_slot_status ? "Abn" : "N");

      printf("  ASC/ASCQ ....................... %02X%02X\n",
        p_slot_status->asc,p_slot_status->ascq);

      printf("  Media Present .................. %s\n",
        p_slot_status->full ? "Yes" : "No");

      printf("  Robot Access Allowed ........... %s\n",
        p_slot_status->access ? "Yes" : "No");

      if(p_slot_status->svalid)
        printf("  Source Element Address ......... %d\n",
          p_slot_status->source);
      else
        printf("  Source Element Address Valid ... No\n");

      printf("  Media Inverted ................. %s\n",
        p_slot_status->invert ? "Yes" : "No");
      dump_bytes((char*)p_slot_status->volume, 36, "Volume Tag");

      printf("  Cartridge Location Length ...... %d\n",
        p_slot_status->ident_len);

      if(((signed)p_slot_status->ident_len) > 0)
        printf("  Cartridge Location Identifier .. %s\n",
          p_slot_status->identifier);
    } /* for */
  } /* if */

  if(p_begin) free(p_begin);

  return rc;
} /* smcioc_read_cartridge_location() */



/*- name:     smcioc_read_element_devids
 *  function: invoke medium changer IOCTOL SMCIOC_READ_ELEMENT_DEVIDS and print device ids
 *  argument: void
 *  return:   0 if succeeded, -1 or -2 if failed
 -*/
int smcioc_read_element_devids()
{
  int    i, j;
  int    rc;
  struct element_devid *elem_devid, *elemp;
  struct read_element_devids devids;
  struct element_info element_info;
  int id;
  char dev_id[37] = "";

  memset(&devids, 0, sizeof(struct read_element_devids));
  memset(&element_info, 0, sizeof(struct element_info));
  elem_devid = NULL;

   /* get the drive ID */
  id = search_drive_id();

  if(id == -1)
      return -1;

  if(id == IBM_INVALID_DEVICE)
    {
      errno = 999;
      return -2;
    }

  rc = ioctl(fd, SMCIOC_ELEMENT_INFO, &element_info);
  if(rc != 0)
    return rc;

  if(element_info.drives)
    {
      elem_devid = malloc(element_info.drives * sizeof(struct element_devid));
      if(elem_devid == NULL)
        {
          errno = ENOMEM;
          return -2;
        }
      bzero((caddr_t)elem_devid, element_info.drives * sizeof(struct element_devid));

      devids.drive_devid = elem_devid;
      devids.element_address = element_info.drive_addr;
      devids.number_elements = element_info.drives;

      printf("Reading element device ids...\n");

      rc = ioctl(fd, SMCIOC_READ_ELEMENT_DEVIDS, &devids);
      if(rc != 0)
        {
          free(elem_devid);
          return rc;
        }

      j=0;
      elemp = elem_devid;
      for (i = 0; i < element_info.drives; i++, elemp++)
        {
          /* don't overflow screen if menu mode */
          if(opcode && j == 2)
            {
              j=0;
              printf("\nHit <enter> to continue...");
              getchar();
            }

          j++;
          printf("\nDrive Address %d\n", elemp->address);
          if(elemp->except)
            {
              if(id == IBM_3584)
                printf("  Drive State .................... See ASC/ASCQ\n");
              else
                printf("  Drive State .................... Abnormal\n");
            }
          else
            printf("  Drive State .................... Normal\n");
          if(elemp->asc == 0x81 && elemp->ascq ==0x00)
            printf("  ASC/ASCQ ....................... %02X%02X (Drive Present)\n",
                   elemp->asc,elemp->ascq);
          else if(elemp->asc == 0x82 && elemp->ascq ==0x00)
            printf("  ASC/ASCQ ....................... %02X%02X (Drive Not Present)\n",
                   elemp->asc,elemp->ascq);
          else
            printf("  ASC/ASCQ ....................... %02X%02X\n",
                   elemp->asc,elemp->ascq);
          if(elemp->full)
            printf("  Media Present .................. Yes\n");
          else
            printf("  Media Present .................. No\n");
          if(elemp->access)
            printf("  Robot Access Allowed ........... Yes\n");
          else
            printf("  Robot Access Allowed ........... No\n");
          if(elemp->svalid)
            printf("  Source Element Address ......... %d\n", elemp->source);
          else
            printf("  Source Element Address Valid ... No\n");
          if(elemp->invert)
            printf("  Media Inverted ................. Yes\n");
          else
            printf("  Media Inverted ................. No\n");
          if(elemp->notbus)
            printf("  Same Bus as Medium Changer ..... No\n");
          else
            printf("  Same Bus as Medium Changer ..... Yes\n");
          if(elemp->idvalid)
            printf("  SCSI Bus Address ............... %d\n", elemp->scsi);
          else
            printf("  SCSI Bus Address Vaild ......... No\n");
          if(elemp->luvalid)
            printf("  Logical Unit Number ............ %d\n", elemp->lun);
          else
            printf("  Logical Unit Number Valid ...... No\n");
          bcopy (elemp->identifier, dev_id, 36);
          dev_id[37] = '\0';
           printf("  Device ID ...... %s\n", dev_id);
        }

        if(elem_devid)
           free(elem_devid);
    }
  else
    {
      printf("\nNo drives found in element information\n");
      if(opcode)
        {
          printf("\nHit <enter> to continue...");
          getchar();
        }
    }

  return rc;
}

/*-
 * name: sioc_query_path
 * function: display the primary path info and if there is any, the
 *           first alternate path info
 * argument: void
 * return: 0 if success, -1 if fail
 -*/
int sioc_query_path()
{
  int rc = 0;
  struct scsi_path path;

  memset(&path, 0, sizeof(struct scsi_path));

  printf("Querying SCSI paths...\n");

  rc = ioctl(fd, SIOC_QUERY_PATH, &path);
  if(rc == 0)
    show_path(&path);
  else
    rc = -2;

  return rc;
}

/*-
 * name: show_path
 * function: display path info
 * argument: pointer of scsi_path
 * return: void
 -*/
void show_path(struct scsi_path *path)
  {
    printf("\n");
    if (path->alternate_configured)
      printf("Primary Path Information:\n");
    printf("  Logical Device................. %s\n",path->primary_name);
    printf("  SCSI Host Adapter.............. %s\n",path->primary_parent);

    if (path->primary_id_valid)
      {
        printf("  SCSI Host ID................... %d\n",path->primary_host);
        printf("  SCSI Channel................... %d\n",path->primary_bus);
        printf("  Target ID...................... %d\n",path->primary_id);
        printf("  Logical Unit................... %d\n",path->primary_lun);
      }

    if (path->primary_enabled)
      printf("  Path Enabled................... Yes\n");
    else
      printf("  Path Enabled................... No \n");
    if (path->primary_fenced)
      printf("  Path Manually Disabled......... Yes\n");
    else
      printf("  Path Manually Disabled......... No \n");

    if (!path->alternate_configured)
      printf("  Alternate Path Configured...... No\n");
    else
      {
        printf("  Alternate Path Configured...... Yes\n");
        printf("\nAlternate Path Information:\n");
        if(path->alternate_id_valid)
          {
            printf("  Logical Device................. %s\n",path->alternate_name);
            printf("  SCSI Host Adapter.............. %s\n",path->alternate_parent);
            printf("  SCSI Host ID................... %d\n",path->alternate_host);
            printf("  SCSI Channel................... %d\n",path->alternate_bus);
            printf("  Target ID...................... %d\n",path->alternate_id);
            printf("  Logical Unit................... %d\n",path->alternate_lun);
          }

        if(path->alternate_enabled)
          printf("  Path Enabled................... Yes\n");
        else
          printf("  Path Enabled................... No \n");
        if (path->alternate_fenced)
          printf("  Path Manually Disabled......... Yes\n");
        else
          printf("  Path Manually Disabled......... No \n");

      }
  }

/*-
 * name: sioc_device_paths
 * function: display all of the paths
 * argument: void
 * return: 0 if success, -1 if fail
 -*/
int sioc_device_paths()
{
  int rc = 0;
  struct device_paths device_path;
  int i;

  printf("Querying device paths...\n");

  rc = ioctl(fd, SIOC_DEVICE_PATHS, &device_path);
  if(rc == 0)
    {
      printf("\n");
      for (i=0; i < device_path.number_paths; i++)
        {
          if (i == 0)
            printf("Primary Path Number 1\n");
          else
            printf("Alternate Path Number %d\n", i+1);
          printf("  Logical Device................. %s\n",device_path.path[i].name);
          printf("  Host Bus Adapter............... %s\n",device_path.path[i].parent);

          if (device_path.path[i].id_valid)
            {
              printf("  SCSI Host ID................... %d\n",device_path.path[i].host);
              printf("  SCSI Channel................... %d\n",device_path.path[i].bus);
              printf("  Target ID...................... %d\n",device_path.path[i].id);
              printf("  Logical Unit................... %d\n",device_path.path[i].lun);
            }

          if (device_path.path[i].enabled)
            printf("  Path Enabled................... Yes\n");
          else
            printf("  Path Enabled................... No \n");
          if (device_path.path[i].fenced)
            printf("  Path Manually Disabled......... Yes\n");
          else
            printf("  Path Manually Disabled......... No \n");

          printf("\n");
        }

      printf("Total paths configured........... %d\n",device_path.number_paths);
    }
  else
    rc = -2;

  return rc;
}


/*-
 * name: sioc_enable_path
 * function: enable a logical path
 * argument: path
 * return: 0 if success, -1 if fail
 -*/
int sioc_enable_path(int path)
{
  int rc = 0;

  if (path == PRIMARY_SCSI_PATH)
    printf("Enabling primary SCSI path 1...\n");
  else
    printf("Enabling alternate SCSI path %d...\n",path);

  rc = ioctl(fd, SIOC_ENABLE_PATH, path);
  if (rc == -1)
    rc = -2;

  return rc;
}

/*-
 * name: sioc_disable_path
 * function: disable a logical path
 * argument: path
 * return: 0 if success, -1 if fail
 -*/
int sioc_disable_path(int path)
{
  int rc = 0;

  if (path == PRIMARY_SCSI_PATH)
    printf("Disabling primary SCSI path 1...\n");
  else
    printf("Disabling alternate SCSI path %d...\n",path);

  rc = ioctl(fd, SIOC_DISABLE_PATH, path);
  if(rc == -1)
    rc = -2;

  return rc;
}



/****************************************************************************/
/*                Support Functions                                         */
/****************************************************************************/

/*- name:     dump_bytes
 *  function: print out *b with length len
 *  argument: b     -- pointer to a struct
 *            len   -- length of data
 *            title -- title of data
 *  return:   none
 -*/
void dump_bytes(char *b, int len, char *title)
{
  unsigned char *ptr;
  short         str_index;
  size_t        buf_index;
  size_t        off;
  size_t        lines;
  size_t        frames;
  size_t        remainder;
  int           index = 0;
  char          smsg[48];
  char          tmsg[48];
  int           start = 0;

  ptr = (unsigned char*)b + start;                    /* point to buffer location to start  */
  buf_index = 0;
  frames    = (len/256);              /* calculate number of frames         */
  remainder = (len%256);
  if(remainder)                      /* if there are bytes left over       */
    {
      frames++;                       /* then we need another frame and..   */
      remainder = (remainder-1) / 16; /* calculate how many lines are       */
      remainder++;                    /* in the last frame                  */
    }

  if(title)                         /* if a title has been supplied       */
    printf("\n%s,  Length %u\n\n", title, len);
  else
    printf("\nData,  Length %u\n\n", len);

  printf( "            0 1  2 3  4 5  6 7  8 9  A B  C D  E F   0123456789ABCDEF\n");

  while (frames > 0)
    {
    /* if last frame and more lines are required get number of lines */
    if(frames == 1 && remainder > 0)
      lines = remainder;
    else
      lines = 16;

    while (lines > 0)
      {
        memset(smsg,' ',48);
        memset(tmsg,' ',48);

        for (str_index=0,off=0; str_index<16;
             str_index++, buf_index++, off+=3)
          {
            if(buf_index >= len)
              break;               /* if we are at the end of the buffer */
            htoc(&tmsg[off],ptr[buf_index]); /* convert hex to string    */
            smsg[str_index]=ptr[buf_index];  /* save in ascii string     */
            if(smsg[str_index] < ' ')
              smsg[str_index]='.'; /* convert non printable chars to '.' */
            buf_index++;           /* index to next byte in buffer       */
            if(buf_index >= len)
              break;               /* if we are at the end of buffer     */
            str_index++;           /* index to next position in string   */
            off+=2;                /* index to next position in hex str  */
            htoc(&tmsg[off],ptr[buf_index]); /* convert hex to string    */
            smsg[str_index]=ptr[buf_index];  /* save in ascii string     */
            if(smsg[str_index] < ' ')
              smsg[str_index]='.'; /* convert non printable chars to '.' */
          }
        tmsg[39]=0;
        smsg[16]=0;
        printf("    %04X - %s  [%s]\n", index, tmsg, smsg);
        index += 16;
        lines--;                 /* finished with this line              */
      } /* end of while(lines>0) */

    frames--;                    /* finished with this frame             */
    } /* end of while(frame>0) */

  printf("\n");
}

/*- name:     htoc
 *  function: convert a char from hex to ascii
 *  argument: s -- string for ascii chars
 *            h -- hex character
 *  return:   none
 -*/
void  htoc(char *s,unsigned char h)
{
  char   *ttable="0123456789ABCDEF"; /* hex character translate table    */
  short  x;

  x = (h & 0xf0) >> 4;               /* isolate hi nibble for conversion */
  s[0] = ttable[x];                  /* index into table for ascii char  */
  x = h & 0x0f;                      /* isolate lo nibble for conversion */
  s[1] = ttable[x];                  /* index into table for ascii char  */
}


/*- name:     genpat
 *  function: generate data for write. set global array pat.
 *  argument: void
 *  return:   none.
 -*/
void genpat(void)
{
  pat[0x00] = 0x01;
  pat[0x01] = 0x13;
  pat[0x13] = 0x37;
  pat[0x37] = 0x7F;
  pat[0x7F] = 0xFE;
  pat[0xFE] = 0xFC;
  pat[0xFC] = 0xE8;
  pat[0xE8] = 0xC0;
  pat[0xC0] = 0x81;
  pat[0x81] = 0x03;
  pat[0x03] = 0x07;
  pat[0x07] = 0x1F;
  pat[0x1F] = 0x3E;
  pat[0x3E] = 0x7C;
  pat[0x7C] = 0xF8;
  pat[0xF8] = 0xF0;
  pat[0xF0] = 0xE1;
  pat[0xE1] = 0xC3;
  pat[0xC3] = 0x87;
  pat[0x87] = 0x0F;
  pat[0x0F] = 0x0E;
  pat[0x0E] = 0x1C;
  pat[0x1C] = 0x38;
  pat[0x38] = 0x70;
  pat[0x70] = 0xF1;
  pat[0xF1] = 0xF3;
  pat[0xF3] = 0xE7;
  pat[0xE7] = 0xCF;
  pat[0xCF] = 0x8E;
  pat[0x8E] = 0x0C;
  pat[0x0C] = 0x08;
  pat[0x08] = 0x10;
  pat[0x10] = 0x31;
  pat[0x31] = 0x73;
  pat[0x73] = 0xF7;
  pat[0xF7] = 0xFF;
  pat[0xFF] = 0xEE;
  pat[0xEE] = 0xCC;
  pat[0xCC] = 0x88;
  pat[0x88] = 0x00;
}


/*- name:      search_drive_id
 *  function:  invoke IOCTL SIOC_INQUIRY to find vandor id and product id
 *  arguments: void
 *  return:    drive device ID or -1 is SIOC_INQUIRY is failed
 */
int search_drive_id()
{
  struct inquiry_data inquiry_data;
  int    id = IBM_INVALID_DEVICE;
  int    rc;
  int    i;

  /* get inquiry data */
  memset(&inquiry_data, 0, sizeof(struct inquiry_data));
  rc = ioctl(fd, SIOC_INQUIRY, &inquiry_data);
  if(rc != 0)
    return rc;

  /* get vendor id */
  i = 0;
  while(SUPPORTED_VENDORS[i]!= NULL &&
        strncmp((char*)inquiry_data.vid, SUPPORTED_VENDORS[i], strlen(SUPPORTED_VENDORS[i])) != 0 )
    {
      i += 1;
    }

  if(SUPPORTED_VENDORS[i] == NULL)
    {
      printf("Vendor %s is not supported.\n", inquiry_data.vid);
      return id;
    }

  /* get drive id */
  i = 0;
  while(SUPPORTED_TAPE_DRIVES[i] != NULL &&
        strncmp((char*)inquiry_data.pid, SUPPORTED_TAPE_DRIVES[i], strlen(SUPPORTED_TAPE_DRIVES[i])) != 0)
    {
      i += 1;
    }

  if(SUPPORTED_TAPE_DRIVES[i] == NULL)
    {
      printf("Product %s is not supported.\n", inquiry_data.pid);
      return id;
    }

  id = SUPPORTED_TAPE_DRIVE_ID[i];
  return id;
}


/*- name:     isnumber
 *  function: test if a string is a valid decimal number
 *  argumant: string
 *  return:   1 if string is a valid decimal number, otherwise, 0.
 -*/
int isnumber(char * str)
{
  int length;
  int i, j;

  length = strlen(str);

  i = 0;
  while(isspace(str[i]))
    i++;

  for(j = i; j < length; j++)
    if(!isdigit(str[j]))
       break;

  if(j != length)
    return 0;
  else
    return 1;
}

/*- name:     ekm_test
 *  function: Perform the three EKM tests.
 *  argument: void
 *  return:   status
 -*/
int ekm_test()
{
  struct encryption_diagnostics diag;

  memset(&diag, 0x00, sizeof (struct encryption_diagnostics));

  /* test server config and connections */
  printf("Testing server configuration and connections...\n");
  diag.diag_type = SERVER_PING_DIAG;

  if (ioctl (fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag))
    return errno;

  if (diag.diag_errno)
  {
    errno = diag.diag_errno;
    perror("Server test failed");
    errno = 0;
    return errno;
  }
  else
    printf("Testing complete, servers available %d\n",diag.diag_cc);

  /* run basic drive to server encryption test */
  printf("Running basic drive to server encryption test...\n");
  diag.diag_type = BASIC_ENCRYPTION_DIAG;

  if (ioctl (fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag))
  {
    perror("Basic Encryption ioctl failed");
    errno =0;
  }
  else
  {
    if (diag.diag_errno)
    {
      errno = diag.diag_errno;
      perror("Basic Encryption test failed");
      errno = 0;
    }
    else
      printf("Testing complete, completion code %d\n",diag.diag_cc);
  }

  /* run full drive to server encryption test */
  printf("Running full drive to server encryption test...\n");
  diag.diag_type = FULL_ENCRYPTION_DIAG;

  if (ioctl (fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag))
  {
    perror("Full Encryption ioctl failed");
    errno =0;
  }
  else
  {
    if (diag.diag_errno)
    {
      errno = diag.diag_errno;
      perror("Full Encryption test failed");
      errno = 0;
    }
    else
      printf("Testing complete, completion code %d\n",diag.diag_cc);
  }

  return 0;
}


/*---------------------------------------------------------------------------*/
/* Name: get_sernum command                                                  */
/*---------------------------------------------------------------------------*/
int get_sernum()
{
   struct inquiry_page inqdata;
   char serial_num[20] = "";

   inqdata.page_code = (unsigned char) 0x80;
   if (!ioctl(fd, SIOC_INQUIRY_PAGE, &inqdata)) {
      bcopy(&inqdata.data[4], serial_num, inqdata.data[3]);
      serial_num[(int)inqdata.data[3]] = '\0';
      printf("%s\n", serial_num);
   }
   else {
      printf ("Get serial number failed (errno=%d).\n", errno);
      sioc_reqsense();
   }

   return errno;
}

/*---------------------------------------------------------------------------*/
/* Name: rekey_command                                                       */
/*---------------------------------------------------------------------------*/
int rekey_command(void)
  {
  #define SCMD_MODE_SEL_10                0x55

  int rc;
  struct sioc_pass_through *passthru;

  passthru = malloc(sizeof (struct sioc_pass_through));
  if (!passthru) {
     errno = ENOMEM;
     printf ("Error %d (%s) Fail to allocate memory.\n",
            errno, strerror(errno));
     return errno;
  }
  memset(passthru, 0, sizeof(struct sioc_pass_through));

  passthru->buffer = malloc(56);
  if (!passthru->buffer){
     errno = ENOMEM;
     printf ("Error %d (%s) Fail to allocate memory for buffer.\n",
            errno, strerror(errno));
     free(passthru);
     return errno;
  }
  memset(passthru->buffer, 0, 56);

  passthru->cdb = malloc(10);
  if (!passthru->cdb){
     errno = ENOMEM;
     printf ("Error %d (%s) Fail to allocate memory for cdb.\n",
            errno, strerror(errno));
     free(passthru->buffer);
     free(passthru);
     return errno;
  }
  memset(passthru->cdb, 0, 10);

  passthru->sense = malloc(125);
  if (!passthru->sense){
     errno = ENOMEM;
     printf ("Error %d (%s) Fail to allocate memory for sense.\n",
            errno, strerror(errno));
     free(passthru->buffer);
     free(passthru->cdb);
     free(passthru);
     return errno;
  }
  memset(passthru->sense, 0, 125);

  /* build mode select x25/xEE in short form */
  passthru->cdb[0]=SCMD_MODE_SEL_10;
  passthru->cdb[1]=0x10;
  passthru->cdb[8]=56;
  passthru->cmd_length=10;
  passthru->timeout=600;
  passthru->buffer_length=56;
  passthru->sense_length=125;
  passthru->data_direction=SCSI_DATA_OUT;

  passthru->buffer[3]=0x10;
  passthru->buffer[8]=0x65;
  passthru->buffer[9]=(char)0xEE;
  passthru->buffer[11]=0x2C;
  passthru->buffer[12]=0x21;
  passthru->buffer[13]=(char)0xEE;
  passthru->buffer[15]=0x28;
  passthru->buffer[20]=0x18;
  passthru->buffer[28]=0x18;
  passthru->buffer[31]=0x04;
  passthru->buffer[39]=0x02;
  passthru->buffer[40]=0x10;
  passthru->buffer[44]=0x19;
  passthru->buffer[48]=0x10;
  passthru->buffer[52]=0x19;

  printf ("Issuing rekey command...\n");
  if (!(rc = ioctl(fd, SIOC_PASS_THROUGH, passthru))) {
     printf("SCSI passthru command succeeded!\n");
     if (passthru->result)
        printf("rekey command failed!\n");
     else
        printf("rekey command succeeded!\n");
  }
  else
     printf("rekey command failed!\n");

  if (passthru->sense_length)
    dump_bytes ((char *)passthru->sense,passthru->sense_length, "Sense data");
  printf("\n");

  free(passthru->sense);
  free(passthru->cdb);
  free(passthru->buffer);
  free(passthru);

  return (rc);
  }



