/* tapeutil for rekeying use                                                 */
/*****************************************************************************/
/*****************************************************************************/
/**                                                                         **/
/**     IBM Tape & Medium Changer Utility Program / Solaris 8, 9 and 10     **/
/**                                                                         **/
/**                   (C) COPYRIGHT IBM Corporation, 1994 -                 **/
/**                                                                         **/
/**                tapeutil.c - Utility Program Source File                 **/
/**                  Licensed Materials - Property of IBM                   **/
/**                           All Rights Reserved                           **/
/**                                                                         **/
/**    US Government Users Restricted Rights  -  Use, duplication, or       **/
/**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.    **/
/**                                                                         **/
/**           ** Notice to users of the source code examples: **            **/
/**                                                                         **/
/**  The source code examples provided by IBM are only intended to assist   **/
/**  in the development of a working software program.  The source code     **/
/**  examples may not function as written; additional code is required.     **/
/**  In addition, the source code examples may not compile and/or bind      **/
/**  successfully as written.                                               **/
/**                                                                         **/
/**  International Business Machines Corporation provides the source code   **/
/**  examples "as is" without warranty of any kind, either expressed or     **/
/**  implied, including, but not limited to the implied warranties of       **/
/**  merchantability and fitness for a particular purpose.  The entire      **/
/**  risk as to the quality and performance of the source code examples     **/
/**  is with you.  Should any part of the source code examples prove        **/
/**  defective, you (and not IBM) assume the entire cost of all necessary   **/
/**  servicing, repair or correction.                                       **/
/**                                                                         **/
/**  IBM does not warrant that the contents of the source code examples     **/
/**  will meet your requirements or that the source code examples are       **/
/**  error-free.                                                            **/
/**                                                                         **/
/**  IBM may make improvements and/or changes in the source code examples   **/
/**  at any time.  Changes may be made periodically to the information in   **/
/**  the source code examples; these changes may be reported, for the       **/
/**  device drivers included herein, in new editions of the examples.       **/
/**                                                                         **/
/**  References in the source code examples to IBM products, programs, or   **/
/**  services do not imply that IBM intends to make these available in all  **/
/**  countries in which IBM operates.  Any reference to an IBM licensed     **/
/**  program in the source code examples is not intended to state or imply  **/
/**  that only IBM's licensed program may be used.  Any functionally        **/
/**  equivalent program may be used.                                        **/
/**                                                                         **/
/*****************************************************************************/
/*****************************************************************************/


/*****************************************************************************/
/*                               INCLUDE FILES                               */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <strings.h>
#include <sys/ioccom.h>
#include <sys/mtio.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <sys/scsi/generic/commands.h>

#include "../src/st.h"
#include "../src/smc.h"
#include "../src/svc.h"

/*
#include <sys/st.h>
#include <sys/smc.h>
#include <sys/svc.h>
*/


/*****************************************************************************/
/*                                GLOBAL DATA                                */
/*****************************************************************************/

#define VERSION         4.1.6.0

#define TRUE            1
#define FALSE           0
#define EXTENDED        1
#define SHORT           0

#define IOBUFSZ         0x10000
#define RWBUFSZ         0x500000
#define END_BUFLIST     -1  /* End of read buffer list for read_buflist() */

#define ELEMT_ROBOT     1
#define ELEMT_CELL      2
#define ELEMT_IMPEXP    3
#define ELEMT_DRIVE     4

#define PRINTF          if (verbose)  printf
#define PERROR          if (verbose)  perror
#define FPRINTF         if (verbose)  fprintf

/* Defines for dump_bytes & fmt_bytes */
#define dump_bytes(a,b)    fmt_bytes((a), (b), 0, (char *)(&""), 16)
#define DUMP_BYTES      if (verbose)  dump_bytes

#ifndef MIN
#define MIN(a,b) ((a) < (b)) ? (a) : (b)
#endif  MIN

typedef struct {
   int opcode;                           /* user selected operation           */
   int mode;                             /* open mode flag                    */
   int src;                              /* source for move command           */
   int dst;                              /* destination for move command      */
   int dst1;                             /* destination 1 for exchange media cmd */
   int dst2;                             /* destination 2 for exchange media cmd */
   int blk_size;                         /* block size for data transfers     */
   int blk_fact;                         /* blocking factor                   */
   int multiplier;                       /* multiple of buffer to read/write  */
   int seed;                             /* random number generator seed      */
   int mt_op;                            /* MTIOCTOP command operation field  */
   int mt_count;                         /* MTIOCTOP command count field      */
   int type;                             /* msg/pos/prm type field            */
   int value;                            /* pos/prm value field               */
   char msg_0[MSGLEN+1];                 /* message 0 text                    */
   char msg_1[MSGLEN+1];                 /* message 1 text                    */
   char filename[64];                    /* ucode/dump/pattern file name      */
} parm_t;

typedef struct dev_typemod_s {          /* device type/model indicator       */
   uint
         t3490       : 1,
         t3570       : 1,
         t3590       : 1,
         t3592       : 1,
         t3592e05    : 1,
         ultrium     : 1,
         lto1        : 1,
         lto2        : 1,
         lto3        : 1,
         lto4        : 1,
         ult3581     : 1,
         ult3582     : 1,
         ult3583     : 1,
         ult3584     : 1,
         ult3581_ta2 : 1,
         ult3576     : 1,
         ult3573     : 1,
         lib3577     : 1,
         reserved    : 14;
} dev_typemod_t;
static dev_typemod_t device;

#define PIDLEN          16
static char pid[PIDLEN + 1];            /* Product ID from Inquiry Data     */

static uchar   tape_drv    = 0,
               medium_chgr = 0,
               attached_medium_chgr = 0;

extern int errno;                       /* system error number variable      */
static int dev_errno, smc_errno;        /* errno for drive or smc open()     */
static int interactive;                 /* execute mode (menu or cmdline)    */
static int verbose;                     /* output mode (noisy or quiet)      */
static int opened = FALSE;              /* special file status               */
static int dev_fd;                      /* device special file descriptor    */
static char dev_name[128] = "";         /* tape special file name            */
static char smc_name[128] = "";         /* medium changer special file name  */
static char tmp_name[128] = "";         /* medium changer special file name  */
static char *str_dev_p;                 /* ptr to device special file name   */
static char message[64] = "";           /* display message text buffer       */
static element_info_t element_info;     /* medium changer element status     */
static char *lastc_p, **lastc_pp;       /* string operation utility ptrs     */
uchar  pat[256] = { 0x00 };             /* data pattern buffer               */
static int max_capacity;                /* maximun tape capacity             */
static inquiry_data_t inquiry_data;
static char ucode_level[32];


/*****************************************************************************/
/*                            FUNCTION PROTOTYPES                            */
/*****************************************************************************/

/*------------------------*/
/* Main Control Functions */
/*------------------------*/

extern int main (int, char **);
static int cmd_line (int, char **);
static int new_cmdline(int argc,char ** argv);
static int interact (void);
static int get_cmd (parm_t *);
static int exec_cmd (parm_t *);
static void menu (void);
static void usage (int extended);
static void newusage (void);
void fmt_bytes ( char *data, int count, int start_offset, char *id, int across );
static int get_pid(char *pid );
static void fmt_elem (int elem_type,  element_status_t *elemp, char *str_elem );
static void genpat(void);
static void show_path (device_path_t *path);

/*-------------------*/
/* Service Functions */
/*-------------------*/

static int svc_device_sn (void);
static int svc_ucode_level (void);
static int svc_force_dump (void);
static int svc_store_dump (char *);
static long long read_buflist ( int fp, buffer_io_t *biop, int desc_mode, int data_mode, int  bufid, ...);
static int svc_load_ucode ( char *filename );
static int svc_format_tape (void);
static int svc_check_device (void);
static int get_driver_info (void);
static int encryption_test (void);
static int rekey_command (void);
static int get_sernum (void);


/*----------------------*/
/* Basic SCSI Functions */
/*----------------------*/

static int scsi_test_unit_ready (void);
static int scsi_inquiry (void);
static int scsi_inquiry_page (int);
static int scsi_request_sense (void);
static int scsi_log_sense_page (uchar, uchar);
static int scsi_mode_sense_page (uchar);
static int scsi_mode_sense_subpage (uchar, uchar);
static int scsi_enable_path (void);
static int scsi_query_device_paths (void);
static int scsi_reserve (void);
static int scsi_release (void);

/*--------------------------*/
/* Medium Changer Functions */
/*--------------------------*/

static int smc_move_medium (int, int);
static int smc_exchange_medium (int, int, int);
static int smc_pos_to_elem (int);
static int smc_element_info (void);
static int smc_inventory (void);
static int smc_audit (void);
static int smc_audit_range (int, int);
static int smc_lock_door (int);
static int smc_read_element_devids (void);

/*----------------------*/
/* Tape Drive Functions */
/*----------------------*/

static int st_read (int, int, int);
static int st_write (int, int, int, int, char *);
static int st_tape_op (int, int);
static int st_get_drivetype (void);
static int st_get_device_status (void);
static int st_get_device_info (void);
static int st_get_media_info (void);
static int st_get_position (int);
static int st_set_position (int, int);
static int st_get_all_parms (void);
static int st_set_all_parms (void);
static int st_get_parm (int);
static int st_set_parm (int, int);
static int st_get_density (void);
static int st_sync_buffer (void);
static int st_display_msg (int, char *, char *);
static int st_report_density_support (void);
static int st_read_block_limits (void);
static int rwtest(int blksize, int blocks, int repetition);
static int twrite(char *fname);
static int tread(char *fname, int number);
static int rwtest_menu(void);
static int rwtestfile(char *file, int count);
static int wtest(int blksize, int blocks, int repetition);
static int rtest(int blksize, int blocks, int repetition);
static int st_get_encryption_state (void);

/*****************************************************************************/
/*                          MAIN CONTROL FUNCTIONS                           */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/*        Name: main                                                         */
/*    Synopsis: Determine run mode (interactive or command line)             */
/*              and invoke appropriate controlling function.                 */
/*     Returns: RC returned by interact() or cmd_line().                     */
/*---------------------------------------------------------------------------*/

extern int main ( int argc, char **argv )
{
   int rc;

   (void) memset (pid, 0, PIDLEN + 1);
   *((int *)&device) = 0;

   setbuf (stdout, '\0');  /* turn off buffering */
   opened  = FALSE;  /* special file not opened yet */

   if (argc == 1) {  /* interactive mode execution */
      interactive = TRUE;
      verbose = TRUE;
      rc = interact ();
   }

   else {  /* command line mode execution */
      interactive = FALSE;
      verbose = FALSE;
      rc = cmd_line (argc, argv);
   }

   return (rc);

}


/*---------------------------------------------------------------------------*/
/*        Name: cmd_line                                                     */
/*    Synopsis: Central controller for command line mode.                    */
/*     Returns: RC returned by exec_cmd().                                   */
/*---------------------------------------------------------------------------*/

static int cmd_line ( int argc, char **argv)
{
   int rc = 0;
   int i, j;
   int opt;
   long  longarg;
   char  *first_char;
   char  **last_char;
   extern int opterr;
   extern char *optarg;
   parm_t parm;
   char flags[] = "f:o:w:s:d:c:b:n:m:r:z:t:x:y:vh";
   char *cmds[] = {
      "" ,           /*  0 undefined                           */
      "" ,           /*  1 open                                */
      "" ,           /*  2 close                               */
      "qsn",         /*  3 Query Subsystem Number (For use w/  */
                     /*    lmcpd and IBM 3494 Tape Subsystem)  */
      "qmc",         /*  4 Query microcode level               */
      "fdp",         /*  5 Force microcode dump                */
      "sdp",         /*  6 Save microcode dump to file         */
      "dmc",         /*  7 Download microcode to drive         */
      "fmt",         /*  8 SCSI Format                         */
      "tur",         /*  9 SCSI Test unit ready                */
      "inq",         /* 10 SCSI Inquiry                        */
      "req",         /* 11 SCSI Request sense                  */
      "res",         /* 12 SCSI Reserve                        */
      "rel",         /* 13 SCSI Release                        */
      "mov",         /* 14 SCSI Move medium                    */
      "pos",         /* 15 SCSI Position to element            */
      "ele",         /* 16 SCSI Element info from mode data    */
      "inv",         /* 17 SCSI Read element status            */
      "aud",         /* 18 SCSI Init element status            */
      "lck",         /* 19 SCSI Prevent allow medium access    */
      "rea",         /* 20 SCSI Read                           */
      "wri",         /* 21 SCSI Write                          */
      "eof",         /* 22 SCSI Write file mark                */
      "era",         /* 23 SCSI Erase                          */
      "rew",         /* 24 SCSI Rewind                         */
      "ret",         /* 25 SCSI Retension                      */
      "off",         /* 26 Offline: Rewind and unload tape     */
      "lod",         /* 27 SCSI Load/unload                    */
      "fsf",         /* 28 SCSI Forward space file             */
      "bsf",         /* 29 SCSI Backward space file            */
      "fsr",         /* 30 SCSI Forward space record           */
      "bsr",         /* 31 SCSI Backward space record          */
      "eod",         /* 32 SCSI Space to end of data           */
      "grs",         /* 33 Get record size                     */
      "srs",         /* 34 Set record size                     */
      "gds",         /* 35 Get device status                   */
      "gdi",         /* 36 Get device informaton               */
      "gmi",         /* 37 Get media informaton                */
      "gpo",         /* 38 Get position                        */
      "spo",         /* 39 Set position                        */
      "gpa",         /* 40 Get parameter                       */
      "spa",         /* 41 Set parameter                       */
      "syn",         /* 42 Sync; flush buffered data to tape   */
      "msg",         /* 43 Display message                     */
      "chk",         /* 44 Check device                        */
      "rds",         /* 45 Report Density Support              */
      "did",         /* 46 Read Element Device ID              */
      "aur",         /* 47 Audit with range                    */
      NULL           /* End of List                            */
   };

   opterr = 0;
   memset ((char *)&parm, 0, sizeof (parm_t));
   parm.type = -1;

   while ((opt = getopt (argc, argv, flags)) != EOF) {

      first_char = optarg;
      last_char = &first_char;

      switch (opt) {

      case 'f':  /* device special file name */

         dev_name[0] = '\0';
         smc_name[0] = '\0';
         (void) strcpy(tmp_name, optarg);
         /*
         ** If the tape drive is identified by only the decimal portion of
         ** its special file name, we'll use the /dev/rmt/..stbn device.
         */
         lastc_p  = tmp_name;
         lastc_pp = &lastc_p;
         (void) (uint)(strtoul(tmp_name, lastc_pp, 10));
         if ((strlen(lastc_p) - strlen(tmp_name)) >= strlen(tmp_name)) {
            str_dev_p = smc_name;
            (void) strcpy(str_dev_p, "/dev/rmt/");
            str_dev_p += strlen("/dev/rmt/");
            (void) strcpy(str_dev_p, tmp_name);
            str_dev_p += strlen(tmp_name);
            (void) strcpy(str_dev_p, "smc");

            str_dev_p = dev_name;
            (void) strcpy(str_dev_p, "/dev/rmt/");
            str_dev_p += strlen("/dev/rmt/");
            (void) strcpy(str_dev_p, tmp_name);
            str_dev_p += strlen(tmp_name);
            (void) strcpy(str_dev_p, "stbn");
         }
         else {
            if (strstr(tmp_name, "smc") != (char *)NULL) {
               (void) strcpy(smc_name, tmp_name);
               str_dev_p = smc_name;
            }
            else {
               (void) strcpy(dev_name, tmp_name);
               str_dev_p = dev_name;
            }
         }
      break;

      case 'o':  /* operation */
         for (i = 1; cmds[i]; i++) {
            if ((*optarg != '\0') && (strncmp (cmds[i], optarg, 3) == 0)) {
            parm.opcode = i;
            break;
            }
         }
      break;

      case 'w':  /* open mode */
         i = atoi (optarg);
         if (i == 1)  parm.mode = O_RDWR;
         if (i == 2)  parm.mode = O_RDONLY;
         if (i == 3)  parm.mode = O_WRONLY;
         if (i == 4)  parm.mode = O_APPEND | O_WRONLY;
      break;

      case 's':  /* move medium source */
         parm.src = atoi (optarg);
      break;

      case 'd':  /* move medium destination */
         parm.dst = atoi (optarg);
      break;

      case 'c':  /* tape op count */
         parm.mt_count = atoi (optarg);
      break;

      case 'b':  /* block size for data transfers */
         parm.blk_size = atoi (optarg);
      break;

      case 'n':  /* blocking factor for data transfers */
         parm.blk_fact = atoi (optarg);
      break;

      case 'm':  /* multiple of buffer to read/write */
         parm.multiplier = atoi (optarg);
      break;

      case 'r':  /* random seed */
         parm.seed = atoi (optarg);
      break;

      case 'z':  /* ucode/dump/pattern file name */
         strcpy (parm.filename, optarg);
      break;

      case 't':  /* msg/gpo/spo/gpa/spa/inq/lod type */
         longarg = strtol(optarg, last_char, 10);
         if (**last_char == 0) {
            parm.type = (int)longarg;
         }
         else {
            printf("-t flag requires an integer argument; \"%s\" is invalid\n", optarg);
         }
      break;

      case 'x':  /* lck/spo/spa value/page code */
         longarg = strtol(optarg, last_char, 10);
         if (**last_char == 0) {
            parm.value = (int)longarg;
         }
         else {
            last_char = &optarg;
            longarg = strtol(optarg, last_char, 16);
            if (**last_char == 0) {
               parm.value = (int)longarg;
            }
            else {
               printf("-x flag requires an integer argument in hex; \"%s\" is invalid\n", optarg);
            }
         }
      break;

      case 'y':  /* display message text */
         strcpy (message, optarg);
         for (i=0, j=0; i<8; i++, j++) {
            if ((message[j] == ',') || (message[j] == '\0'))  break;
            parm.msg_0[i] = message[j];
         }
         for (i=0, j++; i<8; i++, j++) {
            if ((message[j] == ',') || (message[j] == '\0'))  break;
            parm.msg_1[i] = message[j];
         }
      break;

      case 'v':  /* verbose */
         verbose = TRUE;
      break;

      case 'h':
      case '?':  /* help flag */
         newusage();
         return (0);
         /*NOTREACHED*/
      break;

      default:  /* invalid flag */
         newusage();
         return (0);
         /*NOTREACHED*/
      break;

      }

   }

   /* validate input parameters */

   if ((dev_name[0] == '\0') && (smc_name[0] == '\0')) {
      PRINTF ("\nSpecial File Name [-f] always required.\n");
      newusage();
      return (EINVAL);
   }

   if (parm.opcode < 1) {
      rc = new_cmdline(argc,argv);
      return(rc);
   }

   if (parm.opcode > 47) {
      PRINTF ("\nOperation [-o] invalid.\n");
      PRINTF ("\nOpcode is: %i\n",parm.opcode);
      usage (EXTENDED);
      return (EINVAL);
   }

   if (parm.blk_fact * parm.blk_size > RWBUFSZ) {
      PRINTF ("\nBlock Size [-b] or Blocking Factor [-n] out of range.\n");
      usage (EXTENDED);
      return (EINVAL);
   }

   if (parm.multiplier > RWBUFSZ) {
      PRINTF ("\nMultiplier [-m] out of range.\n");
      usage (EXTENDED);
      return (EINVAL);
   }

   /* check for missing parameters required by specific  */
   /* operations and initialize op field for mt commands */

   switch (parm.opcode) {

      case 1:  /* open device */
      break;

      case 2:  /* close device */
      break;

      case 3:  /* query subsystem number */
      break;

      case 4:  /* query microcode level */
      break;

      case 5:  /* force dump */
      break;

      case 6:  /* store dump */
      if (parm.filename[0] == '\0') {
         PRINTF ("\nSpecified operation requires [-z] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 7:  /* download microcode */
      if (parm.filename[0] == '\0') {
         PRINTF ("\nSpecified operation requires [-z] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 8:  /* format cartridge */
      break;

      case 9:  /* test unit ready */
      break;

      case 10:  /* inquiry */
         if ((parm.type < -1) || ( parm.type > 1 )) {
            PRINTF ( "\nSpecified operation requires [-t] parm of 0 or 1.\n ");
         usage (EXTENDED);
         return (EINVAL);
         }
      break;

      case 11:  /* request sense */
      break;

      case 12:  /* reserve */
      break;

      case 13:  /* release */
      break;

      case 14:  /* move medium */
      break;

      case 15:  /* position to element */
      break;

      case 16:  /* element information */
      break;

      case 17:  /* inventory */
      break;

      case 18:  /* audit */
      break;

      case 19:  /* lock/unlock door */
      break;

      case 20:  /* read data */
      if ((parm.blk_size == 0) ||
            (parm.blk_fact == 0) ||
            (parm.multiplier == 0)) {
         PRINTF ("\nSpecified operation requires [-b], [-n], and [-m] parms.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 21:  /* write data */
      if ((parm.blk_size == 0) ||
            (parm.blk_fact == 0) ||
            (parm.multiplier == 0)) {
         PRINTF ("\nSpecified operation requires [-b], [-n], and [-m] parms.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 22:  /* write file mark */
      parm.mt_op = MTWEOF;
      break;

      case 23:  /* erase tape */
      parm.mt_op = MTERASE;
      break;

      case 24:  /* rewind */
      parm.mt_op = MTREW;
      break;

      case 25:  /* retension */
      parm.mt_op = MTRETEN;
      break;

      case 26:  /* offline */
      parm.mt_op = MTOFFL;
      break;

      case 27:  /* load/unload tape */
      if ((parm.type < 1) || (parm.type > 2)) {
         PRINTF ("\nSpecified operation requires [-t] parm of 1 or 2.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      else if (parm.type == 1) {
         parm.mt_op = STLOAD;
      }
      else {
         parm.mt_op = STUNLOAD;
      }
      break;

      case 28:  /* forward space file */
      parm.mt_op = MTFSF;
      break;

      case 29:  /* backward space file */
      parm.mt_op = MTBSF;
      break;

      case 30:  /* forward space record */
      parm.mt_op = MTFSR;
      break;

      case 31:  /* backward space record */
      parm.mt_op = MTBSR;
      break;

      case 32:  /* locate end of data */
      parm.mt_op = MTEOM;
      break;

      case 33:  /* get record size */
      parm.mt_op = MTGRSZ;
      break;

      case 34:  /* set record size */
      parm.mt_op = MTSRSZ;
      break;

      case 35:  /* get device status */
      break;

      case 36:  /* get device information */
      break;

      case 37:  /* get media information */
      break;

      case 38:  /* get position */
      if (parm.type == -1) {
         PRINTF ("\nGet position requires [-t] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 39:  /* set position */
      if (parm.type == -1) {
         PRINTF ("\nSet position requires [-t] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 40:  /* get parameter */
      if (parm.type == -1) {
         PRINTF ("\nGet parameter requires [-t] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 41:  /* set parameter */
      if (parm.type == -1) {
         PRINTF ("\nSet parameter requires [-t] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 42:  /* sync buffer */
      break;

      case 43:  /* display message */
      if ((parm.msg_0[0] == '\0') || (parm.msg_1[0] == '\0')) {
         PRINTF ("\nSpecified operation requires [-y] parm.\n");
         usage (EXTENDED);
         return (EINVAL);
      }
      break;

      case 44:  /* check device */
      break;

      case 45:  /* report density support */
      break;

      case 46:  /* read element device id */
      break;

      case 47:  /* audit range */
      break;

   }

   /* open device, issue command, and close device */
   dev_fd = -1;
   dev_errno = 0;
   smc_errno = 0;
   if (strlen(dev_name) > 0) {
      dev_fd = open (dev_name, parm.mode | O_NDELAY);
      if (dev_fd < 0) {
         dev_errno = errno;
      }
      else {
         opened    = TRUE;
         str_dev_p = dev_name;
      }
   }
   if ((dev_fd < 0) && (strlen(smc_name) > 0)) {
      dev_fd = open (smc_name, O_RDONLY | O_NDELAY);
      if (dev_fd < 0) {
         smc_errno = errno;
      }
      else {
         opened    = TRUE;
         str_dev_p = smc_name;
         if (parm.mode != O_RDONLY) {
            PRINTF ("Open mode forced read-only for medium changer.\n");
         }
      }
   }
   if (opened) {
      rc = get_pid(pid);
      if (rc != 0) {
         (void) fprintf (stderr, "Failure obtaining device PID\n");
         return (EINVAL);
      }
   }
   else {
      if (((dev_errno == ENOENT) && (smc_errno == ENOENT))
         || ((dev_errno && (dev_errno != ENOENT))
               && (smc_errno && (smc_errno != ENOENT)))) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n"
            "Error %d (%s) opening %s.\n",
            dev_errno, strerror(dev_errno), dev_name,
            smc_errno, strerror(smc_errno), smc_name);
         return (dev_errno);
      }
      else if (dev_errno && (dev_errno != ENOENT)) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            dev_errno, strerror(dev_errno), dev_name);
         if (dev_errno == EACCES) {
            if ((parm.mode & O_WRONLY) || (parm.mode & O_RDWR)) {
               FPRINTF (stderr, "   (If tape is write-protected, try opening read-only.)\n");
               FPRINTF (stderr, "   (Or, the encryption setup may be incorrect if the drive is 3592 E05.)\n");
            }
            else {
               FPRINTF (stderr, "   (The encryption setup may be incorrect if the drive is 3592 E05.)\n");
            }
         }
         return (dev_errno);
      }
      else if (smc_errno && (smc_errno != ENOENT)) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            smc_errno, strerror(smc_errno), smc_name);
         return (smc_errno);
      }
      else if (dev_errno == ENOENT) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            dev_errno, strerror(dev_errno), dev_name);
         return (dev_errno);
      }
      else if (smc_errno == ENOENT) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            smc_errno, strerror(smc_errno), smc_name);
         return (smc_errno);
      }
      else {
         (void) fprintf (stderr, "tapeutil error opening %s.\n", smc_name);
         return (-1);
      }
   }
   PRINTF ("%s: \n", str_dev_p);
   rc = exec_cmd (&parm);
   PRINTF ("\n");
   close (dev_fd);
   return (rc);
}



/*---------------------------------------------------------------------------*/
/* Name:        cmdline                                                      */
/*---------------------------------------------------------------------------*/
static int new_cmdline(int argc,char ** argv)
   {
   int i,j;
   int rc = 0;
   int arg_count = 1;
   char *argp;
   int st_count = 1;
   int source;
   int dest;
   int number;
   int read_mode = FALSE;
   int wrt_mode = FALSE;
   int append_mode = FALSE;
   int page;
   int subpage;
   int loop_count=0;
   int loop_pass=0;
   int loop_start=0;
   char devname[40] = "";
   parm_t parm;
   int err_stop = TRUE;
   char *sname;
   char *dname;
   int curpos;
   char msg0[MSGLEN+1];                 /* message 0 text                    */
   char msg1[MSGLEN+1];                 /* message 1 text                    */

   memset ((char *)&parm, 0, sizeof (parm_t));
   parm.type = -1;
   verbose = TRUE;

   /* help subcommand */
   if ((strcmp(argv[1], "-?") == 0) ||
      (strcmp(argv[1], "?") == 0))
      {
      /* help(); */
      exit(0);
      }

   /* check first arg for device */
   if (strncmp("-f",argv[1],2))
      {
      /* help(); */
      exit(1);
      }

   if (strlen(argv[1]) == 2)
      {
      arg_count++;
      if (arg_count > argc)
      {
      printf("Missing argument for device\n");
      exit(1);
      }
      strcpy(devname,argv[2]);
      }
   else
      strcpy(devname,argv[1]+2);


      dev_name[0] = '\0';
      smc_name[0] = '\0';
      (void) strcpy(tmp_name, devname);
      /*
      ** If the tape drive is identified by only the decimal portion of
      ** its special file name, we'll use the /dev/rmt/..stbn device.
      */
      lastc_p  = tmp_name;
      lastc_pp = &lastc_p;
      (void) (uint)(strtoul(tmp_name, lastc_pp, 10));
      if ((strlen(lastc_p) - strlen(tmp_name)) >= strlen(tmp_name)) {
         str_dev_p = smc_name;
         (void) strcpy(str_dev_p, "/dev/rmt/");
         str_dev_p += strlen("/dev/rmt/");
         (void) strcpy(str_dev_p, tmp_name);
         str_dev_p += strlen(tmp_name);
         (void) strcpy(str_dev_p, "smc");

         str_dev_p = dev_name;
         (void) strcpy(str_dev_p, "/dev/rmt/");
         str_dev_p += strlen("/dev/rmt/");
         (void) strcpy(str_dev_p, tmp_name);
         str_dev_p += strlen(tmp_name);
         (void) strcpy(str_dev_p, "stbn");
      }
      else {
        if (strstr(tmp_name, "smc") != (char *)NULL) {
           (void) strcpy(smc_name, tmp_name);
           str_dev_p = smc_name;
        }
        else {
           (void) strcpy(dev_name, tmp_name);
           str_dev_p = dev_name;
        }
      }

   arg_count++;
   if (arg_count > argc)
      {
      printf("Missing subcommand\n");
      exit(1);
      }

   /* see how we should open */
   for (i = arg_count; i < argc; i++) {
      if ((strcmp(argv[i],"rtest") == 0) ||
         (strcmp(argv[i],"rwtest") == 0) ||
         (strcmp(argv[i],"qrypart") == 0)  ||
         (strcmp(argv[i],"chgpart") == 0)  ||
         (strcmp(argv[i],"list") == 0)  ||
         (strcmp(argv[i],"sili") == 0)  ||
         (strcmp(argv[i],"nosili") == 0)  ||
         (strcmp(argv[i],"read") == 0))
      read_mode = TRUE;

      if ((strcmp(argv[i],"wtest") == 0) ||
         (strcmp(argv[i],"rwtest") == 0) ||
         (strcmp(argv[i],"write") == 0)  ||
         (strcmp(argv[i],"erase") == 0)  ||
         (strcmp(argv[i],"sync") == 0)  ||
         (strcmp(argv[i],"eof") == 0)  ||
         (strcmp(argv[i],"eofimm") == 0)  ||
         (strcmp(argv[i],"erg") == 0)  ||
         (strcmp(argv[i],"sdp") == 0)  ||
         (strcmp(argv[i],"idp") == 0)  ||
         (strcmp(argv[i],"weof") == 0) ||
         (strcmp(argv[i],"weofimm") == 0))
      wrt_mode = TRUE;

      if (strcmp(argv[i],"append") == 0) {
         append_mode = TRUE;
      }
   }

   if (read_mode && wrt_mode)
      parm.mode = O_RDWR;
   else if (read_mode)
      parm.mode = O_RDONLY;
   else if (wrt_mode)
      parm.mode = O_WRONLY;
   else if (append_mode)
      parm.mode = O_APPEND | O_WRONLY;


   /* open device, issue command, and close device */
   dev_fd = -1;
   dev_errno = 0;
   smc_errno = 0;
   if (strlen(dev_name) > 0) {
      dev_fd = open (dev_name, parm.mode | O_NDELAY);
      if (dev_fd < 0) {
         dev_errno = errno;
      }
      else {
         opened    = TRUE;
         str_dev_p = dev_name;
      }
   }
   if ((dev_fd < 0) && (strlen(smc_name) > 0)) {
      dev_fd = open (smc_name, O_RDONLY | O_NDELAY);
      if (dev_fd < 0) {
         smc_errno = errno;
      }
      else {
         opened    = TRUE;
         str_dev_p = smc_name;
         if (parm.mode != O_RDONLY) {
            PRINTF ("Open mode forced read-only for medium changer.\n");
         }
      }
   }
   if (opened) {
      rc = get_pid(pid);
      if (rc != 0) {
         (void) fprintf (stderr, "Failure obtaining device PID\n");
         return (EINVAL);
      }
   }
   else {
      if (((dev_errno == ENOENT) && (smc_errno == ENOENT))
         || ((dev_errno && (dev_errno != ENOENT))
               && (smc_errno && (smc_errno != ENOENT)))) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n"
            "Error %d (%s) opening %s.\n",
            dev_errno, strerror(dev_errno), dev_name,
            smc_errno, strerror(smc_errno), smc_name);
         return (dev_errno);
      }
      else if (dev_errno && (dev_errno != ENOENT)) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            dev_errno, strerror(dev_errno), dev_name);
         if (dev_errno == EACCES) {
            if ((parm.mode & O_WRONLY) || (parm.mode & O_RDWR)) {
               FPRINTF (stderr, "   (If tape is write-protected, try opening read-only.)\n");
               FPRINTF (stderr, "   (Or, the encryption setup may be incorrect if it is a 3592 E05 or LTO-4 drive.)\n");
            }
            else {
               FPRINTF (stderr, "   (The encryption setup may be incorrect if it is a 3592 E05 or LTO-4 drive.)\n");
            }
         }
         return (dev_errno);
      }
      else if (smc_errno && (smc_errno != ENOENT)) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            smc_errno, strerror(smc_errno), smc_name);
         return (smc_errno);
      }
      else if (dev_errno == ENOENT) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            dev_errno, strerror(dev_errno), dev_name);
         return (dev_errno);
      }
      else if (smc_errno == ENOENT) {
         (void) fprintf (stderr,
            "Error %d (%s) opening %s.\n",
            smc_errno, strerror(smc_errno), smc_name);
         return (smc_errno);
      }
      else {
         (void) fprintf (stderr, "tapeutil error opening %s.\n", smc_name);
         return (-1);
      }
   }

   for (i = arg_count;; i++)
   {
           fflush(stdout);
           if (rc != 0)
           {
                   /* sprintf(msg,"Operation failed with errno %d",errno);
                          perror (msg);
                          scsi_request_sense(); */
                   if (err_stop)
                           break;
           }

           if (i == argc)
           {
                   if (loop_start)
                   {
                           loop_pass++;
                           if (loop_count && (loop_pass > loop_count))
                                   break;
                           else
                                   i = loop_start;
                   }
                   else
                           break;
           }

           errno = 0;
           rc = 0;
           argp = argv[i];

           /* help subcommand */
           if ((strcmp(argp, "-?") == 0) ||
                           (strcmp(argp, "-??") == 0) ||
                           (strcmp(argp, "/?") == 0) ||
                           (strcmp(argp, "?") == 0) ||
                           (strcmp(argp, "help") == 0)) {
                   newusage();
                   continue;
           }

           /* append subcommand handled on open */
           if (strcmp(argp, "append") == 0)
           {
                   printf("Open in append mode...\n");
                   continue;
           }

           /* passthru subcommand handled on open */
           if (strcmp(argp, "passthru") == 0)
                   continue;

           if (strcmp(argp, "nostop") == 0)
           {
                   err_stop=FALSE;
                   continue;
           }


           /* loop subcommand */
           if (strcmp(argp, "loop") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           if (isdigit(argv[i+1][0]))
                           {
                                   loop_count = atoi(argv[i+1]);
                                   i++;
                                   if ((i + 1) == argc)
                                   {
                                           printf("Missing argument for loop subcommand\n");
                                           exit(1);
                                   }
                           }
                           else
                                   loop_count = 0;
                   }
                   else
                   {
                           printf("Missing argument for loop subcommand\n");
                           exit(1);
                   }
                   loop_pass = 1;
                   loop_start = i + 1;
                   continue;
           }

           /* sleep subcommand */
           if (strcmp(argp, "sleep") == 0)
           {
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for sleep subcommand\n");
                           exit(1);
                   }
                   if (!isdigit(argv[i][0]))
                   {
                           printf("Invalid argument for sleep subcommand\n");
                           exit(1);
                   }
                   number = atoi(argv[i]);
                   sleep(number);
                   continue;
           }

           /* density subcommand */
           if (strcmp(argp, "density") == 0)
           {
                   rc = st_report_density_support();
                   continue;
           }

           /* load ucode subcommand */
           if (strcmp(argp, "ucode") == 0)
           {
                   i++;
                   if (i < argc)
                           svc_load_ucode(argv[i]);
                   else
                   {
                           printf("Missing argument for ucode subcommand\n");
                           exit(1);
                   }
                   continue;
           }

           /* dump subcommand */
           if (strcmp(argp, "dump") == 0)
           {
                   /* if another arg, assume it's a dump filename */
                   if ((i + 1) < argc)
                   {
                           i++;
                           rc = svc_store_dump(argv[i]);
                   }
                   else
                           rc = svc_force_dump();
                   continue;
           }

           /* vpd subcommand */
           if (strcmp(argp, "vpd") == 0)
           {
                   rc = svc_check_device();
                   rc = svc_device_sn();
                   rc = svc_ucode_level();
                   continue;
           }

           /* print subcommand */
           if (strcmp(argp, "print") == 0)
           {
                   i++;
                   if (i < argc)
                           printf("%s\n",argv[i]);
                   else
                   {
                           printf("Missing argument for print subcommand\n");
                           exit(1);
                   }
                   continue;
           }


           /* set read sili bit subcommand */
           if (strcmp(argp, "sili") == 0)
           {
                   printf("Setting Read SILI bit ON\n");
                   if (rc == 0)
                   {
                           rc = st_set_parm(SILI, ON);
                   }
                   continue;
           }

           /* set read sili bit subcommand */
           if (strcmp(argp, "nosili") == 0)
           {
                   printf("Setting Read SILI bit OFF\n");
                   if (rc == 0)
                   {
                           rc = st_set_parm(SILI, OFF);
                   }
                   continue;
           }

           /* set compression subcommand */
           if (strcmp(argp, "compress") == 0)
           {
                   printf("Setting compression ON\n");
                   if (rc == 0)
                   {
                           rc = st_set_parm(COMPRESSION, ON);
                   }
                   continue;
           }

           /* set no compression subcommand */
           if (strcmp(argp, "nocompress") == 0)
           {
                   printf("Setting compression OFF\n");
                   if (rc == 0)
                   {
                           rc = st_set_parm(COMPRESSION, OFF);
                   }
                   continue;
           }

           /* display message subcommand */
           if (strcmp(argp, "display") == 0)
           {
                   i++;
                   if (i < argc) {
                           if (strlen(argv[i]) > 8) {
                                   strcpy(msg0, argv[i]);
                                   strcpy(msg1, argv[i]+8);
                                   rc = st_display_msg(5, msg0, msg1);
                           } else {
                                   strcpy(msg0, argv[i]);
                                   strcpy(msg1, "");
                                   rc = st_display_msg(1, msg0, msg1);
                           }
                   }
                   else
                   {
                           printf("Missing argument for display subcommand\n");
                           exit(1);
                   }
                   continue;
           }

           /* inquiry subcommand */
           if (strncmp(argp, "inquiry",3) == 0)
           {
                   /* check for optional page code */
                   if ((i + 1) < argc)
                   {
                           page = (int)strtol(argv[i+1],NULL,16);
                           if (isdigit(argv[i+1][0]) || page)
                           {
                                   i++;
                                   rc = scsi_inquiry_page(page);
                           }
                           else
                                   rc = scsi_inquiry();
                   }
                   else
                           rc = scsi_inquiry();
                   continue;
           }


           /* mode sense page subcommand */
           if (strcmp(argp, "modepage") == 0)
           {
                   subpage = 0;
                   i++;
                   if (i >= argc)
                   {
                      printf("Missing argument for modepage subcommand\n");
                      exit(1);
                   }
                   if (isdigit(argv[i][0]))
                      page = (int)strtol(argv[i],NULL,16);

                   if ((i + 1) < argc)
                   {
                      subpage = (int)strtol(argv[i+1],NULL,16);
                      if (isdigit(argv[i+1][0]) || subpage)
                      i++;
                   else
                      subpage = 0;
                   }

                   if (subpage)
                      rc = scsi_mode_sense_subpage(page, subpage);
                   else
                      rc = scsi_mode_sense_page(page);
                   continue;
           }


           /* request sense subcommand */
           if (strncmp(argp, "reqsense", 3) == 0)
           {
                   rc = scsi_request_sense();
                   continue;
           }

           /* read position subcommand */
           if (strcmp(argp, "readpos") == 0)
           {
                   rc = st_get_position(LOGICAL_BLK);
                   continue;
           }

           /* locate subcommand */
           if (strcmp(argp, "locate") == 0)
           {
                   rc = st_get_position(LOGICAL_BLK);
                   continue;
           }

           /* query position subcommand */
           if (strcmp(argp, "qrypos") == 0)
           {
                   rc = st_get_position(LOGICAL_BLK);
                   continue;
           }

           /* set position subcommand */
           if (strcmp(argp, "setpos") == 0)
           {
                   /* check for optional bockid arg in decimal */
                   if (((i + 1) < argc) && isdigit(argv[i+1][0]))
                   {
                           i++;
                           /* set the global position structure used for qrypos/setpos defaults */
                           curpos = atoi(argv[i]);
                   }
                   /* check for optional bockid arg in hex */
                   else if (((i + 1) < argc) && argv[i+1][0] == 'x')
                   {
                           i++;
                           /* set the global position structure used for qrypos/setpos defaults */
                           curpos = (int)strtol(argv[i]+1,NULL,16);
                   }
                   rc = st_set_position( LOGICAL_BLK, curpos );
                   if (rc == 0)
                           rc = st_get_position(LOGICAL_BLK);
                   continue;
           }

           /* get devinfo subcommand */
           if (strcmp(argp, "devinfo") == 0)
           {
                   rc = st_get_device_info();
                   continue;
           }

           /* test unit ready subcommand */
           if (strcmp(argp, "tur") == 0)
           {
                   rc = scsi_test_unit_ready();
                   continue;
           }

           /* reserve subcommand */
           if (strcmp(argp, "reserve") == 0)
           {
                   rc = scsi_reserve();
                   continue;
           }

           /* release subcommand */
           if (strcmp(argp, "release") == 0)
           {
                   rc = scsi_release();
                   continue;
           }

           /* get element information subcommand */
           if (strncmp(argp, "elementinfo", 3) == 0)
           {
                   rc = smc_element_info();
                   continue;
           }

           /* initialize element status subcommand */
           if (strcmp(argp, "audit") == 0)
           {
                   /* check for audit range */
                   if ((i + 1) < argc && isdigit(argv[i+1][0]))
                   {
                           i++;
                           source = atoi(argv[i]);
                           if ((i + 1) < argc && isdigit(argv[i+1][0]))
                           {
                                   i++;
                                   st_count = atoi(argv[i]);
                           }
                           else
                                   st_count = 1;
                           rc = smc_audit_range(source,st_count);
                   }
                   else
                           rc = smc_audit();
                   continue;
           }

           /* prevent medium removal subcommand */
           if (strcmp(argp, "prevent") == 0)
           {
                   /* check if it's an smc or tape device */
                   rc = smc_lock_door(LOCK);
                   continue;
           }

           /* allow medium removal subcommand */
           if (strcmp(argp, "allow") == 0)
           {
                   /* check if it's an smc or tape device */
                   rc = smc_lock_door(UNLOCK);
                   continue;
           }

           /* read element status subcommand */
           if (strncmp(argp, "inventory", 3) == 0)
           {
                   rc = smc_inventory();
                   continue;
           }

           /* read element device ids subcommand */
           if (strncmp(argp, "devids", 5) == 0)
           {
                   rc = smc_read_element_devids();
                   continue;
           }

           /* move medium subcommand */
           if (strcmp(argp, "move") == 0)
           {
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for move subcommand\n");
                           exit(1);
                   }
                   /* check for new format */
                   if (isdigit(argv[i][0]))
                           source = atoi(argv[i]);
                   else
                   {
                           argp = argv[i];
                           if (strncmp("-s",argp,2) != 0)  /* source argument */
                           {
                                   printf("Missing argument for move subcommand\n");
                                   exit(1);
                           }
                           if (strlen(argp) == 2)
                           {
                                   i++;
                                   source = atoi(argv[i]);
                           }
                           else
                                   source = atoi(argv[i]+2);
                   }
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for move subcommand\n");
                           exit(1);
                   }
                   /* check for new format */
                   if (isdigit(argv[i][0]))
                           dest = atoi(argv[i]);
                   else
                   {
                           argp = argv[i];
                           if (strncmp("-d",argp,2) != 0)  /* destination argument */
                           {
                                   printf("Missing argument for move subcommand\n");
                                   exit(1);
                           }
                           if (strlen(argp) == 2)
                           {
                                   i++;
                                   dest = atoi(argv[i]);
                           }
                           else
                                   dest = atoi(argv[i]+2);
                   }
                   rc = smc_move_medium(source,dest);
                   continue;
           }

           /* position to element subcommand */
           if (strncmp(argp, "position",3) == 0)
           {
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for position subcommand\n");
                           exit(1);
                   }
                   /* check for new format */
                   if (isdigit(argv[i][0]))
                           dest = atoi(argv[i]);
                   else
                   {
                           argp = argv[i];
                           if (strncmp("-d",argp,2) != 0)  /* destination argument */
                           {
                                   printf("Missing argument for position subcommand\n");
                                   exit(1);
                           }
                           if (strlen(argp) == 2)
                           {
                                   i++;
                                   dest = atoi(argv[i]);
                           }
                           else
                                   dest = atoi(argv[i]+2);
                   }
                   rc = smc_pos_to_elem(dest);
                   continue;
           }

           /* synchronize buffers subcommand */
           if (strcmp(argp, "sync") == 0)
           {
                   rc = st_sync_buffer();
                   continue;
           }

           /* log sense page subcommand */
           if (strcmp(argp, "logpage") == 0)
           {
                   subpage = 0;
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for logpage subcommand\n");
                           exit(1);
                   }
                   if (isdigit(argv[i][0]))
                      page = (int)strtol(argv[i],NULL,16);

                   if ((i + 1) < argc)
                   {
                      subpage = (int)strtol(argv[i+1],NULL,16);
                      if (isdigit(argv[i+1][0]) || subpage)
                      i++;
                   else
                      subpage = 0;
                   }

                   rc = scsi_log_sense_page(page, subpage);
                   continue;
           }

           /* get drive type subcommand */
           if (strcmp(argp, "drvtype") == 0)
           {
                   rc = st_get_drivetype();
                   continue;
           }

           /* get status subcommand */
           if (strcmp(argp, "status") == 0)
           {
                   rc = st_get_device_status();
                   continue;
           }

           /* get parms subcommand */
           if (strcmp(argp, "parms") == 0)
           {
                   rc = st_get_all_parms();
                   continue;
           }

           /* load tape subcommand */
           if (strcmp(argp, "load") == 0)
           {
                   rc = st_tape_op(STLOAD,0);
                   continue;
           }

           /* unload tape subcommand */
           if (strcmp(argp, "offline") == 0  ||
                           strcmp(argp, "rewoffl") == 0  ||
                           strcmp(argp, "unload") == 0)
           {
                   rc = st_tape_op(MTOFFL,0);
                   continue;
           }

           /* rewind tape subcommand */
           if (strcmp(argp, "rewind") == 0  ||
                           strcmp(argp, "retension") == 0)
           {
                   rc = st_tape_op(MTREW,0);
                   continue;
           }

           /* erase tape subcommand */
           if (strcmp(argp, "erase") == 0)
           {
                   rc = st_tape_op(MTERASE,0);
                   continue;
           }

           /* space to end-of-data subcommand */
           if (strcmp(argp, "seod") == 0)
           {
                   rc = st_tape_op(MTEOM,0);
                   continue;
           }

           /* write filemarks subcommand */
           if (strcmp(argp, "eof") == 0  ||
                           strcmp(argp, "weof") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTWEOF,st_count);
                   continue;
           }

           /* forward space filemarks subcommand */
           if (strcmp(argp, "fsf") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTFSF,st_count);
                   continue;
           }

           /* backward space filemarks subcommand */
           if (strcmp(argp, "bsf") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTBSF,st_count);
                   continue;
           }

           /* backward space beginning file subcommand */
           if (strcmp(argp, "nbsf") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTNBSF,st_count);
                   continue;
           }

           /* forward space records subcommand */
           if (strcmp(argp, "fsr") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTFSR,st_count);
                   continue;
           }

           /* backward space records subcommand */
           if (strcmp(argp, "bsr") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTBSR,st_count);
                   continue;
           }

           /* read and write tests */
           if (strcmp(argp, "wtest") == 0 ||
                           strcmp(argp, "rtest") == 0 ||
                           strcmp(argp, "rwtest") == 0)
           {
                   int blksize = 10240;
                   int blocks = 20;
                   int repetition = 1;
                   char *file = NULL;

                   /* look for r/w options */
                   rc = 0;
                   for (j = 0; (j < 3) && ((i + 1) < argc); j++)
                   {
                           if (strncmp("-b",argv[i+1],2) == 0)  /* blocksize argument */
                           {
                                   i++;
                                   if (strlen(argv[i]) == 2)
                                   {
                                           i++;
                                           if (i >= argc)
                                           {
                                                   printf("Missing argument for blocksize option\n");
                                                   exit(1);
                                           }
                                           blksize = atoi(argv[i]);
                                   }
                                   else
                                           blksize = atoi(argv[i]+2);
                                   continue;
                           }
                           if (strncmp("-c",argv[i+1],2) == 0)  /* count argument */
                           {
                                   i++;
                                   if (strlen(argv[i]) == 2)
                                   {
                                           i++;
                                           if (i >= argc)
                                           {
                                                   printf("Missing argument for count option\n");
                                                   exit(1);
                                           }
                                           blocks = atoi(argv[i]);
                                   }
                                   else
                                           blocks = atoi(argv[i]+2);
                                   continue;
                           }
                           if (strncmp("-r",argv[i+1],2) == 0)  /* repetition argument */
                           {
                                   i++;
                                   if (strlen(argv[i]) == 2)
                                   {
                                           i++;
                                           if (i >= argc)
                                           {
                                                   printf("Missing argument for repetition option\n");
                                                   exit(1);
                                           }
                                           repetition = atoi(argv[i]);
                                   }
                                   else
                                           repetition = atoi(argv[i]+2);
                                   continue;
                           }
                           if (strncmp("-s",argv[i+1],2) == 0)  /* file argument */
                           {
                                   i++;
                                   if (strlen(argv[i]) == 2)
                                   {
                                           i++;
                                           if (i >= argc)
                                           {
                                                   printf("Missing argument for file option\n");
                                                   exit(1);
                                           }
                                           file = argv[i];
                                   }
                                   else
                                           file = argv[i]+2;
                                   continue;
                           }
                           break;
                   }
                   if (rc)
                           continue;
                   if (strcmp(argp, "wtest") == 0)
                           rc = wtest(blksize,blocks,repetition);
                   else if (strcmp(argp, "rtest") == 0)
                           rc = rtest(blksize,blocks,repetition);
                   else if (!file)
                           rc = rwtest(blksize,blocks,repetition);
                   else
                           rc = rwtestfile(file, repetition);
                   continue;
           }

           /* write  */
           if (strcmp(argp, "write") == 0)
           {
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for write subcommand\n");
                           exit(1);
                   }
                   argp = argv[i];
                   if (strncmp("-s",argp,2) != 0)  /* source argument */
                   {
                           printf("Missing argument for write subcommand\n");
                           exit(1);
                   }
                   if (strlen(argp) == 2)
                   {
                           i++;
                           sname = argv[i];
                   }
                   else
                           sname = argv[i]+2;
                   rc = twrite(sname);
                   continue;
           }

           /* read  */
           if (strcmp(argp, "read") == 0)
           {
                   i++;
                   if (i >= argc)
                   {
                           printf("Missing argument for read subcommand\n");
                           exit(1);
                   }
                   argp = argv[i];
                   if (strncmp("-d",argp,2) != 0)  /* destination argument */
                   {
                           printf("Missing argument for read subcommand\n");
                           exit(1);
                   }
                   if (strlen(argp) == 2)
                   {
                           i++;
                           dname = argv[i];
                   }
                   else
                           dname = argv[i]+2;
                   /* check for optional -c count  */
                   if (((i + 1) < argc) && strncmp("-c",argv[i+1],2) == 0)
                   {
                           i++;
                           if (strlen(argv[i]) == 2)
                           {
                                   i++;
                                   if (i >= argc)
                                   {
                                           printf("Missing argument for count option\n");
                                           exit(1);
                                   }
                                   number = atoi(argv[i]);
                           }
                           else
                                   number = atoi(argv[i]+2);
                           if (number == 0)
                           {
                                   printf("Invalid argument for count option\n");
                                   exit(1);
                           }
                   }
                   else
                           number = 0;
                   rc = tread(dname, number);
                   continue;
           }

           /* serialnum subcommand */
           if (strcmp(argp, "serialnum") == 0)
           {
                   rc = svc_device_sn();
                   continue;
           }

           /* sernum subcommand */
           if (strcmp(argp, "sernum") == 0)
           {
                   rc = get_sernum();
                   continue;
           }

           /* rekey subcommand */
           if (strcmp(argp, "rekey") == 0)
           {
                   rc = rekey_command();
                   continue;
           }

           /* subsysnum subcommand */
           if (strcmp(argp, "subsysnum") == 0)
           {
                   rc = svc_device_sn();
                   continue;
           }

           /* ucodelevel subcommand */
           if (strcmp(argp, "ucodelevel") == 0)
           {
                   rc = svc_ucode_level();
                   continue;
           }

           /* format subcommand */
           if (strcmp(argp, "format") == 0)
           {
                   rc = svc_format_tape();
                   continue;
           }

           /* get record size subcommand */
           if (strcmp(argp, "getrecsize") == 0)
           {
                   rc = st_tape_op(MTGRSZ,0);
                   continue;
           }

           /* set record size subcommand */
           if (strcmp(argp, "setrecsize") == 0)
           {
                   if ((i + 1) < argc)
                   {
                           st_count = atoi(argv[i+1]);
                           if (st_count > 0)
                                   i++;
                           else
                                   st_count = 1;
                   }
                   else
                           st_count = 1;
                   rc = st_tape_op(MTSRSZ, st_count);
                   continue;
           }

           /* get encryption state subcommand */
           if (strcmp(argp, "encryption") == 0)
           {
              rc = st_get_encryption_state();
                   continue;
           }

           /* ekm test subcommand */
           if (strcmp(argp, "ekmtest") == 0)
           {
              rc = encryption_test ();
                   continue;
           }

           /* get media information */
           if (strcmp(argp, "mediainfo") == 0)
           {
                   rc = st_get_media_info();
                   continue;
           }

           /* read block limits */
           if (strcmp(argp, "readblk") == 0)
           {
                   rc = st_read_block_limits();
                   continue;
           }

           printf("Invalid argument: %s\n",argp);
           exit(1);
   }

   close(dev_fd);                /* Close the device */
   return rc;
   }




/*---------------------------------------------------------------------------*/
/*        Name: interact                                                     */
/*    Synopsis: Central controller for interactive mode.                     */
/*     Returns: RC returned by exec_cmd().                                   */
/*---------------------------------------------------------------------------*/
static int interact ()
{
   int rc = 0;
   parm_t parm;

   printf ("\n");
   menu ();
   do {
      memset ((char *)&parm, 0, sizeof (parm_t));
      while (!get_cmd (&parm));  /* get valid command and all parms */
      if (parm.opcode) {
      rc = exec_cmd (&parm);  /* execute cmd if not quit */
      }
   } while (parm.opcode);  /* if 0, then command was quit */
   if (opened) {
      close (dev_fd);  /* close special file if opened */
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: get_cmd                                                      */
/*    Synopsis: Obtain command selection and appropriate parameters.         */
/*     Returns: TRUE for valid option, FALSE for invalid/missing option.     */
/*---------------------------------------------------------------------------*/
static int get_cmd ( parm_t *parm )
{
   int temp;
   char buf[64] = "";

   printf ("\nENTER COMMAND ('M' for Menu) ==> ");
   fflush (stdin);
   gets (buf);
   printf ("\n");
   fflush (stdout);

   if (*buf == '\0') {
      return (FALSE);
   }
   if ((*buf == 'D') || (*buf == 'd')) {  /* device type */
      sprintf (buf, "%d", 44);
   }
   if ((*buf == 'M') || (*buf == 'm')) {  /* menu refresh */
      menu ();
      return (FALSE);
   }
   if ((*buf == 'Q') || (*buf == 'q')) {  /* quit program */
      parm->opcode = 0;
      return (TRUE);
   }
   if (*buf == '!') {  /* Pass a command to the system. */
      (void) system ((char *)(&buf[1]));
      return (FALSE);
   }

   parm->opcode = atoi (buf);
   if ((parm->opcode < 1) || (parm->opcode > 60)) {
      printf ("Invalid command selection.\n");
      return (FALSE);
   }
   if ((!opened) && (parm->opcode != 1)) { /* must be opened first */
      printf ("Device not opened.\n");
      return (FALSE);
   }

   switch (parm->opcode) {
      case 1:  /* open device */
         if (opened) {
            printf ("Device %s already open.\n", str_dev_p);
            return (FALSE);
         }
         dev_name[0] = '\0';
         smc_name[0] = '\0';
         printf ("Device special file to open"
            " (e.g., /dev/rmt/0stbn): ");
         gets (tmp_name);
         /*
         ** If the tape drive is identified by only the decimal portion of
         ** its special file name, we'll use the /dev/rmt/..stbn device.
         */
         lastc_p  = tmp_name;
         lastc_pp = &lastc_p;
         (void) (uint)(strtoul(tmp_name, lastc_pp, 10));
         if ((strlen(lastc_p) - strlen(tmp_name)) >= strlen(tmp_name)) {
            str_dev_p = dev_name;
            (void) strcpy(str_dev_p, "/dev/rmt/");
            str_dev_p += strlen("/dev/rmt/");
            (void) strcpy(str_dev_p, tmp_name);
            str_dev_p += strlen(tmp_name);
            (void) strcpy(str_dev_p, "stbn");

            str_dev_p = smc_name;
            (void) strcpy(str_dev_p, "/dev/rmt/");
            str_dev_p += strlen("/dev/rmt/");
            (void) strcpy(str_dev_p, tmp_name);
            str_dev_p += strlen(tmp_name);
            (void) strcpy(str_dev_p, "smc");
         }
         else {
            if (strstr(tmp_name, "smc") != (char *)NULL) {
               (void) strcpy(smc_name, tmp_name);
               str_dev_p = smc_name;
            }
            else {
               (void) strcpy(dev_name, tmp_name);
               str_dev_p = dev_name;
            }
         }
         if ((strlen(smc_name) > 0) && (strlen(dev_name) == 0)) {
            printf ("Only read-only mode allowed for medium changer.\n");
            parm->mode = O_RDONLY;
         }
         else {
            printf ("Select open mode");
            printf (" (1=read-write, 2=read, 3=write, 4=append): ");
            gets (buf);
            temp = atoi (buf);
            if ((temp < 1) || (temp > 4)) {
               printf ("\nInvalid open mode selection.\n");
               return (FALSE);
            }
            if (temp == 1)  parm->mode = O_RDWR;
            if (temp == 2)  parm->mode = O_RDONLY;
            if (temp == 3)  parm->mode = O_WRONLY;
            if (temp == 4)  parm->mode = O_APPEND | O_WRONLY;
            printf ("\n");
         }
      break;

      case 2:  /* close device */
      break;

      case 3:  /* query subsystem number */
      break;

      case 4:  /* query microcode level */
      break;

      case 5:  /* force dump */
      break;

      case 6:  /* store dump */
         printf ("Enter dump destination file name: ");
         /* printf (" ('tapedrive' = store on cartridge in drive): "); */
         gets (parm->filename);
         if (parm->filename[0] == '\0') {
            printf ("\nInvalid file name.\n");
            return (FALSE);
         }
         printf ("\n");
      break;

      case 7:  /* download microcode */
         printf ("Enter microcode source file name: ");
         gets (parm->filename);
         if (parm->filename[0] == '\0') {
            printf ("\nInvalid file name.\n");
            return (FALSE);
         }
         printf ("\n");
      break;

      case 8:  /* format cartridge */
      break;

      case 9:  /* test unit ready */
      break;

      case 10:  /* inquiry */
         printf ("Enter page code in hex or <enter> for standard inquiry: ");
         gets (buf);
         if ( buf[0] == '\0' )
            parm->type = 0;
         else {
            parm->type = 1;
            parm->value = (int) strtol ( buf, NULL, 16);
         }
         printf ("\n");
      break;

      case 11:  /* request sense */
      break;

      case 12:  /* reserve */
      break;

      case 13:  /* release */
      break;

      case 14:  /* move medium */
         printf ("Enter source address: ");
         gets (buf);
         parm->src = atoi (buf);
         printf ("Enter destination address: ");
         gets (buf);
         parm->dst = atoi (buf);
         printf ("\n");
      break;

      case 15:  /* position to element */
         printf ("Enter destination address: ");
         gets (buf);
         parm->dst = atoi (buf);
         printf ("\n");
      break;

      case 16:  /* element information */
      break;

      case 17:  /* inventory */
      break;

      case 18:  /* audit */
      break;

      case 19:  /* lock/unlock door */
         printf ("Enter selection");
         printf (" (1=lock, 2=unlock): ");
         gets (buf);
         parm->value = atoi (buf);
         if ((parm->value < 1) || (parm->value > 2)) {
            printf ("\nInvalid selection.\n");
            return (FALSE);
         }
         if (parm->value == 1)  parm->value = LOCK;
         else  parm->value = UNLOCK;
         printf ("\n");
      break;

      case 20:  /* read data */
         printf ("Total Bytes Read = ");
         printf ("(block size) * (blocking factor) * (multiplier).\n\n");
         printf ("Enter block size: ");
         gets (buf);
         parm->blk_size = atoi (buf);
         if ((parm->blk_size < 1) || (parm->blk_size > RWBUFSZ)) {
            printf ("\nBlock size out of range.\n");
            return (FALSE);
         }
         printf("Enter blocking factor: ");
         gets (buf);
         parm->blk_fact = atoi (buf);
         if ((parm->blk_fact < 1) ||
               ((parm->blk_fact * parm->blk_size) > RWBUFSZ)) {
            printf ("\nBlocking factor out of range.\n");
            return (FALSE);
         }
         printf ("Enter multiplier: ");
         gets (buf);
         parm->multiplier = atoi (buf);
         if ((parm->multiplier < 1) || (parm->multiplier > RWBUFSZ)) {
            printf ("\nMultiplier out of range.\n");
            return (FALSE);
         }
         printf ("\n");
      break;

      case 21:  /* write data */
         printf ("Total Bytes Written = ");
         printf ("(block size) * (blocking factor) * (multiplier).\n\n");
         printf ("Enter block size: ");
         gets (buf);
         parm->blk_size = atoi (buf);
         if ((parm->blk_size < 1) || (parm->blk_size > RWBUFSZ)) {
            printf ("\nBlock size out of range.\n");
            return (FALSE);
         }
         printf ("Enter blocking factor: ");
         gets (buf);
         parm->blk_fact = atoi (buf);
         if ((parm->blk_fact < 1) ||
               ((parm->blk_fact * parm->blk_size) > RWBUFSZ)) {
            printf ("\nBlocking factor out of range.\n");
            return (FALSE);
         }
         printf ("Enter multiplier: ");
         gets (buf);
         parm->multiplier = atoi (buf);
         if ((parm->multiplier < 1) || (parm->multiplier > RWBUFSZ)) {
            printf ("\nMultiplier out of range.\n");
            return (FALSE);
         }
         printf ("Enter pattern file name");
         printf (" (<ret> = random pattern): ");
         gets (parm->filename);
         if (parm->filename[0] == '\0') {  /* create our own random pattern */
            printf ("Enter random seed");
            printf (" (0 = auto generated seed): ");
            gets (buf);
            parm->seed = atoi (buf);
            if (parm->seed < 0) {
               printf ("\nRandom seed out of range.\n");
               return (FALSE);
            }
         }
         printf ("\n");
      break;

      case 22:  /* write file mark */
         printf ("Enter number of file marks to write: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         parm->mt_op = MTWEOF;
         printf ("\n");
      break;

      case 23:  /* erase tape */
         parm->mt_op = MTERASE;
      break;

      case 24:  /* rewind */
         parm->mt_op = MTREW;
      break;

      case 25:  /* retension */
         parm->mt_op = MTRETEN;
      break;

      case 26:  /* offline */
         parm->mt_op = MTOFFL;
      break;

      case 27:  /* load/unload tape */
         printf ("Enter selection");
         printf (" (1=load, 2=unload): ");
         gets (buf);
         parm->mt_op = atoi (buf);
         if ((parm->mt_op < 1) || (parm->mt_op > 2)) {
            printf ("\nInvalid selection.\n");
            return (FALSE);
         }
         if (parm->mt_op == 1)  parm->mt_op = STLOAD;
         else  parm->mt_op = STUNLOAD;
         printf ("\n");
      break;

      case 28:  /* forward space file */
         printf ("Enter number of files to space forward: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         parm->mt_op = MTFSF;
         printf ("\n");
      break;

      case 29:  /* backward space file */
         printf ("Enter number of files to space backward: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         parm->mt_op = MTBSF;
         printf ("\n");
      break;

      case 30:  /* forward space record */
         printf ("Enter number of records to space forward: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         parm->mt_op = MTFSR;
         printf ("\n");
      break;

      case 31:  /* backward space record */
         printf ("Enter number of records to space backward: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         parm->mt_op = MTBSR;
         printf ("\n");
      break;

      case 32:  /* locate end of data */
         parm->mt_op = MTEOM;
      break;

      case 33:  /* get record size */
         parm->mt_op = MTGRSZ;
      break;

      case 34:  /* set record size */
         printf ("Enter new record size: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         parm->mt_op = MTSRSZ;
         printf ("\n");
      break;

      case 35:  /* get device status */
      break;

      case 36:  /* get device information */
      break;

      case 37:  /* get media information */
      break;

      case 38:  /* get position */
         printf ("Select block type");
         printf (" (1=logical, 2=physical): ");
         gets (buf);
         parm->type = atoi (buf);
         if ((parm->type < 1) || (parm->type > 2)) {
            printf ("\nInvalid block type selection.\n");
            return (FALSE);
         }
         printf ("\n");
      break;

      case 39:  /* set position */
         printf ("Select block type");
         printf (" (1=logical, 2=physical): ");
         gets (buf);
         parm->type = atoi (buf);
         if ((parm->type < 1) || (parm->type > 2)) {
            printf ("\nInvalid block type selection.\n");
            return (FALSE);
         }
         printf ("Enter position value: ");
         gets (buf);
         parm->value = atoi (buf); printf ("\n");
      break;

      case 40:  /* get parameter */
         printf ("Select parameter type\n");
         printf ("(1=blocksize, 2=compression, 3=buffering, 4=immediate, \n"
            "5=trailer, 6=writeprotect, 7=acfmode, 8=capacity, 9=sili): ");
         gets (buf);
         parm->type = atoi (buf);
         if ((parm->type < 1)
            || ((parm->type > 9) && (parm->type < 12))
            || (parm->type > 14)) {
            printf ("\nInvalid parameter type selection.\n");
            return (FALSE);
         }
         printf ("\n");
      break;

      case 41:  /* set parameter */
         printf ("Select parameter type\n");
         printf ("(1=blocksize, 2=compression, 3=buffering, 4=immediate, \n"
            "5=trailer, 6=writeprotect, 7=acfmode, 8=capacity, 9=sili): ");
         gets (buf);
         parm->type = atoi (buf);
         if ((parm->type < 1)
            || ((parm->type > 9) && (parm->type < 12))
            || (parm->type > 14)
            || (parm->type == 7)) {
            printf ("\nInvalid parameter type selection.\n");
            return (FALSE);
         }
         printf ("Enter parameter value: ");
         gets (buf);
         parm->value = atoi (buf);
         printf ("\n");
      break;

      case 42:  /* sync buffer */
      break;

      case 43:  /* display message */
         printf ("Select message type");
         printf (" (1=msg0, 2=msg1, 3=flash0, 4=flash1, 5=alternate): ");
         gets (buf);
         parm->type = atoi (buf);
         if ((parm->type < 1) || (parm->type > 5)) {
            printf ("\nInvalid message type selection.\n");
            return (FALSE);
         }
         printf ("Enter message-0 text (8 characters or less): ");
         gets (buf);
         if ((strlen(buf) < 1) || (strlen(buf) > 8)) {
            printf ("\nInvalid message-0 length.\n");
            return (FALSE);
         }
         strcpy (parm->msg_0, buf);
         printf ("Enter message-1 text (8 characters or less): ");
         gets (buf);
         if ((strlen (buf) < 1) || (strlen (buf) > 8)) {
            printf ("\nInvalid message-1 length.\n");
            return (FALSE);
         }
         strcpy (parm->msg_1, buf);
         printf ("\n");
      break;

      case 44:  /* check device */
      break;

      case 45:  /* report density support */
      break;

      case 46:  /* read element device id */
      break;

      case 47:  /* audit range */
         printf ("Enter address of first element: ");
         gets (buf);
         parm->src = atoi (buf);
         printf ("Enter number of elements: ");
         gets (buf);
         parm->mt_count = atoi (buf);
         printf ("\n");
      break;

      case 48:  /* set all parms */
      break;

      case 49:  /* read/write tests */
      break;

      case 50:  /* Log Sense Page/Subpage */
         printf ("Enter page code in hex: ");
         gets (buf);
         parm->type = (int) strtol ( buf, NULL, 16);
         printf ("\n");
         printf ("Enter subpage code in hex: ");
         gets (buf);
         parm->value = (int) strtol (buf, NULL, 16);
         printf ("\n");
      break;

      case 51:  /* Mode Sense Page/Subpage */
         printf ("Enter page code in hex: ");
         gets (buf);
         parm->type = (int) strtol ( buf, NULL, 16);
         printf ("\n");
         printf ("Enter subpage code in hex: ");
         gets (buf);
         parm->value = (int) strtol (buf, NULL, 16);
         printf ("\n");
      break;

      case 52:  /* Get Driver Information */
      break;

      case 53:  /* Enable/Disable path */
      break;

      case 54:  /* Query Device Paths */
      break;

      case 55:  /* Read Block Limits */
      break;

      case 56:  /* exchange medium */
         printf ("Enter source address: ");
         gets (buf);
         parm->src = atoi (buf);
         printf ("Enter first destination address: ");
         gets (buf);
         parm->dst1 = atoi (buf);
         printf ("Enter second destination address: ");
         gets (buf);
         parm->dst2 = atoi (buf);
         printf ("\n");
      break;

      case 57:  /* Get Encryption State  */
      break;

      case 58: /* EKM test */
      break;

      case 59:  /* Get Density */
      break;

   }

   return (TRUE);
}


/*---------------------------------------------------------------------------*/
/*        Name: exec_cmd                                                     */
/*    Synopsis: Execute command indicated by parm->opcode.                   */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int exec_cmd ( parm_t *parm )
{
   int page;
   int subpage;
   int rc;

   switch (parm->opcode) {
      case 1:  /* open device */
         dev_fd = -1;
         dev_errno = 0;
         smc_errno = 0;
         if (strlen(dev_name) > 0) {
            dev_fd = open (dev_name, parm->mode | O_NDELAY);
            if (dev_fd < 0) {
               dev_errno = errno;
            }
            else {
               opened    = TRUE;
               str_dev_p = dev_name;
               PRINTF ("Device %s opened.\n", str_dev_p);
            }
         }
         if ((!opened) && (strlen(smc_name) > 0)) {
            dev_fd = open (smc_name, O_RDONLY | O_NDELAY);
            if (dev_fd < 0) {
               smc_errno = errno;
            }
            else {
               opened    = TRUE;
               str_dev_p = smc_name;
               PRINTF ("Device %s opened.\n", str_dev_p);
               if (parm->mode != O_RDONLY) {
                  PRINTF ("Open mode forced read-only for medium changer.\n");
               }
            }
         }
         if (opened) {
            rc = get_pid(pid);
            if (rc != 0) {
               (void) fprintf (stderr, "Failure obtaining device PID\n");
               return (EINVAL);
            }
         }
         else {
            if (((dev_errno == ENOENT) && (smc_errno == ENOENT))
               || ((dev_errno && (dev_errno != ENOENT))
                     && (smc_errno && (smc_errno != ENOENT)))) {
               (void) fprintf (stderr,
                  "Error %d (%s) opening %s.\n"
                  "Error %d (%s) opening %s.\n",
                  dev_errno, strerror(dev_errno), dev_name,
                  smc_errno, strerror(smc_errno), smc_name);
            }
            else if (dev_errno && (dev_errno != ENOENT)) {
               (void) fprintf (stderr,
                  "Error %d (%s) opening %s.\n",
                  dev_errno, strerror(dev_errno), dev_name);
               if (dev_errno == EACCES) {
                  if ((parm->mode & O_WRONLY) || (parm->mode & O_RDWR)) {
                    FPRINTF (stderr, "   (If tape is write-protected, try opening read-only.)\n");
                    FPRINTF (stderr, "   (Or, the encryption setup may be incorrect if the drive is 3592 E05.)\n");
                  }
                  else {
                    FPRINTF (stderr, "   (The encryption setup may be incorrect if the drive is 3592 E05.)\n");
                  }
               }
            }
            else if (smc_errno && (smc_errno != ENOENT)) {
               (void) fprintf (stderr,
                  "Error %d (%s) opening %s.\n",
                  smc_errno, strerror(smc_errno), smc_name);
            }
            else if (dev_errno == ENOENT) {
               (void) fprintf (stderr,
                  "Error %d (%s) opening %s.\n",
                  dev_errno, strerror(dev_errno), dev_name);
            }
            else if (smc_errno == ENOENT) {
               (void) fprintf (stderr,
                  "Error %d (%s) opening %s.\n",
                  smc_errno, strerror(smc_errno), smc_name);
            }
            else {
               (void) fprintf (stderr, "tapeutil error in open routine.\n");
               return (-1);
            }
         }
      if(interactive) menu();
      break;

      case 2:  /* close device */
         rc = close (dev_fd);
         if (rc < 0) {
            (void) fprintf (stderr, "Error %d (%s) closing %s\n",
               errno, strerror(errno), str_dev_p);
         }
         else {
            PRINTF ("Device %s closed.\n", dev_name);
            opened = FALSE;
         }
      break;

      case 3:  /* query subsystem number */
         rc = svc_device_sn ();
      break;

      case 4:  /* query microcode level */
         rc = svc_ucode_level ();
      break;

      case 5:  /* force dump */
         rc = svc_force_dump ();
      break;

      case 6:  /* store dump */
         rc = svc_store_dump (parm->filename);
      break;

      case 7:  /* download microcode */
         rc = svc_load_ucode (parm->filename);
      break;

      case 8:  /* format cartridge */
         rc = svc_format_tape ();
      break;

      case 9:  /* test unit ready */
         rc = scsi_test_unit_ready ();
      break;

      case 10:  /* inquiry */
         if (( parm->type == 0 ) || ( parm->type == -1 ))
               rc = scsi_inquiry ();
         else {
               page = parm->value & 0x00FF;    /* Mask for page code */
               rc = scsi_inquiry_page (page);
         }
      break;

      case 11:  /* request sense */
         rc = scsi_request_sense ();
      break;

      case 12:  /* reserve */
         rc = scsi_reserve ();
      break;

      case 13:  /* release */
         rc = scsi_release ();
      break;

      case 14:  /* move medium */
         rc = smc_move_medium (parm->src, parm->dst);
      break;

      case 15:  /* position to element */
         rc = smc_pos_to_elem (parm->dst);
      break;

      case 16:  /* element information */
         rc = smc_element_info ();
      break;

      case 17:  /* inventory */
         rc = smc_inventory ();
      break;

      case 18:  /* audit */
         rc = smc_audit ();
      break;

      case 19:  /* lock/unlock door */
         rc = smc_lock_door (parm->value);
      break;

      case 20:  /* read data */
         rc = st_read (parm->blk_size, parm->blk_fact, parm->multiplier);
      break;

      case 21:  /* write data */
         rc = st_write (parm->blk_size, parm->blk_fact,
            parm->multiplier, parm->seed, parm->filename);
      break;

      case 22:  /* write file mark */
      case 23:  /* erase tape */
      case 24:  /* rewind */
      case 25:  /* retension */
      case 26:  /* offline */
      case 27:  /* load/unload tape */
      case 28:  /* forward space file */
      case 29:  /* backward space file */
      case 30:  /* forward space record */
      case 31:  /* backward space record */
      case 32:  /* locate end of data */
      case 33:  /* get record size */
      case 34:  /* set record size */
         rc = st_tape_op (parm->mt_op, parm->mt_count);
      break;

      case 35:  /* get device status */
         rc = st_get_device_status ();
      break;

      case 36:  /* get device information */
         rc = st_get_device_info ();
      break;

      case 37:  /* get media information */
         rc = st_get_media_info ();
      break;

      case 38:  /* get position */
         rc = st_get_position (parm->type);
      break;

      case 39:  /* set position */
         rc = st_set_position (parm->type, parm->value);
      break;

      case 40:  /* get parameter */
         rc = st_get_parm (parm->type);
      break;

      case 41:  /* set parameter */
         rc = st_set_parm (parm->type, parm->value);
      break;

      case 42:  /* sync buffer */
         rc = st_sync_buffer ();
      break;

      case 43:  /* display_message */
         rc = st_display_msg (parm->type, parm->msg_0, parm->msg_1);
      break;

      case 44:  /* check device */
         rc = svc_check_device ();
      break;

      case 45:  /* report density support */
         rc = st_report_density_support ();
      break;

      case 46:  /* read element device id */
         rc = smc_read_element_devids ();
      break;

      case 47:  /* audit range */
         rc = smc_audit_range (parm->src, parm->mt_count);
      break;

      case 48:  /* query/set all parameters */
         rc = st_set_all_parms();
      break;

      case 49:  /* read/write tests */
         rc = rwtest_menu();
      break;

      case 50:  /* Log Sense Page */
         page = parm->type & 0x00FF;    /* Mask for page code */
         subpage = parm->value & 0x00FF;    /* Mask for subpage code */
         rc = scsi_log_sense_page (page, subpage);
      break;

      case 51:  /* Mode Sense Page */
         page = parm->type & 0x00FF;    /* Mask for page code */
         subpage = parm->value & 0x00FF;    /* Mask for subpage code */
         if (subpage)
            rc = scsi_mode_sense_subpage (page, subpage);
         else
            rc = scsi_mode_sense_page (page);
      break;

      case 52:  /* Get Driver Information */
         rc = get_driver_info ();
      break;

      case 53:  /* Enable/Disable path */
         rc = scsi_enable_path ();
      break;

      case 54:  /* Query Device Paths */
         rc = scsi_query_device_paths ();
      break;

      case 55:  /* Read Block Limits */
         rc = st_read_block_limits ();
      break;

      case 56:  /* exchange medium  */
         rc = smc_exchange_medium (parm->src, parm->dst1, parm->dst2);
      break;

      case 57:  /* Get Encryption State  */
         rc = st_get_encryption_state ();
      break;

      case 58: /* EKM test */
         rc = encryption_test();
      break;

      case 59:  /* Get Density */
         rc = st_get_density ();
      break;

   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: menu                                                         */
/*    Synopsis: Display menu.                                                */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static void menu ()
{
      printf (" +----------------------------------------------------------------------------+\n");
      printf (" |               IBM SCSI TAPE & MEDIUM CHANGER UTILITY PROGRAM               |\n");
      printf (" +------------------------+--------------------------+------------------------+\n");
      printf (" |                    << GENERAL COMMANDS >>                                  |\n");
      printf (" |  1: Open Device           M: Menu Refresh         D: Device Type           |\n");
      printf (" |  2: Close Device          Q: Quit Program                                  |\n");
      printf (" |                                                                            |\n");
      printf (" |                    << SERVICE COMMANDS >>                                  |\n");
      printf (" |  3: Serial/Subsys Number  6: Store Dump          53: Enable/Disable Path   |\n");
      printf (" |  4: Query Microcode Level 7: Download Microcode  54: Query Device Paths    |\n");
      printf (" |  5: Force Dump           52: Get Driver Info     58: EKM Test              |\n");
      printf (" |                                                                            |\n");
      printf (" |                    << BASIC COMMANDS >>                                    |\n");
      printf (" |  9: Test Unit Ready      12: Reserve             50: Log Sense Page        |\n");
      printf (" | 10: Inquiry              13: Release             51: Mode Sense Page       |\n");
      printf (" | 11: Request Sense                                                          |\n");
      printf (" +------------------------+--------------------------+------------------------+\n");


   /* if ((opened) && (strstr(str_dev_p, "smc") != (char *)NULL)) { */

   if (opened && medium_chgr) {
      printf (" |                    << MEDIUM CHANGER COMMANDS >>                           |\n");
      printf (" | 14: Move Medium          17: Inventory            46: Read DeviceIDs       |\n");
      printf (" | 15: Position To Element  18: Audit                47: Audit Range          |\n");
      printf (" | 16: Element Information  19: Lock/Unlock Door     56: Exchange Medium      |\n");
      printf (" +----------------------------------------------------------------------------+\n");
   }

   if (opened && attached_medium_chgr) {
      printf (" |                    << ATTACHED MEDIUM CHANGER COMMANDS >>                  |\n");
      printf (" | 14: Move Medium          17: Inventory            46: Read DeviceIDs       |\n");
      printf (" | 16: Element Information  19: Lock/Unlock Door                              |\n");
      printf (" +----------------------------------------------------------------------------+\n");
   }

   if (opened && tape_drv) {

      printf (" |                    << TAPE DRIVE COMMANDS >>                               |\n");
      printf (" | 20: Read Data            29: Backward Space File  38: Get Position         |\n");
      printf (" | 21: Write Data           30: Forward Space Rec    39: Set Position         |\n");
      printf (" | 22: Write File Mark      31: Backward Space Rec   42: Sync Buffer          |\n");
      printf (" | 23: Erase Tape           32: Locate End Of Data   43: Display Message      |\n");
      printf (" | 24: Rewind               33: Get Record Size      45: Report Density       |\n");
      printf (" | 25: Retension            34: Set Record Size      48: Query/Set Parameters |\n");
      printf (" | 26: Offline              35: Get Device Status    49: Read/Write Tests     |\n");
      printf (" | 27: Load/Unload Tape     36: Get Device Info      55: Read Block Limits    |\n");
      printf (" | 28: Forward Space File   37: Get Media Info       57: Get Encryption State |\n");
      printf (" | 59: Get Density                                                            |\n");
      printf (" +----------------------------------------------------------------------------+\n");

   }
}


/*---------------------------------------------------------------------------*/
/*        Name: newusage                                                     */
/*    Synopsis: Display program usage information.                           */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static void newusage ()
{
   printf ("Usage: tapeutil [-f Device Subcommand [Subcommand ...]]      \n");
   printf ("                                                             \n");
   printf (" General Subcommands:                                        \n");
   printf ("   devinfo     inquiry [Page]  print \"Text\"                \n");
   printf ("   reserve     release         reqsense                      \n");
   printf ("   tur         vpd             sleep \"Seconds\"             \n");
   printf ("                               loop [Count]                  \n");
   printf ("                                                             \n");
   printf (" Medium Changer Subcommands:                                 \n");
   printf ("   allow       prevent         audit [Address [Count]]       \n");
   printf ("   inventory   elementinfo     position \"Destination\"      \n");
   printf ("   devids      move \"Source\" \"Destination\"               \n");
   printf ("                                                             \n");
   printf (" Tape Subcommands:                                           \n");
   printf ("   parms       fsf [Count]     read -d Destination [-c Count]\n");
   printf ("   autoload    bsf [Count]     write -s Source               \n");
   printf ("   noautoload  fsr [Count]     append                        \n");
   printf ("   compress    bsr [Count]     setpos [Blockid]              \n");
   printf ("   nocompress  eof [Count]     qrypos                        \n");
   printf ("   sili        weof [Count]    display \"Message\"           \n");
   printf ("   nosili      erase           load                          \n");
   printf ("   status      rewind          unload                        \n");
   printf ("   density     retension       sync                          \n");
   printf ("   seod        offline         modepage \"Page\" \"Subpage\" \n");
   printf ("   mediainfo   rewoffl         logpage \"Page\" \"Subpage\"  \n");
   printf ("   serialnum   subsysnum       readblk                       \n");
   printf ("   encryption  getrecsize      setrecsize [size]             \n");
   printf ("   rtest  [-b Blocksize] [-c Count] [-r Repetition]          \n");
   printf ("   wtest  [-b Blocksize] [-c Count] [-r Repetition]          \n");
   printf ("   rwtest [-b Blocksize] [-c Count] [-r Repetition] [-s File]\n");
   printf ("                                                             \n");
   printf (" Service Aid Subcommands:                                    \n");
   printf ("   ucodelevel   dump [\"Filename\"]   ucode \"Filename\"     \n");
   printf ("   ekmtest                                                   \n");
   printf ("   \n");
}


/*---------------------------------------------------------------------------*/
/*        Name: usage                                                        */
/*    Synopsis: Display program usage information.                           */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static void usage (int extended)
{
   if (!verbose)  return;
      printf ("                                                               \n");
      printf (" USAGE:                                                        \n");
      printf ("   tapeutil -f device-file -o operation [-w open-mode]         \n");
      printf ("     [-s move-source] [-d move-destination] [-c op-count]      \n");
      printf ("     [-b block-size] [-n blocking-factor] [-m multiplier]      \n");
      printf ("     [-r random-seed] [-z ucode/dump/pattern-filename]         \n");
      printf ("     [-t msg/gpo/spo/lod/gpa/spa-type/inq-type]                \n");
      printf ("     [-x lck/spo/spa-value/page-code]                          \n");
      printf ("     [-y msg0,msg1] [-v (verbose-mode)] [-h,-? (usage-help)]   \n");
      printf ("                                                               \n");
      printf (" OPERATIONS:                                                   \n");
      printf ("   Service Commands:        qsn, qmc, fdp, sdp, dmc, fmt       \n");
      printf ("   Basic SCSI Commands:     tur, inq, req, res, rel, chk       \n");
      printf ("   Medium Changer Commands: mov, pos, ele, inv, aud, lck       \n");
      printf ("                            did                                \n");
      printf ("   Tape Drive Commands:     rea, wri, eof, era, rew, ret       \n");
      printf ("                            off, lod, fsf, bsf, fsr, bsr       \n");
      printf ("                            eod, grs, srs, gds, gdi, gmi       \n");
      printf ("                            gpo, spo, gpa, spa, syn, msg       \n");
      printf ("                            chk, rds                           \n");
      printf (" OPEN-MODES:                                                   \n");
      printf ("   1=r/w, 2=rdo, 3=wro, 4=app                                  \n");
      printf (" TYPES:                                                        \n");
      printf ("   msg:     1=ms0, 2=ms1, 3=fl0, 4=fl1, 5=alt                  \n");
      printf ("   gpo/spo: 1=log, 2=phy                                       \n");
      printf ("   gpa/spa: 1=blk, 2=cmp, 3=buf, 4=imm, 5=trl                  \n");
      printf ("            6=wpr, 7=acf, 8=cap, 9=sil                         \n");
      printf ("   inq:     0=std, 1=pgc                                       \n");
      printf ("   lod:     1=load, 2=unload                                   \n");
      printf ("                                                               \n");
      if (!verbose) printf ("SCCS ID: %s\n", sccsid);
}


/*****************************************************************************/
/*                             SERVICE FUNCTIONS                             */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/*        Name: srv_device_sn                                                */
/*    Synopsis: Display tape drive subsystem number, as used by lmcpd        */
/*              and IBM 3494 Tape Subsystem.                                 */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_device_sn ()
{
   int rc;
   device_sn_t device_sn;
   inquiry_page_t inquiry_page;
   char serial_num[20];

   memset ((char *)&inquiry_page, (char)0, sizeof (inquiry_page_t));
   inquiry_page.page_code = (uchar) 0x80;
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY_PAGE, &inquiry_page))) {
      bcopy(&inquiry_page.data[4], serial_num, inquiry_page.data[3]);
      serial_num[inquiry_page.data[3]] = '\0';
      PRINTF ("Serial Number: %s\n", serial_num);
      /* dump_bytes ((char *)&inquiry_page.data, inquiry_page.data[3]+4); */
   }
   else {
      PRINTF ("IOC_INQUIRY_PAGE for page 0x%X failed.\n", 0x80);
      PRINTF ("\n");
      scsi_request_sense ();
   }

   if (medium_chgr) {
      return (1);
   }

   if (!(rc = ioctl (dev_fd, STIOC_DEVICE_SN, &device_sn))) {
      PRINTF ("IBM Library Device Subsystem Number: %x\n", device_sn);
   }
   else {
      PERROR ("Failure obtaining IBM Library Device Subsystem Number");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: svc_ucode_level                                              */
/*    Synopsis: Display tape device microcode level.                         */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_ucode_level ()
{
   int rc;
   sense_data_t sense_data;
   int save_verbose;

   save_verbose = verbose;
   verbose = FALSE;
   (void) scsi_inquiry ();
   verbose = save_verbose;
   if ((device.ultrium)
      || (device.ult3581)
      || (device.ult3582)
      || (device.ult3583)
      || (device.ult3584)
      || (device.ult3581_ta2)
      || (device.ult3576)
      || (device.ult3573)
      || (device.lib3577)
      || (device.t3490)) {
      strncpy (ucode_level, (const char *)inquiry_data.rev, 4);
      ucode_level[4] = '\0';
      PRINTF ("   Firmware Revision Level: %s\n", ucode_level);
      PRINTF ("   (from Inquiry Standard Data)\n");
   }
   else if ((device.t3570) || (device.t3590) || (device.t3592)) {
      if (!(rc = ioctl (dev_fd, IOC_REQUEST_SENSE, &sense_data))) {
         strncpy (ucode_level, (const char *)&sense_data.vendor[54], 7);
         ucode_level[7] = '\0';
         PRINTF ("%s microcode level: %s\n", dev_name, ucode_level);
         PRINTF ("   (from Sense Data)\n");
         strncpy (ucode_level, (const char *)inquiry_data.rev, 4);
         ucode_level[4] = '\0';
         PRINTF ("Firmware Revision Level: %s.\n", ucode_level);
         PRINTF ("   (from Inquiry Standard Data)\n");
      }
      else {
         PERROR ("Failure obtaining device microcode level");
         PRINTF ("\n");
         scsi_request_sense ();
      }
   }
   else {
      (void) fprintf (stderr, "Cannot identify device type.\n");
      return (-1);
   }
   return (rc);
}
/* svc_ucode_level */


/*---------------------------------------------------------------------------*/
/*        Name: svc_force_dump                                               */
/*    Synopsis: Force tape device dump.                                      */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_force_dump ()
{
   int rc;

   if ((device.ult3581)
      || (device.ult3582)
      || (device.ult3583)
      || (device.ult3584)
      || (device.ult3581_ta2)
      || (device.ult3576)
      || (device.ult3573)
      || (device.lib3577)
      || ((device.t3590) && (medium_chgr))
      || ((device.t3570) && (medium_chgr))
      ) {
      FPRINTF (stderr, "%s (%s)does not support force_dump.\n", str_dev_p, pid);
      return (1);
   }

   if (!(rc = ioctl (dev_fd, IOC_FORCE_DUMP, 0))) {
      PRINTF ("Dump completed successfully on tape device %s.\n", dev_name);
   }
   else {
      PERROR ("Failure performing tape device dump");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*------------------------------------------------------------
** Name:       read_buflist
**    Synopsis:Read from list of device buffers to specified file descriptor.
**     Returns:Total bytes read from all specified device buffers.
** Called by:  svd_store_dump()
** Call with:  read_buflist(int fp, buffer_io_t *biop, int desc_mode,
**                           int data_mode, int  bufid, ..., END_BUFLIST );
**------------------------------------------------------------*/
static long long read_buflist(int fp, buffer_io_t *biop, int desc_mode, int data_mode, int  bufid, ...)
{
   #define     MAXBUFS 20
   va_list     ap;
   uchar       uc_desc_mode;
   uchar       uc_data_mode;
   uint        next_read;
   int         tmpbufid;
   uint        offset;
   uint        tot_buflen;
   int         bufcount    = 0;
   int         rc          = 0;
   long long   tot_dumplen = 0;
   uchar *bufp;
   uchar       bufid_list[MAXBUFS];
   int         bufidx      = 1;

   uc_desc_mode = (uchar) desc_mode;
   uc_data_mode = (uchar) data_mode;
   va_start(ap, bufid);
   bufid_list[0] = (uchar) bufid;
   while((bufidx < MAXBUFS) && ((tmpbufid = va_arg(ap, int)) >= 0)) {
      bufid_list[bufidx++] = (uchar) tmpbufid;
   }
   va_end(ap);
   bufcount = bufidx;
   bufp = biop->buffer;

   for (bufidx = 0; bufidx < bufcount; bufidx++) {
      biop->mode   = uc_desc_mode;
      biop->offset = 0;
      biop->size   = 4;
      biop->id     = bufid_list[bufidx];
      if (rc = ioctl (dev_fd, IOC_READ_BUFFER, biop)) {
         PERROR ("Failure retrieving dump size");
         PRINTF ("\n");
         errno = rc;
         return (-1);
      }
      tot_buflen = ((bufp[1] << 16) + (bufp[2] << 8) + (bufp[3]));
      offset    = 0;
      next_read = IOBUFSZ;
      biop->mode   = uc_data_mode;
      while (offset < tot_buflen) {
         if ((tot_buflen - offset) < next_read)  next_read = tot_buflen - offset;

         biop->offset = offset;
         biop->size   = next_read;

         PRINTF ("Reading Buffer 0x%2.2X,  Offset %6d,  Bytes %6d \n",
            bufid_list[bufidx], offset, next_read );

         if (rc = ioctl (dev_fd, IOC_READ_BUFFER, biop)) {
            PERROR ("Failure retrieving dump buffer");
            PRINTF ("\n");
            errno = rc;
            return (-1);
         }

         if (write (fp, bufp, next_read) != next_read) {
            PERROR ("Failure writing dump to destination file");
            errno = rc;
            return (-1);
         }
         offset += next_read;
      }
      tot_dumplen += tot_buflen;
   }
   return (tot_dumplen);
}
/* read_buflist */

/*---------------------------------------------------------------------------*/
/*        Name: svc_store_dump                                               */
/*    Synopsis: Store tape device dump to tape cartridge or host file.       */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_store_dump ( char *filename )
{
   int rc;
   int file;
   long long dump_size;
   buffer_io_t buffer_io;
   uchar *buf;
   int            log_data_len, data_len = 0;
   log_sns_pg_t   log_sns_pg;
   log_page_hdr_t *log_page_hdr_p = (log_page_hdr_t *)(&log_sns_pg.data);

   if ((device.t3590 && medium_chgr)
    || (device.t3570 && medium_chgr)
    || (device.ult3576)) {
      FPRINTF (stderr, "%s (%s) does not support store_dump.\n", str_dev_p, pid);
      FPRINTF (stderr, "Use the associated tape drive for this device, instead.\n");
      return (1);
   }
   if (!strcmp (filename, "tapedrive")) {
      if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) {
         if (!(rc = ioctl (dev_fd, IOC_STORE_DUMP, 0))) {
            PRINTF ("Dump stored successfully on tape device %s.\n", dev_name);
         }
         else {
            PERROR ("Failure storing dump on tape device");
            PRINTF ("\n");
            scsi_request_sense ();
         }
      }
      else {
         PERROR ("Tape device not ready or tape not mounted");
         PRINTF ("\n");
         scsi_request_sense ();
      }
   }
   else {
      if (!(buf = malloc (IOBUFSZ))) {
         PERROR ("Failure allocating IO buffer");
         return (errno);
      }
      memset (buf, 0, IOBUFSZ);

      if ((file = open (filename, O_WRONLY | O_CREAT | O_EXCL, 0644)) < 0) {
         (void) fprintf (stderr, "Error %d (%s) opening dump destination file.\n",
            errno, strerror(errno));
         if (errno == EEXIST) {
            (void) fprintf (stderr, "Erase this file or specify another name.\n");
         }
         free (buf);
         return (errno);
      }
      memset ((char *)&buffer_io, 0, sizeof(buffer_io_t));
      buffer_io.buffer = buf;
      if (device.ultrium) {
         dump_size = read_buflist( file, &buffer_io,
            ULT_DESC_ONLY_MODE, ULT_DATA_ONLY_MODE, ULT_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ult3581) {
         dump_size = read_buflist( file, &buffer_io,
            ULT3581_DESC_ONLY_MODE, ULT3581_DATA_ONLY_MODE, ULT3581_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ult3582) {
         dump_size = read_buflist( file, &buffer_io,
            ULT3582_DESC_ONLY_MODE, ULT3582_DATA_ONLY_MODE, ULT3582_DUMP_BUF, ULT3582_CMD_BUF, END_BUFLIST );
      }
      else if (device.ult3583) {
         dump_size = read_buflist( file, &buffer_io,
            ULT3583_DESC_ONLY_MODE, ULT3583_DATA_ONLY_MODE, ULT3583_DUMP_BUF, ULT3583_CMD_BUF, END_BUFLIST );
      }
      else if (device.ult3584) {
         dump_size = read_buflist( file, &buffer_io,
            ULT3584_DESC_ONLY_MODE, ULT3584_DATA_ONLY_MODE, ULT3584_DUMP_BUF, END_BUFLIST );
      }
      else if (device.ult3581_ta2) {
         dump_size = read_buflist( file, &buffer_io,
           ULT3581_TA2_DESC_ONLY_MODE, ULT3581_TA2_DATA_ONLY_MODE, ULT3581_TA2_DUMP_BUF, END_BUFLIST );
      }
      else if ((device.t3570) || (device.t3590) || (device.t3592)) {
         dump_size = read_buflist( file, &buffer_io,
            MAG_DESC_ONLY_MODE, MAG_DATA_ONLY_MODE, MAG_DUMP_BUF, END_BUFLIST );
      }
      else if (device.lib3577 || device.ult3573) {
         memset ((char *)&log_sns_pg, 0, sizeof (log_sns_pg_t));
         if (device.lib3577) log_sns_pg.page_code = 0x6;
         if (device.ult3573) log_sns_pg.page_code = 0x33;

         rc = ioctl (dev_fd, IOC_LOG_SENSE_PAGE, &log_sns_pg);

         if (!rc) {
            data_len = (uint) (((log_page_hdr_p->len[0] << 8) | log_page_hdr_p->len[1]) + 4);
            log_data_len = MIN(data_len, sizeof log_sns_pg.data);
            dump_size = write (file, (uchar *)(&log_sns_pg.data), log_data_len);
            if (dump_size == -1) {
               PERROR ("Failure writing dump to destination file");
               rc = errno;
            }
         }
      }
      else {
         (void) fprintf (stderr, "Cannot identify device type.\n");
         return (-1);
      }

      free (buf);
      close (file);
      if (dump_size > 0) {
         PRINTF ("Dump stored to %s successfully (%lld bytes).\n", filename, dump_size);
      }
      else {
         PRINTF ("Dump failure: dump size is 0. \n");
         return (errno);
      }
   }
   return (rc);
}
/* svc_store_dump */


/*---------------------------------------------------------------------------*/
/*        Name: svc_load_ucode                                               */
/*    Synopsis: Download microcode to tape device.                           */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_load_ucode ( char *filename )
{
   int rc = 0;
   int i;
   int file;
   uint ucode_size = 0;
   uint remaining;
   uint offset = 0;
   long iobuf_sz;
   long uc_chunksz;  /* Ucode chunksize */
   buffer_io_t buffer_io;
   tape_op_t tape_op;
   uchar *buf;
   int wait_reset = 30;
   struct stat f_info;
   uint retry_time;
   char cur_ucode_level[32] = "";

   if (device.ult3573)
      iobuf_sz = IOBUFSZ - 1;
   else
      iobuf_sz = IOBUFSZ;

   if (!(buf = malloc (iobuf_sz))) {
      PERROR ("Failure allocating IO buffer");
      return (errno);
   }
   if ((file = open (filename, O_RDONLY)) < 0) {
      PERROR ("Failure opening microcode source file");
      rc = errno;
      free (buf);
      return (rc);
   }
   if (fstat(file, &f_info)) {
      PERROR ("Failure statting microcode source file");
      rc = errno;
      free (buf);
      return (rc);
   }
   if ((uc_chunksz = read (file, buf, iobuf_sz)) < 0) {
      PERROR ("Failure reading microcode image.");
      rc = errno;
      free (buf);
      close (file);
      return (rc);
   }
   if ((! medium_chgr) && (! device.ultrium) && (! device.t3592)) {
      if (!(ioctl (dev_fd, IOC_TEST_UNIT_READY))) {
         tape_op.mt_op = MTOFFL;
         tape_op.mt_count = 0;
         PRINTF ("Unloading device.\n");
         if (rc = ioctl (dev_fd, STIOC_TAPE_OP, &tape_op)) {
           PERROR ("Failure unloading device");
           PRINTF ("\n");
           scsi_request_sense ();
           free (buf);
           close (file);
           return (rc);
         }
      }
   }
   PRINTF ("\nCurrent microcode level...\n");
   (void) svc_ucode_level();
   strcpy (ucode_level, cur_ucode_level);

   if (device.ultrium) {
      wait_reset       = 60;
      buffer_io.mode   = ULT_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3581) {
      wait_reset       = 60;
      buffer_io.mode   = ULT3581_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3581_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3582) {
      wait_reset       = 120;
      buffer_io.mode   = ULT3582_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3582_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3583) {
      wait_reset       = 120;
      buffer_io.mode   = ULT3583_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3583_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3584) {
      buffer_io.mode   = ULT3584_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3584_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3581_ta2) {
      wait_reset       = 60;
      buffer_io.mode   = ULT3581_TA2_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3581_TA2_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3573) {
      wait_reset       = 60;
      buffer_io.mode   = ULT3573_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3573_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.ult3576) {
      wait_reset       = 120;
      buffer_io.mode   = ULT3576_DWNLD_UCODE_MODE;
      buffer_io.id     = ULT3576_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if (device.lib3577) {
      wait_reset       = 120;
      buffer_io.mode   = LIB3577_DATA_ONLY_MODE;
      buffer_io.id     = LIB3577_UCODE_BUF;
      ucode_size       = f_info.st_size;
   }
   else if ((device.t3570) || (device.t3590) || (device.t3592)) {
      buffer_io.mode   = MAG_DWNLD_UCODE_MODE;
      buffer_io.id     = MAG_UCODE_BUF;
      ucode_size = ((buf[0]) + (buf[1] << 8) + (buf[2] << 16) + (buf[3] << 24));
   }
   else {
      (void) fprintf (stderr, "Cannot identify device type.\n");
      return (-1);
   }
   buffer_io.buffer = buf;
   remaining = ucode_size;

   if (device.ult3583) {
      for ( i = 0; i < 36; i++) {
         if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) break;
      }
      if (rc) {
         (void) fprintf (stderr, "Device not ready.\n");
         /* return (rc) */;
      }
   }
   while (remaining) {
      buffer_io.offset = offset;
      buffer_io.size   = (uint) MIN(uc_chunksz, remaining);
      PRINTF ("Downloading %7d bytes at offset %7d.\n", uc_chunksz, offset);
      if (remaining <= iobuf_sz) {
         PRINTF ("Waiting for device to write flash-prom.\n");
      }
      if (rc = ioctl (dev_fd, IOC_WRITE_BUFFER, &buffer_io)) {
         PERROR ("Downloading microcode to device");
         PRINTF ("\n");
         scsi_request_sense ();
         free (buf);
         close (file);
         return (rc);
      }
      offset += uc_chunksz;
      if ((uc_chunksz = read (file, buf, iobuf_sz)) < 0) {
         PERROR ("Failure reading microcode file");
         rc = errno;
         free (buf);
         close (file);
         return (rc);
      }
      if ((ucode_size - offset) < uc_chunksz)  uc_chunksz = ucode_size - offset;
      remaining = ucode_size - offset;

      if (!uc_chunksz) break;
  }

  if (device.lib3577) {
     buffer_io.mode   = LIB3577_DWNLD_UCODE_MODE;
     buffer_io.id     = LIB3577_UCODE_BUF;
     buffer_io.size   = 0;
     buffer_io.offset = 0;

     if (rc = ioctl (dev_fd, IOC_WRITE_BUFFER, &buffer_io)) {
        PERROR ("Flush microcode from buffer to device");
        PRINTF ("\n");
        scsi_request_sense ();
        free (buf);
        close (file);
        return (rc);
     }
  }

  free (buf);
  close (file);

  PRINTF ("Microcode downloaded to %s successfully (%d bytes).\n", str_dev_p, offset);
  rc = 1;

  if (device.ult3584)
     retry_time = 10;
  else
     retry_time = 3;

  printf ("Waiting 5 minutes for device to reset (will retry up to %d times)...\n", retry_time);
  (void) sleep(300);

  for (i = 0; i < retry_time && rc != 0; i++) {
      if (i > 0) {
         PRINTF ("Waiting %d seconds for device to reset...\n", wait_reset);
         (void) sleep(wait_reset);
      }
      PRINTF ("Trying INQUIRY...\n");
      rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data);
      if (rc == 0  && device.ult3584) {
         if ((strcmp (ucode_level, cur_ucode_level)) == 0) rc = 5;
      }
  }
  if (rc != 0) {
      rc = errno;
      (void) fprintf (stdout, "INQUIRY rc %d (%s)\n", rc, strerror(rc));
  }
  else {
      PRINTF ("\nNew microcode level...\n");
      (void) svc_ucode_level();
  }
  PRINTF ("The tape is unloaded from the drive while updating the microcode.\n");
  return (rc);
}
/* svc_load_ucode */


/*---------------------------------------------------------------------------*/
/*        Name: svc_format_tape                                              */
/*    Synopsis: Format tape cartridge mounted in tape device.                */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_format_tape ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) {
      if (!(rc = ioctl (dev_fd, STIOC_FORMAT_TAPE, 0))) {
      PRINTF ("Tape cartridge formatted successfully.\n");
      }
      else {
      PERROR ("Failure formatting tape cartridge");
      PRINTF ("\n");
      scsi_request_sense ();
      }
   }
   else {
      PERROR ("Tape device not ready or tape not mounted");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: svc_check_device                                             */
/*    Synopsis: Check device attachment.                                     */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int svc_check_device ()
{
   #define VIDSZ 8
   int rc;
   char vid[VIDSZ + 1];
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data))) {
      strncpy (vid, (const char *)inquiry_data.vid, VIDSZ);
      vid[VIDSZ] = '\0';
      if (((inquiry_data.type == 1) || (inquiry_data.type == 8)) &&
         (!(strncmp (vid, "IBM", 3)))) {
         vid[3] = '\0';
         PRINTF ("%s %s configured at %s.\n", vid, pid, str_dev_p);
      }
      else {
      PRINTF ("Unsupported %s %s detected at %s.\n", vid, pid, str_dev_p);
      rc = ENODEV;
      }
   }
   else {
      PERROR ("Unable to verify device attachment");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}

/*---------------------------------------------------------------------------*/
/* Name:        get_driver_info                                              */
/* Synopsis:    Display tape driver name and version.                        */
/* Returns:     Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/

static int get_driver_info ()
{
  int rc;
  IBMtape_info_t IBMtape_info;

  if (!(rc = ioctl (dev_fd, IOC_DRIVER_INFO, &IBMtape_info))) {
    PRINTF ("IBMtape tape device driver information:\n");
    PRINTF ("Name: %s\n", IBMtape_info.name);
    PRINTF ("Version: %s\n", IBMtape_info.version);
    PRINTF ("Short version string: %s\n", IBMtape_info.sver);
    PRINTF ("Operating System: %s\n", IBMtape_info.os);
    PRINTF ("System Architecture: %s\n", IBMtape_info.arch);
  }
  else {
    PERROR ("Failure obtaining the information of IBMtape");
    PRINTF ("\n");
    scsi_request_sense ();
  }

  return (rc);
}
/* get_driver_info */


/*****************************************************************************/
/*                         BASIC SCSI FUNCTIONS                              */
/*****************************************************************************/


/*---------------------------------------------------------------------------*/
/*        Name: scsi_test_unit_ready                                         */
/*    Synopsis: Verify device ready.                                         */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_test_unit_ready ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, IOC_TEST_UNIT_READY, 0))) {
      PRINTF ("IOC_TEST_UNIT_READY succeeded.\n");
   }
   else {
      PERROR ("IOC_TEST_UNIT_READY ioctl failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: scsi_inquiry                                                 */
/*    Synopsis: Display device standard inquiry data.                        */
/*              See also scsi_inquiry_page().                                */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_inquiry ()
{
   int rc;
   inquiry_data_t inquiry_data;

   memset ((char *)&inquiry_data, (char)0, sizeof (inquiry_data_t));
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data))) {
      PRINTF ("IOC_INQUIRY succeeded.\n");
      PRINTF ("\nInquiry Data (Standard Data):\n");
      if (verbose ) {
         dump_bytes ((char *)&inquiry_data, sizeof (inquiry_data_t));
      }
   }
   else {
      PERROR ("IOC_INQUIRY failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}
/* scsi_inquiry */


/*---------------------------------------------------------------------------*/
/*        Name: scsi_inquiry_page                                            */
/*    Synopsis: Display device inquiry page data.                            */
/*              See also scsi_inquiry().                                     */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_inquiry_page (int page)
{
   int rc;
   inquiry_page_t inquiry_page;

   memset ((char *)&inquiry_page, (char)0, sizeof (inquiry_page_t));
   inquiry_page.page_code = (uchar) page;
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY_PAGE, &inquiry_page))) {
      PRINTF ("\nInquiry Data (Page 0x%02x):\n", page);
      dump_bytes ((char *)&inquiry_page.data, inquiry_page.data[3]+4);
   }
   else {
      PRINTF ("IOC_INQUIRY_PAGE for page 0x%X failed.\n", page);
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}
/* scsi_inquiry_page */


/*---------------------------------------------------------------------------*/
/*        Name: scsi_request_sense                                           */
/*    Synopsis: Display device sense data.                                   */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_request_sense ()
{
   int rc;
   sense_data_t sense_data;


   if (!(rc = ioctl (dev_fd, IOC_REQUEST_SENSE, &sense_data))) {
      PRINTF ("IOC_REQUEST_SENSE succeeded.\n");
      PRINTF ("\nThe request sense data is:\n");
      DUMP_BYTES ((char *)&sense_data, sizeof (sense_data_t));
   }
   else {
      PERROR ("IOC_REQUEST_SENSE failed");
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: scsi_reserve                                                 */
/*    Synopsis: Reserve device.                                              */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_reserve ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, IOC_RESERVE, 0))) {
      PRINTF ("IOC_RESERVE succeeded.\n");
   }
   else {
      PERROR ("IOC_RESERVE failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: scsi_release                                                 */
/*    Synopsis: Release device.                                              */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_release ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, IOC_RELEASE, 0))) {
      PRINTF ("IOC_RELEASE succeeded.\n");
   }
   else {
      PERROR ("IOC_RELEASE failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: scsi_enable_path                                             */
/*    Synopsis: Enable and disable failover path.                            */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_enable_path ( )
{
   int i,rc;
   int path;
   int enable;
   char buf[10];
   check_path_t check_path;
   path_enable_t path_enable;

   if (!(rc = ioctl (dev_fd, IOC_CHECK_PATHS, &check_path))) {
      printf ("IOC_CHECK_PATHS succeeded.\n");
      printf ("Path enable information for the device :\n");
      for (i=0; i<check_path.number_paths; i++) {
          path = check_path.path_enable[i].path;
          enable = check_path.path_enable[i].enable;
          if (path == 0xFF) {
             if (enable) printf ("  Enable, but no failover path\n");
             else printf (" Disable, but no failover path\n");
             return (1);
          }
          else if (path == 0) {
             printf ("  Path %d : Primary", path);
          }
          else {
             printf ("  Path %d : Alt_path_%d", path, path);
          }

          if (enable) printf ("  Enable\n");
          else printf ("  Disable\n");
      }

      bzero (&path_enable, sizeof(path_enable_t));
      printf ("Enter the path to enable or disable: ");
      gets (buf);
      if (*buf != '\0') {
         path_enable.path = atoi (buf);
         if ((path_enable.path<0) || (path_enable.path>=check_path.number_paths)) {
            printf ("\nInvalid selection.\n");
            return (FALSE);
         }
      }
      else  printf ("\n");

      printf ("Enter selection (0=disable, 1=enable): ");
      gets (buf);
      if (*buf != '\0') {
         path_enable.enable = atoi (buf);
         if ((path_enable.enable < 0) || (path_enable.enable > 1)) {
            printf ("\nInvalid selection.\n");
            return (FALSE);
         }
      }
      else  printf ("\n");

      if (*buf != '\0') {
      if (path_enable.enable) {
         if (rc = ioctl (dev_fd, IOC_ENABLE_PATH, &path_enable)) {
            perror ("IOC_ENABLE_PATH failed");
            printf ("\n");
            return (rc);
         }
      }
      else {
         if (rc = ioctl (dev_fd, IOC_DISABLE_PATH, &path_enable)) {
            perror ("IOC_DISABLE_PATH failed");
            printf ("\n");
            return (rc);
         }
      }
      }

      if (!(rc = ioctl (dev_fd, IOC_CHECK_PATHS, &check_path))) {
         printf ("\nPath enable information for the device :\n");
         for (i=0; i<check_path.number_paths; i++) {
             path = check_path.path_enable[i].path;
             enable = check_path.path_enable[i].enable;
             if (path == 0xFF) {
                if (enable) printf ("  Enable, but no failover path\n");
                else printf (" Disable, but no failover path\n");
                return (1);
             }
             else if (path == 0) {
                printf ("  Path %d : Primary", path);
             }
             else {
                printf ("  Path %d : Alt_path_%d", path, path);
             }

             if (enable) printf ("  Enable\n");
             else printf ("  Disable\n");
         }
      }
   }
   else {
      perror ("IOC_CHECK_PATHS failed");
      printf ("\n");
   }

   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: scsi_query_device_paths                                      */
/*    Synopsis: Query failover paths for the device .                        */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int scsi_query_device_paths ( )
{
   int i, rc;
   char buf[10];
   device_paths_t device_paths;

   if (rc = ioctl (dev_fd, IOC_DEVICE_PATHS, &device_paths)) {
      perror ("IOC_DEVICE_PATHS failed");
      printf ("\n");
      return (rc);
   }

   printf ("Path list in the device: \n");
   for (i=0; i<device_paths.number_paths; i++) {
       printf ("  Path %d", i);
   }
   printf ("\nCurrent active path: Path %d\n\n", device_paths.cur_path);

   printf ("\nEnter path number or <enter> for all of the paths: ");
   gets (buf);
   if ( buf[0] == '\0' ) {
      for (i=0; i<device_paths.number_paths; i++) {
        show_path (&device_paths.device_path[i]);
        printf ("\n--- more ---");
        if (interactive)  getchar ();
      }
   }
   else {
      i = atoi (buf);
      if ((i>=device_paths.number_paths) || (i<0)) {
          printf ("\nInvalid Path Number selection.\n");
          return (FALSE);
      }
      show_path (&device_paths.device_path[i]);
   }

   return (0);
}


/*****************************************************************************/
/*                         MEDIUM CHANGER FUNCTIONS                          */
/*****************************************************************************/


/*---------------------------------------------------------------------------*/
/*        Name: smc_move_medium                                              */
/*    Synopsis: Move cartridge from source to destination.                   */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_move_medium ( int src, int dst )
{
   int rc;
   move_medium_t move_medium;

   move_medium.robot = 0;
   move_medium.invert = NO_FLIP;
   move_medium.source = (short)src;
   move_medium.destination = (short)dst;
   if (!(rc = ioctl (dev_fd, SMCIOC_MOVE_MEDIUM, &move_medium))) {
      PRINTF ("SMCIOC_MOVE_MEDIUM succeeded.\n");
   }
   else {
      PERROR ("SMCIOC_MOVE_MEDIUM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: smc_exchange_medium                                          */
/*    Synopsis: Move cartridge from source to destination 1 and move tape    */
/*        which provious occupied in destination 1 to destination 2.         */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_exchange_medium ( int src, int dst1, int dst2 )
{
   int rc;
   exchange_medium_t exchange_medium;

   exchange_medium.robot = 0;
   exchange_medium.invert1 = NO_FLIP;
   exchange_medium.invert2 = NO_FLIP;
   exchange_medium.source = (short)src;
   exchange_medium.destination1 = (short)dst1;
   exchange_medium.destination2 = (short)dst2;

   if (!(rc = ioctl (dev_fd, SMCIOC_EXCHANGE_MEDIUM, &exchange_medium))) {
      PRINTF ("SMCIOC_EXCHANGE_MEDIUM succeeded.\n");
   }
   else {
      PERROR ("SMCIOC_EXCHANGE_MEDIUM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: smc_pos_to_elem                                              */
/*    Synopsis: Move robot to element.                                       */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_pos_to_elem ( int dst )
{
   int rc;
   pos_to_elem_t pos_to_elem;

   pos_to_elem.robot = 0;
   pos_to_elem.invert = NO_FLIP;
   pos_to_elem.destination = (short)dst;
   if (!(rc = ioctl (dev_fd, SMCIOC_POS_TO_ELEM, &pos_to_elem))) {
      PRINTF ("SMCIOC_POS_TO_ELEM succeeded.\n");
   }
   else {
      PERROR ("SMCIOC_POS_TO_ELEM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: smc_element_info                                             */
/*    Synopsis: Display device element information.                          */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_element_info ()
{
   int rc;

   PRINTF("Getting element information...\n");
   if (!(rc = ioctl (dev_fd, SMCIOC_ELEMENT_INFO, &element_info))) {
      PRINTF ("\nThe element information data is:\n");
      /* dump_bytes ((char *)&element_info, sizeof (element_info_t)); */

      PRINTF("\n");
      PRINTF("  Number of Robots .............. %d\n",element_info.robot_count);
      if (element_info.robot_count)
      PRINTF("  First Robot Address ........... %d\n",element_info.robot_address);

      PRINTF("  Number of Storage Cells ....... %d\n",element_info.cell_count);
      if (element_info.cell_count)
      PRINTF("  First Cell Address ............ %d\n",element_info.cell_address);

      PRINTF("  Number of Import/Exports ...... %d\n",element_info.port_count);
      if (element_info.port_count)
      PRINTF("  First Import/Export Address ... %d\n",element_info.port_address);

      PRINTF("  Number of Drives .............. %d\n",element_info.drive_count);
      if (element_info.drive_count)
      PRINTF("  First Drive Address ........... %d\n",element_info.drive_address);
      PRINTF("\n");
   }
   else {
      PERROR ("SMCIOC_ELEMENT_INFO failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: smc_inventory                                                */
/*    Synopsis: Display device element status.                               */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_inventory ()
{
   int rc;
   ushort i;
   inventory_t inventory;

   PRINTF("Getting inventory...\n");
   /* get element information first */
   if (rc = ioctl (dev_fd, SMCIOC_ELEMENT_INFO, &element_info)) {
      PERROR ("SMCIOC_INVENTORY failed");
      PRINTF ("\n");
      scsi_request_sense ();
      return (rc);
   }
   inventory.robot_status = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.robot_count);
   inventory.cell_status  = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.cell_count );
   inventory.port_status  = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.port_count );
   inventory.drive_status = (element_status_t *)malloc
      (sizeof (element_status_t) * element_info.drive_count);
   if (!inventory.robot_status || !inventory.cell_status ||
      !inventory.port_status  || !inventory.drive_status) {
      PRINTF ("SMCIOC_INVENTORY failed: Memory allocation failure.\n");
      return (ENOMEM);
   }
   if (!(rc = ioctl (dev_fd, SMCIOC_INVENTORY, &inventory))) {
      for (i = 0; i < element_info.robot_count; i++) {
      /* dump_bytes ((char *)(&inventory.robot_status[i]),
            sizeof (element_status_t)); */
      fmt_elem (ELEMT_ROBOT, &inventory.robot_status[i], "Robot");
      PRINTF ("\n--- more ---");
      if (interactive)  getchar ();
      }
      for (i = 0; i < element_info.cell_count; i++) {
      /* dump_bytes ((char *)(&inventory.cell_status[i]),
            sizeof (element_status_t)); */
      fmt_elem (ELEMT_CELL, &inventory.cell_status[i], "Storage Cell");
      PRINTF ("\n--- more ---");
      if (interactive)  getchar ();
      }
      for (i = 0; i < element_info.port_count; i++) {
      /* dump_bytes ((char *)(&inventory.port_status[i]),
            sizeof (element_status_t)); */
      fmt_elem (ELEMT_IMPEXP, &inventory.port_status[i], "Import/Export");
      PRINTF ("\n--- more ---");
      if (interactive)  getchar ();
      }
      for (i = 0; i < element_info.drive_count; i++) {
      /* dump_bytes ((char *)(&inventory.drive_status[i]),
            sizeof (element_status_t)); */
      fmt_elem (ELEMT_DRIVE, &inventory.drive_status[i], "Tape Drive");
      PRINTF ("\n--- more ---");
      if (interactive)  getchar ();
      }
   }
   else {
      PERROR ("SMCIOC_INVENTORY failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/* Name:        smc_read_element_devids                                      */
/*---------------------------------------------------------------------------*/
static int smc_read_element_devids(void)
   {
   int rc;
   int i,j;
   element_devids_t *elem_devid, *elemp;
   read_element_devids_t devids;
   element_info_t element_info;

   if (ioctl(dev_fd, SMCIOC_ELEMENT_INFO, &element_info))
      return errno;

   if (element_info.drive_count)
      {
      elem_devid = malloc(element_info.drive_count * sizeof(element_devids_t));
      if (elem_devid == NULL)
      {
      errno = ENOMEM;
      return errno;
      }
      bzero((caddr_t)elem_devid,element_info.drive_count * sizeof(element_devids_t));

      devids.drive_devid = elem_devid;
      devids.element_address = element_info.drive_address;
      devids.number_elements = element_info.drive_count;

      printf("Reading element device ids...\n");

      if (rc = ioctl (dev_fd, SMCIOC_READ_ELEMENT_DEVIDS, &devids))
      {
      free(elem_devid);
      PERROR ("SMCIOC_READ_ELEMENT_DEVIDS failed");
      PRINTF ("\n");
      scsi_request_sense ();

      return rc;
      }

      j=0;
      elemp = elem_devid;

      for (i = 0; i < element_info.drive_count; i++, elemp++)
      {
      /* don't overflow screen if menu mode */
      if (interactive && j == 2)
         {
         j=0;
         printf ("\nHit <enter> to continue...");
         getchar();
         }

      j++;
      printf("\nDrive Address %d\n",elemp->address);
      if (elemp->except)
         if (device.t3590 || device.ult3584)
           printf("  Drive State .................... See ASC/ASCQ\n");
         else
           printf("  Drive State .................... Abnormal\n");
      else
         printf("  Drive State .................... Normal\n");
      if (elemp->asc == 0x81 && elemp->ascq ==0x00)
         printf("  ASC/ASCQ ....................... %02X%02X (Drive Present)\n",
               elemp->asc,elemp->ascq);
      else if (elemp->asc == 0x82 && elemp->ascq ==0x00)
         printf("  ASC/ASCQ ....................... %02X%02X (Drive Not Present)\n",
               elemp->asc,elemp->ascq);
      else
         printf("  ASC/ASCQ ....................... %02X%02X\n",
               elemp->asc,elemp->ascq);
      if (elemp->full)
         printf("  Media Present .................. Yes\n");
      else
         printf("  Media Present .................. No\n");
      if (elemp->access)
         printf("  Robot Access Allowed ........... Yes\n");
      else
         printf("  Robot Access Allowed ........... No\n");
      if (elemp->svalid)
         printf("  Source Element Address ......... %d\n",elemp->source);
      else
         printf("  Source Element Address Valid ... No\n");
      if (elemp->invert)
         printf("  Media Inverted ................. Yes\n");
      else
         printf("  Media Inverted ................. No\n");
      if (elemp->notbus)
         printf("  Same Bus as Medium Changer ..... No\n");
      else
         printf("  Same Bus as Medium Changer ..... Yes\n");
      if (elemp->idvalid)
         printf("  SCSI Bus Address ............... %d\n",elemp->scsi);
      else
         printf("  SCSI Bus Address Vaild ......... No\n");
      if (elemp->luvalid)
         printf("  Logical Unit Number ............ %d\n",elemp->lun);
      else
         printf("  Logical Unit Number Valid ...... No\n");

      printf("  Identifier Type ................ %d\n", elemp->idtype);
      printf("  Device ID Info\n");
      printf("      Vendor ..................... %0.8s\n", elemp->vendorid);
      printf("      Model ...................... %0.16s\n", elemp->devtype);
      if ((!strncmp ((char *)elemp->devtype, "0359", 4)) ||
          (!strncmp ((char *)elemp->devtype, "0349", 4)) ||
          (!strncmp ((char *)elemp->devtype, "0357", 4)))
          printf("      Serial Number .............. %0.12s\n", elemp->serialnum);
      else
          printf("      Serial Number .............. %0.10s\n", elemp->serialnum);
      }
      }
   else
      {
      printf("\nNo drives found in element information\n");
      if (interactive)
      {
      printf ("\nHit <enter> to continue...");
      getchar();
      }
      }

   free(elem_devid);
   return errno;
   }

/*---------------------------------------------------------------------------*/
/*        Name: smc_audit                                                    */
/*    Synopsis: Initialize device element status.                            */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_audit ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, SMCIOC_AUDIT, 0))) {
      PRINTF ("SMCIOC_AUDIT succeeded.\n");
   }
   else {
      PERROR ("SMCIOC_AUDIT failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: smc_audit_range                                              */
/*    Synopsis: Initialize device element status with range.                 */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_audit_range (int element_address, int number_elements)
{
   int rc;
   element_range_t element_range;

   element_range.element_address = (ushort)element_address;
   element_range.number_elements = (ushort)number_elements;

   if (!(rc = ioctl (dev_fd, SMCIOC_AUDIT_RANGE, &element_range))) {
      PRINTF ("SMCIOC_AUDIT_RANGE succeeded.\n");
   }
   else {
      PERROR ("SMCIOC_AUDIT_RANGE failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}

/*---------------------------------------------------------------------------*/
/*        Name: smc_lock_door                                                */
/*    Synopsis: Lock or unlock device door.                                  */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int smc_lock_door ( int lock )
{
   int rc;
   lock_door_t lock_door;

   lock_door = (uchar)lock;
   if (!(rc = ioctl (dev_fd, SMCIOC_LOCK_DOOR, &lock_door))) {
      PRINTF ("SMCIOC_LOCK_DOOR succeeded.\n");
   }
   else {
      PERROR ("SMCIOC_LOCK_DOOR failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*****************************************************************************/
/*                           TAPE DRIVE FUNCTIONS                            */
/*****************************************************************************/

/*---------------------------------------------------------------------------*/
/*        Name: st_read                                                      */
/*    Synopsis: Read data from device.                                       */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_read ( int blk_size, int blk_fact, int multiplier )
{
   int rc;
   int i;
   int count;
   long long total_bytes;
   struct timeval start;
   struct timeval stop;
   struct timezone zone;
   char *data;

   if (!(data = malloc (RWBUFSZ))) {
      PERROR ("Failure allocating read data buffer");
      return (errno);
   }
   total_bytes = (long long)0;
   gettimeofday (&start, &zone);
   for (i = 1; i <= multiplier; i++) {
      count = read (dev_fd, data, (uint)(blk_size * blk_fact));
      rc = errno;
      if (count < 0) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("The READ command failed (request %d, count = %d, errno = %d).",
            i, count, errno);
      break;
      }
      if (count != (blk_size * blk_fact)) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("Short READ (request %d, count = %d).", i, count);
      count = -1;
      break;
      }
      total_bytes = total_bytes + (long long)count;
      PRINTF ("*");
      if (!(i % 32)) {
      PRINTF ("  %lld\n", total_bytes);
      }
   }
   gettimeofday (&stop, &zone);
   if ((i % 32) || (count < 0)) {
      PRINTF ("\n");
   }
   PRINTF ("\n");
   if (count < 0) {
      scsi_request_sense ();
   }
   else {
      PRINTF ("Read %lld bytes \n", total_bytes );
   }
   free (data);
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_write                                                     */
/*    Synopsis: Write data to device.                                        */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_write ( int blk_size, int blk_fact, int multiplier, int seed, char *filename )
{
   int rc;
   int i;
   int fd;
   int count;
   int length;
   long long total_bytes;
   struct timeval start;
   struct timeval stop;
   struct timezone zone;
   char *data;
   char *ptrn;

   if (!(data = malloc (RWBUFSZ))) {
      PERROR ("Failure allocating write data buffer");
      return (errno);
   }
   if (filename[0] != '\0') {  /* generate data using pattern file */
      if ((fd = open (filename, O_RDWR)) < 0) {  /* open pattern file */
      PERROR ("Failure opening pattern file");
      rc = errno;
      free (data);
      return (rc);
      }
      if (!(ptrn = malloc (RWBUFSZ))) {
      PERROR ("Failure allocating pattern data buffer");
      rc = errno;
      close (fd);
      free (data);
      return (rc);
      }
      if ((length = read (fd, &ptrn, RWBUFSZ)) <= 0) {  /* read pattern */
      PERROR ("Failure reading pattern file");
      rc = errno;
      close (fd);
      free (ptrn);
      free (data);
      return (rc);
      }
      close (fd);
      /* build data buffer as multiple of pattern */
      for (i = 0; i < (blk_size * blk_fact); i++) {
      data[i] = ptrn[i % length];
      }
      free (ptrn);
   }
   else {  /* generate data using random pattern */
      while (!seed) {  /* 0 < seed < 32768 */
      gettimeofday (&start, &zone);
      seed = (32768 + start.tv_usec) % 32768;
      }
      srand (seed);
      for (i = 0; i < (int)(blk_size * blk_fact); i++) {
      data[i] = (char)rand ();
      }
      PRINTF ("Generated %d bytes of random data using seed %d.\n\n",
         (blk_size * blk_fact), seed);
   }
   total_bytes = (long long)0;
   gettimeofday (&start, &zone);
   for (i = 1; i <= multiplier; i++) {
      count = write (dev_fd, data, (uint)(blk_size * blk_fact));
      rc = errno;
      if (count < 0) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("The WRITE command failed (request %d, count = %d, errno = %d).",
            i, count, errno);
      break;
      }
      if (count != (blk_size * blk_fact)) {
      if (i > 1)  PRINTF ("\n\n");
      PRINTF ("Short WRITE (request %d, count = %d).", i, count);
      count = -1;
      break;
      }
      total_bytes = total_bytes + (long long)count;
      PRINTF ("*");
      if (!(i % 32)) {
      PRINTF ("  %lld\n", total_bytes);
      }
   }
   gettimeofday (&stop, &zone);
   if ((i % 32) || (count < 0)) {
      PRINTF ("\n");
   }
   PRINTF ("\n");
   if (count < 0) {
      scsi_request_sense ();
   }
   else {
      PRINTF ("Wrote %lld bytes \n", total_bytes);
   }
   free (data);
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_tape_op                                                   */
/*    Synopsis: Perform tape movement operations.                            */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_tape_op ( int mt_op, int mt_count )
{
   int rc;
   tape_op_t tape_op;
   tape_op.mt_op = (short)mt_op;
   tape_op.mt_count = mt_count;

   if (!(rc = ioctl (dev_fd, STIOC_TAPE_OP, &tape_op))) {
      PRINTF ("STIOC_TAPE_OP succeeded.\n");
      if (tape_op.mt_op == MTGRSZ) {
      PRINTF ("\nThe record size data is:\n");
      dump_bytes ((char *)&tape_op.mt_count, sizeof (long));
      }
   }
   else {
      PERROR ("STIOC_TAPE_OP failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_drivetype                                             */
/*    Synopsis: Drive type.                                                  */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_drivetype ()
{
  int rc;
  struct mtdrivetype mtdt;
  struct mtdrivetype_request mtreq;

  mtreq.size = sizeof(struct mtdrivetype);
  mtreq.mtdtp = &mtdt;
  if (!(rc = ioctl(dev_fd, MTIOCGETDRIVETYPE ,&mtreq))) {
    PRINTF ("MTIOCGETDRIVETYPE succeeded.\n");
    PRINTF ("\nThe drive type data is:\n");
    dump_bytes ((char *)&mtdt, sizeof (struct mtdrivetype));
  }
  else {
    PERROR ("MTIOCGETDRIVETYPE failed");
    PRINTF ("\n");
    scsi_request_sense ();
  }
  return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_device_status                                         */
/*    Synopsis: Display device status.                                       */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_device_status ()
{
   int rc;
   device_status_t device_status;

   if (!(rc = ioctl (dev_fd, STIOC_GET_DEVICE_STATUS, &device_status))) {
      PRINTF ("STIOC_GET_DEVICE_STATUS succeeded.\n");
      PRINTF ("\nThe device status data is:\n");
      dump_bytes ((char *)&device_status, sizeof (device_status_t));
   }
   else {
      PERROR ("STIOC_GET_DEVICE_STATUS failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_device_info                                           */
/*    Synopsis: Display device information.                                  */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_device_info ()
{
   int rc;
   device_info_t device_info;

   if (!(rc = ioctl (dev_fd, STIOC_GET_DEVICE_INFO, &device_info))) {
      PRINTF ("STIOC_GET_DEVICE_INFO succeeded.\n");
      PRINTF ("\nThe device information is:\n");
      dump_bytes ((char *)&device_info, sizeof (device_info_t));
   }
   else {
      PERROR ("STIOC_GET_DEVICE_INFO failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_media_info                                            */
/*    Synopsis: Display media information.                                   */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static int st_get_media_info ()
{
   int rc;
   media_info_t media_info;

   if (!(rc = ioctl (dev_fd, STIOC_GET_MEDIA_INFO, &media_info))) {
      PRINTF ("STIOC_GET_MEDIA_INFO succeeded.\n");
      PRINTF ("\nThe media information is:\n");
      dump_bytes ((char *)&media_info, sizeof (media_info_t));
   }
   else {
      PERROR ("STIOC_GET_MEDIA_INFO failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_position                                              */
/*    Synopsis: Display tape position information.                           */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static int st_get_position ( int type )
{
   int rc;
   position_data_t position_data;

   position_data.block_type = (uchar)type;
   if (!(rc = ioctl (dev_fd, STIOC_GET_POSITION, &position_data))) {
      PRINTF ("STIOC_GET_POSITION succeeded.\n");
      PRINTF ("\nTape position data:\n");
      /* dump_bytes ((char *)&position_data, sizeof (position_data_t)); */
      PRINTF ( "%20s: %10s \n" "%20s: %10d \n" "%20s: %10d \n" "%20s: %10d \n"
         "%20s: %10d \n" "%20s: %10d \n" "%20s: %10d \n" "%20s: %10d \n" ,
         "Block Type" , (position_data.block_type == 1 ? "Logical" : "Physical") ,
         "BOP"              , position_data.bot  ,
         "EOP"              , position_data.eot  ,
         "Partition Num"    , position_data.partition  ,
         "Current Position" , position_data.position  ,
         "Last Block Pos"   , position_data.last_block  ,
         "Block Count"      , position_data.block_count  ,
         "Byte Count"       , position_data.byte_count );
   }
   else {
      PERROR ("STIOC_GET_POSITION failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_set_position                                              */
/*    Synopsis: Set tape position.                                           */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
static int st_set_position ( int type, int value )
{
   int rc;
   position_data_t position_data;

   position_data.block_type = (uchar)type;
   position_data.position = value;
   if (!(rc = ioctl (dev_fd, STIOC_SET_POSITION, &position_data))) {
      PRINTF ("STIOC_SET_POSITION succeeded.\n");
   }
   else {
      PERROR ("STIOC_SET_POSITION failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_parm                                                  */
/*    Synopsis: Display device specific parameter.                           */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_parm ( int type )
{
   int rc;
   parm_data_t parm_data;

   parm_data.type = (uchar)type;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      PRINTF ("STIOC_GET_PARM succeeded.\n");
      if (parm_data.type == WRITEPROTECT) {
         if (!parm_data.value)
            { PRINTF ("Tape is write-enabled.\n"); }
         if (parm_data.value & WORM_PROTECT)
            { PRINTF ("Tape is WORM-protected.\n"); }
         if (parm_data.value & PERS_PROTECT)
            { PRINTF ("Tape is software write-protected across tape mounts.\n"); }
         if (parm_data.value & ASSC_PROTECT)
            { PRINTF ("Tape is software write-protected for this tape mount only.\n"); }
         if (parm_data.value & PHYS_PROTECT)
            { PRINTF ("Tape is physically write-protected.\n"); }
      }
      else {
         PRINTF ("\nThe parameter data is:\n");
         dump_bytes ((char *)&parm_data.value, sizeof (int));
      }
   }
   else {
      PERROR ("STIOC_GET_PARM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_set_parm                                                  */
/*    Synopsis: Set device specific parameter.                               */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_set_parm ( int type, int value )
{
   int rc;
   parm_data_t parm_data;

   parm_data.type = (uchar)type;
   parm_data.value = value;
   if (!(rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data))) {
      PRINTF ("STIOC_SET_PARM succeeded.\n");
   }
   else {
      PERROR ("STIOC_SET_PARM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_density                                               */
/*    Synopsis: Get tape density.                                            */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_density()
{
   int rc;
   density_data_t density_data;

   if (!(rc = ioctl (dev_fd, STIOC_GET_DENSITY, &density_data))) {
      PRINTF ("STIOC_GET_DENSITY succeeded.\n");
      PRINTF ("\nTape density data:\n");
      if (device.t3592e05) {
         PRINTF ("density_code ------ 0x%X\n", density_data.density_code);
         PRINTF ("default_density --- 0x%X\n", density_data.default_density);
         PRINTF ("pending_density --- 0x%X\n", density_data.pending_density);
      }
      else {
         PRINTF ("density_code ------ 0x%X\n", density_data.density_code);
      }
   }
   else {
      PERROR ("STIOC_GET_DENSITY failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_sync_buffer                                               */
/*    Synopsis: Flush device buffer.                                         */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_sync_buffer ()
{
   int rc;

   if (!(rc = ioctl (dev_fd, STIOC_SYNC_BUFFER, 0))) {
      PRINTF ("STIOC_SYNC_BUFFER succeeded.\n");
   }
   else {
      PERROR ("STIOC_SYNC_BUFFER failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: st_display_msg                                               */
/*    Synopsis: Display message on device operator panel.                    */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_display_msg ( int type, char *msg_0, char *msg_1 )
{
   int rc;
   msg_data_t msg_data;

   char msg_type[] = { DISPMSG0, DISPMSG1, FLASHMSG0, FLASHMSG1, ALTERNATE };
   msg_data.function = msg_type[type-1];
   memcpy (msg_data.msg_0, msg_0, 8);
   memcpy (msg_data.msg_1, msg_1, 8);
   if (!(rc = ioctl (dev_fd, STIOC_DISPLAY_MSG, &msg_data))) {
      PRINTF ("STIOC_DISPLAY_MSG succeeded.\n");
   }
   else {
      PERROR ("STIOC_DISPLAY_MSG failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: get_pid                                                      */
/*    Synopsis: Get product ID into null-terminated                          */
/*              string, then convert all trailing blanks to nulls.           */
/*---------------------------------------------------------------------------*/
static int get_pid(char *pid )
{
   int rc = 0;
   int i;

   rc = ioctl (dev_fd, IOC_INQUIRY, &inquiry_data);
   if (rc != 0) {
      (void) fprintf (stderr, "INQUIRY rc %d (%s)\n", errno, strerror(errno));
      scsi_request_sense ();
      return errno;
   }
   tape_drv = inquiry_data.type & INQ_DRV_TYPE;
   medium_chgr = inquiry_data.type & INQ_SMC_TYPE;
   attached_medium_chgr = inquiry_data.mchngr;
   (void) memset (pid, 0, PIDLEN + 1);
   *((int *)&device) = 0;
   bcopy ((const char *)inquiry_data.pid, pid, PIDLEN);

   /* Start at last char before final \0, look for blanks, change them to \0s. */
   for ( i = PIDLEN - 1; i >= 0; i-- ) {
      if (pid[i] == ' ') pid[i] = '\0';
      else break;
   }
   if (!(strncmp (pid, "ULT3580-TD", 10))) device.ultrium = TRUE;
   if (!(strncmp (pid, "ULTRIUM-TD", 10))) device.ultrium = TRUE;
   if (!(strncmp (pid, "ULT3580-HH", 10))) device.ultrium = TRUE;
   if (!(strncmp (pid, "ULTRIUM-HH", 10))) device.ultrium = TRUE;
   if (!strcmp (pid, "ULT3580-TD1"))  device.lto1    = TRUE;
   if (!strcmp (pid, "ULTRIUM-TD1"))  device.lto1    = TRUE;
   if (!strcmp (pid, "ULT3580-TD2"))  device.lto2    = TRUE;
   if (!strcmp (pid, "ULTRIUM-TD2"))  device.lto2    = TRUE;
   if (!strcmp (pid, "ULT3580-TD3"))  device.lto3    = TRUE;
   if (!strcmp (pid, "ULTRIUM-TD3"))  device.lto3    = TRUE;
   if (!strcmp (pid, "ULT3580-HH3"))  device.lto3    = TRUE;
   if (!strcmp (pid, "ULTRIUM-HH3"))  device.lto3    = TRUE;
   if (!strcmp (pid, "ULT3580-TD4"))  device.lto4    = TRUE;
   if (!strcmp (pid, "ULTRIUM-TD4"))  device.lto4    = TRUE;
   if (!strcmp (pid, "ULT3580-HH4"))  device.lto4    = TRUE;
   if (!strcmp (pid, "ULTRIUM-HH4"))  device.lto4    = TRUE;
   if (!strcmp (pid, "ULT3581-TA"))   device.ult3581 = TRUE;
   if (!strcmp (pid, "ULT3582-TL"))   device.ult3582 = TRUE;
   if (!strcmp (pid, "ULT3583-TL"))   device.ult3583 = TRUE;
   if (!strcmp (pid, "3576-MTL"))     device.ult3576 = TRUE;
   if (!strcmp (pid, "3573-TL"))      device.ult3573 = TRUE;
   if (!(strncmp (pid, "03584", 5)))  device.ult3584 = TRUE;
   if (!strcmp (pid, "3577-TL"))      device.lib3577 = TRUE;
   if (!(strncmp (pid, "03490", 5)))  device.t3490   = TRUE;
   if (!(strncmp (pid, "03570", 5)))  device.t3570   = TRUE;
   if (!(strncmp (pid, "03590", 5)))  device.t3590   = TRUE;
   if (!(strncmp (pid, "03592", 5)))  device.t3592   = TRUE;
   if (!(strncmp (pid, "03592E05", 8)))  device.t3592e05 = TRUE;
   if (!(strncmp (pid, "ULT3581-TA2",11))) device.ult3581_ta2  = TRUE;

   return (rc);
}
/* get_pid */


/*---------------------------------------------------------------------------*/
/*        Name: st_report_density_support                                    */
/*    Synopsis: Report the supported densities for the device.               */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_report_density_support ()
{
   int rc;
   int i;
   rpt_dens_sup_t density;

   int bits_per_mm = 0;
   int media_width = 0;
   int tracks = 0;
   int capacity = 0;

   PRINTF ("Issuing Report Density Support for ALL supported media...\n");

   density.media = ALL_MEDIA_DENSITY;
   density.number_reports = 0;

   if (!(rc = ioctl (dev_fd, STIOC_REPORT_DENSITY_SUPPORT, &density))) {
      PRINTF ("STIOC_REPORT_DENSITY_SUPPORT succeeded.\n");
      PRINTF ("Total number of densities reported: %d\n",density.number_reports);
   }
   else {
      PERROR ("STIOC_REPORT_DENSITY_SUPPORT failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }

   for (i = 0; i < density.number_reports; i++)
   {
      bits_per_mm  = (int)density.reports[i].bits_per_mm[0] << 16;
      bits_per_mm |= (int)density.reports[i].bits_per_mm[1] << 8;
      bits_per_mm |= (int)density.reports[i].bits_per_mm[2];

      media_width  = density.reports[i].media_width[0] << 8;
      media_width |= density.reports[i].media_width[1];

      tracks  = density.reports[i].tracks[0] << 8;
      tracks |= density.reports[i].tracks[1];

      capacity  = density.reports[i].capacity[0] << 24;
      capacity |= density.reports[i].capacity[1] << 16;
      capacity |= density.reports[i].capacity[2] << 8;
      capacity |= density.reports[i].capacity[3];

      PRINTF ("\n");
      PRINTF ("  Density Name................. %0.8s\n",density.reports[i].density_name);
      PRINTF ("  Assigning Organization....... %0.8s\n",density.reports[i].assigning_org);
      PRINTF ("  Description.................. %0.20s\n",density.reports[i].description);
      PRINTF ("  Primary Density Code......... %02X\n",density.reports[i].primary_density_code);
      PRINTF ("  Secondary Density Code....... %02X\n",density.reports[i].secondary_density_code);

      if (density.reports[i].wrtok)
         PRINTF ("  Write OK..................... Yes\n");
      else
         PRINTF ("  Write OK..................... No\n");

      if (density.reports[i].dup)
         PRINTF ("  Duplicate.................... Yes\n");
      else
         PRINTF ("  Duplicate.................... No\n");

      if (density.reports[i].deflt)
         PRINTF ("  Default...................... Yes\n");
      else
         PRINTF ("  Default...................... No\n");

      PRINTF ("  Bits per MM.................. %d\n",bits_per_mm);
      PRINTF ("  Media Width.................. %d\n",media_width);
      PRINTF ("  Tracks....................... %d\n",tracks);
      PRINTF ("  Capacity (megabytes)......... %d\n",capacity);


      if (interactive) {
         PRINTF ("\nHit <enter> to continue...");
         getchar ();
      }

   } /* end for */

   PRINTF ("\nIssuing Report Density Support for CURRENT media...\n");

   density.media = CURRENT_MEDIA_DENSITY;
   density.number_reports = 0;

   if (!(rc = ioctl (dev_fd, STIOC_REPORT_DENSITY_SUPPORT, &density))) {
      PRINTF ("STIOC_REPORT_DENSITY_SUPPORT succeeded.\n");
      PRINTF ("Total number of densities reported: %d\n",density.number_reports);
   }
   else {
      PERROR ("STIOC_REPORT_DENSITY_SUPPORT failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }

   for (i = 0; i < density.number_reports; i++)
   {

      bits_per_mm  = density.reports[i].bits_per_mm[0] << 16;
      bits_per_mm |= density.reports[i].bits_per_mm[1] << 8;
      bits_per_mm |= density.reports[i].bits_per_mm[2];

      media_width  = density.reports[i].media_width[0] << 8;
      media_width |= density.reports[i].media_width[1];

      tracks  = density.reports[i].tracks[0] << 8;
      tracks |= density.reports[i].tracks[1];

      capacity  = density.reports[i].capacity[0] << 24;
      capacity |= density.reports[i].capacity[1] << 16;
      capacity |= density.reports[i].capacity[2] << 8;
      capacity |= density.reports[i].capacity[3];

      max_capacity = capacity;

      PRINTF ("\n");
      PRINTF ("  Density Name................. %0.8s\n",density.reports[i].density_name);
      PRINTF ("  Assigning Organization....... %0.8s\n",density.reports[i].assigning_org);
      PRINTF ("  Description.................. %0.20s\n",density.reports[i].description);
      PRINTF ("  Primary Density Code......... %02X\n",density.reports[i].primary_density_code);
      PRINTF ("  Secondary Density Code....... %02X\n",density.reports[i].secondary_density_code);

      if (density.reports[i].wrtok)
         PRINTF ("  Write OK..................... Yes\n");
      else
         PRINTF ("  Write OK..................... No\n");

      if (density.reports[i].dup)
         PRINTF ("  Duplicate.................... Yes\n");
      else
         PRINTF ("  Duplicate.................... No\n");

      if (density.reports[i].deflt)
         PRINTF ("  Default...................... Yes\n");
      else
         PRINTF ("  Default...................... No\n");

      PRINTF ("  Bits per MM.................. %d\n",bits_per_mm);
      PRINTF ("  Media Width.................. %d\n",media_width);
      PRINTF ("  Tracks....................... %d\n",tracks);
      PRINTF ("  Capacity (megabytes)......... %d\n",capacity);

      if (interactive) {
         PRINTF ("\nHit <enter> to continue...");
         getchar ();
      }
   }


   return (rc);
}
/* st_report_density_support */

/*---------------------------------------------------------------------------*/
/*        Name: st_read_block_limits                                         */
/*    Synopsis: read block limits support for the drive.                     */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_read_block_limits ()
{
   int rc;
   int max_blk_lim = 0;
   int min_blk_lim = 0;
   rd_blklim_t rd_blklim;

   PRINTF ("Issuing Read Block Limits...\n");

   if (!(rc = ioctl (dev_fd, STIOC_READ_BLKLIM, &rd_blklim))) {
      PRINTF ("STIOC_READ_BLKLIM succeeded.\n");

      max_blk_lim  = rd_blklim.max_blk_lim[0] << 16;
      max_blk_lim |= rd_blklim.max_blk_lim[1] << 8;
      max_blk_lim |= rd_blklim.max_blk_lim[2];

      min_blk_lim  = rd_blklim.min_blk_lim[0] << 8;
      min_blk_lim |= rd_blklim.min_blk_lim[1];

      PRINTF ("Maximun Block Length Limit: %d\n", max_blk_lim);
      PRINTF ("Minimun Block Length Limit: %d\n", min_blk_lim);
   }
   else {
      PERROR ("STIOC_READ_BLKLIM failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }

   return (rc);
}
/* st_read_block_limits */

/*---------------------------------------------------------------------------*/
/*        Name: fmt_elem                                                     */
/*    Synopsis: Display inventory info.                                      */
/*     Returns: none.                                                        */
/*---------------------------------------------------------------------------*/
static void fmt_elem (int elem_type,  element_status_t *elemp, char *str_elem )
{

   printf("\n%s Address %d\n", str_elem, elemp->address);

   if (elemp->except)
      printf("  %s State ........... Abnormal\n", str_elem);
      else
      printf("  %s State ........... Normal\n", str_elem);

   printf("  ASC/ASCQ ....................... %02X%02X\n",
         elemp->asc,elemp->ascq);

   if (elemp->full)
      printf("  Media Present .................. Yes\n");
      else
      printf("  Media Present .................. No\n");

   if (elem_type == ELEMT_IMPEXP) {
      if (elemp->inenab)
         printf("  Import Enabled ................. Yes\n");
         else
         printf("  Import Enabled ................. No\n");
      if (elemp->exenab)
         printf("  Export Enabled ................. Yes\n");
         else
         printf("  Export Enabled ................. No\n");
   }

   if (elem_type != ELEMT_ROBOT) {
      if (elemp->access)
         printf("  Robot Access Allowed ........... Yes\n");
         else
         printf("  Robot Access Allowed ........... No\n");
   }

   if (elemp->svalid)
      printf("  Source Element Address ......... %d\n",elemp->source);
      else
      printf("  Source Element Address Valid ... No\n");

   if (elemp->invert)
      printf("  Media Inverted ................. Yes\n");
      else
      printf("  Media Inverted ................. No\n");

   if (elem_type == ELEMT_DRIVE) {
      if (elemp->notbus)
         printf("  Same Bus as Medium Changer ..... No\n");
         else
         printf("  Same Bus as Medium Changer ..... Yes\n");
      if (elemp->idvalid)
         printf("  SCSI Bus Address ............... %d\n",elemp->scsi);
         else
         printf("  SCSI Bus Address Vaild ......... No\n");
      if (elemp->luvalid)
         printf("  Logical Unit Number ............ %d\n",elemp->lun);
         else
         printf("  Logical Unit Number Valid ...... No\n");
   }

   printf("  Volume Tag ..................... %0.36s\n", elemp->volume);

   return;
}


/*---------------------------------------------------------------------------*/
/*        Name: show_path                                                    */
/*    Synopsis: Display Path info.                                           */
/*     Returns: none.                                                        */
/*---------------------------------------------------------------------------*/
static void show_path (device_path_t *path)
{
   printf ("\n Instance Number .......... %d\n", path->instance);
   printf (" SCSI Target .............. %d\n", path->tgt);
   printf (" SCSI LUN ................. %d\n", path->lun);
   printf (" Serial Number ............ %-s\n", path->dev_ser);
   printf (" Ucode Level .............. %-s\n", path->ucode_level);

   if (path->wwnn)
      printf (" Path World Wide NN ....... %16llX\n", path->wwnn);
   else
      printf (" Path World Wide NN ....... N/A\n");

   if (path->wwpn)
      printf (" Path World Wide PN ....... %16llX\n", path->wwpn);
   else
      printf (" Path World Wide PN ....... N/A\n");

   if (path->path_type == 0xFF)
      printf (" Path Type ................ N/A\n");
   else {
      if (path->path_type == 0)
         printf (" Path Type ................ Primary\n");
      else
         printf (" Path Type ................ Alt_path_%d\n", path->path_type);
   }

   if (path->enable)
      printf (" Path ..................... Enable\n");
   else
      printf (" Path ..................... Disable\n");

   if (device.t3592e05 || device.lto4) {
      if (path->encryption_capable)
         printf (" Encryption Capable ....... Yes\n");
      else
         printf (" Encryption Capable ....... No\n");
      if (path->encryption_enable)
         printf (" Encryption Enable ........ Yes\n");
      else
         printf (" Encryption Enable ........ No\n");
   }

   if (path->devpath[0] != '\0')
      printf (" Device Physical Path ..... /devices%-s\n", path->devpath);
   else
      printf (" Device Physical Path ..... N/A\n");

   return;
}


/*---------------------------------------------------------------------------*/
/*        Name: rwtest                                                       */
/*    Synopsis: Write random data to tape, then read it in.  Compare.        */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int rwtest(int blksize, int blocks, int repetition)
{
   int rc;
   int i, j, k;
   char *wbuf;
   char *rbuf;
   long long total = 0;
   int bytes;
   int count;

   printf("\nRead/Write test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

   if (blksize == 0) {
      count = blocks;
      blocks = 1;
   } else
      count = blocks * blksize;

   wbuf = malloc(count);
   if (wbuf == NULL) {
      errno = ENOMEM;
            printf ("Error %d (%s) trying to allocate memory.\n",
            errno, strerror(errno));
      return errno;
   }

   rbuf = malloc(count);
   if (rbuf == NULL) {
      errno = ENOMEM;
            printf ("Error %d (%s) trying to allocate memory.\n",
            errno, strerror(errno));
      free(wbuf);
      return errno;
   }

   verbose = FALSE;
   printf("Setting block size to %d............... ",blksize);
   if (rc = st_tape_op(MTSRSZ, blksize)) {
            printf ("Error %d (%s) setting block size.\n",
            errno, strerror(errno));
      verbose = TRUE;
      free(wbuf);
      free(rbuf);
      return errno;
   }
   printf("done.\n");

   /* generate data pattern */
   genpat();
   j=1;
   for (i=0;i<count;i++) {
      wbuf[i] = (char)j;
      j=pat[j];
   }

   for (j=0,total=0;j<repetition;j++) {
      errno = 0;
      printf("\nRead/Write test pass %d\n",j+1);
      printf("Writing %d bytes, %d blocks........",count, blocks);
      bytes = write(dev_fd,wbuf,count);
      if (bytes == -1) {
         printf("ERROR %d (%s) trying to write file.", errno, strerror(errno));
         rc = -1;
         verbose = TRUE;
         break;
      }

      total += bytes;
      if (blksize == 0)
         printf("done. Bytes written %d, blocks 1.\n", bytes);
      else
         printf("done. Bytes written %d, blocks %d.\n", bytes, bytes/blksize);

      if (bytes != count) {
         printf("Error on write, bytes written not equal to bytes requested.\n");
         printf("bytes %d count %d\n", bytes, count);
         errno = 999;
         rc = -1;
         verbose = TRUE;
         break;
      }

      verbose = FALSE;
      printf("Writing 2 file marks.............. ");
      if (rc = st_tape_op(MTWEOF, 2)) {
         printf("\nError performing tape operation, return code is %d (%s).\n",
               errno, strerror(errno));
         verbose = TRUE;
         break;
      }
      printf("done.\n");

      printf("Backward spacing 2 file marks..... ");
      if (rc = st_tape_op(MTBSF, 2)) {
         printf("\nError performing tape operation, return code is %d (%s).\n",
               errno, strerror(errno));
         verbose = TRUE;
         break;
      }
      printf("done.\n");

      printf("Backward spacing %d records.......... ",blocks);
      if (rc = st_tape_op(MTBSR, blocks)) {
         printf("\nError performing tape operation, return code is %d (%s).\n",
               errno, strerror(errno));
         verbose = TRUE;
         break;
      }
      printf("done.\n");

      /* read the data back from tape and compare to original */
      bzero(rbuf,count);
      printf("Reading %d bytes, %d blocks ..... ",count, blocks);
      bytes = read(dev_fd,rbuf,count);
      if (bytes == -1) {
         printf("ERROR %d (%s) trying to read file.", errno, strerror(errno));
         verbose = TRUE;
         break;
      }

      if (blksize == 0)
         printf("done. Bytes read %d, blocks 1.\n", bytes);
      else
         printf("done. Bytes read %d, blocks %d.\n", bytes, bytes/blksize);

      if (bytes != count) {
         printf("Error on read, bytes read not equal to bytes written.\n");
         errno = 999;
         rc = -1;
         break;
      }

      printf("Forward spacing file mark......... ");
      if (rc = st_tape_op(MTFSF, 1)) {
         printf("\nError performing tape operation, return code is %d (%s).\n",
               errno, strerror(errno));
         verbose = TRUE;
         break;
      }
      printf("done.\n");

      verbose = TRUE;

      for (k=0; k < count; k++) {
         if (wbuf[k] != rbuf[k]) {
            printf("Read data miscompare, ");
            printf("byte read at position %d is %02X, should be %02X\n", k,rbuf[k],wbuf[k]);
            errno = 999;
            break;
         }
      }
      if (k != count) {
         rc=0;
         continue;
      }
      printf("Data verification complete, no discrepancies found\n");
   }

   if (rc)
      printf("\nRead/Write test terminated, total bytes written %lld\n",total);
   else
      printf("\nRead/Write test completed, total bytes written %lld\n",total);

   verbose = TRUE;
   free(wbuf);
   free(rbuf);
   return errno;
}


/*---------------------------------------------------------------------------*/
/*        Name: rwtestfile                                                   */
/*    Synopsis: Write a file to tape, then read it in.  Compare the files.   */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int rwtestfile(char *file, int count)
{
   int k;
   int rc;
   char command[256];

   errno=0;
   printf("\nRead/Write test, input file is %s, output is temp.txt \n", file);

   strcpy(command, "/usr/bin/diff ");
   strcat(command, file);
   strcat(command, " temp.txt > diff.out");

   for (k=0; k < count; k++) {

      printf("\nRead/Write test pass %d\n",k+1);

      verbose = FALSE;
      printf("Writing file mark................. ");
      if (rc = st_tape_op(MTWEOF, 1)) {
         printf("ERROR %d (%s). Return code is %d.\n",
               errno, strerror(errno), rc);
         break;
      }
      printf("done.\n");

      if (twrite(file)) {
         printf("Error writing %s to tape.\n", file);
         errno = -1;
         break;
      }

      verbose = FALSE;
      printf("Writing file mark................. ");
      if (rc = st_tape_op(MTWEOF, 1)) {
         printf("ERROR %d (%s). Return code is %d.\n",
               errno, strerror(errno), rc);
         break;
      }
      printf("done.\n");

      verbose = FALSE;
      printf("Backward spacing 2 file marks..... ");
      if (rc = st_tape_op(MTBSF, 2)) {
         printf("ERROR %d (%s). Return code is %d.\n",
               errno, strerror(errno), rc);
         break;
      }
      printf("done.\n");

      verbose = FALSE;
      printf("Forward spacing file mark......... ");
      if (rc = st_tape_op(MTFSF, 1)) {
         printf("ERROR %d (%s). Return code is %d.\n",
               errno, strerror(errno), rc);
         break;
      }
      printf("done.\n");

      tread("temp.txt", 0);

      printf("Comparing Files................... ");
      rc = (uchar)(system(command) >> 8);

      if(!rc) {
         printf("Files identical.\n\nData verification complete, no discrepancies found\n\n");
         (void)system("rm temp.txt");
         (void)system("rm diff.out");
      } else
         printf("Files not identical.\n\nExamine diff.out for a list of discrepancies.\n\n");

   }
   verbose = TRUE;

   return errno;
}


/*---------------------------------------------------------------------------*/
/*        Name: wtest                                                        */
/*    Synopsis: Write a specified number of blocks to tape.                  */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int wtest(int blksize, int blocks, int repetition)
{
   int rc = 0;
   int i, j;
   char *wbuf;
   long long total = 0;
   int bytes;
   int count;

   errno=0;
   printf("\nWrite test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

   if (blksize == 0) {
      count = blocks;
      blocks = 1;
   } else
      count = blocks * blksize;

   wbuf = malloc(count);
   if (wbuf == NULL) {
      errno = ENOMEM;
            printf ("Error %d (%s) trying to allocate memory.\n",
            errno, strerror(errno));
      return errno;
   }

   printf("Setting block size...");
   if (rc = st_tape_op(MTSRSZ, blksize)) {
            printf ("Error %d (%s) setting block size. Return code %d.\n",
            errno, strerror(errno), rc);
      free(wbuf);
      return errno;
   }

   /* generate data pattern */
   genpat();
   j=01;
   for (i=0;i < count;i++) {
      wbuf[i] = (char)j;
      j=pat[j];
   }

   for (j=0,total=0;j<repetition;j++) {

      printf("\nWrite test pass %d\n",j+1);
      printf("Writing %d bytes, %d blocks...\n",count,blocks);
      bytes = write(dev_fd,wbuf,count);
      if (bytes == -1)
         break;

      total += bytes;
      if (blksize == 0)
         printf("Write complete, bytes written %d, blocks 1\n", bytes);
      else
         printf("Write complete, bytes written %d, blocks %d\n", bytes, bytes/blksize);

      if (bytes != count) {
         printf("Error on write, bytes written not equal to bytes requested\n");
         errno = 999;
         rc = -1;
         break;
      }
   }

   if (errno)
      printf("\nERROR: Write test terminated abnormally, total bytes written %lld, errno %d\n",total, errno);
   else
      printf("\nWrite test completed, total bytes written %lld\n",total);

   free(wbuf);
   return errno;
}

/*---------------------------------------------------------------------------*/
/*        Name: rtest                                                        */
/*    Synopsis: Read specified number of blocks from tape.                   */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int rtest(int blksize, int blocks, int repetition)
{
   int rc = 0;
   int i, j, k;
   char *wbuf;
   char *rbuf;
   long long total = 0;
   int bytes;
   int count;

   errno=0;
   printf("\nRead test, blocksize %u count %u repetition %u\n",
         blksize,blocks,repetition);

   if (blksize == 0) {
      count = blocks;
      blocks = 1;
   } else
      count = blocks * blksize;

   wbuf = malloc(count);
   if (wbuf == NULL) {
      errno = ENOMEM;
            printf ("Error %d (%s) trying to allocate memory.\n",
            errno, strerror(errno));
      return errno;
   }

   rbuf = malloc(count);
   if (rbuf == NULL) {
      errno = ENOMEM;
            printf ("Error %d (%s) trying to allocate memory.\n",
            errno, strerror(errno));
      free(wbuf);
      return errno;
   }

   printf("Setting block size...");
   if (rc = st_tape_op(MTSRSZ, blksize)) {
            printf ("Error %d (%s) setting block size. Return code %d.\n",
            errno, strerror(errno), rc);
      free(wbuf);
      free(rbuf);
      return errno;
   }

   /* generate data pattern */
   genpat();
   j=01;
   for (i=0;i < count;i++) {
      wbuf[i] = (char)j;
      j=pat[j];
   }

   for (j=0,total=0;j<repetition;j++) {

      printf("\nRead test pass %d\n",j+1);

      /* read the data back from tape and compare to expected write data */
      bzero(rbuf,count);
      printf("Reading %d bytes, %d blocks...\n",count,blocks);
      bytes = read(dev_fd,rbuf,count);
      if (bytes == -1) {
         if(errno == ENOMEM) {
            /* this is overlength read */
            printf("Overlength read.\n");
            dump_bytes(rbuf, count);
          }

        rc = -1;
        break;
      }

      total += bytes;
      if (blksize == 0)
         printf("Read complete, bytes read %d, blocks 1\n", bytes);
         else
         printf("Read complete, bytes read %d, blocks %d\n", bytes, bytes/blksize);

      for (k=0; k < count; k++) {
         if (wbuf[k] != rbuf[k]) {
            printf("Read data miscompare, ");
            printf("byte read at position %d is %02X, should be %02X\n", k,rbuf[k],wbuf[k]);
            errno = 999;
            break;
         }
      }

      if (k != count) {
         errno=0;
         continue;
      }
      printf("Data verification complete, no discrepancies found\n");
   }

   if (errno)
      printf("\nERROR: Read test terminated abnormally, total bytes read %lld\n",total);
      else
      printf("\nRead test completed, total bytes read %lld\n",total);

   free(wbuf);
   free(rbuf);
   return errno;
}

/*---------------------------------------------------------------------------*/
/*        Name: genpat                                                       */
/*    Synopsis: Random pattern for read/write tests.                         */
/*     Returns: none.                                                        */
/*---------------------------------------------------------------------------*/
void genpat(void)
{
   pat[0x00] = 0x01;
   pat[0x01] = 0x13;
   pat[0x13] = 0x37;
   pat[0x37] = 0x7F;
   pat[0x7F] = 0xFE;
   pat[0xFE] = 0xFC;
   pat[0xFC] = 0xE8;
   pat[0xE8] = 0xC0;
   pat[0xC0] = 0x81;
   pat[0x81] = 0x03;
   pat[0x03] = 0x07;
   pat[0x07] = 0x1F;
   pat[0x1F] = 0x3E;
   pat[0x3E] = 0x7C;
   pat[0x7C] = 0xF8;
   pat[0xF8] = 0xF0;
   pat[0xF0] = 0xE1;
   pat[0xE1] = 0xC3;
   pat[0xC3] = 0x87;
   pat[0x87] = 0x0F;
   pat[0x0F] = 0x0E;
   pat[0x0E] = 0x1C;
   pat[0x1C] = 0x38;
   pat[0x38] = 0x70;
   pat[0x70] = 0xF1;
   pat[0xF1] = 0xF3;
   pat[0xF3] = 0xE7;
   pat[0xE7] = 0xCF;
   pat[0xCF] = 0x8E;
   pat[0x8E] = 0x0C;
   pat[0x0C] = 0x08;
   pat[0x08] = 0x10;
   pat[0x10] = 0x31;
   pat[0x31] = 0x73;
   pat[0x73] = 0xF7;
   pat[0xF7] = 0xFF;
   pat[0xFF] = 0xEE;
   pat[0xEE] = 0xCC;
   pat[0xCC] = 0x88;
   pat[0x88] = 0x00;
}


/*---------------------------------------------------------------------------*/
/*        Name: tread                                                        */
/*    Synopsis: Read a file from tape.                                       */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int tread(char *fname, int number)
{
   int rdfd, rc=0;
   int bytes,count,blksize;
   char *rbuf;
   long long total = 0;
   int rec_count = 0;

   printf("Opening destination file ......... ");
   rdfd = open(fname, O_RDWR|O_CREAT|O_TRUNC|O_LARGEFILE,0666);
   if (rdfd == -1) {
      printf("ERROR %d (%s) trying to open file.\n", errno, strerror(errno));
      return errno;
   } else
      printf("%s open.\n", fname);

   verbose = FALSE;
   if (rc = scsi_test_unit_ready()) {
            printf ("ERROR %d (%s) issuing test unit ready.\n",
            errno, strerror(errno));
      close(rdfd);
      verbose = TRUE;
      return errno;
   }

   blksize = RWBUFSZ;

   printf("Setting block size to variable.... ");
   if (rc = st_tape_op(MTSRSZ, 0)) {
            printf ("ERROR %d (%s) setting block size.\n",
            errno, strerror(errno));
      close(rdfd);
      verbose = TRUE;
      return errno;
   }
   printf("done.\n");
   verbose = TRUE;

   rbuf = malloc(blksize);
   if (rbuf == NULL) {
      close(rdfd);
      printf("ERROR allocating memory.\n");
      errno = ENOMEM;
      return errno;
   }

   if (number)
      printf("Reading %d records from tape...",number);
      else
      printf("Reading file from tape............ ");

   for(;;) {

      if (number && (rec_count == number))
         break;
      bytes = read(dev_fd,rbuf,blksize);
      if (bytes == -1) {
         rc = errno;
         break;
      }
      total += bytes;
      if (bytes == 0)
         break;
      rec_count++;
      count = write(rdfd,rbuf,bytes);
      if (count == -1) {
         rc = errno;
         break;
      }
      if (count != bytes) {
         rc = EIO;
         break;
      }
   }

   if (rc || (number && (rec_count != number)))
      printf("ERROR: Read terminated abnormally. %d records, %lld total bytes read.\n",
            rec_count,total);
   else
      printf("done. %d records, %lld total bytes read.\n",
            rec_count,total);

   close(rdfd);
   free(rbuf);
   return rc;
}

/*---------------------------------------------------------------------------*/
/*        Name: twrite                                                       */
/*    Synopsis: Write a file to tape.                                        */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int twrite(char *fname)
{
   int wrtfd, rc=0;
   int bytes,count,blksize;
   char *wbuf;
   long long total = 0;

   printf("Opening source file............... ");
   wrtfd = open(fname, O_RDONLY|O_LARGEFILE);
   if (wrtfd == -1) {
      printf("ERROR %d (%s) trying to open file.\n", errno, strerror(errno));
      return errno;
   } else
      printf("%s open.\n", fname);

   verbose = FALSE;
   if (rc = scsi_test_unit_ready()) {
            printf ("ERROR %d (%s) issuing test unit ready.\n",
            errno, strerror(errno));
      close(wrtfd);
      verbose = TRUE;
      return errno;
   }

   blksize = RWBUFSZ;

   printf("Setting block size to variable.... ");
   if (rc = st_tape_op(MTSRSZ, 0)) {
            printf ("ERROR %d (%s) setting block size.\n",
            errno, strerror(errno));
      close(wrtfd);
      verbose = TRUE;
      return errno;
   }
   printf("done.\n");
   verbose = TRUE;

   wbuf = malloc(blksize);
   if (wbuf == NULL) {
      printf("ERROR allocating memory.\n");
      close(wrtfd);
      errno = ENOMEM;
      return errno;
   }

   printf("Writing file to tape.............. ");

   for(;;) {

      bytes = read(wrtfd,wbuf,blksize);
      if (bytes == -1) {
         printf("Read error.\n");
         rc = errno;
         break;
      }
      if (bytes == 0)
         break;
      count = write(dev_fd,wbuf,bytes);
      if (count == -1) {
         printf("Write error.\n");
         rc = errno;
         break;
      }
      total += count;
      if (count != bytes) {
         rc = EIO;
         break;
      }
   }

   if (rc)
      printf("ERROR: Write terminated abnormally. %lld total bytes written.\n",total);
   else
      printf("done. %lld total bytes written.\n",total);

   close(wrtfd);
   free(wbuf);
   return rc;
}


/*---------------------------------------------------------------------------*/
/*        Name: st_get_all_parms                                             */
/*    Synopsis: Display all device specific parameters.                      */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_get_all_parms ()
{
   int rc;
   parm_data_t parm_data;

   if(medium_chgr) {
      printf("Invalid command for medium changer.\n");
      return 1;
   }

   printf("\n-- Tape Parameters ---------------------\n\n");

   parm_data.type = BLOCKSIZE;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    1. Block Size......... %-10d\n",
         parm_data.value);
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = COMPRESSION;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    2. Compression........ %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = BUFFERING;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    3. Buffering.......... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = IMMEDIATE;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    4. Immediate.......... ");
      switch(parm_data.value) {
        case NO_IMMEDIATE:
           printf("None\n");
        break;
        case GEN_IMMEDIATE:
           printf("General Immediate\n");
        break;
        case REW_IMMEDIATE:
           printf("Rewind Immediate\n");
        break;
        default:
           printf("Unsupported\n");
      }
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = TRAILER;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    5. Trailer Labels..... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = WRITEPROTECT;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    6. Write Protection... ");
      switch(parm_data.value) {
      case NO_PROTECT:
         printf("None\n");
         break;
      case PHYS_PROTECT:
         printf("Physical\n");
         break;
      case WORM_PROTECT:
         printf("Logical (WORM)\n");
         break;
      case PERS_PROTECT:
         printf("Logical (PERS)\n");
         break;
      case ASSC_PROTECT:
         printf("Logical (ASSC)\n");
         break;
      default:
         printf("Unsupported\n");
      }
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = ACFMODE;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    7. ACF Mode........... ");
      switch(parm_data.value) {
      case NO_ACF:
         printf("No ACF\n");
         break;
      case SYSTEM_MODE:
         printf("System\n");
         break;
      case RANDOM_MODE:
         printf("Random\n");
         break;
      case MANUAL_MODE:
         printf("Manual\n");
         break;
      case ACCUM_MODE:
         printf("Accumulate\n");
         break;
      case AUTO_MODE:
         printf("Automatic\n");
         break;
      case LIB_MODE:
         printf("Library\n");
         break;
      default:
         printf("Unsupported\n");
      }
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = SCALING;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    8. Capacity Scaling... ");
      if ((device.lto2) || (device.lto3) || (device.lto4)) {
         verbose = FALSE;
         interactive = FALSE;
         if (!(rc = st_report_density_support()))
            printf("%2.2f%%(%d megabytes)\n", (float)parm_data.value/max_capacity*100, parm_data.value);
         else
            printf("%d(megabytes)\n", parm_data.value);
         verbose = TRUE;
         interactive = TRUE;
      }
      else {
         switch(parm_data.value) {
         case SCALE_100:
            printf("100%%\n");
         break;
         case SCALE_75:
            printf("75%%\n");
         break;
         case SCALE_50:
            printf("50%%\n");
         break;
         case SCALE_25:
            printf("25%%\n");
         break;
         default:
            if (device.t3592) {
               if (parm_data.value == 0)
                  printf("100%%\n");
               else
               printf("%2.2f%%\n", (float)parm_data.value/256*100);
            }
            else
               printf("100%%\n");
         }
      }
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   parm_data.type = SILI;
   if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
      printf("    9. SILI Mode.......... %-10s\n",
         (parm_data.value ? "Yes" : "No") );
   }
   else {
      printf ("STIOC_GET_PARM failed.\n");
      printf ("\n");
      scsi_request_sense ();
   }

   printf("\n----------------------------------------\n\n");
   return (rc);
}

/*---------------------------------------------------------------------------*/
/*        Name: st_set_all_parms                                             */
/*    Synopsis: Allow interactive setting of parameters.                     */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int st_set_all_parms ()
{
   int rc = 0;
   parm_data_t parm_data;
   char buf[10];

   for(;;) {
      st_get_all_parms();

      printf ("Enter parameter to change or <enter> to quit: ");
      gets (buf);
      if ( buf[0] == '\0' || buf[0] == 'q' )
         return(rc);
      else {
         parm_data.type = atoi(buf);

         if((parm_data.type == 2)||(parm_data.type == 3)||(parm_data.type == 5)||(parm_data.type == 9)){
            if (!(rc = ioctl (dev_fd, STIOC_GET_PARM, &parm_data))) {
               parm_data.value = !parm_data.value;
               if (rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data)) {
                  printf ("STIOC_SET_PARM failed.\n");
                  printf ("\n");
                  scsi_request_sense ();
                  return rc;
               }
            }
            else {
               printf ("STIOC_GET_PARM failed.\n");
               printf ("\n");
               scsi_request_sense ();
               return rc;
            }
         }
         else if(parm_data.type == IMMEDIATE) {
            printf("Enter Immediate Mode (1=NONE, 2=General, 3=Rewind): ");
            gets (buf);
            switch(atoi(buf)) {
            case 1:
               parm_data.value = NO_IMMEDIATE;
               break;
            case 2:
               parm_data.value = GEN_IMMEDIATE;
               break;
            case 3:
               parm_data.value = REW_IMMEDIATE;
               break;
            default:
               continue;
            }
            rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data);
         }
         else if(parm_data.type == WRITEPROTECT) {
            if(device.ultrium) {
               printf("Write protection cannot be set in software for Ultrium devices.\n");
               printf ("Hit <enter> to continue...\n");
               getchar ();
               continue;
            }
            printf("Enter write protect (1=NONE, 2=ASSOCIATED, 3=PERSISTENT, 4=WORM): ");
            gets (buf);
            switch(atoi(buf)) {
            case 1:
               parm_data.value = NO_PROTECT;
               break;
            case 2:
               parm_data.value = ASSC_PROTECT;
               break;
            case 3:
               parm_data.value = PERS_PROTECT;
               break;
            case 4:
               parm_data.value = WORM_PROTECT;
               break;
            default:
               continue;
            }
            rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data);
         }
         else if(parm_data.type == BLOCKSIZE) {
            printf("Enter new block size: ");
            gets(buf);
            parm_data.value = atoi(buf);
            rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data);
         }
         else if(parm_data.type == ACFMODE) {
            printf("ACF modes can only be set via the tape drive operator panel.\n");
            printf ("Hit <enter> to continue...\n");
            getchar ();
         }
         else if(parm_data.type == SCALING) {
            if(device.lto1) {
               printf("Scaling not supported for Ultrium GEN 1 drive.\n");
               printf ("Hit <enter> to continue...\n");
               getchar ();
               continue;
            }
            if (device.t3592)
               printf("Enter capacity scaling (1=100%%, 2=75%%, 3=50%%, 4=25%% or other hexadecimal value): ");
            else if ((device.lto2) ||(device.lto3) || (device.lto4))
               printf("Enter capacity scaling (Enter a percent value(10 for 10%)): ");
            else
               printf("Enter capacity scaling (1=100%%, 2=75%%, 3=50%%, 4=25%%): ");
            gets (buf);
            if ((device.lto2) || (device.lto3) || (device.lto4))
               parm_data.value = (uint) atoi(buf);
            else {
               switch(atoi(buf)) {
                 case 1:
                   parm_data.value = SCALE_100;
                 break;
                 case 2:
                   parm_data.value = SCALE_75;
                 break;
                 case 3:
                   parm_data.value = SCALE_50;
                 break;
                 case 4:
                   parm_data.value = SCALE_25;
                 break;
                 default:
                   if (device.t3592) {
                      parm_data.value = (int) strtol(buf, NULL, 16);
                      break;
                   }
                   else
                      continue;
               }
            }
            printf("Warning!  All data on tape will be lost with this change!\n");
            printf("Continue (y or n): ");
            gets(buf);
            if(buf[0] == 'y') {
               printf("Rewinding the tape...\n");
               if (st_tape_op(MTREW, 0))
               {
                  return errno;
               }
               verbose = TRUE;

               rc = ioctl (dev_fd, STIOC_SET_PARM, &parm_data);
            }
         }

         if (rc) {
            printf ("STIOC_SET_PARM failed.\n");
            printf ("\n");
            scsi_request_sense ();
            return (rc);
         }
      } /* else */
   } /* for */
}

/*---------------------------------------------------------------------------*/
/*        Name: rwtest_menu                                                  */
/*    Synopsis: Gather options for interactive read/write tests.             */
/*     Returns: Error code from /usr/include/sys/errno.h.                    */
/*---------------------------------------------------------------------------*/
static int rwtest_menu() {

   char buf[80];
   char file[80];
   int type;

   int blocksize;
   int num_blocks;
   int repetitions;

   for(;;) {

      printf("-- Read/Write Tests --------------------\n");
      printf("                                        \n");
      printf("    1. Read/Write Test (Random Data)    \n");
      printf("    2. Write Random Data                \n");
      printf("    3. Read Random Data                 \n");
      printf("    4. Read/Write Test (File)           \n");
      printf("    5. Write File                       \n");
      printf("    6. Read File                        \n");
      printf("                                        \n");
      printf("----------------------------------------\n");

      printf ("Select operation or <enter> to quit: ");
      gets (buf);
      if ( buf[0] == '\0' || buf[0] == 'q' )
         return(0);
      else {
         type = atoi(buf);

         if( (type >= 1) && (type <= 3)) {
            printf ("Total Bytes = ");
            printf ("(block size) * (number of blocks) * (repetitions).\n\n");

            printf ("Enter block size [10240(default)]: ");
            gets (buf);
            if( buf[0] == '\0' )
               blocksize = 10240;
            else
               blocksize = atoi (buf);

            if (blocksize == 0) {
               printf("\nNumber of bytes per read/write[65536(default)]: ");
               gets(buf);
               num_blocks = (int)atoi(buf);
               if (num_blocks == 0)
                  num_blocks = 65536;
            }
            else {
               printf ("Number of blocks [20(default)]: ");
               gets (buf);
               if( buf[0] == '\0' )
                 num_blocks = 20;
               else
                 num_blocks = atoi (buf);
            }

            printf ("Number of repetitions [1(default)]: ");
            gets (buf);
            if( buf[0] == '\0' )
               repetitions = 1;
            else {
               repetitions = atoi (buf);
               if (repetitions < 1) {
                  printf ("\nInvalid Number %d\n", repetitions);
                  printf ("Number of repetitions is changed to the default.\n");
                  repetitions = 1;
                  continue;
               }
            }
         }

         switch (type) {

         case 1:
            rwtest(blocksize, num_blocks, repetitions);
            break;

         case 2:
            wtest(blocksize, num_blocks, repetitions);
            break;

         case 3:
            rtest(blocksize, num_blocks, repetitions);
            break;

         case 4:
            printf ("Enter file name: ");
            gets (buf);
            if( buf[0] == '\0' )
               continue;
            else {
               strcpy(file, buf);
            }

            printf ("Number of repetitions [1]: ");
            gets (buf);
            if( buf[0] == '\0' )
               repetitions = 1;
            else {
               repetitions = atoi (buf);
               if (repetitions < 1) {
                  printf ("\nNumber of repetitions out of range.\n");
                  continue;
               }
            }

            rwtestfile(file, repetitions);
            break;

         case 5:
            printf ("Enter file name: ");
            gets (buf);
            if( buf[0] == '\0' )
               continue;
            else {
               strcpy(file, buf);
            }

            twrite(file);
            break;

         case 6:
            printf ("Enter file name: ");
            gets (buf);
            if( buf[0] == '\0' )
               continue;
            else {
               strcpy(file, buf);
            }

            printf ("Number of blocks [0=entire file]: ");
            gets (buf);
            if( buf[0] == '\0' )
               num_blocks = 0;
            else {
               num_blocks = atoi (buf);
               if (num_blocks < 0) {
                  printf ("\nNumber of blocks out of range.\n");
                  continue;
               }
            }

            tread(file, num_blocks);
            break;

         default:
            continue;

         } /* switch */
      } /* if */
   } /* for */
}

/*---------------------------------------------------------------------------*/
/*        Name: fmt_bytes                                                    */
/*    Synopsis: Display byte data.                                           */
/*     Returns: Nothing.                                                     */
/*---------------------------------------------------------------------------*/
void fmt_bytes ( char *data, int count, int start_offset, char *id, int across )
{
   #define HEX_BUFMAX 64
   #define ASCII_BUFLEN 26
   int i;
   int hex_index;
   int ascii_index;
   int hex_buflen;
   uchar hex_buf[HEX_BUFMAX];
   char hex_adrfmt[8];
   int hex_adrmask;
   char hex_buffmt[16];
   uchar ascii_buf[ASCII_BUFLEN];

   if ((across != 16) && (across != 10)) {
      (void) fprintf ( stderr, "%s: Invalid bytes-per-line value %d!\n", id, across );
      return;
   }

   hex_buflen = (across == 16? 53 : 35);

   (void) sprintf(hex_buffmt, "%%s%%-%ds", hex_buflen);

   (void) fprintf ( stdout, "%s\n", id );
   if (across == 16) {
      (void) fprintf ( stdout, "%s%s %s %s \n",
         id,
         "   ",
         " _0 _1 _2 _3 _4 _5 _6 _7  _8 _9 _A _B _C _D _E _F",
         "   01234567 89ABCDEF"
         );
      (void) fprintf ( stdout, "%s%s %s %s \n",
         id,
         "   ",
         " -----------------------  -----------------------",
         "   -------- --------"
         );
   }
   else {
      (void) fprintf ( stdout, "%s%s %s %s \n",
         id,
         "   ",
         " _0 _1 _2 _3 _4  _5 _6 _7 _8 _9",
         "   01234 56789"
         );
      (void) fprintf ( stdout, "%s%s %s %s \n",
         id,
         "   ",
         " --------------  --------------",
         "   ----- -----"
         );
   }

   hex_index   = 0;
   ascii_index = 0;
   (void) memset (hex_buf  , 0, HEX_BUFMAX);
   (void) memset (ascii_buf, 0, ASCII_BUFLEN);
   if (across == 10) {
      (void) sprintf(hex_adrfmt, "%s", "%4.2d  ");
      hex_adrmask = 0xFFFF;
   }
   else if (count + start_offset > 4096) {
      (void) sprintf(hex_adrfmt, "%s", "%4.2X  ");
      hex_adrmask = 0xFFF0;
   }
   else if (count + start_offset > 256) {
      (void) sprintf(hex_adrfmt, "%s", "%3.2X  ");
      hex_adrmask = 0xFF0;
   }
   else {
      (void) sprintf(hex_adrfmt, "%s", "%2.2X  ");
      hex_adrmask = 0xF0;
   }

   for (i = 0; i < count; i++) {
      if (!(i % across)) {
         /* Print high nibble for new line of dump. */
         (void) sprintf ((char *)(&hex_buf[hex_index]), hex_adrfmt ,
            i + start_offset & hex_adrmask );
         hex_index += 4;
      }
      else if (!(i % (across / 2))) {
         /* Visually separate line into two sections. */
         (void) sprintf ((char *)(&hex_buf[hex_index]), " ");
         hex_index += 1;
         (void) sprintf ((char *)(&ascii_buf[ascii_index]), " ");
         ascii_index += 1;
      }

      /* Write current char to hex buffer... */
      (void) sprintf ((char *)(&hex_buf[hex_index]), " %2.2X", (uchar)data[i]);
      hex_index   += 3;

      /* ...and to ascii buffer. */
      if ((uchar)data[i] < 0x20) {
         ascii_buf[ascii_index] = '.';
      }
      else if ( ((uchar)data[i] >= 0x7F) && ((uchar)data[i] <= 0xA0) ) {
         ascii_buf[ascii_index] = '.';
         /* ascii_buf[ascii_index] = (((uchar)rdbuf[ch_idx]) + 0x20); */
      }
      else if ((uchar)data[i] >= 0xff) {
         ascii_buf[ascii_index] = '.';
      }
      else {
         ascii_buf[ascii_index] = data[i];
      }
      ascii_index += 1;

      /* We just wrote the last char of a line or of the input data.
      ** Print out the two buffers */
      if (((i + 1) % across == 0) || (i + 1 == count)) {
         (void) fprintf ( stdout, hex_buffmt, id, hex_buf);
         /*
         if ((i + 1) % across) {
            (void) fprintf ( stdout, "  ");
         }
         */
         (void) fprintf ( stdout, "    %-16s\n", ascii_buf);
         hex_index   = 0;
         ascii_index = 0;
         (void) memset (hex_buf  , 0, HEX_BUFMAX);
         (void) memset (ascii_buf, 0, ASCII_BUFLEN);
      }
   }
}
/* fmt_bytes */



/*---------------------------------------------------------------------------*/
/*        Name: scsi_log_sense_page                                          */
/*    Synopsis: Request log sense page; print out result.                    */
/*     Returns: Requested log sense page.                                    */
/*---------------------------------------------------------------------------*/
static int scsi_log_sense_page (uchar page, uchar subpage)
{
   int            rc;
   int            log_data_len, data_len = 0;
   log_sns_pg_t   log_sns_pg;
   uchar          *log_page_p = (uchar *)(&log_sns_pg.data);
   log_page_hdr_t *log_page_hdr_p = (log_page_hdr_t *)(&log_sns_pg.data);

   memset ((char *)&log_sns_pg, 0, sizeof (log_sns_pg_t));
   log_sns_pg.page_code = page;
   log_sns_pg.subpage_code = subpage;

   if (!(rc = ioctl (dev_fd, IOC_LOG_SENSE_PAGE, &log_sns_pg))) {
      data_len = (uint) (((log_page_hdr_p->len[0] << 8)
         | log_page_hdr_p->len[1]) + 4);
      log_data_len = MIN(data_len, sizeof log_sns_pg.data);
      PRINTF ("IOC_LOG_SENSE_PAGE succeeded.\n");
      if (subpage == 0)
         PRINTF ("\nLog page 0x%X data, length %d (%d returned):\n", log_sns_pg.page_code, data_len, log_data_len);
      else
         PRINTF ("\nLog page 0x%X subpage 0x%X data, length %d (%d returned):\n", log_sns_pg.page_code,
            log_sns_pg.subpage_code, data_len, log_data_len);
      DUMP_BYTES ((char *)log_page_p, log_data_len);
   }
   else {
      PERROR ("log sense");
      PRINTF ("\n");
      scsi_request_sense ();
   }

   return (rc);
}


/*---------------------------------------------------------------------------*/
/*        Name: scsi_mode_sense_page                                         */
/*    Synopsis: Request mode sense page; print out result.                   */
/*     Returns: Requested mode sense page.                                   */
/*---------------------------------------------------------------------------*/
static int scsi_mode_sense_page (uchar page_code)
{
   int               rc;
   int               header_len;
   int               blk_dsc_len = 0;
   int               mode_data_len = 0;
   int               mode_data_returned_len = 0;
   int               max_mdsnspg_data_len = 0;
   uchar             cmd_code;
   uchar             medium_type;
   uchar             density_code;
   uchar             wrt_prot;
   char              *header_p;
   char              *blkdsc_p;
   void              *mode_data_p;
   char              * dens_str_p;
   char              * med_str_p;
   char              * wrt_prot_p;
   char              err_msg[1024];
   mode_sns_t        mode_page;

   memset (err_msg,            (char)0, sizeof(err_msg));
   memset ((char *)&mode_page, 0, sizeof(mode_sns_t));

   mode_page.page_code = page_code;
   if (!(rc = ioctl (dev_fd, IOC_MODE_SENSE, &mode_page))) {
      PRINTF ("IOC_MODE_SENSE succeeded.\n");
      header_p = (char *)&mode_page.data;
      cmd_code = mode_page.cmd_code;
      if ( cmd_code == MODESNS_6_CMD ) {
         header_len             = sizeof(mode_hdr6_t);
         mode_data_len          = (uint) ((mode_hdr6_t *)header_p)->data_len;
         blk_dsc_len            = (uint) ((mode_hdr6_t *)header_p)->blk_dsc_len;
         max_mdsnspg_data_len   = MAX_MSDATA - header_len - blk_dsc_len;
         mode_data_returned_len = MIN( mode_data_len + 1, max_mdsnspg_data_len);
         medium_type            = (uchar)((mode_hdr6_t *)(header_p))->medium_type;
         wrt_prot                = (uchar)((mode_hdr6_t *)(header_p))->wrt_prot;
      }
      else if ( cmd_code == MODESNS_10_CMD ) {
         header_len    = sizeof(mode_hdr10_t);
         mode_data_len = (uint) ((((mode_hdr10_t *)header_p)->data_len[0] << 8)
               | ((mode_hdr10_t *)header_p)->data_len[1] );
         blk_dsc_len   = (uint) ((((mode_hdr10_t *)header_p)->blk_dsc_len[0] << 8)
               | ((mode_hdr10_t *)header_p)->blk_dsc_len[1] );
         max_mdsnspg_data_len   = MAX_MSDATA - header_len - blk_dsc_len;
         mode_data_returned_len = MIN( mode_data_len + 2, max_mdsnspg_data_len);
         medium_type            = (uchar)((mode_hdr10_t *)(header_p))->medium_type;
         wrt_prot                = (uchar)((mode_hdr10_t *)(header_p))->wrt_prot;
      }
      else {
         fprintf (stderr, "mode sense: Unknown mode sense command code '0x%X'.\n", cmd_code);
         return (1);
      }

      blkdsc_p     = header_p + header_len;
      mode_data_p  = blkdsc_p + blk_dsc_len;
      density_code = (blk_dsc_len
                     ? ( unsigned char )((blkdsc_t *)(blkdsc_p))->density_code
                     : 0 );

      switch (density_code) {
         case 0x28:
         case 0x80:
            dens_str_p = (char *)&("IBM 3480-2 XF Density (36 Track, 1491 cpmm)");
            break;

         case 0x29:
            dens_str_p = (char *)&("IBM TotalStorage 3590 Model B Density");
            break;

         case 0x2A:
            dens_str_p = (char *)&("IBM TotalStorage 3590 Model E Density");
            break;

         case 0x2c:
            dens_str_p = (char *)&("IBM TotalStorage 3590 Model H Density");
            break;

         case 0x51:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model J1A Density");
            break;

         case 0x52:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model E05 Density");
            break;

         case 0x71:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model J1A Encryption Density");
            break;

         case 0x72:
            dens_str_p = (char *)&("IBM TotalStorage 3592 Model E05 Encryption Density");
            break;

         case 0x82:
            dens_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, B-Format Density");
            break;

         case 0x83:
            dens_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, C-Format Density");
            break;

         case 0x40:
            dens_str_p = (char *)&("IBM Ultrium LTO-1 1/8T Density");
            break;

         case 0x42:
            dens_str_p = (char *)&("IBM Ultrium LTO-2 2/8T Density");
            break;

         case 0x44:
            dens_str_p = (char *)&("IBM Ultrium LTO-3 3/16T Density");
            break;

         case 0x46:
            dens_str_p = (char *)&("IBM Ultrium LTO-4 4/16T Density");
            break;

         case 0x00:
            dens_str_p = (char *)&("Medium Not Present");
            break;

         default:
            dens_str_p = (char *)&("? Unrecognized Medium Density Code");
            break;
      }

      switch (medium_type) {
         case 0x81:
            med_str_p = (char *)&("IBM TotalStorage 3590 High Performance Cartridge");
         break;

         case 0x82:
            med_str_p = (char *)&("IBM TotalStorage 3590 High Performance Extended Length Cartridge");
         break;

         case 0x91:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A Tape Cartridge(JA)");
         break;

         case 0x92:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B Tape Cartridge(JB)");
         break;

         case 0xB1:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A Short Cartridge(JJ)");
         break;

         case 0xB2:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B Short Cartridge(JK)");
         break;

         case 0xA1:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A WORM Cartridge(JW)");
         break;

         case 0xA2:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B WORM Cartridge(JX)");
         break;

         case 0xC1:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen A Short WORM Cartridge(JR)");
         break;

         case 0xC2:
            med_str_p = (char *)&("IBM TotalStorage 3592 Gen B Short WORM Cartridge(JS)");
         break;

         case 0x42:
            med_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, B-Format");
         break;

         case 0x43:
            med_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, C-Format");
         break;

         case 0x44:
            med_str_p = (char *)&("IBM Magstar MP Fast Access Linear Tape, C-Format XL");
         break;

         case 0x18:
            med_str_p = (char *)&("IBM Ultrium LTO-1 1/8T Cartridge");
         break;

         case 0x28:
            med_str_p = (char *)&("IBM Ultrium LTO-2 2/8T Cartridge");
         break;

         case 0x38:
            med_str_p = (char *)&("IBM Ultrium LTO-3 3/16T Cartridge");
         break;

         case 0x3C:
            med_str_p = (char *)&("IBM Ultrium LTO-3 WORM Cartridge");
         break;

         case 0x48:
            med_str_p = (char *)&("IBM Ultrium LTO-4 4/16T Cartridge");
         break;

         case 0x4C:
            med_str_p = (char *)&("IBM Ultrium LTO-4 WORM Cartridge");
         break;

         case 0x00:
            if (device.ultrium) med_str_p = (char *)&("Unknown");
            else med_str_p = (char *)&("Medium Not Present");
         break;

         default:
            med_str_p = (char *)&("? Unrecognized Medium Type Code");
         break;
      }

      switch (wrt_prot) {
         case 0x01:
            wrt_prot_p = (char *)&("Medium is Read Only");
         break;

         default:
            wrt_prot_p = (char *)&("Medium is Write-Enabled");
         break;
      }

      PRINTF ("\n");
      PRINTF ("Page Code      x'%2.2X' \n", page_code );
      PRINTF ("Command Code   x'%2.2X' \n", mode_page.cmd_code );
      PRINTF ("Mode Data Len  %4d \n",      mode_data_len);
      PRINTF ("Mode Page Len  %4d\n",     ((*((modep_prefix_t *)mode_data_p)).page_len));
      PRINTF ("Blk Desc Len   %4d \n",      blk_dsc_len );
      PRINTF ("Returned Len   %4d \n",      mode_data_returned_len );
      /*
      */
      PRINTF ("Write Protect  x'%2.2X'\t(%s) \n", wrt_prot, wrt_prot_p );
      PRINTF ("Medium Type    x'%2.2X'\t(%s) \n", medium_type, med_str_p );
      PRINTF ("Density Code   x'%2.2X'\t(%s) \n", density_code, dens_str_p );

      /*
      PRINTF ("&mode_page     x'%X' \n", (char *)&mode_page );
      PRINTF ("header_p       x'%X' \n", header_p );
      PRINTF ("blkdsc_p       x'%X' \n", blkdsc_p );
      PRINTF ("mode_data_p    x'%X' \n", mode_data_p );
      */

      PRINTF ("\nHeader:\n");
      DUMP_BYTES ((char *)(header_p), header_len);
      PRINTF ("\nBlock Descriptor:\n");
      DUMP_BYTES ((char *)(blkdsc_p), blk_dsc_len);
      PRINTF ("\nMode Page:\n");
      DUMP_BYTES ((char *)(mode_data_p), mode_data_returned_len - header_len - blk_dsc_len);
   }
   else {
      PERROR ("mode sense");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}

/*---------------------------------------------------------------------------*/
/*        Name: st_get_encryption_state                                      */
/*    Synopsis: Query drive encrytion status; print out result.              */
/*     Returns: Returns: Error code from /usr/include/sys/errno.h.           */
/*---------------------------------------------------------------------------*/
static int st_get_encryption_state ( )
{
   int rc;
   struct encryption_status  ep_state;

   if (!(rc = ioctl (dev_fd, GET_ENCRYPTION_STATE, &ep_state))) {
      PRINTF ("\nGET_ENCRYPTION_STATE command succeeded.\n");
      PRINTF ("Encryption settings:\n");
      PRINTF ("%20s %s\n", "  Encryption Capable...", (ep_state.encryption_capable == 1 ? "Yes" : "No"));
      PRINTF ("%20s ", "  Encryption Method....");
      switch (ep_state.encryption_method) {
         case 0: PRINTF ("None\n"); break;
         case 1: PRINTF ("Library\n"); break;
         case 2: PRINTF ("System\n"); break;
         case 3: PRINTF ("Application\n"); break;
         case 4: PRINTF ("Custom\n"); break;
         case 5: PRINTF ("Unknown\n"); break;
      }
      PRINTF ("%20s ", "  Encryption State.....");
      switch (ep_state.encryption_state) {
         case 0: PRINTF ("Off\n"); break;
         case 1: PRINTF ("On\n"); break;
         case 2: PRINTF ("NA\n"); break;
      }
   }
   else {
      PERROR ("GET_ENCRYPTION_STATE failed");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}

/*---------------------------------------------------------------------------*/
/*        Name: encryption_test                                              */
/*    Synopsis: Diagnotic drive, ekm server encrytion status                 */
/*     Returns: Returns: Error code from /usr/include/sys/errno.h.           */
/*---------------------------------------------------------------------------*/
int encryption_test()
  {
  int rc;
  struct encryption_diagnostics diag;

  /* test server config and connections */
  printf("Testing server configuration and connections...\n");
  diag.diag_type = SERVER_PING_DIAG;

  if (rc = ioctl (dev_fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag)) {
     printf ("DATA_ENCRYPTION_DIAGNOSTICS ioctl call failed.\n");
     return (rc);
  }

  if (diag.diag_errno) {
    errno = diag.diag_errno;
    perror("Server test failed");
    errno = 0;
    return errno;
  }
  else
    printf("Testing complete, servers available %d\n",diag.diag_cc);

  /* run basic drive to server encryption test */
  printf("Running basic drive to server encryption test...\n");
  diag.diag_type = BASIC_ENCRYPTION_DIAG;

  if (rc = ioctl (dev_fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag)) {
    printf ("Basic Encryption ioctl failed.\n");
    scsi_request_sense ();
  }
  else {
    if (diag.diag_errno) {
      errno = diag.diag_errno;
      perror("Basic Encryption test failed");
      errno = 0;
    }
    else
      printf("Testing complete, completion code %d\n",diag.diag_cc);
  }

  /* run full drive to server encryption test */
  printf("Running full drive to server encryption test...\n");
  diag.diag_type = FULL_ENCRYPTION_DIAG;

  if (rc =ioctl (dev_fd, DATA_ENCRYPTION_DIAGNOSTICS, &diag)) {
    printf ("Full Encryption ioctl failed.\n");
    scsi_request_sense ();
  }
  else
    {
    if (diag.diag_errno)
      {
      errno = diag.diag_errno;
      perror("Full Encryption test failed");
      errno = 0;
      }
    else
      printf("Testing complete, completion code %d\n",diag.diag_cc);
   }

  return (rc);
  }


/*---------------------------------------------------------------------------*/
/*        Name: scsi_mode_sense_subpage                                      */
/*    Synopsis: Request mode sense subpage; print out result.                */
/*     Returns: Requested mode sense subpage.                                */
/*---------------------------------------------------------------------------*/
static int scsi_mode_sense_subpage (uchar page, uchar subpage)
{
   int               rc;
   int               header_len;
   int               blk_dsc_len = 0;
   int               mode_data_len = 0;
   int               mode_data_returned_len = 0;
   int               max_mdsnspg_data_len = 0;
   uchar             cmd_code;
   uchar             medium_type;
   uchar             density_code;
   uchar             wrt_prot;
   char              *header_p;
   char              *blkdsc_p;
   void              *mode_data_p;
   mode_sns_subpage_t   mode_subpage;

   memset ((char *)&mode_subpage, 0, sizeof(mode_sns_subpage_t));

   mode_subpage.page_code = page;
   mode_subpage.subpage_code = subpage;

   if (!(rc = ioctl (dev_fd, IOC_MODE_SENSE_SUBPAGE, &mode_subpage))) {
      PRINTF ("IOC_MODE_SENSE_SUBPAGE succeeded.\n");
      header_p = (char *)&mode_subpage.data;
      cmd_code = mode_subpage.cmd_code;
      if ( cmd_code == MODESNS_6_CMD ) {
         header_len             = sizeof(mode_hdr6_t);
         mode_data_len          = (uint) ((mode_hdr6_t *)header_p)->data_len;
         blk_dsc_len            = (uint) ((mode_hdr6_t *)header_p)->blk_dsc_len;
         max_mdsnspg_data_len   = MAX_MS_SUBDATA - header_len - blk_dsc_len;
         mode_data_returned_len = MIN( mode_data_len + 1, max_mdsnspg_data_len);
         medium_type            = (uchar)((mode_hdr6_t *)(header_p))->medium_type;
         wrt_prot               = (uchar)((mode_hdr6_t *)(header_p))->wrt_prot;
      }
      else if ( cmd_code == MODESNS_10_CMD ) {
         header_len    = sizeof(mode_hdr10_t);
         mode_data_len = (uint) ((((mode_hdr10_t *)header_p)->data_len[0] << 8)
               | ((mode_hdr10_t *)header_p)->data_len[1]);

         blk_dsc_len   = (uint) ((((mode_hdr10_t *)header_p)->blk_dsc_len[0] << 8)
               | ((mode_hdr10_t *)header_p)->blk_dsc_len[1] );
         max_mdsnspg_data_len   = MAX_MS_SUBDATA - header_len - blk_dsc_len;
         mode_data_returned_len = MIN(mode_data_len+2, max_mdsnspg_data_len);
         medium_type            = (uchar)((mode_hdr10_t *)(header_p))->medium_type;
         wrt_prot               = (uchar)((mode_hdr10_t *)(header_p))->wrt_prot;
      }
      else {
         fprintf (stderr, "mode sense: Unknown mode sense command code '0x%X'.\n", cmd_code);
         return (1);
      }

      /* dump_bytes ((char *)&mode_subpage, mode_data_returned_len+16); */

      blkdsc_p     = header_p + header_len;
      mode_data_p  = blkdsc_p + blk_dsc_len;
      density_code = (blk_dsc_len
                     ? ( unsigned char )((blkdsc_t *)(blkdsc_p))->density_code
                     : 0 );

      PRINTF ("\n");
      PRINTF ("Page Code      x'%2.2X'\n", page);
      PRINTF ("SubPage Code   x'%2.2X'\n", subpage);
      PRINTF ("Command Code   x'%2.2X'\n", mode_subpage.cmd_code);
      PRINTF ("Mode Data Len  %4d\n", mode_data_len);
      PRINTF ("Blk Desc Len   %4d\n", blk_dsc_len);
      PRINTF ("Returned Len   %4d\n", mode_data_returned_len);

      PRINTF ("Write Protect  x'%2.2X'\t\n", wrt_prot);
      PRINTF ("Medium Type    x'%2.2X'\t\n", medium_type);
      if (blk_dsc_len != 0)
         PRINTF ("Density Code   x'%2.2X'\t\n", density_code);

      PRINTF ("\nHeader:\n");
      DUMP_BYTES ((char *)(header_p), header_len);
      if (blk_dsc_len != 0) {
         PRINTF ("\nBlock Descriptor:\n");
         DUMP_BYTES ((char *)(blkdsc_p), blk_dsc_len);
      }
      PRINTF ("\nMode Page:\n");
      DUMP_BYTES ((char *)(mode_data_p), (mode_data_returned_len - header_len - blk_dsc_len));
   }
   else {
      PERROR ("mode sense subpage");
      PRINTF ("\n");
      scsi_request_sense ();
   }
   return (rc);
}

/*---------------------------------------------------------------------------*/
/* Name: rekey_command                                                       */
/*---------------------------------------------------------------------------*/
int rekey_command(void)
  {
  #define SCMD_MODE_SEL_10                0x55

  int rc;
  drive_command_t passthru;
  char *buffer;

  buffer = malloc(255);
  if (buffer == NULL) {
     errno = ENOMEM;
     printf ("Error %d (%s) Fail to allocate memory.\n",
            errno, strerror(errno));
     return errno;
   }

  /* build mode select x25/xEE in short form */
  bzero(&passthru, sizeof(drive_command_t));
  bzero(buffer,255);
  passthru.scsi_cdb[0]=SCMD_MODE_SEL_10;
  passthru.scsi_cdb[1]=0x10;
  passthru.scsi_cdb[8]=56;
  passthru.command_length=10;
  passthru.timeout_value=600;
  passthru.buffer=buffer;
  passthru.buffer_length=56;
  passthru.number_bytes=56;

  passthru.buffer[3]=0x10;
  passthru.buffer[8]=0x65;
  passthru.buffer[9]=(char)0xEE;
  passthru.buffer[11]=0x2C;
  passthru.buffer[12]=0x21;
  passthru.buffer[13]=(char)0xEE;
  passthru.buffer[15]=0x28;
  passthru.buffer[20]=0x18;
  passthru.buffer[28]=0x18;
  passthru.buffer[31]=0x04;
  passthru.buffer[39]=0x02;
  passthru.buffer[40]=0x10;
  passthru.buffer[44]=0x19;
  passthru.buffer[48]=0x10;
  passthru.buffer[52]=0x19;


  printf ("Issuing rekey command...\n");
  if (!(rc = ioctl(dev_fd, SIOC_PASSTHRU_COMMAND, &passthru)))
     printf("rekey command succeeded!\n");
  else
     printf("rekey command failed!\n");

  if (passthru.sense_length)
    dump_bytes ((char *)passthru.sense,passthru.sense_length);
  printf("\n");

  free(buffer);

  return (rc);
  }



/*---------------------------------------------------------------------------*/
/* Name: get_sernum command                                                  */
/*---------------------------------------------------------------------------*/
static int get_sernum()
{
   int rc;
   inquiry_page_t inquiry_page;
   char serial_num[20];

   inquiry_page.page_code = (uchar) 0x80;
   if (!(rc = ioctl (dev_fd, IOC_INQUIRY_PAGE, &inquiry_page))) {
      bcopy(&inquiry_page.data[4], serial_num, inquiry_page.data[3]);
      serial_num[inquiry_page.data[3]] = '\0';
      printf("%s\n", serial_num);
   }
   else {
      printf ("Get serial number failed\n");
      scsi_request_sense ();
   }

   return (rc);
}

